class A_LIST [G]
	-- association list:
	-- a table in which each key (a STRING) has an
	--  associated value of the form LINKED_LIST[G]
	-- each key keeps each of its acquired G associations in
	--  the original acquisition order

create
	make

feature {NONE} -- the actual storage space -- a pair of parallel lists

	the_keys: ARRAYED_LIST[STRING]

	the_values: ARRAYED_LIST[LINKED_LIST[G]]

feature {NONE} -- internal lookup (not affecting the iterator)

	locate_key(k: STRING): INTEGER is
		-- find index of k; 0 if not present
	do  Result := the_keys.index_of(k, 1)
	end;

	locate_or_insert(k: STRING): INTEGER is
		-- find or add (to end)
	local v: LINKED_LIST[G];
	do  Result := locate_key(k);
	    if Result = 0
	     then the_keys.extend(k);
		  create v.make;		-- and an empty list for values
		  the_values.extend(v);
		  Result := the_keys.count;
		  check the_values.count = Result end
	    end;
	ensure k.is_equal(the_keys.i_th(Result))
	end;

feature {NONE} -- creation

	make(siz: INTEGER) is
	do  create the_keys.make(siz);
	    create the_values.make(siz);
	    the_keys.compare_objects
	end;

feature -- status

	count: INTEGER is
	do  Result := the_keys.count
	end;

	capacity: INTEGER is
	do  Result := the_keys.capacity
	end;

	valid_key(key: STRING): BOOLEAN is
	do  Result := key /= Void
	end;

feature -- access

	has_any(key: STRING): BOOLEAN is
	require valid_key(key)
	do  Result := locate_key(key) > 0
	end;

	has(key: STRING): INTEGER is
		-- return the number of values associated with key
		-- (0 if key not in table at all or list is empty;
		--  empty is possible only via replace call)
	require  valid_key(key)
	local v: LINKED_LIST[G];
	do v := items(key);
	   if v /= Void
	    then  Result := v.count
	   end
	ensure Result >= 0
	end;

	items, infix "@"(key: STRING): LINKED_LIST[G] is
	require  valid_key(key)
	local i: INTEGER
	do  i := locate_key(key);
	    if i > 0 then Result := the_values.i_th(i) end;
	end;

	item(key: STRING; n: INTEGER): G is
		-- return the n-th value associated with key
		-- Void if no such value (or item value is Void)
	require  valid_key(key);  n > 0;
	local v: LINKED_LIST[G];
	do v := items(key);
	   if v /= Void and then n <= v.count
	    then  Result := v.i_th(n);
	   end
	end;

	current_keys: ARRAY[STRING] is
		-- all the current keys, in the order first presented
	local i, cnt: INTEGER
	do  from  i := 1;
		  cnt := count;
		  create Result.make(1,cnt);
	    until i > cnt
	    loop  Result.put(the_keys.i_th(i), i)
		  i := i + 1
	    end
	end;
	
feature -- alteration

	put(new: G; key: STRING) is
		-- add another value to (end of) key's value list
		-- create a new association entry if key not already present
	require  valid_key(key)
	local v: LINKED_LIST[G];
	do v := the_values.i_th(locate_or_insert(key))
	   v.extend(new);	-- add new value at end
	ensure has(key) = old has(key) + 1
	end;

	put_all(new: LIST[G]; key: STRING) is
		-- add all these values to (end of) key's value list
		-- create a new association entry if key not already present
	require  valid_key(key); new /= Void
	local v: LINKED_LIST[G];
	do v := the_values.i_th(locate_or_insert(key))
	   v.append(new);	-- add new value at end
	ensure has(key) = old has(key) + new.count
	end;

	replace(new: LINKED_LIST[G]; key: STRING) is
		-- associate key with new; old association (if any) is lost
	require  valid_key(key); new /= Void
	local  i: INTEGER
	do  i := locate_or_insert(key);
	    the_values.put_i_th(new, i);
	ensure items(key) = new
	end;

	remove(key: STRING) is
		-- remove everything associated with key
	require  has_any(key)
	local i: INTEGER
	do  i := locate_key(key);
	    the_keys.go_i_th(i);    the_keys.remove;
	    the_values.go_i_th(i);  the_values.remove;
	ensure count = (old count) - 1
	end;

	wipe_out is
		-- wipe out the list
	do  the_keys.wipe_out
	    the_values.wipe_out
	end;

feature -- iteration

	position: INTEGER

	off: BOOLEAN is
		-- true if position is off the list
	do  Result := (position < 1 or position > count)
	end;

	start is
	do  position := 1;
	end;

	forth is
	require  not off
	do  position := position + 1
	end;

	back is
	require  not off
	do  position := position - 1
	end;

	key_for_iteration: STRING is
	do  if not off
	     then Result := the_keys.i_th(position) end;
	end;

	item_for_iteration: LINKED_LIST[G] is
	do  if not off
	     then Result := the_values.i_th(position) end;
	end;

end -- A_LIST
