
indexing
	
	description: "A run closing step operation that assigns a packing %
		%to the run"
	author: "David Hollenberg"
	date: "May 5, 2006"

class ASSIGN_PACKING_OP
inherit
	RC_PRIMITIVE_STEP_OP
	RC_APPLICABLE_STEP
	HAS_USER_INTERFACE
	NO_EXTERNAL_STATE

create
	make

feature -- Properties

	input_data_items: LIST [STRING]
			-- Data items required before step can be executed
		do
			Result := array_as_list (<<
					data_item.Run_type,
					data_item.Is_educational,
					data_item.Wafer_techn,
					data_item.Field_set,
					data_item.Sched_run_no,
					data_item.Effective_other_border,
					data_item.Effective_overhead_styles
				 >>)
		end

feature {NONE} -- Implementation

	do_step
			-- Execute step
		local
			r: RUNS_ROW
		do
			cache.uncache_run
			r := cache.runs_row
			if r.is_null_packing_id then
				-- No packing assigned to run
				fetch_compatible_packing
			else
				fetch_packing
			end
			if packings_row /= Void then
				check_compatibility_with_run
				assign_packing_to_run
				call_completion_actions
			else
				get_packing_id
			end
		end

	undo_step
			-- Undo execution of step
		do
			deassign_projects_from_run
			if not deassign_failed then
				deassign_packing
				call_completion_actions
			else
				abort_actions.call ([])
			end
		end

	set_state
			-- Set `state'
		do
			state := Void
		end

feature {NONE} -- Implementation

	fetch_compatible_packing
			-- Fetch list of packings that are not assigned to
			-- a run and that are compatible with this run.
			-- If more than one, narrow list to packings whose
			-- intended run is this run.  If still more than one,
			-- signal error.  Otherwise,
			-- put single compatible packing (if any) in
			-- `packings_row'
		local
			p, mrow: like packings_row
			list, comp_list: LIST [like packings_row]
			r: STRING
		do
			create mrow.make_template
			list := mrow.unassigned_compatible_packings (cache.runs_row, connection, expected_count.At_least_zero)
			if list.is_empty then
				-- No compatible packings
				packings_row := Void
			elseif list.count >= 1 then
				-- One or more compatible packings
				
				-- FIXME: should let user pick the correct
				-- packing or create a new one, but use
				-- intended run ID to find the right one
				-- for now
				
				packings_row := Void
				r := session.run_id
				comp_list := list.twin
				comp_list.wipe_out
				from
					list.start
				until
					list.after
				loop
					p := list.item
					if p.intended_run.is_equal (r) then
						comp_list.extend (p)
					end
					list.forth
				end
				
				if comp_list.count = 1 then
					-- Found exactly one compatible
					-- packing whose intended run is this
					--  run's run ID
					packings_row := list.first
				elseif comp_list.count = 0 then
					-- Found no compatible packing
					-- whose intended run is this
					-- run's run ID
					packings_row := Void
				else	-- More than one
					handle_hard_error ("don't know how to find correct compatible packing for run.%N" + list.count.out + " unassigned packings are compatible with run but more than one (" + comp_list.count.out + ") have run ID as intended_run")
				end
			end
		end

	get_packing_id
			-- Ask user for name to use for packing ID
		local
			ifc: SINGLE_FREE_VALUE_IFC
		do
			ifc := factory.single_free_value_interface (session.title_with_run_id ("Select New Packing Name"))
			ifc.set_label ("New Packing ID")
			ifc.enable_uppercase_only
			ifc.set_minimum_width_in_chars (10)
			ifc.set_default_value (session.run_id)
			ifc.build
			
			ifc.accept_actions.extend (agent ifc.execute_action (agent call_safely_with_string (?, agent set_new_packing, agent reset (ifc))))
			ifc.decline_actions.append (abort_actions)
		end

	set_new_packing (s: STRING)
			-- Create a new packing with packing ID `s'
			-- and assign it to run
		require
			packing_id_exists: s /= Void
		local
			len, max_len: INTEGER
			row: PACKINGS_ROW
		do
			packing_id := s
			
			if packing_id.is_empty then
				handle_user_error ("Packing ID cannot be empty")
			end
			
			if not su.is_identifier (packing_id) then
				handle_user_error ("Packing ID not valid - must start with letter%Nand contain only letters, digits and underscores")
			end
			
			len := packing_id.count
			max_len := {PACKINGS_ROW}.Maxlen_packing_id
			if len > max_len then
				handle_user_error ("Packing ID length (" + len.out + ") exceeds maximum length (" + max_len.out + ")")
			end
			
			create row.make_by_key (connection, Option.Old_no_exception_row, packing_id)
			if row.is_old then
				handle_user_error ("Packing with ID " + packing_id + " already exists")
			end
			
			create_new_packing
			check_compatibility_with_run
			assign_packing_to_run
			call_completion_actions
		end

	create_new_packing
			-- Create new packing with `packing_id' and initialize
			-- it with attributes of the run
		do
			create packings_row.make_null (connection)
			packings_row.set_packing_id (packing_id)
			packings_row.initialize_from_run (cache.runs_row)
		end

	assign_packing_to_run
			-- Assign packing indicated by `packings_row' to
			-- the run
		local
			r: RUNS_MROW
			packing: RC_PACKING
		do
			r := cache.runs_mrow
			
			if packings_row.is_null_assigned_run_id then
				packings_row.set_assigned_run_id (r.run_id)
			elseif not packings_row.assigned_run_id.is_equal (r.run_id) then
				handle_hard_error ("attempt to assign packing " + su.double_quoted (packings_row.packing_id) + " to run " + session.run_id + " but packing is already assigned to run " + packings_row.assigned_run_id)
			end
			
			if r.is_null_packing_id then
				r.set_packing_id (packings_row.packing_id)
				r.set_null_packing_version_no
			elseif not r.packing_id.is_equal (packings_row.packing_id) then
				handle_hard_error ("attempt to assign packing " + su.double_quoted (packings_row.packing_id) + " to run " + session.run_id + " but that run already has an assigned packing " + su.double_quoted (r.packing_id))
			end
			
			if r.is_null_packing_version_no then
				create packing
				packing.set_packings_row (packings_row)
				packing.refresh_versions
				if packing.versions.is_empty then
					packing.create_initial_version
					packing.save_new_version ("Initial version", r)
					packing.refresh_versions
				end
				r.set_packing_version_no (packing.versions.first.version_no)
			end
		end

	fetch_packing
			-- Fetch packing assigned to run and put in
			-- `packings_row'
		local
			r: RUNS_ROW
		do
			r := cache.runs_row
			create packings_row.make_by_key (connection, Option.Old_row, r.packing_id)
		end
	
	check_compatibility_with_run
			-- Check that packing assigned to run is compatible
			-- with run
		do
			packings_row.check_compatibility_with_run (cache.runs_row)
			if not packings_row.is_compatible_with_run then
				handle_hard_error ("packing " + su.double_quoted (packings_row.packing_id) + " is not compatible with run " + session.run_id + "%N%N" + packings_row.incompatibilities_with_run)
			end
		end
	
	deassign_projects_from_run
			-- Deassign projects assigned to this run,
			-- catching abort
		do
			create operation.make (agent deassign_projects)
			operation.enable_abort_caught
			operation.execute
			if operation.aborted then
				deassign_failed := True
			end
		end
	
	deassign_projects
			-- Deassign projects assigned to this run
		local
			ok, try_again: BOOLEAN
		do
			deassigned_count := 0
			from
				ok := False
				try_again := True
			until
				ok or not try_again
			loop
				try_to_deassign_projects_from_run
				ok := failed_lock_rows.is_empty and in_rework_rows.is_empty
				if not ok then
					try_again := question (lock_failure_text)
				end
			end
			deassign_failed := not ok
			if deassigned_count > 0 then
				inform (deassigned_count.out + " " + su.number (deassigned_count, "project", "projects") + " deassigned from run")
			end
		end
	
	try_to_deassign_projects_from_run
			-- Try to deassign all projects assigned to
			-- this run Increment `deassigned_count' for
			-- each project deassigned from run.  Leave
			-- list of rows where we failed to lock fab
			-- instance in `failed_lock_rows'
		local
			row: IC_FAB_INSTANCES_MROW
			ifis: LIST [IC_FAB_INSTANCES_MROW]
			ifi: IC_FAB_INSTANCES_MROW
			changed: BOOLEAN
		do
			create row.make_template
			ifis := row.fab_instances_for_run (session.run_id, connection, expected_count.At_least_zero)
			create {ARRAYED_LIST [IC_FAB_INSTANCES_MROW]} failed_lock_rows.make (10)
			create {ARRAYED_LIST [IC_FAB_INSTANCES_MROW]} in_rework_rows.make (10)
			from
				ifis.start
			until
				ifis.after
			loop
				ifi := ifis.item
				ifi.try_to_lock_nowait
				if ifi.is_locked_and_unchanged then
					changed := True
					if not ifi.is_reworkable then
						ifi.set_null_run_id
						deassigned_count := deassigned_count + 1
					else
						in_rework_rows.extend (ifi)
					end
				else
					failed_lock_rows.extend (ifi)
				end
				ifis.forth
			end
			if changed then
				session.commit
			end
		end
	
	lock_failure_text: STRING
			-- Text of question to ask if there is a lock failure
			-- when deassigning projects
		local
			n: INTEGER
			list: like failed_lock_rows
			row: IC_FAB_INSTANCES_ROW
		do
			n := failed_lock_rows.count + in_rework_rows.count
			Result := n.out + " " + su.number (n, "project", "projects") + " could not be deassigned from run:%N%N"
			
			from
				list := failed_lock_rows
				list.start
			until
				list.after
			loop
				row := list.item
				Result.append ("   " + row.design_no.out + "   " + "Fab instance lock failure" + "%N")
				list.forth
			end
			
			from
				list := in_rework_rows
				list.start
			until
				list.after
			loop
				row := list.item
				Result.append ("   " + row.design_no.out + "   " + "Project is in rework" + "%N")
				list.forth
			end
			Result.append ("%NRetry?")
		end

	deassign_packing
			-- Deassign packing from run, if any.  Break link
			-- from run to packing and link from packing to run
		local
			r: RUNS_MROW
		do
			cache.uncache_run
			r := cache.runs_mrow
			if not r.is_null_packing_id then
				fetch_packing
				packings_row.set_null_assigned_run_id
				r.set_null_packing_id
				r.set_null_packing_version_no
				r.set_null_packer_box_full_filename
			end
		end
	
	packings_row: PACKINGS_MROW
			-- Row for packing assigned to run

	failed_lock_rows: LIST [IC_FAB_INSTANCES_MROW]
			-- Fab instances for which lock failed

	in_rework_rows: LIST [IC_FAB_INSTANCES_MROW]
			-- Fab instances which are in rework

	deassigned_count: INTEGER
			-- Number of fab instances which were assigned to
			-- run but now are not

	deassign_failed: BOOLEAN
			-- Did we fail to deassign at least one project
			-- from run when user gave up?

	operation: ANONYMOUS_OPERATION

	packing_id: STRING
			-- Candidate new packing ID entered by user

end
