
indexing
	
	author: "David Hollenberg"
	date:	"December 2, 1998"
	description: "An encoder that encodes a stream of data in %
		%base64 format"

class BASE64_ENCODER

inherit
	DATA_TRANSFORM_SYNCHRONOUS
		rename
			consume as encode,
			consume_eof as finish,
			consume_internal as encode_internal,
			consume_eof_internal as finish_internal
		redefine
			reset
		end
	C_MEMORY_ALLOCATION
		export
			{NONE} all
		end
	MEMORY
		export
			{NONE} all
		redefine
			dispose
		end
	STRING_FILLER_ACCESS
		export
			{NONE} all
		end

create
	
	make

feature {NONE} -- Creation

	make
		do
			encoder := allocate_nongc_memory (base64_encoder_length)
			strfiller := strfill
			max_chars_per_line := base64_max_line
			reset
		ensure
			binary_mode: not text_mode
		end

feature -- Properties

	in_initial_state: BOOLEAN
			-- Is `Current' in its initial state?
	
	text_mode: BOOLEAN
			-- Will `Current' be used to encode text files
			-- rather than binary files?
	
	Default_name: STRING = "Base64 encode"
			-- Default name of `Current'

feature -- Modification

	set_text_mode (b: BOOLEAN)
			-- Set `text_mode' to `b'
		do
			text_mode := b
		ensure
			mode_set: text_mode = b
		end

feature -- Operations

	reset
			-- Reset `Current' to its initial state
		do
			precursor
			initialize_base64_encoder (encoder, text_mode)
			in_initial_state := True
		end
	
feature {NONE} -- Implementation
	
	encoder: POINTER
			-- Address of C memory with encoder state

	max_chars_per_line: INTEGER
			-- Maximum number of characters in each base64-encoded
			-- line, not counting newline

	strfiller: STRING_FILLER
			-- Local copy of string filler object

	encode_internal (the_data: STRING)
			-- Encode some more data in base64
			-- format in `the_data' and put results in
			-- `output_buffer'
		local
			in_area, out_area: ANY
			outbuf: POINTER
			in_count, out_count, needed, new_cap, added: INTEGER
		do
			in_initial_state := False
			in_count := the_data.count
			out_count := output_buffer.count
			needed := bytes_needed (in_count)
			new_cap := out_count + needed
			if new_cap > output_buffer.capacity then
				output_buffer.resize (new_cap)
			end
			in_area := the_data.area
			out_area := output_buffer.area
			outbuf := offset ($out_area, out_count)
			base64_encode (encoder, $in_area, in_count, outbuf)
			added := base64_encode_bytes_added (encoder)
			strfiller.set_count (output_buffer, added + out_count)
		end

	finish_internal
			-- Finish base64 encode operation
		local
			out_area: ANY
			outbuf: POINTER
			out_count, needed, new_cap, added: INTEGER
		do
			in_initial_state := False
			out_count := output_buffer.count
			needed := bytes_needed (0)
			new_cap := out_count + needed
			if new_cap > output_buffer.capacity then
				output_buffer.resize (new_cap)
			end
			out_area := output_buffer.area
			outbuf := offset ($out_area, out_count)
			base64_encode_final (encoder, outbuf)
			added := base64_encode_bytes_added (encoder)
			strfiller.set_count (output_buffer, added + out_count)
		end

	bytes_needed (in_count: INTEGER): INTEGER
			-- Number of bytes needed to store output generated
			-- from encoding `in_count' more bytes of input
		local
			count: INTEGER
		do
			
			if text_mode then
				count := 2 * in_count
					-- Every character could be
					-- a LF, which must be converted
					-- to CR and LF before encoding
			else
				count := in_count
			end
			Result := ((count + 2) * 4.0 / 3.0).ceiling
				-- Account for up to 2 chars pending
				-- in decoder.  Every 3 chars are expanded
				-- to 4 after encoding
			Result := Result + (Result / max_chars_per_line).ceiling
				-- Account for newlines
		end

	dispose
			-- Deallocate resources when object is
			-- garbage collected
		local
			null_ptr: POINTER
		do
			if encoder /= null_ptr then
				free_nongc_memory (encoder)
				encoder := null_ptr
			end
		end

feature {NONE} -- Externals

	base64_encoder_length: INTEGER
			-- Length of C memory area required to store
			-- encoder's state
		external
			"C"
		end
	
	base64_max_line: INTEGER
			-- Maximum number of characters in base64-encoded
			-- line, not counting newline
		external
			"C"
		end
	
	initialize_base64_encoder (e: POINTER text: BOOLEAN)
			-- Initialize encoder C memory area at address `e'.
			-- Prepare to encode text data if `text' is
			-- true, binary data otherwise
		external
			"C"
		end
	
	base64_encode (e: POINTER inp: POINTER count: INTEGER outp: POINTER)
			-- Encode in base64 format some more input
			-- at address `inp' and of length `count', using
			-- encoder whose C memory is at address `e'.
			-- Put the resulting output at `outp'
		external
			"C"
		end
	
	base64_encode_final (e: POINTER outp: POINTER)
			-- Get final base64 output and put in
			-- `outp', using encoder whose C memory is at
			-- address `e'
		external
			"C"
		end
	
	base64_encode_bytes_added (e: POINTER): INTEGER
			-- Number of bytes added to output buffer by
			-- last call to `base64_encode' or
			-- `base64_encode_final', for encoder described
			-- by C memory at address `e'
		external
			"C"
		end
	
end
