
indexing
	
	description: "An operation that generates bonding diagrams for a run"
	author: "David Hollenberg"
	date: "March 12, 2003"

class BONDING_GENERATION_OP
inherit
	OPERATION
		rename
			try_operation as try_to_generate_diagrams
		end
	DB_QUERY_OPTIONS
		export
			{NONE} all
		undefine
			default_create
		end
	DB_OPTIONS_ACCESS
		export
			{NONE} all
		undefine
			default_create
		end
	STRING_UTILITIES_ACCESS
		export
			{NONE} all
		undefine
			default_create
		end
	ITEM_COUNT_RANGES_ACCESS
		export
			{NONE} all
		undefine
			default_create
		end
	RC_UTILITIES_ACCESS
		export
			{NONE} all
		end
	GUI_UTILITIES_ACCESS
		export
			{NONE} all
		undefine
			default_create
		end
	USER_INTERFACE_ACCESS
		export
			{NONE} all
		undefine
			default_create
		end

create
	default_create

feature -- Operations

	generate_diagrams (s: STRING sess: RC_RUN_CLOSING_SESSION)
			-- Generate bonding diagrams for run with run id `s',
			-- using `sess'
		require
			name_exists: s /= Void
			session_exists: sess /= Void
			session_has_connection: sess.connection /= Void
			session_connection_open: sess.connection.is_open
		do
			session := sess
			run_id := s
			connection := sess.connection
			cache := sess.cache
			execute
		end

	cleanup_after_abort
			-- Do cleanup needed after operation is aborted
		do
			if progress_window /= Void then
				progress_window.destroy
			end
		end

feature {NONE} -- Implementation

	session: RC_RUN_CLOSING_SESSION
			-- Session to use

	connection: DB_CONNECTION
			-- Database connection to use

	cache: RUN_CACHE
			-- Database cache

	run_id: STRING
			-- Run id for which to try to generate bonding diagrams

	packaging_lots: LIST [RUN_PACKAGING_LOTS_ROW]
			-- All packaging lots for run

	all_packagings: LIST [IC_PACKAGINGS_ROW]
			-- All packagings for run that need bonding

	unassigned_packagings: LIST [IC_PACKAGINGS_ROW]
			-- All packagings for run that need bonding but that
			-- are not yet associated with a packaging lot

	whole_document: DRW_DOCUMENT
			-- Whole document being built from documents
			-- for individual bonding diagrams

	progress_window: EV_DIALOG
			-- Window to show progress of bonding diagram
			-- generation

	progress_bar: EV_HORIZONTAL_PROGRESS_BAR
			-- Bar to show progress of bonding diagram
			-- generation graphically

	lot_box: EV_HORIZONTAL_BOX
			-- Box for displaying packaging lot description

	die_id_box: EV_HORIZONTAL_BOX
			-- Box for displaying die id

	package_id_box: EV_HORIZONTAL_BOX
			-- Box for displaying package id

	bonding_id_box: EV_HORIZONTAL_BOX
			-- Box for displaying bonding id

	remaining_box: EV_HORIZONTAL_BOX
			-- Box for displaying number of diagrams remaining

	generating_box: EV_HORIZONTAL_BOX
			-- Box for displaying what kind of entity we are
			-- currently generating (diagram, PostScript file,
			-- PDF file)

	message: STRING
			-- Error or warning message being constructed

	count: INTEGER
			-- Number of errors or warnings in message

	total_item_count: INTEGER
			-- Total number of items to be generated (bonding
			-- diagrams, PostScript files, PDF files, CIF files,
			-- DXF files)

	finished_item_count: INTEGER
			-- Number of items which have been generated

	remaining_item_count: INTEGER
			-- Remaining number of items to be generated
		do
			Result := total_item_count - finished_item_count
		end

	try_to_generate_diagrams
			-- Try to generate bonding diagrams for run
		do
			do_validation
			session.lock
			session.open_log
			publish_generation_started_event
			cache.uncache_packaging_lots
			packaging_lots := cache.packaging_lot_rows
			fetch_unassigned_packagings
			set_all_packagings
			if all_packagings.is_empty then
				inform ("No packagings that need a bonding diagram for " + run_id)
			else
				check_missing_bondings
				warn_unapproved_ic_packagings
				delete_existing_diagrams
				generate_new_diagrams
				publish_generation_ended_event
				inform ("Bonding diagrams generated for " + run_id)
			end
			session.terminate
		end

	do_validation
			-- Do validation before generating bonding diagrams
			-- for run (authorization checks, etc)
		local
			role_setter: ROLE_SET_OP
			runs_row: RUNS_ROW
			ifi: IC_FAB_INSTANCES_ROW
			wo: CUST_WOS_ROW
			ifi_list: LIST [IC_FAB_INSTANCES_ROW]
			wo_list: LIST [CUST_WOS_ROW]
			env: RC_ENVIRONMENT
		do
			-- Verify that user is authorized to do bonding
			-- diagram generation.  Then set back to default
			-- role since no database modification will be done

			env := session.environment
			create role_setter
			role_setter.set_role (connection, env.db_run_packager_role, env.db_access_control)
			role_setter.set_role (connection, env.db_default_role, Void)
			
			-- Check whether run already exists

			create runs_row.make_by_key (connection, Old_no_exception_row, run_id)
			if runs_row.is_empty then
				handle_user_error ("Run does not exist: " + run_id)
			end
			
			-- Check whether there are any fab instances on run yet

			create ifi.make_template
			ifi_list := ifi.fab_instances_for_run (run_id, connection, expected_count.At_least_zero)
			if ifi_list.is_empty then
				handle_user_error ("There are no projects on run " + run_id + " yet.%NCannot generate bonding diagrams.")
			end
			
			-- Check whether there are any work orders with this
			-- production id yet

			create wo.make_template
			wo_list := wo.work_orders_for_production (run_id, connection, expected_count.At_least_zero)
			if wo_list.is_empty then
				handle_user_error ("Production IDs have not been set in work orders for run " + run_id + " yet.%NCannot generate bonding diagrams.")
			end
			
			confirm ("Generate bonding diagrams for " + run_id)
			gui.process_events
		end

	fetch_unassigned_packagings
			-- Fetch IC packagings for run that need bonding,
			-- but are no associated with a packaging lot yet
		local
			row: IC_PACKAGINGS_ROW
		do
			create row.make_template
			unassigned_packagings := row.rows_needing_diagram_for_pkg_lot (session.run_id, 0, True, connection, expected_count.At_least_zero)
		end

	check_missing_bondings
			-- Issue error for any IC packagings that need bonding,
			-- are MOSIS-generated, but do not have any wires.
			-- The wirings must be generated and made to
			-- match what the old bonding algorithm would
			-- have produced (and marked as "custom" if modified)
		local
			msg: STRING
		do
			create message.make (0)
			count := 0
			packaging_lots.do_all (agent check_missing_lot_bondings)
			unassigned_packagings.do_all (agent check_missing_lot_bonding)
			if count > 0 then
				msg := count.out + " packagings need MOSIS-generated bonding but have no wires:%N%N" + message + "%N"
				if cache.is_run_complete then
					msg.append ("The diagrams for the packagings listed above will NOT include wires.")
					warn (msg)
					gui.process_events
				elseif session.environment.is_production then
					msg.append ("This condition must be rectified manually by software development.%NPlease click the Notify button below to send a message.")
					handle_hard_error (msg)
				else
					msg.append ("Please generate wirings for this run, then generate diagrams.")
					warn (msg)
					gui.process_events
				end
			end
		end

	check_missing_lot_bondings (lot: RUN_PACKAGING_LOTS_ROW)
			-- Check lot for any IC packagings that need bonding,
			-- are MOSIS-generated, but do not have any wires
		do
			lot.ic_packagings_needing_bonding.do_all (agent check_missing_lot_bonding)
		end

	check_missing_lot_bonding (p: IC_PACKAGINGS_ROW)
			-- Check whether `p' (which needs a bonding diagram)
			-- has a MOSIS-generated diagram but no wires
		do
			if not p.customer_providing_diagram and not p.have_a_wiring then
				count := count + 1
				message.append ("Die " + p.fab_instance_bonding_description + "%N")
			end
		end

	warn_unapproved_ic_packagings
			-- Warn about IC packagings that are not approved
		do
			create message.make (0)
			count := 0
			packaging_lots.do_all (agent check_packaging_approvals)
			unassigned_packagings.do_all (agent check_packaging_approval)
			if count > 0 then
				warn (count.out + " unapproved bondings:%N%N" + message)
				gui.process_events
			end
		end

	check_packaging_approvals (lot: RUN_PACKAGING_LOTS_ROW)
			-- Check approvals for all IC packagings for `lot'
			-- that need a bonding diagram
		do
			lot.ic_packagings_needing_bonding.do_all (agent check_packaging_approval)
		end

	check_packaging_approval (p: IC_PACKAGINGS_ROW)
			-- Check whether bonding diagram for `p' has
			-- been approved
		do
			if p.is_null_date_diagram_approved then
				count := count + 1
				message.append ("Die " + p.fab_instance_bonding_description + "%N")
			end
		end

	delete_existing_diagrams
			-- Delete any existing bonding diagrams
		local
			fgen: RC_FILE_GENERATOR
		do
			fgen := session.file_generator
			fgen.bond_diagrams_ps_wildcard.delete_matching_files
			fgen.bond_diagrams_pdf_wildcard.delete_matching_files
			fgen.bond_diagrams_cif_wildcard.delete_matching_files
			fgen.bond_diagrams_dxf_wildcard.delete_matching_files
			fgen.individual_diagrams_ps_wildcard.delete_matching_files
			fgen.individual_diagrams_pdf_wildcard.delete_matching_files
		end

	generate_new_diagrams
			-- Generate PostScript and PDF files for all
			-- bonding diagrams for run
		local
			by_run: BOOLEAN
		do
			if not session.environment.is_production then
				by_run := question ("Generate one PostScript file for entire run,%Ninstead of one file per packaging lot?")
				gui.process_events
			else
				by_run := False
			end
			set_item_counts (by_run)
			session.clear_abort_requested
			show_progress_window
			if by_run then
				generate_new_diagrams_by_run
			else
				generate_new_diagrams_by_lot
			end
			update_progress_window ("", Void, "", "", "")
			gui.application.sleep (300)
			progress_window.destroy
		end

	set_all_packagings
			-- Set `all_packagings'
		local
			list: LIST [IC_PACKAGINGS_ROW]
		do
			create {ARRAYED_LIST [IC_PACKAGINGS_ROW]} list.make (100)
			packaging_lots.do_all (agent append_packagings (?, list))
			list.append (unassigned_packagings)
			all_packagings := list
		end

	set_item_counts (by_run: BOOLEAN)
			-- Initialize counts of total items and those
			-- that are finished, depending on whether we
			-- are generating one PostScript/PDF file pair
			-- per run or one per lot
		local
			n, lot_count, manual_bond_count: INTEGER
			individual_count: INTEGER
			lot: RUN_PACKAGING_LOTS_ROW
			icp: IC_PACKAGINGS_ROW
		do
			n := all_packagings.count
			if by_run then
				lot_count := 1
			else
				from
					lot_count := 0
					packaging_lots.start
				until
					packaging_lots.after
				loop
					lot := packaging_lots.item
					if not lot.ic_packagings_needing_bonding.is_empty then
						lot_count := lot_count + 1
					end
					packaging_lots.forth
				end
				if not unassigned_packagings.is_empty then
					lot_count := lot_count + 1
				end
			end
			
			from
				manual_bond_count := 0
				individual_count := 0
				all_packagings.start
			until
				all_packagings.after
			loop
				icp := all_packagings.item
				if icp.needs_manual_bonding then
					manual_bond_count := manual_bond_count + 1
				end
				if icp.needs_individual_diagram then
					individual_count := individual_count + 1
				end
				all_packagings.forth
			end
			
			total_item_count := n + 2 * lot_count + 4 * manual_bond_count + 2 * individual_count
				-- One PostScript and one PDF file per lot.
				-- One PostScript and one PDF file per
				-- packaging where customer provides diagram.
				
			finished_item_count := 0
		end

	generate_new_diagrams_by_run
			-- Generate one PostScript and one PDF files for all
			-- bonding diagrams for run
		local
			list: LIST [IC_PACKAGINGS_ROW]
			sorter: LIST_BUBBLE_SORTER [IC_PACKAGINGS_ROW]
			comparator: IC_PACKAGINGS_ROW_COMPARATOR
			ps, pdf: RC_FILE
		do
			create {ARRAYED_LIST [IC_PACKAGINGS_ROW]} list.make (100)
			packaging_lots.do_all (agent append_packagings (?, list))
			list.append (unassigned_packagings)
			create sorter
			create comparator
			list := sorter.sorted (list, comparator)
			build_whole_document (list, Void)
			ps := session.file_generator.bond_diagrams_run_ps_file
			pdf := session.file_generator.bond_diagrams_run_pdf_file
			write_whole_document (Void, list.count, ps.file_pathname, pdf.file_pathname)
		end

	append_packagings (lot: RUN_PACKAGING_LOTS_ROW list: LIST [IC_PACKAGINGS_ROW])
			-- Append packagings needing bonding for `lot' to `list'
		do
			list.append (lot.ic_packagings_needing_bonding)
		end

	generate_new_diagrams_by_lot
			-- Generate PostScript and PDF files for all
			-- bonding diagrams for run, with at most one pair of
			-- files per packaging lot, plus one for packagings
			-- that need bonding but are not assigned to a lot
		do
			packaging_lots.do_all (agent generate_diagrams_for_packaging_lot)
			if not unassigned_packagings.is_empty then
				-- Generate map for bond diagrams whose
				-- packagings are not assigned
				-- to any packaging lot
				generate_diagrams_for_packaging_lot (Void)
			end
		end

	generate_diagrams_for_packaging_lot (lot: RUN_PACKAGING_LOTS_ROW)
			-- Generate PostScript and PDF files for bond diagrams
			-- for `lot' (or for unassigned bond diagrams if `lot'
			-- is Void) if it has any associated packagings that
			-- need a bonding diagram
		local
			ps, pdf: RC_FILE
			list: LIST [IC_PACKAGINGS_ROW]
		do
			if (lot /= Void and then not lot.ic_packagings_needing_bonding.is_empty) or (lot = Void and not unassigned_packagings.is_empty) then
				if lot /= Void then
					list := lot.ic_packagings_needing_bonding
				else
					list := unassigned_packagings
				end
				
				build_whole_document (list, lot)
				ps := postscript_file_for_lot (lot)
				pdf := pdf_file_for_lot (lot)
				write_whole_document (lot, list.count, ps.file_pathname, pdf.file_pathname)
			end
		end

	build_whole_document (list: LIST [IC_PACKAGINGS_ROW] lot: RUN_PACKAGING_LOTS_ROW)
			-- Build `whole_document' from `list', which is a list
			-- of packagings that need bonding for `lot'.
			-- Also, write files with just the chip for packagings
			-- that require a manual bonding
		require
			list_not_void: list /= Void
		local
			diagram: BONDING_DIAGRAM
			map_generator: BONDING_DIAGRAM_GENERATOR
			first: BOOLEAN
			icp: IC_PACKAGINGS_ROW
			pages: LIST [DRW_PAGE]
			converter: POSTSCRIPT_CONVERTER
			ps, pdf: RC_FILE
			ps_name, pdf_name: STRING
		do
			create map_generator.make (connection)
			first := True
			from
				list.start
			until
				list.after
			loop
				icp := list.item
				update_progress_window ("Diagram", lot, icp.fab_instance.die_id, icp.package_id, icp.bonding_id)
				-- FIXME: should be session.production_id
				diagram := map_generator.vendor_bonding_diagram (icp, session.run_id)
				diagram.set_drawing_rotatable (False)
				diagram.generate_document
				pages := diagram.last_document.pages
				pages.do_all (agent label_page (?, icp))
				if first then
					whole_document := diagram.last_document
					first := False
				else
					whole_document.add_pages (pages)
				end
				finished_item_count := finished_item_count + 1
				
				if icp.needs_individual_diagram then
					-- Generate PostScript and PDF files
					-- for individual diagram
					ps := postscript_file_for_packaging (icp)
					pdf := pdf_file_for_packaging (icp)
					ps_name := ps.file_pathname
					pdf_name := pdf.file_pathname
					update_progress_window ("Individual PostScript file", lot, icp.fab_instance.die_id, icp.package_id, icp.bonding_id)
					write_postscript_file (diagram.last_document, ps_name)
					finished_item_count := finished_item_count + 1
					update_progress_window ("Individual PDF file", lot, icp.fab_instance.die_id, icp.package_id, icp.bonding_id)
					create converter.make (session.environment.ghostscript_program)
					converter.convert_named_file_to_pdf (ps_name, pdf_name)
					finished_item_count := finished_item_count + 1
				end
				
				if icp.needs_manual_bonding then
					-- Generate PostScript and PDF files
					ps := postscript_file_for_chip (icp)
					pdf := pdf_file_for_chip (icp)
					ps_name := ps.file_pathname
					pdf_name := pdf.file_pathname
					diagram.set_title_omitted (True)
					diagram.set_chip_only (True)
					diagram.generate_document
					update_progress_window ("Chip PostScript file", lot, icp.fab_instance.die_id, icp.package_id, icp.bonding_id)
					write_postscript_file (diagram.last_document, ps_name)
					finished_item_count := finished_item_count + 1
					update_progress_window ("Chip PDF file", lot, icp.fab_instance.die_id, icp.package_id, icp.bonding_id)
					create converter.make (session.environment.ghostscript_program)
					converter.convert_named_file_to_pdf (ps_name, pdf_name)
					finished_item_count := finished_item_count + 1
					
					-- Generate CIF and DXF files
					update_progress_window ("Chip CIF file", lot, icp.fab_instance.die_id, icp.package_id, icp.bonding_id)
					generate_autocad_cif_file (icp)
					finished_item_count := finished_item_count + 1
					update_progress_window ("AutoCAD DXF file", lot, icp.fab_instance.die_id, icp.package_id, icp.bonding_id)
					convert_cif_to_dxf (icp)
					finished_item_count := finished_item_count + 1
				end
				
				list.forth
			end
			update_progress_window ("", lot, "", "", "")
		end

	label_page (page: DRW_PAGE icp: IC_PACKAGINGS_ROW)
			-- Label page with identifying information
		do
			page.set_page_label (icp.fab_instance.die_id + " " + icp.bonding_name)
		end

	write_whole_document (lot: RUN_PACKAGING_LOTS_ROW a_count: INTEGER ps_name, pdf_name: STRING)
			-- Write `whole_document' to PostScript file `ps_name'
			-- and PDF file `pdf_name'.  Document is for `lot'
			-- if not Void and contains `a_count' bonding diagrams
		require
			a_count_positive: a_count > 0
			ps_name_not_void: ps_name /= Void
			ps_name_not_empty: not ps_name.is_empty
			pdf_name_not_void: ps_name /= Void
			pdf_name_not_empty: not ps_name.is_empty
		local
			converter: POSTSCRIPT_CONVERTER
		do
			update_progress_window (with_count ("PostScript file", a_count), lot, "", "", "")
			write_postscript_file (whole_document, ps_name)
			finished_item_count := finished_item_count + 1
			update_progress_window (with_count ("PDF file", a_count), lot, "", "", "")
			create converter.make (session.environment.ghostscript_program)
			converter.convert_named_file_to_pdf (ps_name, pdf_name)
			finished_item_count := finished_item_count + 1
		end

	write_postscript_file (doc: DRW_DOCUMENT fname: STRING)
			-- Write PostScript representation of `doc'
			-- into `fname'
		require
			doc_not_void: doc /= Void
			fname_not_void: fname /= Void
			doc_not_void: doc /= Void
		local
			open_write_op: FILE_OPEN_WRITE_OP
			close_op: FILE_CLOSE_OP
			projector: DRW_POSTSCRIPT_PROJECTOR
			f: LARGE_PLAIN_TEXT_FILE
		do
			create open_write_op
			create close_op
			create f.make (fname)
			open_write_op.open (f)
			create projector.make_with_file (f)
			projector.project (doc)
			close_op.close (f)
		end

	generate_autocad_cif_file (icp: IC_PACKAGINGS_ROW)
			-- Generate AutoCAD CIF file for `icp'
		require
			icp_not_void: icp /= Void
			needs_manual_bonding: icp.needs_manual_bonding
		local
			fgen: RC_FILE_GENERATOR
			fname: STRING
			-- open_write_op: FILE_OPEN_WRITE_OP
			-- close_op: FILE_CLOSE_OP
			-- f: LARGE_PLAIN_TEXT_FILE
			ppr: PACKAGING_PAD_WRITER
		do
			fgen := session.file_generator
			fname := fgen.manual_bond_diagram_cif_file (icp).file_pathname
			
			-- FIXME: Restore when CIF writer can take an open file
			-- create open_write_op
			-- create close_op
			-- create f.make (fname)
			-- open_write_op.open (f)
			
			-- write CIF file for chip to `f', which is
			-- already open for write
			create  ppr.make (icp, connection)
			ppr.set_default_technology (connection)
			ppr.build_library
			-- FIXME: Restore when CIF write can take an open file
			-- ppr.output_as_cif (f)
			ppr.output_as_cif (fname)
			
			-- close_op.close (f)
		end	

	convert_cif_to_dxf (icp: IC_PACKAGINGS_ROW)
			-- Convert CIF file for `icp' to AutoCAD DXF file
		require
			icp_not_void: icp /= Void
			needs_manual_bonding: icp.needs_manual_bonding
		local
			converter: CIF_TO_DXF_CONVERTER
			env: RC_ENVIRONMENT
			fgen: RC_FILE_GENERATOR
			in_name, out_name: STRING
		do
			env := session.environment
			fgen := session.file_generator
			in_name := fgen.manual_bond_diagram_cif_file (icp).file_pathname
			out_name := fgen.manual_bond_diagram_dxf_file (icp).file_pathname
			create converter.make (env.rshell_program, env.cif_to_dxf_program, env.cif_to_dxf_host)
			converter.convert_cif_to_dxf (in_name, out_name)
		end

	postscript_file_for_lot (lot: RUN_PACKAGING_LOTS_ROW): RC_FILE
			-- PostScript file to write for `lot'
		do
			Result := session.file_generator.bond_diagrams_map_ps_file (lot, lot_short_org (lot), lot_etest_id (lot))
		ensure
			result_not_void: Result /= Void
		end

	with_count (s: STRING a_count: INTEGER): STRING
			-- `s' with text added to describe number of diagrams
		require
			s_not_void: s /= Void
		do
			Result := s + "  (" + a_count.out + " " + su.number (a_count, "diagram", "diagrams") + ")"
		ensure
			result_not_void: Result /= Void
		end

	pdf_file_for_lot (lot: RUN_PACKAGING_LOTS_ROW): RC_FILE
			-- PDF file to write for `lot'
		do
			Result := session.file_generator.bond_diagrams_map_pdf_file (lot, lot_short_org (lot), lot_etest_id (lot))
		ensure
			result_not_void: Result /= Void
		end

	lot_short_org (lot: RUN_PACKAGING_LOTS_ROW): STRING
			-- Organization short id for `lot'
		do
			if lot /= Void then
				Result := lot.organization.short_id
			else
				Result := ""
			end
		ensure
			result_exists: Result /= Void
		end

	lot_etest_id (lot: RUN_PACKAGING_LOTS_ROW): STRING
			-- Wafer etest id for `lot'
		do
			if lot /= Void then
				Result := lot.process.wafer_etest_id
			else
				Result := ""
			end
		ensure
			result_exists: Result /= Void
		end

	postscript_file_for_packaging (icp: IC_PACKAGINGS_ROW): RC_FILE
			-- PostScript file to write for `icp'
		do
			Result := session.file_generator.bond_diagram_ps_file (icp)
		ensure
			result_not_void: Result /= Void
		end

	pdf_file_for_packaging (icp: IC_PACKAGINGS_ROW): RC_FILE
			-- PDF file to write for `icp'
		do
			Result := session.file_generator.bond_diagram_pdf_file (icp)
		ensure
			result_not_void: Result /= Void
		end

	postscript_file_for_chip (icp: IC_PACKAGINGS_ROW): RC_FILE
			-- PostScript file to write for `icp'
			-- for picture of just the chip (no title)
		do
			Result := session.file_generator.manual_bond_diagram_ps_file (icp)
		ensure
			result_not_void: Result /= Void
		end

	pdf_file_for_chip (icp: IC_PACKAGINGS_ROW): RC_FILE
			-- PDF file to write for `icp'
			-- for picture of just the chip (no title)
		do
			Result := session.file_generator.manual_bond_diagram_pdf_file (icp)
		ensure
			result_not_void: Result /= Void
		end

	publish_generation_started_event
		local
			event: BONDING_GENERATION_STARTED_EVENT
		do
			create event
			event.set_run_id (run_id)
			event.publish
		end

	publish_generation_ended_event
		local
			event: BONDING_GENERATION_ENDED_EVENT
		do
			create event
			event.set_run_id (run_id)
			event.publish
		end

	show_progress_window
			-- Create and display progress window
		local
			vbox: EV_VERTICAL_BOX
			box: EV_BOX
			cols: WIDGET_COLUMN_SET
			buttons: HORIZONTAL_BUTTON_SET
		do
			create progress_window
			progress_window.set_title ("Diagram Generation Progress for " + run_id)
			progress_window.set_size (500, 250)
			create vbox
			vbox.set_padding (user_interface.padding)
			create cols.make (2)
			cols.set_columns_padding (5)
			cols.set_row_padding (user_interface.padding)
			create lot_box
			create die_id_box
			create package_id_box
			create bonding_id_box
			create remaining_box
			create generating_box
			cols.append_row (<< gui.label_with_text ("Generating:"), generating_box >>)
			cols.append_row (<< gui.label_with_text ("Packaging Lot:"), lot_box >>)
			cols.append_row (<< gui.label_with_text ("Die ID:"), die_id_box >>)
			cols.append_row (<< gui.label_with_text ("Package ID:"), package_id_box >>)
			cols.append_row (<< gui.label_with_text ("Bonding ID:"), bonding_id_box >>)
			cols.append_row (<< gui.label_with_text ("Remaining:"), remaining_box >>)
			cols.disable_column_expand (1)
			cols.align_column_left (1)
			cols.align_column_left (2)
			box := cols.box
			vbox.extend (box)
			vbox.disable_item_expand (box)
			create progress_bar
			progress_bar.value_range.resize_exactly (0, total_item_count)
			progress_bar.set_value (0)
			progress_bar.disable_segmentation
			vbox.extend (progress_bar)
			vbox.disable_item_expand (progress_bar)
			
			create buttons.make_with_text_and_actions (<< gui.ev_cancel >>, << agent process_cancel >>)
			box := buttons.column_set.box
			
			vbox.extend (box)
			vbox.disable_item_expand (box)
			
			vbox.set_border_width (user_interface.border_width)
			progress_window.extend (vbox)
			progress_window.show_relative_to_window (session.application.current_window)
			gui.process_events
		end

	update_progress_window (gen: STRING lot: RUN_PACKAGING_LOTS_ROW die, pkg, bnd: STRING)
			-- Update progress window to show current info
		local
			lot_id: STRING
		do
			if lot /= Void then
				lot_id := lot.identification
			else
				lot_id := ""
			end
			generating_box.wipe_out
			generating_box.extend (gui.label_with_text (gen))
			lot_box.wipe_out
			lot_box.extend (gui.label_with_text (lot_id))
			die_id_box.wipe_out
			die_id_box.extend (gui.label_with_text (die))
			package_id_box.wipe_out
			package_id_box.extend (gui.label_with_text (pkg))
			bonding_id_box.wipe_out
			bonding_id_box.extend (gui.label_with_text (bnd))
			remaining_box.wipe_out
			remaining_box.extend (gui.label_with_text (remaining_item_count.out))
			progress_bar.set_value (finished_item_count)
			gui.process_events
			session.process_pending_abort
		end

	process_cancel
			-- Process request to cancel diagram generation
		do
			session.set_abort_requested
		end

end
