class CATS_COMPILATION_ENVIRONMENT

inherit
	CATS_CONSTANTS

	CATS_OP_SUPPORT

	CATS_LOG_CONSTANTS

	STRING_UTILITIES_ACCESS

	FILE_UTILITIES_ACCESS

create
	make

feature	-- compilation environment input

	operation_template: CATS_OPERATION
		-- we clone this guy to start a new operation
		-- all the "environmental" items should be set here,
		--  but nothing "functional"

	leaf_input_template: CATS_OP_FILE
		-- we clone this guy to input the leaf again
		-- (then fix up the layer specification)
		--  ((which is in the input name for a flat input))
		-- this should be either a hierarchical or a flat file
		-- with an optional orientation

	second_leaf_input_template: CATS_OP_FILE
		-- if leaves are a merger of two layout files (e.g.,
		--  design and overhead), then this is the second file.
		-- fixed up as above

	techn: DESIGN_TECHNOLOGY
		-- for translating layer names into (Cats) layer numbers
		-- in hierarchical input
		-- or into mebes extensions for flat input

	final_fracture_prefix: STRING
		-- beginning of file name for all fractured levels.
		-- to this prefix, we append "<mebes_extension>.cflt"

	generated_temp_prefix: STRING
		-- beginning of file name for all generated temps.
		-- to this prefix, we append "<temp_number>.cflt"
		-- if not provided, final_fracture_prefix is used

	explicit_temp_prefix: STRING
		-- beginning of file name for all explicit temps and
		-- (when there are two inputs) all merge temps.
		-- to this prefix, we append "<layer_name>.cflt"
		-- if not provided, final_fracture_prefix is used

	additional_comments: LINKED_LIST[STRING]
		-- added comment/documentation/statistics lines to stick
		-- into the script (as shell script comments)

feature	-- compilation environment constructed items

	completed_operations: ARRAYED_LIST[CATS_SCRIPT_STEP]
		-- extend this list to file away completed operations

	last_generated_temp_no: INTEGER
		-- generated temp numbers go from 1 to this value

feature	-- status

	is_ready: BOOLEAN is
		-- is the minimal stuff in place to compile?
	do  Result := operation_template /= Void and 
			leaf_input_template /= Void and
			final_fracture_prefix /= Void and
			techn /= Void;
	end;

	known_layer(lyr: STRING): BOOLEAN is
		-- is lyr a valid layer name
	require is_ready;
	do  Result := techn.layer_by_name(lyr) /= Void
	end;

	find_op_for_layer(lyr: STRING): CATS_OPERATION is
		-- find the CATS_OPERATION that claims to be producing
		-- layer lyr
	local ndx: INTEGER; cop: CATS_OPERATION
	do  from  ndx := 1;	-- (don't use internal index)
	    until ndx > completed_operations.count or Result /= Void
	    loop  cop ?= completed_operations @ ndx;
		  if cop /= Void and then	-- just see cats_ops
		     (equal(lyr, cop.layer_level_name) and
			(cop.is_explicit or cop.is_merge))
		   then Result := cop
		  end;
		  ndx := ndx + 1
	    end
	end;

feature	-- setup stuff

	set_operation_template(cop: CATS_OPERATION) is
		-- sets the (top level) modes for format, resolution,
		-- compaction, etc.
	require cop /= Void; not (cop.modifies_primary)
	do  operation_template := cop
	end;

	set_leaf_input_templates(cof, scof: CATS_OP_FILE) is
		-- defines the input file (set) (s) for
		--  leaf expression input
	require have_file: cof /= Void;
	do  leaf_input_template := cof;
	    second_leaf_input_template := scof	-- may be Void
	end;

	set_design_technology(dt: DESIGN_TECHNOLOGY) is
		-- for the interpretation of leaf layer names
	require dt /= Void
	do  techn := dt
	end;

	set_final_fracture_prefix(prf: STRING) is
		-- file name prefix for top level fracture statements
		--  generated here.
		-- file names are <prefix><mebes_ext>.cflt
	do  final_fracture_prefix := prf
	end;

	set_generated_temp_prefix(prf: STRING) is
		-- file name prefix for (single-use) temporary files
		--  generated here.
		-- file names are <prefix><temp_no>.cflt
	do  generated_temp_prefix := prf
	end;

	set_explicit_temp_prefix(prf: STRING) is
		-- file name prefix for (multi-use) temporary files
		--  generated by ctemp statements
		-- file names are <prefix><layer_name>.cflt
	do  explicit_temp_prefix := prf
	end;

	add_a_comment(a_line: STRING) is
		-- add a line to additional_comments
	do  if additional_comments = Void then
		create additional_comments.make
	    end;
	    additional_comments.extend(a_line)
	end;

feature	-- adding of non-compiled steps

	add_a_step(cop: CATS_SCRIPT_STEP) is
		-- add a (non-CATS_OPERATION) step
	do  completed_operations.extend(cop)
	end;

	add_writefile_steps(mebes_name_prefix: STRING; use_mbs: BOOLEAN) is
		-- create a writefile step for each fractured level
		-- mebes files are named either
		--  "<mebes_name_prefix><mebes_ext>" (if use_mbs) or
		--  "<mebes_name_prefix><layer_level_name> (if not use_mbs)
	local ndx: INTEGER; cop: CATS_OPERATION;
		mbs: STRING; wf: CATS_WRITEMEBES_STEP; frst: BOOLEAN
	do  from  ndx := 1;	-- (don't use internal index)
	    until ndx > completed_operations.count
	    loop  cop ?= completed_operations @ ndx;
		  if cop /= Void and then cop.is_final
		   then -- pick output extension
			if use_mbs
			 then mbs := cop.output.twin;
			      mbs.keep_tail(7)	-- xx.cflt
			      mbs.keep_head(2)	-- xx
			 else mbs := cop.layer_level_name.twin
			end;
			mbs.to_lower
			-- writefile step
			create wf.make(cop.output,
				 su.string_of_2(mebes_name_prefix, mbs));
			wf.set_my_producer(cop);
			add_a_step(wf);
			-- tag the first step
			if not frst
			 then wf.set_a_tag(section_beg_writefile, true);
			      frst := true
			end;
		  end;
		  ndx := ndx + 1
	    end;
	    -- and tag the last one
	    if frst
	     then wf.set_a_tag(section_end_writefile, false)
	    end;
	end;

feature -- shell script output

	write_shell_script(sfil: LARGE_PLAIN_TEXT_FILE; rmv: BOOLEAN) is
		-- if rmv,
		--  then remove each temp file after its last step
	local ndx: INTEGER; op: CATS_SCRIPT_STEP
	do  -- first, fix up resolution stuff
	    do_resolution_resolution
	    -- now write out the script, step by step
	    from  ndx := 1;
	    until ndx > completed_operations.count
	    loop  op := completed_operations @ ndx
		  -- now put out the cats commands
		  sfil.putstring(shell_script_separator);
		  op.write_to_script(sfil);
		  -- exit and clean up
		  if rmv
		   then remove_exhausted_temps(sfil)
		  end;
		  ndx := ndx + 1
	    end;
	    if additional_comments /= Void then
		additional_comments.do_all(agent add_a_line(sfil, "# ", ?))
	    end;
	end;

	shell_script_separator: STRING is
	"# ====================================%N";

feature	-- workers

	input_for_layer(lyr: STRING): CATS_OP_FILE is
		-- create a CATS_OP_FILE to input "layer" lyr of
		-- either the leaf_input_template or an
		-- explicit temporary
	local xcop: CATS_OPERATION;
	do  xcop := find_op_for_layer(lyr);
	    if xcop /= Void and then xcop.is_explicit
	     then -- input the explicit temp as the "layer"
		  create Result.make_intermediate
				(xcop.output, false, xcop.output_scale_factor);
		  Result.set_my_producer(xcop)
	    elseif second_leaf_input_template /= Void
	     then Result := set_merged_inputs(lyr, xcop)
	     else Result := set_input_leaf(leaf_input_template, lyr)
	    end
	end;

	fracture_to_temp(op: CATS_OPERATION): CATS_OP_FILE is
		-- op is complete (except for output file)
		-- generate a temporary output file name, add the op
		-- to completed operations, and return a CATS_OP_FILE
		-- that will input the temporary
	local num: INTEGER; tnam: STRING
	do  -- get next temp number
	    num := last_generated_temp_no + 1;
	    last_generated_temp_no := num;
	    -- generate temp file name
	    tnam := generated_temp_name(num);
	    -- make it output file of op
	    op.set_output_generated(tnam);
	    op.set_command(cmd_is_do);
	    add_a_step(op);
	    -- and input file of whatever follows
	    create Result.make_intermediate(tnam, true, op.output_scale_factor)
	    Result.set_my_producer(op)
	end;

	new_operation(op1: CATS_OP_FILE): CATS_OPERATION is
		-- clone the operation_template and stick op1 into it.
	do  create Result.copy_from_template(operation_template);
	    Result.set_primary(op1)
	ensure not Result.modifies_primary
	end;

feature -- helpers

	final_fracture_name(mext: STRING): STRING is
		-- mext is the correct mebes extension.
	require mext.count = 2
	do  Result := su.string_of_3
			(final_fracture_prefix,
			 mext, dot_cflt)
	end;

	explicit_temp_name(lyr: STRING): STRING is
	do  Result := su.string_of_3
			(su.empty_value
			   (explicit_temp_prefix, final_fracture_prefix),
			 lyr, dot_cflt)
	end;

	generated_temp_name(tno: INTEGER): STRING is
	do  Result := su.string_of_3
			(su.empty_value
			   (generated_temp_prefix, final_fracture_prefix),
			 tno.out, dot_cflt)
	end;

	tag_completed_ops(btag, etag: STRING) is
		-- tag the first and last ops with these tags
	do  if not completed_operations.is_empty
	     then completed_operations.first.set_a_tag(btag, true);
		  completed_operations.last.set_a_tag(etag, false)
	    end
	end;

feature {NONE} -- helpers for cleanup

	remove_exhausted_temps(sfil: LARGE_PLAIN_TEXT_FILE) is
		-- remove all unremoved temps whose consumers have all
		-- been processed
	local ndx: INTEGER; op: CATS_SCRIPT_STEP
	do  from  ndx := 1;
	    until ndx > completed_operations.count
	    loop  op := completed_operations @ ndx;
		  if not op.processed
		   then -- bail since the rest are not yet processed
			ndx := completed_operations.count
		  elseif op.output_is_removable
		   then sfil.putstring("rm ");
			sfil.putstring(op.output);
			sfil.new_line;
			op.set_output_removed
		  end;
		  ndx := ndx + 1
	    end;
	end;
	
feature {NONE} -- helpers

	set_input_leaf(fil: CATS_OP_FILE; lyr: STRING): CATS_OP_FILE is
		-- input from file in fil on layer lyr
	require has_layer: known_layer(lyr)
	local dtlyr: DESIGN_TECHN_LAYER; ext, fnam: STRING
	do  dtlyr := techn.layer_by_name(lyr);
	    if fil.is_hierarchical
	     then create Result.copy_from_template(fil);
		  if fil.use_cif_layers
		   then Result.set_layers(dtlyr.cif_layer_list_for_Cats)
	           else Result.set_layers(dtlyr.gds_layer_list_for_Cats)
		  end
	     else fnam := fil.input;	-- prefix
		  ext := dtlyr.mebes_extension
		  if fil.is_mebes	-- expand to full name
		   then fnam := su.string_of_2(fnam, ext)
		   else fnam := su.string_of_3(fnam, ext, dot_cflt)
		  end;
		  create Result.make_intermediate
				(fnam, false, fil.input_scale_factor)
		  if dtlyr.have_dimensional_data
		   then Result.set_resolution
				(dtlyr.dimensions.mebes_spot_size)
		  end
	    end
	end;

	set_merged_inputs(lyr: STRING;
			  xcop: CATS_OPERATION): CATS_OP_FILE is
		-- input is the merge on lyr of the two leaves
		-- xcop, if not Void, is merge that has already been done
	require second_leaf_input_template /= Void;
	local mnam: STRING; cop: CATS_OPERATION;
	do  mnam := explicit_temp_name(lyr);
	    if xcop = Void
	     then -- create merge op for lyr
		  create cop.copy_from_template(operation_template);
		  cop.set_primary(set_input_leaf
				(leaf_input_template, lyr));
		  cop.set_function(cats_op_or,
				  set_input_leaf
				(second_leaf_input_template, lyr));
		  cop.set_output_merge(mnam, lyr);
		  cop.set_descr("Merge for layer " + lyr);
		  cop.set_command(cmd_is_do);
		  add_a_step(cop);
	     else	-- mnam already created in xcop; reuse
			-- but reduce the spot size as needed
		  cop := xcop;
	    end;
	    create Result.make_intermediate
				(mnam, false, cop.output_scale_factor);
	    Result.set_my_producer(cop)
	end;

	do_resolution_resolution is
		-- fix up resolution for temps based on need
	local ndx: INTEGER; op: CATS_OPERATION; cof: CATS_OP_FILE;
		them: LIST[CATS_FILE_CONSUMER]; res: DOUBLE
	do  from  ndx := completed_operations.count
	    until ndx < 1
	    loop  op ?= completed_operations @ ndx;
		  if op /= Void and then op.consumers /= Void
		   then -- determine my fracture resolution
			them := op.consumers;
			from  them.start;
			      res := op.resolution
			until them.after
			loop  res := common_spot
					(res, 
					 them.item.resolution_reqd)
			      them.forth
			end;
			op.set_resolution(res);
			-- and make it everyone's input resolution
			from  them.start
			until them.after
			loop  cof ?= them.item;
			      if cof /= Void
			       then cof.set_resolution(res)
			      end;
			      them.forth
			end
		  end;
		  ndx := ndx - 1
	    end
	end;

feature {NONE} -- creation

	make is
	do  create completed_operations.make(100);
	    last_generated_temp_no := 0;
	end;

end -- CATS_COMPILATION_ENVIRONMENT
