class CATS_EXPR_OPTIMIZER
		-- goes through all of the source expressions (and
		-- subexpressions) in a list of cats statements to 
		-- find all non-trivial repeated subexpressions.
		-- For each such subexpression found, a new ctemp statement
		-- is created to evaluate the subexpression, and each
		-- occurence of the subexpression is replaced by a leaf
		-- that uses that temp.

inherit
	CATS_CONSTANTS

	STRING_UTILITIES_ACCESS
		export {NONE} all end;

	UNIX_OS_ACCESS
		export {NONE} all end;

create
	perform_optimization

feature {NONE} -- creation

	perform_optimization(stmts: DYNAMIC_LIST[CATS_STATEMENT]) is
		-- does the whole banana:
		-- source expressions in stmts are modified and
		-- ctemp statements are added to stmts
	local ocnt: INTEGER; stime: DOUBLE
	do  stime := unix_os.current_time_in_fine_seconds
	    create summary.make(100);
	    from  ocnt := -1
	    until count = ocnt	-- nothing happened
	    loop  ocnt := count;
		  perform_optimization_pass(stmts)
	    end
	    summary.append(fd4.formatted(unix_os.current_time_in_fine_seconds - stime))
	ensure parity: stmts.count = count + (old stmts.count)
	end;

feature	-- statistics

	summary: STRING
		-- a printable summary of the optimization performed

	count: INTEGER
		-- the number of common subexpressions that we located
		-- and replaced with ctemp statements
		-- (also used to name the created temporaries)

feature {CATS_EXPR_HOLDER} -- registration (call-back)

	add_holder(ce: CATS_EXPRESSION; hldr: CATS_EXPR_HOLDER) is
		-- register hldr as a holder of expression ce
	require have_holder: hldr /= Void
		have_statement_no: hldr.first_statement_no > 0
	local cmplx: INTEGER; ht: like one_level; 
		foo: like one_list
	do  cmplx := ce.complexity;
	    if min_complexity <= cmplx and cmplx <= max_complexity
	     then ht := get_a_level(cmplx);
		  if ht.has(ce)
		   then foo := ht @ ce
		   else create foo.make(agent {CATS_EXPR_HOLDER}.compare_statement_no )
			ht.put(foo, ce)
		  end;
		  foo.insert(hldr)
	    end;
	end;

feature {NONE} -- the details

	min_complexity: INTEGER is 3
		-- min complexity worthy of separate evaluation step
		-- MUST BE > 1
		-- probably should always be 2 or 3

	max_complexity: INTEGER is 20
		-- max complexity we attempt to bite off in a single
		-- bite

	one_list: COMPARATOR_SORTED_LIST[CATS_EXPR_HOLDER] is
	do  
	ensure Result = Void
	end;

	one_level: HASH_TABLE[like one_list, CATS_EXPRESSION] is
		-- table of expressions and their holders
		-- all of the cats_expressions in one_level are of the same complexity
	do
	ensure Result = Void
	end;

	all_levels: ARRAY[like one_level]
		-- array of all expressions and holders, indexed by
		-- expression complexity

	get_a_level(cmplx: INTEGER): like one_level is
		-- get the cmplx-th entry from all_levels
		-- (create and enter if Void)
	require min_complexity <= cmplx and cmplx <= max_complexity
	do  Result := all_levels @ cmplx
	    if Result = Void
	     then create Result.make(37);
		  all_levels.enter(Result, cmplx)
	    end;
	ensure Result /= Void
	end;

	perform_optimization_pass(stmts: DYNAMIC_LIST[CATS_STATEMENT]) is
	local cmplx: INTEGER; ht: like one_level
	do  create all_levels.make(min_complexity, max_complexity);
	    -- register all (sub)expressions and holders in all_levels
	    from  stmts.start
	    until stmts.after
	    loop  stmts.item.register_all_exprs(Current, stmts.index)
		  stmts.forth
	    end;
	    -- find greatest complexity level with common expressions
	    from  cmplx := max_complexity
	    until cmplx < min_complexity
	    loop  ht := all_levels @ cmplx;
		  if ht /= Void and then
		     has_exprs_to_optimize(ht)
		   then optimize_these(ht, stmts);
			summary.append(su.string_of_4(count.out,"@",cmplx.out,su.semi_space))
			cmplx := 0	-- kill loop
		  end;
		  cmplx := cmplx - 1
	    end;
	end;

	has_exprs_to_optimize(ht: like one_level): BOOLEAN is
		-- are there any repeated expressions in ht?
	local foo: like one_list
	do  from  ht.start
	    until ht.after or Result
	    loop  foo := ht.item_for_iteration;
		  Result := foo.count > 1
		  ht.forth
	    end
	end;

	optimize_these(ht: like one_level;
			stmts: DYNAMIC_LIST[CATS_STATEMENT]) is
	local ce: CATS_EXPRESSION; foo: like one_list	-- list of expr holders
	do  from  ce := pick_next_item(ht)
	    until ce = Void
	    loop  foo := ht @ ce
		  optimize_this(ce, foo, stmts)
		  foo.wipe_out		-- make it unpickable again
		  ce := pick_next_item(ht)
	    end
	end;

	pick_next_item(ht: like one_level): CATS_EXPRESSION is
		-- pick next ht item to optimize
		-- return Void if ht has been exhausted
	local foo: like one_list; stmt_no1: INTEGER
	do  from  ht.start
	    until ht.after
	    loop  foo := ht.item_for_iteration	-- list of holders
		  if foo.count > 1 and then 
		     foo.first.first_statement_no > stmt_no1 then
			Result := ht.key_for_iteration;
			stmt_no1 := foo.first.first_statement_no
		  end;
		  ht.forth
	    end
	end;

	optimize_this(ce: CATS_EXPRESSION; 
			hldrs: like one_list
			stmts: DYNAMIC_LIST[CATS_STATEMENT]) is
	require have_holders: not hldrs.is_empty
	local	lnam: STRING;		  -- our new temp pseudo-layer
		ctmp: CATS_TEMP_FRACTURE; -- new temp statement to create lnam
		leaf: CATS_EXPR_LEAF;	  -- replacement leaf for all the holders
	do  count := count + 1;		-- one more statement
	    lnam := "CSE_" + count.out;	-- invent a layer name
	    create ctmp.make(ce, lnam, 0); -- and a statement
	    create leaf.make(lnam);	-- and a substitute expression
	    -- replace all occurences of ce with leaf
	    hldrs.do_all(agent {CATS_EXPR_HOLDER}.replace(ce, leaf))
	    -- now insert the new statement just before stmt_no1
	    stmts.go_i_th(hldrs.first.first_statement_no)
	    stmts.put_left(ctmp);
	end;

end -- CATS_EXPR_OPTIMIZER
