
indexing
	
	description: "Log file with output from one CATS job"
	author: "David Hollenberg"
	date: "December 9, 2003"

class CATS_LOG_FILE
inherit
	CATS_LOG_CONSTANTS

	CATS_MESSAGE_PROCESSOR

	REGULAR_EXPRESSION_ELEMENTS
		export
			{NONE} all
		end

	STRING_UTILITIES_ACCESS
		export
			{NONE} all
		end

create
	make

feature {NONE} -- Creation

	make (f: LARGE_PLAIN_TEXT_FILE)
		require
			file_not_void: f /= Void
			file_open_read: f.is_open_read
		do
			file := f
			initialize_sections
		end

feature -- Properties

	is_ok: BOOLEAN
			-- Is log file OK (does it not indicate any errors),
			-- as indicated by last call to `analyze'?
		do
			Result := errors /= Void and then errors.is_empty
		end

	errors: LIST [STRING]
			-- Errors found by last call to `analyze'

	pattern_files: LIST [PATTERN_FILES_MROW]
			-- Information about pattern files found in log
			-- by last call to `analyze'

	-- the following property is specific to is_user_readfile_present

	readfile_messages: LIST[DESIGN_CHECK_MESSAGES_MROW]

	-- the following properties are specific to is_dc_check_present

	top_structure_is_missing: BOOLEAN
			-- is the (declared) top structure AWOL?

	top_structure_name_reported: STRING
			-- when the top structure _was_ found,
			-- what was its real name? (Cats has some
			-- structure name recognition enhancements)

	root_structures_found: LIST[STRING]
			-- list of the root structures Cats found

	datalayers_found: LIST[STRING]
			-- list of the datalayers that were found

	computed_extent: CONST_RECTANGLE
			-- extent computed by Cats

	checkfile_messages: LIST[DESIGN_CHECK_MESSAGES_MROW]
			-- list of worthwhile items noted by checkfile
			--  (no N and summarized)

feature -- Operations

	analyze
			-- Analyze log file and put any errors in
			-- `errors'.  Also, set `is_ok' and `pattern_files'
		local
			name: STRING
			err_pattern: RX_PCRE_REGULAR_EXPRESSION
		do
			create {ARRAYED_LIST [STRING]} errors.make (0)
			create {ARRAYED_LIST [PATTERN_FILES_MROW]} pattern_files.make (10)
			create file_table.make (10)
			-- create the compiled RX_PCRE_REGULAR_EXPRESSIONs
			err_pattern := compiled_expression ("could|programming|internal|error|warning|output file deleted|cannot")
			input_file_pattern := compiled_expression (input_file_expr)
			area_extent_command_pattern := compiled_expression (area_extent_command_expr)
			area_pattern := compiled_expression (area_coverage_expr)
			writing_pattern := compiled_expression (writing_file_expr)
			counts_pattern := compiled_expression (figure_counts_expr)
			arrayed_counts_pattern := compiled_expression (arrayed_figure_counts_expr)
			ignored_merge_warning_pattern := compiled_expression (ignored_jobsmash_warning_expr)
			ignored_license_warning_pattern := compiled_expression (ignored_license_warning_expr)
			limits_output_pattern := compiled_expression (limits_output_expr)
			from
				file.read_line
			until
				file.end_of_file
			loop
				line := file.last_string
				line_number := line_number + 1
				if line.is_empty then
					-- nada
				elseif current_section = Void then
					if is_section_tag (line) then
						name := section_name_in_tag (line)
						sections.search (name)
						if sections.found then
							current_section := sections.found_item
						else
							append_error ("unknown fracture log section " + name)
						end
					end
				elseif is_section_tag (line) then
					name := section_name_in_tag (line)
					if not name.is_equal (current_section.end_tag) then
						append_error ("expecting end tag " + current_section.end_tag + " but found tag " + name)
					end
					if current_section.is_user_readfile_present then
						terminate_user_readfile_log
					end
					current_section := Void
				else
					if current_section.is_checked then
						err_pattern.match (line)
						if err_pattern.has_matched then
							ignored_merge_warning_pattern.match (line);
							ignored_license_warning_pattern.match (line);
							if not (ignored_merge_warning_pattern.has_matched or
								ignored_license_warning_pattern.has_matched or
								line.is_equal (ignored_warning)) then
								append_error (line)
							end
						end
					end
					if current_section.is_density_present then
						extract_density
					end
					if current_section.is_figure_count_present then
						extract_figure_counts
					end
					if current_section.is_user_readfile_present then
						extract_user_readfile_log
					end
					if current_section.is_dc_check_present then
						extract_dc_check_info
					end
				end
				file.read_line
			end
			if current_section /= Void then
				append_error ("section end tag " + current_section.end_tag + " not found before end of file")
			end
			if is_ok then
				set_pattern_files
			end
		end

feature {NONE} -- Implementation

	file: LARGE_PLAIN_TEXT_FILE
			-- File containing fracture log

	line: STRING
			-- Current line from log

	line_number: INTEGER
			-- Current log line number

	current_section: FRACTURE_LOG_SECTION
			-- Current section we're analyzing, if any

	sections: HASH_TABLE [FRACTURE_LOG_SECTION, STRING]
			-- All possible sections

	file_table: HASH_TABLE [PATTERN_FILES_MROW, STRING]
			-- Rows with information about pattern files,
			-- indexed by string

	initialize_sections
			-- Initialize `sections'
		local
			s: FRACTURE_LOG_SECTION
		do
			create sections.make (13)
			create s.make (section_beg_cats_ovrhd, section_end_cats_ovrhd)
			s.disable_checked
			sections.put (s, s.start_tag)
			create s.make (section_beg_read_acctng, section_end_read_acctng)
			s.disable_checked
			sections.put (s, s.start_tag)
			create s.make (section_beg_source_acctng, section_end_source_acctng)
			s.disable_checked
			sections.put (s, s.start_tag)
			create s.make (section_beg_cats_project, section_end_cats_project)
			s.disable_checked
			sections.put (s, s.start_tag)
			create s.make (section_beg_cats_op, section_end_cats_op)
			s.enable_checked
			sections.put (s, s.start_tag)
			create s.make (section_beg_writefile, section_end_writefile)
			s.enable_checked
			s.enable_figure_count_present
			sections.put (s, s.start_tag)
			create s.make (section_beg_is_done, section_end_is_done)
			s.disable_checked
			sections.put (s, s.start_tag)
			
			-- Sections specific to level merge log
			
			create s.make (section_beg_readfile, section_end_readfile)
			s.enable_checked
			sections.put (s, s.start_tag)
			create s.make (section_beg_merge, section_end_merge)
			s.enable_checked
			sections.put (s, s.start_tag)
			create s.make (section_beg_area_measure, section_end_area_measure)
			s.enable_checked
			s.enable_density_present
			sections.put (s, s.start_tag)

			-- section specific to design check

			create s.make (section_beg_user_readfile, section_end_user_readfile)
			s.enable_user_readfile_present
			sections.put (s, s.start_tag)

			create s.make (section_beg_dc_check, section_end_dc_check)
		--	s.enable_checked
			s.enable_dc_check_present
			sections.put (s, s.start_tag)

		end

	extract_density
			-- Extract density information from `file' if `line'
			-- matches "input" command pattern that indicates
			-- start of area measurement for particular file name
			-- and field name
		local
			file_name, field_name, cmd_name: STRING
		do
			input_file_pattern.match (line)
			area_extent_command_pattern.match(line)
			if input_file_pattern.has_matched then
				file_name := input_file_pattern.captured_substring (1)
				field_name := input_file_pattern.captured_substring (2)
				next_line
				next_line
				next_line
				if not file.end_of_file then
					really_extract_density(file_name, field_name)
				end
			elseif area_extent_command_pattern.has_matched then
				cmd_name := area_extent_command_pattern.captured_substring(1)
				file_name := area_extent_command_pattern.captured_substring(2)
				field_name := "Z"	-- irrelevant
				next_line
				if file.end_of_file then
				elseif equal(cmd_name, "AREA")
				 then really_extract_density(file_name, field_name)
				 else really_extract_extent(file_name)
				end
			end
		end

	really_extract_density(file_name, field_name: STRING)
		-- worker for above
	local
		density: DOUBLE
		row: PATTERN_FILES_MROW
	do
		area_pattern.match (line)
		if area_pattern.has_matched then
			density := area_pattern.captured_substring (1).to_double / 100.0
			add_file (file_name)
			row := file_table.item (file_name)
			row.set_field_name (field_name)
			row.set_density (density)
		else
			append_mismatch_error (line,area_coverage_expr)
		end
	end

	really_extract_extent(file_name: STRING)
	local	ext: CONST_RECTANGLE
		row: PATTERN_FILES_MROW
	do  ext := a_limits_box
	    if ext /= Void
	     then add_file (file_name)
		  row := file_table.item (file_name)
		  row.set_extent(ext)
	    end
	end

	extract_figure_counts
			-- Extract figure count information from `file' if
			-- `line' matches pattern that indicates
			-- writing of a MEBES file
		local
			file_name: STRING
			rects, traps, total: DOUBLE
			array_rects, array_traps, grand_total: DOUBLE
			found: BOOLEAN
			row: PATTERN_FILES_MROW
		do
			found := False
			writing_pattern.match (line)
			if writing_pattern.has_matched then
				file_name := writing_pattern.captured_substring (1)
				file_name.to_lower
				next_line
				next_line
				if not file.end_of_file then
					counts_pattern.match (line)
					if counts_pattern.has_matched then
						rects := counts_pattern.captured_substring (1).to_double
						traps := counts_pattern.captured_substring (2).to_double
						total := counts_pattern.captured_substring (3).to_double
						if rects + traps /= total then
							-- Have arrayed counts
							next_line
							if not file.end_of_file then
								arrayed_counts_pattern.match (line)
								if arrayed_counts_pattern.has_matched then
									found := True
									array_rects := arrayed_counts_pattern.captured_substring (1).to_double
									array_traps := arrayed_counts_pattern.captured_substring (2).to_double
									grand_total := arrayed_counts_pattern.captured_substring (3).to_double
								else
									append_mismatch_error (line, arrayed_figure_counts_expr)
								end
							end
						else
							found := True
							array_rects := 0
							array_traps := 0
							grand_total := total
						end
					else
						append_mismatch_error (line, figure_counts_expr)
					end
				end
			end
			if found then
				add_file (file_name)
				row := file_table.item (file_name)
				row.set_rectangle_count (rects)
				row.set_trapezoid_count (traps)
				row.set_array_rectangle_count (array_rects)
				row.set_array_trapezoid_count (array_traps)
				row.set_grand_total_count (grand_total)
			end
		end

	prior_line: STRING
	curr_msg_collector: like readfile_messages

	file_cats_message(sev: CHARACTER text: STRING)
	local msg: DESIGN_CHECK_MESSAGES_MROW
	do  -- file everything that's not 'N'
	    if sev /= 'N' then
		create msg.make_null(Void)
		if sev = unclassified then
			msg.set_severity('F')
			msg.set_text(su.string_of_2
				("Unknown: ", text))
		else
			msg.set_severity(sev)
			msg.set_text(text)
		end
		curr_msg_collector.extend(msg)
	    end
	end

	extract_user_readfile_log
		-- extract the entire log of the readfile/ciflib session
	require have_cats_msg_templates
		local tlin: STRING
		do
		    tlin := line.twin
		    tlin.left_adjust
		    tlin.right_adjust
		    if readfile_messages = Void then
			-- first line
			create {ARRAYED_LIST[DESIGN_CHECK_MESSAGES_MROW]}
				readfile_messages.make(20)
			curr_msg_collector := readfile_messages
		    elseif su.not_empty(prior_line) then
			-- subsequent line: process prior
			process_cats_message(prior_line, tlin)	
		    end
		    prior_line := tlin
		end

	terminate_user_readfile_log
		-- get in final call to process_cats_message
	do  if su.not_empty(prior_line) then
		process_cats_message(prior_line, Void)
	    end
	    prior_line := Void
	    add_cats_message_summaries
	end

	extract_dc_check_info
		-- extract the specific dc_check data from the log
		-- each routine gobbles its entire subsection
		do
			if equal(line, section_beg_root) then
				extract_root_structures
			elseif equal(line, section_beg_structure) then
				extract_top_structure_is_missing
			elseif equal(line, section_beg_datalayers) then
				extract_datalayers
			elseif equal(line, section_beg_extent) then
				extract_computed_extent
			elseif equal(line, section_beg_checkfile) then
				extract_checkfile_messages
			end
		end

	extract_root_structures
		local filt: RX_PCRE_REGULAR_EXPRESSION
		do	filt := compiled_expression(is_a_root_structure_expr)
			create {ARRAYED_LIST[STRING]}
				root_structures_found.make(10)
			from  next_line
			until equal(line, section_end_root) or
				equal(line, error_forced_abort)
			loop  filt.match(line)
			      if filt.has_matched
				then root_structures_found.extend(filt.captured_substring(1))
			      end
			      next_line
			end
		end

	extract_top_structure_is_missing
		local filt, rnam: RX_PCRE_REGULAR_EXPRESSION
		do	filt := compiled_expression(structure_not_found_expr)
			rnam := compiled_expression(structure_real_name_expr)
			top_structure_is_missing := false
			from  next_line
			until equal(line, section_end_structure) or
			      (top_structure_is_missing and
				equal(line, error_forced_abort))
			loop  filt.match(line)
			      if filt.has_matched then
				top_structure_is_missing := true
			      end
			      rnam.match(line)
			      if rnam.has_matched then
				top_structure_name_reported := rnam.captured_substring(1)
			      end
			      next_line
			end
		end

	extract_datalayers
		local filt: RX_PCRE_REGULAR_EXPRESSION
		do	filt := compiled_expression(datalayer_expr)
			create {ARRAYED_LIST[STRING]} datalayers_found.make(60)
			from  next_line
			until equal(line, section_end_datalayers) or
				equal(line, error_forced_abort)
			loop  filt.match(line)
			      if filt.has_matched then
				datalayers_found.extend(su.string_of_3
						(filt.captured_substring(1),
						 ";",
						 filt.captured_substring(2)))
			      end
			      next_line
			end
		end

	extract_computed_extent
		local box: CONST_RECTANGLE
		do	computed_extent := Void
			from  next_line
			until equal(line, section_end_extent) or
			      equal(line, error_forced_abort)
			loop  box := a_limits_box
			      if box /= Void then computed_extent := box end
			      next_line
			end
		end

	a_limits_box: CONST_RECTANGLE
		-- if line matches limit_output_pattern,
		-- return the rectangle implied
	local filt: RX_PCRE_REGULAR_EXPRESSION
	do  filt := limits_output_pattern
	    filt.match(line)
	    if filt.has_matched then
		create Result.make_by_corners
				(filt.captured_substring(1).to_double,
				 filt.captured_substring(2).to_double,
				 filt.captured_substring(3).to_double,
				 filt.captured_substring(4).to_double)
	    end
	end

	extract_checkfile_messages
		local tlin: STRING
		do	create {ARRAYED_LIST[DESIGN_CHECK_MESSAGES_MROW]}
				checkfile_messages.make(100)
			curr_msg_collector := checkfile_messages
			prior_line := Void
			from  next_line
			until equal(line, section_end_checkfile)
			loop  tlin := line.twin
			      if prior_line /= Void then
				process_cats_message(prior_line, tlin)
			      end
			      prior_line := tlin
			      next_line
			end
			if prior_line /= Void then
				 process_cats_message(prior_line, Void)
			end
			add_cats_message_summaries
		end

	append_error (s: STRING)
			-- Append `s' to `errors'
		require
			text_exists: s /= Void
		do
			errors.extend ("Line " + line_number.out + ": " + s)
		end

	append_mismatch_error (text, pat: STRING)
			-- Append pattern mismatch error where `text' did not
			-- match `pat' to `errors'
		require
			text_exists: text /= Void
			pattern_exists: pat /= Void
		do
			append_error ("line " + su.double_quoted (text) + " does not match pattern " + su.double_quoted (pat))
		end

	next_line
			-- Read one line from `file' and put it in `line'
			-- Append error message to `errors' if reached
			-- end of file
			-- skip empty lines
		require
			in_section: current_section /= Void
		do
			if not file.end_of_file then
				file.read_line
				line := file.last_string
				if file.end_of_file then
					append_error ("unexpected end of file in section with start tag " + current_section.start_tag)
				elseif line.is_empty then
					next_line
				else	-- normal case
					line_number := line_number + 1
					if is_section_tag (line) then
						append_error ("unexpected premature occurrence of tag line " + su.double_quoted (line))
					end
				end
			end
		end

	compiled_expression (pat: STRING): RX_PCRE_REGULAR_EXPRESSION
			-- Compiled anchored regular expression to match `pat'
			-- in case-independent manner
		require
			pattern_exists: pat /= Void
		local
			regexp_op: REGEXP_OP
		do
			create regexp_op
			create Result.make
			Result.set_anchored (True)
			Result.set_caseless (True)
			Result.set_strict (True)
			regexp_op.compile (Result, pat)
		end

	add_file (fname: STRING)
			-- Add file name `fname' to `file_table' if
			-- it is not already there
		require
			fname_exists: fname /= Void
		local
			row: PATTERN_FILES_MROW
		do
			file_table.search (fname)
			if not file_table.found then
				create row.make_null (Void)
				row.set_pattern_filename (fname)
				file_table.put (row, fname)
			end
		end

	set_pattern_files
			-- Set `pattern_files' from information in
			-- `file_table'
		do
			from
				file_table.start
			until
				file_table.after
			loop
				pattern_files.extend (file_table.item_for_iteration)
				file_table.forth
			end
		end

	ignored_warning: STRING is
		"Warning: Center justification incompatible with file limits:"
			-- Warning that is ignored (no reasonable way to
			-- prevent it from being issued)

	input_file_expr: STRING
			-- Regular expression for matching lines that
			-- mark the beginning of an area measurement
	once	Result :=
		"==>" + optional_white_space +
		"INPUT" + mandatory_white_space +
		mandatory_black_space + mandatory_white_space + "#" +
		optional_white_space + captured(mandatory_black_space) +
		mandatory_white_space + captured(mandatory_black_space) +
		optional_white_space + ending_anchor
	end

	area_extent_command_expr: STRING
			-- Regular expression for matching lines
			-- that mark an area measurement or an extent
	once	Result :=
		"==>" + optional_white_space +
		captured("AREA|EXTENT") + mandatory_white_space +
		"#" + mandatory_white_space +
		captured(mandatory_black_space)
	end

	area_coverage_expr: STRING
			-- Regular expression for matching lines that
			-- are an area measurement output
	once	Result :=
		optional_white_space + mandatory_integer +
		optional_white_space + "Figures" + mandatory_white_space +
		"Area:" + optional_white_space + mandatory_pseudo_fixed +
		optional_white_space + "Coverage:" + optional_white_space +
		captured(mandatory_integer + "\." + mandatory_integer) + "%%" +
		optional_white_space + ending_anchor
	end

	writing_file_expr: STRING
			-- Regular expression for matching lines that
			-- write MEBES files
	once	Result :=
		"Writing file with internal name " +
		captured(mandatory_black_space) + optional_white_space + ending_anchor
	end

	figure_counts_expr: STRING
			-- Regular expression for matching lines with
			-- figure counts and total
	once	Result :=
		optional_white_space +
		"Rectangles:" + optional_white_space +
		captured(mandatory_integer) + optional_white_space +
		"Trapezoids:" + optional_white_space +
		captured(mandatory_integer) + optional_white_space +
		"Total:" + optional_white_space +
		captured(mandatory_integer) + optional_white_space + ending_anchor
	end
	
	arrayed_figure_counts_expr: STRING
			-- Regular expression for matching lines with
			-- arrayed figure counts and grand total
	once	Result :=
		optional_white_space +
		"Array Rects:" + optional_white_space +
		captured(mandatory_integer) + optional_white_space +
		"Array Traps:" + optional_white_space +
		captured(mandatory_integer) + optional_white_space +
		"Grand Total:" + optional_white_space +
		captured(mandatory_integer) + optional_white_space +
		ending_anchor
	end

	ignored_jobsmash_warning_expr: STRING
			-- Regular expression for matching lines with
			-- job smash warning to be ignored
	once	Result :=
		"Warning:" + optional_white_space +
		"Chip " + mandatory_integer +
		" layers " + mandatory_integer +
		" have size \(" + mandatory_pseudo_fixed + "," +
		mandatory_pseudo_fixed + "\)" + optional_white_space +
		ending_anchor
	end

	ignored_license_warning_expr: STRING
			-- Regular expression for matching lines with
			-- license daemon warnings
	once Result := 
		"Warning:" + optional_white_space +
		"(" +	"Maximum # users for " + mandatory_black_space + " reached" +
		"|" +	"waiting for " + mandatory_black_space + " license" +
		"|" +	"user:" + mandatory_black_space + " node:" +
		")"
	end

	is_a_root_structure_expr: STRING
		-- RE for discovered root structures
	once	Result :=
		"structure" + mandatory_white_space +
		captured(mandatory_black_space) + mandatory_white_space +
		"is a root structure"
	end

	structure_not_found_expr: STRING
		-- RE for not found (we don't care about the name itself)
	once	Result :=
		"structure name" + mandatory_white_space +
		mandatory_black_space + mandatory_white_space +
		"was not found"
	end

	structure_real_name_expr: STRING
		-- RE for evaluated structure name
	once	Result :=
		structure_real_name_header + mandatory_white_space +
		captured(mandatory_black_space)
	end

	error_forced_abort: STRING is
		"Error during non-interactive processing forced abort"

	datalayer_expr: STRING
		-- RE for one discovered datalayer
	once	Result :=
		"layer" + mandatory_white_space +
		captured(mandatory_integer) + optional_white_space +
		":" + optional_white_space +
		"datatypes" + mandatory_white_space +
		captured(mandatory_black_space)
	end

	limits_output_expr: STRING
		-- RE for the reply to extent query
	once	Result :=
		"limits" + optional_white_space + "\(" +
		captured(mandatory_signed_fixed) + "," +
		captured(mandatory_signed_fixed) + "\)" +
		optional_white_space + "\(" +
                captured(mandatory_signed_fixed) + "," +
                captured(mandatory_signed_fixed) + "\)"
	end

	input_file_pattern: RX_PCRE_REGULAR_EXPRESSION
	
	area_extent_command_pattern: RX_PCRE_REGULAR_EXPRESSION

	area_pattern: RX_PCRE_REGULAR_EXPRESSION
	
	writing_pattern: RX_PCRE_REGULAR_EXPRESSION
	
	counts_pattern: RX_PCRE_REGULAR_EXPRESSION
	
	arrayed_counts_pattern: RX_PCRE_REGULAR_EXPRESSION
	
	ignored_merge_warning_pattern: RX_PCRE_REGULAR_EXPRESSION

	ignored_license_warning_pattern: RX_PCRE_REGULAR_EXPRESSION

	limits_output_pattern: RX_PCRE_REGULAR_EXPRESSION
	
end
