deferred class CATS_MESSAGE_PROCESSOR
	-- process messages produced by Cats (and readfile/ciflib/writefile)
	-- against the contents of CATS_MESSAGE_TEMPLATES to discover
	-- their severity

inherit

	SEVERITY_LEVELS

	STRING_UTILITIES_ACCESS

feature	-- setup: get the message templates

	have_cats_msg_templates: BOOLEAN is
	do  Result := cats_msg_templates /= Void
	end;

	get_cats_msg_templates(db: DB_CONNECTION) is
	local r: CATS_MESSAGE_TEMPLATES_ROW; 
		l: LIST[CATS_MESSAGE_TEMPLATES_ROW]
	do  create r.make_template;
	    l := r.all_rows(db);
	    -- now sort by weight into cats_msg_templates
	    create cats_msg_templates.make
		(agent {CATS_MESSAGE_TEMPLATES_ROW}.compare_by_weight);
	    cats_msg_templates.set_sort_decreasing;
	    cats_msg_templates.insert_list(l)
	    create {ARRAYED_LIST[STRING]}unknown_CIF_layers.make(5)
	ensure have_cats_msg_templates
	end;

	unknown_CIF_layers: LIST[STRING]

feature -- processing: batch

	process_cats_messages(msgs: LIST[STRING]) is
		-- classify a list of returned Cats messages
		-- per the directions in cats_msg_templates
		-- and file them away via file_cats_message.
		-- summaries are also generated for templates that
		-- have gone over their max_occurrences
	require have_cats_msg_templates
	local lin1, lin2: STRING
	do  from  if msgs.is_empty
		   then lin1 := Void
		   else msgs.start; lin1 := msgs.first
		  end
	    until lin1 = Void
	    loop  -- get lin2 for the 2-liners
		  msgs.forth; 
		  if msgs.after
		   then lin2 := Void
		   else lin2 := msgs.item;
		  end;
		  process_cats_message(lin1, lin2)
		  -- move on
		  lin1 := lin2;		-- for the next round
	    end;
	end;

feature	-- processing: on the fly
	-- if there are n lines to match, call process_cats_message n times:
	--	process_cats_message(l(1), l(2))
	--	process_cats_message(l(2), l(3))
	--	  ...
	--	process_cats_message(l(n-1), l(n))
	--	process_cats_message(l(n), Void)

	process_cats_message(lin1, lin2: STRING) is
		-- this is one message (which may be 1 or 2 lines long)
		-- find its matching template and publish
	require have_cats_msg_templates
	local cmt: CATS_MESSAGE_TEMPLATES_ROW; mtxt: STRING
	do  if skip_second_line
	     then -- lin1 already consumed as a lin2
		  skip_second_line := false
	     else cmt := locate_template(lin1, lin2);
		  if cmt /= Void
		   then skip_second_line := cmt.line_count = 2;
			if cmt.is_publishable and cmt.severity /= 'N'
			 then -- publish
			      if skip_second_line
			       then mtxt := su.string_of_3(lin1,su.space,lin2)
			       else mtxt := lin1
			      end;
			      file_cats_message(cmt.severity, mtxt)
			 else -- nothing (it's already counted)
			end;
			if cmt.the_unknown_cif_layer /= Void
			 then unknown_CIF_layers.extend
					(cmt.the_unknown_cif_layer)
			end;
		   else -- unknown message: fold
			file_cats_message(unclassified, lin1)
		  end
	    end
	end;

feature	-- processing: adding summaries

	add_cats_message_summaries is
		-- add summary lines for all templates with unpublished
		-- messages
	local cmt: CATS_MESSAGE_TEMPLATES_ROW;
	do  from  cats_msg_templates.start
	    until cats_msg_templates.after
	    loop  cmt := cats_msg_templates.item
		  if cmt.do_publish_summary
		   then file_cats_message(cmt.severity, cmt.published_summary)
		  end;
		  cmt.reset_match_count;
		  cats_msg_templates.forth
	    end;
	end;

feature	-- disposal of classified messages

	unclassified: CHARACTER is '?';

	file_cats_message(sev: CHARACTER; text: STRING) is
		-- routine to put away the classified messages
		-- a pseudo-severity of unclassified is reserved for
		--  messages that were not classified
	require sev = unclassified or is_severity_char(sev);
	deferred
	end;

feature {NONE} -- details

	cats_msg_templates: COMPARATOR_SORTED_LIST[CATS_MESSAGE_TEMPLATES_ROW]

	skip_second_line: BOOLEAN

	locate_template(lin1, lin2: STRING): CATS_MESSAGE_TEMPLATES_ROW is
		-- returns Void if no match
	local cmt: CATS_MESSAGE_TEMPLATES_ROW
	do  from  cats_msg_templates.start
	    until cats_msg_templates.after or Result /= Void
	    loop  cmt := cats_msg_templates.item
		  if (cmt.line_count = 1 or lin2 /= Void)
			and then cmt.match(lin1, lin2)
		   then Result := cmt	-- take first hit
		  end
		  cats_msg_templates.forth
	    end
	end;

end -- CATS_MESSAGE_PROCESSOR
