class CATS_MESSAGE_TEMPLATES_ROW
		-- NOTE: we "know" here that all of the templates are
		-- either one-liners or two-liners, and that this
		-- attribute is not conditional
		-- (i.e., <newline> is never in an alternative clause)

inherit
	CATS_MESSAGE_TEMPLATES_RAW

	REGULAR_EXPRESSION_ELEMENTS

	STRING_UTILITIES_ACCESS
		export {NONE} all end

create
	make_template, make_null, make_by_key

feature	-- status

	is_prepared: BOOLEAN is
		-- (preparation is automatic)
	do  Result := (first_regexp /= Void)
	end;

	line_count: INTEGER is
		-- number of log lines that this template covers
	do  if not is_prepared then prepare end;
	    if second_regexp = Void
	     then Result := 1
	     else result := 2
	    end
	end;

	weight: INTEGER is
		-- an ordering metric for applying templates
	do  if cached_weight = 0 then set_weight end;
	    Result := cached_weight
	end;

	compare_by_weight(other: like Current): INTEGER is
		-- comparator for comparator_sorted_list
	require other /= Void
	do  Result := weight.three_way_comparison(other.weight)
	end;

feature	-- preparation for use

	prepare is
		-- (called automatically when needed, so external
		--  calls are not really necessary)
	local rexp: STRING; rexps: LIST[STRING]
	do  rexp := cats_message_template.twin
	    -- pick off the CIF unknown layer message
	    is_cif_unknown_layer_template := 
			equal(rexp, cmt_unk_cif_template);
	    -- quote RE sensitive characters
	    rexp.replace_substring_all("\", "\\");	-- (do first!)
	    rexp.replace_substring_all(".", "\.");
	    rexp.replace_substring_all("(", "\(");
	    rexp.replace_substring_all(")", "\)");
	    rexp.replace_substring_all("^", "\^");
	    -- change { and } to ( and )
	    rexp.replace_substring_all("{", "(");
	    rexp.replace_substring_all("}", ")");
	    -- replace <...> constructs with RE equivalents
	    rexp.replace_substring_all(cmt_integer, 
					mc(mandatory_signed_integer));
	    rexp.replace_substring_all(cmt_word, mc(mandatory_black_space));
	    rexp.replace_substring_all(cmt_real, mc(mandatory_signed_real));
	    rexp.replace_substring_all(cmt_string, ".*");
	    -- break at <newline>
	    rexp.replace_substring_all(cmt_newline, su.eol);
	    rexps := su.broken_at(rexp, '%N');
	    check rexps.count <= 2 end;
	    -- and build the regexp(s)
	    first_regexp := prepare_one_regexp(rexps.first);
	    if rexps.count > 1 
	     then second_regexp := prepare_one_regexp(rexps @ 2)
	    end;
	ensure is_prepared
	end;

feature	-- matching

	match(this_line, next_line: STRING): BOOLEAN is
		-- next_line is used only when line_count = 2
	do  if not is_prepared then prepare end;
	    first_regexp.match(this_line);
	    Result := first_regexp.has_matched;
	    if is_cif_unknown_layer_template and Result
	     then the_unknown_CIF_layer :=
				first_regexp.captured_substring(2);
	    end;
	    if Result and (second_regexp /= Void)
	     then second_regexp.match(next_line);
		  Result := second_regexp.has_matched
	    end;
	    if Result
	     then match_count := match_count + 1;
	    end
	end;

	the_unknown_CIF_layer: STRING
	    -- set ONLY after a match to the ciflib unknown message
	    --  template

feature	-- counting and publishing

	match_count: INTEGER
		-- number of messages that this template has matched so far

	is_publishable: BOOLEAN is
		-- should we publish this (match_count-th) message?
		-- answer: just the first max_occurrences worth
	do  Result := is_null_max_occurrences or else
			(match_count <= max_occurrences)
	end;

	do_publish_summary: BOOLEAN is
		-- should we publish this summary message?
		-- answer: yes if anyone was not publishable
	do  Result := not is_publishable
	end;

	published_summary: STRING is
		-- 
	do  Result := su.string_of_3
			(match_count.out, su.space, summary_text)
	end;

	reset_match_count is
	do  match_count := 0
	ensure match_count = 0
	end;

feature {NONE} -- implementation

	is_cif_unknown_layer_template: BOOLEAN

	mc(rexp: STRING): STRING is
	do  if is_cif_unknown_layer_template
	     then Result := captured(rexp)
	     else Result := rexp
	    end
	end;

	cached_weight: INTEGER

	set_weight is
	local i, brkt: INTEGER; ch: CHARACTER;
	do  cached_weight := 1;
	    if line_count = 2 then cached_weight := 101 end;
	    from  i := 1; brkt := 0
	    until i > cats_message_template.count
	    loop  ch := cats_message_template @ i;
		  if ch = '{' or ch = '<'
		   then brkt := brkt + 1
		  elseif ch = '}' or ch = '>'
		   then brkt := brkt - 1
		  elseif brkt = 0
		   then cached_weight := cached_weight + 1
		  end;
		  i := i + 1
	    end;
	    check brkt = 0 end;
	end;

	first_regexp: RX_PCRE_REGULAR_EXPRESSION
		-- always have at least one

	second_regexp: RX_PCRE_REGULAR_EXPRESSION
		-- have a second one iff this is a 2-line template

	prepare_one_regexp(rex: STRING): RX_PCRE_REGULAR_EXPRESSION is
	do  create Result.make
	    Result.set_caseless(true);
	    Result.set_strict(true);
	    Result.set_anchored(true);	-- anchors at the beginning
	    rex.extend('$');		-- anchors at the end
	    Result.compile(rex)
	end;

	cmt_integer:	STRING is "<integer>"
	cmt_word:	STRING is "<word>"
	cmt_real:	STRING is "<real>"
	cmt_string:	STRING is "<string>"
	cmt_newline:	STRING is "<newline>"

	cmt_unk_cif_template: STRING is
	"Line <integer>: layer is undefined - Layer <word> is being added as layer <integer>"

end -- CATS_MESSAGE_TEMPLATES_ROW
