class CATS_OP_FILE
	-- a holder for all of the attributes associated with an input
	-- file (primary or secondary) of a CATS_OP

inherit
	CATS_FILE_CONSUMER

	CATS_OP_SUPPORT
	rename origin as GS_origin	-- (we don't need it here)
	end;

create
	make_hierarchical,
	make_flat_set,
	make_intermediate,
	copy_from_template

feature	-- the attributes

	input:		STRING
		-- the file can be hierarchical (gds or clib)
		-- or flat (mebes or cflt)

	structure:	STRING
	layers:		STRING
	limits:		CONST_RECTANGLE
		-- creator must provide these for a hierarchical input

	origin:		CONST_POINT
	sizing:		DOUBLE
	scale:		DOUBLE

	orientation:	MATRIX
		-- considered part of the "template" rather than a 
		-- modification of the input

	selections:	LIST[STRING]

	resolution:	DOUBLE
		-- the "input" resolution (for a flat file), used solely
		-- to avoid the "input resolution differs" warning

feature	-- status

	is_generated_temp: BOOLEAN
		-- is this a generated single-use cflt temporary?

	is_hierarchical: BOOLEAN is
	do  Result := structure /= Void
	end;

	is_flat: BOOLEAN is
	do  Result := structure = Void
	end;

	is_mebes: BOOLEAN is
	do  Result := is_flat and then (input @ input.count = '.')
	end;

	is_cflt: BOOLEAN is
	do  Result := is_flat and then (input @ input.count = '_')
	end;

	is_modified: BOOLEAN is
		-- do any of the other parameters modify the input?
		-- (orientation is not a modification in our sense)
	do  Result := (sizing /= 0) or (scale /= 1) or
			-- (not orientation.is_identity) or
			(not selections.is_empty)
	end;

	is_valid_flat_name(s: STRING): BOOLEAN is
		-- a flat file set name must end in either '_' for cflt
		-- or '.' for mebes, so that we can append the correct
		-- suffix
	require s /= Void
	local cnt: INTEGER;
	do  cnt := s.count;
	    Result := cnt >= 2 and then flat_terminators.has(s @ cnt)
	end;

	flat_terminators: STRING is "_."

	use_cif_layers: BOOLEAN
		-- we are using cif layers (meaning that the 
		-- layer/datatype values of which Cats is cognizant
		-- use the layer_order rather than the mapped gds pairs

feature	-- modification for each of the above

	set_layers(var: STRING) is
	do  layers := var
	end;

	set_limits(var: CONST_RECTANGLE) is
	do  limits := var
	end;

	set_origin(var: CONST_POINT) is
	do  origin := var
	end;

	set_sizing(var: DOUBLE) is
	do  sizing := var
	end;

	set_scale(var: DOUBLE) is
	require var > 0
	do  scale := var
	end;

	set_orientation(var: MATRIX) is
	require is_valid_cats_orientation(var)
	do  orientation := var
	end;

	set_selections(var: LIST[STRING]) is
	require var /= Void
	do  selections := var
	end;

	add_selection(var: STRING) is
	do  selections.extend(var)
	end;

	set_cif_layers is
	require is_hierarchical
	do  use_cif_layers := true
	end;

	set_gds_layers is
	require is_hierarchical
	do  use_cif_layers := false
	end;

	set_resolution(var: DOUBLE) is
	require is_flat; var > 0
	do  resolution := var
	end;

feature	-- scale factor of this cats_op_file

	input_scale_factor: DOUBLE
		-- the scale factor (relative to 1X drawn) of the input file
		-- set by creation

	output_scale_factor: DOUBLE
		-- the scale factor of the output data
	do  Result := scale * input_scale_factor
	end;

feature	-- cats_file_consumer features

	my_step: CATS_OPERATION
		-- link back to my CATS_OPERATION

	resolution_reqd: DOUBLE is
		-- the resolution that I need from this file
		-- (0 = unknown / indefinite)
	do  if my_step /= Void
	     then -- output resolution unscaled
		  Result := my_step.resolution / scale
	    end
	end;

	processed: BOOLEAN is
		-- has my consumption of the file been processed yet?
	do  Result := my_step /= Void and then my_step.processed
	end;

	set_my_step(stp: CATS_OPERATION) is
	do   my_step := stp
	end;

feature	-- output 

	add_cats_commands(csh: LARGE_PLAIN_TEXT_FILE) is
		-- each string is a Cats command;
		-- when executed in order, a Cats input is set up
	local
	do  -- input
	    if resolution > 0
	     then add_a_line_dbl4(csh, "resolution", resolution)
	    end;
	    add_a_line(csh, "input", input);
	    -- structure, layers, limits
	    maybe_add_a_line(csh, "structure", structure);
	    maybe_add_a_line(csh, "layers", layers);
	    if limits /= Void
	     then add_a_line(csh, "limits", limits.Cats_out(fd4))
	    elseif is_hierarchical
	     then add_a_line(csh, "extent", "all")
	    end;
	    -- selects
	    from  selections.start
	    until selections.after
	    loop  add_a_line(csh, "select primary",
			     su.string_of_3(selections.index.out, su.space,
					    selections.item))
		  selections.forth
	    end;
	    -- origin, orientation, scale, sizing
	    if origin /= Void
	     then add_a_line(csh, "origin", origin.formatted(fd4))
	    end;
	    if not orientation.is_identity
	     then add_a_line(csh, "orient",
				Cats_orientation_name(orientation))
	    end;
	    if scale /= 1
	     then add_a_line_dbl4(csh, "scale", scale)
	    end;
	    if sizing /= 0
	     then add_a_line_dbl4(csh, "sizing", sizing)
	    end;
	end;

feature {NONE} -- creation

	make_the_rest is
	do  -- everything else unmodified
	    scale := 1.0
	    orientation := identity
	    create {LINKED_LIST[STRING]} selections.make
	    use_cif_layers := false
	end;

	make_flat_set(fnam: STRING; s_factor: DOUBLE) is
		-- fnam is either a mebes or a cflt prefix
		-- (used to make a template for compilation)
		-- s_factor is the scale factor of the flat set
	require flat_name: fnam /= Void and then is_valid_flat_name(fnam)
		s_factor: s_factor > 0
	do  input := fnam;
	    input_scale_factor := s_factor
	    make_the_rest
	end;

	make_hierarchical(fnam, stru: STRING;
				lims: CONST_RECTANGLE) is
		-- fnam is the full file name (gds or clib)
		-- (used to make a template for compilation)
		-- if lims = Void, the compiler issues "extent all"
	require fnam /= Void and stru /= Void
	do  input := fnam;
	    structure := stru;
	    limits := lims;
	    input_scale_factor := 1.0
	    make_the_rest
	end;

	make_intermediate(fnam: STRING; gen_temp: BOOLEAN; s_factor: DOUBLE) is
		-- fnam is the full file name (of a cflt)
		-- (used in direct construction)
	do  input := fnam;
	    is_generated_temp := gen_temp;
	    input_scale_factor := s_factor
	    make_the_rest
	end;

	copy_from_template(tmplt: CATS_OP_FILE) is
		-- clone the template
	require tmplt /= Void
	do  copy(tmplt);	-- good enough for all but:
	    selections := tmplt.selections.twin
	end;

end -- CATS_OP_FILE
