
indexing
	
	description: "A run closing step operation that checks  %
		%projects on a run for unresolved issues (both policy %
		%violations, such as projects that don't have a funding %
		%allocation or don't have a signed customer agreement, %
		%and technical issues, such as unbondable pads or mismatch %
		%between what customer asked for and what we're ordering)"
	author: "David Hollenberg"
	date: "February 23, 2004"

class CHECK_UNRESOLVED_ISSUES_OP
inherit
	RC_PRIMITIVE_STEP_OP
	RC_APPLICABLE_STEP
	NO_USER_INTERFACE
	NO_EXTERNAL_STATE

create
	make

feature -- Properties

	input_data_items: LIST [STRING]
			-- Data items required before step can be executed
		do
			Result := array_as_list (<<
					data_item.Assigned_wafer_processes,
					data_item.Assigned_die_ids,
					data_item.Cut_patterns,
					data_item.Initial_wafer_lots,
					data_item.Test_vehicle_orders,
					data_item.Production_job
				 >>)
		end

feature {NONE} -- Implementation

	do_step
			-- Execute of step
		do
			-- Check work orders and paying accounts

			check_work_orders
			check_delinquent_accounts
			check_cust_agreements
			check_export_control

			-- Check project to ensure customer won't get
			-- a nasty surprise (wrong number of parts, wrong
			-- package, etc)
			
			check_no_bond_metal
			check_potential_scmos_layer_errors
			check_no_die_id
			check_unapproved_bond_diagrams
			check_missing_requested_packages
			check_quantity_differences
			check_package_quantity_differences
			check_cut_die_quantities
			
			call_completion_actions
		end

	undo_step
			-- Undo execution of step
		do
			-- Nothing to undo
			call_completion_actions
		end

	set_state
			-- Set `state'
		do
			state := Void
		end

	check_work_orders
			-- Check work orders for all projects on run
			-- and warn about any that do not have a valid
			-- funding allocation
		do
			check_project_accounts (
				"  and m_funding_alloc_no < 0 %N",
				"",
				"without a valid funding allocation",
				"projects without valid funding allocations")
		end

	check_delinquent_accounts
			-- Check work orders for all projects on run
			-- and warn about any whose paying account is
			-- delinquent
		do
			check_project_accounts (
				"  and ca.date_acct_delinquent is not null %N",
				"",
				"whose paying account is delinquent",
				"delinquent accounts")
		end

	check_cust_agreements
			-- Check that all commercial projects on run
			-- have a completed customer agreement and warn
			-- about any that do not
		do
			check_project_accounts (
			   "  and ca.acct_type = 'COM' %N%
			   %  and not exists ( %N%
			   %     select 1 from cust_acct_legal_agrs cala, %N%
			   %                   legal_agrs la %N%
			   %     where cala.legal_agr_no = la.legal_agr_no %N%
			   %       and cala.cust_acct_no = ca.cust_acct_no %N%
			   %       and la.legal_agr_type in ('CST', 'GOV') %N%
			   %       and la.legal_agr_complete = 'Y')%N",
			   "commercial",
			   "without a completed customer agreement",
			   "missing customer agreements")
		end

	check_project_accounts (where, type, problem, subject: STRING)
			-- Execute query with partial where clause `where',
			-- to retrieve design number and account string
			-- for accounts restricted to `type' (no restriction
			-- if `type' is empty).
			-- If query returns any rows, issue warning
			-- about `problem', using `subject' for e-mail
			-- message
		require
			where_exists: where /= Void
			where_nonempty: not where.is_empty
			type_exists: type /= Void
			problem_exists: problem /= Void
			problem_nonempty: not problem.is_empty
			subject_exists: subject /= Void
			subject_nonempty: not subject.is_empty
		local
			q: SQL_QUERY
			n: INTEGER
			text, design_no, acct_string, msg, proj_type: STRING
		do
			text :=
				"select design_no, m_lookup.cust_acct_display_name (cw.cust_acct_no) %N%
				%from ic_fab_instances ifi, cust_wos cw, cust_accts ca %N%
				%where ifi.m_fab_no = cw.m_fab_no %N%
				%  and cw.cust_acct_no = ca.cust_acct_no %N"
				+ where +
				"  and ifi.run_id = :1 %N%
				%  and cw.production_id = :2 %N%
				%order by design_no"
			q := connection.sql_query (text)
			q.set_numeric_bind_capacity (2)
			q.bind_num_to_string (1, session.run_id, False)
			q.bind_num_to_string (2, session.run_id, False)
			q.set_select_count (2)
			q.select_integer (1)
			q.select_string (2, 255)
			q.execute
			create msg.make (0)
			from
				q.fetch
			until
				q.end_of_data
			loop
				n := n + 1
				design_no := q.integer_value (1).out
				acct_string := q.string_value (2)
				su.left_pad (design_no, 9)
				if n = 1 then
					msg.append ("Design_no  Account%N")
					msg.append ("---------  -------%N%N")
				end
				msg.append (design_no + Spacer + acct_string + "%N")
				q.fetch
			end
			q.release
			if n > 0 then
				if not type.is_empty then
					proj_type := " " + type + " "
				else
					proj_type := " "
				end
				msg := n.out + proj_type + su.number (n, "project", "projects") + " on run " + session.run_id + " " + problem + "%N%N" + msg
				publish_account_warning (subject, msg)
			end
		end

	check_export_control
			-- Check orders for customer projects
			-- that require an ECCN.  Report on projects
			-- for which we have not received an export
			-- control questionnaire or for which the
			-- ECCN has not been approved
		local
			q: SQL_QUERY
			fab_instance_type: IC_FAB_INSTANCE_TYPE_VALUES
			export_control_type: EXPORT_CONTROL_TYPE_VALUES
			n: INTEGER
			closer, die_id,design_no, acct_string, eccn_status, msg: STRING
		do
			q := connection.sql_query (
				"select %N%
				%	r.closer, %N%
				%	ifi.run_id, %N%
				%	ifi.die_id, %N%
				%	ifi.design_no, %N%
				%	m_lookup.cust_acct_display_name(cid.cust_acct_no) account, %N%
				%	decode(cid.export_control_type, :3, 'ITAR - NOT RCVD', 'NOT RECEIVED') eccn_status %N%
				%from runs r, wafer_techns wt, ic_fab_instances ifi, cust_ic_designs cid, cust_wos cw %N%
				%where cid.design_no = ifi.design_no %N%
				%  and ifi.run_id = r.run_id %N%
				%  and cw.m_fab_no = ifi.m_fab_no %N%
				%  and r.wafer_techn = wt.wafer_techn %N%
				%  and cw.production_id = :1 %N%
				%  and ifi.run_id = :1 %N%
				%  and ifi.ic_fab_instance_type = :2 %N%
				%  and (wt.is_eccn_required = 'Y' or %N%
				%       cust_wo_utilities.is_export_order (cw.m_order_no) = 'Y') %N%
				%  and not exists (select 1 from design_ecc_questionnaires where design_no = ifi.design_no) %N%
				%  and cw.has_eccn = 'N' %N%
				%union %N%
				%select %N%
				%	r.closer, %N%
				%	ifi.run_id, %N%
				%	ifi.die_id, %N%
				%	ifi.design_no, %N%
				%	m_lookup.cust_acct_display_name(cid.cust_acct_no) account, %N%
				%	decode(cid.export_control_type, :3, 'ITAR - NOT APRVD', 'NEEDS APPROVAL') eccn_status %N%
				%from runs r, wafer_techns wt, ic_fab_instances ifi, cust_ic_designs cid, cust_wos cw %N%
				%where cid.design_no = ifi.design_no %N%
				%  and ifi.run_id = r.run_id %N%
				%  and cw.m_fab_no = ifi.m_fab_no %N%
				%  and r.wafer_techn = wt.wafer_techn %N%
				%  and cw.production_id = :1 %N%
				%  and ifi.run_id = :1 %N%
				%  and ifi.ic_fab_instance_type = :2 %N%
				%  and (wt.is_eccn_required = 'Y' or %N%
				%       cust_wo_utilities.is_export_order (cw.m_order_no) = 'Y') %N%
				%  and exists (select 1 from design_ecc_questionnaires where design_no = ifi.design_no) %N%
				%  and cw.has_eccn = 'N' %N%
				%order by run_id, die_id, eccn_status")
			q.set_numeric_bind_capacity (3)
			q.bind_num_to_string (1, session.run_id, False)
			q.bind_num_to_string (2, fab_instance_type.Customer_fab, False)
			q.bind_num_to_string (3, export_control_type.Itar, False)
			q.set_select_count (6)
			q.select_string (1, {RUNS_ROW}.Maxlen_closer)
			q.select_string (2, {RUNS_ROW}.Maxlen_run_id)
			q.select_string (3, 20)
			q.select_integer (4)
			q.select_string (5, 255)
			q.select_string (6, 50)
			q.execute
			create msg.make (0)
			from
				q.fetch
			until
				q.end_of_data
			loop
				n := n + 1
				closer := q.string_value (1)
				die_id := q.string_value (3)
				design_no := q.integer_value (4).out
				acct_string := q.string_value (5)
				eccn_status := q.string_value (6)
				acct_string.keep_head (28)
				su.right_pad (closer, 8)
				su.right_pad (die_id, 3)
				su.right_pad (acct_string, 28)
				su.left_pad (design_no, 9)
				if n = 1 then
					msg.append ("Closer    Run   Die  Design No  Account                         ECCN Status%N")
					msg.append ("--------  ----  ---  ---------  ----------------------------  ----------------%N%N")
				end
				msg.append (closer + Spacer + session.run_id + Spacer + die_id + Spacer + design_no + Spacer + acct_string + Spacer + eccn_status + "%N")
				q.fetch
			end
			q.release
			if n > 0 then
				msg := n.out + " " + su.number (n, "project", "projects") + " on run " + session.run_id + " with missing or unapproved ECCNs%N%N" + msg
				publish_project_warning ("projects with ECCN problems", msg)
			end
		end

	check_no_bond_metal
			-- Check for designs with pads that have no
			-- bond metal underneath and warn about them
		local
			q: SQL_QUERY
			n: INTEGER
			design_no, acct_string, pad_count, msg: STRING
		do
			q := connection.sql_query (
				"select cid.design_no, m_lookup.cust_acct_display_name (cid.cust_acct_no), no_bond_metal_pad_count %N%
				%from ic_fab_instances ifi, cust_ic_designs cid %N%
				%where ifi.design_no = cid.design_no %N%
				%  and no_bond_metal_pad_count > 0 %N%
				%  and ifi.run_id = :1 %N%
				%order by design_no")
			q.set_numeric_bind_capacity (1)
			q.bind_num_to_string (1, session.run_id, False)
			q.set_select_count (3)
			q.select_integer (1)
			q.select_string (2, 255)
			q.select_integer (3)
			q.execute
			create msg.make (0)
			from
				q.fetch
			until
				q.end_of_data
			loop
				n := n + 1
				design_no := q.integer_value (1).out
				acct_string := q.string_value (2)
				pad_count := q.integer_value (3).out
				su.left_pad (design_no, 9)
				su.left_pad (pad_count, 9)
				if n = 1 then
					msg.append ("Design_no  Pad Count  Account%N")
					msg.append ("---------  ---------  -------%N%N")
				end
				msg.append (design_no + Spacer + pad_count + Spacer + acct_string + "%N")
				q.fetch
			end
			q.release
			if n > 0 then
				msg := n.out + " " + su.number (n, "project", "projects") + " on run " + session.run_id + " with pads without bond metal%N%N" + msg
				publish_project_warning ("projects missing bond metal", msg)
			end
		end

	check_potential_scmos_layer_errors
			-- Check for GDS SCMOS designs with an unknown
			-- layer number that matches a layer number in
			-- another SCMOS design technology (possible wrong
			-- technology code)
		local
			q: SQL_QUERY
			n: INTEGER
			design_no, design_techn_id, layer_num_string, layer_name, msg: STRING
		do
			q := connection.sql_query ("[
select	
	cid.design_no, cida.final_design_techn,
	replace(dnl.layer_name, '?') gds_layer_number, scmos_gds.layer_name
  from	design_no#final_layers dnl,
	ic_fab_instances ifi,
	cust_ic_design_attributes cida,
	cust_ic_designs cid,
	(select distinct gds_layer, gds_datatype, layer_name
	  from gds#design_techn_id_layers
	  where design_techn_id like 'SC%'
	    and layer_name not in ('MEMS_OPEN', 
				   'MEMS_ETCH_STOP',
				   'N_HIGH_VOLTAGE', 
				   'P_HIGH_VOLTAGE')) scmos_gds
  where dnl.design_no = cid.design_no
   and  ifi.design_no = cid.design_no
   and	cida.design_no = cid.design_no
   and	cida.final_design_techn like 'SC%'
   and	cida.final_geometry_format = 'GDS'
   and  cid.design_cancelled = 'N'
   and	dnl.layer_name like '?%'
   and	to_char(scmos_gds.gds_layer) =
		substr(dnl.layer_name,2,instr(dnl.layer_name,';')-2)
   and	(scmos_gds.gds_datatype is null or
	  to_char(scmos_gds.gds_datatype) =
		substr(dnl.layer_name,instr(dnl.layer_name,';')+1))
   and  ifi.run_id = :1
 order by
	cid.design_no, dnl.layer_name
				]")
			q.set_numeric_bind_capacity (1)
			q.bind_num_to_string (1, session.run_id, False)
			q.set_select_count (4)
			q.select_integer (1)
			q.select_string (2, 255)
			q.select_string (3, 255)
			q.select_string (4, 255)
			q.execute
			create msg.make (0)
			from
				q.fetch
			until
				q.end_of_data
			loop
				n := n + 1
				design_no := q.integer_value (1).out
				design_techn_id := q.string_value (2)
				layer_num_string := q.string_value (3)
				layer_name := q.string_value (4)
				su.left_pad (design_no, 9)
				su.right_pad (design_techn_id, 20)
				su.right_pad (layer_num_string, 20)
				su.right_pad (layer_name, 20)
				if n = 1 then
					msg.append ("Design_no  Design Techn          GDS Layer             Layer Name%N")
					msg.append ("---------  --------------------  ---------             ----------%N%N")
				end
				msg.append (design_no + Spacer + design_techn_id + Spacer + layer_num_string + Spacer + layer_name + "%N")
				q.fetch
			end
			q.release
			if n > 0 then
				msg := n.out + " SCMOS " + su.number (n, "project", "projects") + " on run " + session.run_id + " with potential layer problems.%N%NThe following projects have an unknown layer number that matches a layer number%Nin another SCMOS design technology.  They may have been submitted with%Nthe wrong technology code.%N%N" + msg
				publish_project_warning ("SCMOS projects with potential layer problems", msg)
			end
		end

	check_no_die_id
			-- Check for fab instances whose effective
			-- overhead style is NONE (no die ID).
			-- Skip check completely if run overhead style is NONE
			-- (user got warned earlier when they selected this
			-- style)
		local
			q: SQL_QUERY
			n: INTEGER
			design_no, acct_string, design_name, msg: STRING
			style: OVERHEAD_STYLE_VALUES
		do
			q := connection.sql_query (
				"select cid.design_no, substr(cid.design_name, 1, 30), m_lookup.cust_acct_display_name (cid.cust_acct_no) %N%
				%from ic_fab_instances ifi, cust_ic_designs cid, runs r %N%
				%where ifi.design_no = cid.design_no %N%
				%  and ifi.run_id = r.run_id %N%
				%  and ifi.run_id = :1 %N%
				%  and nvl (r.overhead_style, :2 || 'X') <> :2 %N%
				%  and nvl(r.overhead_style, ifi.overhead_style) = :2 %N%
				%order by design_no")
			q.set_numeric_bind_capacity (2)
			q.bind_num_to_string (1, session.run_id, False)
			q.bind_num_to_string (2, style.None, False)
			q.set_select_count (3)
			q.select_integer (1)
			q.select_string (2, 255)
			q.select_string (3, 255)
			q.execute
			create msg.make (0)
			from
				q.fetch
			until
				q.end_of_data
			loop
				n := n + 1
				design_no := q.integer_value (1).out
				design_name := q.string_value (2)
				acct_string := q.string_value (3)
				su.left_pad (design_no, 9)
				su.right_pad (design_name, 30)
				if n = 1 then
					msg.append ("Design_no  Design Name                     Account%N")
					msg.append ("---------  ------------------------------  -------%N%N")
				end
				msg.append (design_no + Spacer + design_name + Spacer + acct_string + "%N")
				q.fetch
			end
			q.release
			if n > 0 then
				msg := n.out + " " + su.number (n, "project", "projects") + " on run " + session.run_id + " with no die ID (overhead style " + style.None +")%N%N" + msg
				publish_project_warning ("projects with no die ID", msg)
			end
		end

	check_unapproved_bond_diagrams
			-- Check for customer designs with packagings whose
			-- bonding diagram has not been approved and
			-- warn about them
		local
			q: SQL_QUERY
			n: INTEGER
			design_no, acct_string, msg: STRING
		do
			q := connection.sql_query (
				"select distinct cid.design_no, m_lookup.cust_acct_display_name (cid.cust_acct_no) %N%
				%from ic_fab_instances ifi, cust_ic_designs cid, ic_packaging_orders ipo, cust_wos cw %N%
				%where ifi.design_no = cid.design_no %N%
				%  and ifi.m_fab_no = ipo.m_fab_no %N%
				%  and ipo.m_order_no = cw.m_order_no %N%
				%  and ic_fab_instance_type = 'CUSTOMER_FAB' %N%
				%  and nvl(ipo.packaged, 'N') = 'Y' %N%
				%  and ipo.quantity_ordered > 0 %N%
				%  and (diagram_approved_by is null or date_diagram_approved is null) %N%
				%  and ifi.run_id = :1 %N%
				%  and cw.production_id = :1 %N%
				%order by design_no")
			q.set_numeric_bind_capacity (1)
			q.bind_num_to_string (1, session.run_id, False)
			q.set_select_count (2)
			q.select_integer (1)
			q.select_string (2, 255)
			q.execute
			create msg.make (0)
			from
				q.fetch
			until
				q.end_of_data
			loop
				n := n + 1
				design_no := q.integer_value (1).out
				acct_string := q.string_value (2)
				su.left_pad (design_no, 9)
				if n = 1 then
					msg.append ("Design_no  Account%N")
					msg.append ("---------  -------%N%N")
				end
				msg.append (design_no + Spacer + acct_string + "%N")
				q.fetch
			end
			q.release
			if n > 0 then
				msg := n.out + " " + su.number (n, "project", "projects") + " on run " + session.run_id + " with unapproved bonding diagrams%N%N" + msg
				publish_project_warning ("unapproved bonding diagrams", msg)
			end
		end

	check_missing_requested_packages
			-- Check for designs for which a package was
			-- requested but the package is not being ordered and
			-- warn about them
		local
			q: SQL_QUERY
			n: INTEGER
			design_no, pkg_req, quant, acct_string, msg: STRING
		do
			q := connection.sql_query (
				"select ifi.design_no, icp.package_requested, icp.quantity_packaged_requested, m_lookup.cust_acct_display_name (cid.cust_acct_no) %N%
				%from ic_fab_instances ifi, cust_ic_designs cid, ic_packagings icp %N%
				%where ifi.design_no = cid.design_no %N%
				%  and ifi.m_fab_no = icp.m_fab_no %N%
				%  and nvl(icp.package_requested, 'NONE') <> 'NONE' %N%
				%  and nvl(icp.quantity_packaged_requested, 1) > 0 %N%
				%  and icp.package_id is null %N%
				%  and ifi.run_id = :1 %N%
				%order by ifi.design_no")
			q.set_numeric_bind_capacity (1)
			q.bind_num_to_string (1, session.run_id, False)
			q.set_select_count (4)
			q.select_integer (1)
			q.select_string (2, 255)
			q.select_integer (3)
			q.select_string (4, 255)
			q.execute
			create msg.make (0)
			from
				q.fetch
			until
				q.end_of_data
			loop
				n := n + 1
				design_no := q.integer_value (1).out
				pkg_req := q.string_value (2)
				if q.is_null (3) then
					quant := ""
				else
					quant := q.integer_value (3).out
				end
				acct_string := q.string_value (4)
				su.left_pad (design_no, 9)
				su.left_pad (quant, 8)
				su.right_pad (pkg_req, 20)
				if n = 1 then
					msg.append ("Design_no  Requested Package     Quantity  Account%N")
					msg.append ("---------  --------------------  --------  -------%N%N")
				end
				msg.append (design_no + Spacer + pkg_req + Spacer + quant + Spacer + acct_string + "%N")
				q.fetch
			end
			q.release
			if n > 0 then
				msg := n.out + " requested " + su.number (n, "package", "packages") + " on run " + session.run_id + " not ordered (reason unknown)%N%N" + msg
				publish_project_warning ("requested packages not ordered", msg)
			end
		end

	check_quantity_differences
			-- Check for designs on run whose ordered IC
			-- quantity differs from either requested IC
			-- quantity or total requested packaging
			-- quantity and warn about them
		local
			q: SQL_QUERY
			n: INTEGER
			design_no, quant_req, quant_ordered: STRING
			quant_packaged_req, acct_string, msg: STRING
		do
			q := connection.sql_query (quantity_differences_query)
			q.set_numeric_bind_capacity (1)
			q.bind_num_to_string (1, session.run_id, False)
			q.set_select_count (5)
			q.select_integer (1)
			q.select_integer (2)
			q.select_integer (3)
			q.select_integer (4)
			q.select_string (5, 255)
			q.execute
			create msg.make (0)
			from
				q.fetch
			until
				q.end_of_data
			loop
				n := n + 1
				design_no := q.integer_value (1).out
				if q.is_null (2) then
					quant_req := ""
				else
					quant_req := q.integer_value (2).out
				end
				if q.is_null (3) then
					quant_packaged_req := ""
				else
					quant_packaged_req := q.integer_value (3).out
				end
				if q.is_null (4) then
					quant_ordered := ""
				else
					quant_ordered := q.integer_value (4).out
				end
				acct_string := q.string_value (5)
				su.left_pad (design_no, 9)
				su.left_pad (quant_req, 5)
				su.left_pad (quant_packaged_req, 9)
				su.left_pad (quant_ordered, 9)
				if n = 1 then
					msg.append ("Design_no  # Req  # Pkg Req  # Ordered  Account%N")
					msg.append ("---------  -----  ---------  ---------  -------%N%N")
				end
				msg.append (design_no + Spacer + quant_req + Spacer + quant_packaged_req + Spacer + quant_ordered + Spacer + acct_string + "%N")
				q.fetch
			end
			q.release
			if n > 0 then
				msg := n.out + " " + su.number (n, "project", "projects") + " on run " + session.run_id + " with possible quantity problems%N%N" + msg
				publish_project_warning ("possible quantity problems", msg)
			end
		end

	check_package_quantity_differences
			-- Check for designs on run with packagings
			-- whose quantity requested does not match the
			-- quantity ordered and warn about them
		local
			q: SQL_QUERY
			n: INTEGER
			design_no, package, quant_ordered: STRING
			quant_packaged_req, acct_string, msg: STRING
		do
			q := connection.sql_query (quantity_packaged_differences_query)
			q.set_numeric_bind_capacity (1)
			q.bind_num_to_string (1, session.run_id, False)
			q.set_select_count (5)
			q.select_integer (1)
			q.select_string (2, 255)
			q.select_integer (3)
			q.select_integer (4)
			q.select_string (5, 255)
			q.execute
			create msg.make (0)
			from
				q.fetch
			until
				q.end_of_data
			loop
				n := n + 1
				design_no := q.integer_value (1).out
				package := q.string_value (2)
				if q.is_null (3) then
					quant_packaged_req := ""
				else
					quant_packaged_req := q.integer_value (3).out
				end
				if q.is_null (4) then
					quant_ordered := ""
				else
					quant_ordered := q.integer_value (4).out
				end
				acct_string := q.string_value (5)
				su.left_pad (design_no, 9)
				su.left_pad (quant_packaged_req, 5)
				su.left_pad (quant_ordered, 5)
				su.right_pad (package, 20)
				if n = 1 then
					msg.append ("Design_no  Package               # Req  # Ord  Account%N")
					msg.append ("---------  --------------------  -----  -----  -------%N%N")
				end
				msg.append (design_no + Spacer + package + Spacer + quant_packaged_req + Spacer + quant_ordered + Spacer + acct_string + "%N")
				q.fetch
			end
			q.release
			if n > 0 then
				msg := n.out + " project " + su.number (n, "packaging", "packagings") + " on run " + session.run_id + " whose requested and ordered quantities differ%N%N" + msg
				publish_project_warning ("possible packaging quantity problems", msg)
			end
		end

	check_cut_die_quantities
			-- Check for designs on run whose quantity of cut
			-- dies is not sufficient to satisfy sum of
			-- quantity ordered and quantity excess needed
		local
			q: SQL_QUERY
			n: INTEGER
			msg, die_id, fab_no, design_no: STRING
			quant_ordered, quant_excess, quant_cut, total_needed, shortage: INTEGER
			quant_ordered_str, quant_excess_str, quant_cut_str, total_needed_str, shortage_str: STRING
		do
			q := connection.sql_query (quantity_cut_dies_query)
			q.set_numeric_bind_capacity (1)
			q.bind_num_to_string (1, session.run_id, False)
			q.set_select_count (6)
			q.select_integer (1)
			q.select_integer (2)
			q.select_string (3, 255)
			q.select_integer (4)
			q.select_integer (5)
			q.select_integer (6)
			q.execute
			create msg.make (0)
			from
				q.fetch
			until
				q.end_of_data
			loop
				n := n + 1
				fab_no := q.integer_value (1).out
				design_no := q.integer_value (2).out
				die_id := q.string_value (3)
				quant_ordered := q.integer_value (4)
				quant_excess := q.integer_value (5)
				quant_cut := q.integer_value (6)
				total_needed := quant_ordered + quant_excess
				shortage := total_needed - quant_cut
				quant_ordered_str := quant_ordered.out
				quant_excess_str := quant_excess.out
				quant_cut_str := quant_cut.out
				total_needed_str := total_needed.out
				shortage_str := shortage.out
				su.left_pad (fab_no, 6)
				su.left_pad (design_no, 6)
				su.left_pad (quant_ordered_str, 7)
				su.left_pad (quant_excess_str, 6)
				su.left_pad (quant_cut_str, 9)
				su.left_pad (total_needed_str, 12)
				su.left_pad (shortage_str, 8)
				su.right_pad (die_id, 6)
				if n = 1 then
					msg.append ("Die ID  Design  Fab_no  Ordered  Excess  Total Needed  Available  Shortage%N")
					msg.append ("------  ------  ------  -------  ------  ------------  ---------  --------%N%N")
				end
				msg.append (die_id + Spacer + design_no + Spacer + fab_no + Spacer + quant_ordered_str + Spacer + quant_excess_str + Spacer + total_needed_str + Spacer + quant_cut_str + Spacer + shortage_str + "%N")
				q.fetch
			end
			q.release
			if n > 0 then
				msg := n.out + " " + su.number (n, "project", "projects") + " on run " + session.run_id + " may not have enough parts%N%N" + msg
				publish_project_warning ("possible cut die quantity problems", msg)
			end
		end

	quantity_differences_query: STRING =
			-- Query to check whether quantity differs
			-- from either requested IC quantity or total
			-- requested packaging
		"select ifi.design_no, ifi.quantity_requested, %N%
		%      decode(count(icp.ic_packaging_no), sum(decode(icp.package_requested, 'NONE', 1, decode(icp.quantity_packaged_requested,null, 0, 1))), ifi.quantity_unpackaged_requested + sum(nvl(icp.quantity_packaged_requested,0)), null), %N%
		%      ifq.quantity_ordered, %N%
		%      m_lookup.cust_acct_display_name (cid.cust_acct_no) %N%
		%from (select cwo.m_fab_no, sum(cwli.part_quantity_ordered*mp.quantity_in_part) quantity_ordered from m_parts mp, cust_wo_line_items cwli, cust_wos cwo where mp.m_part_no = cwli.m_part_no and mp.part_type = 'IC' and cwli.m_order_no = cwo.m_order_no group by cwo.m_fab_no) ifq, %N%
		%     ic_fab_instances ifi, ic_packagings icp, cust_ic_designs cid %N%
		%where ifq.m_fab_no = ifi.m_fab_no %N%
		%  and ifi.design_no = cid.design_no %N%
		%  and ifi.m_fab_no = icp.m_fab_no %N%
		%  and ifi.run_id = :1 %N%
		%  and (ifi.quantity_requested is not null or %N%
		%       (ifi.quantity_unpackaged_requested is not null and %N%
		%       not exists (select 1 from ic_packagings where m_fab_no = ifi.m_fab_no and package_requested <> 'NONE' and quantity_packaged_requested is null))) %N%
		%group by ifi.design_no, cid.cust_acct_no, ifi.quantity_requested, ifi.quantity_unpackaged_requested, ifq.quantity_ordered %N%
		%having ifi.quantity_requested <> ifq.quantity_ordered or %N%
		%       (count(icp.ic_packaging_no) = count(all icp.quantity_packaged_requested) and %N%
		%       ifi.quantity_unpackaged_requested + sum(icp.quantity_packaged_requested) <> ifq.quantity_ordered) or %N%
		%       (sum(decode(icp.package_requested, 'NONE', 1, 0)) = 1 and %N%
		%       ifi.quantity_unpackaged_requested <> ifq.quantity_ordered) %N%
		%order by ifi.design_no"

	quantity_packaged_differences_query: STRING =
			-- Query to check for project packagings whose
			-- requested and ordered quantities differ
		"select ifi.design_no, icp.package_id, %N%
		%      icp.quantity_packaged_requested, %N%
		%      sum(decode(mp.packaged, 'Y', cwli.part_quantity_ordered*mp.quantity_in_part, 0)), %N%
		%      m_lookup.cust_acct_display_name (cid.cust_acct_no) %N%
		%from m_parts mp, cust_wo_line_items cwli, ic_packagings icp, ic_fab_instances ifi, cust_ic_designs cid %N%
		%where ifi.design_no = cid.design_no %N%
		%  and mp.m_part_no (+) = cwli.m_part_no %N%
		%  and cwli.ic_packaging_no (+) = icp.ic_packaging_no %N%
		%  and icp.m_fab_no = ifi.m_fab_no %N%
		%  and icp.package_id is not null %N%
		%  and nvl(icp.quantity_packaged_requested, 0) > 0 %N%
		%  and ifi.run_id = :1 %N%
		%group by ifi.design_no, ifi.m_fab_no, icp.package_id, icp.quantity_packaged_requested, cid.cust_acct_no %N%
		%having icp.quantity_packaged_requested <> sum(decode(mp.packaged, 'Y', cwli.part_quantity_ordered*mp.quantity_in_part,0)) %N%
		%order by ifi.design_no"

	quantity_cut_dies_query: STRING =
			-- Query to check for projects whose cut die quantity
			-- does not satisfy sum of quantity ordered and
			-- quantity excess needed
			"[
			select ifi.m_fab_no, ifi.design_no, ifi.die_id,
				mfoq.quantity_ordered, 
				mfeq.quantity_excess_needed,
				dcq.quantity_cut_die
			from
				ic_fab_instances ifi, 
				m_fab_no#order_quantities mfoq, 
				m_fab_no#excess_quantities mfeq, 
				die#cut_quantities dcq
			where	ifi.m_fab_no = mfoq.m_fab_no 
			  and 	ifi.m_fab_no = mfeq.m_fab_no 
			  and 	ifi.run_id = dcq.run_id 
			  and 	ifi.die_id = dcq.die_id 
			  and 	dcq.run_id = :1 
			  and	dcq.quantity_cut_die < mfoq.quantity_ordered + mfeq.quantity_excess_needed
			order by ifi.die_id
			]"

	publish_account_warning (subj, msg: STRING)
			-- Publish message `msg' warning about unresolved issue
			-- for an account, with `subj' in case e-mail is sent
		require
			subject_exists: subj /= Void
			message_exists: msg /= Void
		local
			e: UNRESOLVED_ACCT_ISSUE_EVENT
		do
			create e
			e.set_subject (subj)
			e.set_message (msg)
			e.publish
			if preferences.is_unattended_mode_disabled then
				warn_user (msg)
			end
		end

	publish_project_warning (subj, msg: STRING)
			-- Publish message `msg' warning about unresolved issue
			-- for a project, with `subj' in case e-mail is sent
		require
			subject_exists: subj /= Void
			message_exists: msg /= Void
		local
			e: UNRESOLVED_PROJ_ISSUE_EVENT
		do
			create e
			e.set_subject (subj)
			e.set_message (msg)
			e.publish
			if preferences.is_unattended_mode_disabled then
				warn_user (msg)
			end
		end

	warn_user (msg: STRING)
		require
			message_exists: msg /= Void
		local
			cond: EXPLICIT_CONDITION
		do
			create cond
			cond.set_warning
			cond.set_text (msg)
			cond.disable_retry
			cond.disable_notify
			cond.enable_column_alignment
			cond.handle
		end

	Spacer: STRING = "  "
			-- White space between columns

end
