class CIF_FILE_READER

inherit

	LAYOUT_IO_SUPPORT

	GEOMETRY_SUPPORT
	export {NONE} all
	end;

create
	make

feature {NONE} -- cif reader input management

	my_file: DATA_SOURCE
		-- 

	last_character: CHARACTER;
	last_string: STRING
		-- current char & line from my_file

	line_number: INTEGER;
		-- line number (of line currently in last_string)

	next_char_pos: INTEGER
		-- position (in last_string) of first unexamined character

	get_next_line is
	do  if my_file.after
	     then log_io_error(error_is_eof, Void);
		  last_string := ");E"
	     else my_file.read_line;
		  last_string := my_file.last_string
		  line_number := line_number + 1;
	    end;
	    next_char_pos := 1
	end;

	get_next_character is
		-- next character of the cif file now in last_character
	do  if next_char_pos <= last_string.count
	     then last_character := last_string @ next_char_pos;
		  next_char_pos := next_char_pos + 1
	     else last_character := '%N';
		  get_next_line;
	    end
	end;

feature {NONE} -- some parser pieces

	skip_to_semi is
		-- skip to next semicolon
	do  from
	    until last_character = ';'
	    loop  get_next_character
	    end
	ensure last_character = ';'
	end;

	must_be_at_semi is
		-- statement is parsed; we should be at {blank} semi
	do  if not is_cif_nonblank(last_character)
	     then get_next_nonblank
	    end;
	    if last_character /= ';'
	     then log_io_error(error_is_nosemi, Void);
		  skip_to_semi
	    end;
	end;

feature {NONE} -- cif character characterizers
	-- (in addition to is_upper and is_digit)

	is_cif_nonblank(c: CHARACTER): BOOLEAN is
	do  Result := c.is_digit or c.is_upper or c = '-' or
			c = '(' or c = ')' or c = ';' or
			(lioopt.allow_named_calls and c = '%'')
	end;

	is_cif_sep(c: CHARACTER): BOOLEAN is
		-- separator between (before) number
	do  Result := c.is_upper or not is_cif_nonblank(c)
	end;

	is_cif_c(c: CHARACTER): BOOLEAN is
		-- charcter legal in shortname
	do  Result := c.is_upper or c.is_digit
	end;

	get_next_nonblank is
		-- non-blank now in last_character
	do  from  get_next_character
	    until is_cif_nonblank(last_character)
	    loop  get_next_character
	    end
	end;

feature {NONE} -- error message generator

	error_is_deldef: STRING is "DELDEF";
	error_is_dir000: STRING is "DIR000";
	error_is_dfnods: STRING is "DFNODS";
	error_is_dsinds: STRING is "DSINDS";
	error_is_dsnodf: STRING is "DSNODF";
	error_is_halfsc: STRING is "HALFSC";
	error_is_nosemi: STRING is "NOSEMI";
	error_is_nosyno: STRING is "NOSYNO";

	file_position: STRING is
	do  Result := su.string_of_2("line ", line_number.out)
	end;

	stru_position: STRING is
	do  if curr_structure = Void
	     then Result := "not in symbol"
	     else Result := su.string_of_2("in symbol ",
						curr_structure.name)
	    end
	end;

feature {NONE} -- cif reader globals for collecting layout

	top_level_body: LAYER_SORTED_COLLECTION;
		-- collector for read layout

	curr_scale,
	top_level_scale: DOUBLE
		-- multiply incoming integers by this to get doubles
		-- includes both library unit and DS scale factor

	make_top_current is
		-- not in a structure; collect top_level
	do  curr_structure := Void;
	    curr_body := top_level_body;
	    curr_scale := top_level_scale
	end;

	name_table: HASH_TABLE[STRING, STRING]
		-- key is cif name (an integer); val is symbolic name

	get_top_level_name: STRING is
		-- return a unique name for the top level structure
	do  Result := the_lib.generate_new_structure_name("TOP")
	end;

feature {NONE} -- minor parsers

	skip_cif_sep is
		-- skip to the next non-sep character
	do  from
	    until not is_cif_sep(last_character)
	    loop  get_next_character
	    end;
	ensure not is_cif_sep(last_character)
	end;

	last_integer: INTEGER;

	read_cif_integer: BOOLEAN is
		-- we leave last_character at first non-digit
		-- and value in last_integer
	local i: INTEGER
	do  skip_cif_sep;
	    if last_character.is_digit
	     then from  
		  until not last_character.is_digit
		  loop  i := i * 10 + (last_character |-| '0')
			get_next_character
		  end;
		  last_integer := i;
		  Result := true
	    end;
	end;

	read_cif_sinteger: BOOLEAN is
		-- we leave last_character at first non-digit
		-- and value in last_integer
	local i: INTEGER; neg: BOOLEAN
	do  skip_cif_sep;
	    if last_character = '-'
	     then neg := true;
		  get_next_character
	    end;
	    if last_character.is_digit
	     then from  
		  until not last_character.is_digit
		  loop  i := i * 10 + (last_character |-| '0')
			get_next_character
		  end;
		  if neg then last_integer := -i
			 else last_integer := i end;
		  Result := true
	    end;
	end;

	last_X, last_Y: DOUBLE;

	read_cif_point: BOOLEAN is
		-- point values are scaled by curr_scale
	do  if read_cif_sinteger
	     then last_X := last_integer * curr_scale;
		  if read_cif_sinteger
		   then last_Y := last_integer * curr_scale
			Result := true
		   else log_io_error(error_is_xxnoyy, Void)
		  end
	    end;
	end;

	read_path_into(pth: LIST_OF_POINTS) is
		-- parse path  ==  point {sep point} into pth
	do  from  
	    until not read_cif_point
	    loop  pth.extend_xy(last_X, last_Y)
	    end
	end;

	read_shortname: STRING is
		-- parse c[c][c][c] (layer name)
	do  from  get_next_nonblank;  create Result.make(4);
	    until not is_cif_c(last_character)
	    loop  Result.extend(last_character);
		  get_next_character
	    end;
	end;

	read_name: STRING is
		-- (in user commands) any string inside white space or ;
	do  from	-- skip leading white space
	    until not su.is_white(last_character)
	    loop  get_next_character
	    end;
	    from  create Result.make(32)
	    until last_character = ';' or su.is_white(last_character)
	    loop  Result.extend(last_character);
		  get_next_character
	    end;
	end;

	read_qname: STRING is
		-- quoted string; gobble to next '%''
	do  create Result.make(20);
	    from  get_next_character	-- step over opening quote
	    until last_character = '%''
	    loop  Result.extend(last_character);
		  get_next_character
	    end;
	    get_next_character		-- step over closing quote
	end;

feature {NONE} -- the cif reader commands

	add_to_current_layout(lay: LAYOUT) is
		-- add an item to the current_body (if legal)
	do  if lay.is_unlayered and curr_body.current_layer = Void
	     then log_io_error(error_is_notlyr, Void)
	     else curr_body.extend(lay)
	    end;
	end;

	temp_polygon: LAYOUT_POLYGON;
	temp_path: LAYOUT_PATH;
		-- collectors before we make actual copies to store

	polygon_command is
	local  pts: LAYOUT_POLYGON
	do  -- parse
	    if temp_polygon = Void
	     then create temp_polygon.make(40)
	     else temp_polygon.wipe_out
	    end;
	    read_path_into(temp_polygon)
	    must_be_at_semi;
	    --
	    if temp_polygon.is_empty
	     then log_io_error(error_is_nopnts, Void)
	     else create pts.make_copy(temp_polygon);
		  if pts.count < 3 then log_io_error(error_is_fewpts, Void) end;
		  add_to_current_layout(pts)
	    end;
	end;

	box_command is
	local  box: LAYOUT_RECTANGLE; rbox: LAYOUT_RELOCATION; OK: BOOLEAN;
		sx, sy, cx, cy, dx, dy: DOUBLE; m1,m2,m3,mat: MATRIX
	do  -- parse
	    if read_cif_point		-- size
	     then sx := last_X; sy := last_Y;
		  if read_cif_point	-- center
		   then cx := last_X; cy := last_Y;
			OK := true;
		  end
	    end
	    if OK
	     then if read_cif_point 	-- direction (optional)
		   then dx := last_X; dy := last_Y
		   else dx := 1;      dy := 0
		  end;
		  if dy = 0		-- as is
		   then if dx = 0 then log_io_error(error_is_dir000, Void) end;
			create box.make_by_size_and_center(sx, sy, cx, cy);
			add_to_current_layout(box)
		  elseif dx = 0		-- 90 degrees off
		   then create box.make_by_size_and_center(sy, sx, cx, cy);
					-- switch:    ^^  ^^
			add_to_current_layout(box)
		   else			-- other: just rotate it
			create box.make_by_size_and_center(sx, sy, cx, cy);
			create m1.make_translation_XY(-cx, -cy);
			create m2.make_orientation_XY(false, dx, dy);
			create m3.make_translation_XY(cx, cy);
			mat := (m1.and_then(m2)).and_then(m3);
			create rbox.make(box, mat);
			add_to_current_layout(rbox)
		  end
	     else log_io_error(error_is_fewpts, Void)
	    end
	    must_be_at_semi;
	end;

	round_command is
	local m1, m2, mat: MATRIX; ccc: LAYOUT_POLYGON;
	do  if read_cif_integer		-- diameter
	     then create m1.make_magnification(last_integer*curr_scale)
		  if read_cif_point	-- center
		   then create m2.make_translation_XY(last_X, last_Y);
			mat := m1.and_then(m2);
			create ccc.make_transformed(the_unit_octagon, mat);
			add_to_current_layout(ccc);
		   else log_io_error(error_is_nopnts, Void)
		  end
	     else log_io_error(error_is_nowidt, Void)
	    end;
	    must_be_at_semi
	end;

	wire_command is
	local  w: LAYOUT_PATH; wid: DOUBLE;
	do  if read_cif_integer
	     then wid := last_integer*curr_scale
		  if temp_path = Void
		   then create temp_path.make(wid, half_square_ends)
		   else temp_path.wipe_out;
			temp_path.set_width(wid)
		  end;
		  read_path_into(temp_path);
		  create w.make_copy(temp_path);
		  add_to_current_layout(w);
		  if w.width = 0 then log_io_error(error_is_wire00, Void) end
	     else log_io_error(error_is_nowidt, Void)
	    end;
	    must_be_at_semi;
	end;

	layer_command is
	local lyr: STRING; dtl: LAYER;
	do  lyr := read_shortname;
	    must_be_at_semi;
	    -- process lyr
	    dtl := layer_by_cif(lyr);
	    curr_body.set_current_layer(dtl.layer_name)
	end;

	definition_command is
		-- includes DS, DF, and DD
	do  get_next_nonblank;
	    inspect last_character
	     when 'S' then ds_command;
	     when 'F' then df_command;
	     when 'D' then log_io_error(error_is_deldef, Void);
			   skip_to_semi
	     else log_io_error(error_is_unkcmd, Void);
		  skip_to_semi
	    end
	    must_be_at_semi
	end;

	ds_command is
	local  nam, a, b: INTEGER
	do  if read_cif_integer
	     then -- have a new structure and its name
		  nam := last_integer;
		  a := 1; b := 1;
		  if read_cif_integer	-- a , b [optional]
		   then a := last_integer
			if read_cif_integer
			 then b := last_integer
			 else log_io_error(error_is_halfsc, Void);
			end
		  end
	     else log_io_error(error_is_nosyno, Void)
	    end;
	    if curr_structure /= Void
	     then -- DS within DS
		  log_io_error(error_is_dsinds, Void)
	     else -- begin new structure
		  create curr_structure.make(nam.out);
		  curr_structure.set_cif_number(nam);
		  create curr_body.make(20);
		  curr_body.set_is_caching_bb(true)
		  curr_scale := (top_level_scale * a ) / b
	    end;
	end

	df_command is
	do  get_next_character;
	    if curr_structure = Void
	     then log_io_error(error_is_dfnods, Void)
	     else -- close it out and revert to top_level
		  curr_body.trim(100)
		  curr_structure.set_body(curr_body);
		  if the_lib.has_structure(curr_structure.name)
		   then log_io_error(error_is_dupnam, curr_structure.name)
		   else the_lib.add_structure(curr_structure)
		  end;
		  make_top_current
	    end;
	end;

	call_command is
	local rf: LAYOUT_REFERENCE; mat: MATRIX; trf: LAYOUT_RELOCATION;
			nam: STRING; nrf: LAYOUT_NAMING; it: LAYOUT;
	do  if read_cif_integer
	     then create rf.make_SREF(last_integer.out);
		  if not is_cif_nonblank(last_character)
		   then get_next_nonblank end;
		  -- now loop thru the transformations (if any)
		  from  mat := identity;
		  until last_character = ';'
		  loop  inspect last_character
			 when 'M'
			  then mat := mat.and_then(call_mirroring);
			 when 'R'
			  then mat := mat.and_then(call_rotation);
			 when 'T'
			  then mat := mat.and_then(call_translation);
			 when '%''
			  then nam := read_qname
			  else log_io_error(error_is_unktrn, Void);
				skip_to_semi
			end
			if not is_cif_nonblank(last_character)
			 then get_next_nonblank end;
		  end;
		  -- and store away the reference
		  it := rf;
		  if not mat.is_identity
		   then create trf.make(rf, mat); it := trf
		  end;
		  if nam /= Void
		   then create nrf.make(it, nam); it := nrf
		  end;
		  add_to_current_layout(it)
	     else log_io_error(error_is_nosyno, Void)
	    end;
	    must_be_at_semi
	end;

	call_mirroring: MATRIX is
		-- last_character is 'M' for Mirror
	do  get_next_nonblank;
	    inspect last_character
	     when 'X' then Result := mirror_in_y;
	     when 'Y' then Result := mirror_in_x;
	     else log_io_error(error_is_unktrn, Void)
	    end;
	    if last_character /= ';' then get_next_nonblank end;
	end;

	call_rotation: MATRIX is
		-- last_character is 'R' for Rotation
	do  if read_cif_point
	     then create Result.make_orientation_XY(false, last_X, last_Y)
	     else log_io_error(error_is_unktrn, Void);
	    end;
	end;

	call_translation: MATRIX is
		-- last_character is 'T' for Translation
	do  if read_cif_point
	     then create Result.make_translation_XY(last_X, last_Y)
	     else log_io_error(error_is_unktrn, Void);
	    end;
	end;

	comment_command is
	local lev: INTEGER
	do  from  lev := 1;
	    until lev = 0
	    loop  get_next_character
		  if last_character = '(' then lev := lev + 1
		  elseif last_character = ')' then lev := lev - 1
		  end
	    end;
	    get_next_character;		-- step past the final right paren
	    must_be_at_semi
	end;

	user_command is
	local val: STRING; lyr: LAYER; it: LAYOUT;
	do  if read_cif_integer		-- should always be true here!!
	     then end;
	    if last_integer = 0
	     then -- no match
	    elseif last_integer = lioopt.symbol_name_command
	     then if curr_structure /= Void and then  -- in symbol
		     (not name_table.has(curr_structure.name))
		   then -- (just use first cmd after DS)
			val := read_name;
			if su.not_empty(val)
			 then name_table.extend(val, curr_structure.name)
			end
		  end;
	    elseif last_integer = lioopt.port_name_command
	     then val := read_name;
		  if read_cif_point
		   then create {LAYOUT_PORT} it.make_by_size_and_center(0, 0,
							last_X, last_Y)
		  end;
		  if su.not_empty(val) and it /= Void
		   then create {LAYOUT_NAMING} it.make(it, val)
		  end;
		  lyr := the_techn.layer_by_cif(read_shortname);
		  if lyr /= Void and it /= Void
		   then create {LAYOUT_LAYERING} it.make(it, lyr.layer_name);
			add_to_current_layout(it)
		  end;
	    end;
	    skip_to_semi;
	end;

	process_cif_command is
		-- last_character should be a cif command (other than end);
		--  read it and process that command
	do  inspect last_character
	     when 'P' then polygon_command;
	     when 'B' then box_command;
	     when 'R' then round_command;
	     when 'W' then wire_command;
	     when 'L' then layer_command;
	     when 'D' then definition_command;
	     when 'C' then call_command;
	     when '(' then comment_command;
	     when '0'..'9' then user_command;
	     else log_io_error(error_is_unkcmd, Void);
		  skip_to_semi;
	    end;
	end;

feature -- the reader

	read_library is
	require the_techn /= Void
	local rnam: HASHED_OBJECT_MODIFIER[STRING];
		rref: LC_RENAME_REFERENCES;
		libu: DOUBLE
	do  -- initialize everything
	    create the_messages.make(reader_application);
	    unknown_layers := lioopt.init_unknown_layers;
	    if the_lib = Void
	     then create the_lib.make("CIF_LIBRARY", 0)
	    end;
	    if the_lib.unit_size = 0
	     then if lioopt.input_library_unit = 0
		   then libu := lioopt.one_nanometer
		   else libu := lioopt.input_library_unit
		  end;
		  the_lib.resize(libu)
	    end;
	    set_other;
	    top_level_scale := lioopt.one_centimicron / the_lib.unit_size;
	    create top_level_body.make(1);
	    top_level_body.set_is_caching_bb(true);
	    make_top_current;
	    line_number := 0;
	    get_next_line;
	    create name_table.make(41);
	    -- read on, one cif command at a time
	    from  get_next_nonblank
	    until last_character = 'E'
	    loop  process_cif_command;
		  get_next_nonblank
	    end;
	    -- termination errors (still in DS)
	    if curr_structure /= Void
	     then log_io_error(error_is_dsnodf, Void) end
	    -- switch to the symbolic names for the structures
	    if not name_table.is_empty	-- (before inserting top cell)
	     then create rnam.make(name_table);
		  create rref.make(name_table);
		  the_lib.modify_structure_names(rnam);
		  top_level_body.traverse(rref, identity)
	    end;
	    -- save top_level_body in its own structure
	    create curr_structure.make(get_top_level_name);
	    curr_structure.set_body(top_level_body);
	    the_lib.add_structure(curr_structure);
	    the_lib.set_root_structure(curr_structure.name);
	end;

feature {NONE} -- creation

	make(src: DATA_SOURCE) is
		-- NOTE: we neither open nor close the source
	do  my_file := src
	end;

end -- CIF_FILE_READER
