
indexing
	
	description: "Converter that converts CIF to AutoCAD DXF format"
	author: "David Hollenberg";
	date: "April 14, 2005"

class CIF_TO_DXF_CONVERTER
inherit
	ANY
	CONDITION_UTILITIES
		export
			{NONE} all
		end
	UNIX_OS_ACCESS
		export
			{NONE} all
		end

create
	make

feature {NONE} -- Creation

	make (rsh_path, c2d_path, c2d_host: STRING) is
			-- Create with `rsh_path' as pathname of rshell command
			-- `c2d_path' as pathname of CIF to DXF conversion
			-- program and `c2d_host' as fully qualified host
			-- name on which conversion program must be run
		require
			rsh_path_exists: rsh_path /= Void
			rsh_path_not_empty: not rsh_path.is_empty
			c2d_path_exists: c2d_path /= Void
			c2d_path_not_empty: not c2d_path.is_empty
			c2d_host_exists: c2d_host /= Void
			c2d_host_not_empty: not c2d_host.is_empty
		do
			rshell_pathname := rsh_path
			cif_to_dxf_pathname := c2d_path
			cif_to_dxf_host := c2d_host
		end

feature -- Properties

	rshell_pathname: STRING
			-- Path name of rshell command

	cif_to_dxf_pathname: STRING
			-- Path name of CIF to DXF conversion program

	cif_to_dxf_host: STRING
			-- Fully qualified name of host on which conversion
			-- program must be run

feature -- Operations

	convert_cif_to_dxf (in_fname, out_fname: STRING) is
			-- Convert CIF file named `in_fname' to DXF and
			-- write to `out_fname'
		require
			in_fname_not_void: in_fname /= Void
			in_fname_not_empty: not in_fname.is_empty
			out_fname_not_void: out_fname /= Void
			out_fname_not_empty: not out_fname.is_empty
		local
			f: LARGE_RAW_FILE
			process: UNIX_PROCESS
			checker: FILE_EXISTENCE_CHECK_OP
			delete_op: FILE_DELETE_OP
			temp_in_fname, temp_out_fname: STRING
		do
			create checker
			
			checker.check_existence (rshell_pathname)
			if not checker.path_exists then
				handle_hard_error (rshell_description + " not found")
			end
			create f.make (rshell_pathname)
			if not f.is_plain then
				handle_hard_error (rshell_description + " is not a plain file")
			elseif not f.is_executable then
				handle_hard_error (rshell_description + " is not executable")
			end
			
			checker.check_existence (cif_to_dxf_pathname)
			if not checker.path_exists then
				handle_hard_error (program_description + " not found")
			end
			create f.make (cif_to_dxf_pathname)
			if not f.is_plain then
				handle_hard_error (program_description + " is not a plain file")
			elseif not f.is_executable then
				handle_hard_error (program_description + " is not executable")
			else
				checker.check_existence (out_fname)
				if checker.path_exists then
					create delete_op
					delete_op.delete (out_fname)
				end
				
				if unix_os.local_host_name.is_equal (cif_to_dxf_host) then
					-- No rshell needed
					create process.make (cif_to_dxf_pathname)
					process.set_arguments (<< in_fname, out_fname >>)
				else
					-- Quote '#' characters so that they get past rshell
					-- and are not treated as "comment start" characters
					temp_in_fname := in_fname.twin
					temp_out_fname := out_fname.twin
					temp_in_fname.replace_substring_all ("#", "\#")
					temp_out_fname.replace_substring_all ("#", "\#")
					create process.make (rshell_pathname)
					process.set_arguments (<< cif_to_dxf_host, cif_to_dxf_pathname, temp_in_fname, temp_out_fname >>)
				end
		    		process.set_input_file_name (unix_os.null_file_name)
		    		process.set_output_file_name (unix_os.null_file_name)
		    		process.set_error_same_as_output
				
		    		process.spawn_wait
				
		    		if process.is_exit_status (process.status) then
					if process.exit_code (process.status) /= 0 then
						handle_hard_error (program_description + " exited with code " + process.exit_code (process.status).out)
					end
				else
					handle_hard_error (program_description + " terminated or stopped by signal " + process.signal_number (process.status).out)
				end
				
				checker.check_existence (out_fname)
				if not checker.path_exists then
					handle_hard_error (program_description + " did not convert CIF to DXF.%NOutput file " + out_fname + " not found")
				end
			end
		end

feature {NONE} -- Implementation

	rshell_description: STRING is
			-- Description of rshell program
		do
			Result := "Rshell program file " + rshell_pathname
		ensure
			result_exists: Result /= Void
		end

	program_description: STRING is
			-- Description of conversion program
		do
			Result := "CIF to DXF conversion program file " + cif_to_dxf_pathname
		ensure
			result_exists: Result /= Void
		end

end
