
indexing
	
	description: "Action that computes cut die for a project"
	author: "David Hollenberg"
	date: "October 31, 2006"

class COMPUTE_CUT_DIE_ACTION
inherit
	VIEW_ELEMENT_ACTION
	GRIDABLE

feature -- Properties

	default_scribe_width: DOUBLE
			-- Width of default scribe (used for implied scribes
			-- outside field of view edges)

	half_scribe_width: DOUBLE
			-- Width of default half scribe (used for
			-- implied scribes outside field of view
			-- edges)

	horizontal_scribes: LIST [RETICLE_VIEW_ELEMENT]
			-- Horizontal scribes, sorted by ascending center in Y

	vertical_scribes: LIST [RETICLE_VIEW_ELEMENT]
			-- Vertical scribes, sorted by ascending center in X

	horizontal_cuts: LIST [RETICLE_VIEW_ELEMENT]
			-- Horizontal cuts, sorted by ascending center in Y

	vertical_cuts: LIST [RETICLE_VIEW_ELEMENT]
			-- Vertical cuts, sorted by ascending center in X

feature -- Modification

	set_default_scribe_width (a_width: like default_scribe_width)
		do
			default_scribe_width := a_width
			half_scribe_width := a_width / 2.0
		ensure
			default_scribe_width_set: default_scribe_width = a_width
		end

	set_horizontal_scribes (a_list: like horizontal_scribes)
		require
			a_list_not_void: a_list /= Void
		do
			horizontal_scribes := a_list
		ensure
			horizontal_scribes_set: horizontal_scribes = a_list
		end

	set_vertical_scribes (a_list: like vertical_scribes)
		require
			a_list_not_void: a_list /= Void
		do
			vertical_scribes := a_list
		ensure
			vertical_scribes_set: vertical_scribes = a_list
		end

	set_horizontal_cuts (a_list: like horizontal_cuts)
		require
			a_list_not_void: a_list /= Void
		do
			horizontal_cuts := a_list
		ensure
			horizontal_cuts_set: horizontal_cuts = a_list
		end

	set_vertical_cuts (a_list: like vertical_cuts)
		require
			a_list_not_void: a_list /= Void
		do
			vertical_cuts := a_list
		ensure
			vertical_cuts_set: vertical_cuts = a_list
		end

feature -- Operations

	apply (an_element: RETICLE_VIEW_ELEMENT)
			-- Apply action to `an_element'
		local
			l_fields: LIST [RETICLE_FIELD]
			l_field: RETICLE_FIELD
			l, r, t, b: DOUBLE
			diff_l, diff_r, diff_t, diff_b: DOUBLE
			diff_x, diff_y: DOUBLE
			sl, sr, st, sb: DOUBLE
			el, er, et, eb: DOUBLE
			cl, cr, ct, cb: DOUBLE
			scribe_cut_die, edge_cut_die, cut_cut_die: CONST_RECTANGLE
			has_surrounding_cuts: BOOLEAN
		do
			l_fields := an_element.intersecting_fields
			if l_fields.count = 1 then
				-- Intersects exactly one field
				l_field := l_fields.first
				l := grid.rounded_edge (an_element.left_x)
				r := grid.rounded_edge (an_element.right_x)
				b := grid.rounded_edge (an_element.bottom_y)
				t := grid.rounded_edge (an_element.top_y)

				-- Compute cut die using center of cuts

				is_isolated_site := an_element.is_isolated_in_current_cut_pattern

				compute_low_cut_edge (False, l, b, t, -Very_distant_position, vertical_cuts)
				cl := center_cut_edge
				set_isolating_cut
				compute_high_cut_edge (False, r, b, t, Very_distant_position, vertical_cuts)
				cr := center_cut_edge
				set_isolating_cut
				compute_low_cut_edge (True, b, l, r, -Very_distant_position, horizontal_cuts)
				cb := center_cut_edge
				set_isolating_cut
				compute_high_cut_edge (True, t, l, r, Very_distant_position, horizontal_cuts)
				ct := center_cut_edge
				set_isolating_cut
				has_surrounding_cuts := 
				   cl > -Very_distant_position and
				   cr < Very_distant_position and
				   cb > -Very_distant_position and
				   ct < Very_distant_position
				if has_surrounding_cuts and
				   not has_cuts_in_range (cl, cr, vertical_cuts) and
				   not has_cuts_in_range (cb, ct, horizontal_cuts) then
					-- Has cuts on all four sides
					-- and there are no cuts
					-- through computed cut die
					create cut_cut_die.make_by_corners (cl, cb, cr, ct)
				else
					cut_cut_die := Void
				end

				-- Compute cut die using center of scribes
				-- and cut die using edge of scribes

				compute_low_cut_edge (False, l, b, t, l_field.left_x - half_scribe_width, vertical_scribes)
				sl := center_cut_edge
				el := edge_cut_edge
				compute_high_cut_edge (False, r, b, t, l_field.right_x + half_scribe_width, vertical_scribes)
				sr := center_cut_edge
				er := edge_cut_edge
				compute_low_cut_edge (True, b, l, r, l_field.bottom_y - half_scribe_width, horizontal_scribes)
				sb := center_cut_edge
				eb := edge_cut_edge
				compute_high_cut_edge (True, t, l, r, l_field.top_y + half_scribe_width, horizontal_scribes)
				st := center_cut_edge
				et := edge_cut_edge

				-- Temporarily move die the least amount
				-- possible in X and in Y so that die just
				-- abuts one of the horizontal and one of the
				-- vertical edges discovered above.  Then
				-- redo the computation.  This gives a better
				-- result when die is not yet completely inside
				-- the surrounding scribes or field of view edges,
				-- so that centering works when die overlaps an
				-- adjacent scribe or field edge.  It gives the
				-- same result (no change) if die is completely
				-- inside surrounding scribes and field of
				-- view edges.

				diff_l := el - l
				diff_r := er - r
				if grid.rounded_edge (diff_l.abs) <= grid.rounded_edge (diff_r.abs) then
					diff_x := diff_l
				else
					diff_x := diff_r
				end
				l := grid.rounded_edge (l + diff_x)
				r := grid.rounded_edge (r + diff_x)

				diff_t := et - t
				diff_b := eb - b
				if grid.rounded_edge (diff_t.abs) <= grid.rounded_edge (diff_b.abs) then
					diff_y := diff_t
				else
					diff_y := diff_b
				end
				t := grid.rounded_edge (t + diff_y)
				b := grid.rounded_edge (b + diff_y)

				compute_low_cut_edge (False, l, b, t, l_field.left_x - half_scribe_width, vertical_scribes)
				sl := center_cut_edge
				el := edge_cut_edge
				compute_high_cut_edge (False, r, b, t, l_field.right_x + half_scribe_width, vertical_scribes)
				sr := center_cut_edge
				er := edge_cut_edge
				compute_low_cut_edge (True, b, l, r, l_field.bottom_y - half_scribe_width, horizontal_scribes)
				sb := center_cut_edge
				eb := edge_cut_edge
				compute_high_cut_edge (True, t, l, r, l_field.top_y + half_scribe_width, horizontal_scribes)
				st := center_cut_edge
				et := edge_cut_edge

				create scribe_cut_die.make_by_corners (sl, sb, sr, st)
				create edge_cut_die.make_by_corners (el, eb, er, et)
				
			else
				scribe_cut_die := Void
				edge_cut_die := Void
				cut_cut_die := Void
				has_surrounding_cuts := False
			end
			an_element.set_scribe_cut_die (scribe_cut_die)
			an_element.set_edge_cut_die (edge_cut_die)
			an_element.set_cut_cut_die (cut_cut_die)
			an_element.set_has_surrounding_cuts (has_surrounding_cuts)
		end

feature {NONE} -- Implementation

	compute_low_cut_edge (is_horiz: BOOLEAN a_position, a_low_limit, a_high_limit: DOUBLE an_ext_edge: DOUBLE scribes: LIST [RETICLE_VIEW_ELEMENT])
			-- Compute left or bottom `center_cut_edge' and
			-- `edge_cut_edge' for `an_element', whose left or
			-- bottom edge is at `a_position' and extends from
			-- `a_low_limit' to `a_high_limit'.
			-- `an_ext_edge' is the position of the center
			-- of the implied scribe just outside the
			-- appropriate field edge
		local
			done: BOOLEAN
			scribe: RETICLE_VIEW_ELEMENT
			sp, cen, wid, low, high, half_width: DOUBLE
		do
			cut := Void
			from
				if an_ext_edge /= -Very_distant_position then
					center_cut_edge := grid.rounded_center (an_ext_edge)
				else
					center_cut_edge := an_ext_edge
				end
				half_width := half_scribe_width
				scribes.start
			until
				scribes.after or done
			loop
				scribe := scribes.item
				if is_horiz then
					if scribe.has_zero_size_in_y then
						sp := grid.rounded_center (scribe.center_y)
						wid := 0.0
					else
						sp := grid.rounded_edge (scribe.top_y)
						wid := grid.rounded_edge (scribe.size_y)
					end
					cen := grid.rounded_center (scribe.center_y)
					low := grid.rounded_edge (scribe.left_x)
					high := grid.rounded_edge (scribe.right_x)
				else
					if scribe.has_zero_size_in_x then
						sp := grid.rounded_center (scribe.center_x)
						wid := 0.0
					else
						sp := grid.rounded_edge (scribe.right_x)
						wid := grid.rounded_edge (scribe.size_x)
					end
					cen := grid.rounded_center (scribe.center_x)
					low := grid.rounded_edge (scribe.bottom_y)
					high := grid.rounded_edge (scribe.top_y)
				end
				if sp <= a_position then
					if cen >= an_ext_edge and truly_intersects (low, high, a_low_limit, a_high_limit) then
						cut := scribe
						center_cut_edge := cen
						half_width := wid / 2.0
					end
				else
					done := True
				end
				scribes.forth
			end
			edge_cut_edge := grid.rounded_edge (center_cut_edge + half_width)
		end

	compute_high_cut_edge (is_horiz: BOOLEAN a_position, a_low_limit, a_high_limit: DOUBLE an_ext_edge: DOUBLE scribes: LIST [RETICLE_VIEW_ELEMENT])
			-- Compute right or top `center_cut_edge' and
			-- `edge_cut_edge' for `an_element', whose right or
			-- top edge is at `a_position' and extends from
			-- `a_low_limit' to `a_high_limit'.
			-- `an_ext_edge' is the position of the center
			-- of the implied scribe just outside the
			-- appropriate field edge
		local
			done: BOOLEAN
			scribe: RETICLE_VIEW_ELEMENT
			sp, cen, low, high, wid, half_width: DOUBLE
		do
			cut := Void
			from
				if an_ext_edge /= Very_distant_position then
					center_cut_edge := grid.rounded_center (an_ext_edge)
				else
					center_cut_edge := an_ext_edge
				end
				half_width := half_scribe_width
				scribes.finish
			until
				scribes.before or done
			loop
				scribe := scribes.item
				if is_horiz then
					if scribe.has_zero_size_in_y then
						sp := grid.rounded_center (scribe.center_y)
						wid := 0.0
					else
						sp := grid.rounded_edge (scribe.bottom_y)
						wid := grid.rounded_edge (scribe.size_y)
					end
					cen := grid.rounded_center (scribe.center_y)
					low := grid.rounded_edge (scribe.left_x)
					high := grid.rounded_edge (scribe.right_x)
				else
					if scribe.has_zero_size_in_x then
						sp := grid.rounded_center (scribe.center_x)
						wid := 0.0
					else
						sp := grid.rounded_edge (scribe.left_x)
						wid := grid.rounded_edge (scribe.size_x)
					end
					cen := grid.rounded_center (scribe.center_x)
					low := grid.rounded_edge (scribe.bottom_y)
					high := grid.rounded_edge (scribe.top_y)
				end
				if sp >= a_position then
					if cen <= an_ext_edge and truly_intersects (low, high, a_low_limit, a_high_limit) then
						cut := scribe
						center_cut_edge := cen
						half_width := wid / 2.0
					end
				else
					done := True
				end
				scribes.back
			end
			edge_cut_edge := grid.rounded_edge (center_cut_edge - half_width)
		end

	truly_intersects (low, high, a_low_limit, a_high_limit: DOUBLE): BOOLEAN
			-- Does segment (`low', `high') intersect edge
			-- defined by `a_low_limit' and `a_high_limit', not
			-- including end points of edge?
		do
			Result := low < a_high_limit and high > a_low_limit
		end

	has_cuts_in_range (low, high: DOUBLE cuts: LIST [RETICLE_VIEW_ELEMENT]): BOOLEAN
			-- Do any cuts in `cuts' have position between
			-- low and high, not including end points?
		local
			l_cut: RETICLE_VIEW_ELEMENT
			pos: DOUBLE
			done: BOOLEAN
		do
			from
				cuts.start
			until
				cuts.after or Result or done
			loop
				l_cut := cuts.item
				pos := grid.rounded_center (l_cut.cut_position)
				Result := Result or (pos > low and pos < high)
				done := pos >= high
				cuts.forth
			end
		end

	set_isolating_cut
			-- Set `cut' to be an isolating cut if not Void
		do
			if cut /= Void and is_isolated_site then
				cut.set_is_isolating_cut (True)
			end
		end

	center_cut_edge: DOUBLE
			-- Position of last cut edge computed, using scribe
			-- or cut center
	
	edge_cut_edge: DOUBLE
			-- Position of last cut edge computed, using scribe
			-- closest edge

	cut: RETICLE_VIEW_ELEMENT
			-- Cut or scribe for last computed `center_cut_edge'

	is_isolated_site: BOOLEAN
			-- Are we working on finding the cut dies for an
			-- isolated site?

	Very_distant_position: DOUBLE = 1.0E100
			-- Impossibly distant position

end
