
indexing
	
	description: "Action that computes point to be used for a Measure %
		%Distance operation"
	author: "David Hollenberg"
	date: "November 30, 2006"

class COMPUTE_MEASURE_POINT_ACTION
inherit
	VIEW_ELEMENT_ACTION
	DOUBLE_MATH
		export
			{NONE} all
		end

feature -- Properties

	view_x: INTEGER
			-- X coordinate of point user selected for measurement

	view_y: INTEGER
			-- X coordinate of point user selected for measurement

	limit_x: INTEGER
			-- X coordinate of rightmost visible pixel in
			-- drawing area

	limit_y: INTEGER
			-- Y coordinate of bottommost visible pixel in
			-- drawing area

	corner_threshold: INTEGER
			-- Use closest visible corner instead of side
			-- if distance to corner is <= this value

	is_first_pass: BOOLEAN
			-- Is this the first pass?  First pass finds the
			-- closest edge to point where user clicked and
			-- second pass finds point along that edge where
			-- measure marker should be placed

	found_visible_edge: BOOLEAN
			-- Did first pass find any visible edges at all?
		do
			Result := closest_visible_element /= Void
		end

	measure_point: POINT
			-- Computed point for measurement or Void if no
			-- visible object edge in drawing area

feature -- Modification

	set_view_coordinates (a_x, a_y: INTEGER)
			-- Set `view_x' and `view_y' to `a_x' and `a_y'
		do
			view_x := a_x
			view_y := a_y
		ensure
			view_x_set: view_x = a_x
			view_y_set: view_y = a_y
		end

	set_limits (a_limit_x, a_limit_y: INTEGER)
			-- Set `limit_x' and `limit_y' to `a_limit_x' and
			-- `a_limit_y'
		do
			limit_x := a_limit_x
			limit_y := a_limit_y
		ensure
			limit_x_set: limit_x = a_limit_x
			limit_y_set: limit_y = a_limit_y
		end

	set_corner_threshold (a_value: INTEGER)
			-- Set `corner_threshold' to `a_value'
		require
			a_value_nonnegative: a_value >= 0
		do
			corner_threshold := a_value
		ensure
			corner_threshold_set: corner_threshold = a_value
		end

	set_is_first_pass (a_value: like is_first_pass)
			-- Set `is_first_pass' to `a_value'
		do
			is_first_pass := a_value
		ensure
			first_pass_set: is_first_pass = a_value
		end

feature -- Operations

	initialize
			-- Initialize before applying action to any
			-- elements
		do
			if is_first_pass then
				measure_point := Void
				closest_visible_element := Void
				minimum_distance := 1.0E100
			else
				found_intersection := False
				minimum_intersection_distance := 1.0E100
				clicked_point := closest_visible_element.view.pixel_center_model_coords (view_x, view_y)
			end
		end

	finish
			-- Action to do after applying action to element
		local
			l_x, l_y: DOUBLE
			view: RETICLE_VIEW
		do
			if closest_visible_element /= Void then
				view := closest_visible_element.view
				if is_vertical_edge then
					l_x := closest_edge_model_position
					if is_point_between_ends and found_intersection and minimum_intersection_view_distance <= corner_threshold then
						l_y := closest_intersecting_edge_model_position
					elseif is_corner_visible then
						l_y := closest_corner_model_position
					else
						l_y := clicked_point.y
					end
				else
					if is_point_between_ends and found_intersection and minimum_intersection_view_distance <= corner_threshold then
						l_x := closest_intersecting_edge_model_position
					elseif is_corner_visible then
						l_x := closest_corner_model_position
					else
						l_x := clicked_point.x
					end
					l_y := closest_edge_model_position
				end
				create measure_point.make_xy (l_x, l_y)
			end
		end

	apply (an_element: RETICLE_VIEW_ELEMENT)
			-- Apply action to `an_element'
		do
			if is_first_pass then
				apply_first_pass (an_element)
			elseif is_point_between_ends then
				apply_second_pass (an_element)
			end
		end

feature {NONE} -- Implementation

	apply_first_pass (an_element: RETICLE_VIEW_ELEMENT)
			-- Apply action to `an_element' when on first pass
			-- (finding closest edge to point where user clicked)
		local
			low_x, high_x, low_y, high_y: INTEGER
			left, right, bottom, top: DOUBLE
		do
			low_x := an_element.x
			high_x := an_element.x + an_element.width
			low_y := an_element.y
			high_y := an_element.y + an_element.height
			left := an_element.gridded_left_x
			right := an_element.gridded_right_x
			bottom := an_element.gridded_bottom_y
			top := an_element.gridded_top_y
			check_edge (an_element, left, bottom, top, low_x, low_y, high_y, view_x, view_y, limit_x, limit_y, True)
			check_edge (an_element, right, bottom, top, high_x, low_y, high_y, view_x, view_y, limit_x, limit_y, True)
			check_edge (an_element, top, left, right, low_y, low_x, high_x, view_y, view_x, limit_y, limit_x, False)
			check_edge (an_element, bottom, left, right, high_y, low_x, high_x, view_y, view_x, limit_y, limit_x, False)
		end

	apply_second_pass (an_element: RETICLE_VIEW_ELEMENT)
			-- Apply action to `an_element' when on second pass
			-- (finding best point along closest edge to place
			-- measurement marker)
		local
			low_x, high_x, low_y, high_y: INTEGER
			left, right, bottom, top: DOUBLE
		do
			low_x := an_element.x
			high_x := an_element.x + an_element.width
			low_y := an_element.y
			high_y := an_element.y + an_element.height
			left := an_element.gridded_left_x
			right := an_element.gridded_right_x
			bottom := an_element.gridded_bottom_y
			top := an_element.gridded_top_y
			if is_vertical_edge then
				-- Closest edge is vertical - check horizontal
				-- edge intersections
				check_intersection (an_element, top, left, right, clicked_point.y, low_y, low_x, high_x, view_y, limit_y, limit_x, False)
				check_intersection (an_element, bottom, left, right, clicked_point.y, high_y, low_x, high_x, view_y, limit_y, limit_x, False)
			else
				check_intersection (an_element, left, bottom, top, clicked_point.x, low_x, low_y, high_y, view_x, limit_x, limit_y, True)
				check_intersection (an_element, right, bottom, top, clicked_point.x, high_x, low_y, high_y, view_x, limit_x, limit_y, True)
			end
		end

	check_edge (an_element: RETICLE_VIEW_ELEMENT a_real_edge, a_real_low_end, a_real_high_end: DOUBLE an_edge, a_low_end, a_high_end, a_value, an_other_direction_value, an_edge_limit, a_visible_limit: INTEGER is_vert: BOOLEAN)
		local
			diff: DOUBLE
			low_dist, high_dist: INTEGER
			between_ends: BOOlEAN
		do
			if an_edge >= 0 and an_edge <= an_edge_limit and a_low_end <= a_visible_limit and a_high_end >= 0 then
				-- At least some part of edge is visible
				if an_other_direction_value >= a_low_end and an_other_direction_value <= a_high_end then
					between_ends := True
					diff := (an_edge - a_value).abs
				elseif an_other_direction_value < a_low_end then
					between_ends := False
					diff := sqrt ((an_edge - a_value) ^ 2 + (a_low_end - an_other_direction_value) ^ 2)
				else
					between_ends := False
					diff := sqrt ((an_edge - a_value) ^ 2 + (a_high_end - an_other_direction_value) ^ 2)
				end
				if diff < minimum_distance then
					closest_visible_element := an_element
					minimum_distance := diff
					closest_edge_view_position := an_edge
					closest_edge_model_position := a_real_edge
					closest_edge_low_end := a_real_low_end
					closest_edge_high_end := a_real_high_end
					is_vertical_edge := is_vert
					is_point_between_ends := between_ends
					is_corner_visible := False
					low_dist := (a_low_end - an_other_direction_value).abs
					high_dist := (a_high_end - an_other_direction_value).abs
					if (not between_ends or low_dist <= corner_threshold) and a_low_end >= 0 and a_low_end <= a_visible_limit then
						if is_vert then
							closest_corner_model_position := a_real_high_end
						else
							closest_corner_model_position := a_real_low_end
						end
						is_corner_visible := True
					end
					if (not between_ends or high_dist <= corner_threshold) and a_high_end >= 0 and a_high_end <= a_visible_limit and high_dist < low_dist then
						if is_vert then
							closest_corner_model_position := a_real_low_end
						else
							closest_corner_model_position := a_real_high_end
						end
						is_corner_visible := True
					end
				end
			end
		end

	check_intersection (an_element: RETICLE_VIEW_ELEMENT a_real_edge, a_real_low_end, a_real_high_end, a_clicked_point: DOUBLE an_edge, a_low_end, a_high_end, a_value, an_edge_limit, a_visible_limit: INTEGER is_vert: BOOLEAN)
		local
			diff: DOUBLE
		do
			if an_edge >= 0 and an_edge <= an_edge_limit and a_low_end <= a_visible_limit and a_high_end >= 0 then
				-- At least some part of edge is visible
				if a_real_edge >= closest_edge_low_end and a_real_edge <= closest_edge_high_end and closest_edge_model_position >= a_real_low_end and closest_edge_model_position <= a_real_high_end then
					-- Edge intersects previously found
					-- closest edge
					diff := (a_real_edge - a_clicked_point).abs
					if diff < minimum_intersection_distance then
						found_intersection := True
						minimum_intersection_distance := diff
						minimum_intersection_view_distance := (an_edge - a_value).abs
						closest_intersecting_edge_model_position := a_real_edge
					end
				end
			end
		end

	clicked_point: CONST_POINT
			-- Model coordinates of point on which user clicked
	
	minimum_distance: DOUBLE
			-- Distance to closest edge found so far, in pixels
			-- (first pass)
	
	minimum_intersection_distance: DOUBLE
			-- Distance along closest edge to nearest intersection
			-- found so far, in model coordinates (second pass)
	
	minimum_intersection_view_distance: INTEGER
			-- Distance along closest edge to nearest intersection
			-- found so far, in pixels
	
	found_intersection: BOOLEAN
			-- Have we found an edge (of opposite orientation)
			-- intersecting the closest edge to point where user
			-- clicked?
	
	closest_edge_view_position: INTEGER
			-- Position of closest edge found so far, in
			-- view coordinates
	
	closest_edge_model_position: DOUBLE
			-- Position of closest edge found so far, in
			-- model coordinates
	
	closest_intersecting_edge_model_position: DOUBLE
			-- Position of edge that intersects edge
			-- closest to point where user clicked, in
			-- model coordinates
	
	closest_edge_low_end: DOUBLE
			-- Low end of closest edge found so far, in
			-- model coordinates
	
	closest_edge_high_end: DOUBLE
			-- High end of closest edge found so far, in
			-- model coordinates
	
	closest_corner_model_position: DOUBLE
			-- Position of closest visible corner for closest
			-- edge found so far, in model coordinates, if
			-- `is_corner_visible'
	
	is_vertical_edge: BOOLEAN
			-- Is closest edge found so far a vertical edge?
	
	is_corner_visible: BOOLEAN
			-- Does closest edge found so far have a corner that
			-- is visible?
	
	is_point_between_ends: BOOlEAN
			-- Is point user clicked on between ends of closest
			-- edge found so far?

	closest_visible_element: RETICLE_VIEW_ELEMENT
			-- Element with visible edge which is closest to the
			-- point user clicked on

end
