class CONST_RECTANGLE
	-- a Manhattan rectangle

inherit
	MORE_DOUBLE_MATH
	export {NONE} all
	redefine out
	end;

	GEOMETRY_RESOLUTION_ACCESS
	redefine out
	end;

create 
	make_corners, make_by_corners, 
	make_size_and_center, make_by_size_and_center,
	make_copy

feature -- the size and center

	center_x: DOUBLE

	center_y: DOUBLE

	center: POINT is
	do  create Result.make_xy(center_x, center_y)
	end;

	size_x: DOUBLE

	size_y: DOUBLE

	size: POINT is
	do  create Result.make_xy(size_x, size_y)
	end;

feature -- the four edge coordinates

	low_x: DOUBLE is
	do  Result := center_x - size_x/2
	end;

	low_y: DOUBLE is
	do  Result := center_y - size_y/2
	end;

	high_x: DOUBLE is
	do  Result := center_x + size_x/2
	end;

	high_y: DOUBLE is
	do  Result := center_y + size_y/2
	end;

feature -- the four corners as points

	ll: POINT is
	do  create Result.make_xy(low_x, low_y)
	end;

	ul: POINT is
	do  create Result.make_xy(low_x, high_y)
	end;

	lr: POINT is
	do  create Result.make_xy(high_x, low_y)
	end;

	ur: POINT is
	do  create Result.make_xy(high_x, high_y)
	end;

feature	-- the four sides as line_segments, each heading ccw

	top_seg: LINE_SEGMENT is
	do  create Result.make_xy(high_x, high_y, low_x, high_y)
	end;

	left_seg: LINE_SEGMENT is
	do  create Result.make_xy(low_x, high_y, low_x, low_y)
	end;

	bottom_seg: LINE_SEGMENT is
	do  create Result.make_xy(low_x, low_y, high_x, low_y)
	end;

	right_seg: LINE_SEGMENT is
	do  create Result.make_xy(high_x, low_y, high_x, high_y)
	end;

feature	-- conversion to an equivalent polygon

	to_polygon: POLYGON is
	do  create Result.make(4);
	    Result.extend_xy( low_x,  low_y);	-- ll
	    Result.extend_xy(high_x,  low_y);	-- lr
	    Result.extend_xy(high_x, high_y);	-- ur
	    Result.extend_xy( low_x, high_y);	-- ul
	ensure Result.count = 4
	end;

feature -- basic attributes

	is_empty: BOOLEAN
		-- area is 0 
	do  Result := (size_y = 0) or (size_x = 0)
	end;

	area: DOUBLE
	do  Result := size_y * size_x
	end;

feature -- 1 rectangle computations

	contains(p: VIRTUAL_POINT): BOOLEAN is
		-- is p within me (including on the edge)?
	require have_point: p /= Void
	local  px, py: DOUBLE;
	do  px := p.x;  py := p.y;
	    Result := gres.le(low_x, px) and
		      gres.le(px, high_x) and
		      gres.le(low_y, py) and
		      gres.le(py, high_y)
	end;

	truly_contains(p: VIRTUAL_POINT): BOOLEAN is
		-- is p within my interior (not on an edge)?
	require have_point: p /= Void
	local  px, py: DOUBLE;
	do  px := p.x;  py := p.y;
	    Result := gres.lt(low_x, px) and
		      gres.lt(px, high_x) and
		      gres.lt(low_y, py) and
		      gres.lt(py, high_y)
	end;

	sized(dx, dy: DOUBLE): like Current is
		-- new box is me grown by dx, dy (mebes onesided sizing)
	require size_x + dx >= 0; size_y + dy >= 0;
	do  create Result.make_by_size_and_center(size_x + dx, size_y + dy,
					center_x, center_y)
	ensure	Result.size_x = dx + size_x;
		Result.size_y = dy + size_y
	end;

	at(m: MATRIX): like Current is
		-- produce box after being transformed by m
	require m.is_Manhattan
	local  lx, ly, hx, hy: DOUBLE
	do  -- since m is Manhattan, we can just do two opposite corners
	    --  to determine the new rectangle
	    lx := m.transformed_X(low_x, low_y);
	    ly := m.transformed_Y(low_x, low_y);
	    hx := m.transformed_X(high_x, high_y);
	    hy := m.transformed_Y(high_x, high_y);
	    create Result.make_by_corners(lx, ly, hx, hy)
		-- the edges will be resorted into low and high
	ensure Result /= Void
	end;

feature	-- computations with a line segment

	contains_seg(seg: CONST_LINE_SEGMENT): BOOLEAN is
		-- is seg within me (including touching the edge)?
	require have_seg: seg /= Void
	do  Result := contains(seg.pt1) and contains(seg.pt2)
	end;

	contained_seg(seg: CONST_LINE_SEGMENT): LINE_SEGMENT is
		-- the sub-segment of seg that lies within me:
		--  if seg does not touch me, then Void
		--  otherwise, a new segment (in the same direction as seg)
	require have_seg: seg /= Void
	local fracs: VALUE_INTERVAL[DOUBLE]
	do  -- fracs are fractions within seg that mark the ends of
	    --  the contained_seg
	    create fracs.make_empty;
	    -- look for contained ends
	    if contains(seg.pt1) then fracs.extend(0) end;
	    if contains(seg.pt2) then fracs.extend(1) end;
	    -- now look for edge crossings
	    if not seg.is_horizontal
	     then check_one_edge(seg, top_seg, fracs)
		  check_one_edge(seg, bottom_seg, fracs)
	    end;
	    if not seg.is_vertical
	     then check_one_edge(seg, left_seg, fracs)
		  check_one_edge(seg, right_seg, fracs)
	    end;
	    -- this is it
	    if not fracs.is_empty
	     then create Result.make(seg.point_at_fraction(fracs.low),
				     seg.point_at_fraction(fracs.high))
	     else -- Void
	    end
	end;

feature {NONE} -- helper for contained_seg

	check_one_edge(seg, edge: CONST_LINE_SEGMENT;
			fracs: VALUE_INTERVAL[DOUBLE]) is
	local sfr, efr: DOUBLE
	do  sfr := seg.fraction_at_intersection(edge);
	    efr := edge.fraction_at_intersection(seg);
	    if 0 <= sfr and sfr <= 1 and edge.is_within_segment(efr)
	     then fracs.extend(sfr)
	    end;
	end;

feature -- 2 rectangle computations

	is_same_box(other: CONST_RECTANGLE): BOOLEAN is
		-- tests for same rectangle by coordinates
	do  if other /= Void then
	      Result :=	gres.eq(center_x, other.center_x) and
			gres.eq(center_y, other.center_y) and
			gres.eq(size_x, other.size_x) and
			gres.eq(size_y, other.size_y)
	    end
	end;

	is_in_list(lst: LIST[CONST_RECTANGLE]): BOOLEAN
		-- am I is_same_box for any element of lst?
		-- if true, first matching box is lst.item
	require have_list: lst /= Void
	do  from  lst.start
	    until lst.after or Result
	    loop  if is_same_box(lst.item) then
			Result := true
		  else	lst.forth
		  end
	    end;
	    ensure is_item: Result implies is_same_box(lst.item)
	end;

	touches(other: CONST_RECTANGLE): BOOLEAN is
		-- rectangles touch or overlap
		-- (intersection_with exists)
	local lx, ly, hx, hy: DOUBLE;
	do  lx := low_x.max(other.low_x);
	    ly := low_y.max(other.low_y);
	    hx := high_x.min(other.high_x);
	    hy := high_y.min(other.high_y);
	    Result := gres.ge(hx, lx) and gres.ge(hy, ly)
	end;
	
	intersects(other: CONST_RECTANGLE): BOOLEAN is
		-- true overlap of the rectangles
		-- (intersection_with has non-zero area)
	local lx, ly, hx, hy: DOUBLE;
	do  lx := low_x.max(other.low_x);
	    ly := low_y.max(other.low_y);
	    hx := high_x.min(other.high_x);
	    hy := high_y.min(other.high_y);
	    Result := gres.gt(hx, lx) and gres.gt(hy, ly)
	end;

	intersection_with(other: CONST_RECTANGLE): like Current is
		-- new rectangle with the intersection
		-- (Result has zero area if the two rectangles just touch;
		--  Void if not even that)
	require other /= Void
	local lx, ly, hx, hy: DOUBLE;
	do  lx := low_x.max(other.low_x);
	    ly := low_y.max(other.low_y);
	    hx := high_x.min(other.high_x);
	    hy := high_y.min(other.high_y);
	    if gres.ge(hx, lx) and gres.ge(hy, ly)
	     then create Result.make_by_corners(lx, ly, hx, hy)
	    end;
	end;

	encloses(other: CONST_RECTANGLE): BOOLEAN is
		-- other is within me (can touch edges)
	do  Result :=	gres.le(low_x, other.low_x) and
			gres.le(low_y, other.low_y) and
			gres.le(other.high_x, high_x) and
			gres.le(other.high_y, high_y)
	end;

	truly_encloses(other: CONST_RECTANGLE): BOOLEAN is
		-- other is within me (cannot even touch edges)
	do  Result :=	gres.lt(low_x, other.low_x) and
			gres.lt(low_y, other.low_y) and
			gres.lt(other.high_x, high_x) and
			gres.lt(other.high_y, high_y)
	end;

	union_with(other: CONST_RECTANGLE): like Current is
		-- new rectangle which is smallest rectangle that
		-- can contain both me and other
	require other /= Void
	do  create Result.make_by_corners(
			low_x.min(other.low_x), low_y.min(other.low_y),
			high_x.max(other.high_x), high_y.max(other.high_y))
	ensure Result /= Void
	end;

	minus(other: CONST_RECTANGLE): LIST[CONST_RECTANGLE] is
		-- set of rectangles (up to 4) that collectively cover
		--  all of Current that is _outside_ of other
		-- (returns Current if not intersects(other))
		-- (returns empty list if Current is within other)
	require other /= Void
	local xsect, foo: CONST_RECTANGLE
	do  create {LINKED_LIST[RECTANGLE]} Result.make;
	    xsect := intersection_with(other);
	    if xsect = Void or else xsect.is_empty
	     then -- return all of me
		Result.extend(Current)
	     else -- return parts to left, above, to right, and below xsect
		if gres.lt(low_x, xsect.low_x) then
		    create foo.make_by_corners
			(low_x, low_y, xsect.low_x, xsect.high_y);
		    Result.extend(foo)
		end;
		if gres.gt(high_y, xsect.high_y) then
		    create foo.make_by_corners
			(low_x, xsect.high_y, xsect.high_x, high_y);
		    Result.extend(foo)
		end;
		if gres.gt(high_x, xsect.high_x) then
		    create foo.make_by_corners
			(xsect.high_x, xsect.low_y, high_x, high_y);
		    Result.extend(foo)
		end;
		if gres.lt(low_y, xsect.low_y) then
		    create foo.make_by_corners
			(xsect.low_x, low_y, high_x, xsect.low_y);
		    Result.extend(foo)
		end;
	    end
	end;

	minus_all(lst: LIST[CONST_RECTANGLE]): LIST[CONST_RECTANGLE]
		-- Successively remove each element of lst from me, or
		--  from what is left of me, returning a set of rectangles
		--  that cover the inverse within me of the coverage of lst.
	require have_list: lst /= Void
	do  create {LINKED_LIST[CONST_RECTANGLE]} Result.make;
	    Result.extend(Current)
	    from  lst.start
	    until lst.after
	    loop  Result := lst.item.remove_from_all(Result)
		  lst.forth
		  end;
	    end

	remove_from_all(lst: LIST[CONST_RECTANGLE]): LIST[CONST_RECTANGLE]
		-- Remove (minus) me from each element of lst,
		-- then merge the Results, removing duplicates.
		-- Result collectively covers all of lst outside of me.
		-- Note that in minus and minus_all, Current is the whole
		--  and the argument is the subtrahend,
		--  but here Current is the subtrahend!
	require have_list: lst /= Void
	local  itm: CONST_RECTANGLE; res: LIST[CONST_RECTANGLE]
	do  create {ARRAYED_LIST[CONST_RECTANGLE]} Result.make(4*lst.count)
	    from  lst.start
	    until lst.after
	    loop  itm := lst.item;
		  if itm /= Void then
			res := itm.minus(Current);
			from  res.start
			until res.after
			loop  itm := res.item;
			      if not itm.is_in_list(Result) then
				Result.extend(itm)
			      end;
			      res.forth
			end
		  end;
		  lst.forth
	    end;
	end;

	moved_into(other: CONST_RECTANGLE): MATRIX is
		-- the matrix that scales and translates me to fit
		-- "nicely" into other
	require have_other: other /= Void;
		have_space: not other.is_empty
	local scale: DOUBLE
	do  -- find scale for best (max) fit
	    if size_x = 0 and size_y = 0
	     then scale := 1	-- cannot scale, so skip it
	    elseif size_y = 0
	     then scale := other.size_x / size_x
	    elseif size_x = 0
	     then scale := other.size_y / size_y
	     else scale := (other.size_x / size_x).min(other.size_y / size_y)
	    end;
	    -- and do it (scale and translate)
	    create Result.make_arbitrary(scale, 0, 0, scale, 
				other.center_x - scale * center_x,
				other.center_y - scale * center_y)
	end;

feature -- display

	out: STRING is
		-- depict as two corners (ll and ur)
	do  create Result.make(80);
	    Result.append(LL_seg);  Result.append(ll.out);
	    Result.append(UR_seg);  Result.append(ur.out);
	    Result.append(close)
	end;

	Cats_out(fmt: FORMAT_DOUBLE): STRING is
		-- "(<low_x>,<low_y>) (<high_x>,<high_y>)"
		-- use fmt if provided; otherwise use .out
	do  create Result.make(50);
	    Result.extend('(');   Result.append(dbl_to_str(low_x, fmt));
	    Result.extend(',');   Result.append(dbl_to_str(low_y, fmt));
	    Result.append(") ("); Result.append(dbl_to_str(high_x, fmt));
	    Result.extend(',');   Result.append(dbl_to_str(high_y, fmt));
	    Result.extend(')')
	end;

	out_SC: STRING is
		-- depict as center and size
	do  create Result.make(80);
	    Result.append(SIZE_seg);  Result.append(size.out);
	    Result.append(CENT_seg);  Result.append(center.out);
	    Result.append(close)
	end;

feature	{NONE} -- stuff for the above

	LL_seg:   STRING is "LL: ("
	UR_seg:   STRING is "), UR: ("
	SIZE_seg: STRING is "SIZE: ("
	CENT_seg: STRING is "), CENT: ("
	close:    STRING is ")"

	dbl_to_str(dbl: DOUBLE; fmt: FORMAT_DOUBLE): STRING is
		-- use fmt or use .out
	do  if fmt /= Void
	     then Result := fmt.formatted(dbl)
	     else Result := dbl.out
	    end
	end;

feature {NONE} -- creation

	make_copy(r: CONST_RECTANGLE) is
		-- make a copy of r
	require present: r /= Void
	do  make_by_size_and_center(r.size_x, r.size_y, r.center_x, r.center_y)
	ensure is_same_box: is_same_box(r)
	end;

	make_corners(crnr1, crnr2: VIRTUAL_POINT) is
		-- one corner is crnr1; the other is crnr2
		-- ANY two diagonally opposed corners work just fine!
	require present: crnr1 /= Void and crnr2 /= Void
	do  make_by_corners(crnr1.x, crnr1.y, crnr2.x, crnr2.y)
	end;

	make_by_corners(x1, y1, x2, y2: DOUBLE) is
		-- one corner is (x1, y1); other is (x2, y2)
		-- ANY two diagonally opposed corners work just fine!
	do  center_x := (x1 + x2)/2;
	    center_y := (y1 + y2)/2;
	    if gres.eq(x1, x2) then
		size_x := 0
	    else
		size_x := (x1 - x2).abs
	    end;
	    if gres.eq(y1, y2) then
		size_y := 0
	    else
		size_y := (y1 - y2).abs
	    end;
	ensure	match_x: (gres.eq(low_x, x1) and gres.eq(high_x, x2)) or
			 (gres.eq(low_x, x2) and gres.eq(high_x, x1))
		match_y: (gres.eq(low_y, y1) and gres.eq(high_y, y2)) or
			 (gres.eq(low_y, y2) and gres.eq(high_y, y1))
	end;

	make_size_and_center(siz, cen: VIRTUAL_POINT) is
		-- size is siz; center point is cen
	require present: siz /= Void and cen /= Void;
		non_neg: siz.x >= 0 and siz.y >= 0;
	do  make_by_size_and_center(siz.x, siz.y, cen.x, cen.y)
	ensure	same_size: size.is_same_point(siz)
		same_center: center.is_same_point(cen)
	end;

	make_by_size_and_center(sx, sy, cx, cy: DOUBLE) is
		-- size is (sx, sy); center point is (cx, cy)
	require non_neg: sx >= 0 and sy >= 0;
	do  
	    size_x   := sx;  size_y   := sy;
	    center_x := cx;  center_y := cy;
	ensure
	    gres.eq(size_x,   sx);  gres.eq(size_y,   sy);
	    gres.eq(center_x, cx);  gres.eq(center_y, cy);
	end;

invariant
	size_x >= 0;
	size_y >= 0

end -- CONST_RECTANGLE
