
indexing
	
	description: "A run closing step operation that asks the %
		%run closer to make sure that all SCMOS layout has %
		%been converted to native rules, if a GDS merge is required"
	author: "David Hollenberg"

class CONVERT_SCMOS_LAYOUT_OP
inherit
	RC_PRIMITIVE_STEP_OP
	NO_USER_INTERFACE
	NO_EXTERNAL_STATE

create
	make

feature -- Properties

	input_data_items: LIST [STRING]
			-- Data items required before step can be executed
		do
			Result := array_as_list (<<
				data_item.Test_vehicle_fab_instances
				>>)
		end

feature {NONE} -- Implementation

	is_applicable: BOOLEAN
			-- Is step applicable to run for session?
		do
			Result := cache.run_needs_merged_gds_files
		end

	do_step
		local
			msg, full_msg: STRING
			list: like scmos_dies_to_merge
			n: INTEGER
		do
			uncache_changeable_data
			create msg.make (200)
			list := scmos_dies_to_merge
			if not list.is_empty then
				n := list.count
				msg.append ("Found " + n.out + " " + su.number (n, "die", "dies") + " that almost certainly " + su.number (n, "needs", "need") + " to be converted from SCMOS to native rules:%N%N")
				list.do_all (agent append_die (?, msg))
				full_msg := msg.twin
				full_msg.append ("%NHave these dies been converted to native rules?%N%N")
				full_msg.append ("Click on No to notify vendor engineering wizard to do conversions.%N")
				full_msg.append ("Click on Yes when conversions to native rules have been done.%N")
				if not question (full_msg) then
					msg.append ("%NPlease reply to this message when the conversions have been done.%N")
					publish_scmos_conversion (msg)
					inform ("Vendor engineering wizard notified about required SCMOS to native rule conversions")
					abort
				end
				verify_conversion
				uncache_changeable_data
			end
			call_completion_actions
		end

	undo_step
			-- Undo execution of step
		do
			uncache_changeable_data
			call_completion_actions
		end

	uncache_changeable_data
			-- Remove from cache things which might change
			-- due to conversion
		do
			uncache_scmos_dies_to_merge
			cache.uncache_dies
			cache.uncache_die_locations
		end

	verify_conversion
			-- Verify that all dies that need to be
			-- converted to native rules have been
			-- converted
		local
			msg: STRING
			list: like scmos_dies_to_merge
			n: INTEGER
		do
			uncache_scmos_dies_to_merge
			cache.die_rows.do_all (agent {DIES_ROW}.uncache_attributes)
			create msg.make (200)
			list := scmos_dies_to_merge
			n := list.count
			msg.append (n.out + " " + su.number (n, "die", "dies") + " still " + su.number (n, "needs", "need") + " to be converted from SCMOS to native rules:%N%N")
			list.do_all (agent append_die (?, msg))
			msg.append ("%N")
			msg.append ("If you proceed, we will send designs in non-native rules to the foundry.%N")
			msg.append ("Do NOT proceed unless a vendor engineering wizard says it is OK.%N%N")
			msg.append ("Proceed anyway?")
			if not question (msg) then
				abort
			end
		end

	append_die (die: DIES_ROW dest: STRING)
			-- Append description of `die' to `dest'
		do
			dest.append ("    " + su.right_padded (die.description, 6) + "   " + die.die_id + "   " + die.attributes.design_techn + "%N")
		end

	scmos_dies_to_merge: LIST [DIES_ROW]
			-- List of dies in non-native design technology
			-- that need to be merged
		do
			if cached_scmos_dies_to_merge = Void then
				fetch_scmos_dies_to_merge
			end
			Result := cached_scmos_dies_to_merge
		end

	fetch_scmos_dies_to_merge
			-- Fetch list of dies that are not in a
			-- native design techn and that need to
			-- be merged and put in `scmos_dies_to_merge'
		local
			list_filter: LIST_UTILITIES [DIES_ROW]
			list: like scmos_dies_to_merge
			die: DIES_ROW
			attr: DIE_ATTRIBUTES_ROW
			dtwt_row: DESIGN_TECHNXWAFER_TECHNS_ROW
		do
			create list_filter
			-- Start with all dies
			list := cache.die_rows
			if cache.is_die_merge then
				-- Keep only dies that are projects
				list := list_filter.items_passing_filter (list, agent {DIES_ROW}.is_project)
			else
				-- Keep only dies that are not vendor
				-- ghosts, that are internal to payload
				-- and that are not scribes
				list := list_filter.items_passing_filter (list, agent {DIES_ROW}.is_not_vendor_ghost)
				list := list_filter.items_passing_filter (list, agent {DIES_ROW}.is_internal_to_payload)
				list := list_filter.items_passing_filter (list, agent {DIES_ROW}.is_non_scribe)
			end
			-- Keep only dies that are not in
			-- a native design techn
			
			create {ARRAYED_LIST [DIES_ROW]} cached_scmos_dies_to_merge.make (list.count)
			from
				list.start
			until
				list.after
			loop
				die := list.item
				attr := die.attributes
				create dtwt_row.make_template
				dtwt_row := dtwt_row.row_for_dt_and_wt (attr.design_techn, cache.wafer_techn, connection, expected_count.Exactly_one)
				if not dtwt_row.is_native_design_techn then
					-- SCMOS die (not in native design techn)
					cached_scmos_dies_to_merge.extend (die)
				end
				list.forth
			end
		end

	uncache_scmos_dies_to_merge
			-- Remove `scmos_dies_to_merge' from cache
		do
			cached_scmos_dies_to_merge := Void
		end

	publish_scmos_conversion (msg: STRING)
			-- Publish an "SCMOS conversion required" event
		local
			e: SCMOS_CONVERSION_REQUIRED_EVENT
		do
			create e
			e.set_subject ("SCMOS conversions")
			e.set_message ("SCMOS conversions required for run " + session.run_id + ".%N%N" + msg)
			e.publish
		end

	set_state
			-- Set `state'
		do
			state := Void
		end

	cached_scmos_dies_to_merge: LIST [DIES_ROW]
			-- List of dies in non-native design technology
			-- that need to be merged, cached
			-- in memory
			
end
