
indexing
	
	description: "A run closing step operation that creates a %
		%work order line items, packagings and charges for each %
		%test vehicle on run"
	author: "David Hollenberg"
	date: "February 27, 2004"

class CREATE_TEST_VEHICLE_ORDERS_OP
inherit
	CREATE_TEST_VEHICLE_DATA_OP

create
	make

feature -- Properties

	input_data_items: LIST [STRING]
			-- Data items required before step can be executed
		do
			Result := array_as_list (<<
				data_item.Test_vehicle_fab_instances,
				data_item.Initial_wafer_lots,
				data_item.Cut_patterns
				>>)
		end

feature {NONE} -- Implementation

	do_step
			-- Execute of step
		do
			create_test_vehicle_orders
			call_completion_actions
		end

	undo_step
			-- Undo execution of step
		do
			delete_test_vehicle_orders
			call_completion_actions
		end

	create_test_vehicle_orders
			-- Create work order line items, packagings and
			-- charges for each test vehicle on run
		local
			list: LIST [DIES_ROW]
			tv: DIES_ROW
			ifi: IC_FAB_INSTANCES_MROW
		do
			order_date := connection.transaction_date
			fetch_isolated_die_ids
			list := cache.test_vehicle_die_rows
			from
				list.start
			until
				list.after
			loop
				tv := list.item
				fetch_test_vehicle (tv)
				fab_instance := tv.ic_fab_instance
				m_fab_no := fab_instance.m_fab_no
				create_test_vehicle_order (tv)

				-- Update quantity requested in fab
				-- instance so that it will match IC
				-- quantity in line items, to avoid
				-- spurious unresolved issue warning
				-- int later step.  No need to undo
				-- this change when undo step

				create ifi.make_by_key (connection, option.Old_row, m_fab_no)
				ifi.set_quantity_requested (quantity_ordered)
				list.forth
			end
		end

	create_test_vehicle_order (tv: DIES_ROW)
			-- Create an order for `tv'
		require
			test_vehicle_exists: tv /= Void
			is_test_vehicle: tv.is_test_vehicle
		do
			fetch_cust_acct_no (tv)
			fetch_funding_info (cust_acct_no)
			fetch_quantities (tv)
			wo_line_item_number := 0
			
			m_order_no := fab_instance.work_order.m_order_no
			if quantity_ordered > 0 then
				create_wo_line_item (tv, test_vehicle.m_part_no_ic, quantity_ordered)
				create_cust_charge (tv, last_job_no)
			end
			create_packaging_line_item (tv, test_vehicle.m_part_no_package_isolated, quantity_packaged)
			create_packaging_line_item (tv, test_vehicle.m_part_no_package_unisolated, quantity_unpackaged)
		end

	fetch_isolated_die_ids
			-- Fetch die ids of test vehicle dies that are
			-- isolated by at least one cut pattern and
			-- put in `isolated_die_ids'
		
		local
			b: DB_BIND_SS
			die_type: DIE_TYPE_VALUES
		do
			create b.make (session.run_id, die_type.Test_vehicle)
			isolated_die_ids := connection.selected_strings (
				"d.die_id",
				"cut_die_sites cds, die_locations dl, dies d",
				"cds.run_id = d.run_id and %N%
				%d.run_id = dl.run_id and %N%
				%cds.mask_site_no = dl.mask_site_no and %N%
				%cds.run_id = :1 and %N%
				%d.die_type = :2",
				Void, False, b)
		end

	fetch_quantities (tv: DIES_ROW)
			-- Set `quantity_ordered' for `tv'
		require
			test_vehicle_exists: tv /= Void
			is_test_vehicle: tv.is_test_vehicle
		local
			list: LIST [DIEXCUT_QUANTITIES_ROW]
			row: DIEXCUT_QUANTITIES_ROW
			quantity_cut: INTEGER
			is_isolated: BOOLEAN
		do
			create row.make_template
			list := row.rows_for_run_die (session.run_id, tv.die_id, connection, expected_count.At_most_one)
			if list.is_empty then
				quantity_cut := 0
			else
				quantity_cut := list.first.quantity_cut_die
			end
			quantity_ordered := quantity_cut.min (max_test_vehicle_quantity)
			is_isolated := isolated_die_ids.has (tv.die_id)
			if not is_isolated or test_vehicle.m_part_no_package_isolated = test_vehicle.m_part_no_package_unisolated then
				quantity_packaged := 0
			else
				quantity_packaged := quantity_ordered.min (max_test_vehicle_packaged_quantity)
			end
			quantity_unpackaged := quantity_ordered - quantity_packaged
		end

	create_wo_line_item (tv: DIES_ROW m_part_no, quantity: INTEGER)
			-- Create a work order line item for `tv', with
			-- part number `m_part_no' to order `quantity' parts.
			-- Use next line number
			-- in sequence.  Put job number in `last_job_no'
			-- and put created line item in `last_wo_line_item'
		require
			test_vehicle_exists: tv /= Void
			is_test_vehicle: tv.is_test_vehicle
		local
			sequence: SEQUENCE_NAMES
			woli: CUST_WO_LINE_ITEMS_MROW
		do
			wo_line_item_number := wo_line_item_number + 1
			last_job_no := connection.sequence_next_value (sequence.M_job_no_seq)
			create woli.make_null (connection)
			woli.set_m_job_no (last_job_no)
			woli.set_m_order_no (m_order_no)
			woli.set_cust_wo_line_no (wo_line_item_number)
			woli.set_m_part_no (m_part_no)
			woli.set_part_quantity_ordered (quantity)
			woli.set_cust_wo_line_comment ("Run " + session.run_id + ", test vehicle " + tv.test_vehicle_design_no.out)
			last_wo_line_item := woli
		end

	create_packaging_line_item (tv: DIES_ROW m_part_no, quantity: INTEGER)
			-- Create a work order line item for packaging
			-- `tv', with part number `m_part_no' to order
			-- `quantity' parts, if `quantity' is greater
			-- than 0.  Use next line number in sequence.
			-- Also, create a packaging and put its number
			-- in the packaging line item
		require
			test_vehicle_exists: tv /= Void
			is_test_vehicle: tv.is_test_vehicle
			quantity_nonnegative: quantity >= 0
		do
			if quantity > 0 then
				create_packaging (tv, m_part_no, quantity)
				create_wo_line_item (tv, m_part_no, quantity)
				last_wo_line_item.set_ic_packaging_no (last_ic_packaging_no)
				create_cust_charge (tv, last_job_no)
			end
		end

	create_cust_charge (tv: DIES_ROW m_job_no: INTEGER)
			-- Create a cust charge for `tv', with job
			-- number m_job_no
		require
			test_vehicle_exists: tv /= Void
			is_test_vehicle: tv.is_test_vehicle
		local
			sequence: SEQUENCE_NAMES
			cust_charge_no: INTEGER
			charge: CUST_CHARGES_MROW
		do
			cust_charge_no := connection.sequence_next_value (sequence.M_cust_charge_no_seq)
			create charge.make_null (connection)
			charge.set_m_cust_charge_no (cust_charge_no)
			charge.set_m_adjustment_no (0)
			charge.set_m_job_no (m_job_no)
			charge.set_m_funding_alloc_no (m_funding_alloc_no)
			charge.set_m_funding_auth_no (m_funding_auth_no)
			charge.set_distribution_rank (distribution_rank)
			charge.set_date_entered (order_date)
			charge.set_amount_total_no_tax (0.0)
			charge.set_amount_tax (0.0)
			charge.set_edu_units (0)
		end

	create_packaging (tv: DIES_ROW m_part_no, quantity: INTEGER)
			-- Create an IC packaging for `tv' using
			-- `m_part_no' for `quantity' parts.  Leave
			-- packaging number in `last_ic_packaging_no'
		require
			test_vehicle_exists: tv /= Void
			is_test_vehicle: tv.is_test_vehicle
			quantity_nonnegative: quantity >= 0
		local
			sequence: SEQUENCE_NAMES
			packaging: IC_PACKAGINGS_MROW
			package_id, lid_attachment: STRING
			parts_row: M_PARTS_ROW
			packages_row: PACKAGES_ROW
			packages: LIST [PACKAGES_ROW]
			lead_frame: BOOLEAN
			defaults_row: IC_PACKAGING_DEFAULTS_ROW
			ic_fab_instance_type: IC_FAB_INSTANCE_TYPE_VALUES
		do
			create parts_row.make_template
			package_id := parts_row.parts_with_number (m_part_no, connection, expected_count.Exactly_one).first.package_id
			
			create packages_row.make_template
			packages := packages_row.packages_with_id (package_id, connection, expected_count.At_most_one)
			if packages.is_empty then
				lead_frame := False
			else
				lead_frame := packages.first.leadframe
			end
			
			create defaults_row.make_template
			lid_attachment := defaults_row.rows_for_class_and_leadframe (ic_fab_instance_type.Test_vehicle, lead_frame, connection, expected_count.Exactly_one).first.lid_attachment
			
			last_ic_packaging_no := connection.sequence_next_value (sequence.Ic_packaging_no_seq)
			create packaging.make_null (connection)
			packaging.set_ic_packaging_no (last_ic_packaging_no)
			packaging.set_m_fab_no (m_fab_no)
			packaging.set_customer_providing_diagram (False)
			packaging.set_quantity_packaged_requested (quantity)
			packaging.set_lid_attachment (lid_attachment)
			packaging.set_bake_and_bag (False)
			packaging.set_package_id (package_id)
			packaging.set_design_package_rotation (test_vehicle.design_package_rotation)
			packaging.set_diagram_approved_by (test_vehicle.diagram_approved_by)
			packaging.set_date_diagram_approved (test_vehicle.date_diagram_approved)
		end

	delete_test_vehicle_orders
			-- Delete all line items, packagings and charges
			-- for test vehicles on run
		local
			breakdown_row: CUST_CHARGE_BREAKDOWNS_MROW
			charge_row: CUST_CHARGES_MROW
			line_item_row: CUST_WO_LINE_ITEMS_MROW
			packaging_row: IC_PACKAGINGS_MROW
			list: LIST [DB_DELETABLE_ROW]
		do
			-- Delete cust charge breakdowns (temporary,
			-- until there are no more un-COMPLETE runs with
			-- breakdowns)

			create breakdown_row.make_template
			list := breakdown_row.locked_test_vehicle_breakdowns_for_run (session.run_id, connection, busy_action.return, expected_count.at_least_zero)
			list.do_all (agent {DB_DELETABLE_ROW}.delete)
			
			-- Delete charges

			create charge_row.make_template
			list := charge_row.locked_test_vehicle_charges_for_run (session.run_id, connection, busy_action.return, expected_count.at_least_zero)
			list.do_all (agent {DB_DELETABLE_ROW}.delete)
		
			-- Delete work order line items

			create line_item_row.make_template
			list := line_item_row.locked_test_vehicle_line_items_for_run (session.run_id, connection, busy_action.return, expected_count.at_least_zero)
			list.do_all (agent {DB_DELETABLE_ROW}.delete)
		
			-- Delete IC packagings

			create packaging_row.make_template
			list := packaging_row.locked_test_vehicle_packagings_for_run (session.run_id, connection, busy_action.return, expected_count.at_least_zero)
			list.do_all (agent {DB_DELETABLE_ROW}.delete)
		end
	
	isolated_die_ids: LIST [STRING]
			-- Die ids for test vehicles that are isolated by
			-- at least one cut pattern

	fab_instance: IC_FAB_INSTANCES_ROW
			-- Fab instance for test vehicle being processed
	
	quantity_ordered: INTEGER
			-- Number of parts to order for test vehicle
	
	quantity_packaged: INTEGER
			-- Number of packaged parts to order for test vehicle
	
	quantity_unpackaged: INTEGER
			-- Number of unpackaged parts to order for test vehicle
	
	last_job_no: INTEGER
			-- Job number for last work order line item created

	last_wo_line_item: CUST_WO_LINE_ITEMS_MROW
			-- Last work order line item created

	last_ic_packaging_no: INTEGER
			-- Packaging number for last packaging created

	wo_line_item_number: INTEGER
			-- Number of last last work order line item created

	max_test_vehicle_quantity: INTEGER = 25
			-- Maximum number of test vehicles to order

	max_test_vehicle_packaged_quantity: INTEGER = 10
			-- Maximum number of test vehicles to package

end
