
indexing
	
	description: "Map that shows cut patterns"
	author: "David Hollenberg"
	date: "January 21, 2005"

class CUT_PATTERNS_MAP
inherit
	DETAILED_PAYLOAD_MAP
		redefine
			set_sizes,
			set_stroke_and_fill_colors,
			generate_drawing_specific_part,
			current_page_label
		end
	DOUBLE_MATH
		export
			{NONE} all
		end
	NUMBER_UTILITIES_ACCESS
		export
			{NONE} all
		end

create
	make

feature -- Properties

	title: STRING = "Cut Pattern"
			-- Title to include on map
	
	packaging_lot: RUN_PACKAGING_LOTS_ROW
			-- Packaging lot for which map is to be generated
			-- (Void for unassigned cut patterns)

	cut_patterns: LIST [RUN_CUT_PATTERNS_ROW]
			-- Cut patterns to be included in map,
			-- one per page, in order of inclusion

	is_draft: BOOLEAN
			-- Is map a draft (i.e., is some title information
			-- not available)?
		do
			Result := packaging_lot = Void or else packaging_lot.map_wafer_lot_ids.is_equal (Unknown_value) or else packaging_lot.is_null_die_thickness_mils
		end

feature -- Modification

	set_cut_patterns (list: like cut_patterns)
		require
			list_not_void: list /= Void
		do
			cut_patterns := list
		ensure
			value_set: cut_patterns = list
		end

	set_packaging_lot (lot: like packaging_lot)
		do
			packaging_lot := lot
		ensure
			value_set: packaging_lot = lot
		end

feature -- Operations

	generate_page_for_cut_pattern (pat: RUN_CUT_PATTERNS_ROW)
			-- Generate page for cut pattern `pat' with
			-- both drawing and title and put in
			-- `last_page'
		require
			pat_not_void: pat /= Void
		do
			set_cut_pattern (pat)
			generate_page
		end

	generate_drawing_for_cut_pattern (pat: RUN_CUT_PATTERNS_ROW)
			-- Generate drawing for cut pattern `pat' and put
			-- in `last_drawing'
		require
			pat_not_void: pat /= Void
		do
			set_cut_pattern (pat)
			generate_drawing
		end

	generate_title_for_cut_pattern (pat: RUN_CUT_PATTERNS_ROW)
			-- Generate title for cut pattern `pat' and put
			-- in `last_drawing'
		require
			pat_not_void: pat /= Void
		do
			set_cut_pattern (pat)
			generate_title
		end

feature {NONE} -- Implementation

	generate_whole_document
			-- Generate entire document representing map
			-- and put in `last_document'
		do
			cut_patterns.do_all (agent add_page_for_cut_pattern)
		end

	add_page_for_cut_pattern (pat: RUN_CUT_PATTERNS_ROW)
			-- Add to `last_document' a page for cut pattern `pat'
		do
			generate_page_for_cut_pattern (pat)
			last_document.add_page (last_page)
		end

	current_page_label: STRING
			-- Descriptive label for current page of map
			-- (alternative to page number)
		do
			Result := "Pattern " + cut_pattern.cut_pattern_no.out + " (field " + field_name + ")"
		end

	set_sizes
			-- Set various sizes
		do
			precursor
			original_title_fraction := title_fraction
			set_new_title_fraction
		end

	set_new_title_fraction
			-- Set new value of `title_fraction' to be used
		do
			title_fraction := title_fraction * Title_scale_factor
		end

	set_stroke_and_fill_colors
			-- Set colors used for stroke and fill
		do
			precursor
			create unisolated_die_stroke.make_with_rgb (0.5, 0.5, 0.5)
			create draft_stroke.make_with_rgb (1.0, 0.0, 0.0)
			create cut_stroke.make_with_rgb (0.0, 0.0, 1.0)
			create isolated_die_fill.make_with_rgb (0.6, 1.0, 0.6)
		end

	generate_whole_title
			-- Generate entire title and put in `last_title'
		local
			whole_box, upper_box: DRW_BASIC_RECTANGLE
			main_box, pkg_box, proc_box, thickness_box, lot_box, count_box: DRW_BASIC_RECTANGLE
			bg_thickness_box, saw_width_box: DRW_BASIC_RECTANGLE
			left_draft_box, right_draft_box: DRW_BASIC_RECTANGLE
			box: DRW_RECTANGLE
			text: DRW_TEXT
			w, h, old_tf: DOUBLE
			t, bg_thickness, saw_width: STRING
			pkg_lot: like packaging_lot
		do
			-- Add invisible box that covers area
			-- for run id, cut pattern number and
			-- DRAFT label (if appropriate)
			
			title_fraction := original_title_fraction
			w := available_paper_width
			h := title_height * (Title_scale_factor - 1.0)
			create box.make (w, h, w / 2.0, title_height / 2.0 + h / 2.0)
			box.disable_stroke
			last_title.extend (box)
			upper_box := box.bounding_box.twin
			main_box := horizontal_section (upper_box, 0.2, 0.8)
			shrink_box (main_box)
			
			if is_draft then
				left_draft_box := horizontal_section (upper_box, 0.0, 0.2)
				right_draft_box := horizontal_section (upper_box, 0.8, 1.0)
				shrink_box (left_draft_box)
				shrink_box (right_draft_box)
				text := text_in_box (Draft_label, left_draft_box)
				text.set_stroke (draft_stroke)
				last_title.extend (text)
				text := text_in_box (Draft_label, right_draft_box)
				text.set_stroke (draft_stroke)
				last_title.extend (text)
			end
			
			w := available_paper_width - (title_aux_width + title_logo_width)
			create box.make (w, title_height, w / 2.0 + title_logo_width, 0)
			box.set_stroke_width (thick_width)
			last_title.extend (box)
			whole_box := box.bounding_box.twin
			
			lot_box := nth_vertical_section (box, 5, 5)
			pkg_box := nth_vertical_section (box, 4, 5)
			proc_box := nth_vertical_section (box, 3, 5)
			thickness_box := nth_vertical_section (box, 2, 5)
			count_box := nth_vertical_section (box, 1, 5)
			
			last_title.extend (thick_box (pkg_box))
			last_title.extend (thick_box (proc_box))
			last_title.extend (thick_box (lot_box))
			last_title.extend (thick_box (thickness_box))
			last_title.extend (thick_box (count_box))
			
			old_tf := text_fraction
			text_fraction := 0.6
			shrink_wide_box (pkg_box)
			shrink_wide_box (proc_box)
			shrink_wide_box (lot_box)
			shrink_wide_box (thickness_box)
			shrink_wide_box (count_box)
			text_fraction := old_tf
			
			bg_thickness_box := nth_horizontal_section (thickness_box, 1, 2)
			saw_width_box := horizontal_section (thickness_box, 0.55, 1.0)
			
			t := run_id + " " + title + " " + cut_pattern.cut_pattern_no.out
			if has_multiple_fields then
				t.append (" - Field " + field_name)
			end
			last_title.extend (text_in_box (t, main_box))
			pkg_lot := packaging_lot
			if pkg_lot /= Void then
				if pkg_lot.is_null_die_thickness_mils then
					bg_thickness := Unknown_value
				else
					bg_thickness := nu.std_formatted_number (pkg_lot.die_thickness_mils) + " mils"
				end
				if pkg_lot.is_null_saw_blade_width_mils then
					saw_width := Unknown_value
				else
					saw_width := nu.std_formatted_number (pkg_lot.saw_blade_width_mils) + " mils"
				end
				last_title.extend (left_text_in_box ("Packager:  " + pkg_lot.packager_short_id, pkg_box))
				last_title.extend (left_text_in_box ("Process:  " + pkg_lot.wafer_etest_id, proc_box))
				last_title.extend (left_text_in_box ("Backgrind thickness:  " + bg_thickness, bg_thickness_box))
				last_title.extend (right_text_in_box ("Saw blade width:  " + saw_width, saw_width_box))
				last_title.extend (left_text_in_box ("Wafer lot IDs:  " + pkg_lot.map_wafer_lot_ids, lot_box))
			else
				last_title.extend (left_text_in_box ("Packager:  " + Unknown_value, pkg_box))
				last_title.extend (left_text_in_box ("Process:  " + Unknown_value, proc_box))
				last_title.extend (left_text_in_box ("Backgrind thickness:  " + Unknown_value, bg_thickness_box))
				last_title.extend (right_text_in_box ("Saw blade width:  " + Unknown_value, saw_width_box))
				last_title.extend (left_text_in_box ("Wafer lot IDs:  " + Unknown_value, lot_box))
			end
			last_title.extend (left_text_in_box ("Wafers to cut:  " + cut_pattern.quantity_wafers_needed.out, count_box))
			
			append_mosis_logo
			append_std_title_info
			set_new_title_fraction
		end

	generate_drawing_specific_part
			-- Generate part of drawing that is specific
			-- to a particular map
		do
			precursor
			set_cut_extension
			cut_pattern.cuts.do_all (agent draw_cut)
		end

	set_cut_extension
			-- Set `cut_extension'
		local
			max_hw, len: DOUBLE
		do
			max_hw := payload_box.size_x.max (payload_box.size_y)
			cut_extension := max_hw * cut_extension_fraction
			sawtooth_length := max_hw * sawtooth_length_fraction
			len := sawtooth_length
			sawtooth_height := sqrt ((len / 4.0) ^ 2 - (len / 8.0) ^ 2)
		end

	generate_drawing_for_location (loc: DIE_LOCATIONS_EROW)
			-- Generate part of drawing for `loc' that is specific
			-- to a particular map
		local
			box: DRW_BASIC_RECTANGLE
			size: CONST_POINT
			desc: STRING
			group: DRW_GROUP
		do
			if not loc.is_scribe then
				size := loc.printed_size_on_wafer
				create box.make (size.x, size.y, loc.location_x, loc.location_y)
				desc := loc.die_id
				group := optional_text_pair_in_box (desc, "", box, loc.non_null_die_wafer_rotation)
				if not cut_pattern.isolated_sites.has (loc.mask_site_no) then
					group.do_all (agent {DRW_SHAPE}.set_stroke (unisolated_die_stroke))
				end
				last_drawing.extend (group)
			end
		end

	draw_cut (cut: CUT_LOCATIONS_ROW)
			-- Generate representation of `cut' and add to
			-- `last_drawing'
		require
			cut_not_void: cut /= Void
		local
			line: DRW_LINE
			loc: DOUBLE
		do
			loc := cut.location
			if cut.is_horizontal then
				create line.make (payload_box.ll.x - cut_extension, loc, payload_box.lr.x + cut_extension, loc)
				add_horizontal_sawtooth (payload_box.ll.x - (cut_extension + sawtooth_length), loc)
				add_horizontal_sawtooth (payload_box.lr.x + cut_extension, loc)
			else
				create line.make (loc, payload_box.ll.y - cut_extension, loc, payload_box.ul.y + cut_extension)
				add_vertical_sawtooth (loc, payload_box.ll.y - (cut_extension + sawtooth_length))
				add_vertical_sawtooth (loc, payload_box.ul.y + cut_extension)
			end
			line.set_stroke (Cut_stroke)
			line.set_stroke_width (Cut_scale_factor * thick_width)
			last_drawing.extend (line)
		end

	add_horizontal_sawtooth (x, y: DOUBLE)
			-- Add horizontal sawtooth starting at (`x', `y')
		local
			path: DRW_PATH
			line: DRW_DRAWTO
			len, h: DOUBLE
		do
			len := sawtooth_length / 8.0
			h := sawtooth_height
			create path.make (4)
			path.set_stroke (Cut_stroke)
			path.set_stroke_width (Cut_scale_factor * thick_width)
			create line.make_moveto_abs (x, y)
			path.extend (line)
			create line.make_lineto_abs (x + 1.0 * len, y + h)
			path.extend (line)
			create line.make_lineto_abs (x + 3.0 * len, y - h)
			path.extend (line)
			create line.make_lineto_abs (x + 5.0 * len, y + h)
			path.extend (line)
			create line.make_lineto_abs (x + 7.0 * len, y - h)
			path.extend (line)
			create line.make_lineto_abs (x + 8.0 * len, y)
			path.extend (line)
			last_drawing.extend (path)
		end

	add_vertical_sawtooth (x, y: DOUBLE)
			-- Add vertical sawtooth starting at (`x', `y')
		local
			path: DRW_PATH
			line: DRW_DRAWTO
			len, h: DOUBLE
		do
			len := sawtooth_length / 8.0
			h := sawtooth_height
			create path.make (4)
			path.set_stroke (Cut_stroke)
			path.set_stroke_width (Cut_scale_factor * thick_width)
			create line.make_moveto_abs (x, y)
			path.extend (line)
			create line.make_lineto_abs (x + h, y + 1.0 * len)
			path.extend (line)
			create line.make_lineto_abs (x - h, y + 3.0 * len)
			path.extend (line)
			create line.make_lineto_abs (x + h, y + 5.0 * len)
			path.extend (line)
			create line.make_lineto_abs (x - h, y + 7.0 * len)
			path.extend (line)
			create line.make_lineto_abs (x,     y + 8.0 * len)
			path.extend (line)
			last_drawing.extend (path)
		end

	set_cut_pattern (pat: like cut_pattern)
			-- Set `cut_pattern' to `pat'
		require
			pat_not_void: pat /= Void
		do
			cut_pattern := pat
			field_name := pat.field_name
		end

	thick_box (box: DRW_BASIC_RECTANGLE): DRW_RECTANGLE
			-- Rectangle made from `box', with thick lines
		require
			box_not_void: box /= Void
		do
			create Result.make_from_rectangle (box)
			Result.set_stroke_width (thick_width)
		end

	location_fill_color (loc: DIE_LOCATIONS_EROW): DRW_COLOR
			-- Fill color to be used for interior of `loc'
		do
			if cut_pattern.isolated_sites.has (loc.mask_site_no) then
				Result := isolated_die_fill
			else
				Result := Void
			end
		end

	cut_pattern: RUN_CUT_PATTERNS_ROW
			-- Cut pattern for which page is being generated

	original_title_fraction: DOUBLE
			-- Value of title fraction in parent

	cut_stroke: DRW_COLOR
			-- Color for cut lines

	unisolated_die_stroke: DRW_COLOR
			-- Color for dies that are not isolated by cut pattern

	draft_stroke: DRW_COLOR
			-- Color for text labeling drawing as a draft

	isolated_die_fill: DRW_COLOR
			-- Fill for interior of dies that are isolated
			-- by cut pattern

	cut_extension_fraction: DOUBLE = 0.03
			-- Fraction of drawing maximum dimension (width or
			-- height) to use for distance drawn cuts extend past
			-- edge of payload

	sawtooth_length_fraction: DOUBLE = 0.02
			-- Fraction of drawing maximum dimension (width or
			-- height) to use for length of drawn sawtooth

	cut_extension: DOUBLE
			-- Distance to extend drawn cuts past edge of payload,
			-- in microns

	sawtooth_length: DOUBLE
			-- Length of drawn sawtooth, in microns

	sawtooth_height: DOUBLE
			-- Height of drawn sawtooth (above cut line and
			-- below cut line), in microns

	Cut_scale_factor: DOUBLE = 2.0
			-- Factor by which to thicken cut lines

	Title_scale_factor: DOUBLE = 1.5
			-- Factor by which scale title height

	Unknown_value: STRING = "UNKNOWN"
			-- What to display for value that is not known

	Draft_label: STRING = "DRAFT"
			-- Text identifying map as a draft

end
