
indexing
	
	author: "David Hollenberg"
	date: "November 20, 1998"
	description: "A consumer of data from some arbitrary source"

deferred class DATA_CONSUMER

feature -- Operations

	initialize
			-- Initialize `Current'.  Must be called before
			-- any data is passed to it.  May be redefined in
			-- descendant
		do
			reset
		ensure
			initialized: in_initial_state
			no_errors: not local_errors_or_warnings
		end

	consume (the_data: STRING)
			-- Consume some more data in `the_data'.
			-- If there was an error consuming the data,
			-- set `local_ok' to false and add a description of the
			-- error to `local_errors'.  Does not return until
			-- the data has been consumed or an error has
			-- occurred.
		require
			data_exists: the_data /= Void
		deferred
		end

	consume_eof
			-- Consume an "end of file" indication.
			-- If there was an error consuming the data,
			-- set `local_ok' to false and add a description of the
			-- error to `local_errors'
		deferred
		end

	reset
			-- Reset `Current' so that it may be reused.
			-- May be redefined in descendant
		do
			local_ok := True
			local_errors := Void
			local_warnings := Void
			end_of_input := False
		ensure
			no_errors: not local_errors_or_warnings
		end

	terminate
			-- Terminate operation prematurely, releasing
			-- any resources obtained.  May be redefined
			-- in descendant
		do
		end

feature -- Modification

	set_explicit_name (s: STRING)
			-- Set `explicit_name' to `s'
		do
			if s /= Void then
				explicit_name := s.twin
			else
				explicit_name := Void
			end
		ensure
			name_set: equal (name, s)
		end

feature -- Properties

	in_initial_state: BOOLEAN
			-- Is `Current' in its initial state?
		deferred
		end

	name: STRING
			-- Current name of this consumer (`explicit_name'
			-- if non-Void, otherwise `default_name').  Prepended
			-- to error and warning messages
		do
			if explicit_name /= Void then
				Result := explicit_name
			else
				Result := default_name
			end
		end

	explicit_name: STRING
			-- Explicit name for `Current' (overrides
			-- `default_name')
			
	default_name: STRING
			-- Default name of this consumer
		deferred
		end

	always_ok: BOOLEAN
			-- Is this consumer always OK (never any errors
			-- or warnings from it or from the consumer of
			-- its output, if any)?
		deferred
		end
	
	local_ok: BOOLEAN
			-- Have all calls made to `consume' or `consume_eof'
			-- since initialization or reset succeeded
			-- for this consumer?
			-- If this becomes false, all further calls to
			-- `consume' or `consume_eof' do nothing

	local_errors: LIST [STRING]
			-- Errors that occurred for this consumer
			-- in a call to `consume' or `consume_eof'.
			-- Void unless `local_ok' is false

	local_warnings: LIST [STRING]
			-- Warnings that occurred for this consumer
			-- during calls to `consume' and `consume_eof'
			-- or Void if none

	other_ok: BOOLEAN
			-- Have all calls made to `consume' or `consume_eof'
			-- since initialization or reset succeeded
			-- for the consumer of my output, if any?
		deferred
		end

	other_errors: LIST [STRING]
			-- Errors that occurred for the consumer of my output
			-- (if any) in a call to `consume' or `consume_eof'.
			-- Void unless `other_ok' is false
		deferred
		end

	other_warnings: LIST [STRING]
			-- Warnings that occurred for the consumer of
			-- my output during calls to `consume' and
			-- `consume_eof' or Void if none
		deferred
		end

	all_ok: BOOLEAN
			-- Have all calls made to `consume' or `consume_eof'
			-- since initialization or reset succeeded
			-- for this consumer and all its output consumers?
		deferred
		end

	all_errors: LIST [STRING]
			-- Errors that occurred for this consumer and all
			-- its output consumers in a call to `consume' or
			-- `consume_eof'.
			-- Void unless `all_ok' is false
		deferred
		end

	all_warnings: LIST [STRING]
			-- Warnings that occurred for this consumer
			-- and all its output consumers during calls
			-- to `consume' and `consume_eof', or Void if
			-- none
		deferred
		end

feature {DATA_TRANSFORM} -- Implementation

	end_of_input: BOOLEAN
			-- Has end of input been reached?
			-- This can happen either when an EOF indication
			-- is consumed or when there is a transform feeding
			-- input to `Current' and it gets an error
			-- Only meaningful for a DATA_TRANSFORM

	set_end_of_input (b: BOOLEAN)
			-- Set `end_of_input' to `b'
		do
			end_of_input := b
		ensure
			value_set: end_of_input = b
		end

feature {NONE} -- Implementation

	local_errors_or_warnings: BOOLEAN
			-- Have any local errors or warnings occurred
			-- since `Current' was initialized or reset?
		do
			Result := not (local_ok and local_errors = Void and local_warnings = Void)
		end

	append_error (s: STRING)
			-- Append consumption error described by `s'
			-- to `local_errors'
		local
			n, err: STRING
		do
			if s /= Void then
				if local_errors = Void then
					create {ARRAYED_LIST [STRING]} local_errors.make (5)
				end
				create err.make (100)
				n := name
				if not n.is_empty then
					err.append (n)
					err.append (": ")
				end
				err.append (s)
				local_errors.extend (err)
			end
			local_ok := False
		end

	append_errors (list: LIST [STRING])
			-- Append errors in `list' to `local_errors'
		do
			if list /= Void then
				from
					list.start
				until
					list.after
				loop
					append_error (list.item)
					list.forth
				end
			end
		end

	append_warning (s: STRING)
			-- Append consumption warning `s'
			-- to `local_warnings'
		local
			n, warn: STRING
		do
			if s /= Void then
				if local_warnings = Void then
					create {ARRAYED_LIST [STRING]} local_warnings.make (5)
				end
				create warn.make (100)
				n := name
				if not n.is_empty then
					warn.append (n)
					warn.append (": ")
				end
				warn.append (s)
				local_warnings.extend (warn)
			end
		end

	append_warnings (list: LIST [STRING])
			-- Append warnings in `list' to `local_warnings'
		do
			if list /= Void then
				from
					list.start
				until
					list.after
				loop
					append_warning (list.item)
					list.forth
				end
			end
		end

invariant
	default_name_exists: default_name /= Void
	no_local_errors_equals_ok: local_ok = (local_errors = Void)
	no_other_errors_equals_ok: other_ok = (other_errors = Void)
	no_all_errors_equals_ok: all_ok = (all_errors = Void)
	local_ok_if_always_ok: always_ok implies all_ok

end
