
indexing
	
	author: "Joel Goldberg (reworked by David Hollenberg)"
	date: "January 18, 2000"
	description: "Arbitrary source of data, which can be read %
		%a character, line or buffer at a time"

deferred class DATA_SOURCE

inherit
	FILABLE

feature -- Access

	last_char: CHARACTER
			-- Last character read by `read_char'

	last_string: STRING
			-- Last string read by `read_line' or `read_stream'
	
feature -- Status

	after: BOOLEAN
			-- Is there no more data available in data source?
			-- For a source containing N characters, will 
			-- be true after the N+1 calls to `read_char'
			-- (or equivalent in `read_line' or `read_stream')
		deferred
		end

	position: INTEGER_64
			-- Position of next character to read from source,
			-- relative to 0 (i.e., number of characters read
			-- so far)
		deferred
		end

	count: INTEGER_64
			-- Total number of characters in data source
		deferred
		ensure
			valid_count: count >= 0
		end
	
feature -- Modification

	go (pos: INTEGER_64)
			-- Reset `position' to `pos'
		require
			position_large_enough: pos >= 0
		deferred
		end

	reset
			-- Reset `position' to 0, so that next read
			-- starts from beginning of source
		do
			go (0)
		ensure
			at_start: position = 0
			not_end_of_data: not after
		end

	read_char
			-- Read next character of source and put in
			-- `last_char'.  If no more characters, `after'
			-- becomes true and value of `last_char' is
			-- not meaningful
		require
			not_end_of_data: not after
		deferred
		ensure
			position_changed: (position = old position + 1) or after
		end

	read_line
			-- Read characters until newline encountered or no
			-- more data.  Leave characters read (not including
			-- newline) in `last_string'
		require
			not_end_of_data: not after
		deferred
		ensure
			position_changed: position >= (old position) + last_string.count
		end

	read_stream (n: INTEGER)
			-- Read characters until `n' have been read or 
			-- no more data and put in `last_string'
		require
			not_end_of_data: not after
			count_large_enough: n >= 0
		deferred
		ensure
			valid_count: last_string.count = n or (after and last_string.count < n)
			position_changed: position = (old position) + last_string.count
		end

	copy_to_consumer (c: DATA_CONSUMER)
			-- Copy entire contents of source to consumer `c'.
			-- Quit on any consumption error.
			-- Does not initialize or terminate `c'
		do
			from
				reset
			until
				after or not c.all_ok
			loop
				read_stream (4096)
				if last_string.count > 0 then
					c.consume (last_string)
				end
			end
		end

feature -- Segments

	segment (pos, cnt: INTEGER_64): DATA_SOURCE
			-- Segment of current source consisting of 
			-- `cnt' characters starting at position `pos'
		require
			valid_position: pos >= 0
			valid_count: cnt >= 0
			window_short_enough: count >= 0 implies pos + cnt <= count
		deferred
		end

	tail: DATA_SOURCE
			-- Tail of current source, with all characters
			-- that have not yet been read
		local
			pos: INTEGER_64
		do
			if after then
				create {DATA_SOURCE_EMPTY} Result
			else
				pos := position
				Result := segment (pos, count - pos)
			end
		end
	
invariant

	valid_position: position >= 0
	valid_count: position <= count

end
