
indexing
	
	author: "Joel Goldberg (reworked by David Hollenberg)"
	date: "January 18, 2000"
	description: "Source of data that comes from a file"

class DATA_SOURCE_FILE

inherit
	DATA_SOURCE

create
	make_entire,
	make_partial,
	make_from_filename

feature {NONE} -- Creation

	make_entire (f: LARGE_FILE)
			-- Create data source consisting of all characters
			-- of `f'
		require
			file_exists: f /= Void
			file_readable: f.is_open_read
		do
			source_file := f
			first_position := 0
			last_position := f.count
		end

	make_partial (f: LARGE_FILE pos, cnt: INTEGER_64)
			-- Create data source starting at position `pos'
			-- in `f' and `cnt' bytes long
		require
			file_exists: f /= Void
			file_readable: f.is_open_read
		do
			source_file := f
			first_position := pos
			last_position := pos + cnt
		end

	make_from_filename (fname: STRING)
			-- Create data source consisting of all characters
			-- of file named `fname'
		require
			file_name_exists: fname /= Void
		local
			f: LARGE_PLAIN_TEXT_FILE
		do
			create f.make_open_read (fname)
			make_entire (f)
		end
	
feature -- Status

	after: BOOLEAN
			-- Is there no more data available in data source?
			-- For a source containing N characters, will 
			-- be true after N+1 calls to `read_char'
			-- (or equivalent in `read_line' or `read_stream')

	position: INTEGER_64
			-- Position of next character to read from source,
			-- relative to 0 (i.e., number of characters read
			-- so far)
		do
			Result := source_file.position - first_position
		end

	count: INTEGER_64
			-- Total number of characters in data source
		do
			Result := last_position - first_position
		end

feature -- Modification

	go (pos: INTEGER_64)
			-- Reset `position' to `pos'
		do
			after := pos > count
			if not after then
				source_file.go (first_position + pos)
			end
		end

	read_char
			-- Read next character of source and put in
			-- `last_char'.  If no more characters, `after'
			-- becomes true and value of `last_char' is
			-- not meaningful
		do
			if source_file.position < last_position then
				source_file.read_character
				last_char := source_file.last_character
			else
				after := true
			end
		end

	read_line
			-- Read characters until newline encountered or no
			-- more data.  Leave characters read (not including
			-- newline) in `last_string'
		local
			curr_pos: INTEGER_64
		do
			curr_pos := source_file.position
			if curr_pos < last_position then
				source_file.read_line
				last_string := source_file.last_string
				if source_file.position > last_position or source_file.end_of_file then
					source_file.go (last_position)
					last_string.keep_head ((last_position - curr_pos).to_integer)
					after := true
				end
			else
				create last_string.make (0)
				after := true
			end
		end

	read_stream (n: INTEGER)
			-- Read characters until `n' have been read or 
			-- no more data and put in `last_string'
		local
			diff: INTEGER_64
			num: INTEGER
		do
			diff := last_position - source_file.position
			if diff < n then
				num := diff.to_integer
			else
				num := n
			end
			if num > 0 then
				source_file.read_stream (num)
				last_string := source_file.last_string
			else
				create last_string.make (0)
			end
			if num < n then
				after := true
			end
		end

feature -- Segments

	segment (pos, cnt: INTEGER_64): DATA_SOURCE_FILE
			-- Segment of current source consisting of 
			-- `cnt' characters starting at position `pos'
		do
			create Result.make_partial (source_file, first_position + pos, cnt)
		end

	close
			-- Close data source and release all resources
		do
			source_file.close
			source_file := void
		end
	
feature {NONE} -- Implementation

	source_file: LARGE_FILE
			-- File that is source of data

	first_position: INTEGER_64
			-- Position in file of first character of source

	last_position: INTEGER_64
			-- Position in file immediately after last 
			-- character of source

end
