
indexing
	
	author: "Joel Goldberg (reworked by David Hollenberg)"
	date: "January 18, 2000"
	description: "Source of data that comes from a string"

class DATA_SOURCE_STRING

inherit
	DATA_SOURCE

create
	make

feature {NONE} -- Creation

	make (s: STRING)
			-- Create data source whose characters are
			-- those of `s'.  String is not copied or
			-- cloned, so it must not be changed as long
			-- as `Current' is reachable
		require
			string_exists: s /= Void
		do
			source_string := s
		end
	
feature -- Status

	after: BOOLEAN
			-- Is there no more data available in data source?
			-- For a source containing N characters, will 
			-- be true after the N+1 calls to `read_char'
			-- (or equivalent in `read_line' or `read_stream')

	position: INTEGER_64
			-- Position of next character to read from source,
			-- relative to 0 (i.e., number of characters read
			-- so far)

	count: INTEGER_64
			-- Total number of characters in data source
		do
			Result := source_string.count
		end

feature -- Modification

	go (pos: INTEGER_64)
			-- Reset `position' to `pos'
		do
			position := pos.min (count)
			after := pos > count
		end

	read_char
			-- Read next character of source and put in
			-- `last_char'.  If no more characters, `after'
			-- becomes true and value of `last_char' is
			-- not meaningful
		do
			if position < count then
				position := position + 1
				last_char := source_string.item (position.to_integer)
			else
				after := true
			end
		end

	read_line
			-- Read characters until newline encountered or no
			-- more data.  Leave characters read (not including
			-- newline) in `last_string'
		local
			end_pos, num: INTEGER
		do
			if position < count then
				end_pos := source_string.index_of ('%N', (position + 1).to_integer)
			end
			if end_pos = 0 then
				num := (count - position).to_integer
			else
				num := end_pos - 1 - position.to_integer
			end
			put_into_last_string (num)
			if end_pos = 0 then
				after := true
			else
				position := position + 1
			end
		end

	read_stream (n: INTEGER)
			-- Read characters until `n' have been read or 
			-- no more data and put in `last_string'
		local
			num: INTEGER
		do
			num := n.min ((count - position).to_integer)
			put_into_last_string (num)
			if num < n then
				after := true
			end
		end

	close
			-- Close data source and release all resources
		do
		end
	
feature -- Segments

	segment (pos, cnt: INTEGER_64): DATA_SOURCE_STRING
			-- Segment of current source consisting of 
			-- `cnt' characters starting at position `pos'
		local
			temp_str: STRING
		do
			create temp_str.make (cnt.to_integer)
			temp_str.set (source_string, pos.to_integer, (pos + cnt - 1).to_integer)
			create Result.make (temp_str)
		end

feature {NONE} -- Implementation

	source_string: STRING
			-- String which is source of data

	put_into_last_string (num: INTEGER)
			-- Set `last_string' to the next `num' characters
			-- of source and update `position'
		require
			valid_number: num >= 0
		do
			if last_string = Void then
				create last_string.make (num)
			end
			last_string.set (source_string, (position + 1).to_integer, (position + num).to_integer)
			position := position + num
		end

invariant
	source_exists: source_string /= Void

end
