
indexing
	
	author: "David Hollenberg"
	date: "November 20, 1998"
	description: "A data transformation"

deferred class DATA_TRANSFORM

inherit
	DATA_CONSUMER
		redefine
			reset, initialize, terminate
		end

feature -- Properties

	output_buffer: STRING
			-- Transformed output produced by last call
			-- to `consume' or `consume_eof',
			-- if `output_consumer' is Void

	error_buffer: STRING
			-- Error output produced by last call
			-- to `consume' or `consume_eof',
			-- if `error_consumer' is Void

	output_capacity: INTEGER
			-- Requested capacity of output buffer
			
	error_capacity: INTEGER
			-- Requested capacity of error buffer
			
	output_consumer: DATA_CONSUMER
			-- Consumer of output produced by transformation,
			-- if not Void.  If Void, output is left in
			-- `output_buffer'

	error_consumer: DATA_CONSUMER
			-- Consumer of error output produced by transformation,
			-- if not Void.  If Void, error output is left
			-- in `error_buffer'

	immediate_error_propagation: BOOLEAN
			-- Should `Current' propagate any errors of
			-- `output_consumer' immediately, making
			-- `all_ok' false?
			-- If false, errors of `output_consumer' will
			-- only become visible in `all_errors'
			-- when an EOF indication is consumed or when
			-- `Current' gets an error

	always_ok: BOOLEAN = False
			-- Is this consumer always OK?
			-- (Answer: no, because output consumer could fail)
	
	other_ok: BOOLEAN
			-- Have all calls made to `consume' or `consume_eof'
			-- since initialization or reset succeeded
			-- for the consumer of my output, if any?
		do
			if output_consumer = Void then
				Result := True
			elseif immediate_error_propagation or end_of_input then
				Result := output_consumer.all_ok
			else
				Result := True
			end
		end

	other_errors: LIST [STRING]
			-- Errors that occurred for the consumer of my output
			-- (if any) in a call to `consume' or `consume_eof'.
			-- Void unless `other_ok' is false
		do
			if output_consumer = Void then
				Result := Void
			elseif immediate_error_propagation or end_of_input then
				Result := output_consumer.all_errors
			else
				Result := Void
			end
		end

	other_warnings: LIST [STRING]
			-- Warnings that occurred for the consumer of
			-- my output during calls to `consume' and
			-- `consume_eof' or Void if none
		do
			if output_consumer = Void then
				Result := Void
			elseif immediate_error_propagation or end_of_input then
				Result := output_consumer.all_warnings
			else
				Result := Void
			end
		end

	all_ok: BOOLEAN
			-- Have all calls made to `consume' or `consume_eof'
			-- since initialization or reset succeeded
			-- for this consumer and all its output consumers?
		do
			Result := local_ok and other_ok
		end

	all_errors: LIST [STRING]
			-- Errors that occurred for this consumer and all
			-- its output consumers in a call to `consume' or
			-- `consume_eof'.
			-- Void unless `all_ok' is false
		local
			others: LIST [STRING]
		do
			others := other_errors
			if others /= Void and local_errors /= Void then
				create {ARRAYED_LIST [STRING]} Result.make (others.count + local_errors.count)
				Result.append (local_errors)
				Result.append (others)
			elseif local_errors /= Void then
				Result := local_errors
			elseif others /= Void then
				Result := others
			else
				Result := Void
			end
		end

	all_warnings: LIST [STRING]
			-- Warnings that occurred for this consumer
			-- and all its output consumers during calls
			-- to `consume' and `consume_eof', or Void if
			-- none
		local
			others: LIST [STRING]
		do
			others := other_warnings
			if others /= Void and local_warnings /= Void then
				create {ARRAYED_LIST [STRING]} Result.make (others.count + local_warnings.count)
				Result.append (local_warnings)
				Result.append (others)
			elseif local_warnings /= Void then
				Result := local_warnings
			elseif others /= Void then
				Result := others
			else
				Result := Void
			end
		end

feature -- Modification

	set_output_consumer (c: DATA_CONSUMER)
			-- Set `output_consumer' to `c'
		do
			output_consumer := c
		ensure
			consumer_set: output_consumer = c
		end

	set_error_consumer (c: DATA_CONSUMER)
			-- Set `error_consumer' to `c'
		require
			consumer_always_ok: c /= Void implies c.always_ok
		do
			error_consumer := c
		ensure
			consumer_set: error_consumer = c
		end

	set_output_capacity (n: INTEGER)
			-- Set capacity of `output_buffer' to `n'
		require
			positive_value: n > 0
		do
			output_capacity := n
			if output_buffer /= Void then
				output_buffer.resize (n)
			end
		ensure
			output_capacity_set: output_capacity = n
			output_buffer_resized: output_buffer /= Void implies output_buffer.count >= output_capacity
		end

	set_error_capacity (n: INTEGER)
			-- Set capacity of `error_buffer' to `n'
		require
			positive_value: n > 0
		do
			error_capacity := n
			if error_buffer /= Void then
				error_buffer.resize (n)
			end
		ensure
			error_capacity_set: error_capacity = n
			error_buffer_resized: error_buffer /= Void implies error_buffer.count >= error_capacity
		end

	set_immediate_error_propagation (b: BOOLEAN)
			-- Set `immediate_error_propagation' to `b'
		do
			immediate_error_propagation := b
		ensure
			value_set: immediate_error_propagation = b
		end

feature -- Operations

	initialize
			-- Initialize `Current'.  Must be called before
			-- any data is passed to it.
		do
			precursor
			create output_buffer.make (output_capacity)
			create error_buffer.make (error_capacity)
			if output_consumer /= Void then
				output_consumer.initialize
			end
			if error_consumer /= Void then
				error_consumer.initialize
			end
		--	precursor
		end

	consume (the_data: STRING)
			-- Consume some more data in `the_data' and
			-- propagate the resulting output to
			-- `output_consumer', if not Void.
			-- If there was an error consuming the data,
			-- set `local_ok' to false and add a description of the
			-- error to `local_errors'.  Does not return until
			-- the data has been consumed or an error has
			-- occurred.
		do
			if local_ok then
				consume_internal (the_data)
				propagate_data
				if not local_ok and output_consumer /= Void then
					output_consumer.set_end_of_input (True)
				end
			end
		end

	consume_eof
			-- Consume an "end of file" indication and
			-- propagate any resulting output and the EOF
			-- indication to `output_consumer', if not Void.
			-- If there was an error consuming the data,
			-- set `local_ok' to false and add a description of the
			-- error to `local_errors'.
		do
			if local_ok then
				consume_eof_internal
				end_of_input := True
				propagate_data	
					-- Propagate any data generated as
					-- a result of EOF indication
				propagate_eof
			end
		end

	reset
			-- Reset `Current' to its initial state so
			-- that it may be reused.  May be redefined in
			-- descendant
		do
			precursor
			if output_buffer /= Void then
				output_buffer.keep_head (0)
			end
			if error_buffer /= Void then
				error_buffer.keep_head (0)
			end
			if output_consumer /= Void then
				output_consumer.reset
			end
			if error_consumer /= Void then
				error_consumer.reset
			end
		end

	terminate
			-- Terminate transformation prematurely
		do
			if output_consumer /= Void then
				output_consumer.terminate
			end
			if error_consumer /= Void then
				error_consumer.terminate
			end
		end

feature {NONE} -- Implementation

	consume_internal (the_data: STRING)
			-- Consume some more data in `the_data'
		require
			data_exists: the_data /= Void
		deferred
		end

	consume_eof_internal
			-- Consume an "end of file" indication
		deferred
		end

	propagate_data
			-- Propagate data in `output_buffer' to
			-- `output_consumer', if not Void
		deferred
		end

	propagate_eof
			-- Propagate "end of file" indication to
			-- `output_consumer', if not Void
		deferred
		end

end

