
indexing

	description: "Delta run closing system"
	author: "David Hollenberg"
	date: "June 20, 2002"

class DELTA
inherit
	GENERIC_SYSTEM
		rename
			vars as rc_env
		undefine
			rc_env
		redefine
			make
		end
	RC_SESSION_MANAGER_ACCESS
		export
			{NONE} all
		end
	RC_ENVIRONMENT_ACCESS
		export
			{NONE} all
		end
	STAT_CONSTANTS
		export
			{NONE} all
		end
	RESOURCE_LIMITS
		export
			{NONE} all
		end
	PUBLICATION_MANAGER_ACCESS
		export
			{NONE} all
		end
	CONTROL_FLOW_EXCEPTIONS_ACCESS
		export
			{NONE} all
		end
	GUI_UTILITIES_ACCESS
		export
			{NONE} all
		end
	USER_INTERFACE_ACCESS
		export
			{NONE} all
		end
	EXCEPTIONS_ACCESS
		export
			{NONE} all
		end

create
	make

feature {NONE} -- Creation

	make (args: ARRAY[STRING])
			-- Run system
		do
			is_silent_on_error := True
			precursor (args)
		end

feature -- Operations

	run_a_system
			-- Run system after reading configuration
		do
			is_silent_on_error := False
			glo.set_system_internal_error ("DELTA: internal error")
			glo.set_is_logging (False)
			set_no_session_current_step_name ("Initialization")
			set_no_session_current_activity ("Initializing")
			initialize
			set_no_session_current_activity ("Checking configuration")
			setup_event_handlers
			check_configuration
			set_display
			open_log_and_error_files
			get_delta_lock
			run_delta
		end

	setup_event_handlers
			-- Setup handlers for published events
		local
			handler: DELTA_EVENT_HANDLER
		do
			create handler.make (session_manager)
			handler.subscribe_to_events
			pubman.enable_event_processing
		end

	run_delta
			-- Run Delta
		local
			session: RC_SESSION
			version: DELTA_LATEST_VERSION
		do
			set_no_session_current_activity  ("Creating interface")
			session := session_manager.new_session
			session_manager.set_current_session (session)
			if rc_env.is_graphical then
				create version
				rc_env.log_operation ("Starting graphical application (version " + version.number + " compiled on " + version.release_date + ")")
				exception := Void
				run_graphic_delta_with_error_handling (session)
				if exception /= Void then
					exc.raise (exception_explanation)
				end
			else
				report_fatal_error ("Non-graphical version of Delta not implemented", "configuration")
			end
			set_no_session_current_step_name ("Finalization")
			set_no_session_current_activity ("Exiting")
		end

	run_graphic_delta_with_error_handling (sess: RC_SESSION)
			-- Run Delta as a graphical application, using
			-- `sess' as initial session.  If an exception
			-- is raised, catch it, set all windows insensitive
			-- and create a new window to display the error
		local
			delta: DELTA_GRAPHIC_APPLICATION
			tried: BOOLEAN
		do
			if not tried then
				create delta.make (Current)
				if su.not_empty (delta.initialization_errors) then
					report_fatal_error (delta.initialization_errors, "initialization")
					exc.die (0)
				end

				set_user_interface_parameters
				sess.set_factory (delta)
				sess.set_application (delta)
				delta.set_background_color
				delta.build
				delta.uncaught_exception_actions.extend (agent handle_uncaught_exception (delta, ?))
				delta.launch
				if exception = Void then
					rc_env.log_operation ("Application terminated normally")
				else
					rc_env.log_operation ("Application terminated by fatal exception")
				end
			end
		rescue
			tried := True
			handle_delta_exception (delta, Void)
			if needs_retry then
				retry
			end
		end

	handle_uncaught_exception (delta: DELTA_GRAPHIC_APPLICATION e: EXCEPTION)
			-- Handle uncaught exception `e' that occurred while
			-- trying to execute an action sequence from
			-- within Vision2 event loop.
		do
			handle_delta_exception (delta, e)
		end

	handle_delta_exception (delta: DELTA_GRAPHIC_APPLICATION e: EXCEPTION)
			-- Handle exception `e' that occurred while
			-- trying to run `delta', using
			-- `e'.`trace_as_string' as a description of
			-- the exception trace if not Void.  Set
			-- `needs_retry' to indicate whether caller
			-- should execute a retry instruction if in a
			-- rescue clause
		local
			a_trace: STRING
		do
			needs_retry := False
			if unix_os.is_software_termination then
				try_to_log ("Application terminated by kill signal")
				exc.die (0)
			elseif unix_os.is_interrupt then
				try_to_log ("Application terminated by interrupt")
				exc.die (0)
			elseif unix_os.is_hangup then
				try_to_log ("Application terminated by hangup")
				exc.die (0)
			else
				if cfe.is_abort_exception then
					try_to_log ("Application detected uncaught abort")
				else
					try_to_log ("Handling fatal exception")
				end
				exception := e
				if e /= Void then
					exception_explanation := mu.exception_explanation
					a_trace := e.exception_trace
				else
					exception_explanation := "Uncaught exception"
					a_trace := Void
				end
				delta.report_fatal_exception (a_trace)
			end
		end

	try_to_log (s: STRING)
			-- Try to log message `s'.  Do not raise an exception
		require
			message_exists: s /= Void
		local
			tried: BOOLEAN
		do
			if not tried then
				rc_env.log_operation (s)
			end
		rescue
			tried := True
			retry
		end

	disable_error_notification
			-- Set `is_silent_on_error' to true
		do
			is_silent_on_error := True
		ensure
			no_notification: is_silent_on_error
		end

	initialize
			-- Do any needed initialization
		local
			old_mask, new_mask: INTEGER
		do
			set_all_soft_resource_limits_to_max
			new_mask := (S_irusr | S_iwusr | S_irgrp | S_iwgrp).bit_not
			old_mask := unix_os.set_file_creation_mask (new_mask)
		end

	check_configuration
			-- Check that configuration parameters were read
			-- without error and that all required parameters
			-- are present
		local
			errs: STRING
		do
			errs := rc_env.configuration_errors
			if su.not_empty (errs) then
				report_fatal_error (errs, "configuration")
				exc.die (0)
			end
		end

	set_display
			-- Set DISPLAY environment variable to value
			-- of fixed_display_name if non-empty (for debugging
			-- only)
		do
			if su.not_empty (rc_env.fixed_display_name) then
				exe.put (rc_env.fixed_display_name, "DISPLAY")
			end
		end

	get_delta_lock
			-- Get lock on lock file that allows running
			-- of Delta.  Get shared lock unless exclusive
			-- lock was requested.
		local
			
			name, sep: STRING
			exclusive: BOOLEAN
			process_id: INTEGER
		do
			sep := "   "
			name := rc_env.delta_lock_name
			exclusive := rc_env.exclusive_delta_lock
			delta_lock := os.new_lock (name)
			if exclusive then
				delta_lock.set_exclusive
				delta_lock.lock_nowait
				if delta_lock.is_locked_by_me then
					delta_lock.set_auxiliary_data (rc_env.session_identifier + sep + "exclusive%N")
					io.put_string ("Hit return to release exclusive Delta lock: ")
					io.read_line
				else
					process_id := delta_lock.conflicting_lock_process_id
					io.put_string ("Unable to get exclusive lock on file " + name + "%N")
					io.put_string ("Lock is held by process id " + process_id.out + "%N")
					show_matching_processes (name, process_id.out)
				end
				exc.die (0)
			else
				delta_lock.set_shared
				delta_lock.lock_nowait
				if delta_lock.is_locked_by_me then
					delta_lock.append_auxiliary_data (rc_env.session_identifier + sep + "shared%N")
				else
					process_id := delta_lock.conflicting_lock_process_id
					io.put_string ("Unable to get shared lock on file " + name + "%N")
					io.put_string ("Release of new Delta version is probably in progress.%N")
					io.put_string ("If problem persists, please contact development staff.%N")
					io.put_string ("Lock is held by process id " + process_id.out + "%N")
					show_matching_processes (name, process_id.out)
					exc.die (0)
				end
			end
		end

	show_matching_processes (fname, pid: STRING)
			-- Show lines from file named `fname' that
			-- match process id `pid'
		local
			line, line_pid, user, host: STRING
			f: LARGE_PLAIN_TEXT_FILE
			words: LIST [STRING]
		do
			f := fu.opened_for_read (fname)
			if f /= Void then
				from
					f.read_line
				until
					f.end_of_file
				loop
					line := f.last_string
					words := su.broken_into_words (line)
					if words.count >= 3 then
						line_pid := words.i_th (1)
						user := words.i_th (2)
						host := words.i_th (3)
						if line_pid.is_equal (pid) then
							print (line)
							io.new_line
						end
					end
					f.read_line
				end
				f.close
			end
		end

	open_log_and_error_files
			-- Open general log and error files for append
		local
			errs: STRING
		do
			rc_env.open_log_and_error_files
			errs := rc_env.open_errors
			if su.not_empty (errs) then
				report_fatal_error (errs, "initialization")
				exc.die (0)
			end
		end

	set_user_interface_parameters
			-- Propagate values of parameters which represent
			-- user interface properties into shared object
		local
			color: EV_COLOR
			s: STRING
		do
			s := rc_env.background_color
			if s /= Void then
				color := gui.rgb_string_to_color (s)
				if color /= Void then
					user_interface.set_background_color (color)
				end
			end
		end

	report_fatal_error (s, err_type: STRING)
			-- Report fatal error `s' of type `err_type'
			-- before graphics has been started
		require
			string_exists: s /= Void
			error_type_exists: err_type /= Void
		do
			io.put_string ("ERROR: fatal " + err_type + " error")
			io.new_line
			io.put_string (s)
			if (s.count > 0 and then s.item (s.count) /= '%N') then
				io.new_line
			end
			io.put_string ("Delta terminated due to fatal error")
			io.new_line
		end

feature -- Modification

	set_no_session_current_step_name (s: STRING)
			-- Set `no_session_current_step_name' to `s'
		require
			step_name_exists: s /= Void
		do
			no_session_current_step_name := s
		ensure
			step_name_set: no_session_current_step_name = s
		end

	set_no_session_current_activity (s: STRING)
			-- Set `no_session_current_activity' to `s'
		require
			activity_exists: s /= Void
		do
			no_session_current_activity := s
		ensure
			activity_set: no_session_current_activity = s
		end

feature {NONE} -- Miscellaneous

	delta_lock: LOCK
			-- Lock that allows running of Delta

	needs_retry: BOOLEAN
			-- Should caller of `handle_delta_exception'
			-- execute a retry instruction if calling from
			-- within a rescue clause

	is_silent_on_error: BOOLEAN
			-- Should e-mail notification be suppressed when
			-- an exception occurs?

	rerun_on_transient_error: BOOLEAN = False
			-- Shall we re-run the system on a "transient" error?

	crash_msg_file_name: STRING
			-- Full name of file to receive crash message
			-- (empty means don't file)
		local
			sess: RC_SESSION
		do
			sess := session_manager.current_session
			if sess /= Void then
				Result := sess.error_file_name
			else
				Result := rc_env.general_error_filename
			end
		end

	crash_msg_to_list: STRING
			-- Address list for crash message
			-- (empty means don't send mail)
		do
			if not is_silent_on_error then
				Result := su.empty_value (rc_env.crash_msg_to_list,
					"mfe-default@mosis.com")
			end
		end

	crash_msg_mailer_options: STRING
			-- Command to send mail, including any options
		do
			Result := rc_env.mailer_command
		end

	current_activity: STRING
			-- What we were processing at crash time
		local
			sess: RC_SESSION
			l_trace: STRING
		do
			sess := session_manager.current_session
			if sess /= Void then
				Result := sess.current_state
			else
				Result := current_state
			end
			if exception /= Void then
				l_trace := exception.exception_trace
				if l_trace /= Void then
					Result.extend ('%N')
					Result.append ("Exception trace:%N%N")
					Result.append (l_trace)
				end
			end
		end

	current_state: STRING
			-- Description of current state, including what
			-- is being done, which user is doing it, date, etc.
			-- Only used if there is no current session
		do
			create Result.make (256)
			Result.append ("Date: ")
			unix_os.append_current_date_string (Result)
			Result.extend ('%N')
			Result.append ("Run ID: ")
			Result.append ("<none>")
			Result.extend ('%N')
			Result.append ("User: ")
			Result.append (rc_env.user_name)
			Result.extend ('%N')
			Result.append ("Host: ")
			Result.append (rc_env.host_name)
			Result.extend ('%N')
			Result.append ("Process ID: ")
			Result.append_integer (rc_env.process_id)
			Result.extend ('%N')
			Result.append ("Step: ")
			if su.not_empty (no_session_current_step_name) then
				Result.append (no_session_current_step_name)
			end
			Result.extend ('%N')
			Result.append ("Activity: ")
			if su.not_empty (no_session_current_activity) then
				Result.append (no_session_current_activity)
			end
			Result.extend ('%N')
		ensure
			result_exists: Result /= Void
		end

	no_session_current_activity: STRING
			-- Current activity of Delta before first session
			-- created

	no_session_current_step_name: STRING
			-- Name of current step being executed before
			-- first session created

	mode: STRING
			-- If the system has various operating modes, the name
			-- can be left here for inclusion in messages
		local
			sess: RC_SESSION
		do
			sess := session_manager.current_session
			if sess /= Void then
				if sess.has_steps then
					Result := sess.current_step_name
				end
			else
				Result := no_session_current_step_name
			end
		end

	exception: EXCEPTION
			-- Uncaught exception during callback action (Void
			-- if none)

	exception_explanation: STRING
			-- Explanation of exception recorded when it happens
			-- so it can be reported correctly later

feature {NONE} -- Inapplicable

	rerun_sleep_interval_in_seconds: INTEGER = 0
			-- If we do re-run, how long to sleep

	report_web_error
			-- Last chance message to web browser
		do
			io.put_string ("ERROR: Delta terminated due to fatal error")
			io.new_line
			if not is_silent_on_error then
				io.put_string ("       Support staff notified by e-mail")
				io.new_line

			end
		end

end
