
indexing
	
	description: "Facilities for traversing a directory tree."
	author: "David Hollenberg"
	date: "July 13, 2000"

deferred class DIRECTORY_TREE_TRAVERSAL

inherit
	ANY
	OS_ACCESS
		export
			{NONE} all
		end
	UNIX_OS_ACCESS
		export
			{NONE} all
		end
	FILE_UTILITIES_ACCESS
		export
			{NONE} all
		end
	EXECUTION_ENVIRONMENT_ACCESS
		export
			{NONE} all
		end
	OPERATING_ENVIRONMENT
		export
			{NONE} all
		end
	MISC_UTILITIES_ACCESS
		export
			{NONE} all
		end
	STRING_UTILITIES_ACCESS
		export
			{NONE} all
		end


feature -- Creation

	make (root: STRING)
			-- Create `Current' with root directory `root',
			-- which must be an absolute path name of
			-- a directory
		require
			root_exists: root /= Void
			root_not_empty: not root.is_empty
		do
			root_directory := root
			set_processed_natural_order
		end

feature -- Properties

	root_directory: STRING
			-- Root of directory tree

	directory_symlinks_followed: BOOLEAN
			-- Should symbolic links pointing to directories
			-- be following during the traversal?

	processed_alphabetically: BOOLEAN
			-- Should entries in each directory be processed
			-- in alphabetical order?

	processed_natural_order: BOOLEAN
			-- Should entries in each directory be processed
			-- in their natural (OS-defined) order?

	root_excluded: BOOLEAN
			-- Should root of directory tree be excluded
			-- from the traversal and therefore not visited?

	files_excluded: BOOLEAN
			-- Should files in directory tree be excluded
			-- from the traversal and therefore not visited?

	directories_excluded: BOOLEAN
			-- Should directories in directory tree be excluded
			-- from the traversal and therefore not visited?

	errors_displayed: BOOLEAN
			-- Should errors (directory open errors and
			-- errors processing a file or directory) be
			-- displayed immediately after they occur?
			-- If not, errors are added to `error_list'

	wildcard_filters: LIST [STRING]
			-- List of wildcard patterns.  If non-void,
			-- only entries that match one of these patterns
			-- are visited.  Use '*' to match any number of
			-- characters and '%' to match a single character
			-- If void, then no entry name filtering is done

	ignored_wildcard_filters: LIST [STRING]
			-- List of wildcard patterns.  If non-void,
			-- all entries that match one of these patterns
			-- are completely ignored, as though they
			-- did not exist. Use '*' to match any number of
			-- characters and '%' to match a single character
			-- If void, then no entry name filtering is done

	error_list: LIST [PATH_ERROR]
			-- Errors encountered during last traversal,
			-- if `errors_displayed' is false

	visited_count: INTEGER
			-- Number of path names visited successfully

	error_count: INTEGER
			-- Number of path names on which an error occurred,
			-- either trying to get status on path name or
			-- while trying to visit path name

	error_list_string (heading, footing, item_prefix: STRING): STRING
			-- String with errors in `error_list', with
			-- each item preceded by `item_prefix'.
			-- Whole list is preceded by `heading' and followed
			-- by `footing' if there is anything in the list
		local
			list: LIST [PATH_ERROR]
		do
			list := error_list
			if list.count > 0 then
				create Result.make (100 * error_list.count)
				Result.append (heading)
				from
					list.start
				until
					list.after
				loop
					Result.append (item_prefix)
					Result.append (list.item.path_name)
					Result.append (su.colon_space)
					Result.append (list.item.description)
					Result.append ("%N")
					list.forth
				end
				Result.append (footing)
			else
				create Result.make (0)
			end
		end

feature -- Modification

	set_directory_symlinks_followed (b: BOOLEAN)
			-- Set `directory_symlinks_followed' to `b'
		do
			directory_symlinks_followed := b
		ensure
			value_set: directory_symlinks_followed = b
		end

	set_processed_alphabetically
			-- Set `processed_alphabetically' to true
		do
			processed_alphabetically := True
			processed_natural_order := False
		ensure
			value_set: processed_alphabetically
		end

	set_processed_natural_order
			-- Set `processed_natural_order' to true
		do
			processed_alphabetically := False
			processed_natural_order := True
		ensure
			value_set: processed_natural_order
		end

	set_root_excluded (b: BOOLEAN)
			-- Set `root_excluded' to `b'
		do
			root_excluded := b
		ensure
			value_set: root_excluded = b
		end

	set_files_excluded (b: BOOLEAN)
			-- Set `files_excluded' to `b'
		do
			files_excluded := b
		ensure
			value_set: files_excluded = b
		end

	set_directories_excluded (b: BOOLEAN)
			-- Set `directories_excluded' to `b'
		do
			directories_excluded := b
		ensure
			value_set: directories_excluded = b
		end

	set_errors_displayed (b: BOOLEAN)
			-- Set `errors_displayed' to `b'
		do
			errors_displayed := b
		ensure
			value_set: errors_displayed = b
		end

	set_wildcard_filters (list: LIST [STRING])
			-- Set `wildcard_filters' to `list'
		do
			wildcard_filters := list
		ensure
			value_set: wildcard_filters = list
		end

	set_ignored_wildcard_filters (list: LIST [STRING])
			-- Set `ignored_wildcard_filters' to `list'
		do
			ignored_wildcard_filters := list
		ensure
			value_set: ignored_wildcard_filters = list
		end

feature -- Operations

	traverse_preorder
			-- Do preorder traversal of directory tree
			-- (parent node visited before children)
		do
			traverse_general (True)
		end

	traverse_postorder
			-- Do postorder traversal of directory tree
			-- (parent node visited after children)
		do
			traverse_general (False)
		end

	display_errors (heading, footing, item_prefix: STRING)
			-- Display errors in `error_list', preceding
			-- each item by `item_prefix'.  Precede the
			-- whole list by `heading' and follow it by
			-- `footing' if there is anything in the list
		do
			io.put_string (error_list_string (heading, footing, item_prefix))
		end

	visit_directory (parent, entry, full_path: STRING)
			-- Visit directory with full pathname
			-- `full_path', whose parent directory name is
			-- `parent' (Void if no parent) and whose
			-- entry in parent directory is `entry'.
		require
			entry_exists: entry /= Void
			entry_not_empty: not entry.is_empty
			full_path_exists: full_path /= Void
			full_path_nonempty: not full_path.is_empty
		deferred
		end

	visit_file (parent, entry, full_path: STRING)
			-- Visit file with full pathname `full_path',
			-- whose parent directory name is `parent'
			-- (empty if unknown) and whose entry in
			-- parent directory is `entry'.
		require
			parent_exists: parent /= Void
			entry_exists: entry /= Void
			entry_not_empty: not entry.is_empty
			full_path_exists: full_path /= Void
			full_path_nonempty: not full_path.is_empty
		deferred
		end

	process_error (full_path, err_text: STRING)
			-- Handle error that occurred on `full_path',
			-- with error description `err_text'.  Only called
			-- if `errors_displayed' is true
		require
			full_path_exists: full_path /= Void
			full_path_nonempty: not full_path.is_empty
			error_text_exists: err_text /= Void
		deferred
		end

feature {NONE} -- Implementation

	last_dir_status: INTEGER
			-- Status from last call to `get_directory_status'.
			-- 0 if `file' is a file or symbolic link pointing
			-- to nonexistent path, 1 if it is a directory
			-- 2 if error occurred getting its status
	
	depth: INTEGER
			-- Depth of traversal

	traverse_general (preorder: BOOLEAN)
			-- Do traversal of directory tree (preorder
			-- traversal if `preorder' is true, otherwise
			-- postorder)
		local
			parent, entry: STRING
		do
			visited_count := 0
			error_count := 0
			depth := 0
			create {ARRAYED_LIST [PATH_ERROR]} error_list.make (10)
			parent := fu.directory_part (root_directory)
			if root_directory.count = 1 and root_directory.item (1) =  Directory_separator then
				parent := Void
			elseif parent.count > 1 then
				parent.keep_head (parent.count - 1)
			end
			entry := fu.file_name_part (root_directory)
			traverse_tree (parent, entry, preorder)
		end

	traverse_tree (parent, entry: STRING preorder: BOOLEAN)
			-- Traverse sub-tree rooted at sub-directory `entry'
			-- of parent directory `parent', using
			-- a preorder traversal if `preorder' is true
			-- and a postorder traversal otherwise
		require
			entry_exists: entry /= Void
		local
			full_path: STRING
			follow: BOOLEAN
		do
			depth := depth + 1
			if parent /= Void then
				full_path := os.full_file_name (parent, entry)
			else
				create full_path.make (1)
				full_path.extend (Directory_separator)
				full_path.append (entry)
			end
			follow := not unix_os.is_symbolic_link (full_path) or directory_symlinks_followed
			if not is_ignored (entry) then
				if preorder then
					if depth > 1 or not root_excluded then
						safe_visit_directory (parent, entry, full_path)
					end
					if follow then
						process_directory (full_path, preorder)
					end
				else
					if follow then
						process_directory (full_path, preorder)
					end
					if depth > 1 or not root_excluded then
						safe_visit_directory (parent, entry, full_path)
					end
				end
			end
			depth := depth - 1
		end

	process_directory (path: STRING preorder: BOOLEAN)
			-- Process directory rooted at `path', using
			-- a preorder traversal if `preorder' is true
			-- and a postorder traversal otherwise
		local
			entry, full_path: STRING
			list: LIST [STRING]
			file: LARGE_RAW_FILE
		do
			list := safe_directory_contents (path)

			if list /= Void then
				create file.make ("dummy")
				from
					list.start
				until
					list.after
				loop
					entry := list.item
					if not is_ignored (entry) then
						full_path := os.full_file_name (path, entry)
						file.make (full_path)
						get_directory_status (file)
						inspect last_dir_status
						when 0 then	-- File or bad link
							safe_visit_file (path, entry, full_path)
						when 1 then	-- Directory
							traverse_tree (path, entry, preorder)
						when 2 then	-- Error occurred
							-- Already handled
						end
					end
					list.forth
				end
			end
		end

	get_directory_status (file: LARGE_RAW_FILE)
			-- Get directory status of `file' and put in
			-- `last_dir_status'.
		require
			file_not_void: file /= Void
		local
			tried, know_symlink, is_symlink: BOOLEAN
		do
			if not tried then
				last_dir_status := 0
				is_symlink := unix_os.is_symbolic_link (file.name)
				know_symlink := True
				if file.is_directory then
					last_dir_status := 1
				else
					last_dir_status := 0
				end
			else
				if know_symlink and is_symlink then
					if directory_symlinks_followed then
						handle_error (file.name)
					end
					last_dir_status := 0
				else
					handle_error (file.name)
					last_dir_status := 2
				end
			end
		rescue
			tried := True
			retry
		end

	safe_directory_contents (path: STRING): LIST [STRING]
			-- Contents of directory `path'.  Will not raise
			-- an exception
		local
			tried: BOOLEAN
		do
			if not tried then
				Result := directory_contents (path)
			else
				Result := Void
				handle_error (path)
			end
		rescue
			tried := True
			retry
		end

	directory_contents (path: STRING): LIST [STRING]
			-- Contents of directory `path'
		local
			dir: LARGE_DIRECTORY
			entry: STRING
		do
			if processed_alphabetically then
				create {SORTED_TWO_WAY_LIST [STRING]} Result.make
			else
				create {ARRAYED_LIST [STRING]} Result.make (50)
			end
			create dir.make_open_read (path)
			from
				dir.readentry
			until
				dir.lastentry = Void
			loop
				entry := dir.lastentry
				if not entry.is_equal (Current_dir_name) and
				   not entry.is_equal (Parent_dir_name) then
					Result.extend (entry)
				end
				dir.readentry
			end
			dir.close
		end

	safe_visit_directory (parent, entry, full_path: STRING)
			-- Visit directory with full pathname `full_path',
			-- whose parent directory name is `parent' and
			-- whose entry in parent directory is `entry'.
			-- Does not raise any exceptions
		require
			entry_exists: entry /= Void
			entry_not_empty: not entry.is_empty
			full_path_exists: full_path /= Void
			full_path_nonempty: not full_path.is_empty
		local
			tried: BOOLEAN
		do
			if not tried then
				if not directories_excluded and matches_wildcard (entry, wildcard_filters) then
					visit_directory (parent, entry, full_path)
					visited_count := visited_count + 1
				end
			else
				handle_error (full_path)
			end
		rescue
			tried := True
			retry
		end

	safe_visit_file (parent, entry, full_path: STRING)
			-- Visit file with full pathname `full_path',
			-- whose parent directory name is `parent' and
			-- whose entry in parent directory is `entry'.
			-- Does not raise any exceptions
		require
			parent_exists: parent /= Void
			entry_exists: entry /= Void
			entry_not_empty: not entry.is_empty
			full_path_exists: full_path /= Void
			full_path_nonempty: not full_path.is_empty
		local
			tried: BOOLEAN
		do
			if not tried then
				if not files_excluded and matches_wildcard (entry, wildcard_filters) then
					visit_file (parent, entry, full_path)
					visited_count := visited_count + 1
				end
			else
				handle_error (full_path)
			end
		rescue
			tried := True
			retry
		end

	matches_wildcard (entry: STRING list: LIST [STRING]): BOOLEAN
			-- Does `entry' match at least one of the filters
			-- in `list'?  Result is true if
			-- `list' is Void (i.e., if there are
			-- no filters)
		require
			entry_exists: entry /= Void
		do
			if list = Void then
				Result := True
			else
				from
					list.start
				until
					list.after or Result
				loop
					Result := Result or su.wildcard_matches (entry, list.item)
					list.forth
				end
			end
		end

	is_ignored (entry: STRING): BOOLEAN
			-- Should entry be completely ignored?
		require
			entry_exists: entry /= Void
		do
			Result := ignored_wildcard_filters /= Void and then
				not ignored_wildcard_filters.is_empty and then
				matches_wildcard (entry, ignored_wildcard_filters)
		end

	handle_error (full_path: STRING)
			-- Handle error that occurred on `full_path'
		local
			err: PATH_ERROR
			descr: STRING
		do
			error_count := error_count + 1
			descr := mu.exception_explanation
			if errors_displayed then
				process_error (full_path, descr)
			else
				create err.make (full_path, descr)
				error_list.extend (err)
			end
		end

	Current_dir_name: STRING = "."
			-- Name for current directory

	Parent_dir_name: STRING = ".."
			-- Name for parent directory


end
