indexing
	description:
		"Rectangular figures."
	author: "Barden E. Smith"
	date:	"September 15, 2004"

class
	DRW_BASIC_RECTANGLE

inherit
	MORE_DOUBLE_MATH
		export
			{NONE} all
		redefine
			out
		end
	ANY
		redefine
			out
		end

create
	make,
	make_by_coordinates


feature {NONE} -- Creation

	make (a_width, a_height, c_x, c_y: DOUBLE) is
			-- Create a rectangle with width `a_width' and height `a_height'
			--  and with center at (`c_x',`c_y') with
		require
			abs_width: a_width >= 0
			abs_height: a_height >= 0
		do
			x := c_x
			y := c_y
			width := a_width
			height := a_height
		end


	make_by_coordinates (c1, c2: DRW_COORDINATE) is
			-- Create a rectangle from two points (`c1', `c2')
			-- i.e. the smallest rectangle containing both points
		require
			c1_exists: c1 /= Void
			c2_exists: c2 /= Void
		do
			x := 0.5 * (c1.x + c2.x)
			y := 0.5 * (c1.y + c2.y)
			width := (c1.x - c2.x).abs
			height := (c1.y - c2.y).abs
		ensure
			has_corners: has (c1) and has (c2)
		end

	make_from_rectangle (r: DRW_BASIC_RECTANGLE) is
			-- Create a rectangle from another rectangle `r'
		require
			r_exists: r /= Void
		do
			x := r.x
			y := r.y
			width := r.width
			height := r.height
		ensure
			same_center: x = r.x and y = r.y
			same_size: width = r.width and height = r.height
		end

	make_from_const_rectangle (r: CONST_RECTANGLE) is
			-- Create a rectangle from a constant rectangle `r'
		require
			r_exists: r /= Void
		do
			x := r.center_x
			y := r.center_y
			width := r.size_x
			height := r.size_y
		ensure
			same_center: x = r.center_x and y = r.center_y
			same_size: width = r.size_x and height = r.size_y
		end

feature -- Access

	x: DOUBLE	-- center x position

	y: DOUBLE	-- center y position

	width: DOUBLE	-- size along x-axis

	height: DOUBLE	-- size along y-axis

	area: DOUBLE is
			-- Area of the rectangle
		do
			Result := width * height
		end

	left: DOUBLE is
			-- Horizontal position of left side
		do
			Result := x - (0.5 * width)
		end
		
	top: DOUBLE is
			-- Vertical position of top.
		do
			Result := y + (0.5 * height)
		end
		
	right: DOUBLE is
			--  Horizontal position of right side.
		do
			Result := x + (0.5 * width)
		end

	bottom: DOUBLE is
			-- Vertical position of bottom.
		do
			Result := y - (0.5 * height)
		end


feature -- Coordinates

	upper_left: DRW_COORDINATE is
			-- Upper-left corner of `Current'.
		do
			create Result.set (left, top)
		ensure
			Result_exists: Result /= Void
			Result_assigned: Result.x = left and then Result.y = top
		end

	upper_right: DRW_COORDINATE is
			-- Upper-right corner of `Current'.
		do
			create Result.set (right, top)
		ensure
			Result_exists: Result /= Void
			Result_assigned: Result.x = right and then Result.y = top
		end

	lower_left: DRW_COORDINATE is
			-- Lower-left corner of `Current'.
		do
			create Result.set (left, bottom)
		ensure
			Result_exists: Result /= Void
			Result_assigned: Result.x = left and then Result.y = bottom
		end

	lower_right: DRW_COORDINATE is
			-- Lower-right corner of `Current'.
		do
			create Result.set (right, bottom)
		ensure
			Result_exists: Result /= Void
			Result_assigned: Result.x = right and then Result.y = bottom
		end

feature -- Status report

	has_xy (a_x, a_y: DOUBLE): BOOLEAN is
			-- Is (`a_x', `a_y') inside `Current'?
		do
			Result :=
				(a_x >= left or else virtually_equal (a_x, left)) and then
				(a_x <= right or else virtually_equal (a_x, right)) and then
				(a_y >= bottom or else virtually_equal (a_y, bottom)) and then
				(a_y <= top) or else virtually_equal (a_y, top)
		end

	has (c: DRW_COORDINATE): BOOLEAN is
			-- Is `c' inside `Current'?
		require
			c_not_void: c /= Void
		do
			Result := has_xy (c.x, c.y)
		end

	intersects (other: like Current): BOOLEAN is
			-- Does `other' at least partially overlap `Current'?
		require
			other_not_void: other /= Void
		local
			y_test: BOOLEAN
		do
			if left <= other.left then
				y_test := right >= other.left
			else
				y_test := left <= other.right
			end
			if y_test then
				if bottom <= other.top then
					Result := top >= other.bottom
				else
					Result := bottom <= other.top
				end
			end
		end

feature -- Element change

	include_coordinate (c: DRW_COORDINATE) is
			-- Enlarge so that `c' is in `Current'.
		require
			c_not_void: c /= Void
		do
			include (c.x, c.y)
		ensure
			has_c: has (c)
		end

	include (a_x, a_y: DOUBLE) is
			-- Enlarge so that `a_x', `a_y' is in `Current'.
		do
			set_left (left.min (a_x))
			set_top (top.max (a_y))
			set_right (right.max (a_x))
			set_bottom (bottom.min (a_y))
		ensure
			has_point: has_xy (a_x, a_y)
		end

	merge (other: DRW_BASIC_RECTANGLE) is
			-- Enlarge `Current' so that `other' fits inside.
		require
			other_not_void: other /= Void
		do
			set_left (left.min (other.left))
			set_top (top.max (other.top))
			set_right (right.max (other.right))
			set_bottom (bottom.min (other.bottom))
		end

	grow_left (u: DOUBLE) is
			-- Increment size by `u' to west.
		require
			width + u > 0
		do
			x := x - (0.5 * u)
			width := width + u
		end

	grow_right (u: DOUBLE) is
			-- Increment size by `u' to east.
		require
			width + u > 0
		do
			x := x + (0.5 * u)
			width := width + u
		end

	grow_bottom (u: DOUBLE) is
			-- Increment size by `u' to south.
		require
			height + u > 0
		do
			y := y - (0.5 * u)
			height := height + u
		end

	grow_top (u: DOUBLE) is
			-- Increment size by `u' to north.
		require
			height + u > 0
		do
			y := y + (0.5 * u)
			height := height + u
		end

	set_left (u: DOUBLE) is
			-- Assign `u' to `left'.
		require
			u <= right
		local
			r: DOUBLE
		do
			r := right
			width := r - u
			x := r - (0.5 * width)
		ensure
			assigned: virtually_equal (left, u)
			right_same: virtually_equal (right, old right)
		end

	set_right (u: DOUBLE) is
			-- Assign `u' to `right'.
		require
			u >= left
		local
			l: DOUBLE
		do
			l := left
			width := u - l
			x := l + (0.5 * width)
		ensure
			assigned: virtually_equal (right, u)
			left_same: virtually_equal (left, old left)
		end

	set_bottom (u: DOUBLE) is
			-- Assign `u' to `bottom'.
		require
			u <= top
		local
			t: DOUBLE
		do
			t := top
			height := t - u
			y := t - (0.5 * height)
		ensure
			assigned: virtually_equal (bottom, u)
			top_same: virtually_equal (top, old top)
		end

	set_top (u: DOUBLE) is
			-- Assign `u' to `top'.
		require
			u >= bottom
		local
			b: DOUBLE
		do
			b := bottom
			height := u - b
			y := b + (0.5 * height)
		ensure
			assigned: virtually_equal (top, u)
			bottom_same: virtually_equal (bottom, old bottom)
		end

	set_x (new_x: DOUBLE) is
			-- Assign `new_x' to `x'.
		do
			x := new_x
		ensure
			x_set: x = new_x
		end

	set_y (new_y: DOUBLE) is
			-- Assign `new_y' to `y'.
		do
			y := new_y
		ensure
			y_set: y = new_y
		end

	set_width (new_width: DOUBLE) is
			-- Assign `new_width' to 'width'.
		require
			new_width_positive: new_width >= 0
		do
			width := new_width
		ensure
			width_assigned: width = new_width
		end

	set_height (new_height: DOUBLE) is
			-- Assign `new_height' to `height'.
		require
			new_height_positive: new_height >= 0
		do
			height := new_height
		ensure
			height_assigned: height = new_height
		end

	resize (a_width, a_height: DOUBLE) is
			-- Resize to `a_width' and `a_height'.
		require
			a_width_positive: a_width >= 0
			a_height_positive: a_height >= 0
		do
			width := a_width
			height := a_height
		ensure
			width_assigned: width = a_width
			height_assigned: height = a_height
		end

	move (a_x, a_y: DOUBLE) is
			-- Move to `a_x' and `a_y'.
		do
			x := a_x
			y := a_y
		ensure
			x_assigned: x = a_x
			y_assigned: y = a_y
		end

	move_and_resize (a_x, a_y, a_width, a_height: DOUBLE) is
			-- Move to `a_x' and `a_y' and resize to `a_width' and `a_height'.
		require
			a_width_positive: a_width >= 0
			a_height_positive: a_height >= 0
		do
			move (a_x, a_y)
			resize (a_width, a_height)
		end

feature -- Tangled web

	as_const_rectangle: CONST_RECTANGLE is
		do
			create Result.make_by_corners (left, bottom, right, top)
		end

feature -- Output

	out: STRING is
			-- Return readable string.
		do
			Result := "(X: " + x.out + ", Y: " + y.out +
				", Width: " + width.out +
				", Height: " + height.out + ")"
		end


invariant
	width_non_negative: width >= 0
	height_non_negative: height >= 0
	left_left_of_right: left <= right
	bottom_below_top: bottom <= top


end -- class DRW_BASIC_RECTANGLE
