indexing
	description:
		"Rectangular figures with rounded (elliptical) corners."
	author: "Barden E. Smith"
	date:	"August 13, 2004"

class
	DRW_SOFT_RECTANGLE

inherit
	DRW_RECTANGLE
		redefine
			as_transformed_points,
			project,
			out
		end

create
	make,
	make_by_coordinates,
	make_by_size_center_corners,
	make_from_rectangle


feature -- Creation

	make_by_size_center_corners (a_width, a_height, a_x, a_y, a_rx, a_ry: DOUBLE) is
			-- Create a rectangle with width `a_width' and height `a_height'
			--  and with center at (`c_x',`c_y') having
			--  rounded corners (elipical) with axis of
			--  `a_rx' in x and `a_ry' in y
		require
			abs_width: a_width >= 0
			abs_height: a_height >= 0
			abs_rx: a_rx >= 0
			small_rx: a_rx <= 0.5 * a_width
			abs_ry: a_ry >= 0
			small_ry: a_ry <= 0.5 * a_height
		do
			make (a_width, a_height, a_x, a_y)
			set_corner_radii (a_rx, a_ry)

		end


feature -- Access


	rx: DOUBLE is
			-- Radius along x-axis.
		do
			if rxf = 0 and ryf > 0 then
				Result := ry
			else
				Result := rxf * width
			end
		end


	ry: DOUBLE is
			-- Radius along y-axis.
		do
			if ryf = 0 and rxf > 0 then
				Result := rx
			else
				Result := ryf * height
			end
		end


feature -- Modification


	set_corner_radii (n_rx, n_ry: DOUBLE) is
		require
			abs_rx: n_rx >= 0
			small_rx: n_rx <= 0.5 * width
			abs_ry: n_ry >= 0
			small_ry: n_ry <= 0.5 * height
		do
			rxf := n_rx / width
			ryf := n_ry / height
		end

	set_rx (n_rx: DOUBLE) is
		require
			abs_r: n_rx >= 0
			small_r: n_rx <= 0.5 * width
		do
			rxf := n_rx / width
		ensure
			fits: 2.0 * rx <= width
		end

	set_ry (n_ry: DOUBLE) is
		require
			abs_r: n_ry >= 0
			small_r: n_ry <= 0.5 * height
		do
			ryf := n_ry / height
		ensure
			fits: 2.0 * ry <= height
		end


feature -- Transformation			


	as_transformed_points (xform: DRW_MATRIX): SIMPLE_LIST_OF_POINTS is
			-- return points as transformed by local transform and
			-- then by `xform'
			-- if `Result' has 4 entries then it is a regular rectangle
			-- otherwise it has 12 points with the 2nd, 5th, 8th, and 11th point
			-- are the control points for the elipical corners (i.e. it is where
			-- the corners would be if the corners were not rounded.
		local
			m: DRW_MATRIX
		do
			if rxf = 0 or ryf = 0 then
				Result := precursor (xform)
			else
				if is_transformed then
					m := transform.and_then (xform)
				else
					m := xform
				end
				create Result.make (12)
				Result.extend (m.transform_xy (right - rx, bottom))
				Result.extend (m.transform_xy (right, bottom))
				Result.extend (m.transform_xy (right, bottom + ry))
				Result.extend (m.transform_xy (right, top - ry))
				Result.extend (m.transform_xy (right, top))
				Result.extend (m.transform_xy (right - rx, top))
				Result.extend (m.transform_xy (left + rx, top))
				Result.extend (m.transform_xy (left, top))
				Result.extend (m.transform_xy (left, top - ry))
				Result.extend (m.transform_xy (left, bottom + ry))
				Result.extend (m.transform_xy (left, bottom))
				Result.extend (m.transform_xy (left + rx, bottom))
			end
		end


feature {DRW_PROJECTOR} -- plotting routine

	project (plotter: DRW_FIGURE_DRAWING_ROUTINES) is
		do
			if is_drawable then
				plotter.draw_figure_soft_rectangle (current)
			end
		end

feature -- Output

	out: STRING is
			-- Return readable string.
		do
			Result := precursor {DRW_RECTANGLE}
			if rxf = 0 and ryf = 0 then
				-- Simple rectangle
			else
				Result.append (" R Corners (")
				Result.append (rx.out)
				Result.append (",")
				Result.append (ry.out)
				Result.append (")")
			end
		end

feature {NONE} -- Implementation

	rxf: DOUBLE
			-- Radius as fraction of width
			--  between 0 and 0.5

	ryf: DOUBLE
			-- Radius as fraction of height
			--  between 0 and 0.5

end -- class DRW_SOFT_RECTANGLE
