
indexing
	
	description: "Action that extends both ends of a scribe to the first %
		%meetable edge if not already at a meetable edge, but %
		%without extending past an unmeetable edge"
	author: "David Hollenberg"
	date: "January 17, 2007"

class EXTEND_SCRIBE_ACTION
inherit
	SCRIBE_EDGE_ACTION
		redefine
			default_create,
			closest_meetable_edge
		end
	SCRIBE_MULTIPLE_CHANGE_ACTION
		redefine
			default_create
		end

feature {NONE} -- Creation

	default_create
		do
			create resize_amount.make_xy (0, 0)
		end

feature -- Properties

	changed: BOOLEAN
			-- Were any scribes extended since last call to
			-- `initialize'?

	is_horizontal: BOOLEAN
			-- Are horizontal scribes being extended, rather
			-- than vertical scribes?

	selected_only: BOOLEAN
			-- Are just selected scribes being extended, rather
			-- than all scribes

feature -- Modification

	set_is_horizontal (a_value: like is_horizontal)
			-- Set `is_horizontal' to `a_value'
		do
			is_horizontal := a_value
		ensure
			is_horizontal_set: is_horizontal = a_value
		end

	set_selected_only (a_value: like selected_only)
			-- Set `selected_only' to `a_value'
		do
			selected_only := a_value
		ensure
			is_selected_only_set: selected_only = a_value
		end

	initialize
			-- Initialize `Current'
		do
			changed := False
		end

feature -- Operations

	apply (an_element: RETICLE_VIEW_ELEMENT)
			-- Apply action to `an_element'
		local
			l_needs_extend, l_was_extended: BOOLEAN
		do
			if (not selected_only or an_element.is_selected) and an_element.intersects_any_field then
				-- Try to extend each end, if any part of
				-- it falls truly inside a field
				
				l_needs_extend := False
				l_was_extended := False
				create {ARRAYED_LIST [RETICLE_SIMPLE_CHANGE]} change_list.make (2)
				
				extend_one_end (an_element, True)
				l_needs_extend := l_needs_extend or needs_extend
				l_was_extended := l_was_extended or was_extended
				extend_one_end (an_element, False)
				l_needs_extend := l_needs_extend or needs_extend
				l_was_extended := l_was_extended or was_extended
				if l_needs_extend then
					if l_was_extended then
						an_element.view.changed_scribes.extend (an_element)
						add_change (an_element)
					else
						an_element.view.locked_scribes.extend (an_element)
					end
				end
			end
		end

feature {NONE} -- Implementation

	extend_one_end (an_element: RETICLE_VIEW_ELEMENT is_low_end: BOOLEAN)
			-- Extend one end of scribe `an_element'.  Extend
			-- left or bottom end if `is_low_end', otherwise
			-- extend top or right end.  Do nothing if no part
			-- of end is truly inside a field of view
		local
			field: RETICLE_FIELD
			edges: LIST [MODEL_ELEMENT_EDGE]
			low, high, pos, dist: DOUBLE
			next_edge: MODEL_ELEMENT_EDGE
		do
			needs_extend := False
			was_extended := False
			if is_horizontal then
				if is_low_end then
					field := an_element.field_containing_left_edge
				else
					field := an_element.field_containing_right_edge
				end
				if field /= Void then
					edges := edges_for_field (field).vertical_edges
					low := grid.rounded_edge (an_element.bottom_y)
					high := grid.rounded_edge (an_element.top_y)
					if is_low_end then
						pos := grid.rounded_edge (an_element.left_x)
					else
						pos := grid.rounded_edge (an_element.right_x)
					end
					next_edge := closest_meetable_edge (pos, low, high, edges, is_low_end)
					if next_edge /= Void then
						-- Needs to be extended
						needs_extend := True
						dist := (next_edge.position - pos).abs
						if is_low_end then
							resize_amount.make_xy (-dist, 0)
						else
							resize_amount.make_xy (dist, 0)
						end
						if an_element.is_movable_in_x then
							-- Extend if not locked
							an_element.resize_by (resize_amount, is_low_end)
							an_element.set_view_coordinates
							change_list.extend (create {SCRIBE_END_POSITION_CHANGE}.make (an_element, resize_amount.x, resize_amount.y, is_low_end))
							was_extended := True
						end
					end
				end
			else
				if is_low_end then
					field := an_element.field_containing_bottom_edge
				else
					field := an_element.field_containing_top_edge
				end
				if field /= Void then
					edges := edges_for_field (field).horizontal_edges
					low := grid.rounded_edge (an_element.left_x)
					high := grid.rounded_edge (an_element.right_x)
					if is_low_end then
						pos := grid.rounded_edge (an_element.bottom_y)
					else
						pos := grid.rounded_edge (an_element.top_y)
					end
					next_edge := closest_meetable_edge (pos, low, high, edges, is_low_end)
					if next_edge /= Void then
			
						-- Needs to be extended
						needs_extend := True
						dist := (next_edge.position - pos).abs
						if is_low_end then
							resize_amount.make_xy (0, -dist)
						else
							resize_amount.make_xy (0, dist)
						end
						if an_element.is_movable_in_y then
							-- Extend if not locked
							an_element.resize_by (resize_amount, is_low_end)
							an_element.set_view_coordinates
							change_list.extend (create {SCRIBE_END_POSITION_CHANGE}.make (an_element, resize_amount.x, resize_amount.y, is_low_end))
							was_extended := True
						end
					end
				end
			end
		end

	closest_meetable_edge (pos, low, high: DOUBLE edges: LIST [MODEL_ELEMENT_EDGE] is_low_end: BOOLEAN): MODEL_ELEMENT_EDGE
			-- Closest edge to end of `an_element' at
			-- `pos' (which direction to search is
			-- indicated by `is_low_end' - bottom or left
			-- end if true) that truly intersects segment
			-- (`low', `high') or Void if closest edge is
			-- same as `pos' or is not meetable (i.e., Void if
			-- end is not to be extended)
		do
			Result := precursor (pos, low, high, edges, is_low_end)
			if Result /= Void and then Result.position = pos then
				-- End already meets closest meetable edge
				Result := Void
			end
		end

	needs_extend: BOOLEAN
			-- Did last scribe passed to `extend_one_end' need
			-- to be extended (though it might not have been
			-- extended if it was locked in length dimension)?

	was_extended: BOOLEAN
			-- Was last scribe passed to `extend_one_end' extended?

	resize_amount: POINT
			-- Point that specifies how much scribe was resized,
			-- if `was_extended'

end
