
indexing
	
	author: "David Hollenberg"
	date: "May 3, 2002"
	description: "A file descriptor which is to be associated with a file %
		%in spawned process"

class FD_FILE

inherit
	PROCESS_FD
		redefine
			is_file, file_name, move_source_descriptor
		end

create
	make_read_text, make_read_binary, make_read_file,
	make_write_text, make_write_binary, make_write_file,
	make_append_text, make_append_binary, make_append_file

feature {NONE} -- Creation

	make_read_text (fd: INTEGER fname: STRING)
			-- Create `Current' for file descriptor `fd',
			-- which is to be associated with text file named
			-- `fname' opened for read
		require
			valid_descriptor: valid_file_descriptor (fd)
			valid_file_name: valid_file_name (fname)
		do
			number := fd
			file_name := fname
			is_text := True
			open_mode := Read_mode
		ensure
			number_set: number = fd
			file_name_set: file_name = fname
		end

	make_read_binary (fd: INTEGER fname: STRING)
			-- Create `Current' for file descriptor `fd',
			-- which is to be associated with binary file
			-- named `fname' opened for read
		require
			valid_descriptor: valid_file_descriptor (fd)
			valid_file_name: valid_file_name (fname)
		do
			number := fd
			file_name := fname
			is_text := False
			open_mode := Read_mode
		ensure
			number_set: number = fd
			file_name_set: file_name = fname
		end

	make_write_text (fd: INTEGER fname: STRING)
			-- Create `Current' for file descriptor `fd',
			-- which is to be associated with text file
			-- named `fname' opened for write
		require
			valid_descriptor: valid_file_descriptor (fd)
			valid_file_name: valid_file_name (fname)
		do
			number := fd
			file_name := fname
			is_text := True
			open_mode := Write_mode
		ensure
			number_set: number = fd
			file_name_set: file_name = fname
		end

	make_write_binary (fd: INTEGER fname: STRING)
			-- Create `Current' for file descriptor `fd',
			-- which is to be associated with binary file
			-- named `fname' opened for write
		require
			valid_descriptor: valid_file_descriptor (fd)
			valid_file_name: valid_file_name (fname)
		do
			number := fd
			file_name := fname
			is_text := False
			open_mode := Write_mode
		ensure
			number_set: number = fd
			file_name_set: file_name = fname
		end

	make_append_text (fd: INTEGER fname: STRING)
			-- Create `Current' for file descriptor `fd',
			-- which is to be associated with text file
			-- named `fname' opened for append
		require
			valid_descriptor: valid_file_descriptor (fd)
			valid_file_name: valid_file_name (fname)
		do
			number := fd
			file_name := fname
			is_text := True
			open_mode := Append_mode
		ensure
			number_set: number = fd
			file_name_set: file_name = fname
		end

	make_append_binary (fd: INTEGER fname: STRING)
			-- Create `Current' for file descriptor `fd',
			-- which is to be associated with binary file
			-- named `fname' opened for append
		require
			valid_descriptor: valid_file_descriptor (fd)
			valid_file_name: valid_file_name (fname)
		do
			number := fd
			file_name := fname
			is_text := False
			open_mode := Append_mode
		ensure
			number_set: number = fd
			file_name_set: file_name = fname
		end

	make_read_file (fd: INTEGER f: LARGE_FILE)
			-- Create `Current' for file descriptor `fd',
			-- which is to be associated with file `f'
			-- opened for read
		require
			valid_descriptor: valid_file_descriptor (fd)
			file_exists: f /= Void
			valid_file: f.is_closed or f.is_open_read
		do
			number := fd
			file := f
			open_mode := Read_mode
		ensure
			number_set: number = fd
		end

	make_write_file (fd: INTEGER f: LARGE_FILE)
			-- Create `Current' for file descriptor `fd',
			-- which is to be associated with file `f'
			-- opened for write
		require
			valid_descriptor: valid_file_descriptor (fd)
			file_exists: f /= Void
			valid_file: f.is_closed or f.is_open_write
		do
			number := fd
			file := f
			open_mode := Write_mode
		ensure
			number_set: number = fd
		end

	make_append_file (fd: INTEGER f: LARGE_FILE)
			-- Create `Current' for file descriptor `fd',
			-- which is to be associated with file `f'
			-- opened for append
		require
			valid_descriptor: valid_file_descriptor (fd)
			file_exists: f /= Void
			valid_file: f.is_closed or f.is_open_append
		do
			number := fd
			file := f
			open_mode := Append_mode
		ensure
			number_set: number = fd
		end

feature -- Properties

	is_file: BOOLEAN = True
			-- Is file descriptor associated with a file?

	file_name: STRING
			-- Name of file to be associated with descriptor
			-- in spawned process, if not Void

	valid_file_name (fname: STRING): BOOLEAN
			-- Is `fname' a valid file name?
		do
			Result := fname /= Void and then not fname.is_empty
		ensure
			correct_result: Result = (fname /= Void and then not fname.is_empty)
		end

feature -- Operations

	prepare_for_fork
			-- Do any preparation needed prior to a fork
			-- to setup this file descriptor in spawned
			-- process
		do
			if file_name /= Void then
				prepare_for_fork_named
			else
				prepare_for_fork_file
			end
			source_descriptor := file.descriptor
		end

	setup_for_parent
			-- Take any actions needed to setup this file
			-- descriptor in parent process after a fork
		do
			file.close
			file := Void
		end

	setup_for_child (fd_set: PROCESS_FD_SET)
			-- Take any actions needed to setup this file
			-- descriptor in child process after a fork,
			-- where `fd_set' is the entire set of file
			-- descriptors to be setup
		do
			if source_descriptor /= number then
				unix_os.duplicate_file_descriptor (source_descriptor, number)
			end
			close_source_descriptor_if_ok (fd_set)
		end

feature {PROCESS_FD_SET} -- Modification

	move_source_descriptor (new_fd: INTEGER)
			-- Move source descriptor to `new_fd', since
			-- it collides with existing descriptor
		do
			precursor (new_fd)
			file.reset (Dummy_name)
			
		end

feature {NONE} -- Implementation

	is_text: BOOLEAN
			-- Is `file_name' to be opened as a text file?

	open_mode: INTEGER
			-- Mode in which `file_name' is to be opened

	Read_mode: INTEGER = 1
			-- Code for Read mode

	Write_mode: INTEGER = 2
			-- Code for Write mode

	Append_mode: INTEGER = 3
			-- Code for Append mode

	file: LARGE_FILE
			-- File object

	Dummy_name: STRING = "dummy"
			-- Dummy file name

	prepare_for_fork_named
			-- Do any preparation needed prior to a fork
			-- to setup this file descriptor in spawned
			-- process, in case where `file_name' is not Void
		local
			tf: LARGE_PLAIN_TEXT_FILE
			bf: LARGE_RAW_FILE
		do
			if is_text then
				inspect open_mode
				when Read_mode then
					create tf.make_open_read (file_name)
				when Write_mode then
					create tf.make_open_write (file_name)
				when Append_mode then
					create tf.make_open_append (file_name)
				end
				file := tf
			else
				inspect open_mode
				when Read_mode then
					create bf.make_open_read (file_name)
				when Write_mode then
					create bf.make_open_write (file_name)
				when Append_mode then
					create bf.make_open_append (file_name)
				end
				file := bf
			end
			moved := True
		end

	prepare_for_fork_file
			-- Do any preparation needed prior to a fork
			-- to setup this file descriptor in spawned
			-- process, in case where `file_name' is Void
			-- and `file' is not Void
		do
			if file.is_closed then
				inspect open_mode
				when Read_mode then
					file.open_read
				when Write_mode then
					file.open_write
				when Append_mode then
					file.open_append
				end
				moved := True
			end
		end

end
