
indexing
	
	author: "David Hollenberg"
	date: "May 3, 2002"
	description: "A file descriptor which is to be associated with a pipe %
		%connecting spawned process to parent process"

class FD_PIPE

inherit
	PROCESS_FD
		redefine
			is_pipe, is_input, pipe_file, setup_for_child_early
		end

create

	make_input, make_output

feature {NONE} -- Creation

	make_input (fd: INTEGER)
			-- Create `Current' for file descriptor `fd',
			-- which is to be associated with pipe from which
			-- spawned process can read input
		require
			valid_descriptor: valid_file_descriptor (fd)
		do
			number := fd
			is_input := True
		ensure
			number_set: number = fd
		end

	make_output (fd: INTEGER)
			-- Create `Current' for file descriptor `fd',
			-- which is to be associated with pipe to which
			-- spawned process can write output
		require
			valid_descriptor: valid_file_descriptor (fd)
		do
			number := fd
			is_input := False
		ensure
			number_set: number = fd
		end

feature -- Properties

	is_pipe: BOOLEAN = True
			-- Is file descriptor associated with a pipe connected
			-- to parent process?
	
	is_input: BOOLEAN
			-- Will spawned process read input from pipe?
			-- rather than writing output?

	is_output: BOOLEAN
			-- Will spawned process write output to pipe?
		do
			Result := not is_input
		end

	pipe_file: LARGE_RAW_FILE
			-- File for writing (if pipe is input pipe to child)
			-- or reading (if pipe is output pipe from child)
			-- to read or write data on pipe

feature -- Operations

	prepare_for_fork
			-- Do any preparation needed prior to a fork
			-- to setup this file descriptor in spawned
			-- process
		do
			pipe := unix_os.new_pipe
			if is_input then
				source_descriptor := pipe.read_descriptor
			else
				source_descriptor := pipe.write_descriptor
			end
			moved := True
		end

	setup_for_parent
			-- Take any actions needed to setup this file
			-- descriptor in parent process after a fork
		local
			name: STRING
		do
			create name.make (Descriptor_name.count + 4)
			name.append (Descriptor_name)
			name.append_integer (number)
			create pipe_file.make (name)
			if is_input then
				pipe_file.fd_open_write (pipe.write_descriptor)
				pipe.erase_write_descriptor
				pipe.close_read_descriptor
			else
				pipe_file.fd_open_read (pipe.read_descriptor)
				pipe.erase_read_descriptor
				pipe.close_write_descriptor
			end
			pipe := Void
		end

	setup_for_child_early
			-- Take any early actions needed to setup this file
			-- descriptor in child process after a fork
		do
			if is_input then
				pipe.close_write_descriptor
			else
				pipe.close_read_descriptor
			end
		end

	setup_for_child (fd_set: PROCESS_FD_SET)
			-- Take any actions needed to setup this file
			-- descriptor in child process after a fork,
			-- where `fd_set' is the entire set of file
			-- descriptors to be setup
		do
			if is_input then
				move_desc (source_descriptor, number)
			else
				move_desc (source_descriptor, number)
			end
		end

feature {NONE} -- Implementation

	pipe: UNIX_PIPE
			-- Pipe object

	Descriptor_name: STRING = "Descriptor_"

invariant
	input_or_output: is_input xor is_output

end
