
indexing
	
	description: "File utilities"
	author: "Joel Goldberg (reworked by David Hollenberg)"
	date: "February 12, 2000"

class FILE_UTILITIES

inherit
	ANY
	STRING_UTILITIES_ACCESS
		export
			{NONE} all
		end
	EXECUTION_ENVIRONMENT_ACCESS
		export
			{NONE} all
		end
	OPERATING_ENVIRONMENT
		export
			{ANY} Directory_separator
			{NONE} all
		end
	EXCEPTIONS_ACCESS
		export
			{NONE} all
		end

feature -- Operations

	expand_shell_tilde (s: STRING)
			-- Change `s' to expand leading shell tilde.  If `s'
			-- starts with "~/" or is just "~", replace
			-- the "~" with current user's home directory.
			-- Replace leading "~user" with "/home/user"
		require
			string_exists: s /= Void
		local
			new_prefix: STRING
		do
			if not s.is_empty and then s.item (1) = '~' then
				if s.count = 1 or else s.item (2) = '/' then
					new_prefix := exe.home_directory_name
				else
					new_prefix := "/home/"
				end
				s.remove (1)
				s.prepend (new_prefix)
			end
		end

	change_file_mode (f: LARGE_FILE s: STRING)
			-- Change mode of `f' to permissions indicated
			-- by `s'.  `s' must be constructed from the
			-- string "rwxrwxrwx", where each character
			-- can be replaced by a '-' character to
			-- indicate that the corresponding permission
			-- is not granted.  First three characters
			-- represent owner's permissions, next three
			-- are group's permissions and last three are
			-- other's permissions
		require
			file_exists: f /= Void
			string_exists: s /= Void
			valid_permission_string: is_valid_permission_string (s)
		do
			f.change_mode (permission_string_to_mask (s))
		end

	opened_for_read (fname: STRING): LARGE_PLAIN_TEXT_FILE
			-- Text file object opened for read with file
			-- name `fname' or Void if `fname' is not a
			-- plain file or open fails
		require
			file_name_exists: fname /= Void
		local
			tried: BOOLEAN
		do
			if not tried then
				create Result.make (fname)
				if Result.exists and then Result.is_plain then
					Result.open_read
				else
					Result := Void
				end
			else
				Result := Void
			end
		ensure
			valid_result: Result = Void or else Result.is_open_read
		rescue
			if not exc.assertion_violation then
				tried := true
				retry
			end
		end

	opened_for_write (fname: STRING): LARGE_PLAIN_TEXT_FILE
			-- Text file object opened for write with file
			-- name `fname' or Void if open fails
		require
			file_name_exists: fname /= Void
		local
			tried: BOOLEAN
		do
			if not tried then
				create Result.make (fname)
				Result.open_write
			else
				Result := Void
			end
		ensure
			valid_result: Result = Void or else Result.is_open_write
		rescue
			if not exc.assertion_violation then
				tried := true
				retry
			end
		end

	opened_for_append (fname: STRING): LARGE_PLAIN_TEXT_FILE
			-- Text file object opened for append with file
			-- name `fname' or Void if open fails
		require
			file_name_exists: fname /= Void
		local
			tried: BOOLEAN
		do
			if not tried then
				create Result.make (fname)
				Result.open_append
			else
				Result := Void
			end
		ensure
			valid_result: Result = Void or else Result.is_open_append
		rescue
			if not exc.assertion_violation then
				tried := true
				retry
			end
		end

	delete_if_exists (fname: STRING)
			-- Delete file `fname' if it exists.  Do not raise
			-- an exception if delete fails
		require
			file_name_exists: fname /= Void
		local
			tried: BOOLEAN
			file: LARGE_PLAIN_TEXT_FILE
		do
			if not tried then
				create file.make (fname)
				if file.exists then
					file.delete
				end
			end
		rescue
			if not exc.assertion_violation then
				tried := true
				retry
			end
		end

	current_size (fname: STRING): INTEGER_64
			-- Current size of file `fname' in bytes or -1
			-- if cannot get size (presumably because file
			-- doesn't exist)
		local
			tried: BOOLEAN
			f: LARGE_RAW_FILE
		do
			if not tried then
				create f.make (fname)
				Result := f.count
			else
				Result := -1
			end
		rescue
			tried := True
			retry
		end

	last_modification_date (fname: STRING): INTEGER
			-- Last modification date of file `fname' or 0
			-- if cannot get date (presumably because file
			-- doesn't exist)
		local
			tried: BOOLEAN
			f: LARGE_RAW_FILE
		do
			if not tried then
				create f.make (fname)
				Result := f.date
			else
				Result := 0
			end
		rescue
			tried := True
			retry
		end

	file_contents (a_name: STRING): STRING
		-- Entire contents of text file as a string or
		-- Void if could not open file to read.  Won't work
		-- for large files (> 128 MB)
		require
			a_name_not_void: a_name /= Void
		local 	
			nm: STRING
			f: LARGE_PLAIN_TEXT_FILE
		do	
			nm := a_name.twin
			f := opened_for_read (nm)
			if f /= Void then
		  		f.read_stream (f.count.to_integer)
		  		Result := f.last_string.twin
		  		f.close
			end
		end

	update_name_in_mebes_header(mfil: LARGE_FILE old_name: STRING)
	-- mfil must be a mebes file. we open it; replace the file_name in
	-- the mebes start-drawing header with the actual file name of mfil
	-- (forced to upper case), and close it again.
	-- if old_name is not Void, the existing file_name in the header
	-- must match old_name.
	-- if there are any issues, an exception is raised.
	require have_mebes: mfil /= Void
		valid_name: old_name = Void or else
			    (old_name.count = 12 and (old_name @ 10) = '.')
	do  mfil.open_read_write
	    update_name_in_open_mebes_header (mfil, old_name)
	    mfil.close
	end

	update_name_in_open_mebes_header (mfil: LARGE_FILE old_name: STRING)
			-- Replace the file name in start-drawing
			-- header of MEBES file `mfil' with the actual
			-- file name of `mfil' (forced to upper case).
			-- If `old_name' is not Void, the existing
			-- file name in the header must match it or an
			-- exception is raised
		require
			mfil_exists: mfil /= Void
			mfil_has_name: mfil.name /= Void and then not mfil.name.is_empty
			mfil_open_read_write: mfil.is_open_read and mfil.is_open_write
			valid_name: old_name = Void or else
			    (old_name.count = 12 and (old_name @ 10) = '.')
		local
			mnam: STRING
			nloc: INTEGER
			mode: INTEGER
		do
	    		mfil.go (0)
	    		mfil.read_character	-- Data-field count
	    		mfil.read_character	-- Mode
	    		mode := mfil.last_character.code

	    		-- Position of the file name is a function of the mode
			
	    		if mode = 3 or mode = 6 then
				nloc := 14
	    		elseif mode = 5 or mode = 1 then
				nloc := 24
	     		else
				exc.raise (su.string_of_3
				("update_name_in_mebes_header: ",
				 mfil.name, " has unknown mode byte"))
	    		end

	    		-- Test for old name if provided
			
	    		if su.not_empty (old_name) then
				mfil.go (nloc)
		  		mfil.read_stream (12)
		  		if not equal (mfil.last_string, old_name) then
					exc.raise (su.string_of_6
						("update_name_in_mebes_header: ",
				 		mfil.name, " internal name is ",
				 		mfil.last_string, "; we expected ",
				 		old_name))
		  		end
	    		end

	    		-- Get new name to put into the header

	    		mnam := file_name_part (mfil.name) -- Use true name
	    		mnam.to_upper			   -- but in UC
	    		if mnam.count /=12 or else (mnam @ 10) /= '.' then
				exc.raise (su.string_of_3
				("update_name_in_mebes_header: ",
				 mfil.name, " not a mebes format name"))
	    		end

			-- Put new name in header

	    		mfil.go (nloc)
	    		mfil.put_string (mnam)
	end

feature -- Properties

	shell_tilde_expanded (s: STRING): STRING
			-- New string with leading shell tilde
			-- expanded.  If `s' starts with "~/" or is
			-- just "~", replace the "~" with current
			-- user's home directory.  Replace leading
			-- "~user" with "/home/user"
		require
			string_exists: s /= Void
		do
			Result := s.twin
			expand_shell_tilde (Result)
		ensure
			result_exists: Result /= Void
		end

	is_valid_permission_string (s: STRING): BOOLEAN
			-- Is `s' a valid file permission string?
			-- A valid string is one constructed from the
			-- string "rwxrwxrwx", where each character can
			-- be replaced by a '-'.  Also the first and second "x"
			-- can be replaced by "s" to indicate set user id and
			-- set group id
		require
			string_exists: s /= Void
		do
			Result := s.count = 9 and
				(s.item (1) = 'r' or s.item (1) = '-') and
				(s.item (2) = 'w' or s.item (2) = '-') and
				(s.item (3) = 'x' or s.item (3) = '-' or s.item (3) = 's') and
				(s.item (4) = 'r' or s.item (4) = '-') and
				(s.item (5) = 'w' or s.item (5) = '-') and
				(s.item (6) = 'x' or s.item (6) = '-' or s.item (6) = 's') and
				(s.item (7) = 'r' or s.item (7) = '-') and
				(s.item (8) = 'w' or s.item (8) = '-') and
				(s.item (9) = 'x' or s.item (9) = '-')
		end
	
	file_exists (fname: STRING): BOOLEAN
			-- Does file with name `fname' exist?
		require
			name_exists: fname /= Void
			name_not_empty: not fname.is_empty
		local
			f: LARGE_RAW_FILE
		do
			create f.make (fname)
			Result := f.exists
		end

	is_file_deletable (fname: STRING): BOOLEAN
			-- Can file with name `fname' be deleted by
			-- effective uid?
		require
			name_exists: fname /= Void
			name_not_empty: not fname.is_empty
		local
			dir: LARGE_DIRECTORY
			dir_name: STRING
		do
			dir_name := directory_part (fname)
			create dir.make (dir_name)
			Result := dir.is_writable
		end

	last_directory_separator_position (fname: STRING): INTEGER
			-- Position of last directory separator in `fname'
			-- or 0 if none
		require
			file_name_exists: fname /= Void
		local
			sep: CHARACTER
			count, n: INTEGER
		do
			sep := Directory_separator
			count := fname.count
			from
				if count > 0 then
					n := fname.index_of (sep, 1)
				end
			until
				n = 0
			loop
				Result := n
				if n < count then
					n := fname.index_of (sep, n + 1)
				else
					n := 0
				end
			end
		end

	directory_part (fname: STRING): STRING
			-- Directory part of file name `fname' (everything
			-- up to and including the last directory separator
			-- character or entire name if no separators)
		require
			file_name_exists: fname /= Void
		local
			pos: INTEGER
		do
			pos := last_directory_separator_position (fname)
			Result := fname.substring (1, pos)
		end

	file_name_part (fname: STRING): STRING
			-- File name part of file name `fname' (everything
			-- after the last directory separator
			-- character or entire name if no separators)
		require
			file_name_exists: fname /= Void
		local
			pos: INTEGER
		do
			pos := last_directory_separator_position (fname)
			Result := fname.substring (pos + 1, fname.count)
		end
	
	permission_string_to_mask (s: STRING): INTEGER
			-- Convert `s' to an integer mask suitable for passing
			-- to change_mode in class FILE.  `s' must
			-- be constructed from the string "rwxrwxrwx",
			-- where each character can be replaced by a '-'
			-- character to indicate that the corresponding
			-- permission is not granted (first and second "x"
			-- can also be "s" to indicate set user id or set
			-- group id).  First three characters
			-- represent owner's permissions, next three are
			-- group's permissions and last three are other's
			-- permissions
		require
			string_exists: s /= void
			valid_permission_string: is_valid_permission_string (s)
		do
			if s.item (1) = 'r' then
				Result := Result + 256
			end
			if s.item (2) = 'w' then
				Result := Result + 128
			end
			if s.item (3) = 'x' then
				Result := Result + 64
			elseif s.item (3) = 's' then
				Result := Result + 64 + 2048
			end
			if s.item (4) = 'r' then
				Result := Result + 32
			end
			if s.item (5) = 'w' then
				Result := Result + 16
			end
			if s.item (6) = 'x' then
				Result := Result + 8
			elseif s.item (6) = 's' then
				Result := Result + 8 + 1024
			end
			if s.item (7) = 'r' then
				Result := Result + 4
			end
			if s.item (8) = 'w' then
				Result := Result + 2
			end
			if s.item (9) = 'x' then
				Result := Result + 1
			end
		ensure
			nonnegative_result: Result >= 0
		end

end
