
indexing
	
	description: "Final job file to be sent to mask shop, to drive %
		%electron beam machine that writes masks"
	author: "David Hollenberg"
	date: "July 15, 2004"

class FINAL_JOB_FILE
inherit
	JOB_FILE

create
	default_create

feature -- Properties

	is_final: BOOLEAN = True
			-- Is job file a final one to be sent to mask shop?
			-- If not, it is only input to CATS for a merge
			-- operation

	title_orientations: LIST [STRING]
			-- List of title orientation commands

	titles: LIST [MASK_TITLES_ROW]
			-- List of titles

	prolog_canned_lines: LIST [STRING]
			-- Canned lines that go at start of section 5

	epilog_canned_lines: LIST [STRING]
			-- Canned lines that go at end of section 5

feature -- Modification

	set_title_orientations (list: LIST [STRING])
			-- Set `title_orientations' to `list'
		do
			title_orientations := list
		ensure
			value_set: title_orientations = list
		end

	set_titles (list: like titles)
			-- Set `titles' to `list'
		do
			titles := list
		ensure
			value_set: titles = list
		end

	set_prolog_canned_lines (list: like prolog_canned_lines)
			-- Set `prolog_canned_lines' to `list'
		do
			prolog_canned_lines := list
		ensure
			value_set: prolog_canned_lines = list
		end

	set_epilog_canned_lines (list: like epilog_canned_lines)
			-- Set `epilog_canned_lines' to `list'
		do
			epilog_canned_lines := list
		ensure
			value_set: epilog_canned_lines = list
		end

feature {NONE} -- Implementation

	write_titles_and_orientations
			-- Write mask titles and title orientations
		do
			title_orientations.do_all (agent write_one_title_orientation)
			titles.do_all (agent write_one_title)
		end

	write_one_title_orientation (s: STRING)
			-- Write mask title orientation `s'
		do
			append (Orient_cmd)
			indent
			append (s)
			flush
		end

	write_one_title (row: MASK_TITLES_ROW)
			-- Write mask title represented by `row'
		local
			lev, title: STRING
		do
			title := parameters.substituted_text (row.title_code_string)
			if not is_legal_title (title) then
				handle_hard_error ("Job file title contains illegal character or exceeds max length of " + Max_title_length.out + "%N" + "Title is: " + title)
			end
			append (row.job_deck_title_command)
			indent
			if row.is_null_level_no then
				-- Written on all levels
				extend ('A')
			else
				lev := row.level_no.out
				su.left_pad (lev, 2)
				append (lev)
			end
			extend (Comma)
			append (title)
			flush
		end

	is_legal_title (s: STRING): BOOLEAN
			-- Is `s' a legal title in a job file
			-- (MTITLE, DTITLE or ITITLE)?
		local
			k, n: INTEGER
		do
			n := s.count
			Result := n <= Max_title_length
			from
				k := 1
			until
				k > n or not Result
			loop
				if not is_legal_title_char (s.item (k)) then
					Result := False
				end
				k := k + 1
			end
		end

	is_legal_title_char (c: CHARACTER): BOOLEAN
			-- Is `c' a legal character in a job file title
			-- (MTITLE, DTITLE or ITITLE)?
		do
			Result := (c.is_alpha and c = c.upper) or
				c.is_digit or c = '-' or c = ' ' or c = '&' or
				c = '.'
		end

	append_extra_global_options
			-- Append extra global options needed in job file
			-- to `line'
		do
			global_mirror := mask_levels.for_all (agent {RUN_LEVELS_ROW}.geometry_mirrored)
			global_title_mirror := mask_levels.for_all (agent {RUN_LEVELS_ROW}.titles_xor_geometry_mirrored)
			global_positive_resist := mask_levels.for_all (agent positive_resist)
			if global_mirror then
				extend (Comma)
				append (all_levels (Mirror_option))
			end
			if global_title_mirror then
				extend (Comma)
				append (all_levels (Title_mirror_option))
			end
			if global_positive_resist then
				extend (Comma)
				append (all_levels (Positive_resist_option))
			end
		end

	write_all_level_options
			-- Write one line per level for levels with
			-- level-specific (nonglobal) options
		do
			mask_levels.do_all (agent write_level_options)
		end

	write_level_options (level: RUN_LEVELS_ROW)
			-- Write options that do not apply globally but
			-- do apply to `level'
		require
			level_exists: level /= Void
		local
			m_opt, t_opt, p_opt: BOOLEAN
			level_no: INTEGER
			started: BOOLEAN
		do
			m_opt := not global_mirror and level.geometry_mirrored
			t_opt := not global_title_mirror and level.titles_xor_geometry_mirrored
			p_opt := not global_positive_resist and mask_techn.positive_resist (level.polarity)
			level_no := level.level_no
			if m_opt or t_opt or p_opt then
				append (Option_cmd)
				indent
				if m_opt then
					append (Mirror_option)
					append_integer (level_no)
					started := True
				end
				if t_opt then
					if started then
						extend (Comma)
					end
					append (Title_mirror_option)
					append_integer (level_no)
				end
				if p_opt then
					if started then
						extend (Comma)
					end
					append (Positive_resist_option)
					append_integer (level_no)
				end
				flush
			end
		end

	positive_resist (level: RUN_LEVELS_ROW): BOOLEAN
			-- Does `level' need positive resist?
		do
			Result := mask_techn.positive_resist (level.polarity)
		end

	write_prolog_lines
			-- Write canned lines in prolog (before MOSIS chips)
		do
			prolog_canned_lines.do_all (agent append_and_flush)
		end

	write_epilog_lines
			-- Write canned lines in epilog (after MOSIS chips)
		do
			epilog_canned_lines.do_all (agent append_and_flush)
		end

	global_mirror: BOOLEAN
			-- Are all levels be mirrored with `Mirror_option'?

	global_title_mirror: BOOLEAN
			-- Do all levels have titles mirrored with
			-- `Title_mirror_option'?

	global_positive_resist: BOOLEAN
			-- Do all levels have positive resist?

end
