deferred class GDS_FILE_READER
	-- read a gds DATA_SOURCE (usually a file) 
	--
	-- Usage is	make(data_source)
	--	either	read_gds_library
	--	or	read_gds_header
	--		if header_is_OK		-- and you want the body
	--		 then read_gds_body
	--		end
	-- Output is	the_messages: RESPONSE_COLLECTOR
	--	and	whatever else the descendent does
		
inherit

	LAYOUT_IO_SUPPORT

	GEOMETRY_SUPPORT
	export {NONE} all
	end;

	GDS_FILE_VALUES
	export {NONE} all
	end;

feature {NONE} -- block level gds buffer management

	my_file: DATA_SOURCE
		-- file (or other source) being read

	curr_buffer: STRING
		-- current buffer (one gds block)

	block_number: INTEGER;
		-- block number (of block currently in curr_buffer)

	next_byte_pos: INTEGER
		-- position (in curr_buffer) of first unexamined byte

	have_syntax_error: BOOLEAN
		-- issue just one syntax error

	log_syntax_error(id, val: STRING) is
		-- log just one syntax error (rest are garbage)
	do  if not have_syntax_error
	     then log_io_error(id, val);
		  have_syntax_error := true
	    end
	end;

	get_next_block is
	do  if my_file.after
	     then log_syntax_error(error_is_eof, Void);
		  create curr_buffer.make(gds_record_length)
	     else my_file.read_stream(gds_record_length);
		  block_number := block_number + 1;
		  curr_buffer := my_file.last_string
	    end;
	    from
	    until curr_buffer.count >= gds_record_length
	    loop  curr_buffer.extend('%U')
	    end
	    next_byte_pos := 1		-- string index
	end;

feature	{NONE} -- gds (16 bit) word level reader

	hi_byte, lo_byte: CHARACTER
		-- current byte pair

	get_2_bytes is
		-- next byte pair of the gds file to hi_byte and lo_byte
	do  if next_byte_pos >= gds_record_length
	     then get_next_block
	    end;
	    hi_byte := curr_buffer @ next_byte_pos;
	    lo_byte := curr_buffer @ (next_byte_pos + 1);
	    next_byte_pos := next_byte_pos + 2
	end;

	unsigned_int16: INTEGER is
		-- current byte pair as 16 bit unsigned (>=0) integer
	do  Result := (hi_byte.code * two_up_8) + lo_byte.code
	end;

	signed_int16: INTEGER is
		-- current byte pair as 16 bit signed integer
	do  Result := unsigned_int16;
	    if Result >= two_up_15
	     then Result := Result - two_up_16
	    end;
	end;

feature {NONE} -- gds record reader

	-- the data below holds the "current" record. whatever routine
	-- "consumes" this record must then call read_gds_record to get
	-- the next one.

	curr_record_type,		-- rec_XXXX
	curr_data_type,			-- dta_XXXX
	curr_data_count: INTEGER	-- data byte count (record - 4)

	curr_bit_array: INTEGER			-- value(s) in one of these
	curr_int_array: R_ARRAYED_LIST[INTEGER] -- depending on
	curr_real_array: R_ARRAYED_LIST[DOUBLE]	-- curr_data_type
	curr_ascii_string: STRING

	read_gds_record is
		-- NOTE: we do not reconcile the record_type and data_type
	do  -- the byte count
	    get_2_bytes;
	    curr_data_count := unsigned_int16 - 4;
	    -- the record and data types
	    get_2_bytes;
	    curr_record_type := hi_byte.code; 
	    curr_data_type := lo_byte.code;
	    -- erase old data
	    if not curr_int_array.is_empty
	     then curr_int_array.reset end;
	    if not curr_real_array.is_empty
	     then curr_real_array.reset end;
	    curr_ascii_string := Void	-- (force new object each string)
	    -- now the data itself
	    inspect curr_data_type
	     when dta_NONE then read_dta_NONE
	     when dta_BITS then  read_dta_BITS
	     when dta_B2INT then read_dta_B2INT
	     when dta_B4INT then read_dta_B4INT
	     when dta_B4REAL then read_dta_B4REAL
	     when dta_B8REAL then read_dta_B8REAL
	     when dta_ASCII then read_dta_ASCII
	     else log_syntax_error(error_is_nodtyp, curr_data_type.out)
	    end
	    -- data count must be even and non-negative
	    if (curr_data_count < 0) or
	       (curr_data_count \\ 2 /= 0)
	     then log_syntax_error(error_is_illcnt, Void)
	    end;
	end;

	read_dta_NONE is
	do  if curr_data_count /= 0
	     then log_syntax_error(error_is_illcnt, Void) end
	end;

	read_dta_BITS is
	do  if curr_data_count /= 2
	     then log_syntax_error(error_is_illcnt, Void) end
	    get_2_bytes;
	    curr_bit_array := unsigned_int16
	end;

	read_dta_B2INT is
	local i: INTEGER;
	do  from  i := 0
	    until i >= curr_data_count
	    loop  get_2_bytes;
		  curr_int_array.extend(signed_int16);
		  i := i + 2
	    end;
	end;

	read_dta_B4INT is
	local i, val: INTEGER;
	do  if (curr_data_count \\ 4) /= 0
	     then log_syntax_error(error_is_illcnt, Void) end;
	    from  i := 0
	    until i >= curr_data_count
	    loop  get_2_bytes;		-- this should be correct on any
		  val := signed_int16;	-- machine where INTEGER is at
		  get_2_bytes;		-- least 32 bits
		  val := (val * two_up_16) + unsigned_int16;
		  curr_int_array.extend(val);
		  i := i + 4
	    end;
	end;

	read_dta_B4REAL is
	local i: INTEGER;
	do  if (curr_data_count \\ 4) /= 0
	     then log_syntax_error(error_is_illcnt, Void) end;
	    from  i := 0
	    until i >= curr_data_count
	    loop  read_one_real(false)
		  i := i + 4
	    end;
	end;

	read_dta_B8REAL is
	local i: INTEGER;
	do  if (curr_data_count \\ 8) /= 0
	     then log_syntax_error(error_is_illcnt, Void) end;
	    from  i := 0
	    until i >= curr_data_count
	    loop  read_one_real(true)
		  i := i + 8
	    end;
	end;

	read_one_real(B8: BOOLEAN) is
		-- read either a 4 byte or an 8 byte real
		-- (only difference is length of mantissa)
	local neg: BOOLEAN; exp16: INTEGER; val: DOUBLE
	do  get_2_bytes;
	    exp16 := hi_byte.code;	-- sign and excess-64 exponent
	    neg := exp16 >= two_up_7;
	    if neg then exp16 := exp16 - two_up_7 end;
	    exp16 := exp16 - 64;		-- true HEX exponenet
	    -- collect the 3 or 7 byte mantissa as an integer value
	    -- adjust exponent by number of hex digits collected
	    val := lo_byte.code;	-- start of mantissa
	    get_2_bytes;		-- 2 more bytes of mantissa
	    val := (val * two_up_16) + unsigned_int16;
	    if B8
	     then get_2_bytes;
		  val := (val * two_up_16) + unsigned_int16;
		  get_2_bytes;
		  val := (val * two_up_16) + unsigned_int16;
		  exp16 := exp16 - 14	-- 7 bytes = 14 hex digits
	     else exp16 := exp16 -  6	-- 3 bytes =  6 hex digits
	    end;
	    val := val * (16 ^ exp16);
	    if neg then val := - val end;
	    -- add to the string
	    curr_real_array.extend(val)
	end;

	read_dta_ASCII is
	local i: INTEGER;
	do  create curr_ascii_string.make(curr_data_count);
	    from  i := 0
	    until i >= curr_data_count
	    loop  get_2_bytes;
		  curr_ascii_string.extend(hi_byte)
		  curr_ascii_string.extend(lo_byte)
		  i := i + 2
	    end;
	    -- remove final null if present
	    if (curr_ascii_string @ curr_data_count) = '%U'
	     then curr_ascii_string.keep_head(curr_data_count-1)
	    end
	end;

feature {NONE} -- basic record checker

	record_is(rec_type, data_type: INTEGER): BOOLEAN is
		-- see if current record is of rec_type
		-- if it is, also insist that the data_type match
	do  Result := curr_record_type = rec_type
	    if Result and (curr_data_type /= data_type)
	     then log_syntax_error(error_is_ildtyp,
				gds_record_type_name(rec_type))
	    end
	end;

	record_must_be(rec_type, data_type: INTEGER) is
		-- as above, but add an error if it isnt
	do  if not record_is(rec_type, data_type)
	     then log_syntax_error(error_is_misrec,
				gds_record_type_name(rec_type))
	    end
	end;

	ignored_record(rec_type, data_type: INTEGER) is
		-- a record we ignore (but parse)
	do  if record_is(rec_type, data_type)
	     then read_gds_record	-- that's all we want from him!
	    end
	end;

feature {NONE} -- error message generator / manager

	error_is_absref: STRING is "ABSREF";
	error_is_illcnt: STRING is "ILLCNT";
	error_is_illdat: STRING is "ILLDAT";
	error_is_illext: STRING is "ILLEXT";
	error_is_ildtyp: STRING is "ILDTYP";
	error_is_misrec: STRING is "MISREC";
	error_is_negwid: STRING is "NEGWID";
	error_is_nodtyp: STRING is "NODTYP";
	error_is_notelm: STRING is "NOTELM";
	error_is_opnbdy: STRING is "OPNBDY";
	error_is_unkend: STRING is "UNKEND";
	error_is_dfaref: STRING is "DFAREF"

	file_position: STRING is
	do  Result := su.string_of_2("block ", block_number.out)
	end;

	stru_position: STRING is
	do  if curr_structure_name = Void
	     then Result := "not in structure"
	     else Result := su.string_of_2("in structure ",
						curr_structure_name)
	    end;
	end;

feature	-- the header attributes

	modification_date, 
	access_date: DB_DATE; 
	library_name: STRING; 
	user_unit, 
	database_unit: DOUBLE

feature -- the reader

	read_gds_header is
		-- read header or generate some errors
	do  -- Initialize everything
	    have_syntax_error := false
	    create the_messages.make(reader_application);
	    unknown_layers := lioopt.init_unknown_layers;
	    block_number := 0;
	    get_next_block;		-- prime the block reader
	    create curr_int_array.make(400);
	    create curr_real_array.make(2);
	    create sref_pt1.make_XY(0,0);	-- temporaries for S/AREF reuse
	    create aref_pt2.make_XY(0,0);
	    create aref_pt3.make_XY(0,0);
	    read_gds_record;		-- prime the gds record reader
	    -- Now start reading 
	    -- HEADER
	    record_must_be(rec_HEADER, dta_B2INT);
		-- FIX_ME: do we use the version for anything?
	    read_gds_record;
	    -- BGNLIB
	    record_must_be(rec_BGNLIB, dta_B2INT);
	    if curr_int_array.count /= 12
	     then log_syntax_error(error_is_illcnt, Void) end;
	    if not have_syntax_error
	     then modification_date := make_db_date(1);
		  access_date := make_db_date(2)
	    end;
	    read_gds_record
	    -- LIBDIRSIZE, SRFNAME, LIBSECUR
	    ignored_record(rec_LIBDIRSIZE, dta_B2INT);
	    ignored_record(rec_SRFNAME, dta_ASCII);
	    ignored_record(rec_LIBSECUR, dta_B2INT);
	    -- LIBNAME
	    record_must_be(rec_LIBNAME, dta_ASCII);
	    library_name := curr_ASCII_string;
	    read_gds_record
	    -- REFLIBS, FONTS, ATTRTABLE, GENERATIONS, FormatType
	    ignored_record(rec_REFLIBS, dta_ASCII);	    
	    ignored_record(rec_FONTS, dta_ASCII);
	    ignored_record(rec_ATTRTABLE, dta_ASCII);	    
	    ignored_record(rec_GENERATIONS, dta_B2INT);
	    ignore_Format_Type;
	    -- UNITS
	    record_must_be(rec_UNITS, dta_B8REAL);
	    if curr_real_array.count = 2
	     then user_unit  := curr_real_array @ 1;
		  database_unit := curr_real_array @ 2
	     else log_syntax_error(error_is_illcnt, Void)
	    end;
	    read_gds_record
	    -- if we have no syntax errors, we read a header successfully
	end;

	header_is_OK: BOOLEAN is
	do  Result := the_messages /= Void and (not have_syntax_error)
	end;

	read_gds_body is
	require header_is_OK
	do  begin_gds_library
	    from  
	    until have_syntax_error or not record_is(rec_BGNSTR, dta_B2INT)
	    loop  read_gds_structure
	    end;
	    record_must_be(rec_ENDLIB, dta_NONE);
	    end_gds_library
	end;

	read_gds_library is
	do  read_gds_header;
	    if header_is_OK then read_gds_body end
	end; 

feature {NONE} -- read_gds_header assistants

	make_db_date(ndx: INTEGER): ORACLE_DATE is
		-- make a date from a 6 integer entry in curr_int_array
	require ndx = 1 or ndx = 2;
	local n0, yr, mo, dy, hr, mn, sc: INTEGER
	do  create Result.set_zero;
	    n0 := 6 * (ndx - 1);
	    yr := curr_int_array @ (n0 + 1);
	    mo := curr_int_array @ (n0 + 2);
	    dy := curr_int_array @ (n0 + 3);
	    hr := curr_int_array @ (n0 + 4);
	    mn := curr_int_array @ (n0 + 5);
	    sc := curr_int_array @ (n0 + 6);
	    -- adjust yr from gds standard to real year value
	    if yr < gds_year_zero // 2 
	     then yr := yr + gds_year_zero
	    end;
	    -- if all values are valid, stuff them into Result
	    if Result.valid_year(yr)   and Result.valid_month(mo) and
	       Result.valid_day(dy, mo, yr)    and Result.valid_hour(hr)  and
	       Result.valid_minute(mn) and result.valid_second(sc)
	     then Result.set_year  (yr);
		  Result.set_month (mo);
		  Result.set_day   (dy);
		  Result.set_hour  (hr);
		  Result.set_minute(mn);
		  Result.set_second(sc)
	     else log_io_error(error_is_illdat, Void)
	    end
	end;

	ignore_Format_Type is
		-- ignore FORMAT, and also MASK and ENDMASKS
	do  if record_is(rec_FORMAT, dta_B2INT)
	     then from	read_gds_record
		  until not record_is(rec_MASK, dta_ASCII)
		  loop  read_gds_record
		  end;
		  ignored_record(rec_ENDMASKS, dta_NONE)
	    end;
	end;

feature {NONE} -- structure reader

	curr_structure_name: STRING
		-- (solely for stru_position)

	read_gds_structure is
	require record_is(rec_BGNSTR, dta_B2INT)
	local cdat, mdat: DB_DATE; snam: STRING
	do  -- BGNSTR
	    if curr_int_array.count /= 12
	     then log_syntax_error(error_is_illcnt, Void) end;
	    cdat := make_db_date(1);
	    mdat := make_db_date(2);
	    read_gds_record;
	    -- STRNAME
	    record_must_be(rec_STRNAME, dta_ASCII);
	    snam := curr_ASCII_string;
	    read_gds_record;
	    curr_structure_name := snam;
	    -- create new structure
	    begin_gds_structure(snam, cdat, mdat);
	    -- STRCLASS
	    ignored_record(rec_STRCLASS, dta_BITS)
	    -- elements
	    from  
	    until record_is(rec_ENDSTR, dta_NONE) or have_syntax_error
	    loop  if record_is(rec_BOUNDARY, dta_NONE)
		   then read_BOUNDARY
		  elseif record_is(rec_PATH, dta_NONE)
		   then read_PATH
		  elseif record_is(rec_AREF, dta_NONE)
		   then read_AorSREF(true)
		  elseif record_is(rec_SREF, dta_NONE)
		   then read_AorSREF(false)
		  elseif record_is(rec_TEXT, dta_NONE)
		   then read_TEXT
		  elseif record_is(rec_NODE, dta_NONE)
		   then read_NODE
		  elseif record_is(rec_BOX,  dta_NONE)
		   then read_BOX
		   else log_syntax_error(error_is_notelm,
				gds_record_type_name(curr_record_type))
		  end
		  -- properties (ignored)
		  from
		  until not record_is(rec_PROPATTR, dta_B2INT)
		  loop  read_gds_record;
			ignored_record(rec_PROPVALUE, dta_ASCII)
		  end
		  -- ENDEL
		  record_must_be(rec_ENDEL, dta_NONE)
		  read_gds_record;
	    end
	    -- ENDSTR
	    read_gds_record;
	    end_gds_structure;
	    curr_structure_name := Void
	end;

feature {NONE} -- the seven structure elements

	read_BOUNDARY is
	local pts: INTEGER;
	do  begin_element;
	    get_layer_and_datatype(rec_DATATYPE);
	    -- XY
	    pts := get_XY;
	    if pts < 3
	     then log_io_error(error_is_fewpts, Void)
	     else if (curr_int_array @ 1) /= (curr_int_array @ (2*pts-1)) 
		    or
		     (curr_int_array @ 2) /= (curr_int_array @ (2*pts))
		   then log_io_error(error_is_opnbdy, Void);
		   else pts := pts - 1
		  end;
		  gds_boundary(pts)
	    end
	    read_gds_record
	end;

	read_PATH is
	local ptyp: INTEGER; e: PATH_OPTIONS;
		wid, bext, eext, pts: INTEGER;
		have_cext: BOOLEAN;
	do  begin_element;
	    get_layer_and_datatype(rec_DATATYPE);
	    -- [PATHTYPE]
	    if record_is(rec_PATHTYPE, dta_B2INT)
	     then ptyp := curr_int_array @ 1;
		  read_gds_record
	     else ptyp := 0	-- the GDS default
	    end;
	    -- [WIDTH]
	    if record_is(rec_WIDTH, dta_B4INT)
	     then wid := curr_int_array @ 1;
		  read_gds_record
	     else wid := 0
	    end;
	    if wid < 0
	     then log_io_error(error_is_negwid, Void);
		  wid := 0
	    end;
	    -- [BGNEXTN], [ENDEXTN]
	    if record_is(rec_BGNEXTN, dta_B4INT)
	     then bext := curr_int_array @ 1;
		  read_gds_record
		  have_cext := true
	    end;
	    if record_is(rec_ENDEXTN, dta_B4INT)
	     then eext := curr_int_array @ 1;
		  read_gds_record
		  have_cext := true
	    end;
	    -- form e from ptyp, bext, eext
	    if ptyp = the_gds_custom_pathtype
	     then create e.make_custom(bext, eext)
	     else e := pathtype_ends(ptyp)	-- use the shared ends
		  if e = Void
		   then log_io_error(error_is_unkend, ptyp.out);
			e := pathtype_ends(0)
		  elseif have_cext
		   then log_io_error(error_is_illext, Void)
		  end
	    end;
	    -- XY
	    pts := get_XY;
	    if pts < 2
	     then log_io_error(error_is_fewpts, Void)
	     else gds_path(wid, e, pts)
	    end
	    read_gds_record
	end;

	read_AorSREF(is_aref: BOOLEAN) is
		-- AREF or SREF
	local m: MATRIX; nam: STRING; pts, nc, nr: INTEGER;
	do  begin_element;
	    -- SNAME
	    record_must_be(rec_SNAME, dta_ASCII);
	    nam := curr_ascii_string;
	    read_gds_record;
	    -- [STRANS]
	    m := get_strans
	    -- COLROW (if is_aref)
	    if is_aref
	     then record_must_be(rec_COLROW, dta_B2INT)
		  if curr_int_array.count /= 2
		   then log_syntax_error(error_is_illcnt, Void) 
		   else nc := curr_int_array @ 1;
			nr := curr_int_array @ 2
		  end
		  read_gds_record
	    end;
	    -- XY
	    pts := get_XY;
	    if is_aref
	     then if pts /= 3
		   then log_syntax_error(error_is_illcnt, Void)
		   else sref_pt1.make_XY(curr_int_array @ 1,
					 curr_int_array @ 2);
			aref_pt2.make_XY((curr_int_array @ 3) - sref_pt1.X,
					 (curr_int_array @ 4) - sref_pt1.Y);
			aref_pt3.make_XY((curr_int_array @ 5) - sref_pt1.X,
					 (curr_int_array @ 6) - sref_pt1.Y);
			gds_aref(nam, m, nc, nr, sref_pt1, aref_pt2, aref_pt3)
		  end
	     else if pts /= 1
		   then log_syntax_error(error_is_illcnt, Void)
		   else sref_pt1.make_XY(curr_int_array @ 1,
					 curr_int_array @ 2);
			gds_sref(nam, m, sref_pt1)
		  end
	    end
	    read_gds_record
	end;

	read_TEXT is
	local m: MATRIX; pres, ptyp, wid, pts: INTEGER
	do  begin_element;
	    get_layer_and_datatype(rec_TEXTTYPE);
	    if record_is(rec_PRESENTATION, dta_BITS)
	     then pres := curr_bit_array
		  read_gds_record;
	    end;
	    -- [PATHTYPE]
	    if record_is(rec_PATHTYPE, dta_B2INT)
	     then ptyp := curr_int_array @ 1;
		  read_gds_record
	     else ptyp := 0	-- the GDS default
	    end;
	    -- [WIDTH]
	    if record_is(rec_WIDTH, dta_B4INT)
	     then wid := curr_int_array @ 1;
		  read_gds_record
	     else wid := 0
	    end;
	    if wid < 0
	     then log_io_error(error_is_negwid, Void);
		  wid := 0
	    end;
	    -- strans and XY
	    m := get_strans;
	    pts := get_XY;
	    if pts /= 1
	     then log_syntax_error(error_is_illcnt, Void) end;
	    sref_pt1.make_XY(curr_int_array @ 1,
				curr_int_array @ 2);
	    read_gds_record
	    record_must_be(rec_STRING, dta_ASCII);
	    gds_text(pres, ptyp, wid, m, sref_pt1, curr_ASCII_string)
	    read_gds_record
	end;

	read_NODE is
	local pts: INTEGER
	do  begin_element;
	    get_layer_and_datatype(rec_NODETYPE);
	    pts := get_XY
	    gds_node(pts)
	    read_gds_record
	end;

	read_BOX is
	local pts: INTEGER
	do  begin_element;
	    get_layer_and_datatype(rec_BOXTYPE);
	    pts := get_XY
	    if pts /= 5
	     then log_io_error(error_is_fewpts, Void)
	     else if (curr_int_array @ 1) /= (curr_int_array @ (2*pts-1)) 
		    or
		     (curr_int_array @ 2) /= (curr_int_array @ (2*pts))
		   then log_io_error(error_is_opnbdy, Void);
		   else pts := pts - 1
		  end;
		  gds_box(pts)
	    end
	    read_gds_record
	end;

feature {NONE} -- element helpers

	begin_element is
	do  -- the intro record (any of the seven; they have no data)
	    read_gds_record;
	    -- ELFLAGS, PLEX
	    ignored_record(rec_ELFLAGS, dta_BITS);
	    ignored_record(rec_PLEX, dta_B4INT);
	end;

	get_layer_and_datatype(rec2_type: INTEGER) is
		-- reads 2 records: a rec_LAYER and a rec2_type
		-- calls gds_layer to process it, passing
		--  the gds layer, type, AND the rec2_type
	local gln, gdt: INTEGER;
	do  record_must_be(rec_LAYER, dta_B2INT);
	    if curr_int_array.count = 1
	     then gln := curr_int_array @ 1
	     else log_syntax_error(error_is_illcnt, Void)
	    end;
	    read_gds_record;
	    record_must_be(rec2_type, dta_B2INT);
	    if curr_int_array.count = 1
	     then gdt := curr_int_array @ 1
	     else log_syntax_error(error_is_illcnt, Void)
	    end;
	    read_gds_record;
	    -- now process it
	    gds_layer(gln, gdt, rec2_type)
	end;

	get_XY: INTEGER is
		-- return is number of POINTs in the XY
	do  record_must_be(rec_XY, dta_B4INT)
	    Result := curr_int_array.count;
	    if (Result \\ 2) > 0
	     then log_syntax_error(error_is_xxnoyy, Void)
	    end;
	    Result := (Result // 2);
	end;

	get_strans: MATRIX is
	local m: MATRIX; p: CONST_POINT
	do  if record_is(rec_STRANS, dta_BITS)
	     then if (curr_bit_array & strans_reflection) = 0
		   then Result := Identity
		   else Result := mirror_in_X
		  end;
		  if (curr_bit_array & 
			(strans_abs_mag | strans_abs_angle)) /= 0
		   then log_io_error(error_is_absref, Void)
		  end;
		  read_gds_record;
		  if record_is(rec_MAG, dta_B8REAL)
		   then create m.make_magnification(curr_real_array @ 1);
			Result := Result.and_then(m)
			read_gds_record
		  end;
		  if record_is(rec_ANGLE, dta_B8REAL)
		   then create p.make_polar_degrees(1, curr_real_array @ 1);
			create m.make_orientation(false, p);
			Result := Result.and_then(m)
			read_gds_record
		  end
	     else Result := Identity
	    end;
	end;

	sref_pt1, aref_pt2, aref_pt3: POINT;
		-- temps for SREF/AREF XY values to avoid object creation

	equal_rotations(a1, a2: DOUBLE): BOOLEAN is
		-- are a1 and a2 the same degree of rotation?
	local df: DOUBLE
	do  df := a1 - a2;
	    Result := df = 0 or df.abs = 360
	end;

	same_either_way(strans: MATRIX; ncol, nrow: INTEGER;
			dp2, dp3: CONST_POINT): BOOLEAN is
		-- is this an AREF that comes out the same under
		-- either setting of te_gds2_arrays?
		-- dp2 = (p2 - p1) and dp3 = (p3 - p1) are the 
		--  lattice defining vectors
	local xang, yang: DOUBLE; 
	do  xang := strans.rotation_degrees;
	    if strans.is_mirrored_in_X
	     then yang := xang - 90
	     else yang := xang + 90
	    end;
	    Result := (ncol = 1 or 
			equal_rotations(xang, dp2.theta_degrees))
		      and
		      (nrow = 1 or
			equal_rotations(yang, dp3.theta_degrees))
	end;

feature {NONE} -- the deferred routines to process the library structures

	begin_gds_library
		-- mark beginning of a gds library (header has been read).
		-- library_name is the library name
		-- user_unit is size of database units in user units
		-- database_unit is size of database units in meters
		-- modification_date and access_date are
		--   dates of last modification and access, respectively
	deferred
	end;

	end_gds_library is
		-- it's all over
	deferred
	end;

	begin_gds_structure(snam: STRING; cdat, mdat: DB_DATE) is
		-- snam is the structure name
		-- cdat and mdat are dates of creation and last modification,
		--  respectively
	deferred
	end;

	end_gds_structure is
		-- marks end of a structure
	deferred
	end;

	gds_layer(layer, datatype, rectype: INTEGER) is
		-- layer and (data)type for whatever follows
		-- rectype is one of rec_DATATYPE, rec_TEXTTYPE, 
		--	rec_NODETYPE, or rec_BOXTYPE
	deferred
	end;

	gds_boundary(pts: INTEGER) is
		-- actual XY values are still in curr_int_array
	deferred
	end;

	gds_path(wid: INTEGER; e: PATH_OPTIONS; pts: INTEGER) is
		-- actual XY values are still in curr_int_array
	deferred
	end;

	gds_sref(snam: STRING; strans: MATRIX; pt1: POINT) is
		-- strans is just the STRANS; pt1 is not included
	deferred
	end;

	gds_aref(snam: STRING; strans: MATRIX; ncol, nrow: INTEGER;
			pt1, dpt2, dpt3: POINT) is
		-- these are the three lattice definition points
	deferred
	end;

	gds_text(pres, ptyp, wid: INTEGER; strans: MATRIX; 
		 xy: VIRTUAL_POINT; text: STRING) is
	deferred
	end;

	gds_node(pts: INTEGER) is
		-- actual XY values are still in curr_int_array
	deferred
	end;

	gds_box	(pts: INTEGER) is
		-- actual XY values are still in curr_int_array
	deferred
	end;

feature	{NONE} -- creation

	make(dsrc: DATA_SOURCE) is
		-- No positioning of our DATA_SOURCE is performed.
		-- We read from the current "file" position as far as
		-- requested, and then leave the source at that position.
	do  my_file := dsrc;
	end;

end -- GDS_FILE_READER
