class GDS_FILE_WRITER
	-- write out a LAYOUT_LIBRARY as GDS Stream Format data
	--
	-- Usage	make(data_consumer)
	--		lioopt.set_output_write_scale(double)	-- optional
	--		write_gds_from_library(layout_library)
	--
	-- Output	the data sent to the data_consumer
	--		the_messages: RESPONSE_COLLECTOR
	--		 if the_messages has any errors, the sent data is
	--		 not a valid gds file; if there are only warnings,
	--		 it is valid.

inherit

	LAYOUT_IO_SUPPORT

	GEOMETRY_SUPPORT
	export {NONE} all
	end;

	GDS_FILE_VALUES
	export {NONE} all
	end;

	LC_OUTPUT_TERMINAL
	export  {NONE} all;
				-- we are our own layout_consumer
				--   (see put_structure)
	redefine output_text
	end;

create 
	make

feature {NONE} -- the output_write_scale

	gds_write_scale: DOUBLE
		-- we write out gds_write_scale x the library numbers

feature {NONE} -- the layout writers (deferred from LC_OUTPUT_TERMINAL)
	-- each item is output on current_layer

	do_reference(r: LAYOUT_REFERENCE; at: MATRIX) is
		-- sref or aref
	local  nX, nY: INTEGER; p1, p2, p3: POINT; mat: MATRIX;
		mangle, yangle, mmag: DOUBLE;
	do  -- AREF or SREF
	    if r.is_aref
	     then emit_no_data_record(rec_AREF);
	     else emit_no_data_record(rec_SREF)
	    end
	    -- SNAME
	    emit_ASCII_record(rec_SNAME, r.structure_name);
	    -- strans part
	    if at.is_gds_legal
	     then mat := at
	     else -- issue error and preserve just the translation part
		  log_io_error(error_is_illmat, at.out);
		  create mat.make_translation_xy(at.tx, at.ty)
	    end;
	    emit_strans(mat);
	    -- COLROW (if aref)
	    p1 := at.translation_point
	    temp_XY_list.wipe_out;
	    temp_XY_list.extend(p1);	-- aref and/or sref
	    if r.is_aref 
	     then nX := r.copies_in_X; nY := r.copies_in_Y
		  emit_b2int_hdr(rec_COLROW, 2);
		  emit_short_int(nX);
		  emit_short_int(nY)
		  -- these are the CALMA mode values
		  mmag := mat.magnification;
		  -- if we only had Calma mode, these values for p2 and p3
		  --  would suffice, but we instead use values that work
		  --  in either mode (by orienting the lattice to match mat)
		  -- create p2.make_xy(p1.x + nX*r.step_in_X*mmag, p1.y);
		  -- create p3.make_xy(p1.x, p1.y + nY*r.step_in_Y*mmag);
		  mangle := mat.rotation_degrees;
		  if mat.is_mirrored_in_X
		   then yangle := mangle - 90
		   else yangle := mangle + 90
		  end;
		  create p2.make_polar_degrees(nX*r.step_in_X*mmag, mangle);
		  create p3.make_polar_degrees(nY*r.step_in_Y*mmag, yangle);
		  temp_XY_list.extend(p1+p2)
		  temp_XY_list.extend(p1+p3)
	    end;
	    -- XY
	    emit_XY_record(temp_XY_list, false);
	    emit_no_data_record(rec_ENDEL)
	end;

	output_rectangle(r: CONST_RECTANGLE) is
		-- no RECTANGLE; use BOUNDARY
	local lx, ly, hx, hy: DOUBLE
	do  lx := r.low_x;  ly := r.low_y;
	    hx := r.high_x; hy := r.high_y;
	    temp_XY_list.wipe_out
	    temp_XY_list.extend_xy(lx, ly);
	    temp_XY_list.extend_xy(lx, hy);
	    temp_XY_list.extend_xy(hx, hy);
	    temp_XY_list.extend_xy(hx, ly);
	    output_polygon(temp_XY_list)
	end;

	output_polygon(p: POLYGON) is
		-- BOUNDARY
	do  emit_no_data_record(rec_BOUNDARY);
	    emit_layer_and_datatype;
	    emit_XY_record(p, true);
	    emit_no_data_record(rec_ENDEL)
	end;

	output_path(w: PATH) is
	local ptype: INTEGER
	do  emit_no_data_record(rec_PATH);
	    emit_layer_and_datatype;
	    ptype := w.ends.gds_pathtype;
	    if ptype /= 0
	     then emit_b2int_hdr(rec_PATHTYPE, 1);
		  emit_short_int(ptype)
	    end;
	    emit_b4int_hdr(rec_WIDTH, 1);
	    emit_scaled_number(w.width);
	    if ptype = the_gds_custom_pathtype
	     then emit_b4int_hdr(rec_BGNEXTN, 1);
		  emit_scaled_number(w.ends.begin_extension);
		  emit_b4int_hdr(rec_ENDEXTN, 1);
		  emit_scaled_number(w.ends.end_extension)
	    end;
	    emit_XY_record(w, false);
	    emit_no_data_record(rec_ENDEL);
	end;

	output_port(p: CONST_RECTANGLE) is
	do  -- ??
	end;

	output_text(t: LAYOUT_TEXT; at: MATRIX)
	local ptype: INTEGER
	do  if lioopt.preserve_gds_text then
		emit_no_data_record(rec_TEXT);
		emit_layer_and_type(rec_TEXTTYPE)
		if t.presentation /= 0 then
			emit_bits_record(rec_PRESENTATION, t.presentation)
		end;
		ptype := t.ends.gds_pathtype;
		if ptype /= 0 then
			emit_b2int_hdr(rec_PATHTYPE, 1);
			emit_short_int(ptype)
		end;
		if t.width > 0 then
			emit_b4int_hdr(rec_WIDTH, 1);
			emit_scaled_number(t.width)
		end;
		emit_strans(at);
		temp_XY_list.wipe_out
		temp_XY_list.extend(at.translation_point)
		emit_XY_record(temp_XY_list, false)
		emit_ascii_record(rec_STRING, t.text)
		emit_no_data_record(rec_ENDEL)
	    end
	end;

	terminate_traversal: BOOLEAN is false

feature {NONE} -- basic output manager

	file_position: STRING is "N/A";

	stru_position: STRING is
	do  if curr_structure = Void
	     then Result := "not in structure"
	     else Result := su.string_of_2("in structure ",
						curr_structure.name)
	    end
	end;

	my_file: DATA_CONSUMER
		-- probably a file, but can be anything

	output_buffer: STRING		-- 

	emit_2_bytes(b1, b2: CHARACTER) is
		-- add two bytes to the buffer and flush to file if full
	do  output_buffer.extend(b1);
	    output_buffer.extend(b2)
	    if output_buffer.count >= gds_record_length
	     then my_file.consume(output_buffer);
		  output_buffer.keep_head(0)
	    end;
	end;

	emit_final_contents is
		-- fill buffer with nulls until full
	do  from
	    until output_buffer.is_empty
	    loop  emit_2_bytes('%U', '%U')
	    end
	end;

	emit_2_ints(i1, i2: INTEGER) is
		-- two 8 bit integers as 1 byte each
	do  emit_2_bytes(i1.to_character_8, i2.to_character_8)
	end;

	emit_short_int(i: INTEGER) is
		-- 16 bit inteer as 2 bytes
	do  emit_2_ints(((i |>>  8) & 255),  (i         & 255))
	end;

	emit_long_int(i: INTEGER) is
		-- 32 bit integer as four bytes
	do  emit_2_ints(((i |>> 24) & 255), ((i |>> 16) & 255));
	    emit_2_ints(((i |>>  8) & 255),  (i         & 255))
	end;

	emit_double(d: DOUBLE) is
	local exp16: INTEGER
	do  if d = 0
	     then emit_long_int(0); emit_long_int(0)
	     else -- extract sign; use absolute value
		  -- hex normalize (with exponent in exp16)
		  from	exp16 := two_up_6;
			if d < 0
			 then exp16 := exp16 + two_up_7 end;
			emitted_double := d.abs;
		  until emitted_double < 1.0
		  loop  emitted_double := emitted_double / 16.0;
			exp16 := exp16 + 1
		  end
		  from
		  until emitted_double >= 0.0625
		  loop  emitted_double := emitted_double * 16.0;
			exp16 := exp16 - 1
		  end;
		  -- emit 1 byte sign & exponent, 7 bytes mantissa
		  emit_2_ints(exp16, next_few_bits(two_up_8));
		  emit_short_int(next_few_bits(two_up_16));
		  emit_short_int(next_few_bits(two_up_16));
		  emit_short_int(next_few_bits(two_up_16));
	    end
	end;

	next_few_bits(m: INTEGER): INTEGER is
		-- m should be a power of two
	do  emitted_double := emitted_double * m
	    Result := emitted_double.floor
	    emitted_double := emitted_double - Result
	end;

	emitted_double: DOUBLE	

feature {NONE} -- gds record headers (first four bytes)

	emit_b2int_hdr(rtyp, cnt: INTEGER) is
	do  emit_short_int(4 + 2*cnt);
	    emit_2_ints(rtyp, dta_B2INT)
	end;

	emit_b4int_hdr(rtyp, cnt: INTEGER) is
	do  emit_short_int(4 + 4*cnt);
	    emit_2_ints(rtyp, dta_B4INT)
	end;

	emit_b4real_hdr(rtyp, cnt: INTEGER) is
	do  emit_short_int(4 + 4*cnt);
	    emit_2_ints(rtyp, dta_B4REAL)
	end;

	emit_b8real_hdr(rtyp, cnt: INTEGER) is
	do  emit_short_int(4 + 8*cnt);
	    emit_2_ints(rtyp, dta_B8REAL)
	end;

feature {NONE} -- entire gds records

	emit_no_data_record(rtyp: INTEGER) is
	do  emit_short_int(4);
	    emit_2_ints(rtyp, dta_NONE)
	end;

	emit_bits_record(rtyp: INTEGER; bit_array: INTEGER) is
		-- NOTE: there is exactly one bit array in all cases
	do  emit_short_int(6);
	    emit_2_ints(rtyp, dta_BITS);
	    emit_short_int(bit_array);
	end;

	emit_one_b2int_record(rtyp: INTEGER; val: INTEGER) is
		-- common case of a single b2int
	do  emit_b2int_hdr(rtyp, 1);
	    emit_short_int(val)
	end;

	emit_ascii_record(rtyp: INTEGER; s: STRING) is
	local i, cnt, pad: INTEGER
	do  cnt := s.count;
	    pad := cnt \\ 2;	-- 0 if even; 1 if odd
	    -- the record header
	    emit_short_int(4 + cnt + pad);
	    emit_2_ints(rtyp, dta_ASCII);
	    -- the ASCII string itself
	    from  i := 1
	    until i >= cnt	-- note: skip odd last byte here
	    loop  emit_2_bytes(s @ i, s @ (i+1))
		  i := i + 2
	    end;
	    if pad > 0		-- do odd last byte here
	     then emit_2_bytes(s @ cnt, '%U')
	    end
	end;

	emit_2_date_record(rtyp: INTEGER; dt1, dt2: DB_DATE) is
	do  emit_b2int_hdr(rtyp, 12);
	    emit_db_date(dt1); emit_db_date(dt2)
	end;

feature {NONE} -- writers of the bits and pieces

	emit_db_date(dt: DB_DATE) is
	local nvd: DB_DATE
	do  if dt = Void
	     then nvd := the_zero_date
	     else nvd := dt
	    end;
	    emit_short_int(nvd.year - gds_year_zero);
	    emit_short_int(nvd.month);
	    emit_short_int(nvd.day);
	    emit_short_int(nvd.hour);
	    emit_short_int(nvd.minute);
	    emit_short_int(nvd.second);
	end;

	the_zero_date: DB_DATE is
	once create {ORACLE_DATE} Result.set_zero
	end;

	emit_layer_and_type(a_type: INTEGER) is
	local lay: LAYER; glay, gtyp: INTEGER;
	do  lay := layer_by_name(current_layer);
	    if lay.have_gds_name
	     then glay := lay.a_gds_layer;
		  gtyp := lay.a_gds_datatype
	     else -- FIX_ME: ??
	    end;
	    emit_one_b2int_record(rec_LAYER, glay);
	    emit_one_b2int_record(a_type, gtyp)
	end;

	emit_layer_and_datatype is
	do  emit_layer_and_type(rec_DATATYPE)
	end;

	emit_XY_record(lst: LIST_OF_POINTS; closed: BOOLEAN) is
		-- if closed, repeat the first point at the end
	require not lst.is_empty
	local cnt: INTEGER;
	do  cnt := lst.count; if closed then cnt := cnt + 1 end;
	    emit_b4int_hdr(rec_XY, 2*cnt);
	    from  lst.start
	    until lst.after
	    loop  emit_scaled_number(lst.item_X);
		  emit_scaled_number(lst.item_Y)
		  lst.forth
	    end;
	    if closed
	     then emit_scaled_number(lst.first_X);
		  emit_scaled_number(lst.first_Y)
	    end
	end;

	emit_scaled_number(x: DOUBLE) is
		-- scale by gds_write_scale (and round)
	do  emit_long_int((x * gds_write_scale).rounded)
	end;

	temp_XY_list: POLYGON		-- reuse as needed

	emit_strans(m: MATRIX) is
	require m.is_gds_legal
	local refl: BOOLEAN; mag, rot: DOUBLE; val: INTEGER
	do  refl := m.is_mirrored_in_X;
	    mag  := m.magnification;
	    rot  := m.rotation_degrees;
	    if refl or (mag /= 1.0) or (rot /= 0)
	     then if refl then val := strans_reflection end;
		  emit_bits_record(rec_STRANS, val);
		  if mag /= 1.0
		   then emit_b8real_hdr(rec_MAG, 1);
			emit_double(mag)
		  end
		  if rot /= 0
		   then -- Cadence wants 0 to 360, not -180 to 180
			if rot < 0 then rot := rot + 360 end;
			emit_b8real_hdr(rec_ANGLE, 1);
			emit_double(rot)
		  end
	    end
	end;

	emit_structure(stru: LIBRARY_STRUCTURE) is
		-- emit stru to the file
	require exists: stru /= Void;
	--	valid_name: valid_structure_name(stru.name);
		body_exists: stru.body /= Void
	do  curr_structure := stru
	    -- structure header
	    emit_2_date_record(rec_BGNSTR, stru.date_created,
					   stru.date_modified);
	    emit_ASCII_record(rec_STRNAME, stru.name);
	    -- the body
	    stru.body.traverse(Current, identity);
	    -- structure trailer
	    emit_no_data_record(rec_ENDSTR)
	    curr_structure := Void
	end;

feature -- the gds writer

	write_gds_from_library(lib: LAYOUT_LIBRARY) is
	require exists: lib /= Void
	local rnam: STRING; root, stru: LIBRARY_STRUCTURE
	do  -- setup
	    set_the_lib(lib);
	    gds_write_scale := lioopt.output_write_scale
	    unknown_layers := lioopt.init_unknown_layers
	    filtering_techn := lioopt.filtering_techn;
	    traversed_layers := lioopt.traversed_layers;
	    create output_buffer.make(2048);
	    create temp_XY_list.make(5)
	    -- write library header stuff
	    emit_one_b2int_record(rec_HEADER, version_6);
	    emit_2_date_record(rec_BGNLIB, lib.date_modified,
					   lib.date_accessed);
	    emit_ascii_record(rec_LIBNAME, lib.name);
	    emit_b8real_hdr(rec_UNITS, 2);
		emit_double(lib.user_units);
						-- db unit in user units
		emit_double(lib.unit_size/gds_write_scale);
						-- db unit in meters
	    -- write out all the structures in the library
	    rnam := lib.root_structure;
	    if su.not_empty(rnam) and lib.has_structure(rnam) 
	     then -- root first
		  root := lib.get_structure(rnam)
		  emit_structure(root)
	    end;
	    from  lib.start	-- then all the others
	    until lib.off
	    loop  stru := lib.item;
		  if stru /= root then emit_structure(stru) end;
		  lib.forth
	    end;
	    -- write End and be done
	    emit_no_data_record(rec_ENDLIB);
	    emit_final_contents;
	    -- remove temporaries
	    output_buffer := Void
	    temp_XY_list := Void
	end;

feature {NONE} -- creation

	make(dcon: DATA_CONSUMER) is
		-- note that we just feed a byte stream to the consumer
		-- we neither initialize nor close it here
	do  my_file := dcon
	    create the_messages.make(writer_application)
	end;

end -- GDS_FILE_WRITER
