
indexing
	
	description: "A run closing step operation that generates %
		%a fill exclusion file, which is a text file with CATS %
		%commands to generate solid blocks over projects and vendor %
		%frame that are not to be filled at all"
	author: "David Hollenberg"
	date: "March 3, 2004"

class GEN_FILL_EXCLUSION_OP
inherit
	RC_PRIMITIVE_STEP_OP
	NO_USER_INTERFACE
	CREATES_EXTERNAL_STATE
	NUMBER_UTILITIES_ACCESS
		export
			{NONE} all
		undefine
			default_create
		end
	STRING_UTILITIES_ACCESS
		export
			{NONE} all
		undefine
			default_create
		end

create
	make

feature -- Properties

	input_data_items: LIST [STRING]
			-- Data items required before step can be executed
		do
			Result := array_as_list (<<
				data_item.Mask_techn,
				data_item.Test_vehicle_fab_instances
				>>)
		end

feature {NONE} -- Implementation

	is_applicable: BOOLEAN
			-- Is step applicable to run for session?
		do
			Result := cache.run_needs_fill
		end

	do_step
		do
			write_fill_exclusion_file
			call_completion_actions
		end

	undo_step
			-- Undo execution of step
		local
			fname: STRING
		do
			fname := fill_exclusion_file.file_pathname
			session.delete_file_if_exists (fname)
			call_completion_actions
		end

	set_state
			-- Set `state'
		do
			state := Void
		end

	write_fill_exclusion_file
			-- Write fill exclusion file
		do
			cache.check_one_scale_factor
			cache.check_one_field
			levels := cache.run_level_rows
			create lines.make (Max_primary_selects)
			append_die_fill_exclusion_lines
			append_frame_fill_exclusion_lines
			write_fill_exclusion_lines
			check_too_many_selects
		end

	check_too_many_selects
			-- Check whether number of selects to be generated
			-- exceeds the maximum supported by CATS
		do
			if lines.count > Max_selects  then
				handle_hard_error ("Number of areas to be excluded from fill (" + lines.count.out + ") exceeds maximum of " + Max_selects.out)
			end
		end

	write_fill_exclusion_lines
			-- Write lines to fill exclusion file
		local
			file: LARGE_PLAIN_TEXT_FILE
			f: RC_FILE
			partial_line, line: STRING
			n: INTEGER
		do
			f := fill_exclusion_file
			create file.make (f.file_pathname)
			session.open_file_for_write (file, f)
			create line.make (100)
			from
				lines.start
			until
				lines.after
			loop
				partial_line := lines.item
				n := n + 1
				build_select_line (n, partial_line, line)
				file.put_string (line)
				lines.forth
			end
			session.close_file (file)
		end

	build_select_line (n: INTEGER partial_line, dest: STRING)
			-- Write "select" number `n' from `partial_line'
			-- and put into `dest'
		require
			line_number_positive: n > 0
			partial_line_exists: partial_line /= Void
			dest_exists: dest /= Void
		local
			select_num: STRING
		do
			dest.keep_head (0)
			dest.append ("select ")
			if n <= Max_primary_selects then
				dest.append ("primary   ")
				select_num := n.out
			elseif n <= Max_selects then
				dest.append ("secondary ")
				select_num := (n - Max_primary_selects).out
			else
				handle_hard_error ("line number too big in build_select_line")
			end
			su.left_pad (select_num, 3)
			dest.append (select_num)
			dest.append (" fill ")
			dest.append (partial_line)
			dest.extend ('%N')
		end

	append_die_fill_exclusion_lines
			-- Append to `lines' lines that each have 4
			-- coordinates that define a box covering die
			-- to be excluded from fill, plus comment with
			-- die id and rotation status
		local
			q: SQL_QUERY
			die_id, ident: STRING
			factor, low_x, low_y, high_x, high_y: DOUBLE
			rotation: INTEGER
			merged_size: CONST_POINT
			loc, half_design_size, offset: POINT
			rect: CONST_RECTANGLE
		do
			factor := levels.first.scale_factor
			merged_size := cache.merged_size
			create loc.make_xy (0, 0)
			create offset.make_xy (0, 0)
			create half_design_size.make_xy (0, 0)
			
			q := connection.sql_query (unfilled_dies_query)
			q.set_numeric_bind_capacity (1)
			q.bind_num_to_string (1, session.run_id, False)
			q.set_select_count (8)
			q.select_double (1)
			q.select_double (2)
			q.select_double (3)
			q.select_double (4)
			q.select_integer (5)
			q.select_string (6, 2)
			q.select_double (7)
			q.select_double (8)
			q.execute
			from
				q.fetch
			until
				q.end_of_data
			loop
				loc.make_xy (q.double_value (1), q.double_value (2))
				half_design_size.make_xy (q.double_value (3) / 2.0, q.double_value (4) / 2.0)
				rotation := q.integer_value (5)
				die_id := q.string_value (6)
				offset.make_xy (q.double_value (7), q.double_value (8))
				if rotation = 0 then
					low_x := loc.x + offset.x - half_design_size.x
					low_y := loc.y + offset.y - half_design_size.y
					high_x := loc.x + offset.x + half_design_size.x
					high_y := loc.y + offset.y + half_design_size.y
				elseif rotation = 90 then
					low_x := loc.x + offset.y - half_design_size.y
					low_y := loc.y - offset.x - half_design_size.x
					high_x := loc.x + offset.y + half_design_size.y
					high_y := loc.y - offset.x + half_design_size.x
				elseif rotation = 180 then
					low_x := loc.x - offset.x - half_design_size.x
					low_y := loc.y - offset.y - half_design_size.y
					high_x := loc.x - offset.x + half_design_size.x
					high_y := loc.y - offset.y + half_design_size.y
				elseif rotation = 270 then
					low_x := loc.x - offset.y - half_design_size.y
					low_y := loc.y + offset.x - half_design_size.x
					high_x := loc.x - offset.y + half_design_size.y
					high_y := loc.y + offset.x + half_design_size.x
				else
					handle_hard_error ("Expecting rotation of 0 or 90 but rotation is " + rotation.out + " for die id " + die_id)
				end
				
				low_x  := low_x  * factor + merged_size.x / 2.0
				low_y  := low_y  * factor + merged_size.y / 2.0
				high_x := high_x * factor + merged_size.x / 2.0
				high_y := high_y * factor + merged_size.y / 2.0
				
				create rect.make_by_corners (low_x, low_y, high_x, high_y)

				ident := die_id
				if rotation /= 0 then
					ident := ident + " rot " + rotation.out
				end
				append_partial_line (rect, ident)
				q.fetch
			end
			q.release
		end

	append_frame_fill_exclusion_lines
			-- Append to `lines' up to 4 lines (for left,
			-- right, top, bottom) that each have 4
			-- coordinates that define a box covering a
			-- portion of the frame (area of merged
			-- pattern file outside of payload)
		local
			factor: DOUBLE
			payload_left, payload_right, payload_bottom, payload_top: DOUBLE
			merged_left, merged_right, merged_bottom, merged_top: DOUBLE
			pbox, merged_box: CONST_RECTANGLE
			payload_box: RECTANGLE
			move: CONST_POINT
		do
			factor := levels.first.scale_factor
			pbox := cache.payload_box
			create payload_box.make_copy (pbox)
			payload_box.scale_coordinates_by (factor)
			merged_box := cache.merged_box
			
			payload_left := payload_box.low_x
			payload_right := payload_box.high_x
			payload_bottom := payload_box.low_y
			payload_top := payload_box.high_y

			merged_left := merged_box.low_x
			merged_right := merged_box.high_x
			merged_bottom := merged_box.low_y
			merged_top := merged_box.high_y
			
			move := -merged_box.ll

			append_frame_exclusion (merged_left, merged_bottom,
				payload_left, merged_top, move, "Left")
			append_frame_exclusion (payload_right, merged_bottom,
				merged_right, merged_top, move, "Right")
			append_frame_exclusion (merged_left, payload_top,
				merged_right, merged_top, move, "Top")
			append_frame_exclusion (merged_left, merged_bottom,
				merged_right, payload_bottom, move, "Bottom")
		end

	append_frame_exclusion (low_x, low_y, high_x, high_y: DOUBLE move: CONST_POINT ident: STRING)
			-- Append a partial "select" line to `lines'
			-- with box defined by `low_x', `low_y',
			-- `high_x' and `high_y' moved by `move' and
			-- identified by `ident', but only if area of
			-- rectangle is greater than 0
		require
			move_exists: move /= Void
			ident_exists: ident /= Void
		local
			rect: RECTANGLE
		do
			create rect.make_by_corners (low_x, low_y, high_x, high_y)
			rect.move_by (move.x, move.y)
			if rect.area > 0 then
				append_partial_line (rect, ident)
			end
		end

	append_partial_line (rect: CONST_RECTANGLE ident: STRING)
			-- Append a partial "select" line to `lines'
			-- with box defined by `rect' and identified by
			-- `ident'
		require
			rectangle_exists: rect /= Void
			ident_exists: ident /= Void
		local
			line: STRING
		do
			create line.make (100)
			line.append (nu.std_formatted_number (rect.low_x))
			line.extend (' ')
			line.append (nu.std_formatted_number (rect.low_y))
			line.extend (' ')
			line.append (nu.std_formatted_number (rect.high_x))
			line.extend (' ')
			line.append (nu.std_formatted_number (rect.high_y))
			su.right_pad (line, 43)
			line.extend (' ')
			line.extend ('!')
			line.extend (' ')
			line.append (ident)
			lines.extend (line)
		end

	fill_exclusion_file: RC_FILE
			-- File with CATS commands to generete solid
			-- blocks for projects and parts of vendor frame
			-- which are not to be filled
		do
			Result := file_generator.fill_exclusion_file
		ensure
			result_exists: Result /= Void
		end

	unfilled_dies_query: STRING =
		"select %N%
		%   location_x, location_y, design_size_x, design_size_y, %N%
		%   die_wafer_rotation, ifi.die_id, design_die_offset_x, design_die_offset_y %N%
		%from cust_ic_designs cid, ic_fab_instances ifi, die_locations dl, die_attributes da %N%
		%where cid.design_no = ifi.design_no %N%
		%  and ifi.run_id = da.run_id %N%
		%  and ifi.die_id = da.die_id %N%
		%  and dl.run_id = da.run_id %N%
		%  and dl.die_id = da.die_id %N%
		%  and ifi.fill_allowed = 'N' %N%
		%  and ifi.run_id = :1 %N%
		%order by ifi.die_id"
			-- Query to retrieve areas of dies to be excluded
			-- from fill and their die ids
	
	levels: LIST [RUN_LEVELS_ROW]
			-- List of read-only rows for wafer techn levels
	
	lines: ARRAYED_LIST [STRING]
			-- List of lines with CATS "select" operations
			-- to exclude areas from being filled

	Max_primary_selects: INTEGER = 99
			-- Maximum number of primary selects (there are
			-- an equal number of secondary selects)
	
	Max_selects: INTEGER
			-- Maximum number of selects (primary plus secondary)
		do
			Result := Max_primary_selects * 2
		end
	
end
