
indexing
	
	description: "A run closing step operation that generates %
		%a merged GDS file for each individual die that is a project or %
		%one for each field"
	author: "David Hollenberg"
	date: "October 7, 2008"

class GEN_MERGED_GDS_FILES_OP
inherit
	RC_PRIMITIVE_STEP_OP
		redefine
			is_currently_undoable, not_undoable_reason
		end
	HAS_USER_INTERFACE
	CREATES_EXTERNAL_STATE
	GUI_UTILITIES_ACCESS
		export
			{NONE} all
		undefine
			default_create
		end
	RC_ENVIRONMENT_ACCESS
		export
			{NONE} all
		undefine
			default_create
		end
	NUMBER_UTILITIES_ACCESS
		export
			{NONE} all
		undefine
			default_create
		end

create
	make

feature -- Properties

	input_data_items: LIST [STRING]
			-- Data items required before step can be executed
		do
			Result := array_as_list (<<
				data_item.Dynamic_vendor_geometry,
				data_item.Scribe_files,
				data_item.Filler_files,
				data_item.Die_overhead_files
				>>)
		end

feature {NONE} -- Implementation

	is_applicable: BOOLEAN
			-- Is step applicable to run for session?
		do
			Result := cache.run_needs_merged_gds_files
		end

	is_currently_undoable: BOOLEAN
			-- Is `Current' currently in a state where it can be
			-- undone?
		do
			Result := not cache.run_gds_file_rows.there_exists (agent {RUN_GDS_FILES_ROW}.is_status_stopable)
		end

	not_undoable_reason: STRING
			-- Reason step cannot currently be undone
		do
			Result := "one or more GDS merge jobs is running or pending.%NPlease select all stoppable jobs and stop them first."
		end

	do_step
		local
			ifc: JOB_CONTROL_IFC
		do
			check_validity
			if not is_valid then
				abort_actions.call ([])
			else
				clear_run_gds_cache
				jobs := cache.run_gds_file_mrows
				if jobs.is_empty then
					insert_gds_merge_jobs
					cache.cache_run_gds_files (jobs)
				end
				if cache.is_die_merge then
					ifc := factory.merged_gds_die_control_interface (session.run_id, session.title_with_run_id ("Generate Merged GDS Files for Dies"))
				else
					ifc := factory.merged_gds_field_control_interface (session.run_id, session.title_with_run_id ("Generate Merged GDS Files for Fields"))
				end
				ifc.set_session (session)
				ifc.success_actions.extend (agent process_jobs_completion)
				ifc.abort_actions.append (abort_actions)
				ifc.build
				if preferences.is_autosubmit_enabled then
					gui.do_once_on_idle (agent ifc.autosubmit_jobs)
				end
			end
		end

	undo_step
			-- Undo execution of step
		local
			lock_op: ROW_LOCK_OP
		do
			clear_run_gds_cache
			fetch_run_gds_rows
			add_commit_action (agent delete_files (generated_files))
			add_commit_action (agent cache.uncache_run_gds_files)
			add_commit_action (agent cache.uncache_run_gds_file_layers)
			add_commit_action (agent cache.uncache_run_gds_merge_results)
			create lock_op
			lock_op.lock_list_nowait (run_gds_merge_result_mrows)
			lock_op.lock_list_nowait (run_gds_file_layer_mrows)
			lock_op.lock_list_nowait (run_gds_file_mrows)
			run_gds_merge_result_mrows.do_all (agent {RUN_GDS_MERGE_RESULTS_MROW}.delete)
			run_gds_file_layer_mrows.do_all (agent {RUN_GDS_FILE_LAYERS_MROW}.delete)
			run_gds_file_mrows.do_all (agent {RUN_GDS_FILES_MROW}.delete)
			call_completion_actions
		end

	check_validity
			-- Check for validity errors (field merge only).
			-- If not valid, issue error message.
			-- Set `is_valid' to indicate whether valid
		do
			is_valid := True
			if cache.is_field_merge then
				-- Field merge
				cache.die_xrows.do_all (agent {DIES_XROW}.extend_xrow_data (cache))
				check_for_multiple_design_techns
				if is_valid then
					check_derived_layers
				end
			end
		end

	check_for_multiple_design_techns
			-- Check whether all dies to be merged in each
			-- field (for a field merge only) have the same
			-- wafer design techn.  If not, issue error and
			-- abort the step since merge will fail
		require
			is_field_merge: cache.is_field_merge
		local
			operation: ANONYMOUS_OPERATION
			msg: STRING
		do
			create multiple_design_techns_message.make (0)
			create distinct_dies.make (5)
			cache.field_names.do_all (agent check_field_for_multiple_design_techns)
			if not multiple_design_techns_message.is_empty then
				is_valid := False
				create msg.make (200)
				msg.append ("Some fields have more than one distinct design techn so merging will fail.%N")
				msg.append ("To fix this problem, you must repack these fields so that all projects%N")
				msg.append ("have the same design techn.  Design techn distribution shown below.%N%N")
				multiple_design_techns_message.prepend (msg)
				create operation.make (agent handle_user_error (multiple_design_techns_message))
				operation.enable_abort_caught
				operation.execute
			end
		end

	check_field_for_multiple_design_techns (fld_name: STRING)
			-- Check whether all dies in `fld_name' have the
			-- same wafer design techn.  If not, append
			-- error message to `multiple_design_techns_message'.
			-- Also, set list of distinct dies in
			-- `fld_name' in `distinct_dies'
		local
			dies: LIST [DIES_XROW]
			list: ARRAYED_LIST [DIES_XROW]
			die: DIES_XROW
			locations: LIST [DIE_LOCATIONS_ROW]
			dt_id: STRING
		do
			create design_techn_dies.make (5)
			dies := cache.die_xrows
			from
				dies.start
			until
				dies.after
			loop
				die := dies.item
				if not die.is_vendor_ghost and not die.is_external_to_payload then
					locations := cache.die_location_rows_for_die_id (die.die_id)
					if locations.there_exists (agent {DIE_LOCATIONS_ROW}.is_in_field (fld_name)) then
						dt_id := die.wafer_design_techn
						design_techn_dies.search (dt_id)
						if design_techn_dies.found then
							design_techn_dies.found_item.extend (die)
						else
							create list.make (20)
							list.extend (die)
							design_techn_dies.put (list, dt_id)
						end
						distinct_dies.search (fld_name)
						if distinct_dies.found then
							distinct_dies.found_item.extend (die)
						else
							create list.make (20)
							list.extend (die)
							distinct_dies.put (list, fld_name)
						end
					end
				end
				dies.forth
			end
			
			if design_techn_dies.count > 1 then
				-- More than one design techn in field
				multiple_design_techns_message.append (design_techn_distribution (fld_name))
			end
		end

	design_techn_distribution (fld_name: STRING): STRING
			-- Text describing how many dies with each
			-- design techn are in `fld_name'
		local
			dt: STRING
			dies: LIST [DIES_ROW]
			n: INTEGER
			pp: PRETTY_PRINT_STRING_MODIFIER
		do
			create Result.make (100)
			Result.append ("Field " + fld_name + "%N%N")
			from
				design_techn_dies.start
			until
				design_techn_dies.after
			loop
				dt := design_techn_dies.key_for_iteration
				dies := design_techn_dies.item_for_iteration
				n := dies.count
				create pp.make (80, "        ", "        ", "%N%N")
				Result.append ("    " + dt + "   (" + n.out + " " + su.number (n, "die", "dies") + ")%N")
				Result.append (pp.modify (su.joined_by (die_ids_from_list (dies), " ")))
				design_techn_dies.forth
			end
		end

	die_ids_from_list (list: LIST [DIES_ROW]): LIST [STRING]
			-- List of die IDs in `list'
		do
			create {ARRAYED_LIST [STRING]} Result.make (list.count)
			list.do_all (agent add_die_id (?, Result))
		end

	add_die_id (row: DIES_ROW dest: LIST [STRING])
			-- Add die ID in `row' to `dest'
		do
			dest.extend (row.die_id)
		end

	check_derived_layers
			-- Check for each field whether all dies to be
			-- merged in that field have the same set of
			-- "derived when absent" layers.  If not,
			-- issue a warning since merging may
			-- not produce a correct result
		require
			is_field_merge: cache.is_field_merge
			distinct_dies_set: distinct_dies /= Void
		local
			operation: ANONYMOUS_OPERATION
		do
			create derived_layers_message.make (0)
			cache.field_names.do_all (agent check_field_for_derived_layers)
			if not derived_layers_message.is_empty then
				derived_layers_message.prepend ("Some fields have a mix of dies with derived layers present in some dies but not all dies.%NThis makes merging potentially unsafe.  Please contact a vendor engineering wizard.%NProblem derived layers noted below.%N%N")
				create operation.make (agent warn_with_notify (derived_layers_message))
				operation.enable_abort_caught
				operation.execute
				is_valid := not operation.aborted
				if is_valid then
					create operation.make (agent warn ("continuing past this point will send an e-mail notification%Nto the vendor engineering wizard, who will check whether this GDS merge is safe.%NDo not trust the results of the merge until the wizard blesses it."))
					operation.enable_abort_caught
					operation.execute
					is_valid := not operation.aborted
					-- Notify mailing list that user 
					-- continued past warning
					if is_valid then
						session.send_email_to_list (rc_env.gds_merge_error_to_list, "potentially unsafe GDS merge", "User continued past the warning below.  Please verify that this GDS merge is safe%Nand then reply to this message if you haven't already responded to a previous%Nnotification.%N%NWarning:%N%N" +  derived_layers_message, rc_env.user_email_address)
					end
				end
			end
		end

	check_field_for_derived_layers (fld_name: STRING)
			-- Check whether all dies in `fld_name' have
			-- the same set of "derived when absent"
			-- layers.  If not, append error message to
			-- `derived_layers_message'.
		local
			dies: LIST [DIES_XROW]
			die: DIES_XROW
			layers, derived_layers, scribe_layers: LIST [STRING]
		do
			dies := distinct_dies.item (fld_name)
				-- External to payload dies and vendor ghosts
				-- already excluded
			if not dies.is_empty then
				scribe_layers := dies.first.design_techn.layer_names_for_scribe
				derived_layers := dies.first.derived_when_absent_xform_layer_names
					-- All dies in field already verified
					-- to have same wafer design techn
				create derived_layer_dies.make (derived_layers.count)
				from
					dies.start
				until
					dies.after
				loop
					die := dies.item
					if die.is_scribe then
						layers := scribe_layers
					elseif die.is_project then
						layers := die.layers
					else
						layers := Void
					end
					if layers /= Void then
						create die_layers.make (layers.count)
						layers.do_all (agent add_layer (?, die_layers))
						derived_layers.do_all (agent record_derived_layer_for_die (?, die, die_layers))
					end
					dies.forth
				end

				add_derived_layers_message (fld_name)
			end
		end

	add_derived_layers_message (fld_name: STRING)
			-- Add error message (if any) for issues with
			-- derived layers for `fld_name' to
			-- `derived_layers_message'
		local
			pair: LIST_PAIR [DIES_ROW]
			layer, msg: STRING
			pp_in, pp_not_in: PRETTY_PRINT_STRING_MODIFIER
		do
			create msg.make (0)
			create pp_in.make (80, "        In:     ", "                ", "%N")
			create pp_not_in.make (80, "        Not In: ", "                ", "%N")
			from
				derived_layer_dies.start
			until
				derived_layer_dies.after
			loop
				pair := derived_layer_dies.item_for_iteration
				layer := derived_layer_dies.key_for_iteration
				if pair.both_lists_non_empty then
					-- Some dies have derived layer and some do not
					if pair.list2.there_exists (agent {DIES_ROW}.is_non_scribe) then
						-- At least one die which is missing
						-- derived layer is a non-scribe
						msg.append ("    " + layer + "%N%N")
						msg.append (pp_in.modify (su.joined_by (die_ids_from_list (pair.list1), " ")))
						msg.append (pp_not_in.modify (su.joined_by (die_ids_from_list (pair.list2), " ")))
						msg.extend ('%N')
					end
				end
				derived_layer_dies.forth
			end

			if not msg.is_empty then
				derived_layers_message.append ("Field " + fld_name + "%N%N")
				derived_layers_message.append (msg)
			end
		end

	add_layer (a_layer: STRING dest: like die_layers)
			-- Add `a_layer' to `dest'
		do
			dest.put (1, a_layer)
		end

	record_derived_layer_for_die (a_layer: STRING a_die: DIES_XROW layers_table: like die_layers)
			-- If `layers_table' contains `a_layer' 
			-- then record `a_die' as having `a_layer'.
			-- Otherwise, record it as not having `a_layer'
		local
			pair: LIST_PAIR [DIES_ROW]
		do
			derived_layer_dies.search (a_layer)
			if derived_layer_dies.found then
				pair := derived_layer_dies.found_item
			else
				create pair.make (20, 20)
				derived_layer_dies.put (pair, a_layer)
			end
			layers_table.search (a_layer)
			if layers_table.found then
				-- Die has `a_layer'
				pair.list1.extend (a_die)
			else
				-- Die does not have `a_layer'
				pair.list2.extend (a_die)
			end
		end

	clear_run_gds_cache
			-- Clear any cached RUN_GDS_* rows since another
			-- process may have modified or inserted them
		do
			cache.uncache_run_gds_files
			cache.uncache_run_gds_file_layers
			cache.uncache_run_gds_merge_results
		end

	fetch_run_gds_rows
		do
			run_gds_file_mrows := cache.run_gds_file_mrows
			run_gds_file_layer_mrows := cache.run_gds_file_layer_mrows
			run_gds_merge_result_mrows := cache.run_gds_merge_result_mrows
		end

	insert_gds_merge_jobs
			-- Insert one row in RUN_GDS_FILES and one row
			-- in RUN_GDS_MERGE_RESULTS for each field or
			-- one of each row for each die that
			-- represents a project, depending on packing
			-- target
		do
			if cache.is_die_merge then
				insert_die_gds_merge_jobs
			else
				insert_field_gds_merge_jobs
			end
			session.commit
		end

	insert_field_gds_merge_jobs
			-- Insert one row in RUN_GDS_FILES for each
			-- for each field for run
		local
			fld_names: LIST [STRING]
			fld_name: STRING
			job: RUN_GDS_FILES_MROW
			job_status: JOB_STATUS_VALUES
			n: INTEGER
		do
			fld_names := cache.field_names
			create {ARRAYED_LIST [RUN_GDS_FILES_MROW]} jobs.make (fld_names.count)
			from
				fld_names.start
			until
				fld_names.after
			loop
				fld_name := fld_names.item
				if not distinct_dies.item (fld_name).is_empty then
					-- There is at least 1 die in field
					n := n + 1
					create job.make_null (session.connection)
					job.set_run_id (session.run_id)
					job.set_file_no (n)
					job.set_field_name (fld_name)
					job.set_gds_filename (cache.field_merged_gds_layout_file (job).file_name)
					job.set_log_filename (cache.field_merged_gds_log_file (job).file_name)
					job.set_job_status (job_status.Created)
					jobs.extend (job)
				end
				fld_names.forth
			end
		end

	insert_die_gds_merge_jobs
			-- Insert one row in RUN_GDS_FILES for each
			-- die that is a project
		local
			dies: LIST [DIES_ROW]
			die: DIES_ROW
			job: RUN_GDS_FILES_MROW
			job_status: JOB_STATUS_VALUES
			n: INTEGER
		do
			dies := cache.die_rows
			create {ARRAYED_LIST [RUN_GDS_FILES_MROW]} jobs.make (dies.count)
			from
				dies.start
			until
				dies.after
			loop
				die := dies.item
				if die.is_project then
					n := n + 1

					create job.make_null (session.connection)
					job.set_run_id (session.run_id)
					job.set_file_no (n)
					job.set_die_id (die.die_id)
					job.set_gds_filename (cache.die_merged_gds_layout_file (job).file_name)
					job.set_log_filename (cache.die_merged_gds_log_file (job).file_name)
					job.set_job_status (job_status.Created)
					jobs.extend (job)
					
				end
				dies.forth
			end
		end

	process_jobs_completion
			-- Do processing needed after all merge jobs have
			-- been completed.
		do
			clear_run_gds_cache
			fetch_run_gds_rows
			build_lookup_tables
			incorporate_merge_job_results
			session.commit_actions.extend (agent publish_merge_results)
			call_completion_actions
		end

	build_lookup_tables
			-- Build lookup tables 
			-- run_gds_merge_results to run_gds_files.
		do
			create merge_results.make (run_gds_file_mrows.count)
			run_gds_merge_result_mrows.do_all (agent add_merge_result)
			create merge_layers.make (run_gds_file_layer_mrows.count)
			run_gds_file_layer_mrows.do_all (agent add_merge_layer)
		end

	add_merge_result (row: RUN_GDS_MERGE_RESULTS_MROW)
			-- Add `row' to `merge_results'
		do
			merge_results.put (row, row.file_no)
		end

	add_merge_layer (row: RUN_GDS_FILE_LAYERS_MROW)
			-- Add `row' to `merge_layers'
		local
			n: INTEGER
			list: ARRAYED_LIST [RUN_GDS_FILE_LAYERS_MROW]
		do
			n := row.file_no
			merge_layers.search (n)
			if merge_layers.found then
				merge_layers.found_item.extend (row)
			else
				create list.make (20)
				list.extend (row)
				merge_layers.put (list, n)
			end
		end

	build_dies_lookup_table
			-- Build lookup tables for dies
		local
			dies: LIST [DIES_ROW]
		do
			dies := cache.die_rows
			create all_dies.make (dies.count)
			dies.do_all (agent add_die)
		end

	add_die (die: DIES_ROW)
			-- Add `row' to `all_dies'
		do
			all_dies.put (die, die.die_id)
		end

	incorporate_merge_job_results
			-- Propagate results from
			-- run_gds_merge_results to run_gds_files.
		do
			run_gds_file_mrows.do_all (agent propagate_job_results)
		end

	propagate_job_results (job: RUN_GDS_FILES_MROW)
			-- Propagate results from merge into `job' if
			-- it has not already been done and delete merge
			-- results row if it exists
		local
			merge_result: RUN_GDS_MERGE_RESULTS_MROW
			n: INTEGER
		do
			n := job.file_no
			merge_results.search (n)
			if merge_results.found then
				merge_result := merge_results.found_item
				job.set_top_structure (merge_result.non_null_top_structure)
				job.set_size_x (merge_result.non_null_size_x)
				job.set_size_y (merge_result.non_null_size_y)
				job.set_center_x (merge_result.non_null_center_x)
				job.set_center_y (merge_result.non_null_center_y)
				merge_result.delete
			elseif job.is_null_top_structure or job.is_null_size_x then
				handle_hard_error ("cannot find merge results row for file_no " + n.out)
			else
				-- No merge results row, but results
				-- were previously propagated into job
				-- so we're OK.  Merge results columns in `job' 
				-- were set to null when job was submitted
				-- so it they're non-null now, then they must
				-- have been previously propagated
			end
		end

	publish_merge_results
			-- Publish a "all GDS merges completed" event
		local
			e: GDS_MERGE_RESULTS_EVENT
			type: STRING
		do
			if cache.is_die_merge then
				type := "die"
			else
				type := "field"
			end
			create e
			e.set_subject ("GDS " + type + " merge results")
			e.set_message ("GDS " + type + " merge results for run " + session.run_id + ".%N%N" + gds_merge_results_text)
			e.publish
		end

	gds_merge_results_text: STRING is
			-- Text for body of message reporting results
			-- of GDS merge
		do
			create Result.make (500 * run_gds_file_mrows.count)
			if not run_gds_file_mrows.is_empty then
				build_dies_lookup_table
				create design_techn_layers.make (5)
				run_gds_file_mrows.do_all (agent append_file_merge_results (?, cache.is_die_merge, Result))
			elseif cache.is_die_merge then
				Result.append ("No merged GDS files were created because there were no customer projects%Nor test vehicles on the run.%N")
			else
				Result.append ("No merged GDS files were created because all fields were empty.%N")
			end
		end

	append_file_merge_results (row: RUN_GDS_FILES_MROW is_die_merge: BOOLEAN dest: STRING)
			-- Append merge results for `row' to `dest'
		local
			fname, low_x, low_y, high_x, high_y, dt_id: STRING
			layer_name, gds_name, gds_layer, gds_datatype: STRING
			size_x, size_y, center_x, center_y: DOUBLE
			file: LARGE_RAW_FILE
			layers: LIST [RUN_GDS_FILE_LAYERS_ROW]
			one_die: DIES_ROW
			layers_table: like layers_for_design_techn
			dtl_row: GDSXDESIGN_TECHN_ID_LAYERS_ROW
			pos: INTEGER
		do
			if is_die_merge then
				fname := cache.die_merged_gds_layout_file (row).file_pathname
				all_dies.search (row.die_id)
				if all_dies.found then
					one_die := all_dies.found_item
				else
					handle_hard_error ("cannot find DIES row for die ID " + row.die_id)
				end
				dest.append ("File: " + fname + "%N")
				dest.append ("Die: " + row.die_id + "%N")
			else
				fname := cache.field_merged_gds_layout_file (row).file_pathname
				dest.append ("File: " + fname + "%N")
				dest.append ("Field: " + row.field_name + "%N")
				one_die := distinct_dies.item (row.field_name).first
			end
			dt_id := one_die.wafer_design_techn
			dest.append ("Design Technology: " + dt_id + "%N")
			create file.make (fname)
			dest.append ("Size: " + file.count.out + "%N")
			dest.append ("Top Structure: " + row.top_structure + "%N")
			size_x := row.size_x
			size_y := row.size_y
			center_x := row.center_x
			center_y := row.center_y
			low_x := nu.formatted_number_or_integer (center_x - size_x / 2.0, Extent_precision)
			low_y := nu.formatted_number_or_integer (center_y - size_y / 2.0, Extent_precision)
			high_x := nu.formatted_number_or_integer (center_x + size_x / 2.0, Extent_precision)
			high_y := nu.formatted_number_or_integer (center_y + size_y / 2.0, Extent_precision)
			dest.append ("Extent: (" + low_x + ", " + low_y + ")  (" + high_x + ", " + high_y + ")%N")
			
			layers_table := layers_for_design_techn (dt_id)
			dest.append ("Layers:%N")
			
			-- Add the known layers (ones that do not
			-- start with `Unknown_layer_prefix')
			
			from
				layers := merge_layers.item  (row.file_no)
				layers.start
			until
				layers.after
			loop
				layer_name := layers.item.layer_name
				if not su.is_prefix (Unknown_layer_prefix, layer_name) then
					-- Known layer
					layers_table.search (layer_name)
					if layers_table.found then
						dtl_row := layers_table.found_item.first
						gds_layer := dtl_row.gds_layer.out
						if not dtl_row.is_null_gds_datatype then
							gds_datatype := dtl_row.gds_datatype.out
						else
							gds_datatype := "0"
						end
						gds_datatype.prepend_character (Gds_datatype_separator)
					else
						handle_hard_error ("no GDS layers found for design techn " + dt_id + " and layer name " + layer_name)
					end
					dest.append (layer_line (layer_name, gds_layer, gds_datatype))
				end
				layers.forth
			end
			
			-- Add the unknown layers (ones that 
			-- start with `Unknown_layer_prefix')
			
			from
				layers := merge_layers.item  (row.file_no)
				layers.start
			until
				layers.after
			loop
				layer_name := layers.item.layer_name
				if su.is_prefix (Unknown_layer_prefix, layer_name) then
					-- Unknown layer
					gds_name := layer_name.substring (Unknown_layer_prefix.count + 1, layer_name.count)
					pos := gds_name.index_of (Gds_datatype_separator, 1)
					if pos > 0 then
						gds_layer := gds_name.substring (1, pos - 1)
						gds_datatype := gds_name.substring (pos + 1, gds_name.count)
					else
						gds_layer := gds_name
						gds_datatype := "???"
					end
					gds_datatype.prepend_character (Gds_datatype_separator)
					layer_name := Unknown_layer_name
					dest.append (layer_line (layer_name, gds_layer, gds_datatype))
				end
				layers.forth
			end
			dest.extend ('%N')
		end

	layers_for_design_techn (dt: STRING): HASH_TABLE [LIST [GDSXDESIGN_TECHN_ID_LAYERS_ROW], STRING]
			-- Layers lookup table for design techn `dt'
		local
			list: LIST [GDSXDESIGN_TECHN_ID_LAYERS_ROW]
			row: GDSXDESIGN_TECHN_ID_LAYERS_ROW
		do
			design_techn_layers.search (dt)
			if design_techn_layers.found then
				Result := design_techn_layers.found_item
			else
				create row.make_template
				list := row.rows_for_design_techn (dt, connection, expected_count.At_least_one)
				create Result.make (list.count)
				list.do_all (agent add_design_techn_layer (?, Result))
				design_techn_layers.put (Result, dt)
			end
		end

	add_design_techn_layer (dtl: GDSXDESIGN_TECHN_ID_LAYERS_ROW dest: like layers_for_design_techn)
			-- Add `dt' to `dest'
		local
			list: ARRAYED_LIST [GDSXDESIGN_TECHN_ID_LAYERS_ROW]
		do
			dest.search (dtl.layer_name)
			if dest.found then
				dest.found_item.extend (dtl)
			else
				create list.make (4)
				list.extend (dtl)
				dest.put (list, dtl.layer_name)
			end
		end

	layer_line (layer_name, gds_layer, gds_datatype: STRING): STRING
			-- Line describing layer `layer_name' with indicated
			-- layer number and datatype
		do
			Result := "    " + su.right_padded (layer_name, 25) + " " + su.left_padded (gds_layer, 3) + gds_datatype + "%N"
		end

	set_state
			-- Set `state'
		do
			state := Void
		end

	generated_files: LIST [RC_FILE]
			-- Files which may have been generated during
			-- execution of step (were generated if step
			-- finished)
		local
			row: RUN_GDS_FILES_ROW
		do
			create {ARRAYED_LIST [RC_FILE]} Result.make (10)
			if cache.is_die_merge then
				from
					run_gds_file_mrows.start
				until
					run_gds_file_mrows.after
				loop
					row := run_gds_file_mrows.item
					Result.extend (cache.die_merged_gds_layout_file (row))
					Result.extend (cache.die_merged_gds_log_file (row))
					run_gds_file_mrows.forth
				end
			else
				from
					run_gds_file_mrows.start
				until
					run_gds_file_mrows.after
				loop
					row := run_gds_file_mrows.item
					Result.extend (cache.field_merged_gds_layout_file (row))
					Result.extend (cache.field_merged_gds_log_file (row))
					run_gds_file_mrows.forth
				end
			end
		end

	delete_files (list: LIST [RC_FILE])
			-- Delete all files in `list' that exist
		do
			list.do_all (agent {RC_FILE}.delete_if_exists)
		end

	run_gds_file_mrows: LIST [RUN_GDS_FILES_MROW]
			-- Rows for run in RUN_GDS_FILES table

	run_gds_file_layer_mrows: LIST [RUN_GDS_FILE_LAYERS_MROW]
			-- Rows for layers for run in RUN_GDS_FILE_LAYERS table

	run_gds_merge_result_mrows: LIST [RUN_GDS_MERGE_RESULTS_MROW]
			-- Rows for run in RUN_GDS_MERGE_RESULTS table

	merge_results: HASH_TABLE [RUN_GDS_MERGE_RESULTS_MROW, INTEGER]
			-- Merge result rows, indexed by file_no

	merge_layers: HASH_TABLE [LIST [RUN_GDS_FILE_LAYERS_MROW], INTEGER]
			-- Lists of merge layer rows, indexed by file_no

	all_dies: HASH_TABLE [DIES_ROW, STRING]
			-- Dies, indexed by die ID

	jobs: LIST [RUN_GDS_FILES_MROW]
			-- Rows to be inserted for jobs to create
			-- merged GDS files

	distinct_dies: HASH_TABLE [LIST [DIES_XROW], STRING]
			-- Lists of distinct dies in field, indexed by
			-- field name.  Used to skip creating a GDS
			-- merge job for fields that do not have any
			-- dies and to determine the wafer design
			-- techn for each merged GDS file in the field
			-- merge case
	
	design_techn_dies: HASH_TABLE [LIST [DIES_ROW], STRING]
			-- Lists of dies with a given wafer design techn,
			-- indexed by wafer design techn (for one field)
	
	die_layers: HASH_TABLE [INTEGER, STRING]
			-- Table to record which layers are present
			-- for a die, for fast lookup

	derived_layer_dies: HASH_TABLE [LIST_PAIR [DIES_ROW], STRING]
			-- Pairs of lists of dies that do and do not
			-- have a derived layer, indexed by layer name

	is_valid: BOOLEAN
			-- Were any validity errors detected (done before
			-- submitting GDS merge jobs)?

	multiple_design_techns_message: STRING
			-- Error message for fields with multiple design
			-- techns

	derived_layers_message: STRING
			-- Error message for fields with mix of dies that
			-- do not all contain the same set of "derived when
			-- absent" layers

	design_techn_layers: HASH_TABLE [HASH_TABLE [LIST [GDSXDESIGN_TECHN_ID_LAYERS_ROW], STRING], STRING]
			-- Hash tables of lists of GDS design techn layer
			-- rows indexed by layer name, indexed by design techn
	
	Extent_precision: INTEGER = 6
			-- Maximum number of digits after decimal point
			-- for extent displayed in merge results message

	Unknown_layer_prefix: STRING = "GDS_IS_"
			-- Prefix for unknown layers in files input to
			-- GDS merge

	Unknown_layer_name: STRING = "<UNKNOWN>"
			-- Layer name to display for unknown layers

	Gds_datatype_separator: CHARACTER = ';'
			-- Separator between GDS layer and datatype

end
