
indexing
	
	description: "A run closing step operation that generates %
		%a survey report of the wafer technology for the run, %
		%which preserves the current data values in a file"
	author: "David Hollenberg"
	date: "February 19, 2004"

class GEN_WAFER_TECHN_SURVEY_OP
inherit
	RC_PRIMITIVE_STEP_OP
	RC_APPLICABLE_STEP
	NO_USER_INTERFACE
	CREATES_EXTERNAL_STATE
	UNIX_OS_ACCESS
		export
			{NONE} all
		undefine
			default_create
		end

create
	make

feature -- Properties

	input_data_items: LIST [STRING]
			-- Data items required before step can be executed
		do
			Result := array_as_list (<<
				data_item.Wafer_techn
				>>)
		end

feature {NONE} -- Implementation

	do_step
			-- Execute of step
		local
			script_fname: STRING
			script_file: LARGE_PLAIN_TEXT_FILE
			process: UNIX_PROCESS
			log_file: RC_FILE
			input: LARGE_RAW_FILE
			sqlplus_fname: STRING
		do
			sqlplus_fname := session.environment.sqlplus_exe_file
			if not session.path_exists (sqlplus_fname) then
				handle_hard_error ("file " + su.double_quoted (sqlplus_fname) + " not found")
			end

			script_fname := session.environment.wafer_techn_survey_script
			-- Make sure script file exists
			create script_file.make (script_fname)
			session.open_file_for_read (script_file)
			session.close_file (script_file)

			log_file := survey_log_file
			create file.make (log_file.file_pathname)
			session.open_file_for_write (file, log_file)
			session.close_file (file)
			process := unix_os.new_process (session.environment.sqlplus_exe_file)
			process.set_arguments (<< "/nolog" >>)
			process.set_close_nonstandard_files (True)
			process.set_input_piped
			process.set_output_file_name (file.name)
			process.set_error_file_name (file.name)
			process.spawn_nowait
			input := process.input_to_child
			input.put_string ("set sqlprompt ''%N")
			input.put_string ("set tab off%N")
			input.put_string ("connect " + connection.user_id + "/" + connection.password + "@" + connection.connect_string + "%N")
			input.put_string ("select " + su.single_quoted (start_tag) + From_dual)
			input.put_string ("@" + session.environment.wafer_techn_survey_script + " " + cache.wafer_techn + "%N")
			input.put_string ("select " + su.single_quoted (end_tag) + From_dual)
			input.put_string ("exit%N")
			session.close_file (input)
			process.get_status_block
			verify_success (file)
			call_completion_actions
		end

	undo_step
			-- Undo execution of step
		local
			fname: STRING
		do
			fname := survey_log_file.file_pathname
			if session.path_exists (fname) then
				session.delete_file (fname)
			end
			call_completion_actions
		end

	set_state
			-- Set `state'
		do
			state := Void
		end

	verify_success (f: LARGE_FILE)
			-- Check contents of `f' to verify that script
			-- ran successfully.  Report error if not
		require
			file_exists: f /= Void
		local
			start_pattern, end_pattern, err_pattern: RX_PCRE_REGULAR_EXPRESSION
			found_start, found_end: BOOLEAN
			err: STRING
			line: STRING
			regexp_op: REGEXP_OP
		do
			create regexp_op

			create start_pattern.make
			start_pattern.set_anchored (True)
			start_pattern.set_strict (True)
			regexp_op.compile (start_pattern, start_tag + Dollar)
			
			create end_pattern.make
			end_pattern.set_anchored (True)
			end_pattern.set_strict (True)
			regexp_op.compile (end_pattern, end_tag + Dollar)
			
			create err_pattern.make
			err_pattern.set_anchored (True)
			err_pattern.set_caseless (True)
			err_pattern.set_strict (True)
			regexp_op.compile (err_pattern, "Not connected|ORA-|Unable to open file")
			
			session.open_file_for_read (f)
			from
				line_number := 0
				f.read_line
			until
				f.end_of_file or err /= Void
			loop
				line := f.last_string
				line_number := line_number + 1
				
				if not found_start then
					start_pattern.match (line)
					if start_pattern.has_matched then
						found_start := True
					end
				end
				if not found_end then
					end_pattern.match (line)
					if end_pattern.has_matched then
						found_end := True
					end
				end
					
				err_pattern.match (line)
				if err_pattern.has_matched then
					err := line.twin
				end
				
				f.read_line
			end
			session.close_file (f)
			if err /= Void then
				report_error (err)
			elseif not found_start then
				report_error ("start tag " + su.double_quoted (start_tag) + " not found")
			elseif not found_end then
				report_error ("end tag " + su.double_quoted (end_tag) + " not found")
			end
		end

	report_error (s: STRING)
			-- Report error with text `s'
		do
			handle_hard_error (s + "%Nat line " + line_number.out + " of file " + file.name)
		end

	survey_log_file: RC_FILE
			-- Log file where survey output is written
		do
			Result := file_generator.wafer_techn_survey_log_file
		ensure
			result_exists: Result /= Void
		end

	start_tag: STRING = "Database login succeeded"

	end_tag: STRING = "Reached end of script"

	From_dual: STRING = " from dual;%N"

	Dollar: STRING = "$"
			-- Regular expression end-of-string character

	file: LARGE_PLAIN_TEXT_FILE
			-- File with output of SQL script
	
	line_number: INTEGER
			-- Current line number in output file

end
