class GENERIC_APPLICATION_PARAMETERS
	-- fetch our application_parameters (which are read-only),
	--   returning in the form desired.
	-- fetch our application_events (which are updateable) as rows
	-- fetch our application_colors as rows
	-- the application for all three tables is set at creation time

inherit

	ANY
	SEVERITY_LEVELS
	export {NONE} all end;

	DB_ACCESS
	export {NONE} all end;

	CONDITION_UTILITIES
	export {NONE} all end;

	DB_QUERY_OPTIONS
	export {NONE} all end;

create
	make

feature	-- application name

	application: STRING	-- set by make

feature -- parameter existence

	has(pnam: STRING): BOOLEAN is
		-- do we have a valid application parameter for pnam?
	do  try_to_get_row(pnam);
	    Result := last_row_is_valid
	ensure Result implies equal(last_row.parameter_name, pnam);
	end;

feature -- parameter fetch routines for PARAMETER_NAME pnam

	as_boolean(pnam: STRING): BOOLEAN is
	require pnam /= Void;
	local vv: STRING;
	do  get_row(pnam);
	    vv := last_row.char_value;
	    if vv.is_empty
	     then Result := last_row.numeric_value > 0
	     else Result := (vv @ 1) = 'Y'
	    end;
	end;

	as_string(pnam: STRING): STRING is
	require pnam /= Void;
	do  get_row(pnam);
	    Result := last_row.char_value
	ensure Result /= Void
	end;

	as_string_or_Void(pnam: STRING): STRING is
		-- returns Void if no such parameter
	require pnam /= Void;
	do  if has(pnam)
	     then get_row(pnam);
		  Result := last_row.char_value
	    end
	end;

	as_character(pnam: STRING): CHARACTER is
	require pnam /= Void;
	local vv: STRING;
	do  get_row(pnam);
	    vv := last_row.char_value;
	    if vv.is_empty
	     then Result := '%U'
	     else Result := vv @ 1
	    end;
	end;

	as_integer(pnam: STRING): INTEGER is
	require pnam /= Void;
	do  get_row(pnam);
	    Result := last_row.numeric_value.rounded
	end;

	as_double(pnam: STRING): DOUBLE is
	require pnam /= Void;
	do  get_row(pnam);
	    Result := last_row.numeric_value
	end;

	as_fixed_length_string(pnam: STRING): STRING is
		-- char_value, but pad to numeric_value length
	require pnam /= Void;
	local cnt: INTEGER;
	do  cnt := as_integer(pnam);
	    Result := last_row.char_value;
	    from
	    until cnt <= Result.count
	    loop  Result.extend(' ')
	    end;
	    if cnt > 0 then Result.keep_head(cnt) end;
	end;

	as_severity_level(pnam: STRING): INTEGER is
	require pnam /= Void
	local sev: CHARACTER
	do  sev := as_character(pnam);
	    Result := severity_level(sev);
	end;

feature -- application event fetching

	event(pnam: STRING): APPLICATION_EVENTS_MROW is
		-- (presumably, one reads and updates the row)
	require pnam /= Void;
	do  create Result.make_by_key(db, old_or_new_row,
					application, pnam)
	end;

feature	-- application color fetching

	as_color(crole: STRING): APPLICATION_COLORS_ROW is
		-- just fetch it
	require crole /= Void;
	do  create Result.make_by_key(db, old_row,
					application, crole)
	end;

feature {NONE} -- implementation

	last_row: APPLICATION_PARAMETERS_ROW
		-- very small cache

	get_row(pnam: STRING) is
	do  try_to_get_row(pnam);
	   if not last_row_is_valid
	    then handle_hard_error("Policy " + pnam +
					" has no value.")
	    end
	end;

	try_to_get_row(pnam: STRING) is
	do  if last_row = Void or else
		not last_row.parameter_name.is_equal(pnam)
	     then create last_row.make_by_key(db, old_no_exception_row,
					 application, pnam)
	    end;
	ensure	equal(last_row.parameter_name, pnam);
	end;

	last_row_is_valid: BOOLEAN is
		-- must have fetched a row
	do  Result := last_row.is_old
	end;

feature {NONE} -- creation

	make(an_app: STRING) is
	do  application := an_app
	    last_row := Void
	end;

end -- GENERIC_APPLICATION_PARAMETERS
