class GENERIC_SYSTEM_GLOBALS
		-- global stuff that is accessed thru glo.xxxx

inherit
	DB_ACCESS

	SEVERITY_LEVELS

	STRING_UTILITIES_ACCESS

	OS_ACCESS

	EXCEPTIONS_ACCESS

feature -- system internal error exception name, error message, and punt

	system_internal_error: STRING
		-- exception string for (unrecoverable) internal error
		-- a "constant"

	set_system_internal_error(s: STRING) is
		-- normally set once at the top and left alone
	require su.not_empty(s)
	do  system_internal_error := s
	end;

	system_error_message: STRING
		-- actual message for this error

	system_punt(s: STRING) is
	do  system_error_message := s;
	    if su.is_empty(system_internal_error)
	     then set_system_internal_error("SYSTEM: internal error")
	    end;
	    exc.raise(system_internal_error);
	end;

feature -- last_chance

	last_chance: BOOLEAN
		-- we are on the last round for database retry before we punt
		-- for those who have their own (better) rescue

	set_last_chance(lc: BOOLEAN) is
		-- set it
	do  last_chance := lc
	ensure last_chance = lc
	end;

feature -- commit_actions & rollback_actions & post_execution_actions
	--  other things to do at the end of a transaction

	extend_pre_commit_actions(item: ACTION_ITEM) is
			-- add item to the pre_commit list
	require not item.done
	do  if pre_commit_actions = Void
	     then create pre_commit_actions.make(2)
	    end;
	    pre_commit_actions.extend(item)
	end;
	
	extend_commit_actions(item: ACTION_ITEM) is
			-- add item to the commit list
	require not item.done
	do  if commit_actions = Void
	     then create commit_actions.make(12)
	    end;
	    commit_actions.extend(item)
	end;
	
	extend_rollback_actions(item: ACTION_ITEM) is
			-- add item to the rollback list
	require not item.done
	do  if rollback_actions = Void
	     then create rollback_actions.make(12)
	    end;
	    rollback_actions.extend(item)
	end;

	extend_post_execution_actions(item: ACTION_ITEM) is
			-- add item to the post_execution list
	require not item.done
	do  if post_execution_actions = Void
	     then create post_execution_actions.make(12)
	    end;
	    post_execution_actions.extend(item)
	end;

feature -- common commit and rollback action constructors

	delete_on_rollback(fn: STRING) is
	local  a: FILE_DELETE_ACTION
	do  create a.make(fn);
	    extend_rollback_actions(a);
	end;

	delete_on_commit(fn: STRING) is
	local  a: FILE_DELETE_ACTION
	do  create a.make(fn);
	    extend_commit_actions(a);
	end;

	rename_on_commit(ofn, nfn: STRING) is
	local  a: FILE_RENAME_ACTION
	do  create a.make(ofn, nfn);
	    extend_commit_actions(a);
	end;

	hard_link_on_commit(ofn, nfn: STRING) is
	local  a: FILE_HARD_LINK_ACTION
	do  create a.make(ofn, nfn);
	    extend_commit_actions(a);
	end;

	unlink_on_commit (fn: STRING) is
	local  a: FILE_UNLINK_ACTION
	do  create a.make(fn);
	    extend_commit_actions(a);
	end;

	rename_or_delete(ofn, nfn: STRING) is
	do  rename_on_commit(ofn, nfn);
	    delete_on_rollback(ofn);
	end;

	mail_on_commit(msg: OUTGOING_MESSAGE) is
	local  a: MESSAGE_SEND_ACTION;
	do  create a.make(msg);
	    extend_commit_actions(a);
	end;

	close_on_rollback(f: LARGE_FILE) is
		-- (in case we abort around the file opener)
	local  a: FILE_CLOSE_ACTION;
	do  create a.make(f);
	    extend_rollback_actions(a);
	end;

	execute_sql_before_commit(sql: SQL_STATEMENT) is
	local  a: SQL_STATEMENT_ACTION
	do  create a.make(sql);
	    extend_pre_commit_actions(a);
	end;

feature	-- commit & rollback
	-- our official commit and rollback interface

	commit is
			-- database commit and then the action items
	do
	    log_cursor_stats
	    log_line_2("Committing #", db.transaction_number.out);
	    db.post;
	    take_action(pre_commit_actions, word_is_pre_commit);
	    db.commit;
	    take_action(commit_actions, word_is_commit)
	    reset_actions	-- reset three of the lists
	    post_execution
	    log_cursor_stats
	end;
	
	rollback is
			-- database rollback and then the action items
	do  log_line_2("Rolling back #", db.transaction_number.out);
	    db.rollback;
	    take_action(rollback_actions, word_is_rollback)
	    reset_actions
	    post_execution
	end;

	reset_actions is
	do  if pre_commit_actions /= Void
	     then pre_commit_actions.wipe_out
	    end;
	    if commit_actions /= Void
	     then commit_actions.wipe_out
	    end;
	    if rollback_actions /= Void
	     then rollback_actions.wipe_out
	    end;
	end;

feature {NONE} -- details

	pre_commit_actions,
	commit_actions,
	rollback_actions,
	post_execution_actions: ARRAYED_LIST[ACTION_ITEM];
			-- the actual action item lists

	word_is_pre_commit: STRING is "PRE-COMMIT action failure: "
	word_is_commit: STRING is "COMMIT action failure: "
	word_is_rollback: STRING is "ROLLBACK action failure: "
	word_is_post_exec: STRING is "POST_EXEC action failure: "

	take_action(it: LIST[ACTION_ITEM]; the_word: STRING) is
		-- try all the actions; punt if any failed
	local foo: ACTION_ITEM; oops: STRING
	do  if it /= Void
	     then from  it.start
		  until it.after
		  loop	foo := it.item;
			foo.try_it;
			if not foo.done
			 then log_line_2(foo.descr, foo.exception_cause);
			      if oops = Void then create oops.make(0) end;
			      oops.append(foo.descr);
			      oops.append(su.colon_space);
			      oops.append(foo.exception_cause);
			      oops.extend(';');
			end;
			it.forth
		  end
	    end;
	    if oops /= Void
	     then -- we are going down
		  oops.prepend(the_word)
		  reset_actions;	-- (get rid of rollback_actions)
		  system_punt(oops)
	    end;
	end;

	post_execution is
			-- AFTER commit or rollback is DONE
	do  if post_execution_actions /= Void
	     then take_action(post_execution_actions, word_is_post_exec)
		  post_execution_actions.wipe_out
	    end
	end;

feature -- logging of random information (on our log file)

	is_logging: BOOLEAN

	set_is_logging(b: BOOLEAN) is
	do  is_logging := b
	end;

	begin_log_line(hdr: STRING) is
	require is_logging
	local foo: STRING;
	do foo := os.current_date_string;	-- "SUN JAN 31 23:59:59 1999"
	   io.putstring(foo.substring(5, 20));	--     "JAN 31 23:59:59 "
	   if su.not_empty(hdr) then io.putstring(hdr) end;
	end;

	end_log_line is
	require is_logging
	do  io.new_line;
	    io.default_output.flush;
	end;
	
	log_line_2(a,b: STRING) is
		-- quickie complete line
	do if is_logging
	    then begin_log_line(a);
		 if su.not_empty(b)
		  then	io.putchar(' ');
			io.putstring(b);
		 end;
		 end_log_line
	   end
	end;

	as_pid: STRING is " as PID ";

	log_pid_line(v, n: STRING; pid: INTEGER) is
	do if is_logging
	    then begin_log_line(v);		-- the verb
		 io.putchar(' ');
		 io.putstring(n);		-- the noun
		 io.putstring(as_pid);
		 io.putint(pid);
		 end_log_line
	    end
	end;

	log_cursor_stats is
	do
	-- if is_logging and db.is_open
	--  then begin_log_line("Cursor stats: in_use: ");
	--	 io.putint(db.in_use_statement_count);
	--	 io.putstring("; parsed: ");
	--	 io.putint(db.cached_parsed_cursor_count);
	--	 io.putstring("; open: ");
	--	 io.putint(db.cached_open_cursor_count);
	--	 io.putstring("; avail: ");
	--	 io.putint(db.available_cursor_count);
	--	 end_log_line
	-- end
	end;

end -- GENERIC_SYSTEM_GLOBALS
