
indexing
	
	author: "David Hollenberg"
	date: "September 15, 1998"
	description: "A virtual interval timer that counts down in %
		%real time.  Multiple virtual timers share the one %
		%actual system timer"

class INTERVAL_TIMER

inherit
	ANY
		redefine
			is_equal
		end
	UNIX_OS_ACCESS
		export
			{NONE} all
		redefine
			is_equal
		end
	INTERVAL_TIMER_SERVER_ACCESS
		export
			{NONE} all
		redefine
			is_equal
		end
	EXCEPTIONS
		export
			{NONE} all
		redefine
			is_equal
		end
	UNIX_SIGNALS
		rename
			meaning as signal_meaning,
			ignore as signal_ignore,
			catch as signal_catch
		export
			{NONE} all
		redefine
			is_equal
		end
	C_MEMORY_ALLOCATION
		export
			{NONE} all
		redefine
			is_equal
		end
	MEMORY
		export
			{NONE} all
		redefine
			is_equal, dispose
		end
		

create
	set_seconds

feature -- Creation

   	set_seconds (t: DOUBLE)
			-- Create new virtual interval timer object
			-- set to expire after `t' seconds
		require
			nonnegative_interval: t >= 0.0
		local
			interval: DOUBLE
		do
			-- Make sure interval doesn't violate Unix
			-- system timer interval requirements
			
			if t < 1.0E-6 then
				interval := 1.0E-6
			elseif t > 100_000_000 then
				interval := 100_000_000
			else
				interval := t
			end
			interval_timer_server.set_new_timer (Current, interval)
		end

feature {INTERVAL_TIMER_SERVER} -- Initialization

   	initialize (t: DOUBLE id: INTEGER)
			-- Create new virtual interval timer object
			-- set for `t' seconds, with unique identifier `id'
		local
			addr: POINTER
		do
			seconds_original := t
			identifier := id
			addr := allocate_nongc_memory (8)
			
			-- Set `is_expired' to False in C memory before
			-- putting its address in `Current'
			
			set_boolean_value (addr, False)
			address := addr
			time_when_set := unix_os.current_time_in_fine_seconds
		ensure
			timer_set: not is_expired
			interval_set: seconds_original = t
			identifier_set: identifier = id
		end

feature -- Properties

	identifier: INTEGER
			-- Unique identifier for `Current'
		
	is_expired: BOOLEAN
			-- Has timer expired?  Expiration of timer
			-- always raises an exception
		do
			if address /= default_pointer then
				Result := boolean_value (address)
			end
		end
	
	is_expiration_exception: BOOLEAN
			-- Is last exception originally due to expiration
			-- of `Current'?  Must only be referenced in
			-- a routine rescue clause
		do
			interval_timer_server.handle_possible_timer_expiration
			Result := is_signal and then signal = Sigalrm and
				is_expired and
				identifier = interval_timer_server.last_expiration_identifier
		end
	
	seconds_original: DOUBLE
			-- Number of seconds timer was originally set for
	
	seconds_remaining: DOUBLE
			-- Number of seconds left until timer expires.
			-- May be <= 0.0 before timer expires due to
			-- delay in operating system delivering alarm signal
			-- or delay in signal being processed
		do
			Result := seconds_remaining_at (unix_os.current_time_in_fine_seconds)
		end

	seconds_elapsed: DOUBLE
			-- Number of seconds that have elapsed since
			-- timer was set.  May be greater than or equal to
			-- `seconds_original' before timer expires
			-- due to delay in delivery of alarm signal
		do
			Result := seconds_elapsed_at (unix_os.current_time_in_fine_seconds)
		end

	is_equal (other: like Current): BOOLEAN
			-- Is `other' attached to an object considered
			-- equal to current object?
		do
			Result := identifier = other.identifier
		end

feature -- Modification

	clear
		require
			timer_set: not is_expired
		do
			set_is_expired (True)
			interval_timer_server.release_cleared_timer (Current)
		ensure
			timer_not_set: is_expired
		end
	
	notify_exception_occurred
			-- Notify timer manager that an exception
			-- occurred, so that it can start next virtual
			-- timer if a timer expiration caused the exception
			-- This routine should be called in the rescue
			-- clause of every routine that uses a timer
			-- (not needed if `is_expiration_exception' is called).
			-- It is safe to call this routine at any time,
			-- in either rescue clause or routine body,
			-- even if no exception has occurred
		do
			interval_timer_server.handle_possible_timer_expiration
		end
	
feature {INTERVAL_TIMER_SERVER} -- Implementation

	seconds_remaining_at (t: DOUBLE): DOUBLE
			-- Number of seconds remaining on timer at time `t'.
			-- May be <= 0.0 seconds
		do
			Result := time_when_set + seconds_original - t
		end
	
	seconds_elapsed_at (t: DOUBLE): DOUBLE
			-- Number of seconds that had elapsed at time `t'
			-- since timer was set
		do
			Result := t - time_when_set
		end
	
	address: POINTER
			-- Address of C memory for `is_expired' indicator

feature {NONE} -- Implementation

	time_when_set: DOUBLE
			-- Time in seconds when timer was set
	
	set_is_expired (b: BOOLEAN)
			-- Set `is_expired' to `b'
		do
			set_boolean_value (address, b)
		end

	dispose
			-- Deallocate resources when object is
			-- garbage collected
		local
			null_ptr: POINTER
		do
			if address /= null_ptr then
				free_nongc_memory (address)
				address := null_ptr
			end
		end

	boolean_value (addr: POINTER): BOOLEAN
			-- Boolean value at address `addr'
		external
			"C"
		end

	set_boolean_value (addr: POINTER b: BOOLEAN)
			-- Set boolean value at address `addr' to `b'
		external
			"C"
		end

end
