
indexing
	
	description: "Window to control jobs that do fracture"
	author: "David Hollenberg"
	date: "November 17, 2003"

deferred class JOB_CONTROL_WINDOW [reference G -> JOB_MULTI_COLUMN_LIST_ROW]

inherit
	JOB_CONTROL_IFC
	WINDOW_INTERFACE
	MOSIS_GRAPHICAL_UTILITIES
		export
			{NONE} all
		undefine
			default_create
		end
		
feature {NONE} -- Creation

	make (w: EV_TITLED_WINDOW)
			-- Create interface in window `w'
		require
			window_exists: w /= Void
			window_valid: not w.is_destroyed
		do
			make_interface
			make_proxy (w)
		end

feature -- Operations

	reset
			-- Reset interface
		do
		end

	autosubmit_jobs
			-- Automatically select all jobs that are submittable,
			-- submit them and then start monitoring
		do
			process_select_submittable
			refresh_selected_count
			if selected_count > 0 then
				process_submit (False)
			end
			if all_rows.there_exists (is_stopable_agent) or all_rows.for_all (is_passed_agent) then
				process_monitor
			end
		end

feature {NONE} -- Deferred implementation

	refresh_corrupted_rows
			-- Refresh requestor for elements of `all_rows'
			-- that were modified and became invalid due
			-- to a rollback
		deferred
		end

	batch_jobs_for_run: LIST [BATCH_JOBS_MROW]
			-- List of batch jobs for this run
		deferred
		end
			
	initialize_total_count
			-- Initialize `total_count'
		deferred
		end

	fetch_needed_data
			-- Fetch needed data when first building interface
		deferred
		end

	fetch_requestor_list
			-- Fetch list of rows "requesting" a batch job
		deferred
		end

	build_all_rows
			-- Build list `all_rows'
		deferred
		end

	add_visible_columns
			-- Add columns to be displayed initially to
			-- `visible_columns'
		deferred
		ensure
			at_least_one_column: not visible_columns.is_empty
		end
			
	add_extra_select_menu_items (select_menu: EV_MENU)
			-- Add any application-specific items to `select_menu'
		require
			select_menu_exists: select_menu /= Void
		deferred
		end

	add_extra_view_row_menu_items (row_menu: EV_MENU)
			-- Add any application-specific items to `row_menu'
		require
			row_menu_exists: row_menu /= Void
		deferred
		end

	has_job_status_agent (s: STRING): FUNCTION [G, TUPLE [G], BOOLEAN]
			-- Agent that tells whether row has job status `s'
		require
			die_status_exists: s /= Void
		deferred
		ensure
			result_exists: Result /= Void
		end

	is_stopable_agent: FUNCTION [G, TUPLE [G], BOOLEAN]
			-- Agent that tells whether row has job that
			-- is stopable
		deferred
		ensure
			result_exists: Result /= Void
		end

	is_unstopable_agent: FUNCTION [G, TUPLE [G], BOOLEAN]
			-- Agent that tells whether row does not have a
			-- job that is stopable
		deferred
		ensure
			result_exists: Result /= Void
		end

	is_resettable_agent: FUNCTION [G, TUPLE [G], BOOLEAN]
			-- Agent that tells whether row has job that
			-- is resettable
		deferred
		ensure
			result_exists: Result /= Void
		end

	is_unresettable_agent: FUNCTION [G, TUPLE [G], BOOLEAN]
			-- Agent that tells whether row does not have
			-- a job that is resettable
		deferred
		ensure
			result_exists: Result /= Void
		end

	is_unsuccessful_agent: FUNCTION [G, TUPLE [G], BOOLEAN]
			-- Agent that tells whether row has job that
			-- was unsuccessful
		deferred
		ensure
			result_exists: Result /= Void
		end

	is_submittable_agent: FUNCTION [G, TUPLE [G], BOOLEAN]
			-- Agent that tells whether row has job that
			-- is submittable
		deferred
		ensure
			result_exists: Result /= Void
		end

	is_unsubmittable_agent: FUNCTION [G, TUPLE [G], BOOLEAN]
			-- Agent that tells whether row is not in a state
			-- where job can be submitted
		deferred
		ensure
			result_exists: Result /= Void
		end

	is_unreleasable_agent: FUNCTION [G, TUPLE [G], BOOLEAN]
			-- Agent that tells whether row is not in a state
			-- where batch job can be released
		deferred
		ensure
			result_exists: Result /= Void
		end

	is_passed_agent: FUNCTION [G, TUPLE [G], BOOLEAN]
			-- Agent that tells whether row has job that
			-- is passed
		deferred
		ensure
			result_exists: Result /= Void
		end

	job_arguments (row: G): STRING
			-- Arguments to be passed to batch job for `row'
		require
			row_exists: row /= Void
		deferred
		end

	handle_submit_event (submitted_rows: LIST [G])
			-- Handle job submission event
		require
			submitted_rows_exists: submitted_rows /= Void
		deferred
		end

	handle_release_event (released_rows: LIST [G])
			-- Handle job release event
		require
			released_rows_exists: released_rows /= Void
		deferred
		end

	handle_changed_event (changed_rows: LIST [G])
			-- Handle job status change event
		require
			changed_rows_exists: changed_rows /= Void
		deferred
		end

	handle_all_passed_event
			-- Handle "all rows passed" event
		deferred
		end

	handle_failure_event (failed_rows, aborted_rows: LIST [G])
			-- Handle job failure event
		require
			failed_rows_exists: failed_rows /= Void
			aborted_rows_exists: aborted_rows /= Void
		deferred
		end

	handle_reset_event (reset_rows: LIST [G])
			-- Handle job reset event
		deferred
		end

	row_to_generic (row: EV_MULTI_COLUMN_LIST_ROW): G
			-- Value of `row' which is of type G
		require
			row_exists: row /= Void
		deferred
		ensure
			result_exists: Result /= Void
		end

feature {NONE} -- Implementation

	build_interface
			-- Build interface in `window'
		local
			main_vbox: EV_VERTICAL_BOX
			buttons: HORIZONTAL_BUTTON_SET
			box: EV_BOX
			sep: EV_HORIZONTAL_SEPARATOR
			cell: EV_CELL
		do
			initialize_state_info
			create main_vbox
			add_menu_bar
			fetch_needed_data
			
			initialize_total_count
			initialize_other_counts
			build_selection_status_box
			refresh_status_line_counts
			
			build_all_rows
			recompute_batch_job_info
			rebuild_visible_list
			
			main_vbox.extend (visible_list)
			
			create_job_status_boxes
			refresh_job_status_counts
			
			create sep
			main_vbox.extend (sep)
			main_vbox.disable_item_expand (sep)
			create cell
			cell.extend (selection_status_box)
			main_vbox.extend (cell)
			main_vbox.disable_item_expand (cell)
			create sep
			main_vbox.extend (sep)
			main_vbox.disable_item_expand (sep)
			
			create cell
			cell.extend (visible_job_status_box)
			main_vbox.extend (cell)
			main_vbox.disable_item_expand (cell)
			create sep
			main_vbox.extend (sep)
			main_vbox.disable_item_expand (sep)
			
			create cell
			cell.extend (all_job_status_box)
			main_vbox.extend (cell)
			main_vbox.disable_item_expand (cell)
			create sep
			main_vbox.extend (sep)
			main_vbox.disable_item_expand (sep)
			
			create buttons.make_with_text_and_actions (<< Submit_text, Monitor_text, Stop_text, Reset_text, gui.ev_cancel >>, << agent process_submit (False), agent process_monitor, agent process_stop, agent process_reset, agent process_cancel >>)
			box := buttons.column_set.box
			submit_button := buttons.i_th (1)
			monitor_button := buttons.i_th (2)
			stop_button := buttons.i_th (3)
			reset_button := buttons.i_th (4)
			
			main_vbox.extend (box)
			main_vbox.disable_item_expand (box)
			gui.set_box_background_recursive (main_vbox, background_color)
			extend (main_vbox)

			gui.add_cancel_accelerator (window, agent process_cancel)
			gui.do_once_on_idle (agent post_display_actions)
			enable_active
		end
		

	fetch_batch_jobs_list
			-- Fetch list of batch jobs for this run
			-- from database and put in `batch_jobs'
		local
			rows: LIST [BATCH_JOBS_MROW]
		do
			rows := batch_jobs_for_run
			create batch_jobs.make (rows.count)
			from
				rows.start
			until
				rows.after
			loop
				batch_jobs.put (rows.item, rows.item.batch_job_no)
				rows.forth
			end
		end
			
	initialize_other_counts
			-- Initialize variables that are counters whose
			-- values are displayed
		do
			selected_count := 0
			hidden_count := 0
			visible_count := total_count
		end

	build_selection_status_box
			-- Build box with selection and visibility status info
		local
			set: WIDGET_COLUMN_SET
		do
			create selection_status_box
			create selected_box
			create visible_box
			create hidden_box
			create total_box
			create set.make (4)
			set.append_row (<< selected_box, visible_box, hidden_box, total_box >>)
			set.set_padding (user_interface.padding)
			set.align_left
			set.align_columns_left
			set.disable_expand
			selection_status_box.extend (set.box)
		end

	refresh_status_line_counts
			-- Refresh all counters displayed in status line,
			-- which has visibility and "selected" counters
		do
			refresh_selected_count
			refresh_visible_counts
			update_counter (total_box, Total_text, total_count)
		end

	queue_selected_count_refresh
			-- Queue action to refresh "selected" counter
			-- when next idle
		do
			if not selected_count_refresh_queued then
				gui.do_once_on_idle (agent refresh_selected_count)
				selected_count_refresh_queued := True
			end
		end

	refresh_selected_count
			-- Refresh "selected" counter
		do
			selected_count_refresh_queued := False
			
			-- FIXME: hack to workaround Vision2 bug (agent not
			-- always called when row selected or deselected)
			
			selected_count := current_selected_count
			update_counter (selected_box, Selected_text, selected_count)
		end

	current_selected_count: INTEGER
			-- Current number of selected rows
		local
			the_row: EV_MULTI_COLUMN_LIST_ROW
			cursor: EV_DYNAMIC_LIST_CURSOR [EV_MULTI_COLUMN_LIST_ROW]
		do
			cursor := visible_list.cursor
			from
				visible_list.start
			until
				visible_list.after
			loop
				the_row := visible_list.item
				if the_row.is_selected then
					Result := Result + 1
				end
				visible_list.forth
			end
			visible_list.go_to (cursor)
		end

	refresh_visible_counts
			-- Refresh "visible" and "hidden" counters
		do
			update_counter (visible_box, Visible_text, visible_count)
			update_counter (hidden_box, Hidden_text, hidden_count)
		end

	set_visible_counts
			-- Set the "visible" and "hidden" counters
		do
			visible_count := visible_list.count
			hidden_count := total_count - visible_count
		end

	update_counter (b: EV_BOX a_text: STRING n: INTEGER)
			-- Update counter displayed in `b' with text `a_text'
			-- and value `n'
		require
			box_exists: b /= Void
			a_text_exists: a_text /= Void
			-- FIXME: add back in when Vision2 bug is fixed
			-- (select_actions and deselect_actions are not
			-- always called)
			-- value_nonnegative: n >= 0
		local
			label: EV_LABEL
		do
			create label.make_with_text (a_text + ":  " + n.out)
			b.wipe_out
			b.extend (label)
		end

	recompute_batch_job_info
			-- Recompute information about associated batch jobs
			-- for elements of `all_rows'
		local
			row: G
			job: BATCH_JOBS_MROW
		do
			from
				all_rows.start
			until
				all_rows.after
			loop
				row := all_rows.item
				job := Void
				if not row.is_null_batch_job_no then
					batch_jobs.search (row.requestor_batch_job_no)
					if batch_jobs.found then
						job := batch_jobs.found_item
					end
				end
				row.set_batch_job (job)
				all_rows.forth
			end
		end

	rebuild_visible_list
			-- Rebuild `visible_list' from rows in `all_rows'
			-- that are marked "visible" and from `visible_columns'
		local
			row: G
			k: INTEGER
			t: STRING
			col: MULTI_COLUMN_LIST_COLUMN [G]
			sort_sym: STRING
			old_visible_count, old_selected_count: INTEGER
		do
			save_selection
			old_visible_count := visible_count
			old_selected_count := selected_count
			visible_list.wipe_out
			
			from
				visible_columns.start
				k := 1
			until
				visible_columns.after
			loop
				col := visible_columns.item
				t := col.title
				if col = last_sort_column then
					if sort_count \\ 2 = 1 then
						sort_sym := Sort_ascending
					else
						sort_sym := Sort_descending
					end
					if col.alignment.is_right_aligned then
						t := Sort_sym + Spacer + t
					else
						t := t + Spacer + Sort_sym
					end
				end
				visible_list.set_column_title (t, k)
				k := k + 1
				visible_columns.forth
			end
			
			
			from
				all_rows.start
			until
				all_rows.after
			loop
				row := all_rows.item
				if row.is_visible then
					row.wipe_out
					from
						visible_columns.start
					until
						visible_columns.after
					loop
						row.extend (visible_columns.item.value (row))

						visible_columns.forth
					end
					visible_list.extend (row)
				end
				all_rows.forth
			end
			
			resize_visible_list_to_content
			
			restore_selection
			set_visible_counts
			if selected_count /= old_selected_count then
				refresh_selected_count
			end
			if visible_count /= old_visible_count then
				refresh_visible_counts
			end
		end
			
	refresh_after_selected_rows_operation
			-- Refresh information after performing an
			-- operation that applies only to the selected row
		do
			fetch_batch_jobs_list
			recompute_batch_job_info
			refresh_only_selected_rows
			resize_visible_list_to_content
			refresh_job_status_counts
		end

	refresh_only_selected_rows
			-- Refresh only rows of `visible_list' that are
			-- currently selected
		do
			refresh_visible_list (True)
		end
			
	refresh_all_visible_rows
			-- Refresh all rows of `visible_list'
		do
			refresh_visible_list (False)
		end
			
	refresh_visible_list (selected_only: BOOLEAN)
			-- Refresh all rows of `visible_list' or
			-- only selected rows if `selected_only'
		local
			row: G
		do
			from
				visible_list.start
			until
				visible_list.after
			loop
				row := row_to_generic (visible_list.item)
				if not selected_only or row.is_selected then
					row.wipe_out
					from
						visible_columns.start
					until
						visible_columns.after
					loop
						row.extend (visible_columns.item.value (row))

						visible_columns.forth
					end
				end
				visible_list.forth
			end
		end
			
	resize_visible_list_to_content
			-- Resize every column of `visible_list' to its contents
		local
			k: INTEGER
		do
			from
				visible_columns.start
				k := 1
			until
				visible_columns.after
			loop
				if k > 1 then
					-- First column always has to be left
					-- aligned - no choice
					visible_list.set_column_alignment (visible_columns.item.alignment, k)
				end
				visible_list.resize_column_to_content (k)
				k := k + 1
				visible_columns.forth
			end
		end
			
	create_job_status_boxes
			-- Create boxes to hold job statuses of visible
			-- row and of all rows
		do
			create visible_job_status_box
			create all_job_status_box
		end

	refresh_job_status_counts
			-- Refresh counters that show number of rows with
			-- each job status (passed, failed, etc), for both
			-- visible rows and all rows
		do
			refresh_visible_job_status_counts
			refresh_all_job_status_counts
		end

	refresh_visible_job_status_counts
			-- Refresh counters that show number of rows with
			-- each job status (passed, failed, etc) for
			-- visible rows
		local
			row: G
			s: STRING
		do
			create job_status_counts.make (10)
			from
				visible_list.start
			until
				visible_list.after
			loop
				row := row_to_generic (visible_list.item)
				s := row.job_status
				job_status_counts.force (job_status_counts.item (s) + 1, s)
				visible_list.forth
			end

			visible_job_status_box.wipe_out
			visible_job_status_box.extend (current_job_status_counts_box (Visible_rows_text))
		end

	refresh_all_job_status_counts
			-- Refresh counters that show number of rows with
			-- each job status (passed, failed, etc) for all
			-- rows
		local
			row: G
			s: STRING
		do
			create job_status_counts.make (10)
			from
				all_rows.start
			until
				all_rows.after
			loop
				row := all_rows.item
				s := row.job_status
				job_status_counts.force (job_status_counts.item (s) + 1, s)
				all_rows.forth
			end

			all_job_status_box.wipe_out
			all_job_status_box.extend (current_job_status_counts_box (All_rows_text))
		end

	current_job_status_counts_box (a_text: STRING): EV_HORIZONTAL_BOX
			-- Box with text describing number of rows with
			-- each job status, based on `job_status_counts'.
			-- Description starts with `a_text'
		require
			a_text_exists: a_text /= Void
		local
			set: WIDGET_COLUMN_SET
			label: EV_LABEL
			list: ARRAY [EV_WIDGET]
			s, the_text: STRING
			k, n: INTEGER
			font: EV_FONT
		do
			create list.make (1, job_status_counts.count + 1)
			create label.make_with_text (a_text)
			label.set_minimum_width (100)
			list.put (label, 1)
			from
				create the_text.make (0)
				k := 2
				known_job_statuses.start
			until
				known_job_statuses.after
			loop
				s := known_job_statuses.item
				n := job_status_counts.item (s)
				if n > 0 then
					the_text.keep_head (0)
					su.append_capitalized (the_text, s)
					create label.make_with_text (the_text + ":  " + n.out)
					if not s.is_equal (job_status.Created) and not s.is_equal (job_status.Pending) and not s.is_equal (job_status.Executing) and not s.is_equal (job_status.Passed) and not s.is_equal (job_status.Stopped) then
						label.set_foreground_color (gui.Red)
						font := label.font
						font.set_weight (gui.Weight_bold)
						label.set_font (font)
					end
					list.put (label, k)
					k := k + 1
				end
				known_job_statuses.forth
			end
			
			create set.make (list.count)
			set.append_row (list)
			set.set_padding (user_interface.padding)
			set.align_left
			set.align_columns_left
			set.disable_expand
			Result := set.box
		end

	post_display_actions
		local
			screen: EV_SCREEN
		do
			create screen
			gui.do_once_on_idle (agent gui.center_window_with_margin (window, 100, 100))
		end

	initialize_state_info
			-- Initialize variables that hold state information
		do
			create all_rows.make (100)
			initialize_visible_columns
			initialize_visible_list
			initialize_known_job_statuses
			initialize_monitoring
		end

	initialize_visible_columns
			-- Initialize `visible_columns'
		do
			create {ARRAYED_LIST [MULTI_COLUMN_LIST_COLUMN [G]]} visible_columns.make (10)
			add_visible_columns
			if not visible_columns.is_empty then
				last_sort_column := visible_columns.first
			else
				last_sort_column := Void
			end
			sort_count := 1
		end
			
	initialize_visible_list
		do
			create visible_list
			visible_list.enable_multiple_selection
			visible_list.column_title_click_actions.extend (agent process_column_title_click)
		end

	initialize_known_job_statuses
			-- Initialize list of known job statuses
		do
			create {ARRAYED_LIST [STRING]} known_job_statuses.make (10)
			known_job_statuses.extend (job_status.Old_status)
			known_job_statuses.extend (job_status.Na)
			known_job_statuses.extend (job_status.Not_created)
			known_job_statuses.extend (job_status.Created)
			known_job_statuses.extend (job_status.Pending)
			known_job_statuses.extend (job_status.Executing)
			known_job_statuses.extend (job_status.Stopped)
			known_job_statuses.extend (job_status.Passed)
			known_job_statuses.extend (job_status.Failed)
			known_job_statuses.extend (job_status.Aborted)
		end

	initialize_monitoring
			-- Initialize attributes related to monitoring
		do
			is_monitoring := False
			monitor_interval := Default_monitor_interval
		end

	add_menu_bar
			-- Add menu bar to window
		local
			file_menu, select_menu, view_menu: EV_MENU
			row_menu, col_menu: EV_MENU
			a_menu_bar: EV_MENU_BAR
		do	
			create file_menu.make_with_text ("File")
			create select_menu.make_with_text ("Select")
			create view_menu.make_with_text ("View")
			create row_menu.make_with_text ("Rows")
			create col_menu.make_with_text ("Columns")
			
			create a_menu_bar
			a_menu_bar.extend (file_menu)
			a_menu_bar.extend (select_menu)
			a_menu_bar.extend (view_menu)
			
			add_menu_item (file_menu, "Submit and hold", agent process_submit (True))
			add_menu_item (file_menu, "Release hold", agent process_release)
			add_menu_item (file_menu, "Reanalyze", agent process_reanalyze)
			add_menu_item (file_menu, "Close", agent process_cancel)
			add_menu_item (file_menu, "Exit", gui.destroy_application_agent)
			
			add_menu_item (select_menu, All_text, agent process_select_all)
			add_menu_item (select_menu, None_text, agent process_select_none)
			add_menu_item (select_menu, Invert_text, agent process_select_invert)
			
			add_separator (select_menu)
			
			add_menu_item (select_menu, Submittable_text, agent process_select_submittable)
			add_menu_item (select_menu, Stopable_text, agent process_select_stopable)
			add_menu_item (select_menu, Resettable_text, agent process_select_resettable)
			add_menu_item (select_menu, Unsuccessful_text, agent process_select_unsuccessful)
			
			add_separator (select_menu)
			
			add_menu_item (select_menu, Aborted_text, agent process_select_row_status (job_status.Aborted))
			add_menu_item (select_menu, Created_text, agent process_select_row_status (job_status.Created))
			add_menu_item (select_menu, Executing_text, agent process_select_row_status (job_status.Executing))
			add_menu_item (select_menu, Failed_text, agent process_select_row_status (job_status.Failed))
			add_menu_item (select_menu, Passed_text, agent process_select_row_status (job_status.Passed))
			add_menu_item (select_menu, Pending_text, agent process_select_row_status (job_status.Pending))
			add_menu_item (select_menu, Stopped_text, agent process_select_row_status (job_status.Stopped))
			
			add_extra_select_menu_items (select_menu)
			
			view_menu.extend (row_menu)
			view_menu.extend (col_menu)
			
			add_menu_item (row_menu, All_text, agent process_view_all_rows)
			add_menu_item (row_menu, None_text, agent process_view_no_rows)
			add_menu_item (row_menu, Invert_text, agent process_view_invert_rows)
			add_menu_item (row_menu, Selected_text, agent process_view_selected_rows)
			add_separator (row_menu)
			
			add_menu_item (row_menu, Submittable_text, agent process_view_submittable)
			add_menu_item (row_menu, Stopable_text, agent process_view_stopable)
			add_menu_item (row_menu, Resettable_text, agent process_view_resettable)
			add_menu_item (row_menu, Unsuccessful_text, agent process_view_unsuccessful)
			add_separator (row_menu)
			
			add_menu_item (row_menu, Aborted_text, agent process_view_row_status (job_status.Aborted))
			add_menu_item (row_menu, Created_text, agent process_view_row_status (job_status.Created))
			add_menu_item (row_menu, Executing_text, agent process_view_row_status (job_status.Executing))
			add_menu_item (row_menu, Failed_text, agent process_view_row_status (job_status.Failed))
			add_menu_item (row_menu, Passed_text, agent process_view_row_status (job_status.Passed))
			add_menu_item (row_menu, Pending_text, agent process_view_row_status (job_status.Pending))
			add_menu_item (row_menu, Stopped_text, agent process_view_row_status (job_status.Stopped))
			
			add_extra_view_row_menu_items (row_menu)
			
			set_menu_bar (a_menu_bar)
		end

	add_separator (a_menu: EV_MENU)
		require
			a_menu_exists: a_menu /= Void
		local
			sep: EV_MENU_SEPARATOR
		do
			create sep
			a_menu.extend (sep)
		end
			
	add_menu_item (a_menu: EV_MENU a_text: STRING an_action: PROCEDURE [ANY, TUPLE])
		require
			a_menu_exists: a_menu /= Void
			a_text_exists: a_text /= Void
			an_action_exists: an_action /= Void
		local
			menu_item: EV_MENU_ITEM
		do
			create menu_item.make_with_text_and_action (a_text, an_action)
			a_menu.extend (menu_item)
		end
			
	process_select_one (row: G)
			-- Take actions needed when user selects a row
		do
			selected_count := selected_count + 1
			queue_selected_count_refresh
		end

	process_deselect_one (row: G)
			-- Take actions needed when user deselects a row
		do
			selected_count := selected_count - 1
			queue_selected_count_refresh
		end

	process_select_all
			-- Take actions needed when user chooses "Select All"
		do
			from
				visible_list.start
			until
				visible_list.after
			loop
				visible_list.item.enable_select
				visible_list.forth
			end
		end

	process_select_invert
			-- Take actions needed when user chooses "Select Invert"
		do
			from
				visible_list.start
			until
				visible_list.after
			loop
				if visible_list.item.is_selected then
					visible_list.item.disable_select
				else
					visible_list.item.enable_select
				end
				visible_list.forth
			end
		end

	process_select_none
			-- Take actions needed when user chooses "Select Invert"
		do
			from
				visible_list.start
			until
				visible_list.after
			loop
				visible_list.item.disable_select
				visible_list.forth
			end
		end

	process_select_row_status (s: STRING)
			-- Take actions needed when user chooses to replace
			-- selection with all rows with status `s'
		require
			die_status_exists: s /= Void
		do
			process_select_rows_filtered (has_job_status_agent (s))
		end

	process_select_stopable
			-- Take actions needed when user chooses to replace
			-- selection with all rows whose jobs are stopable
		do
			process_select_rows_filtered (is_stopable_agent)
		end

	process_select_resettable
			-- Take actions needed when user chooses to replace
			-- selection with all rows whose jobs are resettable
		do
			process_select_rows_filtered (is_resettable_agent)
		end

	process_select_unsuccessful
			-- Take actions needed when user chooses to replace
			-- selection with all rows whose jobs did not
			-- succeed
		do
			process_select_rows_filtered (is_unsuccessful_agent)
		end

	process_select_submittable
			-- Take actions needed when user chooses to replace
			-- selection with all rows whose jobs are submittable
		do
			process_select_rows_filtered (is_submittable_agent)
		end

	process_select_rows_filtered (filter: FUNCTION [G, TUPLE [G], BOOLEAN])
			-- Take actions needed when user chooses to replace
			-- selection with all rows for which `filter' is true
		require
			filter_exists: filter /= Void
		local
			row: G
		do
			from
				visible_list.start
			until
				visible_list.after
			loop
				row := row_to_generic (visible_list.item)
				if filter.item ([row]) then
					row.enable_select
				else
					row.disable_select
				end
				visible_list.forth
			end
		end

	process_view_all_rows
			-- Take actions needed when user chooses to view all
			-- rows
		do
			from
				all_rows.start
			until
				all_rows.after
			loop
				all_rows.item.enable_visible
				all_rows.forth
			end
			
			rebuild_visible_list
			refresh_visible_job_status_counts
		end

	process_view_no_rows
			-- Take actions needed when user chooses to view no
			-- rows
		do
			from
				all_rows.start
			until
				all_rows.after
			loop
				all_rows.item.disable_visible
				all_rows.forth
			end
			
			rebuild_visible_list
			refresh_visible_job_status_counts
		end

	process_view_invert_rows
			-- Take actions needed when user chooses to invert
			-- set of visible rows (making all previously
			-- invisible rows visible and vice versa)
		do
			from
				all_rows.start
			until
				all_rows.after
			loop
				if all_rows.item.is_visible then
					all_rows.item.disable_visible
				else
					all_rows.item.enable_visible
				end
				all_rows.forth
			end
			
			rebuild_visible_list
			refresh_visible_job_status_counts
		end

	process_view_selected_rows
			-- Take actions needed when user chooses to view
			-- only selected rows
		do
			from
				all_rows.start
			until
				all_rows.after
			loop
				if all_rows.item.is_selected then
					all_rows.item.enable_visible
				else
					all_rows.item.disable_visible
				end
				all_rows.forth
			end
			
			rebuild_visible_list
			refresh_visible_job_status_counts
		end

	process_view_stopable
			-- Take actions needed when user chooses to view
			-- only rows whose jobs are stopable
		do
			process_view_rows_filtered (is_stopable_agent)
		end

	process_view_resettable
			-- Take actions needed when user chooses to view
			-- only rows whose jobs are resettable
		do
			process_view_rows_filtered (is_resettable_agent)
		end

	process_view_unsuccessful
			-- Take actions needed when user chooses to view
			-- only rows whose jobs did not succeed
		do
			process_view_rows_filtered (is_unsuccessful_agent)
		end

	process_view_submittable
			-- Take actions needed when user chooses to view
			-- only rows whose jobs are submittable
		do
			process_view_rows_filtered (is_submittable_agent)
		end

	process_view_row_status (s: STRING)
			-- Take actions needed when user chooses to view
			-- only rows with status `s'
		require
			die_status_exists: s /= Void
		do
			process_view_rows_filtered (has_job_status_agent (s))
		end

	process_view_rows_filtered (filter: FUNCTION [G, TUPLE [G], BOOLEAN])
			-- Take actions needed when user chooses to view
			-- only rows for which `filter' is true
		require
			filter_exists: filter /= Void
		local
			row: G
		do
			from
				all_rows.start
			until
				all_rows.after
			loop
				row := all_rows.item
				if filter.item ([row]) then
					row.enable_visible
				else
					row.disable_visible
				end
				all_rows.forth
			end
			rebuild_visible_list
			refresh_visible_job_status_counts
		end

	process_submit (hold: BOOLEAN)
			-- Take actions needed when user clicks Submit button.
			-- If `hold' then put any submitted jobs in HOLDING
			-- state, otherwise put them in WAITING state
		do
			if is_active then
				disable_active
				rollback_required := False
				create operation.make (agent process_submit_request (hold))
				operation.enable_abort_caught
				operation.abort_actions.extend (agent handle_submission_abort)
				operation.success_actions.extend (agent refresh_after_selected_rows_operation)
				operation.execute
				refresh_after_selected_rows_operation
				enable_active
			end
		end

	process_submit_request (hold: BOOLEAN)
			-- Take actions needed to process a Submit request.
			-- If `hold' then put any submitted jobs in HOLDING
			-- state, otherwise put them in WAITING state
		local
			row: G
			conn: DB_CONNECTION
			signal: BATCH_JOB_QUEUE_SIGNAL_VALUES
			script_name, log_name: STRING
			file_op: FILE_OP
			f: LARGE_PLAIN_TEXT_FILE
			state: STRING
			batch_job_state: BATCH_JOB_STATE_VALUES
		do
			rebuild_selected_rows
			if selected_rows.is_empty then
				report_no_rows_selected
			elseif selected_rows.there_exists (is_unsubmittable_agent) then
				report_invalid_row ("submittable", << job_status.Created >>, "You must Reset a job that does not have status " + job_status.Created + " to make it submittable.")
			else
				rollback_required := True
				conn := session.connection
				lock_row_list (selected_rows)
				create file_op
				if hold then
					state := batch_job_state.Holding
				else
					state := batch_job_state.Waiting
				end
				from
					selected_rows.start
				until
					selected_rows.after
				loop
					row := selected_rows.item
					script_name := row.job_script_file
					log_name := row.job_log_file
					create f.make (script_name)
					file_op.verify_existence (f)
					file_op.verify_permissions (f, Minimum_script_permissions)
					row.submit_job (session.run_id, state, script_name, log_name, job_arguments (row))
					selected_rows.forth
				end
				conn.queue_signal (signal.Job_queue_changed, Void)
				session.commit
				handle_submit_event (selected_rows)
			end
		end

	handle_submission_abort
			-- Handle abort of Submit operation
		do
			if rollback_required then
				session.rollback
				fetch_requestor_list
				refresh_corrupted_rows
			end
		end

	process_release
			-- Take actions needed when user selects Release
			-- Holding menu item.
		do
			if is_active then
				disable_active
				rollback_required := False
				create operation.make (agent process_release_request)
				operation.enable_abort_caught
				operation.abort_actions.extend (agent handle_release_abort)
				operation.success_actions.extend (agent refresh_after_selected_rows_operation)
				operation.execute
				refresh_after_selected_rows_operation
				enable_active
			end
		end

	process_release_request
			-- Take actions needed to process a Release
			-- Holding request
		local
			row: G
			conn: DB_CONNECTION
			signal: BATCH_JOB_QUEUE_SIGNAL_VALUES
			batch_job_state: BATCH_JOB_STATE_VALUES
		do
			rebuild_selected_rows
			if selected_rows.is_empty then
				report_no_rows_selected
			elseif selected_rows.there_exists (is_unreleasable_agent) then
				report_invalid_job_row ("releasable", << batch_job_state.Holding >>)
			else
				rollback_required := True
				from
					selected_rows.start
				until
					selected_rows.after
				loop
					row := selected_rows.item
					row.release_job
					selected_rows.forth
				end
				conn := session.connection
				conn.queue_signal (signal.Job_queue_changed, Void)
				session.commit
				handle_release_event (selected_rows)
			end
		end

	handle_release_abort
			-- Handle abort of Release Holding operation
		do
			if rollback_required then
				session.rollback
				fetch_requestor_list
				refresh_corrupted_rows
			end
		end

	process_monitor
			-- Take actions needed when user clicks Monitor button
		do
			if is_active then
				disable_active
				if is_monitoring then
					disable_monitoring
				else
					rollback_required := True
					failure_notification_required := True
					create operation.make (agent process_monitor_request)
					operation.enable_abort_caught
					operation.abort_actions.extend (agent handle_monitor_abort)
					operation.execute
				end
				enable_active
			end
		end

	process_monitor_request
			-- Take actions needed to process a Monitor request
		do
			if not all_rows.there_exists (is_stopable_agent) and not all_rows.for_all (is_passed_agent) then
				-- No pending or executing rows and not all
				-- rows have passed
				operation.inform ("Nothing to monitor%N(no rows with status " + joined_by (<< job_status.Pending, job_status.Executing >>, " or ") + ")")
				operation.abort
			else
				process_select_none
					-- Deselect all rows
				enable_monitoring
				create monitor_timeout
				monitor_timeout.actions.extend (agent monitor_once)
				monitor_once
			end
		end

	monitor_once
			-- Check status of jobs and update
			-- database if any started, finished or aborted
		do
			fetch_batch_jobs_list
			recompute_batch_job_info
			recompute_row_statuses
			refresh_all_visible_rows
			resize_visible_list_to_content
			refresh_job_status_counts
			if all_rows.there_exists (is_stopable_agent) then
				reset_monitor_interval
			else
				disable_monitoring
			end
		end

	handle_monitor_abort
			-- Handle abort of Monitor operation
		do
			if rollback_required then
				session.rollback
				disable_monitoring
				fetch_requestor_list
				refresh_corrupted_rows
				fetch_batch_jobs_list
				recompute_batch_job_info
				refresh_all_visible_rows
				resize_visible_list_to_content
				refresh_job_status_counts
			end
		end

	reset_monitor_interval
			-- Reset monitoring interval
		do
			if monitor_timeout /= Void then
				monitor_timeout.set_interval (monitor_interval * 1000)
			end
		end

	enable_monitoring
		do
			is_monitoring := True
			monitor_button.set_text (Unmonitor_text)
			submit_button.disable_sensitive
			stop_button.disable_sensitive
			reset_button.disable_sensitive
		end

	disable_monitoring
		do
			is_monitoring := False
			monitor_button.set_text (Monitor_text)
			submit_button.enable_sensitive
			stop_button.enable_sensitive
			reset_button.enable_sensitive
			if monitor_timeout /= Void then
				monitor_timeout.destroy
				monitor_timeout := Void
			end
		end

	recompute_row_statuses
			-- Recompute status of each row that has an
			-- associated batch job
		local
			row: G
			changed, all_passed: BOOLEAN
			failed_rows: ARRAYED_LIST [G]
			aborted_rows: ARRAYED_LIST [G]
			changed_rows: ARRAYED_LIST [G]
		do
			changed := False
			create failed_rows.make (0)
			create aborted_rows.make (0)
			create changed_rows.make (0)
			from
				all_rows.start
			until
				all_rows.after
			loop
				row := all_rows.item
				if row.has_batch_job then
					row.update_status
					if row.status_changed then
						changed := True
						changed_rows.extend (row)
						if row.is_failed then
							failed_rows.extend (row)
						end
						if row.is_aborted then
							aborted_rows.extend (row)
						end
					end
				end
				all_rows.forth
			end
				
			session.commit
			all_passed := all_rows.for_all (is_passed_agent)
			if changed then
				handle_changed_event (changed_rows)
			end
			
			if all_passed then
				handle_all_passed_event
				success_actions.call ([])
			end
			
			if changed and failure_notification_required and (not failed_rows.is_empty or not aborted_rows.is_empty) then
				failure_notification_required := False
				handle_failure_event (failed_rows, aborted_rows)
			end
		end

	process_stop
			-- Take actions needed when user clicks Stop button
		do
			if is_active then
				disable_active
				rollback_required := False
				create operation.make (agent process_stop_request)
				operation.enable_abort_caught
				operation.abort_actions.extend (agent handle_stop_abort)
				operation.success_actions.extend (agent refresh_after_selected_rows_operation)
				operation.execute
				enable_active
			end
		end

	process_stop_request
			-- Take actions needed to process a Stop request
		do
			rebuild_selected_rows
			if selected_rows.is_empty then
				report_no_rows_selected
			elseif selected_rows.there_exists (is_unstopable_agent) then
				report_invalid_row ("stopable", << job_status.Pending, job_status.Executing >>, "")
			else
				operation.confirm ("Stop jobs for all selected rows")
				rollback_required := True
				try_to_stop_jobs
				if selected_rows.there_exists (is_stopable_agent) then
					operation.inform ("One or more jobs not stopped (still have status " + joined_by (<< job_status.Pending, job_status.Executing >>, " or ") + ")")
					operation.abort
				end
			end
		end

	try_to_stop_jobs
			-- Try to stop the jobs for `selected_rows'
		local
			row: G
			list, new_list: ARRAYED_LIST [G]
			changed_rows: ARRAYED_LIST [G]
			k, limit: INTEGER
			app: EV_APPLICATION
			changed, need_signal: BOOLEAN
			conn: DB_CONNECTION
			signal: BATCH_JOB_QUEUE_SIGNAL_VALUES
		do
			app := (create {EV_ENVIRONMENT}).application
			conn := session.connection
			create list.make (selected_rows.count)
			list.append (selected_rows)
			from
				limit := 25
				k := 1
			until
				list.is_empty or k > limit
			loop
				create changed_rows.make (0)
				lock_row_list (list)
				changed := False
				need_signal := False
				from
					list.start
				until
					list.after
				loop
					row := list.item
					row.try_to_stop_job
					if row.status_changed then
						changed := True
						changed_rows.extend (row)
						if row.has_batch_job and then row.is_job_stop_requested then
							need_signal := True
						end
					end
					list.forth
				end
				if changed then
					if need_signal then
						conn.queue_signal (signal.Job_queue_stop, Void)
					end
					session.commit
					handle_changed_event (changed_rows)
				else
					session.rollback
				end
				
				app.sleep (200)
				fetch_batch_jobs_list
				recompute_batch_job_info

				create new_list.make (list.count)
				list.do_if (agent new_list.extend, is_stopable_agent)
				list := new_list
				
				k := k + 1
			end
		end

	handle_stop_abort
			-- Handle abort of Stop operation
		do
			if rollback_required then
				session.rollback
				fetch_requestor_list
				refresh_corrupted_rows
				refresh_after_selected_rows_operation
			end
		end

	process_reset
			-- Take actions needed when user clicks Reset button
		do
			if is_active then
				disable_active
				rollback_required := False
				create operation.make (agent process_reset_request)
				operation.enable_abort_caught
				operation.abort_actions.extend (agent handle_reset_abort)
				operation.success_actions.extend (agent refresh_after_selected_rows_operation)
				operation.execute
				enable_active
			end
		end

	process_reset_request
			-- Take actions needed to process a Reset request
		local
			row: G
		do
			rebuild_selected_rows
			if selected_rows.is_empty then
				report_no_rows_selected
			elseif selected_rows.there_exists (is_unresettable_agent) then
				report_invalid_row ("resettable", << job_status.Passed, job_status.Failed, job_status.Aborted, job_status.Stopped >>, "")
			else
				operation.confirm ("Resetting a row deletes all pattern files generated for it%Nand any log files as well.  Are you sure you want to do this")
				rollback_required := True
				lock_row_list (selected_rows)
				from
					selected_rows.start
				until
					selected_rows.after
				loop
					row := selected_rows.item
					row.set_job_status_created
					selected_rows.forth
				end
				session.commit_actions.extend (agent delete_generated_files (selected_rows))
				session.commit
				handle_reset_event (selected_rows)
			end
		end

	delete_generated_files (rows: LIST [G])
			-- Delete existing files generated from
			-- fracture of `rows'
		local
			fname: STRING
			row: G
			out_files: LIST [STRING]
		do
			from
				rows.start
			until
				rows.after
			loop
				row := rows.item
				out_files := row.job_output_files
				from
					out_files.start
				until
					out_files.after
				loop
					fname := out_files.item
					session.delete_file_if_exists (fname)
					out_files.forth
				end
				fname := row.job_log_file
				session.delete_file_if_exists (fname)
				if row.has_job_done_file then
					fname := row.job_done_file
					session.delete_file_if_exists (fname)
				end
				rows.forth
			end
		end

	handle_reset_abort
			-- Handle abort of Reset operation
		do
			if rollback_required then
				session.rollback
				fetch_requestor_list
				refresh_corrupted_rows
			end
		end

	process_reanalyze
			-- Take actions needed when user selects Reanalyze
		do
			if is_active then
				disable_active
				rollback_required := True
				create operation.make (agent process_reanalyze_request)
				operation.enable_abort_caught
				operation.abort_actions.extend (agent handle_reanalyze_abort)
				operation.success_actions.extend (agent refresh_after_selected_rows_operation)
				operation.execute
				enable_active
			end
		end

	process_reanalyze_request
			-- Take actions needed to process a Reanalyze request
		local
			row: G
		do
			rebuild_selected_rows
			if selected_rows.is_empty then
				report_no_rows_selected
			elseif not selected_rows.for_all (is_unsuccessful_agent) then
				report_invalid_row ("reanalyzable", << job_status.Failed, job_status.Aborted >>, "")
			else
				rollback_required := True
				lock_row_list (selected_rows)
				from
					selected_rows.start
				until
					selected_rows.after
				loop
					row := selected_rows.item
					row.reanalyze_failure
					selected_rows.forth
				end
				session.commit
			end
		end

	handle_reanalyze_abort
			-- Handle abort of Reanalyze operation
		do
			if rollback_required then
				session.rollback
				fetch_requestor_list
				refresh_corrupted_rows
				refresh_all_visible_rows
				resize_visible_list_to_content
				refresh_job_status_counts
			end
		end
	
	process_cancel
			-- Take actions needed when user clicks cancel
		do
			if is_active then
				disable_active
				disable_monitoring
				abort_actions.call ([])
			end
		end

	process_column_title_click (n: INTEGER)
			-- Take actions needed when user clicks on column title
			-- for column `n'
		local
			comp: KL_COMPARATOR [G]
			col: like last_sort_column
		do
			col := visible_columns.i_th (n)
			comp := col.comparator
			if last_sort_column = col then
				sort_count := sort_count + 1
			else
				sort_count := 1
			end
			sort_all_rows (comp, sort_count \\ 2 = 0)
			last_sort_column := col
			rebuild_visible_list
		end

	sort_all_rows (a_comparator: KL_COMPARATOR [G] reverse: BOOLEAN)
			-- Sort `all_rows' according to `a_comparator'
		local
			sorted_rows: ARRAY [G]
			sorter: DS_ARRAY_BUBBLE_SORTER [G]
			k: INTEGER
		do
			create sorted_rows.make (1, all_rows.count)
			from
				all_rows.start
				k := 1
			until
				all_rows.after
			loop
				sorted_rows.put (all_rows.item, k)
				all_rows.forth
				k := k + 1
			end
			create sorter.make (a_comparator)
			if reverse then
				sorter.reverse_sort (sorted_rows)
			else
				sorter.sort (sorted_rows)
			end
			create all_rows.make (sorted_rows.count)
			from
				k := sorted_rows.lower
			until
				k > sorted_rows.upper
			loop
				all_rows.extend (sorted_rows.item (k))
				k := k + 1
			end
		end

	save_selection
			-- Save `is_selected' status of each row in
			-- `visible_list'
		local
			is_selected: BOOLEAN
			the_row: G
		do
			create selection_status.make (visible_list.count)
			from
				visible_list.start
			until
				visible_list.after
			loop
				the_row := row_to_generic (visible_list.item)
				is_selected := the_row.is_selected
				selection_status.put (is_selected, the_row.unique_identifier)
				visible_list.forth
			end
		end

	restore_selection
			-- Restore `is_selected' status of each row in
			-- `visible_list'
		local
			is_selected: BOOLEAN
			the_row: G
			n: INTEGER
		do
			from
				visible_list.start
			until
				visible_list.after
			loop
				the_row := row_to_generic (visible_list.item)
				is_selected := selection_status.item (the_row.unique_identifier)
				if is_selected then
					the_row.enable_select
					n := n + 1
				end
				visible_list.forth
			end
			selected_count := n
		end

	rebuild_selected_rows
			-- Make list of all rows in `visible_list' that
			-- are currently selected and put in `selected_rows'
		local
			the_row: G
		do
			create {ARRAYED_LIST [G]} selected_rows.make (selected_count.max (0))
				-- FIXME: change when Vision2 bug is fixed
				-- Work around bug where selected_count can
				-- become negative
			from
				visible_list.start
			until
				visible_list.after
			loop
				the_row := row_to_generic (visible_list.item)
				if the_row.is_selected then
					selected_rows.extend (the_row)
				end
				visible_list.forth
			end
		end

	lock_row_list (list: LIST [G])
			-- Lock the requestor row for every row in
			-- `list'
		local
			row: G
			lock_op: ROW_LOCK_OP
		do
			create lock_op
			from
				list.start
			until
				list.after
			loop
				row := list.item
				lock_op.lock_nowait (row.requestor)
				list.forth
			end
		end

	report_no_rows_selected
			-- Tell user no rows were selected
		do
			operation.inform ("You must select one or more rows first")
			operation.abort
		end

	report_invalid_row (what: STRING states: ARRAY [STRING] a_hint: STRING)
			-- Report that one or more selected rows are
			-- not `what' (not in one of `states')
		require
			what_not_void: what /= Void
			states_not_void: states /= Void
			a_hint_not_void: a_hint /= Void
		local
			msg: STRING
		do
			msg := "All selected rows must be " + what + "%N(have status " + joined_by (states, " or ") + ")."
			if not a_hint.is_empty then
				msg.extend ('%N')
				msg.append (a_hint)
			end
			operation.inform (msg)
			operation.abort
		end

	report_invalid_job_row (what: STRING states: ARRAY [STRING])
			-- Report that one or more selected rows are
			-- not `what' (not in one of `states')
		do
			operation.inform ("All selected rows must be " + what + "%N(have job state " + joined_by (states, " or ") + ")")
			operation.abort
		end

	joined_by (arr: ARRAY [STRING] sep: STRING): STRING
			-- `arr' with adjacent elements separated by `sep'
		local
			list: ARRAYED_LIST [STRING]
			first: BOOLEAN
			k: INTEGER
		do
			create list.make (arr.count)
			from
				first := True
				k := arr.lower
			until
				k > arr.upper
			loop
				list.extend (arr.item (k))
				k := k + 1
			end
			Result := su.joined_by (list, sep)
		end

	fetch_fab_instance_list
			-- Fetch list of fab instances (with extra info)
			-- for dies for run from database
			-- and put in `fab_instances'
		local
			fab_row: IC_FAB_INSTANCES_EROW
			rows: LIST [IC_FAB_INSTANCES_EROW]
		do
			create fab_row.make_template
			rows := fab_row.fab_instances_for_run (session.run_id, session.connection, expected_count.At_least_zero)
			create fab_instances.make (rows.count)
			from
				rows.start
			until
				rows.after
			loop
				fab_instances.put (rows.item, rows.item.die_id)
				rows.forth
			end
		end
			
	batch_jobs: HASH_TABLE [BATCH_JOBS_MROW, INTEGER]
			-- Batch jobs for this run, indexed by batch
			-- job number

	fab_instances: HASH_TABLE [IC_FAB_INSTANCES_EROW, STRING]
			-- Fab instances with extra information from
			-- other tables, indexed by die id, for all
			-- dies on run that have fab instances
			-- (customer dies and test vehicles)

	all_rows: ARRAYED_LIST [G]
			-- Rows for all entities on run

	selected_rows: LIST [G]
			-- Rows for entities that are currently selected
			-- (re-built whenever it is needed)

	visible_list: EV_MULTI_COLUMN_LIST
			-- Multi-column list with rows which are
			-- currently visible

	visible_columns: LIST [MULTI_COLUMN_LIST_COLUMN [G]]
			-- List of columns which are currently visible

	last_sort_column: MULTI_COLUMN_LIST_COLUMN [G]
			-- Last column on which sort was done

	sort_count: INTEGER
			-- Number of times same column has been sorted
			-- consecutively

	selection_status: HASH_TABLE [BOOLEAN, STRING]
			-- `is_selected' status of each entity

	selection_status_box: EV_HORIZONTAL_BOX
			-- Box with selection and visibility status info

	selected_box: EV_HORIZONTAL_BOX
			-- Box to display number of selected items

	visible_box: EV_HORIZONTAL_BOX
			-- Box to display number of visible items

	hidden_box: EV_HORIZONTAL_BOX
			-- Box to display number of hidden items

	total_box: EV_HORIZONTAL_BOX
			-- Box to display total number of items

	selected_count: INTEGER
			-- Number of items currently selected

	visible_count: INTEGER
			-- Number of items currently visible

	hidden_count: INTEGER
			-- Number of items currently hidden

	total_count: INTEGER
			-- Total number of items

	visible_job_status_box: EV_HORIZONTAL_BOX
			-- Box with job status of just the visible
			-- rows for run

	all_job_status_box: EV_HORIZONTAL_BOX
			-- Box with status of all rows for run

	job_status_counts: HASH_TABLE [INTEGER, STRING]
			-- Number of rows with each job status

	known_job_statuses: LIST [STRING]
			-- Known job statuses, in the order in which
			-- counts of each status should be displayed

	submit_button: EV_BUTTON
			-- Button to submit jobs

	monitor_button: EV_BUTTON
			-- Button to start/stop monitoring of jobs

	stop_button: EV_BUTTON
			-- Button to stop jobs

	reset_button: EV_BUTTON
			-- Button to reset jobs

	operation: ANONYMOUS_OPERATION
			-- Current operation in progress, if any

	rollback_required: BOOLEAN
			-- Does routine that handles abort of operation
			-- need to do a database rollback?

	failure_notification_required: BOOLEAN
			-- Is a failure notification required for jobs
			-- that fail or abort?

	is_monitoring: BOOLEAN
			-- Are we currently monitoring jobs
			-- at regular intervals and updating database
			-- when jobs are done?

	selected_count_refresh_queued: BOOLEAN
			-- Is a call to `refresh_selected_count' currently
			-- queued for execution when things are idle?

	monitor_interval: INTEGER
			-- Number of seconds to sleep between monitor updates

	monitor_timeout: EV_TIMEOUT
			-- Action to be executed every `monitor_interval'
			-- seconds

	Default_monitor_interval: INTEGER = 10

	Sort_ascending: STRING = "+"

	Sort_descending: STRING = "-"

	Spacer: STRING = " "

	Submit_text: STRING = "Submit"
	Monitor_text: STRING = "Monitor"
	Unmonitor_text: STRING = "Unmonitor"
	Stop_text: STRING = "Stop"
	Reset_text: STRING = "Reset"
	
	All_text: STRING = "All"
	None_text: STRING = "None"
	Invert_text: STRING = "Invert"
	Selected_text: STRING = "Selected"

	Aborted_text: STRING = "Aborted"
	Created_text: STRING = "Created"
	Executing_text: STRING = "Executing"
	Failed_text: STRING = "Failed"
	Passed_text: STRING = "Passed"
	Pending_text: STRING = "Pending"
	Stopped_text: STRING = "Stopped"

	Submittable_text: STRING = "Submittable (Created)"
	Stopable_text: STRING = "Stopable (Pending or Executing)"
	Resettable_text: STRING = "Resettable (Passed, Failed, Aborted or Stopped)"
	Unsuccessful_text: STRING = "Unsuccessful (Failed or Aborted)"

	Visible_text: STRING = "Visible"
	Hidden_text: STRING = "Hidden"
	Total_text: STRING = "Total"

	Visible_rows_text: STRING = "VISIBLE"
	All_rows_text: STRING = "ALL"
	
	job_status: JOB_STATUS_VALUES

	Minimum_script_permissions: STRING = "---r-x---"
end
