
indexing
	
	description: "Job file that drives ebeam machine that makes masks"
	author: "David Hollenberg"
	date: "July 15, 2004"

deferred class JOB_FILE
inherit
	STRING_UTILITIES_ACCESS
	NUMBER_UTILITIES_ACCESS
		export
			{NONE} all
		end
	CONDITION_UTILITIES
		export
			{NONE} all
		end
	ASCII
		rename
			Comma as Comma_code
		export
			{NONE} all
		end

feature -- Properties

	is_final: BOOLEAN
			-- Is job file a final one to be sent to mask shop?
			-- If not, it is only input to CATS for a merge
			-- operation
		deferred
		end

	slice_code: INTEGER
			-- Slice code that indicates substrate type and size

	slice_id: STRING
			-- Identifier for job

	slice_comment: STRING
			-- Comment for slice command

	parameters: SUBSTITUTION_PARAMETER_SET
			-- Set of substitution parameters used to
			-- replace parameters with their values in titles
			-- and canned lines

	mask_levels: LIST [RUN_LEVELS_ROW]
			-- List of mask levels for run

	mask_techn: MASK_TECHNS_ROW
			-- Mask techn for run

	chips: LIST [JOB_FILE_CHIP]
			-- Chips to be written to job file, in order

feature -- Modification

	set_slice_code (n: INTEGER)
			-- Set `slice_code' to `n'
		require
			code_positive: n > 0
		do
			slice_code := n
		ensure
			value_set: slice_code = n
		end

	set_slice_id (s: STRING)
			-- Set `slice_id' to `s'
		require
			value_exists: s /= Void
			value_alphanumeric: su.is_alphanumeric (s) and is_all_legal_chars (s)
		do
			slice_id := s
		ensure
			value_set: slice_id = s
		end

	set_slice_comment (s: STRING)
			-- Set `slice_comment' to `s'
		require
			valid_value: s = Void or else is_all_legal_chars (s)
		do
			slice_id := s
		ensure
			value_set: slice_id = s
		end

	set_parameters (set: like parameters)
			-- Set `parameters' to `set'
		do
			parameters := set
		ensure
			value_set: parameters = set
		end

	set_mask_levels (list: like mask_levels)
			-- Set `mask_levels' to `list'
		do
			mask_levels := list
		ensure
			value_set: mask_levels = list
		end

	set_mask_techn (mt: like mask_techn)
			-- Set `mask_techn' to `mt'
		do
			mask_techn := mt
		ensure
			value_set: mask_techn = mt
		end

	set_chips (list: like chips)
			-- Set `chip' to `list'
		do
			chips := list
		ensure
			value_set: chips = list
		end

feature -- Contract support

	is_all_legal_chars (s: STRING): BOOLEAN
		require
			value_exists: s /= Void
		local
			k, n, count: INTEGER
			c: CHARACTER
		do
			Result := True
			count := s.count
			from
				k := 1
			until
				k > count or not Result
			loop
				c := s.item (k)
				n := c.code
				if n /= Tabulation and (n < First_printable or
				   n > Last_printable or c.upper /= c) then
					Result := False
				end
				k := k + 1
			end
		end


feature -- Operations

	write (f: LARGE_PLAIN_TEXT_FILE)
		require
			file_not_void: f /= Void
			file_open_write: f.is_open_write
		do
			file := f
			create line.make (Max_line_length)
			write_section_1
			write_section_2
			write_section_3
			write_section_4
			write_section_5
			write_section_6
		end

feature {NONE} -- Implementation

	write_section_1
			-- Write section 1 of job file
		do
			append (Slice_cmd)
			indent
			extend (Space)
			append (slice_id)
			extend (Comma)
			append_integer (slice_code)
			if not su.is_empty (slice_comment) then
				append (Comma_space)
				append (slice_comment)
			end
			flush
		end

	write_section_2
			-- Write section 2 of job file
		do
			write_section_comment (2)
		end

	write_section_3
			-- Write section 3 of job file
		do
			write_section_comment (3)
			append (Reticle_cmd)
			flush
			write_titles_and_orientations
			write_global_options
			write_all_level_options
		end

	write_titles_and_orientations
			-- Write mask titles and title orientations
		deferred
		end

	write_global_options
			-- Write global options
		do
			append (Option_cmd)
			indent
			append (all_levels (Overlap_option))
			append_extra_global_options
			flush
		end

	append_extra_global_options
			-- Append extra global options needed in job file
			-- to `line'
		deferred
		end

	write_all_level_options
			-- Write one line per level for levels with
			-- level-specific (nonglobal) options
		deferred
		end

	write_section_4
			-- Write section 4 of job file
		do
			write_section_comment (4)
		end

	write_section_5
			-- Write section 5 of job file
		do
			write_section_comment (5)
			write_prolog_lines
			write_chips
			write_epilog_lines
		end

	write_prolog_lines
			-- Write canned lines in prolog (before MOSIS chips)
		deferred
		end

	write_epilog_lines
			-- Write canned lines in epilog (after MOSIS chips)
		deferred
		end

	write_chips
			-- Write chips and their rows
		do
			if chips /= Void then
				compute_spot_size_decimals
				chips.do_all (agent write_chip)
			end
		end

	write_chip (chip: JOB_FILE_CHIP)
			-- Write one chip and its rows
		do
			if chip.comment /= Void then
				append_comment
				flush
				append_comment
				append (chip.comment)
				flush
				append_comment
				flush
			end
			append (Chip_cmd)
			indent
			append_integer (chip.number)
			extend (Comma)
			flush
			chip.levels.do_all (agent write_chip_level)
			write_chip_rows (chip)
		end

	compute_spot_size_decimals
			-- Compute value for `spot_size_decimals', based
			-- on actual spot sizes in all chip levels
		local
			size: DOUBLE
			s: STRING
			levels: LIST [JOB_FILE_CHIP_LEVEL]
			n: INTEGER
		do
			spot_size_decimals := 1
			from
				chips.start
			until
				chips.after
			loop
				levels := chips.item.levels
				from
					levels.start
				until
					levels.after
				loop
					size := levels.item.spot_size
					s := nu.std_formatted_number (size)
					n := s.count - s.index_of ('.', 1)
					spot_size_decimals := spot_size_decimals.max (n)
					levels.forth
				end
				chips.forth
			end
		end

	write_chip_level (chip_level: JOB_FILE_CHIP_LEVEL)
			-- Write one chip level
		local
			level_no: STRING
		do
			level_no := chip_level.level_number.out
			su.left_pad (level_no, 2)
			extend (Continuation)
			indent
			append ("   (")
			append (level_no)
			if chip_level.reverse_tone then
				extend (Reverse)
			else
				extend (Space)
			end
			append (Comma_space)
			append (chip_level.formatted_file_name)
			append (Comma_space)
			append (Address_option)
			extend ('=')
			append (nu.fixed_formatted_number (chip_level.spot_size, spot_size_decimals))
			extend (')')
			flush
		end

	write_chip_rows (chip: JOB_FILE_CHIP)
			-- Write ROWS commands for the locations of `chip'
		local
			locations: LIST [CONST_POINT]
			location: CONST_POINT
		do
			locations := chip.locations
			from
				locations.start
			until
				locations.after
			loop
				location := locations.item
				append (Rows_cmd)
				indent
				append (nu.fixed_formatted_number (location.y, location_decimals))
				append (Xy_separator)
				append (nu.fixed_formatted_number (location.x, location_decimals))
				flush
				locations.forth
			end
		end

	write_section_6
			-- Write section 6 of job file
		do
			write_section_comment (6)
			append (End_cmd)
			flush
		end

	all_levels (s: STRING): STRING
			-- Append `s' to `line'
		require
			string_exists: s /= Void
		do
			Result := s + All_levels_id
		end

	append (s: STRING)
			-- Append `s' to `line'
		require
			string_exists: s /= Void
		do
			line.append (s)
		end

	append_and_flush (s: STRING)
			-- Append `s' to `line' and then flush it to file
		require
			string_exists: s /= Void
		do
			append (s)
			flush
		end

	append_integer (n: INTEGER)
			-- Append `n' to `line'
		do
			line.append_integer (n)
		end

	append_comment
			-- Append comment start to `line'
		do
			append (Comment)
		end

	write_section_comment (n: INTEGER)
			-- Write comment for section `n'
		do
			append_comment
			flush
			append_comment
			append ("SECTION")
			extend (Space)
			append_integer (n)
			flush
			append_comment
			flush
		end

	extend (c: CHARACTER)
			-- Extend `line' with `c'
		do
			line.extend (c)
		end

	indent
			-- Append enough spaces to `line' to pad to a
			-- fixed column so things line up nicely
		local
			k, count: INTEGER
		do
			count := Min_command_length - line.count
			if count <= 0 then
				count := 1
			end
			from
				k := 1
			until
				k > count
			loop
				extend (Space)
				k := k + 1
			end
		end

	flush
			-- Append a newline character to `line', then
			-- write it to `file' and then clear it
		do
			check_validity
			extend (Newline)
			file.put_string (line)
			line.keep_head (0)
		end

	check_validity
			-- Check validity of `line' and signal error if
			-- not valid (too long or has invalid characters)
		do
			if line.count > Max_line_length then
				handle_hard_error ("Job file line length (" + line.count.out + ") exceeds maximum length of " + Max_line_length.out + "%N" + "Line is: " + line)
			end
			if not is_all_legal_chars (line) then
				handle_hard_error ("Job file line contains illegal character%N" + "Line is: " + line)
			end
		end

	file: LARGE_PLAIN_TEXT_FILE
			-- File to which job deck should be written

	line: STRING
			-- File to which job deck should be written

	spot_size_decimals: INTEGER
			-- Computed number of digits to show after decimal
			-- point for all spot sizes

	location_decimals: INTEGER = 3
			-- Number of digits to show after decimal
			-- point for all mask locations

	Min_command_length: INTEGER = 8
			-- Minimum length to which command words will be
			-- padded with spaces

	Max_line_length: INTEGER = 80
			-- Maximum length of line in job file

	Max_title_length: INTEGER = 30
			-- Maximum length of title in job file

	Slice_cmd: STRING = "SLICE"

	Reticle_cmd: STRING = "RETICLE"

	Orient_cmd: STRING = "ORIENT"

	Option_cmd: STRING = "OPTION"

	Chip_cmd: STRING = "CHIP"

	Rows_cmd: STRING = "ROWS"

	End_cmd: STRING = "END"

	Address_option: STRING = "AD"

	Overlap_option: STRING = "O"

	Mirror_option: STRING = "M"

	Title_mirror_option: STRING = "T"

	Positive_resist_option: STRING = "P"

	All_levels_id: STRING = "A"

	Comment: STRING = "* "

	Continuation: CHARACTER = '$'

	Xy_separator: STRING = " / "

	Coord_separator: CHARACTER = ';'

	Reverse: CHARACTER = 'R'

	Space: CHARACTER = ' '
	
	Comma: CHARACTER = ','

	Comma_space: STRING = ", "

	Newline: CHARACTER = '%N'

end
