class LAYER_SORTED_COLLECTION
		-- union of arbitrary layout items sorted by layers 
		-- (especially useful for the layout readers)

inherit
	LAYOUT

	CACHED_BOUNDING_BOX

create
	make

feature -- the deferred LAYOUT features

	out: STRING is
	do  Result := su.string_of_2("Layer Collection of ", layers.count.out)
	end;

	is_unlayered: BOOLEAN is false;

	traverse(c: LAYOUT_CONSUMER; m: MATRIX) is
		-- "walk" this layout
	local i: INTEGER; b: LAYOUT_COLLECTION
	do  b := bodies.first		-- the one with Void layer
	    b.traverse(c, m);
	    from  i := 2		-- do not use indices here
	    until i > layers.count or c.terminate_traversal
	    loop  b := bodies.i_th(i);
		  if not b.is_empty	-- dont bother with empty collections
		   then c.do_layer(layers.i_th(i), b, m)
		  end
		  i := i + 1
	    end;
	end;

	calculate_bb: CONST_RECTANGLE
		-- calculate my bounding box
	local  foo: CONST_RECTANGLE; mbb: RECTANGLE
	do  from  bodies.start
	    until bodies.after
	    loop  foo := bodies.item.bounding_box;
		  if foo = Void
		   then		-- we have nothing to enclose
		  elseif mbb = Void
		   then create mbb.make_copy(foo)
		   else mbb.extend_to_include(foo)
		  end;
		  bodies.forth
	    end;
	    Result := mbb	-- (even if Void)
	end;

	change_layers(remapping: HASH_TABLE[STRING, STRING];
			notifier: PROCEDURE[ANY, TUPLE[LAYOUT, STRING]])
		-- remapping defines the layers to be remapped (the keys)
		--   and the new layer for each (the values)
		-- notifier is called once for each item that is remapped,
		--   with the second argument being the NEW layer name
	local i: INTEGER; nlyr: STRING
	do  from  i := 2	-- skip the void_body
	    until i > layers.count
	    loop  remapping.search(layers.i_th(i));
		  if remapping.found then
			nlyr := remapping.found_item
			layers.put_i_th(nlyr, i);
			if notifier /= Void then
				notifier.call([bodies.i_th(i), nlyr])
			end;
		  end;
		  i := i + 1
	    end
	end;

feature -- add to the collection

	set_current_layer(lnam: STRING) is
		-- lnam will be applied to all future unlayered additions
		-- until another set_current_layer
	require lnam /= Void
	local i: INTEGER;
	do  if not equal(lnam, current_layer)
	     then -- switch layers
		  i := index_of_layer(lnam)
		  if i = 0
		   then add_another_layer(lnam)
		   else current_layer := lnam;
			curr_body := bodies.i_th(i)
		  end
	    end
	end;

	extend(l: LAYOUT) is
		-- add l (on the current layer if unlayered)
	require exists: l /= Void;
		have_layer: l.is_unlayered implies (current_layer /= Void)
	do  if l.is_unlayered
	     then curr_body.extend(l)
	     else void_body.extend(l)
	    end;
	    cached_bb := Void
	end;

feature -- the current layer, should anyone care

	current_layer: STRING

feature {NONE} -- rest of the details

	curr_body: LAYOUT_COLLECTION
		-- collection for the unlayered items
		-- (which will all end up on the current_layer)

	void_body: LAYOUT_COLLECTION
		-- collection for the layered items

	layers: R_ARRAYED_LIST[STRING]

	bodies: R_ARRAYED_LIST[LAYOUT_COLLECTION]
		-- parallel lists
		-- the first entry has a Void layer and the void_body
		-- all other entries have a layer and an unlayered body

	index_of_layer(l: STRING): INTEGER is
		-- 0 if not found
	require l /= Void
	local i: INTEGER
	do  from  i := 2		-- dont bother with first one
	    until i > layers.count or Result > 0
	    loop  if equal(l, layers.i_th(i))
		   then Result := i
		   else i := i + 1
		  end
	    end
	end;

	add_another_layer(l: STRING) is
	require l /= Void
	do  create curr_body.make_unlayered(1000)
	    current_layer := l
	    layers.extend(l);
	    bodies.extend(curr_body)
	end;

feature	-- trimming

	trim(thold: INTEGER)
		-- trim self and each held collection for which
		--  capacity - count > thold
	require valid_thold: thold >= 0
	do  from  bodies.start
	    until bodies.after
	    loop  bodies.item.trim(thold)
		  bodies.forth
	    end;
	    bodies.trim(thold)
	    layers.trim(thold)
	end;

feature {NONE} -- creation

	make(n: INTEGER) is
		-- n is (initial) capacity (in number of layers)
	require n > 0
	do  create layers.make(n);
	    create bodies.make(n);
	    create void_body.make_layered(1000);
	    layers.extend(Void);
	    bodies.extend(void_body)
	end;

invariant
	layers.count = bodies.count;
	layers.first = Void;
	bodies.first = void_body;
	
end -- LAYER_SORTED_COLLECTION
