deferred class LAYOUT_IO_SUPPORT
		-- the common things that a layout reader or writer
		--  must have (or deal with) regardless of the format

inherit
	STRING_UTILITIES_ACCESS		-- su.

	LAYOUT_IO_OPTIONS_ACCESS	-- lioopt.

feature -- a design technology (for translating layers)

	the_techn: LAYER_SET [LAYER]

	set_the_techn(t: like the_techn) is
	require exists: t /= Void
		compatible: are_compatible(t, the_lib);
		has_layers: not t.layer_set.is_empty
	do  the_techn := t;
	    last_layer_found := t.layer_set.first	-- random
	    set_other
	end;

feature -- a library (for holding the structures in the file)

	the_lib: LAYOUT_LIBRARY

	set_the_lib(l: LAYOUT_LIBRARY) is
	require exists: l /= Void;
		compatible: are_compatible(the_techn, l)
	do  the_lib := l; 
	    set_other
	end;

	are_compatible(t: like the_techn; l: LAYOUT_LIBRARY): BOOLEAN is
	do  Result := (t = Void or l = Void) or else
			(l.technology = Void or equal(l.technology, t))
	end;

feature {NONE} 

	set_other is
		-- fill in the other one if Void
	require are_compatible(the_techn, the_lib)
	do  if the_lib /= Void
	     then if the_techn = Void
		   then set_the_techn(the_lib.technology)
		  elseif the_lib.technology = Void
		   then the_lib.set_technology(the_techn)
		  end
	    end
	ensure are_compatible(the_techn, the_lib)
	end;

feature	-- possible common attributes

	curr_structure: LIBRARY_STRUCTURE
	curr_body: LAYER_SORTED_COLLECTION
		-- what we are now working on

feature	-- the result(s) of reading a file

	the_messages: RESPONSE_COLLECTOR
		-- created by the reader or writer; contains feedback

	reader_application: STRING is "LAYOUT_READER"
	writer_application: STRING is "LAYOUT_WRITER"

feature -- errors (message_ids) common to all readers and writers

	key_is_pos:      STRING is "POS";
	key_is_stru:     STRING is "STRU";
	key_is_val:      STRING is "VAL";

	error_is_eof:	 STRING is "FILEOF";
	error_is_unklyr: STRING is "UNKLYR";

	error_is_dupnam: STRING is "DUPNAM";
	error_is_fewpts: STRING is "FEWPTS";
	error_is_nopnts: STRING is "NOPNTS";
	error_is_nowidt: STRING is "NOWIDT";
	error_is_notlyr: STRING is "NOTLYR";
	error_is_notrns: STRING is "NOTRNS";
	error_is_unkcmd: STRING is "UNKCMD";
	error_is_unktrn: STRING is "UNKTRN";
	error_is_wire00: STRING is "WIRE00";
	error_is_xxnoyy: STRING is "XXNOYY";

	error_is_illmat: STRING is "ILLMAT";

feature -- standard error insertion

	have_io_error: BOOLEAN is
	do  Result := the_messages.is_in_error
	end;

	log_io_error(mid, val: STRING) is
		-- add message mid to the_messages
	local foo: STRING
	do  foo := file_position;
	    if foo /= Void
	     then the_messages.update_rvalue(key_is_pos, foo)
	    end;
	    foo := stru_position;
	    if foo /= Void
	     then the_messages.update_rvalue(key_is_stru, foo)
	    end;
	    if val /= Void
	     then the_messages.update_rvalue(key_is_val, val)
	    end;
	    the_messages.add_response_item(mid, false);
	end;

	file_position: STRING is
		-- where are we in the file?
	deferred
	end;

	stru_position: STRING is
		-- which structure/symbol are we processing?
	deferred
	end;

feature	-- standard layer lookup and error reporting:
	--
	-- these three lookup routines, layer_by_name, layer_by_cif, 
	--  and layer_by_gds ALWAYS return a layer.
	-- they first try last_layer_found,
	-- then they try the_techn.layer_by_xxx,
	-- then they try unknown_layers.layer_by_xxx (for a repeated unknown),
	-- finally, they create a new unknown, add it and return it,
	--  AND generate an io_error

	layer_by_name(lnam: STRING): LAYER is
	do  if equal(lnam, last_layer_found.layer_name)
	     then Result := last_layer_found
	     else Result := the_techn.layer_by_name(lnam)
	    end;
	    if Result = Void and unknown_layers /= Void
	     then Result := unknown_layers.layer_by_name(lnam)
	    end;
	    if Result = Void
	     then add_and_log_new_unknown(lnam, Void, Void);
		  Result := unknown_layers.layer_by_name(lnam)
	    end;
	    last_layer_found := Result
	ensure equal(lnam, Result.layer_name)
	end;

	layer_by_cif(cnam: STRING): LAYER is
	do  if last_layer_found.is_my_cif_name(cnam)
	     then Result := last_layer_found
	     else Result := the_techn.layer_by_cif(cnam)
	    end;
	    if Result = Void and unknown_layers /= Void
	     then Result := unknown_layers.layer_by_cif(cnam)
	    end;
	    if Result = Void
	     then add_and_log_new_unknown(su.string_of_2("CIF_IS_", cnam),
					cnam, Void);
		  Result := unknown_layers.layer_by_cif(cnam)
	    end;
	    last_layer_found := Result
	ensure Result.is_my_cif_name(cnam)
	end;

	layer_by_gds(glyr, gtyp: INTEGER): LAYER is
	local gpr: GDS_PAIR
	do  if last_layer_found.is_my_gds_name(glyr, gtyp)
	     then Result := last_layer_found
	     else Result := the_techn.layer_by_gds(glyr, gtyp)
	    end;
	    if Result = Void and unknown_layers /= Void
	     then Result := unknown_layers.layer_by_gds(glyr, gtyp)
	    end;
	    if Result = Void
	     then create gpr.make(glyr, gtyp);
		  add_and_log_new_unknown(su.string_of_2("GDS_IS_",
							 gpr.out),
					  Void, gpr);
		  Result := unknown_layers.layer_by_gds(glyr, gtyp)
	    end;
	    last_layer_found := Result
	ensure Result.is_my_gds_name(glyr, gtyp)
	end;

	last_layer_found: LAYER
		-- single item cache

feature	-- the unknown layers (just another layer_list)

	unknown_layers: SIMPLE_LAYER_SET
		-- build list of unknowns here to avoid error overload
		-- (we generate one error for each distinct unknown layer)

	set_unknown_layers(unk: SIMPLE_LAYER_SET) is
		-- if you want to start with a set of known unknowns,
		--  be my guest
	do  unknown_layers := unk
	end;

feature {NONE} -- unknown layer helpers

	add_and_log_new_unknown(lnam, cnam: STRING; gnam: GDS_PAIR) is
		-- cnam and/or gnam may be Void; invent if so
	local cnm: STRING; gnm: GDS_PAIR
	do  log_io_error(error_is_unklyr, lnam);
	    if  unknown_layers = Void
	     then create unknown_layers.make("????")
	    end;
	    if cnam = Void
	     then cnm := new_cif_name(lnam)
	     else cnm := cnam
	    end;
	    if gnam = Void
	     then gnm := new_gds_name
	     else gnm := gnam
	    end;
	    unknown_layers.build_and_add_layer(lnam, cnm, gnm);
	end;

	new_cif_name(hint: STRING): STRING is
		-- (ignoring the hint now)
	local i: INTEGER; c: CHARACTER
	do  from  Result := "ZZZZ"
	    until the_techn.layer_by_cif(Result) = Void and
		  unknown_layers.layer_by_cif(Result) = Void
	    loop  -- generate next name in reverse alphabetic walk
		  from  i := 4
		  until i < 4 and c /= 'Z'
		  loop  c := Result @ i;
			if c = 'A'
			 then c := 'Z'		-- loop and "carry"
			 else c := (c.code - 1).to_character_8
			end;
			Result.put(c, i)
			i := i - 1
		  end
	    end;
	end;

	new_gds_name: GDS_PAIR is
	local glyr, gtyp: INTEGER
	do  from  glyr := 255; gtyp := 169;	-- 13 x 13
	    until the_techn.layer_by_gds(glyr, gtyp) = Void and
		  unknown_layers.layer_by_gds(glyr, gtyp) = Void
	    loop  glyr := glyr -1
	    end;
	create Result.make(glyr, gtyp)
	end;

feature	-- reset (between ops)

	reset is
	do  curr_body := Void;
	    curr_structure := Void;
	    the_messages := Void;
	    unknown_layers := Void;
	end;

end -- LAYOUT_IO_SUPPORT
