class LAYOUT_LIBRARY

inherit
	BASIC_LIBRARY_SUPPORT

	GEOMETRY_SUPPORT

	STRING_UTILITIES_ACCESS

create
	make

feature -- library characteristics

	name: STRING
		-- name of this library

	unit_size: DOUBLE
		-- value of one database unit, in meters
	user_units: DOUBLE
		-- value of one database unit in user units
		-- (not used; here for gds compatibility)

	date_accessed, 
	date_modified: DB_DATE
		-- dates of last access and last modification

	root_structure: STRING
		-- name of the declared root structure, if defined/known

	technology: LAYER_SET [LAYER]
		-- "design technology" of the structures in this library

feature	-- small potatoes

	units_per_micron: DOUBLE is
		-- number of database units per micron
	do  Result := rationalize(1 / (unit_size * 10^6))
	end;

	generate_new_structure_name(sugg: STRING): STRING is
		-- return a valid structure name that is not already
		--  a structure in this library.
		-- sugg is a suggested prefix to which a numeric suffix
		--  will be appended
	local foo: STRING
	do  if su.is_empty(sugg) then
		foo := "ZZYZX"
	    else
		foo := sugg
	    end;
	    -- now add on the magic number and keep trying
	    from  
	    until Result /= Void
	    loop  magic_number := magic_number + 1;
		  Result := foo + magic_number.out;
		  if has_structure(Result) then
			Result := Void
		  end
	    end;
	ensure	--	valid_structure_name(Result);
		not has_structure(Result)
	end;

feature {NONE} -- magic number for generate_new_structure_name

	magic_number: INTEGER

feature -- structure access

	count: INTEGER is
		-- number of structures in the library
	do  Result := structures.count
	ensure Result >= 0
	end;

	has_structure(nam: STRING): BOOLEAN is
		-- do we have a structure named nam?
  --	require valid_structure_name(nam)
	do  Result := structures.has(nam)
	end;

	get_structure(nam: STRING): LIBRARY_STRUCTURE is
		-- get the structure named nam
	require has_structure(nam)
	do  Result := structures.item(nam)
	    date_accessed := now
	ensure Result /= Void; equal(Result.name, nam)
	end;

	all_the_structures: ARRAYED_LIST[STRING] is
		-- (new) list of all the structure names in the library
	local  pos: CURSOR;
	do  pos := structures.cursor
	    create Result.make(structures.count);
	    from  structures.start
	    until structures.after
	    loop  Result.extend(structures.key_for_iteration)
		  structures.forth
	    end;
	    structures.go_to(pos)
	ensure Result.count = count
	end;

	root_layout: LAYOUT is
		-- return the layout of the root structure (may be Void)
	do  if root_structure /= Void
	     then Result := get_structure(root_structure).body
	    end
	end;

	correct_structure_name(nam: STRING): STRING is
		-- if nam is present, return it
		-- if there is a structure of "almost" this name, return that
		-- otherwise, return Void
	do  if has_structure(nam)
	     then Result := nam
	     else Result := nam.twin
		  Result.to_upper;	-- this is our "almost"
		  if not has_structure(Result)
		   then Result := Void
		  end
	    end
	ensure Result = Void or else has_structure(Result)
	end;

feature -- library traversal

	start is
		-- initialize a walk of all the library structures
	do  structures.start
	end;

	forth is
	require not off
	do  structures.forth
	end;

	off: BOOLEAN is
		-- have we walked off the end of the library
	do  Result := structures.off
	end;

	item: LIBRARY_STRUCTURE is
	require not off
	do  Result := structures.item_for_iteration
	    date_accessed := now
	end;

feature	-- modification of library

	set_root_structure(nam: STRING) is
	require nam = Void or else has_structure(nam);
			-- must be in the library
	do  root_structure := nam
	end;

	set_technology(techn: like technology) is
	do  technology := techn
	end;

	set_date_accessed(d: DB_DATE) is
		-- change date_accessed for library
	do  date_accessed := d
	end;

	set_date_modified(d: DB_DATE) is
		-- change date_modified for library
	do  date_modified := d
	end;

	set_name(nam: STRING) is
		-- change library name to name
	require su.not_empty(nam)
	do  name := nam
	end;

	resize(unit: DOUBLE) is
		-- unit is in meters/dbu
		-- change the library unit size to unit
		-- NOTE: if library is not empty, existing structures are
		--  implicitly rescaled!!
	require valid_unit: unit >= 0; 
	do  unit_size := unit
	    if user_units = 0 and unit > 0 then
		set_user_units(unit * 10^6)	-- uu = 1.0 microns
	    end;
	end;

	set_user_units(uu: DOUBLE) is
	require valid_unit: uu > 0;
	do  user_units := uu
	end;

	remove_structure(nam: STRING) is
	require has_structure(nam)
	do  structures.remove(nam);
	    date_modified := now
	ensure count = old count - 1
	end;

	add_structure(s: LIBRARY_STRUCTURE) is
	require not has_structure(s.name); 
	--	valid_structure_name(s.name);
	do  structures.put(s, s.name);
	    if date_modified = Void or else s.date_modified > date_modified
	     then date_modified := s.date_modified
	    end;
	    if is_caching then cache_this_structure(s) end;
	ensure count = old count + 1
	end;

	add_structures(l: LAYOUT_LIBRARY) is
		-- add all the structures of l to our library,
	require exists: l /= Void;
		same_size: unit_size = l.unit_size;
		same_techn: technology = Void or 
			equal(technology, l.technology);
		-- none of the structures in l have name conflicts with me
	do  if technology = Void		-- adopt his
	     then set_technology(l.technology)
	    end;
	    from  l.start
	    until l.off
	    loop  add_structure(l.item)
		  l.forth
	    end
	ensure count = old count + l.count
	end;

feature -- instantiation of references from library structures
	-- ext (if not Void) is an external library used to satisfy any
	--  references not satisfied in this library; ext itself is not
	--  instantiated or otherwise modified

	instantiate_all(ext: LAYOUT_LIBRARY) is
	do  instantiate_from_list(all_the_structures, ext)
	end;

	instantiate_from_root(ext: LAYOUT_LIBRARY) is
	require has_structure(root_structure);
	local  foo: ARRAYED_LIST[STRING]
	do  create foo.make(count);
	    foo.extend(root_structure)
	    instantiate_from_list(foo, ext);
	end;

	instantiate_from_list(names: LIST[STRING]; ext: LAYOUT_LIBRARY) is
		-- 
	local  pos: INTEGER; mia: ARRAYED_LIST[STRING]; tmp: LIST[STRING];
		stru: LIBRARY_STRUCTURE
	do  create mia.make(0);  mia.compare_objects;
	    from  pos := 1
	    until pos > names.count	-- (list may grow as we work)
	    loop  stru := get_structure(names.i_th(pos));
		  -- try satisfying stru in this library
		  instantiate_one_structure(stru);
		  -- for each instantiated structure, add it to names
		  --  so that we also do it
		  if names.count < count
		   then -- but dont bother if we have everyone already
		      make_bunch_of_notes(names, instantiated_references);
		  end;
		  tmp := uninstantiated_references
		  if (ext /= Void) and (not tmp.is_empty)
		   then -- still some gaps; fill from other library
			ext.instantiate_one_structure(stru);
			tmp := ext.uninstantiated_references
		  end;
		  -- tmp now has the truly uninstantiated items
		  make_bunch_of_notes(mia, tmp);
		  -- onward
		  pos := pos + 1
	    end;
	    uninstantiated_references := mia
	    instantiated_references := names
	end;

	instantiate_one_structure(stru: LIBRARY_STRUCTURE) is
		-- fill in all uninstantiated references in stru that
		--  can be found in this library
		-- leaves behind instantiated_references (a list of all
		--  the elements in this library that were successfully
		--  instantiated into stru) and uninstantiated_references
		--  (a list of all remaining uninstantiated references in
		--   stru)
		-- NOTE: stru does not have to be a structure in this
		--   library
	require stru /= Void; stru.body /= Void
	local foo: LC_INSTANTIATE_REFERENCES
	do  create foo.make(Current);
	    stru.body.traverse(foo, identity)
	    instantiated_references := foo.instantiated_references
	    uninstantiated_references := foo.missing_references
	end;

	instantiated_references: LIST[STRING]
	uninstantiated_references: LIST[STRING]
		-- from last instantiate_xxxx routine above

feature	-- instantiated layers

	instantiated_layers: ARRAYED_LIST[STRING]
		-- gets all of the layers that are in all of the
		-- structures in instantiated_references
	require have_list: instantiated_references /= Void
	local stru: LIBRARY_STRUCTURE; stru_layers: LIST[STRING]
	do  create Result.make(50);
	    Result.compare_objects
	    from  instantiated_references.start
	    until instantiated_references.after
	    loop  stru := get_structure(instantiated_references.item)
		  stru_layers := stru.local_layers_present
		  make_bunch_of_notes(Result, stru_layers)
		  instantiated_references.forth
	    end;
	end;

feature -- structure name changes

	modify_structure_names(nmod: OBJECT_MODIFIER[STRING]) is
		-- change all structure names, and all references to those
		--  structures, according to nmod
	require nmod /= Void
	local	xlat: HASH_TABLE[STRING, STRING];
		nstru: HASH_TABLE[LIBRARY_STRUCTURE, STRING];
		stru: LIBRARY_STRUCTURE; nam, nnam: STRING;
		refs: LC_RENAME_REFERENCES
	do  create xlat.make(2*count);		-- for refs
	    create nstru.make(2*count);		-- to replace my body
	    -- change (internal & external) names of all structures
	    -- build xlat using nmod (key = old; value = new)
	    from  start
	    until off
	    loop  stru := item;  nam := stru.name;
		  nnam := nmod.modify(nam);
		  xlat.extend(nnam, nam)
		  stru.rename_me(nnam);		-- inside the structure
		  nstru.extend(stru, nnam);	-- in the new library body
		  forth
	    end;
	    -- now do all the intra-library references
	    create refs.make(xlat);
	    from  start
	    until off
	    loop  item.rename_references(refs)
		  forth
	    end
	    -- finally, install nstru as my new library body
	    structures := nstru;
	    if su.not_empty(root_structure)
	     then root_structure := xlat.item(root_structure)
	    end;
	end;

feature	-- layer remapping changes

	change_layers(remapping: HASH_TABLE[STRING, STRING])
		-- each key in remapping is an extant layer name;
		-- replace all occurences of that layer name (in all
		-- structures in this library) with its respective value.
	require have_remapping: remapping /= Void
	local stru: LIBRARY_STRUCTURE
	do
	    from  start
	    until off
	    loop  stru := item;
		  stru.change_layers(remapping)
		  if stru.body_was_changed and is_caching then
			cache_this_structure(stru)
		  end
		  forth
	    end;
	end;

feature	-- root structure calculator

	unreferenced_structures: LIST[STRING] is
	    -- all the structures that are NOT referenced internally
	    -- NOTE: the declared root_structure is not relevant here
	local reffed: HASH_TABLE[BOOLEAN, STRING]; s: LIBRARY_STRUCTURE;
		rc: LC_MARK_REFERENCES;
	do  create reffed.make(count);	-- key is name; item is referenced
	    create rc.make(reffed);		-- the layout consumer we use
	    from  start
	    until off
	    loop  s := item
		  reffed.put(false, s.name);
		  s.body.traverse(rc, Identity)
		  forth
	    end;
	    -- now build Result from the ones never referenced
	    create {LINKED_LIST[STRING]} Result.make;
	    from  reffed.start
	    until reffed.off
	    loop  if not reffed.item_for_iteration
		   then Result.extend(reffed.key_for_iteration)
		  end
		  reffed.forth
	    end
	end;

feature	-- caching of structure bodies

	is_caching: BOOLEAN is
	do  Result := cache_file_name_prefix /= Void
	end;

	cache_bodies_to_files(fpre: STRING) is
		-- Begin the caching of all structure bodies to a series
		-- of files whose names are all constructed from fpre as a
		-- prefix (fpre must include the full directory path).
		-- A new file is begun approx every 1 GB of cached data.
		-- All structures already in the library are cached now,
		-- and new structures are cached as they are added
	require fpre_OK: su.not_empty(fpre); 
		not_caching: not is_caching
	local pos: CURSOR;
	do  cache_file_name_prefix := fpre;
	    create caching_files.make
	    -- now cache all the existing structures
	    pos := structures.cursor
	    from  structures.start
	    until structures.after
	    loop  cache_this_structure(structures.item_for_iteration)
		  structures.forth
	    end;
	    structures.go_to(pos)
	ensure is_caching
	end;

	wipe_out is
		-- the purpose of this routine is not library reuse, but
		-- rather the closing and deletion of the cache file(s).
		-- Letting a library be simply disposed will leave any
		-- caching files undeleted because ISE's library will not
		-- let us delete the files early.
	local fil: LARGE_FILE
	do  if caching_files /= Void
	     then from  caching_files.start
		  until caching_files.after
		  loop  fil := caching_files.item;
			fil.close; fil.delete;
			caching_files.forth
		  end;
		  caching_files := Void
	    end;
	    structures := Void		-- make self unusable
	end;

feature {NONE} -- the library

	structures: HASH_TABLE[LIBRARY_STRUCTURE, STRING];

	cache_file_name_prefix: STRING;

	caching_files: LINKED_LIST[LARGE_RAW_FILE]
		-- stack: first one is the active file

	cache_this_structure(stru: LIBRARY_STRUCTURE) is
		-- ask this structure to cache_to_file
		-- as soon as we can figure out which file
	require stru /= Void
	local fil: LARGE_RAW_FILE; siz: INTEGER_64;
	do  if not caching_files.is_empty  -- (empty just the first time)
	     then fil := caching_files.first;
		  siz := fil.count
	    end;
	    if fil = Void or siz > 1000000000
	     then -- time for a new file
		  create fil.make_create_read_write
			(su.string_of_2(cache_file_name_prefix,
					stru.name));
		  caching_files.put_front(fil);
		  siz := 0
	    end;
	    stru.cache_to_file(fil, siz);
	end;

feature {NONE} -- creation

	make(nam: STRING; unit: DOUBLE) is
		-- unit is meters/database unit
		-- a unit_size of 0 is allowed as an indicator that
		-- resizing is needed later (after we know more)
	require su.not_empty(nam); unit >= 0;
	do  name := nam
	    resize(unit)
	    date_accessed := now
	    create structures.make(919);
	end;

end -- LAYOUT_LIBRARY
