class LIBRARY_STRUCTURE
	-- a structure element in a LAYOUT_LIBRARY

inherit
	LIBRARY_SUPPORT

create
	make

feature	-- attributes (or pseudo-attributes)

	name: STRING
		-- name of this structure

	cif_number: INTEGER
		-- cif symbol number to use (-1 = not set)

	date_created,
	date_accessed,		-- (mistake??)
	date_modified: DB_DATE

	body: LAYOUT is		-- can be cached in a file
	do  if is_cached
	     then cache_file.go(cache_pos);
		  Result ?= cache_file.retrieved
	     else Result := in_memory_body
	    end
	end;

	is_cached: BOOLEAN is
		-- have we cached this body in a file?
	do  Result := cache_file /= Void
	end;

feature	-- body attributes

	local_layers_present: LIST[STRING]
		-- all of the layers present in the direct layout of my body
		-- (layout_references are NOT followed, whether instantiated
		--  or not)
	local lcoll: LC_LOCAL_LAYERS
	do  if body /= Void then
		create lcoll.make;
		body.traverse(lcoll, Identity)
		Result := lcoll.layers_found
	    end
	end;

feature -- attribute modification

	set_cif_number(i: INTEGER) is
	require i >= 0		-- is zero (0) allowed ???
	do  cif_number := i
	end;

	set_body(b: LAYOUT) is
	require b = Void or else not b.is_unlayered
	do  in_memory_body := b;
	    date_modified := now;
	    cache_file := Void	-- just lose cached version, if any
	end;

	set_date_created(dat: DB_DATE) is
	require dat /= Void
	do  date_created := dat
	end;

	set_date_accessed(dat: DB_DATE) is
	require dat /= Void
	do  date_accessed := dat
	end;

	set_date_modified(dat: DB_DATE) is
	require dat /= Void
	do  date_modified := dat
	end;

	rename_me(nam: STRING) is
		-- rename this structure (needed for library merging)
  --	require valid_structure_name(nam)
	do  make(nam)
	end;

feature -- body modification

	body_was_changed: BOOLEAN
		-- indicates whether last call to a body modifier made
		-- any changes to this body

	rename_references(refs: LC_RENAME_REFERENCES)
		-- refs already has the name translation table
		-- we just need to know whether we were altered
	do  body_was_changed := false;
	    refs.set_notifier(agent note_body_changes);
	    body.traverse(refs, identity);
	    if body_was_changed then
		set_body(body)	-- (may uncache)
	    end
	end;

	change_layers(remapping: HASH_TABLE[STRING, STRING])
		-- remapping defines the layers to be remapped (the keys)
		--  and the new layer for each (the values)
		-- body_was_changed reflects whether any changes were made
		--  to the body of this structure
	require have_map: remapping /= Void
	local nbody: LAYOUT
	do  body_was_changed := false;
	    nbody := body
	    nbody.change_layers(remapping, agent note_body_changes)
	    if body_was_changed then
		set_body(nbody)	-- (may uncache)
	    end;
	end;

feature {NONE} -- implementation

	in_memory_body: LAYOUT	-- if body in virtual memory

	cache_file: LARGE_FILE	-- if body is a stored object in this file
	cache_pos: INTEGER_64	--  and at this position

	note_body_changes(what: LAYOUT; ident: STRING)
	do  if what /= Void then
		body_was_changed := true
	    end
	end;

feature {LAYOUT_LIBRARY} -- cache body on demand

	cache_to_file(f: LARGE_FILE; pos: INTEGER_64) is
		-- cache body to f, starting at pos,
		-- and retrieve from there in the future
	require file_OK: f.is_open_read and f.is_open_write 
			 and f.support_storable;
		not_cached: not is_cached
	do  if in_memory_body /= Void	-- (ignore if body is Void)
	     then cache_pos := pos;	-- 
		  f.go(cache_pos)
		  f.basic_store(in_memory_body)
		  in_memory_body := Void;
		  cache_file := f
	    end
	end;

feature {NONE} -- creation

	make(nam: STRING) is
  --	require valid_structure_name(nam)
	do  name := nam
	    cif_number := -1
	end

invariant
	not_both: in_memory_body = Void or cache_file = Void

end -- LIBRARY_STRUCTURE
