
indexing
	
	description: "Miscellaneous utilities for lists"
	author: "David Hollenberg"
	date: "December 23, 2005"

class LIST_UTILITIES [G]

feature -- Filtering

	items_passing_filter (list: LIST [G] f: FUNCTION [ANY, TUPLE [G], BOOLEAN]): LIST [G]
			-- New list with only the items of `list'
			-- for which filter `f' is true, in the same order.
			-- Does not move cursor of `list'
		require
			list_not_void: list /= Void
			filter_not_void: f /= Void
		local
			c: CURSOR
			an_item: G
		do
			c := list.cursor
			Result := list.twin
			Result.wipe_out
			from
				list.start
			until
				list.after
			loop
				an_item := list.item
				if f.item ([an_item]) then
					Result.extend (an_item)
				end
				list.forth
			end
			list.go_to (c)
		ensure
			result_not_void: Result /= Void
		end
	
	items_failing_filter, items_not_passing_filter (list: LIST [G] f: FUNCTION [ANY, TUPLE [G], BOOLEAN]): LIST [G]
			-- New list with only the items of `list'
			-- for which filter `f' is false, in the same order.
			-- Does not move cursor of `list'
		require
			list_not_void: list /= Void
			filter_not_void: f /= Void
		local
			c: CURSOR
			an_item: G
		do
			c := list.cursor
			Result := list.twin
			Result.wipe_out
			from
				list.start
			until
				list.after
			loop
				an_item := list.item
				if not f.item ([an_item]) then
					Result.extend (an_item)
				end
				list.forth
			end
			list.go_to (c)
		ensure
			result_not_void: Result /= Void
		end
	
	items_passing_all_filters (list: LIST [G] filter_list: LIST [FUNCTION [ANY, TUPLE [G], BOOLEAN]]): LIST [G]
			-- New list with only the items of `list'
			-- for which all filters in `filter_list' are true,
			-- in the same order.
			-- Does not move cursor of `list' or `filter_list'
		require
			list_not_void: list /= Void
			filter_list_not_void: filter_list /= Void
		local
			c, fc: CURSOR
			an_item: G
			passed_all: BOOLEAN
			arg: TUPLE [G]
			f: FUNCTION [ANY, TUPLE [G], BOOLEAN]
		do
			c := list.cursor
			fc := filter_list.cursor
			Result := list.twin
			Result.wipe_out
			from
				list.start
			until
				list.after
			loop
				an_item := list.item
				passed_all := True
				arg := [an_item]
				from
					filter_list.start
				until
					filter_list.after or not passed_all
				loop
					f := filter_list.item
					passed_all := f.item (arg)
					filter_list.forth
				end
				if passed_all then
					Result.extend (an_item)
				end
				list.forth
			end
			list.go_to (c)
			filter_list.go_to (fc)
		ensure
			result_not_void: Result /= Void
		end
	
	items_passing_any_filter (list: LIST [G] filter_list: LIST [FUNCTION [ANY, TUPLE [G], BOOLEAN]]): LIST [G]
			-- New list with only the items of `list'
			-- for which at least one filter in `filter_list'
			-- is true, in the same order.
			-- Does not move cursor of `list' or `filter_list'
		require
			list_not_void: list /= Void
			filter_list_not_void: filter_list /= Void
		local
			c, fc: CURSOR
			an_item: G
			passed_one: BOOLEAN
			arg: TUPLE [G]
			f: FUNCTION [ANY, TUPLE [G], BOOLEAN]
		do
			c := list.cursor
			fc := filter_list.cursor
			Result := list.twin
			Result.wipe_out
			from
				list.start
			until
				list.after
			loop
				an_item := list.item
				passed_one := False
				arg := [an_item]
				from
					filter_list.start
				until
					filter_list.after or passed_one
				loop
					f := filter_list.item
					passed_one := f.item (arg)
					filter_list.forth
				end
				if passed_one then
					Result.extend (an_item)
				end
				list.forth
			end
			list.go_to (c)
			filter_list.go_to (fc)
		ensure
			result_not_void: Result /= Void
		end
	
	items_failing_any_filter, items_not_passing_all_filters (list: LIST [G] filter_list: LIST [FUNCTION [ANY, TUPLE [G], BOOLEAN]]): LIST [G]
			-- New list with only the items of `list'
			-- for which at least one filter in `filter_list'
			-- is false, in the same order.
			-- Does not move cursor of `list' or `filter_list'
		require
			list_not_void: list /= Void
			filter_list_not_void: filter_list /= Void
		local
			c, fc: CURSOR
			an_item: G
			failed_one: BOOLEAN
			arg: TUPLE [G]
			f: FUNCTION [ANY, TUPLE [G], BOOLEAN]
		do
			c := list.cursor
			fc := filter_list.cursor
			Result := list.twin
			Result.wipe_out
			from
				list.start
			until
				list.after
			loop
				an_item := list.item
				failed_one := False
				arg := [an_item]
				from
					filter_list.start
				until
					filter_list.after or failed_one
				loop
					f := filter_list.item
					failed_one := not f.item (arg)
					filter_list.forth
				end
				if failed_one then
					Result.extend (an_item)
				end
				list.forth
			end
			list.go_to (c)
			filter_list.go_to (fc)
		ensure
			result_not_void: Result /= Void
		end
	
	items_failing_all_filters, items_not_passing_any_filter (list: LIST [G] filter_list: LIST [FUNCTION [ANY, TUPLE [G], BOOLEAN]]): LIST [G]
			-- New list with only the items of `list'
			-- for which none of the filters in `filter_list'
			-- is true, in the same order.
			-- Does not move cursor of `list' or `filter_list'
		require
			list_not_void: list /= Void
			filter_list_not_void: filter_list /= Void
		local
			c, fc: CURSOR
			an_item: G
			passed_one: BOOLEAN
			arg: TUPLE [G]
			f: FUNCTION [ANY, TUPLE [G], BOOLEAN]
		do
			c := list.cursor
			fc := filter_list.cursor
			Result := list.twin
			Result.wipe_out
			from
				list.start
			until
				list.after
			loop
				an_item := list.item
				passed_one := False
				arg := [an_item]
				from
					filter_list.start
				until
					filter_list.after or passed_one
				loop
					f := filter_list.item
					passed_one := f.item (arg)
					filter_list.forth
				end
				if not passed_one then
					Result.extend (an_item)
				end
				list.forth
			end
			list.go_to (c)
			filter_list.go_to (fc)
		ensure
			result_not_void: Result /= Void
		end
	
	first_item_passing_filter (list: LIST [G] f: FUNCTION [ANY, TUPLE [G], BOOLEAN]): G
			-- First item in `list' for which filter `f'
			-- is true or Void if none.  If `f' has the
			-- target open, `list' must not contain any
			-- Void items.  Does not move cursor of `list'
		require
			list_not_void: list /= Void
			filter_not_void: f /= Void
		local
			c: CURSOR
			an_item: G
			found: BOOLEAN
		do
			c := list.cursor
			from
				list.start
			until
				list.after or found
			loop
				an_item := list.item
				if f.item ([an_item]) then
					Result := an_item
					found := True
				end
				list.forth
			end
			list.go_to (c)
		end
	
	first_item_failing_filter (list: LIST [G] f: FUNCTION [ANY, TUPLE [G], BOOLEAN]): G
			-- First item in `list' for which filter `f'
			-- is False or Void if none.  If `f' has the
			-- target open, `list' must not contain any
			-- Void items.  Does not move cursor of `list'
		require
			list_not_void: list /= Void
			filter_not_void: f /= Void
		local
			c: CURSOR
			an_item: G
			found: BOOLEAN
		do
			c := list.cursor
			from
				list.start
			until
				list.after or found
			loop
				an_item := list.item
				if not f.item ([an_item]) then
					Result := an_item
					found := True
				end
				list.forth
			end
			list.go_to (c)
		end
	
end
