
indexing
	
	author: "David Hollenberg"
	date: "October 7, 1997"
	description: "A network-wide shared or exclusive advisory lock on %
		%a name, which may have to be the name of the file for some %
		%operating systems."

deferred class LOCK

feature -- Properties

	lock_name: STRING
			-- Name of lock to which locking/unlocking
			-- operations will be applied
	
	is_exclusive: BOOLEAN
			-- Can only one entity at a time hold this lock?
			-- This is the default on creation
	
	is_shared: BOOLEAN
			-- Can multiple entities hold this lock simultaneously?
			-- This is not the default on creation
		do
			Result := not is_exclusive
		end
	
	minimize_retained_resources: BOOLEAN
			-- Should any auxiliary resources required to
			-- implement locking be released when they
			-- are no longer needed, instead of retaining them?

feature -- Status
	
	is_locked_by_me: BOOLEAN
			-- Is `lock_name' currently locked by this process?

	auxiliary_data: STRING
			-- Auxiliary data associated with lock
		deferred
		ensure
			result_exists: Result /= Void
		end

	conflicting_lock_exists: BOOLEAN
			-- Is `lock_name' currently locked by some
			-- other process, in a manner which prevents
			-- this process from obtaining the lock with
			-- the mode specified by `is_exclusive'?

	conflicting_lock_process_id: INTEGER
			-- Process id of the process (or one of the processes
			-- in the case of a shared lock) which is holding
			-- a lock that conflicts with `Current' or 0
			-- if no such process.
			-- May be 0 if referenced after a call to
			-- `lock_nowait' since there is a window where
			-- lock may be released
		require
			conflicting_lock_exists: conflicting_lock_exists
		do
			Result := lock_holder_process_id
		ensure
			nonnegative_result: Result >= 0
		end

	conflicting_lock_system_id: INTEGER
			-- System id of the process (or one of the processes
			-- in the case of a shared lock) which is holding
			-- a lock that conflicts with `Current' or 0
			-- if no such process.
			-- May be 0 if referenced after a call to
			-- `lock_nowait' since there is a window where
			-- lock may be released
		require
			conflicting_lock_exists: conflicting_lock_exists
		do
			Result := lock_holder_system_id
		ensure
			nonnegative_result: Result >= 0
		end

feature -- Modification

	set_exclusive
			-- Change mode of lock so that only one entity
			-- at a time can hold it
		require
			not_locked_by_me: not is_locked_by_me
		do
			is_exclusive := True
		ensure
			exclusive_mode_lock: is_exclusive
			not_locked_by_me: not is_locked_by_me
		end
	
	set_shared
			-- Change mode of lock so that multiple entities
			-- can hold it simultaneously
		require
			not_locked_by_me: not is_locked_by_me
		do
			is_exclusive := False
		ensure
			shared_mode_lock: is_shared
			not_locked_by_me: not is_locked_by_me
		end
	
	lock_wait
			
			-- Obtain network-wide advisory lock on
			-- name `lock_name', with locking mode indicated by
			-- `is_exclusive'.  If lock is currently locked
			-- by another process in a manner which prevents
			-- obtaining the lock, call blocks until lock
			-- is released.
		require
			not_locked_by_me: not is_locked_by_me
		deferred
		ensure
			locked_by_me: is_locked_by_me
			not_locked_by_other: not conflicting_lock_exists
		end
	
	lock_nowait
			-- Attempt to obtain network-wide advisory
			-- lock on name `lock_name', with locking
			-- mode indicated by `is_exclusive'.  Set
			-- `is_locked_by_me' to indicate whether the lock
			-- was gotten.  Also sets `conflicting_lock_exists',
			-- `conflicting_lock_process_id' and
			-- `conflicting_lock_system_id' appropriately.
			-- Lock won't be gotten if it is
			-- currently locked by another process in a manner
			-- which prevents obtaining the lock.  Does
			-- not block if lock is not available
		require
			not_locked_by_me: not is_locked_by_me
		deferred
		ensure
			locked_by_someone: is_locked_by_me xor conflicting_lock_exists
		end
	
	test_lock
			-- Set `conflicting_lock_exists' to indicate whether
			-- the network-wide lock on name `lock_name'
			-- is locked in a manner which would prevent
			-- obtaining the lock with mode specified by
			-- `is_exclusive'.  Also sets
			-- `conflicting_lock_process_id' and
			-- `conflicting_lock_system_id'.  Will not block
		require
			not_locked_by_me: not is_locked_by_me
		deferred
		end
	
	unlock
			-- Release lock on lock `lock_name'
		require
			lock_is_locked: is_locked_by_me
		deferred
		ensure
			lock_not_locked: not is_locked_by_me
		end

	set_auxiliary_data (data: STRING)
			-- Set auxiliary data associated with `Current'
			-- to `data'
		require
			data_exists: data /= Void
			lock_is_locked: is_locked_by_me
		deferred
		ensure
			auxiliary_data_set: auxiliary_data.is_equal (data)
		end

	append_auxiliary_data (data: STRING)
			-- Append `data' to auxiliary data associated with
			-- `Current'
		require
			data_exists: data /= Void
			lock_is_locked: is_locked_by_me
		deferred
		end

	clear_auxiliary_data
			-- Clear auxiliary data associated with `Current'
			-- (set it to the empty string)
		require
			lock_is_locked: is_locked_by_me
		deferred
		ensure
			auxiliary_data_cleared: auxiliary_data.is_equal ("")
		end

	set_minimize_retained_resources (b: BOOLEAN)
			-- Set `minimize_retained_resources' to `b'
		do
			minimize_retained_resources := b
		ensure
			value_set: minimize_retained_resources = b
		end
	
feature {NONE} -- Implementation

	lock_holder_process_id: INTEGER
			-- Process id of the process (or one of the processes
			-- in the case of a shared lock) which is holding
			-- a lock that conflicts with `Current'

	lock_holder_system_id: INTEGER
			-- System id of the process (or one of the processes
			-- in the case of a shared lock) which is holding
			-- a lock that conflicts with `Current'

invariant
	lock_name_exists: lock_name /= Void
	exclusive_or_shared: not (is_exclusive and is_shared)
	locked_or_blocked: not (is_locked_by_me and conflicting_lock_exists)
	process_implies_conflict: lock_holder_process_id > 0 implies conflicting_lock_exists
	system_implies_conflict: lock_holder_system_id > 0 implies conflicting_lock_exists

end
