
indexing
	
	description: "View of a model, which can be displayed on a drawable %
		%canvas using a projector"
	author: "David Hollenberg"
	date: "July 28, 2006"

deferred class MODEL_VIEW
inherit
	ANY
	NUMBER_UTILITIES_ACCESS
		export
			{NONE} all
		end
	PACKER_UTILITIES
		export
			{NONE} all
		end

feature {NONE} -- Creation

	make (a_model: like anchored_model a_drawable: EV_DRAWING_AREA)
			-- Create with `a_model' and `a_drawable'
		require
			a_model_not_void: a_model /= Void
			a_drawable_not_void: a_drawable /= Void
		do
			model := a_model
			drawable := a_drawable
			create upper_left_model_coords.make_xy (0, 0)
			create lower_right_model_coords.make_xy (0, 0)
			create displayed_model_area.make_by_corners (0, 0, 0, 0)
			create {ARRAYED_LIST [like anchored_view_element]} elements.make (0)
			create stock_pixmaps
			create rectangle
			create pixmap_rectangle
			create selected_count_change_actions
			create displayed_dimensions_change_actions
			create area_history_list_position_actions
			create saved_areas.make (1, 0)
			create displayed_area_history_list.make
			create current_window_position.make (0, 0)
			create previous_window_position.make (0, 0)
			selected_count := -1
		ensure
			model_set: model = a_model
			drawable_set: drawable = a_drawable
		end

feature -- Properties

	model: like anchored_model
			-- Model to be displayed by view

	grid: GRID
			-- Grid for object sizes and locations
		do
			Result := model.grid
		end

	drawable: EV_DRAWING_AREA
			-- Canvas on which view can be projected

	palette: VIEW_COLORS

	elements: LIST [like anchored_view_element]
			-- Elements of view, each corresponding to an
			-- element of the model, in stacking order
			-- (first element is at bottom, last element is on
			-- top of all other elements)

	saved_area_capacity: INTEGER
			-- Maximum number of saved areas supported

	displayed_area_history_capacity: INTEGER
			-- Maximum number of displayed areas of model
			-- in history list

	selected_count: INTEGER
			-- Number of items that are currently selected

	unselected_count: INTEGER
			-- Number of selectable items that are currently
			-- not selected
		do
			Result := selectable_count - selected_count
		end

	selectable_count: INTEGER
			-- Number of items that are selectable
		deferred
		end

	selected_count_change_actions: ACTION_SEQUENCE [TUPLE [INTEGER]]
			-- Actions to be performed when `selected_count'
			-- changes

	displayed_dimensions_change_actions: ACTION_SEQUENCE [TUPLE [DOUBLE, DOUBLE]]
			-- Actions to be performed when `displayed model_width'
			-- or `displayed_model_height' changes

	area_history_list_position_actions: ACTION_SEQUENCE [TUPLE [BOOLEAN, BOOLEAN]]
			-- Actions to be performed when current position in
			-- `displayed_area_history_list' may have changed.
			-- Values passed are whether position is at start
			-- and at end of list

	error_count: INTEGER
			-- Number of errors that occurred on last operation
		do
			Result := error_messages.count
		end

	warning_count: INTEGER
			-- Number of warnings that occurred on last operation
		do
			Result := warning_messages.count
		end

	had_error: BOOLEAN
			-- Did any errors occur on last operation?
		do
			Result := not error_messages.is_empty
		end

	had_warning: BOOLEAN
			-- Did any warnings occur on last operation?
		do
			Result := not warning_messages.is_empty
		end

	error_messages: LIST [STRING]
			-- Error messages from last operation

	warning_messages: LIST [STRING]
			-- Warning messages from last operation

	width: INTEGER
			-- Width in pixels
		do
			Result := drawable.width
		end

	height: INTEGER
			-- Height of view in pixels
		do
			Result := drawable.height
		end

	pointer_position: EV_COORDINATE
			-- Position of pointer relative to drawing area
		do
			Result := drawable.pointer_position
		end

	model_units_per_pixel: DOUBLE
			-- Number of units in model each pixel currently
			-- represents.  Changes when zoom in or out

	upper_left_model_coords: POINT
			-- Model coordinates of upper left corner of
			-- view (upper left corner of (0, 0) pixel).  Changes
			-- when display different part of model

	lower_right_model_coords: POINT
			-- Model coordinates of lower right corner of
			-- view (lower right corner of lowermost rightmost
			-- pixel).  Changes when display different part
			-- of model

	displayed_model_width: DOUBLE
			-- Width of area of model that is displayed on view
		do
			Result := lower_right_model_coords.x - upper_left_model_coords.x
		end

	displayed_model_height: DOUBLE
			-- Height of area of model that is displayed on view
		do
			Result := upper_left_model_coords.y - lower_right_model_coords.y
		end

	displayed_model_area: RECTANGLE
			-- Area of model that is displayed on view

	maximum_pixels_per_grid_unit: INTEGER
			-- Maximum number of pixels per model grid unit
			-- (to prevent zooming in too far)

	maximum_model_units_per_pixel: DOUBLE
			-- Maximum number of model units per pixel
			-- (to prevent zooming out too far)

	is_fully_zoomed_in: BOOLEAN
			-- Is view zoomed in as far as it can go?

	is_fully_zoomed_out: BOOLEAN
			-- Is view zoomed out as far as it can go?

	minimum_zoom_dimension_in_pixels: INTEGER
			-- Minimum width and height of zoom box in pixels
			-- for it to be interpreted as a zoom

	minimum_selection_dimension_in_pixels: INTEGER
			-- Minimum width or height of selection box in pixels
			-- for it to be interpreted as an area selection
			-- (select items completely inside box) rather than
			-- a point selection (select top-most item containing
			-- the point)

	minimum_dup_distance_in_pixels: INTEGER
			-- Minimum distance in pixels Duplicate operation
			-- must move from original position for it to be
			-- valid (to prevent accidental duplication)

	pixel_upper_left_model_coords (a_x, a_y: INTEGER): POINT
			-- Model coordinates of upper left corner of
			-- view pixel at (`a_x', `a_y')
		do
			create Result.make_xy (upper_left_model_coords.x + a_x * model_units_per_pixel, upper_left_model_coords.y - a_y * model_units_per_pixel)
		end

	pixel_upper_right_model_coords (a_x, a_y: INTEGER): CONST_POINT
			-- Model coordinates of upper right corner of
			-- view pixel at (`a_x', `a_y')
		do
			create Result.make_xy (upper_left_model_coords.x + (a_x + 1) * model_units_per_pixel, upper_left_model_coords.y - a_y * model_units_per_pixel)
		end

	pixel_center_model_coords (a_x, a_y: INTEGER): CONST_POINT
			-- Model coordinates of center of view pixel
			-- at (`a_x', `a_y')
		do
			create Result.make_xy (upper_left_model_coords.x + (a_x + 0.5) * model_units_per_pixel, upper_left_model_coords.y - (a_y + 0.5) * model_units_per_pixel)
		end

	pixel_lower_left_model_coords (a_x, a_y: INTEGER): POINT
			-- Model coordinates of lower left corner of
			-- view pixel at (`a_x', `a_y')
		do
			create Result.make_xy (upper_left_model_coords.x + a_x * model_units_per_pixel, upper_left_model_coords.y - (a_y + 1) * model_units_per_pixel)
		end

	pixel_lower_right_model_coords (a_x, a_y: INTEGER): POINT
			-- Model coordinates of lower right corner of
			-- view pixel at (`a_x', `a_y')
		do
			create Result.make_xy (upper_left_model_coords.x + (a_x + 1) * model_units_per_pixel, upper_left_model_coords.y - (a_y + 1) * model_units_per_pixel)
		end

	model_point_view_coords (a_x, a_y: DOUBLE): TUPLE [INTEGER_64, INTEGER_64]
			-- View coordinates of model point (`a_x', `a_y')
		local
			vx, vy: INTEGER_64
		do
			vx := nu.floor_integer_64 ((a_x - upper_left_model_coords.x) / model_units_per_pixel)
			vy := nu.floor_integer_64 ((upper_left_model_coords.y - a_y) / model_units_per_pixel)
			Result := [vx, vy]
		end

	set_view_rectangle_from_model_rectangle (a_rectangle: EV_RECTANGLE a_max: INTEGER a_left, a_top, a_right, a_bottom: DOUBLE)
			-- Set `a_rectangle' to the view rectangle
			-- corresponding to model rectangle defined by
			-- `a_left', `a_right', `a_top' and `a_bottom'
			-- and clip it by rectangle whose corners are
			-- (-`a_max', -`a_max') to (`a_max', `a_max')
			
		require
			a_rectangle_not_void: a_rectangle /= Void
			left_right_ok: a_left <= a_right
			top_bottom_ok: a_bottom <= a_top
		local
			ul_x, ul_y, lr_x, lr_y: INTEGER_64
			vals: TUPLE [INTEGER_64, INTEGER_64]
		do
			vals := model_point_view_coords (a_left, a_top)
			ul_x := vals.integer_64_item (1)
			ul_y := vals.integer_64_item (2)
			vals := model_point_view_coords (a_right, a_bottom)
			lr_x := vals.integer_64_item (1)
			lr_y := vals.integer_64_item (2)
			set_clipped_rectangle_from_values (a_rectangle, a_max, ul_x, ul_y, lr_x - ul_x + 1, lr_y - ul_y + 1)
		end

	model_area_for_view_rectangle (r: EV_RECTANGLE): CONST_RECTANGLE
			-- View rectangle `r' translated to model coordinates
		local
			ul, lr: CONST_POINT
		do
			ul := pixel_upper_left_model_coords (r.left, r.top)
			lr := pixel_lower_right_model_coords (r.right, r.bottom)
			create Result.make_by_corners (ul.x, lr.y, lr.x, ul.y)
		end

	zoomed_out (a_rect: CONST_RECTANGLE): CONST_RECTANGLE
			-- Area of model to be displayed when currently
			-- displayed area is fit into `a_rect',
			-- after first increasing width or height of `a_rect'
			-- so that it has the same aspect ratio as the view
		require
			a_rect_not_void: a_rect /= Void
		local
			h, w, aspect_ratio, sf: DOUBLE
			drw_height, drw_width, new_x, new_y: DOUBLE
			ratio: DOUBLE
			r, rect: RECTANGLE
		do
			create r.make_copy (a_rect)
			w := r.size_x
			h := r.size_y
			drw_width := drawable.width.max (1)
			drw_height := drawable.height.max (1)
			aspect_ratio := drw_width / drw_height
			if w / h > aspect_ratio then
				-- Increase height
				h := w / aspect_ratio
			elseif w / h < aspect_ratio then
				-- Increase width
				w := h * aspect_ratio
			end
			
			sf := (lower_right_model_coords.x - upper_left_model_coords.x) / w
			ratio := maximum_model_units_per_pixel / model_units_per_pixel
			if sf > ratio then
				sf := sf / ratio
				w := w * sf
				h := h * sf
			end
			r.make_by_size_and_center (w, h, r.center_x, r.center_y)
			sf := (lower_right_model_coords.x - upper_left_model_coords.x) / w
			create rect.make_copy (displayed_model_area)
			rect.scale_by (sf)
			new_x := upper_left_model_coords.x + sf * (upper_left_model_coords.x - r.low_x)
			new_y := upper_left_model_coords.y + sf * (upper_left_model_coords.y - r.high_y)
			rect.move_to (new_x + sf * displayed_model_width / 2.0, new_y - sf * displayed_model_height / 2.0)
			
			Result := rect
		end

	is_redraw_enabled: BOOLEAN
			-- Should redrawing of model be done?

feature -- Modification

	set_palette (a_palette: like palette)
		require
			a_palette_not_void: a_palette /= Void
		do
			palette := a_palette
		end

	set_maximum_pixels_per_grid_unit_from_dist (a_value: DOUBLE)
			-- Set `maximum_pixels_per_grid_unit' so that
			-- when zoomed all the way in one grid unit
			-- is `a_value' inches (so it will look about the
			-- same regardless of screen resolution)
		require
			a_value_positive: a_value > 0.0
		do
			maximum_pixels_per_grid_unit := inches_to_pixels (a_value)
		end

	set_maximum_model_units_per_pixel (a_model_units, a_distance: DOUBLE)
			-- Set `maximum_model_units_per_pixel' so that
			-- when zoomed all the way out `a_model_units'
			-- model units occupies `a_distance' inches
			-- (so it will look about the same regardless
			-- of screen resolution)
		require
			a_model_units_positive: a_model_units > 0.0
			a_distance_positive: a_distance > 0.0
		do
			maximum_model_units_per_pixel := a_model_units / inches_to_pixels (a_distance)
		end

	set_minimum_zoom_dimension_in_pixels (a_value: like minimum_zoom_dimension_in_pixels)
			-- Set `minimum_zoom_dimension_in_pixels' to `a_value'
		require
			a_value_nonnegative: a_value >= 0
		do
			minimum_zoom_dimension_in_pixels := a_value
		end

	set_minimum_selection_dimension_in_pixels (a_value: like minimum_selection_dimension_in_pixels)
			-- Set `minimum_selection_dimension_in_pixels' to
			-- `a_value'
		require
			a_value_nonnegative: a_value >= 0
		do
			minimum_selection_dimension_in_pixels := a_value
		end

	set_minimum_dup_distance_in_pixels (a_value: like minimum_dup_distance_in_pixels)
			-- Set `minimum_dup_distance_in_pixels' to
			-- `a_value'
		require
			a_value_nonnegative: a_value >= 0
		do
			minimum_dup_distance_in_pixels := a_value
		end

	set_saved_area_capacity (a_value: like saved_area_capacity)
			-- Set `saved_area_capacity' to `a_value'
		require
			a_value_nonnegative: a_value >= 0
		do
			saved_area_capacity := a_value
			saved_areas.conservative_resize (1, a_value)
		end

	set_displayed_area_history_capacity (a_value: like displayed_area_history_capacity)
			-- Set `displayed_area_history capacity' to `a_value'
		require
			a_value_nonnegative: a_value >= 0
		do
			displayed_area_history_capacity := a_value
		end

	set_current_window_position (a_x, a_y: INTEGER)
			-- Set current window position to (`a_x', `a_y')
		do
			current_window_position.set (a_x, a_y)
		end

	save_current_window_position
			-- Save current window position as previous
			-- window position
		do
			previous_window_position.set (current_window_position.x, current_window_position.y)
		end

	enable_redraw
			-- Enable redraw
		do
			is_redraw_enabled := True
		end

	disable_redraw
			-- Disable redraw
		do
			is_redraw_enabled := False
		end

feature -- Operations

	initialize
			-- Initialize view
		deferred
		end

	set_displayed_model_area_without_history (ul_x, ul_y, lr_x, lr_y: DOUBLE)
			-- Set area of model that is displayed to
			-- have upper left (`ul_x', `ul_y') and lower
			-- right (`lr_x', `lr_y').  Do not change
			-- `displayed_area_history_list'
		do
			upper_left_model_coords.make_xy (ul_x, ul_y)
			lower_right_model_coords.make_xy (lr_x, lr_y)
			displayed_model_area.make_by_corners (ul_x, lr_y, lr_x, ul_y)
			set_all_view_coordinates
			displayed_dimensions_change_actions.call ([displayed_model_width, displayed_model_height])
		end

	set_displayed_model_area (ul_x, ul_y, lr_x, lr_y: DOUBLE)
			-- Set area of model that is displayed to
			-- have upper left (`ul_x', `ul_y') and lower
			-- right (`lr_x', `lr_y').  Add the new displayed
			-- area to the history list
		do
			set_displayed_model_area_without_history (ul_x, ul_y, lr_x, lr_y)
			add_displayed_area_to_history
		end

	set_adjusted_displayed_model_area (r: CONST_RECTANGLE)
			-- Set area of model that is displayed from `r',
			-- after first increasing width or height of `r'
			-- so that it has the same aspect ratio as the view.
			-- Add new displayed area to
			-- `displayed_area_history_list'
		require
			r_not_void: r /= Void
		do
			set_adjusted_displayed_model_area_without_history (r)
			add_displayed_area_to_history
		end

	set_adjusted_displayed_model_area_without_history (r: CONST_RECTANGLE)
			-- Set area of model that is displayed from `r',
			-- after first increasing width or height of `r'
			-- so that it has the same aspect ratio as the view.
			-- Do not change `displayed_area_history_list'
		local
			h, w, aspect_ratio, cx, cy: DOUBLE
			sf, drw_height, drw_width: DOUBLE
			l_pixels_per_grid_unit, the_min: DOUBLE
		do
			w := r.size_x
			h := r.size_y
			drw_width := drawable.width.max (1)
			drw_height := drawable.height.max (1)
			aspect_ratio := drw_width / drw_height
			if w / h > aspect_ratio then
				-- Increase height
				h := w / aspect_ratio
			elseif w / h < aspect_ratio then
				-- Increase width
				w := h * aspect_ratio
			end
			
			-- Increase width and height so they are not below
			-- the minimum, keeping same aspect ratio

			model_units_per_pixel := w / drawable.width
			l_pixels_per_grid_unit := grid.edge_resolution / model_units_per_pixel
			is_fully_zoomed_in := l_pixels_per_grid_unit + Pixels_per_grid_unit_tolerance >= maximum_pixels_per_grid_unit
			if is_fully_zoomed_in then
				the_min := grid.edge_resolution / maximum_pixels_per_grid_unit
				sf := (the_min / model_units_per_pixel)
				model_units_per_pixel := the_min
				if sf > 1.0 then
					-- Scale both width and height
					w := w * sf
					h := h * sf
				end
			end
			
			-- Decrease width and height so they are not above
			-- the maximum, keeping same aspect ratio

			model_units_per_pixel := w / drawable.width
			is_fully_zoomed_out := model_units_per_pixel >= maximum_model_units_per_pixel
			if is_fully_zoomed_out then
				sf := maximum_model_units_per_pixel / model_units_per_pixel
				model_units_per_pixel := maximum_model_units_per_pixel
				if sf < 1.0 then
					-- Scale both width and height
					w := w * sf
					h := h * sf
				end
			end
			
			debug ("packer")
				print ("Width = " + w.out + "   Height = " + h.out + "%N")
			end
			
			cx := r.center_x
			cy := r.center_y
			set_displayed_model_area_without_history (cx - 0.5 * w, cy + 0.5 * h, cx + 0.5 * w, cy - 0.5 * h)
		end

	set_all_view_coordinates
			-- Set view coordinates for all elements of view,
			-- after change in displayed model area
		do
			from
				elements.start
			until
				elements.after
			loop
				elements.item.set_view_coordinates
				elements.forth
			end
		end

	set_initial_model_elements (list: LIST [like anchored_model_element] with_redraw: BOOLEAN)
			-- Set initial list of model elements
			-- to `list'.  Redraw model if `with_redraw'
		require
			list_not_void: list /= Void
		do
			create {ARRAYED_LIST [like anchored_view_element]} elements.make (list.count)
			add_model_elements (list, with_redraw)
			displayed_area_history_list.wipe_out
			set_adjusted_displayed_model_area (enlarged_model_bounding_box)
			set_selected_count (0)
		end

	set_saved_area (an_area_number: INTEGER)
			-- Set saved area `an_area_number' to the current
			-- displayed area of model
		require
			an_area_number_positive: an_area_number > 0
			an_area_number_small_enough: an_area_number <= saved_area_capacity
		local
			rect: RECTANGLE
			ul, lr: POINT
		do
			rect := saved_areas.item (an_area_number)
			ul := upper_left_model_coords
			lr := lower_right_model_coords
			if rect = Void then
				create rect.make_copy (displayed_model_area)
				saved_areas.put (rect, an_area_number)
			else
				rect.make_copy (displayed_model_area)
			end
		end

	set_selected_count (a_new_count: INTEGER)
			-- Set `selected_count' to `a_new_count'
		require
			a_new_count_nonnegative: a_new_count >= 0
		deferred
		ensure
			selected_count_set: selected_count = a_new_count
		end

	add_model_elements (list: LIST [like anchored_model_element] with_redraw: BOOLEAN)
			-- Add model elements in `list' to view.  Redraw
			-- model if `with_redraw'
		require
			list_not_void: list /= Void
		deferred
		end

	remove_model_elements (list: LIST [PACKING_DIES_MROW])
			-- Remove model elements corresponding to dies
			-- in `list' from model and view
		require
			list_not_void: list /= Void
		deferred
		end

	show_entire_model
			-- Display the whole model
		do
			show_entire_model_without_redraw
			redraw
		end

	show_entire_model_without_redraw
			-- Display the whole model but don't redraw
		do
			set_adjusted_displayed_model_area (enlarged_model_bounding_box)
		end

	refresh_drawing_area
			-- Refresh the drawing area from the pixmap
		do
			redraw_canvas
		end

	show_saved_area (an_area_number: INTEGER)
			-- Display the area of the model previously saved
			-- under `an_area_number'
		local
			rect: CONST_RECTANGLE
		do
			initialize_status
			rect := saved_areas.item (an_area_number)
			if rect /= Void then
				set_adjusted_displayed_model_area (rect)
				redraw
			else
				add_error ("Could not show saved area " + an_area_number.out + " since it has never been set")
			end
		end

	show_previous_area
			-- Display the area of the model before the current
			-- displayed area in the history list
		do
			initialize_status
			if displayed_area_history_list.off then
				add_error ("Area history list is empty")
			else
				displayed_area_history_list.back
				if displayed_area_history_list.before then
					displayed_area_history_list.forth
					add_error ("No previous area in history list")
				else
					set_adjusted_displayed_model_area_without_history (displayed_area_history_list.item)
					redraw
				end
			end
			call_area_history_list_actions
		end

	show_next_area
			-- Display the area of the model after the current
			-- displayed area in the history list
		do
			initialize_status
			if displayed_area_history_list.off then
				add_error ("Area history list is empty")
			else
				displayed_area_history_list.forth
				if displayed_area_history_list.after then
					displayed_area_history_list.back
					add_error ("No next area in history list")
				else
					set_adjusted_displayed_model_area_without_history (displayed_area_history_list.item)
					redraw
				end
			end
			call_area_history_list_actions
		end

	redraw
			-- Clear pixmap, draw view on pixmap and draw
			-- pixmap on drawable
		do
			pixmap.clear
			draw_view_on_pixmap
			redraw_canvas
		end

	redraw_canvas
			-- Redraw pixmap onto drawable
		do
			drawable.draw_pixmap (0, 0, pixmap)
		end

	redraw_canvas_rectangle (a_x, a_y, a_width, a_height: INTEGER)
			-- Redraw the area of the canvas whose upper left
			-- corner is at (`a_x', `a_y') and whose size
			-- is `a_width' by `a_height'
		do
			rectangle.move_and_resize (a_x, a_y, a_width, a_height)
			-- FIXME: work around Vision2 bug
			-- drawable.draw_sub_pixmap (a_x, a_y, pixmap, rectangle)
			draw_sub_pixmap (drawable, a_x, a_y, pixmap, rectangle)
		end

	handle_canvas_resize (a_x, a_y, a_width, a_height: INTEGER)
			-- Handle "canvas resized" event, where new upper left
			-- corner is at (`a_x', `a_y') and new size
			-- is `a_width' by `a_height'
		require
			a_width_nonnegative: a_width >= 0
			a_height_nonnegative: a_height >= 0
		deferred
		end

	set_clip_area (a_rectangle: EV_RECTANGLE)
			-- Set clip area for `pixmap' to
			-- `a_rectangle', if no clip region is already
			-- defined for view
		do
			if not is_clipping_set then
				pixmap.set_clip_area (a_rectangle)
			end
		end

	remove_clip_area
			-- Remove clip area for `pixmap', if no clip
			-- region is already defined for view
		do
			if not is_clipping_set then
				pixmap.remove_clip_area
			end
		end

	add_error (a_message: STRING)
			-- Add a new error described by `a_message'
		require
			a_message_not_void: a_message /= Void
		do
			error_messages.extend (a_message)
		end

	add_warning (a_message: STRING)
			-- Add a new warning described by `a_message'
		require
			a_message_not_void: a_message /= Void
		do
			warning_messages.extend (a_message)
		end

	set_visible_rectangle (a_rectangle: EV_RECTANGLE a_drawable: EV_DRAWABLE a_x, a_y, a_width, a_height, a_line_width: INTEGER)
			-- Set `a_rectangle' to be the visible portion
			-- in `a_drawable' of rectangle whose upper
			-- left corner is at (`a_x', `a_y') and size
			-- is `a_width' by `a_height' where rectangle
			-- is drawn with `a_line_width'.  May actually
			-- extend outside bounds of drawable a little
			-- bit.  Routine is needed to work around
			-- Vision2 bug
		local
			l_x, l_y, l_width, l_height, diff: INTEGER
			drw_w, drw_h: INTEGER
		do
			l_x := a_x
			l_y := a_y
			l_width := a_width
			l_height := a_height
			drw_w := a_drawable.width
			drw_h := a_drawable.height

			if l_x < -a_line_width then
				diff := -a_line_width - l_x
				l_x := -a_line_width
				l_width := (l_width - diff).max (0)
			end
			if l_x + l_width > drw_w + a_line_width then
				diff := l_x + l_width - (drw_w + a_line_width)
				l_width := (l_width - diff).max (0)
			end

			if l_y < -a_line_width then
				diff := -a_line_width - l_y
				l_y := -a_line_width
				l_height := (l_height - diff).max (0)
			end
			if l_y + l_height > drw_h + a_line_width then
				diff := l_y + l_height - (drw_h + a_line_width)
				l_height := (l_height - diff).max (0)
			end
			a_rectangle.move_and_resize (l_x, l_y, l_width, l_height)
		end
	
feature {VIEW_ELEMENT_DRAWER} -- Implementation
	
	changed_bounding_box: EV_RECTANGLE
			-- Bounding box of changed region
		deferred
		end

feature {NONE} -- Implementation

	draw_sub_pixmap (a_drawable: EV_DRAWABLE a_x, a_y: INTEGER a_pixmap: EV_PIXMAP an_area: EV_RECTANGLE)
			-- Draw `an_area' of `a_pixmap' onto `a_drawable' with
			-- upper left corner at (`a_x', `a_y').  Only handles
			-- case where upper left corner of `an_area' is
			-- (`a_x', `a_y').  Intersect
			-- `an_area' with rectangle covering `a_drawable'
			-- (modifies `an_area').  Routine is needed to work
			-- around Vision2 bug
		
		require
			a_drawable_not_void: a_drawable /= Void
			a_pixmap_not_void: a_pixmap /= Void
			an_area_not_void: an_area /= Void
			same_x: an_area.x = a_x
			same_y: an_area.y = a_y
		do
			if an_area.intersects (pixmap_rectangle) then
				intersect (an_area, pixmap_rectangle)
			end
			a_drawable.draw_sub_pixmap (an_area.x, an_area.y, pixmap, an_area)
		end

	intersect (a_rectangle, other: EV_RECTANGLE)
			-- Set `a_rectangle' to be the intersection of
			-- itself with `other'
		require
			a_rectangle_not_void: a_rectangle /= Void
			other_not_void: other /= Void
		local
			l_top, l_bottom, l_left, l_right: INTEGER
		do
			l_top := a_rectangle.top.max (other.top)
			l_bottom := a_rectangle.bottom.min (other.bottom)
			l_left := a_rectangle.left.max (other.left)
			l_right := a_rectangle.right.min (other.right)
			if l_top < 0 then
				a_rectangle.set_top (l_top)
				a_rectangle.set_bottom (l_bottom)
			else
				a_rectangle.set_bottom (l_bottom)
				a_rectangle.set_top (l_top)
			end

			if l_left < 0 then
				a_rectangle.set_left (l_left)
				a_rectangle.set_right (l_right)
			else
				a_rectangle.set_right (l_right)
				a_rectangle.set_left (l_left)
			end
		end

	draw_rectangle (a_drawable: EV_DRAWABLE a_x, a_y, a_width, a_height, a_line_width: INTEGER)
			-- Draw rectangle on `a_drawable' with upper
			-- left corner at (`a_x', `a_y') and size
			-- `a_width' by `a_height'.  Routine is needed
			-- to work around Vision2 bug
		
		require
			a_drawable_not_void: a_drawable /= Void
			a_width_nonnegative: a_width >= 0
			a_height_nonnegative: a_height >= 0
		do
			set_visible_rectangle (rectangle, a_drawable, a_x, a_y, a_width, a_height, a_line_width)
			a_drawable.draw_rectangle (rectangle.x, rectangle.y, rectangle.width, rectangle.height)
		end


	new_view_element (a_model_element: like anchored_model_element): like anchored_view_element
			-- New view element for `a_model_element'
		deferred
		ensure
			result_not_void: Result /= Void
		end

	rectangle_contained_in (r1, r2: EV_RECTANGLE): BOOLEAN
			-- Is `r1' contained in `r2' (i.e., is intersection
			-- of `r1' and `r2' equal to `r1')?
		require
			rectangle_1_not_void: r1 /= Void
			rectangle_2_not_void: r2 /= Void
		do
			Result := r2.has_x_y (r1.left, r1.top) and then
				r2.has_x_y (r1.left, r1.bottom) and then
				r2.has_x_y (r1.right, r1.top) and then
				r2.has_x_y (r1.right, r1.bottom)
		end

	enlarged_model_bounding_box: CONST_RECTANGLE
			-- Minimum bounding box of model enlarged a bit
			-- so that there is some empty space around it
		deferred
		ensure
			result_not_void: Result /= Void
		end

	draw_view_on_pixmap
			-- Draw all view elements which have any pixels
			-- that fall onto pixmap
		local
			elem: like anchored_view_element
		do
			from
				elements.start
			until
				elements.after
			loop
				elem := elements.item
				if elem.is_transparent then
					elem.draw_fill_only (pixmap, palette)
				end
				elements.forth
			end
			
			from
				elements.start
			until
				elements.after
			loop
				elem := elements.item
				if elem.is_opaque then
					-- Draw everything for opaque elements
					-- on 2nd pass
					elem.draw_fill_only (pixmap, palette)
				end
				elem.draw_without_fill (pixmap, palette)
				elements.forth
			end
		end

	initialize_status
			-- Initialize status attributes before starting
			-- operation that could have errors or warnings
		do
			create {ARRAYED_LIST [STRING]} error_messages.make (0)
			create {ARRAYED_LIST [STRING]} warning_messages.make (0)
		end

	add_displayed_area_to_history
			-- Add the currently displayed area to the end
			-- of the history list, after removing any items
			-- after the cursor position.  Remove first item
			-- if list now exceeds history capacity
		local
			rect: RECTANGLE
			list: like displayed_area_history_list
			c: CURSOR
		do
			if displayed_model_width > 0.0 and displayed_model_height > 0.0 then
				list := displayed_area_history_list
				from
				until
					list.index >= list.count
				loop
					list.remove_right
				end
				if displayed_area_history_capacity > 0 and then (list.off or else not list.item.is_same_box (displayed_model_area)) then
					create rect.make_copy (displayed_model_area)
					list.extend (rect)
					list.forth
					if list.count > displayed_area_history_capacity then
						c := list.cursor
						list.start
						list.remove
						list.go_to (c)
					end
				end
				call_area_history_list_actions
			end
		end

	call_area_history_list_actions
			-- Call `area_history_list_position_actions'
		local
			list: like displayed_area_history_list
		do
			list := displayed_area_history_list
			area_history_list_position_actions.call ([list.is_empty or list.before or list.isfirst, list.is_empty or list.after or list.islast])
		end

	inches_to_pixels (a_value: DOUBLE): INTEGER
			-- `a_value' inches converted to pixels, based on
			-- screen resolution
		require
			a_value_nonnegative: a_value >= 0.0
		local
			s: EV_SCREEN
			h, v: DOUBLE
		do
			create s
			h := (a_value * s.horizontal_resolution)
			v := (a_value * s.vertical_resolution)
			Result := h.max (v).rounded
		end

	is_clipping_set: BOOLEAN
			-- Is there currently a clip region set on `pixmap'?

	anchored_view_element: VIEW_ELEMENT
			-- Type of elements in view
		do
		ensure
			result_void: Result = Void
		end

	anchored_model_element: MODEL_ELEMENT
			-- Type of elements in model
		do
		ensure
			result_void: Result = Void
		end

	anchored_model: MODEL
			-- Type of model
		do
		ensure
			result_void: Result = Void
		end

	pixmap: EV_PIXMAP
			-- Pixmap to which to draw everything, then
			-- refresh view by copying pixmap to it

	rectangle: EV_RECTANGLE
			-- Reusable rectangle.  Avoids need to create new
			-- objects repeatedly

	pixmap_rectangle: EV_RECTANGLE
			-- Rectangle that covers entire pixmap.
			-- Avoids need to create new objects
			-- repeatedly

	stock_pixmaps: EV_STOCK_PIXMAPS
			-- Access to stock pixmaps

	saved_areas: ARRAY [RECTANGLE]
			-- Model areas that have been saved so that we
			-- can quickly go back to displaying that area

	displayed_area_history_list: TWO_WAY_LIST [RECTANGLE]
			-- History list of model areas that have been displayed

	current_window_position: EV_COORDINATE
			-- Current position of upper left corner of window
			-- containing view

	previous_window_position: EV_COORDINATE
			-- Previous position of upper left corner of window
			-- containing view (before window resize)

	Pixels_per_grid_unit_tolerance: DOUBLE = 0.001
			-- How close pixels per grid unit must be to
			-- `maximum_pixels_per_grid_unit' for view to
			-- be considered zoomed in fully

invariant
	drawable_not_void: drawable /= Void
	rectangle_not_void: rectangle /= Void
end
