
indexing
	
	description: "Set of bindings of MODIFIED_KEY (key code plus Ctrl, %
		%Alt, Shift modifier) to action to be executed"
	author: "David Hollenberg"
	date: "March 13, 2007"

class MODIFIED_KEY_BIND_SET
inherit
	EV_KEY_CONSTANTS
	MODIFIER_KEY_CONSTANTS

create
	make

feature {NONE} -- Creation

	make (an_unbound_action: like unbound_key_action a_read_only_error_action: like read_only_error_action)
			-- Create empty set of bindings with `an_unbound_action'
			-- as unbound key action and `a_read_only_error_action'
		require
			a_read_only_error_action_not_void: a_read_only_error_action /= Void
		do
			create lookup_table.make (1, Modifier_combinations)
			set_unbound_key_action (an_unbound_action)
			set_read_only_error_action (a_read_only_error_action)
		end

feature -- Properties

	unbound_key_action: PROCEDURE [ANY, TUPLE]
			-- Action to be executed for unbound key

	read_only_error_action: PROCEDURE [ANY, TUPLE]
			-- Action to be executed when key is pressed in
			-- read-only mode but key is not bound to a read-only
			-- action

	post_call_action: PROCEDURE [ANY, TUPLE]
			-- Action to be executed by `call' after calling
			-- procedure bound to key

feature -- Modification

	set_unbound_key_action (an_action: like unbound_key_action)
			-- Set `unbound_key_action' to `an_action'
		do
			unbound_key_action := an_action
		ensure
			unbound_key_action_set: unbound_key_action = an_action
		end

	set_read_only_error_action (an_action: like read_only_error_action)
			-- Set `read_only_error_action' to `an_action'
		require
			an_action_not_void: an_action /= Void
		do
			read_only_error_action := an_action
		ensure
			read_only_error_action_set: read_only_error_action = an_action
		end

	set_post_call_action (an_action: like post_call_action)
			-- Set `post_call_action' to `an_action'
		do
			 post_call_action := an_action
		ensure
			post_call_action_set: post_call_action = an_action
		end

	bind (a_key_code: INTEGER alt, ctrl, shift, read_only: BOOLEAN an_action: like unbound_key_action)
			-- Bind modified key defined by `a_key_code' and
			-- `alt', `ctrl' and `shift' to `an_action',
			-- which is a read-only action only if `read_only
		require
			valid_key_code: valid_key_code (a_key_code)
		local
			mod_code, pos: INTEGER
			table: ARRAY [KEY_BINDING]
		do
			mod_code := modifier_code (alt, ctrl, shift)
			pos := mod_code - Minimum_modifier_code + 1
			table := lookup_table.item (pos)
			if table = Void then
				create table.make (Minimum_key_code, Maximum_key_code)
				lookup_table.put (table, pos)
			end
			table.put (create {KEY_BINDING}.make (an_action, read_only), a_key_code)
		end

	bind_unmodified (a_key_code: INTEGER an_action: like unbound_key_action)
			-- Bind unmodified key defined by `a_key_code' and
			-- to read-only action `an_action'
		require
			valid_key_code: valid_key_code (a_key_code)
		do
			bind (a_key_code, False, False, False, True, an_action)
		end

	bind_control (a_key_code: INTEGER an_action: like unbound_key_action)
			-- Bind control key defined by `a_key_code' and
			-- to read-only action `an_action'
		require
			valid_key_code: valid_key_code (a_key_code)
		do
			bind (a_key_code, False, True, False, True, an_action)
		end

	bind_shift (a_key_code: INTEGER an_action: like unbound_key_action)
			-- Bind shifted key defined by `a_key_code' and
			-- to read-only `an_action'
		require
			valid_key_code: valid_key_code (a_key_code)
		do
			bind (a_key_code, False, False, True, True, an_action)
		end

	bind_unmodified_change (a_key_code: INTEGER an_action: like unbound_key_action)
			-- Bind unmodified key defined by `a_key_code' and
			-- to non read-only action `an_action'
		require
			valid_key_code: valid_key_code (a_key_code)
		do
			bind (a_key_code, False, False, False, False, an_action)
		end

	bind_control_change (a_key_code: INTEGER an_action: like unbound_key_action)
			-- Bind control key defined by `a_key_code' and
			-- to non read-only action `an_action'
		require
			valid_key_code: valid_key_code (a_key_code)
		do
			bind (a_key_code, False, True, False, False, an_action)
		end

	bind_shift_change (a_key_code: INTEGER an_action: like unbound_key_action)
			-- Bind shifted key defined by `a_key_code' and
			-- to non read-only `an_action'
		require
			valid_key_code: valid_key_code (a_key_code)
		do
			bind (a_key_code, False, False, True, False, an_action)
		end

feature -- Operation

	call (a_key: MODIFIED_KEY read_only_mode: BOOLEAN)
			-- Call action associated with `a_key' if not
			-- `read_only_mode'.  If `read_only_mode', then
			-- call `read_only_error_action' instead if associated
			-- action is not read-only.  If key is
			-- not bound, call `unbound_key_action' if it is not
			-- Void
		require
			a_key_not_void: a_key /= Void
		local
			mod_code, pos: INTEGER
			l_binding: KEY_BINDING
			l_action: like unbound_key_action
			table: ARRAY [KEY_BINDING]
		do
			mod_code := a_key.modifier_code
			pos := mod_code - Minimum_modifier_code + 1
			table := lookup_table.item (pos)
			if table = Void then
				l_binding := Void
			else
				l_binding := table.item (a_key.code)
			end
			if l_binding /= Void then
				if read_only_mode and not l_binding.is_read_only then
					l_action := read_only_error_action
				else
					l_action := l_binding.action
				end
				if l_action.open_count > 0 then
					l_action.call ([a_key])
				else
					l_action.call ([])
				end
			else
				call_unbound_key_action (a_key)
			end
			if post_call_action /= Void then
				post_call_action.call ([])
			end
		end

feature {NONE} -- Implementation

	call_unbound_key_action (a_key: MODIFIED_KEY)
			-- Call `unbound_key_action' if it is not Void
		require
			a_key_not_void: a_key /= Void
		do
			if unbound_key_action /= Void then
				unbound_key_action.call ([a_key])
			end
		end

	Minimum_key_code: INTEGER = 1
			-- Minimum value for a key code

	Maximum_key_code: INTEGER = 99
			-- Maximum value for a key code

	lookup_table: ARRAY [ARRAY [KEY_BINDING]]
			-- Array that maps modifier code (first array) to
			-- array of actions, which is indexed by key code
			

end
