
indexing
	
	description: "A MOSIS map or diagram"
	author: "David Hollenberg"
	date: "February 24, 2005"

deferred class MOSIS_MAP

feature {NONE} -- Creation

	make_with_date (a_date: DB_DATE)
			-- Create with map date `a_date'
		require
			a_date_exists: a_date /= Void
		do
			date := a_date
		end

feature -- Properties

	date: DB_DATE
			-- Date to include on map

	title: STRING
			-- Title to include on map
		deferred
		end

	title_omitted: BOOLEAN
			-- Should title be omitted from map?

	last_document: DRW_DOCUMENT
			-- Document generated by last call to
			-- `generate_document'?

	drawing_rotatable: BOOLEAN
			-- May drawing be rotated to make it appear
			-- larger on the paper, if appropriate?

feature -- Modification

	set_drawing_rotatable (b: BOOLEAN)
			-- Set `drawing_rotatable' to `b'
		do
			drawing_rotatable := b
		ensure
			value_set: drawing_rotatable = b
		end

	set_title_omitted (b: BOOLEAN)
			-- Set `title_omitted' to `b'
		do
			title_omitted := b
		ensure
			value_set: title_omitted = b
		end

feature -- Operations

	generate_document
			-- Generate document representing map and put
			-- in `last_document'
		do
			initialize
			create_last_document
			generate_whole_document
		ensure
			last_document_exists: last_document /= Void
		end

	write_postscript_file (fname: STRING)
			-- Write PostScript representation of `last_document'
			-- into `fname'
		require
			fname_not_void: fname /= Void
			last_document_not_void: last_document /= Void
		local
			open_write_op: FILE_OPEN_WRITE_OP
			close_op: FILE_CLOSE_OP
			projector: DRW_POSTSCRIPT_PROJECTOR
			f: LARGE_PLAIN_TEXT_FILE
		do
			create open_write_op
			create close_op
			create f.make (fname)
			open_write_op.open (f)
			create projector.make_with_file (f)
			projector.project (last_document)
			close_op.close (f)
		end

feature {NONE} -- Implementation

	generate_whole_document
			-- Generate entire document representing map
			-- and put in `last_document'
		require
			last_document_exists: last_document /= Void
			last_document_empty: last_document.pages.is_empty
		deferred
		end

	generate_page
			-- Generate page and put in `last_page'
		local
			box: DRW_BASIC_RECTANGLE
		do
			generate_drawing
			create last_page.make (1000)
			
			create box.make (available_paper_width, available_drawing_height, 0, paper.paper_height - top_margin - available_drawing_height / 2.0)

			last_drawing.set_view_box (box)
			last_page.extend (last_drawing)
			if not title_omitted then
				generate_title
				create box.make (available_paper_width, title_height, 0, bottom_margin + title_height / 2.0)
				last_title.set_view_box (box)
				last_page.extend (last_title)
			end
			last_page.set_page_label (current_page_label)
		ensure
			last_page_exists: last_page /= Void
		end

	generate_drawing
			-- Generate drawing and put in `last_drawing'
		local
			m: DRW_MATRIX
			max_h: DOUBLE
		do
			initialize
			create last_drawing.make (1000)
			generate_drawing_common_part
			generate_drawing_specific_part
			if needs_rotation (last_drawing, available_paper_width, available_drawing_height) then
				create m.make_rotation_degrees (90.0)
				last_drawing.set_transform (m)
			end
			scale_stroke_widths_for_paper (last_drawing, available_paper_width, available_drawing_height)
			max_h := units_per_paper_unit (last_drawing, available_paper_width, available_drawing_height) * maximum_text_height
			limit_text_view_box_heights (last_drawing, max_h)
		ensure
			last_drawing_exists: last_drawing /= Void
		end

	generate_title
			-- Generate title and put in `last_title'
		do
			initialize
			create last_title.make (20)
			generate_whole_title
			scale_stroke_widths_for_paper (last_title, available_paper_width, title_height)
		ensure
			last_title_exists: last_title /= Void
		end

	generate_whole_title
			-- Generate entire title and put in `last_title'
		require
			last_title_not_void: last_title /= Void
		deferred
		end

	generate_drawing_common_part
			-- Generate part of drawing that is common
			-- to a group of maps
		require
			last_drawing_not_void: last_drawing /= Void
		deferred
		end

	generate_drawing_specific_part
			-- Generate part of drawing that is specific
			-- to a particular map
		require
			last_drawing_not_void: last_drawing /= Void
		deferred
		end

	current_page_label: STRING
			-- Descriptive label for current page of map
			-- (alternative to page number) or Void if none
		deferred
		end

	initialize
			-- Do initialization needed before map generation
		do
			if not is_initialized then
				set_margins
				set_sizes
				set_paper
				set_fonts
				set_stroke_widths
				set_stroke_and_fill_colors
			end
		ensure
			is_initialized: is_initialized
		end

	set_margins
			-- Set margins
		do
			top_margin := 0.5
			bottom_margin := 0.5
			left_margin := 0.75
			right_margin := 0.25
		end

	set_sizes
			-- Set various sizes
		do
			title_drawing_separation := 0.5
			title_logo_width := 1.0
			title_aux_width := 1.5
			maximum_text_height := 0.25
			text_fraction := 0.8
			text_horizontal_fraction := 0.9
			label_fraction := 0.02
			title_fraction := 0.1
		end

	set_paper
			-- Set paper size
		local
			p: DRW_LETTER_PAPER
		do
			create p
			p.set_top_margin (top_margin)
			p.set_bottom_margin (bottom_margin)
			p.set_left_margin (left_margin)
			p.set_right_margin (right_margin)
			paper := p
		end

	set_fonts
			-- Set font used for text size calculations
			-- (not the same font as used in PostScript but
			-- should be a good guess) and PostScript font name
		do
			font_name := "Univers"
		end

	set_stroke_widths
			-- Set stroke values to use for various line
			-- thicknesses (relative numbers only - widths
			-- will be scaled later for each group)
		do
			Absolute_thin_width := 0.005
			thin_width := approximate_drawing_width * Small_scale_factor
			thick_width := 2.0 * thin_width
		end

	set_stroke_and_fill_colors
			-- Set colors used for stroke and fill
		deferred
		end

	approximate_drawing_width: DOUBLE
			-- Approximate width of drawing, in unspecified units.
			-- Used only to determine initial stroke width
			-- for various line thicknesses, so that stroke
			-- width will be a small fraction of drawing width
		deferred
		end

	title_height: DOUBLE
			-- Height of title in inches
		do
			Result := paper.paper_height * title_fraction
		end

	available_paper_width: DOUBLE
			-- Width of paper available for drawing and title
			-- (excludes margins)
		do
			Result := paper.width
		end

	available_drawing_height: DOUBLE
			-- Height of paper available for drawing (excludes
			-- margins, titles and separation of title from
			-- drawing)
		do
			Result := paper.height - title_height - title_drawing_separation
		end

	needs_rotation (g: DRW_GROUP w, h: DOUBLE): BOOLEAN
			-- Should `g' be rotated so that it will
			-- be larger when fit into the area with width `w'
			-- and height `h'?
		require
			group_not_void: g /= Void
		local
			rot_scale_factor, unrot_scale_factor: DOUBLE
			box: DRW_BASIC_RECTANGLE
		do
			if drawing_rotatable then
				box := g.bounding_box
				unrot_scale_factor := scale_factor_to_fit (box.width, box.height, w, h)
				rot_scale_factor := scale_factor_to_fit (box.height, box.width, w, h)
				Result := rot_scale_factor > unrot_scale_factor * Rotation_threshold
			else
				Result := False
			end
		end

	units_per_paper_unit (g: DRW_GROUP w, h: DOUBLE): DOUBLE
			-- Number of units per inch to make `g' fit in
			-- box whose width on paper is `w' and whose
			-- height on paper is `h'
		require
			group_not_void: g /= Void
		local
			box: DRW_BASIC_RECTANGLE
		do
			box := g.bounding_box
			if is_width_limited (box.width, box.height, w, h) then
				Result := box.width / w
			else
				Result := box.height / h
			end
		end

	is_width_limited (w, h: DOUBLE view_width, view_height: DOUBLE): BOOLEAN
			-- When scaling entity with width `w'
			-- and height `h' so that it just barely fits
			-- in area with `view_width' and `view_height',
			-- is the scaling width-limited (as opposed to
			-- height-limited)?
		do
			Result := (view_width / w) < (view_height / h)
		end

	scale_factor_to_fit (w, h: DOUBLE view_width, view_height: DOUBLE): DOUBLE
			-- Scale factor to apply to entity with width `w'
			-- and height `h' so that it just barely fits
			-- in area with `view_width' and `view_height'
		do
			Result := (view_width / w).min (view_height / h)
		end

	stroke_scale_factor (w, h, abs_width, abs_height, ratio: DOUBLE): DOUBLE
			-- Scale factor to apply to stroke width(s) for
			-- figure with width `w' and height `h' to be
			-- mapped to physical area of `abs_width' by
			-- `abs_height', so that stroke widths come out to
			-- desired absolute width on paper.  `ratio' is
			-- the ratio of the relative width of a thin stroke
			-- to its desired absolute width on the paper
		do
			if abs_width / w < abs_height / h then
				-- Width limited
				Result := w / abs_width / ratio
			else
				-- Height limited
				Result := h / abs_height / ratio
			end
		end

	scale_stroke_widths_for_paper (g: DRW_GROUP w, h: DOUBLE)
			-- Scale stroke widths of all figures in `g'
			-- so that they come out properly on `paper'
			-- in area of width `w' and height `h'
		require
			group_not_void: g /= Void
		local
			factor: DOUBLE
			box: DRW_BASIC_RECTANGLE
		do
			box := g.bounding_box
			factor := stroke_scale_factor (box.width, box.height,
				w, h, thin_width / Absolute_thin_width)
			scale_stroke_widths (g, factor)
		end

	scale_stroke_widths (g: DRW_GROUP factor: DOUBLE)
			-- Scale stroke widths of all figures in `g'
			-- by `factor' that do not have stroke
			-- disabled, recursing for figures which are
			-- themselves groups
		require
			group_not_void: g /= Void
			factor_positive: factor > 0
		local
			f: DRW_FIGURE
			shape: DRW_SHAPE
			group: DRW_GROUP
		do
			from
				g.start
			until
				g.after
			loop
				f := g.item
				shape ?= f
				if shape /= Void and then shape.is_drawable and then shape.is_stroked then
					shape.set_stroke_width (shape.stroke_width * factor)
				end
				group ?= f
				if group /= Void then
					scale_stroke_widths (group, factor)
				end
				g.forth
			end
		end

	limit_text_view_box_heights (g: DRW_GROUP h: DOUBLE)
			-- Change to `h' the height of the view box for all
			-- text figures in `g' that have a view box whose
			-- height is greater than `h'
		require
			group_not_void: g /= Void
			height_nonnegative: h >= 0
		local
			f: DRW_FIGURE
			text: DRW_TEXT
			group: DRW_GROUP
			box, new_box: DRW_BASIC_RECTANGLE
		do
			from
				g.start
			until
				g.after
			loop
				f := g.item
				text ?= f
				if text /= Void then
					box := text.view_box
					if box /= Void and then box.height > h then
						create new_box.make (box.width, h, box.x, box.y)
						text.set_view_box (new_box)
					end
				end
				group ?= f
				if group /= Void then
					limit_text_view_box_heights (group, h)
				end
				g.forth
			end
		end

	shrink_box (rect: DRW_BASIC_RECTANGLE)
			-- Resize width and height of `rect' by `text_fraction'
		require
			rect_exists: rect /= Void
		do
			rect.resize (rect.width * text_fraction, rect.height * text_fraction)
		end

	shrink_wide_box (rect: DRW_BASIC_RECTANGLE)
			-- Resize height of `rect' by `text_fraction' and
			-- width by `text_horizontal_fraction'
		require
			rect_exists: rect /= Void
		do
			rect.resize (rect.width * text_horizontal_fraction, rect.height * text_fraction)
		end

	horizontal_section (rect: DRW_BASIC_RECTANGLE start_fraction, end_fraction: DOUBLE): DRW_BASIC_RECTANGLE
			-- Horizontal section of `rect' from `start_fraction'
			-- to `end_fraction' of its width
		require
			rect_exists: rect /= Void
			start_fraction_big_enough: start_fraction >= 0.0
			start_fraction_small_enough: start_fraction <= 1.0
			end_fraction_big_enough: end_fraction >= 0.0
			end_fraction_small_enough: end_fraction <= 1.0
			valid_start_end: start_fraction <= end_fraction
		local
			diff: DOUBLE
		do
			diff := end_fraction - start_fraction
			create Result.make (rect.width * diff, rect.height, rect.left + rect.width * (start_fraction + diff / 2.0), rect.y)
		end

	vertical_section (rect: DRW_BASIC_RECTANGLE start_fraction, end_fraction: DOUBLE): DRW_BASIC_RECTANGLE
			-- Vertical section of `rect' from `start_fraction'
			-- to `end_fraction' of its height
		require
			rect_exists: rect /= Void
			start_fraction_big_enough: start_fraction >= 0.0
			start_fraction_small_enough: start_fraction <= 1.0
			end_fraction_big_enough: end_fraction >= 0.0
			end_fraction_small_enough: end_fraction <= 1.0
			valid_start_end: start_fraction <= end_fraction
		local
			diff: DOUBLE
		do
			diff := end_fraction - start_fraction
			create Result.make (rect.width, rect.height * diff, rect.x, rect.bottom + rect.height * (start_fraction + diff / 2.0))
		end

	nth_horizontal_section (rect: DRW_BASIC_RECTANGLE m, n: INTEGER): DRW_BASIC_RECTANGLE
			-- `m'th rectangle from left of `n' total, when
			-- rect is divided into `n' equal horizontal sections
		require
			rect_exists: rect /= Void
			m_positive: m > 0
			n_positive: n > 0
			m_below_n: m <= n
		do
			create Result.make (rect.width / n, rect.height, rect.left + (1 + 2 * (m - 1)) * (rect.width / n / 2.0), rect.y)
		end

	nth_vertical_section (rect: DRW_BASIC_RECTANGLE m, n: INTEGER): DRW_BASIC_RECTANGLE
			-- `m'th rectangle from bottom of `n' total, when
			-- rect is divided into `n' equal vertical sections
		require
			rect_exists: rect /= Void
			m_positive: m > 0
			n_positive: n > 0
			m_below_n: m <= n
		do
			create Result.make (rect.width, rect.height / n, rect.x, rect.bottom + (1 + 2 * (m - 1)) * (rect.height / n / 2.0))
		end

	text_in_box (s: STRING box: DRW_BASIC_RECTANGLE): DRW_TEXT
			-- Text figure with `s' placed in view box `box'
		require
			string_exists: s /= Void
			box_exists: box /= Void
		do
			create Result.make (s)
			Result.set_size (1)
			Result.set_font (font_name)
			Result.set_view_box (box)
		end

	left_text_in_box (s: STRING box: DRW_BASIC_RECTANGLE): DRW_TEXT
			-- Text figure with `s' placed in view box `box'
			-- and left-justified within box
		require
			string_exists: s /= Void
			box_exists: box /= Void
		do
			Result := text_in_box (s, box)
			Result.set_mid_x (-1)
		end

	right_text_in_box (s: STRING box: DRW_BASIC_RECTANGLE): DRW_TEXT
			-- Text figure with `s' placed in view box `box'
			-- and right-justified within box
		require
			string_exists: s /= Void
			box_exists: box /= Void
		do
			Result := text_in_box (s, box)
			Result.set_mid_x (1)
		end

	rotated_text_in_box (s: STRING a_box: DRW_BASIC_RECTANGLE rot: INTEGER): DRW_GROUP
			-- Text figure with `s' placed nicely in view
			-- box `a_box', rotated `rot' degrees clockwise
		require
			string_exists: s /= Void
			a_box_exists: a_box /= Void
			rot_valid: is_valid_rotation (rot)
		local
			box: DRW_BASIC_RECTANGLE
			old_width, old_height: DOUBLE
			m: DRW_MATRIX
		do
			box := a_box.twin
			if rot = 90 or rot = 270 then
				old_width := box.width
				old_height := box.height
				box.set_width (old_height)
				box.set_height (old_width)
			end
			shrink_box (box)
			create Result.make (1)
			Result.extend (text_in_box (s, box))
			if rot /= 0 then
				create m.make_rotation_degrees_xy (- rot, box.x, box.y)
				Result.set_transform (m)
			end
		end

	is_valid_rotation (rot: INTEGER): BOOLEAN
			-- Is `rot' a valid supported rotation?
		do
			Result := rot = 0 or rot = 90 or rot = 180 or rot = 270
		end

	append_mosis_logo
			-- Generate MOSIS logo and append to
			-- `last_title'
		require
			last_title_not_void: last_title /= Void
		local
			box: DRW_RECTANGLE
			b: DRW_BASIC_RECTANGLE
			logo: DRW_MOSIS_LOGO
			logo_group: DRW_GROUP
		do
			-- Add logo box

			create box.make (title_logo_width, title_height, title_logo_width / 2.0, 0)
			box.set_stroke_width (thick_width)
			last_title.extend (box)
			
			-- Add the logo itself

			create logo
			logo_group := logo.full_color
			b := box.bounding_box.twin
			shrink_box (b)
			logo_group.set_view_box (b)
			last_title.extend (logo_group)
		end

	create_last_document
			-- Create empty document and put in `last_document'
		do
			create last_document.make_sized (paper)
		ensure
			last_document_exists: last_document /= Void
		end

	is_initialized: BOOLEAN
			-- Has initialization needed before map generation
			-- been done?

	font_name: STRING
			-- Name of PostScript font to use

	paper: DRW_PAPER_SIZE
			-- Paper on which document could be printed

	last_page: DRW_PAGE
			-- Page generated by last call to `generate_page'

	last_drawing: DRW_GROUP
			-- Drawing generated by last call to `generate_drawing'

	last_title: DRW_GROUP
			-- Title generated by last call to `generate_title'

	Small_scale_factor: DOUBLE = 1.0E-6
			-- Scale factor to make stroke width very small
			-- in proportion to drawing

	Rotation_threshold: DOUBLE = 1.1
			-- Factor by which scaling for rotated drawing
			-- must exceed scaling for unrotated drawing in
			-- order to choose to rotate the drawing

	Absolute_thin_width: DOUBLE
			-- Desired absolute width of thin stroke on paper,
			-- in inches

	Thin_width: DOUBLE
			-- Relative width of thin stroke

	Thick_width: DOUBLE
			-- Relative width of thick stroke

	top_margin: DOUBLE
			-- Top margin in inches

	bottom_margin: DOUBLE
			-- Bottom margin in inches

	left_margin: DOUBLE
			-- Left margin in inches

	right_margin: DOUBLE
			-- Right margin in inches

	title_drawing_separation: DOUBLE
			-- Space between title and drawing, in inches

	title_logo_width: DOUBLE
			-- Width of box with MOSIS log

	title_aux_width: DOUBLE
			-- Width of box with auxiliary information, in inches

	maximum_text_height: DOUBLE
			-- Maximum height of text in drawing, in inches

	title_fraction: DOUBLE
			-- Fraction of paper height used for title

	text_fraction: DOUBLE
			-- Fraction of box width and height used for text
			-- (number < 1 ensures space between text and
			-- sides of box)

	text_horizontal_fraction: DOUBLE
			-- Fraction of box width used for text for
			-- some wide boxes

	label_fraction: DOUBLE
			-- Fraction of drawing maximum dimension (width or
			-- height) to use for height of label boxes
			-- added outside of main drawing

	Microns_label: STRING = " um"
			-- Label for quantities whose units are microns

	Millimeters_label: STRING = " mm"
			-- Label for quantities whose units are millimeters

invariant
	date_exists: date /= Void
end
