class MOSIS_OVERHEAD_GENERATOR
	-- creator of LAYOUT that represents
	-- the standard MOSIS overhead strip for a die 
	--  and
	-- the body of a freestanding scribe/seal ring

inherit
	MOSIS_SCRIBE_SYNTHESIZER
		export {NONE} all
		redefine the_technology
		end;

	LAYOUT_TO_CIF

	DB_QUERY_OPTIONS
		export {NONE} all
		end

	MOSIS_PROD_VALUES

	SEVERITY_LEVELS
		export {NONE} all
		end

create
	make

feature {NONE} -- creation

	make(run: RUNS_ROW) is
	require not_void: run /= Void
	do  -- save argument
	    the_run := run
	    the_wt := run.wafer_techn_row;
	    set_include_barcode_fabid(not the_wt.foundry_reviews_layout);
	    -- create (empty) caches
	    create cached_scribe_corners.make(5)
	    create cached_dts.make(8);
	    create the_messages.make("NONE");
	    -- create technology-independent layout
	    strip_height := the_wt.overhead_strip_height
	    if the_wt.foundry_reviews_layout then
		die_id_height := strip_height
	    else	-- grow die_id height a bit
		die_id_height := strip_height + 
				run.other_border.min
				(the_wt.top_border - the_wt.die_id_border)
	    end;
	    standard_trademark := build_trademark("MOSIS")
	    tiny_trademark := build_trademark("TinyChip");
	    build_alignment_shapes;
    	end;

feature -- constructor for project (or test vehicle) die overhead

	die_overhead(die: DIES_ROW): LAYOUT is
		-- standard or non-standard die overhead layout
	require is_valid_die: is_valid_die(die);
		is_design_die: die.cust_ic_design /= Void
	local ohs: STRING; b: DB_BIND_S; ostyl: OVERHEAD_STYLE_VALUES
	do  setup_die(die);
	    ohs := die.overhead_style;
	    if not die.cust_ic_design.is_overhead_viable(ohs)
	     then handle_hard_error(die.die_id + ": die overhead not viable")
	    elseif equal(ohs, ostyl.external_)
	     then Result := external_overhead(die)
	    elseif equal(ohs, ostyl.internal)
	     then Result := internal_overhead(die)
	    elseif equal(ohs, ostyl.none)
	     then create {LAYOUT_COLLECTION} Result.make_layered(0)
	     else handle_hard_error(die.die_id + ": unknown overhead style: " + ohs)
	    end
	    -- check for need for thievings in overhead
	    create b.make(the_wt.wafer_techn_family)
	    if connection.selected_count
				("wafer_techn_level_thievings",
				 "wafer_techn_family = :1", b) > 0
	     then handle_hard_error(the_die.die_id + ": Thieving not supported.")
	    end
	end;

feature	-- construct overhead and write directly to cif

	overhead_to_cif(die: DIES_ROW;
			dcf: DATA_CONSUMER) is
		-- generate overhead for die, and write as cif to dcf
		-- errors and warnings are in the_messages
	require is_valid_die: is_valid_die(die);
		is_design_die: die.cust_ic_design /= Void
		have_dcf: dcf /= Void
	local ohead: LAYOUT
	do  ohead := die_overhead(die);
	    layout_to_cif(ohead, the_technology, dcf)
	end;

feature	-- constructor for a freestanding scribe die

	standard_scribe(die: DIES_ROW): LAYOUT is
		-- scribe for die (which must be a scribe)
		-- mbb of result has ll corner at origin
	require is_valid_die: is_valid_die(die);
		is_scribe: equal(die.die_type, die_type_is_scribe)
	local dloc: DIE_LOCATIONS_ROW
	do  setup_die(die);
	    scribe_layers := the_technology.layers_for_scribe;
	    if die.locations.count = 1
	     then dloc := die.locations.first
	    end;
	    Result := scribe_at_loc(die, dloc)
	end;

feature -- construct scribe and write directly to cif

	scribe_to_cif(die: DIES_ROW; dcf: DATA_CONSUMER) is
	require is_valid_die: is_valid_die(die);
		is_scribe: equal(die.die_type, die_type_is_scribe);
		have_dcf: dcf /= Void
	local scribe: LAYOUT
	do  scribe := standard_scribe(die);
	    layout_to_cif(scribe, the_technology, dcf)
	end;

feature	-- constructor for a filler die

	standard_filler(die: DIES_ROW): LAYOUT is
		-- mbb of result has ll corner at origin
	require is_valid_die: is_valid_die(die);
		is_filler: equal(die.die_type, die_type_is_filler)
	do  setup_die(die);
	    create {LAYOUT_COLLECTION} Result.make_layered(0)
	end;

feature -- construct filler and write directly to cif

	filler_to_cif(die: DIES_ROW; dcf: DATA_CONSUMER) is
	require is_valid_die: is_valid_die(die);
		is_filler: equal(die.die_type, die_type_is_filler)
		have_dcf: dcf /= Void
	local filler: LAYOUT
	do  filler := standard_filler(die);
	    layout_to_cif(filler, the_technology, dcf)
	end;

feature -- miscellaneous interface

	the_run: RUNS_ROW;

	the_wt: WAFER_TECHNS_ROW;

	run_id: STRING is
	do  Result := the_run.run_id
	end;

	is_valid_die(die: DIES_ROW): BOOLEAN is
		-- is this a die we can handle?
	do  Result := die /= Void and then 
			equal(die.run_id, run_id)
	end;

	include_barcode_fabid: BOOLEAN
		-- do we include a barcode in the overhead strip?
		-- (default is true)

	set_include_barcode_fabid(ibf: BOOLEAN) is
	do  include_barcode_fabid := ibf
	end;

feature	-- optional design_technology management
	-- (if not used, we just fetch and cache locally)

	the_technology: DESIGN_TECHNOLOGY
		-- the one selected (internally) by the latest call
		-- to generate layout (overhead or scribe)
		-- (cannot be directly set by a client)
 
	available_design_technologies(master: LIST[DESIGN_TECHNOLOGY]) is
		-- a list of potential design_technologies that we can
		-- use rather than fetch from the db
		-- (here, they are just added to the local cache)
	require master /= Void
	do  from  master.start
	    until master.after
	    loop  available_design_technology(master.item)
		  master.forth
	    end
	end;

	available_design_technology(cand:DESIGN_TECHNOLOGY) is
		-- a  potential design_technology that we can
		-- use rather than fetch from the db
		-- (here, it is just added to the local cache)
	require cand /= Void
	do  if find_dt_in_cache(cand.design_techn) = Void
	     then cached_dts.extend(cand)
	    end
	end;

	uncache_design_technologies is
		-- clear cache to cause re-fetching
	do  cached_dts.wipe_out
	end;

feature {NONE} -- the current die and some of its attributes

	setup_die(die: DIES_ROW) is
	do  the_die := die;
	    find_the_technology(die.wafer_design_techn)
	end;

	the_die: DIES_ROW

	all_source_layers: LIST[DESIGN_TECHN_LAYER]
		-- subset of the_technology.layer_set that are
		-- source layers for the_wt
	barred_source_layers: LIST[DESIGN_TECHN_LAYER]
		-- subset of the above with cd_style = 'BAR'
	bond_source_layers: LIST[DESIGN_TECHN_LAYER]
		-- subset with layer_roles.is_bond_layer

	bunch_of_layers(lst: LIST[DESIGN_TECHN_LAYER]): LIST[DESIGN_TECHN_LAYER] is
		-- return lst, unless it is empty
	do  if lst.is_empty
	     then Result := bond_source_layers
	     else Result := lst
	    end
	end;

	connection: DB_CONNECTION is
	do  Result := the_run.connection
	end;

feature {NONE} -- design_technology management

	cached_dts: ARRAYED_LIST[DESIGN_TECHNOLOGY];

	find_the_technology(dt: STRING) is
	do  -- see if the_technology is already correct
	    if the_technology = Void or else
	       not the_technology.is_my_name(dt)
	     then -- nope; dump all this stuff
		  the_technology := Void;
		  all_source_layers := Void
		  registration_mark := Void;
		  cd_and_alignment_strip := Void;
		  cd_only_strip := Void
	    end;
	    -- otherwise, see if we have already cached the one we want
	    if the_technology = Void
	     then the_technology := find_dt_in_cache(dt)
	    end;
	    -- otherwise, get from the db and cache it
	    if the_technology = Void
	     then add_a_new_dt(dt)
	    end;
	    -- and then recompute all the technology-dependent stuff
	    if all_source_layers = Void
	     then set_the_source_layers
	    end;
	    if registration_mark = Void
	     then build_registration_mark
	    end;
	    if cd_and_alignment_strip = Void
	     then build_cd_strips
	    end
	ensure the_technology.is_my_name(dt)
	end;

	find_dt_in_cache(dt: STRING): DESIGN_TECHNOLOGY is
		-- Void if not there
	do  from  cached_dts.start
	    until cached_dts.after or Result /= Void
	    loop  if cached_dts.item.is_my_name(dt)
		   then Result := cached_dts.item
		  end;
		  cached_dts.forth
	    end;
	end;

	add_a_new_dt(dt: STRING) is
	do  create the_technology.make_by_key
				(connection, old_row, dt)
	    cached_dts.extend(the_technology)
	end;

	set_the_source_layers is
		-- build all_source_layers, barred_source_layers, 
		--  bond_source_layers
	local lyrs: LIST[DESIGN_TECHN_LAYER]; lyr: DESIGN_TECHN_LAYER
		r: DESIGN_TECHN_XFORM_LEVELS_ROW;
		dtxls: LIST[DESIGN_TECHN_XFORM_LEVELS_ROW]
	do  -- get list of design_techn_xform_levels
	    create r.make_template;
	    dtxls := r.levels_for_wt_and_dt
				(the_wt.wafer_techn,
				 the_technology.design_techn,
				 connection);
	    -- find subset of the_technology.layer_set that are sources
	    create {ARRAYED_LIST[DESIGN_TECHN_LAYER]}
			all_source_layers.make(dtxls.count);
	    create {ARRAYED_LIST[DESIGN_TECHN_LAYER]}
			barred_source_layers.make(dtxls.count);
	    create {LINKED_LIST[DESIGN_TECHN_LAYER]}
			bond_source_layers.make;
	    lyrs := the_technology.layer_set;
	    from  lyrs.start
	    until lyrs.after
	    loop  lyr := lyrs.item;
		  if is_dtxls_source(lyr.layer_name, dtxls)
		   then all_source_layers.extend(lyr);
			if equal(lyr.cd_style, cd_style_is_bar)
			 then barred_source_layers.extend(lyr)
			end
		  end;
		  if not lyr.is_null_bond_layer_priority
		   then bond_source_layers.extend(lyr)
		  end;
		  lyrs.forth
	    end
	end;

	is_dtxls_source(lname: STRING; 
			dtxls: LIST[DESIGN_TECHN_XFORM_LEVELS_ROW]
		       ): BOOLEAN is
		-- is lname a source_layer_name in dtxls?
	do  from  dtxls.start
	    until dtxls.after or Result
	    loop  Result := equal(lname, dtxls.item.source_layer_name)
		  dtxls.forth
	    end
	end;

feature {NONE} -- cached fragments that we construct on demand
	-- each lies in the first quadrant (ll of bounding box = (0,0))
	-- these first three are technology-dependent (and are therefore
	--  re-created at each technology switch)

	registration_mark: LAYOUT

	cd_and_alignment_strip: LAYOUT_COLLECTION

	cd_only_strip: LAYOUT_COLLECTION

	-- these last five are technology-independent and not layered;
	-- all are constructed once at creation time

	standard_trademark: LAYOUT

	tiny_trademark: LAYOUT

	alignment_target,
	alignment_bullet,
	cd_background: LAYOUT

feature {NONE} -- constructors for the above

	strip_height: DOUBLE	-- the_wt.overhead_strip_height

	die_id_height: DOUBLE	-- height of the die_id (slightly taller)

	scale_to_fixed_height(bdy: LAYOUT; ht: DOUBLE): LAYOUT is
	local bb: CONST_RECTANGLE; wid: DOUBLE;
	do  bb := bdy.bounding_box;
	    wid := bb.size_x * ht / bb.size_y;
	    Result := bdy.nicely_fitting_in(box(0, 0, wid, ht))
	end;

	scale_to_strip_height(bdy: LAYOUT): LAYOUT is
	do  Result := scale_to_fixed_height(bdy, strip_height)
	end;

	reg_mark_bar_width: DOUBLE = 4.0;

	build_registration_mark is
	local hregm: LAYOUT_COLLECTION; lyr: DESIGN_TECHN_LAYER;
		hrm: LAYOUT_RECTANGLE; hlyr: LAYOUT; 
		ht2, mwid, mptch: DOUBLE;
		asl: like all_source_layers;
	do  if the_wt.foundry_reviews_layout then
		asl := bond_source_layers
	    else
		asl := bunch_of_layers(all_source_layers)
	    end;
	    create hregm.make_layered(asl.count);
	    hrm := box_sc(strip_height, reg_mark_bar_width, 0, 0);
				-- the standard layout for all barred layer
	    ht2 := strip_height/2;
	    -- build a horizontal 1/2 reg mark in hregm
	    from  asl.start
	    until asl.after
	    loop  lyr := asl.item
		  if lyr.min_width > reg_mark_bar_width
		   then hlyr := Void	-- skip "fat" layers
		  elseif lyr.is_box_cd_style
		   then -- one row of squares to the right
			mwid := lyr.min_width;
			mptch := lyr.min_pitch;
			hlyr := stepped_copies(
					box_sc(mwid, mwid, 0, 0),
					create {POINT}.make_xy(mptch, 0),
					(ht2/mptch).floor)
			-- and replicated to the left
			hlyr := u2(hlyr, move(hlyr, Rot180))
		  elseif lyr.check_max_size
		   then -- one box of his width
			hlyr := box_sc(strip_height, lyr.min_width, 0, 0)
		   else -- the standard box
			hlyr := hrm
		  end;
		  -- hlyr is horizontal element for lyr
		  if hlyr /= Void then
			hregm.extend(layer(hlyr, lyr))
		  end;
		  asl.forth
	    end;
	    -- and now build a full registration mark
	    registration_mark := 
			slide(u2(hregm, move(hregm, Rot90)),
			      ht2, ht2)
	end;

	cdam_pos, cdonly_pos: INTEGER

	build_cd_strips is
	local lyr: DESIGN_TECHN_LAYER; asl: like all_source_layers;
	do  asl := bunch_of_layers(all_source_layers);
	    -- prepare for construction
	    create cd_only_strip.make_layered(2*asl.count);
	    create cd_and_alignment_strip.make_layered
				(3*asl.count);
	    cdam_pos := 0; cdonly_pos := 0;
	    -- grow them one layer at a time
	    from  asl.start
	    until asl.after
	    loop  lyr := asl.item;
		  add_am_for_layer(lyr);
		  add_cd_for_layer(lyr);
		  add_id_for_layer(lyr);
		  asl.forth
	    end;
	end;

	cd_width: DOUBLE is 20.0;

	add_cd_for_layer(lyr: DESIGN_TECHN_LAYER) is
	local 
		cd, fcd: LAYOUT_COLLECTION; 
		wid, pitch, lowxy, top, right: DOUBLE;
		cnt: INTEGER; obx: LAYOUT_RECTANGLE
		blyrnms: LIST[STRING]; blyr: DESIGN_TECHN_LAYER;
	do  -- build cd figure itself in cd
	    create cd.make_unlayered(6);	-- cd itself
	    wid := lyr.min_width; pitch := lyr.min_pitch;
	    from  cnt := 1; lowxy := 4; top := strip_height - 14;
			    right := cd_width - 4;
	    until cnt > 3 or lowxy > cd_width - 8
	    loop  if lyr.is_box_cd_style
		   then obx := box(lowxy, lowxy, lowxy+wid, lowxy+wid);
			cd.extend(stepped_copies
					(obx,
					 create {POINT}.make_xy(pitch, 0),
					 ((right-lowxy)/pitch).ceiling));
			cd.extend(stepped_copies
					(obx,
					 create {POINT}.make_xy(0, pitch),
					 ((top-lowxy)/pitch).ceiling))
		   else cd.extend(box(lowxy, lowxy, right, lowxy+wid))
			cd.extend(box(lowxy, lowxy, lowxy+wid, top))
		  end;
		  cnt := cnt + 1; lowxy := lowxy + pitch;
			-- left and bottom up by the pitch
		  top := 2*strip_height - (top + 18);
			-- top alternates between ht-14 and ht-4
			-- right stays unchanged
	    end;
	    -- now combine with backgrounds into fcd
	    blyrnms := lyr.cd_background_layers;
	    create fcd.make_layered(blyrnms.count);
	    fcd.extend(layer(cd, lyr));
	    from  blyrnms.start
	    until blyrnms.after
	    loop  blyr := the_technology.layer_by_name(blyrnms.item);
		  fcd.extend(layer(cd_background, blyr));
		  blyrnms.forth
	    end;
	    -- and finally, add fcd to the two global figures
	    cd_only_strip.extend(slide(fcd, cdonly_pos, 0));
	    cdonly_pos := cdonly_pos + cd_width.ceiling;
	    cd_and_alignment_strip.extend(slide(fcd, cdam_pos, 0));
	    cdam_pos := cdam_pos + cd_width.ceiling;
	end;

	add_am_for_layer(lyr: DESIGN_TECHN_LAYER) is
	local alyrnms: LIST[STRING]; alyr: DESIGN_TECHN_LAYER;
		am: LAYOUT
	do  alyrnms := lyr.alignment_target_layers;
	    from  alyrnms.start
	    until alyrnms.after
	    loop  alyr := the_technology.layer_by_name(alyrnms.item);
		  am := u2(layer(alignment_target, alyr),
			   layer(alignment_bullet, lyr));
		  cd_and_alignment_strip.extend(slide(am, cdam_pos, 0));
		  cdam_pos := cdam_pos + 2 * strip_height.ceiling;
		  alyrnms.forth
	    end;
	end;

	add_id_for_layer(lyr: DESIGN_TECHN_LAYER) is
	local idc: INTEGER; idl: LAYOUT; ids: INTEGER; 
		idlyr: DESIGN_TECHN_LAYER; blyrnms: LIST[STRING];
	do  idc := lyr.layer_id.code
	    if lyr.is_box_cd_style or lyr.check_max_size
	     then -- put ID on a background layer
		  blyrnms := lyr.cd_background_layers
		  if not blyrnms.is_empty
		   then idlyr := the_technology.layer_by_name
						(blyrnms.first)
		   else -- forget it
		  end
	     else idlyr := lyr		-- put ID on self
	    end;
	    if simple_font.is_graphic_code(idc) and idlyr /= Void
	     then idl := scale_to_strip_height(simple_font.depict(idc));
		  idl := layer(idl, idlyr);
		  ids := idl.bounding_box.size_x.ceiling + 8;
		  cd_only_strip.extend(slide(idl, cdonly_pos+4, 0));
		  cdonly_pos := cdonly_pos + ids;
		  cd_and_alignment_strip.extend(slide(idl, cdam_pos+4, 0));
		  cdam_pos := cdam_pos + ids;
	    end
	end;

	build_trademark(txt: STRING): LAYOUT is
	local bdy, tm: LAYOUT
	do  
	  if the_wt.foundry_reviews_layout then
	    Result := scale_to_strip_height(text(txt + " (TM)"))
	  else
	    bdy := scale_to_strip_height(text(txt));
	    tm  := scale_to_fixed_height(text("TM"), strip_height * 0.6)
	    Result := u2(bdy,
			 slide(tm, bdy.bounding_box.high_x+5,
				   bdy.bounding_box.high_y -
					tm.bounding_box.high_y))
	  end
	end;

	build_alignment_shapes is
		-- 2 ht x ht alignment pairs with opposing bullet polarity
		-- to use, layer(alignment_bullet, <this_layer>)
		--  and	   layer(alignment_target, <target_layer>)
		--  and merge
	local ht: DOUBLE; t1, ab: LAYOUT_COLLECTION;
	do  ht := strip_height;		-- width is 2 x strip_height
	    ab :=  u4(  box(   0,   0,  ht,   7),
			box(   0,   0,   7,  ht),
			box(ht-7,   0,  ht,  ht),
			box(   0,ht-7,  ht,  ht));	-- hollow
	    ab.extend(  box(ht+7,   7,2*ht-7,ht-7));	-- solid
	    t1 :=    u4(box(  6,  6, 10, 10),		-- 4 castles
			box( ht-10,  6, ht-6, 10),
			box(  6, ht-10, 10, ht-6),
			box( ht-10, ht-10, ht-6, ht-6));
	    t1.extend(  box(  8,  8, ht-8, ht-8));	-- around target
	    alignment_bullet := ab;
	    alignment_target := u2(t1, slide(t1, ht, 0));
	    cd_background := box( 0, 0, cd_width, ht);
	end;

	barcode_pixel_width: DOUBLE is 3.0;	-- 3 microns per bar

	barcode_gen: BARCODE_GENERATOR;

	barcode_scale: MATRIX;

	build_barcode(txt: STRING): LAYOUT is
	do  if barcode_gen = Void
	     then create barcode_gen.make(Void);	-- use default class
		  barcode_gen.set_with_frame(true);
		  barcode_gen.set_pixel_height
				(strip_height/barcode_pixel_width);
		  barcode_scale := magnify_by(barcode_pixel_width)
	    end;
	    Result := move(barcode_gen.barcode(txt), barcode_scale)
	end;

feature {NONE} -- the overhead constructor

	external_overhead(die: DIES_ROW): LAYOUT is
		-- overhead strip to go at the top of the project
		-- placed in the same coordinate system as the_cid
	require is_valid_die: is_valid_die(die);
		is_design_die: die.cust_ic_design /= Void
	local	cid: CUST_IC_DESIGNS_ROW; wid, chmfr: DOUBLE;
		recon: BOOLEAN; mbb: RECTANGLE;
	do  cid := the_die.cust_ic_design;
	    wid := cid.design_size_x - 2*the_wt.top_border;
	    -- if we are in a technology requiring CHIPEDGE and PROTECT
	    -- reconstruction, shrink width by one chamfer per side 
	    recon := cid.is_chamfer_reconstruction_required
					(die.overhead_style);
	    if recon then
		if cid.is_null_chamfer_edge then
			handle_hard_error
				("cannot handle external overhead for "
				 + die.die_id)
		end;
		chmfr := cid.chamfer_edge;
		wid := wid - 2 * (chmfr + 2);
	    end;
	    -- build strip (with origin at ll)
	    Result := build_strip_of_width(wid, cid.tiny_chip);
	    -- move into place above the project
	    Result := slide(Result, 
			    cid.center_x - wid/2,
			    cid.center_y + cid.design_size_y/2 +
					the_wt.design_strip_separation);
	    if recon then
		-- add 4 new corner chamfers plus a chamfered chipedge
		mbb := cid.design_bounding_box;
		mbb.extend_to_include(Result.bounding_box);
		Result := u2(Result, 
			     new_chamfer_layout(cid.chamfer_layer, chmfr, mbb))
	    end;
	end;

	build_strip_of_width(wid: DOUBLE;
				is_tiny: BOOLEAN): LAYOUT is
		-- 
	local ohead: LAYOUT_COLLECTION; item: LAYOUT; lx, hx, mx: INTEGER;
	do  create ohead.make_layered(7);
	    hx := wid.floor;
	    -- first item is die_id, at far right
	    item := multi_layer
			(scale_to_fixed_height
				(text(the_die.die_id), die_id_height),
			 bond_source_layers);
	    hx := (hx - item.bounding_box.size_x).floor;
	    ohead.extend(slide(item, hx, 0));
	    hx := hx - 20;	-- spacer (enlarged per RW)
	    -- second item is run_id, also at the right
	    item := multi_layer
			(scale_to_strip_height
				(text(the_run.run_id_for_overhead)),
			 bond_source_layers);
	    hx := hx - item.bounding_box.size_x.ceiling;
	    ohead.extend(slide(item, hx, 0));
	    hx := hx - 20;       -- spacer
	    -- third item is (optional) fabid barcode, also at the right
	    if include_barcode_fabid
	     then item := multi_layer
				(build_barcode(the_run.run_id_for_overhead
						+ the_die.die_id),
				 bond_source_layers);
		  hx := hx - (item.bounding_box.size_x.ceiling + 15);
		  ohead.extend(slide(item, hx, 0));
		  hx := hx - 20;       -- spacer (20 um each side)
	    end;
	    -- fourth item is registration mark, at far left
	    item := registration_mark
	    ohead.extend(item);
	    lx := item.bounding_box.size_x.ceiling + 5;
	    -- fifth item is cd figure, also at the left
	    if the_wt.foundry_reviews_layout
	     then item := Void
	    elseif (hx - lx) >= cd_and_alignment_strip.bounding_box.size_x
	     then item := cd_and_alignment_strip
	    elseif (hx - lx) >= cd_only_strip.bounding_box.size_x
	     then item := cd_only_strip
	     else item := Void
	    end;
	    if item /= Void
	     then ohead.extend(slide(item, lx, 0));
		  lx := lx + item.bounding_box.size_x.ceiling + 5
	    end;
	    -- sixth item is trademark, in center of what remains
	    if is_tiny
	     then item := tiny_trademark
	     else item := standard_trademark
	    end;
	    item := multi_layer(item, bond_source_layers);
	    if (hx - lx) >= item.bounding_box.size_x
	     then mx := hx + lx - item.bounding_box.size_x.floor;
		  ohead.extend(slide(item, mx/2, 0))
	    end;
	    -- cough it up
	    Result := ohead
	end;

	new_chamfer_layout(plyr: STRING; chmfr: DOUBLE; mbb: CONST_RECTANGLE): LAYOUT
		-- note: chmfr might be zero!!
	local clyr: STRING; lyrs: LAYER_NAME_VALUES
		crnr: CONST_POINT; 
		chipedge: LAYOUT_POLYGON; tris: LAYOUT_COLLECTION
	do  clyr := lyrs.chipedge
	    create tris.make_unlayered(4)
	    create chipedge.make(8)
	    -- now go around the four corners, building triangles and chipedge
	    crnr := mbb.ll; 
	    tris.extend(triangle(crnr.x, crnr.y, crnr.x, crnr.y+chmfr,
				 crnr.x+chmfr, crnr.y));
	    chipedge.extend_xy(crnr.x, crnr.y+chmfr);
	    chipedge.extend_if_different_xy(crnr.x+chmfr, crnr.y);
	    crnr := mbb.lr;
	    tris.extend(triangle(crnr.x, crnr.y, crnr.x-chmfr, crnr.y,
				 crnr.x, crnr.y+chmfr));
	    chipedge.extend_if_different_xy(crnr.x-chmfr, crnr.y);
	    chipedge.extend_if_different_xy(crnr.x, crnr.y+chmfr);
	    crnr := mbb.ur; 
	    tris.extend(triangle(crnr.x, crnr.y, crnr.x, crnr.y-chmfr,
				 crnr.x-chmfr, crnr.y));
	    chipedge.extend_if_different_xy(crnr.x, crnr.y-chmfr);
	    chipedge.extend_if_different_xy(crnr.x-chmfr, crnr.y);
	    crnr := mbb.ul;
	    tris.extend(triangle(crnr.x, crnr.y, crnr.x+chmfr, crnr.y,
				 crnr.x, crnr.y-chmfr));
	    chipedge.extend_if_different_xy(crnr.x+chmfr, crnr.y);
	    chipedge.extend_if_different_xy(crnr.x, crnr.y-chmfr);
	    -- combine (unless chmfr = 0) and return
	    Result := create {LAYOUT_LAYERING}.make(chipedge, clyr);
	    if chmfr > 0 then
		Result := u2(Result,
			     create {LAYOUT_LAYERING}.make(tris, plyr))
	    end;
	end;

feature {NONE} -- internal overhead constructor

	internal_overhead(die: DIES_ROW): LAYOUT is
		-- overhead is just the fab id placed in the label box
	require is_valid_die: is_valid_die(die);
		is_design_die: die.cust_ic_design /= Void
		has_label_box: not die.cust_ic_design.is_null_label_box_size_x
	local	the_cid: CUST_IC_DESIGNS_ROW;
		runid, dieid: LAYOUT; labelbox: CONST_RECTANGLE
		mtrx: MATRIX;
	do  the_cid := die.cust_ic_design;
	    create labelbox.make_by_size_and_center
				(the_cid.label_box_size_x,
				 the_cid.label_box_size_y,
				 the_cid.label_box_center_x,
				 the_cid.label_box_center_y)
	    -- die_id (full size) over run_id (possibly scaled down) 
	    dieid := scale_to_strip_height
				(text(the_die.die_id));
	    runid := scale_to_strip_height
				(text(the_run.run_id_for_overhead));
	    if runid = Void then
		-- no run_id; just die_id
		Result := dieid
	    elseif runid.bounding_box.size_x > labelbox.size_x-2 then
		-- small logobnd area;
		-- shrink run_id and under die_id
		create mtrx.make_magnification(0.5);
		runid := move(runid, mtrx);
		dieid := slide(dieid, 0, (strip_height + 5)/2)
		Result := u2(dieid, runid)
	    else
		-- big box case
		dieid := slide(dieid, 0, strip_height + 5);
		Result := u2(dieid, runid)
	    end;
	    Result := multi_layer(Result, bond_source_layers);
	    -- place in ll of label box & cough it up
	    Result := slide(Result, labelbox.low_x+0.5, labelbox.low_y+0.5)
	end;

feature {NONE} -- the scribe constructors
	-- if dloc is Void, just build a simple structure without
	--	consideration of abutting scribes
	-- if dloc is not Void, look for other scribes that abut this
	--	one and create corners to meet them

	scribe_layers: LIST[DESIGN_TECHN_LAYER];

	scribe_at_loc(die: DIES_ROW;
			dloc: DIE_LOCATIONS_ROW): LAYOUT is
	local vert: BOOLEAN; len, hwid: DOUBLE; 
		top_int, bot_int: LIST[DOUBLE]; top, bot: LAYOUT
	do  vert := die.is_vertical;
	    if vert
	     then len := die.die_size_printed_y;
		  hwid := die.die_size_printed_x / 2
	     else len := die.die_size_printed_x;
		  hwid := die.die_size_printed_y / 2
	    end;
	    if dloc /= Void
	     then top_int := abutment_points
				(die.run_id, dloc.mask_site_no, vert, 1);
		  bot_int := abutment_points
				(die.run_id, dloc.mask_site_no, vert, -1)
	    end;
	    if dloc = Void or else (top_int.count = 0 and bot_int.count = 0)
	     then top := scribe_segment(0, len, hwid);
		  bot := top;
	     else top := abutted_half_scribe(len, hwid, top_int);
		  bot := abutted_half_scribe(len, hwid, bot_int);
	    end;
	    Result := u2(top, move(bot, Mirror0));
	    Result := slide(Result, 0, hwid);	-- origin to lower left
	    if vert
	     then Result := move(Result, Mirror45)
	    end
	end;

feature {NONE} -- finding the abutting scribe intersections

	abutment_points(runid: STRING; siteno: INTEGER; vert: BOOLEAN;
			side: INTEGER): LIST[DOUBLE] is
	-- runid and siteno define a scribe (which is vert[ical] or not)
	-- find the abutment points on the +1 or -1 side
	-- returned values are between 0 and the scribe length
	require valid_side: side.abs = 1
	local b: DB_BIND_SII;
	do  create b.make(runid, siteno, side)
	    if vert
	     then Result := connection.selected_doubles
					(vert_select, abutment_tables,
					 vert_filter, "1", false, b)
	     else Result := connection.selected_doubles
					(horz_select, abutment_tables,
					 horz_filter, "1", false, b)
	    end
	end;

	vert_select: STRING is
	-- relative location from bottom of (a vertical) scribe
	"tloc.location_y-(sloc.location_y-(sd.die_size_printed_y/2))"

	horz_select: STRING is
	-- relative location from left end of (a horizontal) scribe
	"tloc.location_x-(sloc.location_x-(sd.die_size_printed_x/2))"

	abutment_tables: STRING is
	"die_locations tloc, dies td, dies sd, die_locations sloc"

	vert_filter: STRING is
	once Result :=
	"abs(tloc.location_x - (sloc.location_x + %
	% (:3)*(td.die_size_printed_x+sd.die_size_printed_x)/2)) < 0.0001 and %
	%tloc.location_y between %
	%  sloc.location_y-(sd.die_size_printed_y/2) and %
	%  sloc.location_y+(sd.die_size_printed_y/2) and %
	%td.scribe_orientation = 'HORIZONTAL' and " + common_filter
	end;

	horz_filter: STRING is
	once Result :=
	"abs(tloc.location_y - (sloc.location_y + %
	% (:3)*(td.die_size_printed_y+sd.die_size_printed_y)/2)) < 0.0001 and %
	%tloc.location_x between %
	%  sloc.location_x-(sd.die_size_printed_x/2) and %
	%  sloc.location_x+(sd.die_size_printed_x/2) and %
	%td.scribe_orientation = 'VERTICAL' and " + common_filter
	end;

	common_filter: STRING is
		-- all scribes and locations in same run and field
	"tloc.run_id = td.run_id and %
	%tloc.die_id = td.die_id and %
	%tloc.field_name = sloc.field_name and %
	%td.run_id = sd.run_id and %
	%td.die_type = 'SCRIBE' and %
	%sd.run_id = sloc.run_id and %
	%sd.die_id = sloc.die_id and %
	%sloc.run_id = :1 and %
	%sloc.mask_site_no = :2"

feature	-- scribe constructor

	abutted_half_scribe(len, hwid: DOUBLE;
				crnr_locs: LIST[DOUBLE]): LAYOUT is
		-- create a (horizontal) half-scribe
		-- Result fits in (0,0) to (len, hwid)
		-- the scribe consists of scribe_corners at each of the
		--  crnr_loc positions, connected by plain scribe segments
	-- require each element of crnr_locs is between 0 and len;
	--	   crnr_locs is sorted in ascending order;
	local pcs: LAYOUT_COLLECTION; xcor, xlo, xhi: DOUBLE
	do  create pcs.make_layered(2*crnr_locs.count+1);
	    from  xlo := 0;
		  crnr_locs.start
	    until crnr_locs.after
	    loop  xcor := crnr_locs.item;
		  xhi := xcor - hwid
		  -- the plain segment to the left of xcor
		  if xhi > xlo
		   then pcs.extend(slide(scribe_segment(0, xhi-xlo, hwid),
					 xlo, 0));
		  end;
		  -- and the corner at xcor
		  pcs.extend(slide(scribe_corner(hwid), xcor, 0));
		  crnr_locs.forth;
		  xlo := xcor + hwid;
	    end;
	    -- the final segment to the right of the last xcor
	    if len > xlo
	     then pcs.extend(slide(scribe_segment(0, len-xlo, hwid),
				   xlo, 0));
	    end;
	    Result := pcs
	end;

feature {NONE} -- half-scribe corner constructor and manager

	scribe_corner(hwid: DOUBLE): LAYOUT is
		-- returns layout that represents a horizontal half-scribe's
		-- abutment to a vertical full scribe from above
		-- origin is in the middle of the intersection
		-- bounding box is (-hwid, 0) to (hwid, hwid)
	local qrtr: LAYOUT;
	do  Result := cached_scribe_corners.item(hwid);
	    if Result = Void
	     then -- build a new corner segment of size hwid x hwid
		  qrtr := scribe_segment(1, hwid, hwid)
		  -- and make 4 copies; result is 2*hwid x hwid
		  Result := u4(	qrtr,			-- +x
				move(qrtr, Mirror45),	-- +y on right
				move(qrtr, Rot90),	-- +y on left
				move(qrtr, Mirror90))	-- -x
		  -- and save it for possible reuse
		  cached_scribe_corners.put(Result, hwid)	-- and cache
	    end
	end;

	cached_scribe_corners: HASH_TABLE[LAYOUT, DOUBLE]
		-- this table is used solely by scribe_corner, above

end -- MOSIS_OVERHEAD_GENERATOR
