class MOSIS_RUN

inherit
	MOSIS_PROD_VALUES

	MOSIS_MPRO_VALUES

	DB_ACCESS

	DB_QUERY_OPTIONS

	DB_UTILITIES

	OS_ACCESS

	STRING_UTILITIES_ACCESS

create
	make

feature -- attributes

	run_id: STRING

feature {RUN_STEP_PATTERN} -- almost non-shared attributes

	run: RUNS_MROW

feature {NONE} -- non-shared attributes

	my_binder: DB_BIND_S

	my_filter: STRING = "run_id = :1"

	productions: LIST[PRODUCTIONS_ROW]

	mask_sets: LIST[MASK_SETS_ROW]

	events: LIST[RUN_EVENTS_MROW]

	wfr_lots: LIST[RUN_WAFER_LOTS_ROW]
	do  Result := run.wafer_lots
	end

	pkg_lots: LIST[RUN_PACKAGING_LOTS_ROW]
	do  Result := run.packaging_lots
	end

	my_closer: M_OPERATORS_ROW
		-- cache for closer

feature -- basic runs / productions / mask_sets table info

	exists: BOOLEAN
		-- does the run exist in the database?
	do  Result := not run.is_empty
	end

	wafer_techn: STRING
	do  Result := run.wafer_techn
	end

	mask_techn: STRING
	do  Result := mask_sets.first.mask_techn
	end

	field_set_name: STRING
	do  Result := run.field_set_name
	end

	wafer_part_id: STRING
	do  Result := productions.first.wafer_part_id
	end

feature	-- temporary support of obsolete feature

	run_charges_locked: BOOLEAN
	do  Result := productions.first.production_charges_locked
	end

feature --

	closing_is_complete: BOOLEAN
		-- have we completed run closing (and gone into fab)?
	local step_name: PRIMITIVE_STEP_NAME_VALUES
	      b: DB_BIND_SS
	      cdat: DB_DATE
	do  create b.make(run_id, step_name.Complete_run_assembly)
	    cdat := db.safe_selected_date
			("date_completed", "run_primitive_steps",
			 "run_id = :1 and primitive_step_name= :2", b)
	    Result := is_date(cdat)
	end

	run_directory_root: STRING
		-- directory containing all the run directories
	local scntxt: SYS_CONTEXT_VALUES; b: DB_BIND_SS; dbnam: STRING
	once
	    create b.make(scntxt.userenv, scntxt.db_name);
	    dbnam := db.selected_string
				("sys_context(:1, :2)", "dual", Void, b)
	    if equal(dbnam, "MPRD") then
		Result := "/nfs/m_runs/runs"
	    else
		Result := "/nfs/m_data1/dev_runs"
	    end;
	end;

	run_directory: STRING
		-- run directory name 
	do  
	    Result := os.full_directory_name
				(run_directory_root, run_id.as_lower)
	end

	status_location: STRING
	do  Result := db.safe_selected_string(
				"m_lookup.status_location(run_id)",
				"runs", my_filter, my_binder)
	end

	is_dedicated: BOOLEAN
	do  Result := equal(run.run_type, rtype_is_dedicated)
	end

	is_development: BOOLEAN
	do  Result := equal(run.run_type, rtype_is_development)
	end

	closer: M_OPERATORS_ROW
		-- the run closer
	local nam: STRING
	do  nam := run.closer	-- this value is official - verify the cache
	    if su.is_empty(nam)
	     then my_closer := Void
	    elseif my_closer = Void or else not equal(my_closer.username, nam)
	     then create my_closer.make_by_key(db, old_row, nam)
	    end
	    Result := my_closer
	end

	run_cust_acct_nos: LIST[INTEGER]
		  -- for dedicated runs, the cust_acct_no(s)
		  -- paying the FAB RUN charges
	local b: DB_BIND_SS
	do  create b.make(run_id, part_is_fab_run)
	    Result := db.selected_integers
			("cwo.cust_acct_no",
			 "cust_wos cwo, cust_wo_line_items woli, m_parts mp",
			 "cwo.m_order_no = woli.m_order_no and%
			 % woli.m_part_no = mp.m_part_no and%
			 % cwo.production_id = :1 and mp.part_type = :2",
			 Void,	-- order_by clause
			 true,	-- select distinct
			 b)
	end

	projects_of_type(typ: STRING): INTEGER
		-- typ is type_is_customer_fab or type_is_test_vehicle
	local b: DB_BIND_SS
	do  create b.make(run_id, typ)
	    Result := db.selected_count("ic_fab_instances",
			"run_id = :1 and ic_fab_instance_type = :2", b)
	end

	customer_fabs: INTEGER
	do  Result := projects_of_type(type_is_customer_fab)
	end

	test_vehicles: INTEGER
	do  Result := projects_of_type(type_is_test_vehicle)
	end

	overhead_id: STRING
	do  Result := run.overhead_id
	end;

feature	-- the wafer_process and/or packaging_order_type restrictors.
	-- if these are not_empty, only conforming lots are used by
	-- the date setters (each is treated independently)

	curr_wafer_process: STRING
	curr_packaging_order_type: STRING

	set_curr_process_and_type(wp, pot: STRING)
	do  curr_wafer_process := wp
	    curr_packaging_order_type := pot
	end

feature -- the results of the last date setter call

	last_date: DB_DATE
	last_date_is_est: BOOLEAN

	last_date_has_passed: BOOLEAN
		-- has last_date gone by (est or hard)
	do  Result := 
		last_date /= Void and then not last_date.is_null and then
		last_date.truncated_to_midnight <= db.transaction_date
	end

	last_date_has_happened: BOOLEAN
		-- has last_date really happened (no if estimate)
	do  Result := last_date_has_passed and not last_date_is_est
	end

feature -- the various date setters
	-- each sets last_date (Void if null in database) and last_date_is_est

	tapes_out
	do  set_ld_1(mask_sets.first.date_tapeout)
	end

	masks_out
	do  set_ld_1(mask_sets.first.date_masks_out)
	end

	wafers_in
	local wlot, best: RUN_WAFER_LOTS_ROW
	do  from  wfr_lots.start
	    until wfr_lots.after
	    loop  wlot := wfr_lots.item
		  -- find earliest date_received
		  if (su.not_empty(curr_wafer_process) implies
		      equal(curr_wafer_process, wlot.wafer_process)) and
		     (best = Void or else 
		      best.date_received > wlot.date_received)
		   then best := wlot
		  end
		  wfr_lots.forth
	    end
	    if best = Void
	     then set_ld_1(Void)
	     else set_ld_2(best.date_received, 
			   best.is_null_quantity_wafers_received)
	    end
	end

	wafers_out
	local plot, best: RUN_PACKAGING_LOTS_ROW
	do  from  pkg_lots.start
	    until pkg_lots.after
	    loop  plot := pkg_lots.item
		  if pkg_lot_OK(plot) and
		     (best = Void or else 
		      best.date_sent > plot.date_sent)
		   then best := plot
		  end
		  pkg_lots.forth
	    end
	    if best = Void
	     then set_ld_1(Void)
	     else set_ld_2(best.date_sent, not best.is_actual_date_sent)
	    end
	end

	pkgs_out
	local plot, best: RUN_PACKAGING_LOTS_ROW
	do  from  pkg_lots.start
	    until pkg_lots.after
	    loop  plot := pkg_lots.item
		  if pkg_lot_OK(plot) and
		     (best = Void or else 
		      best.date_received > plot.date_received)
		   then best := plot
		  end
		  pkg_lots.forth
	    end
	    if best = Void
	     then set_ld_1(Void)
	     else set_ld_2(best.date_received, 
				not best.is_actual_date_received)
	    end
	end

	parts_out
		-- parts out to user (wafers or packages)
	do  if any_pkg_lots_OK
	     then pkgs_out
	     else wafers_out
	    end
	end

feature {NONE} -- details

	set_ld_1(d: DB_DATE)
		-- for those cases that lack an est(imated) flag
	do  last_date := d
	    last_date_is_est := not last_date_has_passed
	end

	set_ld_2(d: DB_DATE e: BOOLEAN)
		-- for those cases that have an est(imated) flag
	do  last_date := d
	    last_date_is_est := e
	end

	pkg_lot_OK(pkg: RUN_PACKAGING_LOTS_ROW): BOOLEAN
	-- does this lot pass the curr_wafer_process/curr_order_type test
	do  Result :=
	    (su.not_empty(curr_wafer_process) implies
	     equal(curr_wafer_process, pkg.wafer_process)) and
	    (su.not_empty(curr_packaging_order_type) implies
	     equal(curr_packaging_order_type, pkg.packaging_order_type))
	end

	any_pkg_lots_OK: BOOLEAN
	do  from  pkg_lots.start
	    until pkg_lots.after or Result
	    loop  Result := pkg_lot_OK(pkg_lots.item)
		  pkg_lots.forth
	    end
	end

feature -- other attributes

	is_closed: BOOLEAN
		-- has the run been closed
	do  get_run_event(event_is_closed)
	    Result := last_date_has_happened
	end

	is_eccn_reported: BOOLEAN
		-- have we sent a ECCN report to the fabricator
	do  get_run_event(event_is_eccn_reported)
	    Result := last_date_has_happened
	end

	is_posted: BOOLEAN
		-- has the run been posted (fab id's sent)?
	do  get_run_event(event_is_posted)
	    Result := last_date_has_happened
	end

	is_released_to_manufacturing: BOOLEAN
		-- has the vendor (implicitly or explicitly) accepted the run
	do  get_run_event(event_is_rtm)
	    Result := last_date_has_happened
	end

	is_shipped: BOOLEAN
	do  parts_out
	    Result := last_date_has_happened
	end

	current_stage: STRING
		-- description of where we are in the fab cycle
	do  wafers_out
	    if last_date_has_passed
	     then Result := "packaging"
	     else wafers_in
		  if last_date_has_passed
		   then Result := "wafer testing"
		   else masks_out
			if last_date_has_passed
			 then Result := "wafer fabrication"
			 else tapes_out
			      if last_date_has_passed
			       then Result := "mask generation"
			       else Result := "data preparation"
			      end
			end
		  end
	    end
	end

feature -- run_events access

	last_event: RUN_EVENTS_MROW

	get_run_event(enam: STRING)
		-- leave in last_event; Void if no entry for enam
		-- also leave its data in last_date, last_date_is_est
	local row: like last_event
	do  if last_event = Void or else
		not last_event.event_name.is_equal(enam)
	     then  from  events.start
			 last_event := Void
		   until events.after or last_event /= Void
		   loop  row := events.item
			 if row.event_name.is_equal(enam)
			  then last_event := row end
			 events.forth
		   end
	    end
	    if last_event = Void
	     then last_date := Void
		  last_date_is_est := true
	     else last_date := last_event.date_event
		  last_date_is_est := not last_event.actual_date
	    end
	end

	add_new_event(ev: RUN_EVENTS_MROW)
	require exists: ev /= Void 
		my_run: ev.run_id = run_id
	do  events.extend(ev)
	end

feature -- creation (and modification)

	make(rid: STRING)
	require rid /= Void
	local 	pr: PRODUCTIONS_ROW 
		ms: MASK_SETS_ROW 
		re: RUN_EVENTS_MROW
	do  run_id := rid
	    create run.make_by_key(db, old_no_exception_row, rid)
	    create my_binder.make(rid)
	    last_event := Void
	    last_date := Void
	    -- get my productions
	    create pr.make_template
	    productions := pr.all_rows_filtered_ordered
				(db, my_filter, my_binder,
				 "date_production_closed,production_id")
	    -- get my mask sets
	    create ms.make_template
	    mask_sets := ms.all_rows_filtered_ordered
				(db, my_filter, my_binder, "mask_revision")
	    -- get my events
	    create re.make_template
	    events := re.all_rows_filtered_ordered
				(db, my_filter, my_binder, "date_event")
	end

end
