deferred class MOSIS_SCRIBE_SYNTHESIZER
	-- the official synthesizer of scribe-like layout from the
	-- data in design_techn_layers (near_edge, far_edge, scribe_style)

inherit

	LAYOUT_SUPPORT

	MOSIS_TECH_VALUES

feature	-- the scribe layer list (which must be constructed for us)

	scribe_layers: LIST[DESIGN_TECHN_LAYER] is
	deferred
	end;

feature {NONE} -- half-scribe segment constructor
	-- segments are for a horizontal half-scribe that extends from the
	-- x-axis (center of kerf) upward

	scribe_segment(mnear: INTEGER; len, ht: DOUBLE): LAYOUT is
		-- horizontal half-scribe of length len, 
		-- fitting in (0,0) to (len, ht)
		-- (ht may be less that the implied scribe 1/2 width, in 
		--  which case it is used to omit or truncate layers)
		-- mnear is 1 for a corner segment
		--	(knock off a little at the left)
		--   and 0 for a linear segment
	require valid_mnear: mnear = 0 or mnear = 1
	local segm: LAYOUT_COLLECTION; lyr: DESIGN_TECHN_LAYER;
		seg1: LAYOUT; 
	do  from  create segm.make_layered(scribe_layers.count);
		  scribe_layers.start
	    until scribe_layers.after
	    loop  lyr := scribe_layers.item;
		  seg1 := scribe_seg_layer(lyr, mnear, len, ht)
		  if seg1 /= Void
		   then segm.extend(layer(seg1, lyr))
		  end;
		  scribe_layers.forth
	    end;
	    Result := segm
	end;

	scribe_seg_layer(lyr: DESIGN_TECHN_LAYER;
			 mnear: INTEGER; len, ht: DOUBLE): LAYOUT is
		-- contribution to the segment for lyr
	local near: DOUBLE;
	do  near := lyr.near_edge
	    if near + lyr.min_width > ht -- don't have room for minimal layout
	     then	-- return Void
	    elseif equal(lyr.scribe_style, scribe_is_trench2)
	     then -- customized scribe_style
		  Result := trenched_segment(lyr, mnear, len, ht, 2)
	    elseif equal(lyr.scribe_style, scribe_is_trench3)
	     then -- customized scribe_style
		  Result := trenched_segment(lyr, mnear, len, ht, 3)
	    elseif equal(lyr.scribe_style, scribe_is_boxes)
	     then -- row of boxes
		  Result := boxed_segment(lyr, mnear, len, ht)
	     else -- normal single bar case
		  Result := box(near*mnear, near,
				len, lyr.far_edge.min(ht))
	    end
	end;

	boxed_segment(lyr: DESIGN_TECHN_LAYER;
			mnear: INTEGER; len, ht: DOUBLE): LAYOUT is
		-- lyr.scribe_style is BOXES (or assumed boxes)
		-- scribe layout is row of min_width squares
	local near, far, mwid, ptch: DOUBLE;
		obox: LAYOUT_RECTANGLE; cnt: INTEGER; step: CONST_POINT
	do  near := lyr.near_edge
	    far := lyr.far_edge;
	    mwid := lyr.min_width;
	    if mnear = 1
	     then -- single box in the corner
		  Result := box(near, near, far, far)
	     else -- row of boxes at 4X pitch
		  ptch := 4 * lyr.min_pitch;
		  cnt := (len/ptch).floor;
		  obox := box_sc(mwid, mwid,
					(len-ptch*(cnt-1))/2,
					near+mwid/2);
		  create step.make_xy(ptch, 0);
		  Result := stepped_copies(obox, step, cnt)
	    end
	end;

	trenched_segment(lyr: DESIGN_TECHN_LAYER;
			 mnear: INTEGER; len, ht: DOUBLE; 
			 cnt: INTEGER): LAYOUT is
		-- lyr.scribe_style is TRENCHn
		-- scribe layout is cnt min_width trenches between near and far
	require two_or_three: cnt = 2 or cnt = 3
	local near1, near2, near3, far1, far2, far3, mwid: DOUBLE;
		trenches: LAYOUT_COLLECTION
	do  mwid  := lyr.min_width;
	    near1 := lyr.near_edge	-- near trench limits
	    far1  := near1 + mwid;
	    far2  := lyr.far_edge;	-- far trench limits
	    near2 := far2 - mwid;
	    near3 := (near1 + far2 - mwid) / 2;	-- middle trench limits
	    far3  := (near1 + far2 + mwid) / 2;
	    -- now build it
	    create trenches.make_unlayered(cnt);
	    trenches.extend(box(near1*mnear, near1, len, far1));
	    if cnt = 3 and far3 <= ht
	     then trenches.extend(box(near3*mnear, near3, len, far3))
	    end;
	    if far2 <= ht
	     then trenches.extend(box(near2*mnear, near2, len, far2))
	    end;
	    Result := trenches
	end;

end -- MOSIS_SCRIBE_SYNTHESIZER
