
indexing
	
	description: "EV_MULTI_COLUMN_LIST with additional properties"
	author: "David Hollenberg";
	date: "Ocrtober 11, 2004"

class MULTI_COLUMN_LIST [G -> MULTI_COLUMN_LIST_ROW]
inherit
	EV_MULTI_COLUMN_LIST
		redefine
			initialize
		end

feature {NONE} -- Creation

	initialize is
		do
			precursor
			create {ARRAYED_LIST [G]} all_rows.make (0)
			create {ARRAYED_LIST [MULTI_COLUMN_LIST_COLUMN [G]]} visible_columns.make (0)
			column_title_click_actions.extend (agent process_column_title_click)
			set_visibility_counts_included (True)
		end

feature -- Properties

	all_rows: LIST [G]
			-- Rows for all entities

	visible_rows: LIST [G] is
			-- Rows for all entities that are visible
		do
			from
				Result := all_rows.twin
				Result.wipe_out
				all_rows.start
			until
				all_rows.after
			loop
				if all_rows.item.is_visible then
					Result.extend (all_rows.item)
				end
				all_rows.forth
			end
		end

	visible_columns: LIST [MULTI_COLUMN_LIST_COLUMN [G]]
			-- List of columns which are currently visible

	selected_rows: LIST [G]
			-- Rows for entities that were selected
			-- as of last call to `rebuild_selected_rows'

	selection_status_box: EV_HORIZONTAL_BOX
			-- Box with selection and visibility status info

	row_to_generic_function: FUNCTION [ANY, TUPLE [EV_MULTI_COLUMN_LIST_ROW], G]
			-- Function which translates an 
			-- EV_MULTI_COLUMN_LIST_ROW to object of type G

	last_sort_position: INTEGER
			-- Position of visible column on which last sort
			-- was done

	sort_count: INTEGER
			-- Number of times same column has been sorted
			-- consecutively

	visibility_counts_included: BOOLEAN
			-- Should counts of visible and hidden rows
			-- be included in `selection_status_box'?

feature -- Modification

	set_row_to_generic_function (f: FUNCTION [ANY, TUPLE [EV_MULTI_COLUMN_LIST_ROW], G]) is
		do
			row_to_generic_function := f
		end

	set_visibility_counts_included (b: BOOLEAN) is
			-- Set `visibility_counts_included' to `b'
		do
			visibility_counts_included := b
		end

	set_visible_columns (list: LIST [MULTI_COLUMN_LIST_COLUMN [G]]) is
			-- Set `visible_columns' to `list'
		require
			list_exists: list /= Void
			list_not_empty: not list.is_empty
		do
			visible_columns := list
			last_sort_column := visible_columns.first
			last_sort_position := 1
			sort_count := 0
			process_change
		end
			
	build_selection_status_box is
			-- Build box with selection and visibility status info
			-- and put in `selection_status_box'
		local
			set: WIDGET_COLUMN_SET
		do
			create selection_status_box
			create selected_box
			create visible_box
			create hidden_box
			create total_box
			if visibility_counts_included then
				create set.make (4)
				set.append_row (<< selected_box, visible_box, hidden_box, total_box >>)
			else
				create set.make (2)
				set.append_row (<< selected_box, total_box >>)
			end
			set.set_padding (20)
			set.align_left
			set.align_columns_left
			set.disable_expand
			selection_status_box.extend (set.box)
			refresh_status_line_counts
		end

	set_all_rows (list: LIST [G]) is
			-- Set `all_rows' from `list'
		require
			list_exists: list /= Void
		do
			all_rows.wipe_out
			all_rows.append (list)
			refresh_status_line_counts
			process_change
		end
			
feature -- Operations

	sort_on_column (n: INTEGER) is
			-- Sort rows on column `n'
		require
			n_positive: n >= 1
			n_small_enough: n <= visible_columns.count
		do
			last_sort_column := visible_columns.i_th (n)
			last_sort_position := n
			sort_count := 1
			process_change
		end

	reverse_sort_on_column (n: INTEGER) is
			-- Reverse sort rows on column `n'
		require
			n_positive: n >= 1
			n_small_enough: n <= visible_columns.count
		do
			last_sort_column := visible_columns.i_th (n)
			last_sort_position := n
			sort_count := 2
			process_change
		end

	resize_all_columns_to_content is
			-- Resize every column of `Current' to its contents
		local
			k: INTEGER
		do
			from
				visible_columns.start
				k := 1
			until
				visible_columns.after
			loop
				if k > 1 then
					-- First column always has to be left
					-- aligned - no choice
					set_column_alignment (visible_columns.item.alignment, k)
				end
				resize_column_to_content (k)
				k := k + 1
				visible_columns.forth
			end
		end
			
	process_change is
			-- Process change to `all_rows' or `visible_columns'
		do
			if not visible_columns.is_empty then
				if not all_rows.is_empty then
					if sort_count /= 0 then
						-- Make sort be in same order as before
						-- rather than reversing the order
						sort_count := sort_count - 1
					end
					process_column_title_click (last_sort_position)
				else
					rebuild
				end
			end
		end

	process_select_one (row: G) is
			-- Take actions needed when user selects a row
		do
			selected_count := selected_count + 1
			queue_selected_count_refresh
		end

	process_deselect_one (row: G) is
			-- Take actions needed when user deselects a row
		do
			selected_count := selected_count - 1
			queue_selected_count_refresh
		end

	process_select_all is
			-- Take actions needed when user chooses to replace
			-- selection with all rows
		local
			row: G
		do
			from
				start
			until
				after
			loop
				row := row_to_generic (item)
				row.enable_select
				forth
			end
		end

	process_select_none is
			-- Take actions needed when user chooses to replace
			-- selection with no rows
		local
			row: G
		do
			from
				start
			until
				after
			loop
				row := row_to_generic (item)
				row.disable_select
				forth
			end
		end

	process_select_invert is
			-- Take actions needed when user chooses to invert
			-- selection status of each row
		local
			row: G
		do
			from
				start
			until
				after
			loop
				row := row_to_generic (item)
				if row.is_selected then
					row.disable_select
				else
					row.enable_select
				end
				forth
			end
		end

	process_select_rows_filtered (filter: FUNCTION [G, TUPLE [G], BOOLEAN]) is
			-- Take actions needed when user chooses to replace
			-- selection with all rows for which `filter' is true
		require
			filter_exists: filter /= Void
		local
			row: G
		do
			from
				start
			until
				after
			loop
				row := row_to_generic (item)
				if filter.item ([row]) then
					row.enable_select
				else
					row.disable_select
				end
				forth
			end
		end

	process_view_all_rows is
			-- Take actions needed when user chooses to view all
			-- rows
		do
			from
				all_rows.start
			until
				all_rows.after
			loop
				all_rows.item.enable_visible
				all_rows.forth
			end
			
			rebuild
		end

	process_view_rows_filtered (filter: FUNCTION [G, TUPLE [G], BOOLEAN]) is
			-- Take actions needed when user chooses to view
			-- only rows for which `filter' is true
		require
			filter_exists: filter /= Void
		local
			row: G
		do
			from
				all_rows.start
			until
				all_rows.after
			loop
				row := all_rows.item
				if filter.item ([row]) then
					row.enable_visible
				else
					row.disable_visible
				end
				all_rows.forth
			end
			rebuild
		end

	rebuild_selected_rows is
			-- Make list of all rows that are currently
			-- selected and put in `selected_rows'
		local
			the_row: G
		do
			create {ARRAYED_LIST [G]} selected_rows.make (selected_count.max (0))
				-- FIXME: change when Vision2 bug is fixed
				-- Work around bug where selected_count can
				-- become negative
			from
				start
			until
				after
			loop
				the_row := row_to_generic (item)
				if the_row.is_selected then
					selected_rows.extend (the_row)
				end
				forth
			end
		end

	refresh_after_selected_rows_operation is
			-- Refresh information after performing an
			-- operation that applies only to the selected rows
		do
			refresh_only_selected_rows
			resize_all_columns_to_content
		end

	refresh_row (row: G) is
			-- Refresh `row'
		do
			row.wipe_out
			from
				visible_columns.start
			until
				visible_columns.after
			loop
				row.extend (visible_columns.item.value (row))

				visible_columns.forth
			end
		end
			
	refresh_all_rows is
			-- Refresh all rows
		do
			refresh (False)
		end
			
feature {NONE} -- Implementation

	row_to_generic (row: EV_MULTI_COLUMN_LIST_ROW): G is
			-- Value of `row' which is of type G
		require
			row_exists: row /= Void
		do
			Result := row_to_generic_function.item ([row])
		ensure
			result_exists: Result /= Void
		end

	refresh_status_line_counts is
			-- Refresh all counters displayed in status line,
			-- which has visibility and "selected" counters
		do
			if selection_status_box /= Void then
				refresh_selected_count
				refresh_visible_counts
				update_counter (total_box, Total_text, total_count)
			end
		end

	queue_selected_count_refresh is
			-- Queue action to refresh "selected" counter
			-- when next idle
		local
			env: EV_ENVIRONMENT
		do
			if not selected_count_refresh_queued then
				create env
				env.application.do_once_on_idle (agent refresh_selected_count)
				selected_count_refresh_queued := True
			end
		end

	refresh_selected_count is
			-- Refresh "selected" counter
		do
			selected_count_refresh_queued := False
			
			-- FIXME: hack to workaround Vision2 bug (agent not
			-- always called when row selected or deselected)
			
			selected_count := current_selected_count
			update_counter (selected_box, Selected_text, selected_count)
		end

	current_selected_count: INTEGER is
			-- Current number of selected rows
		local
			the_row: EV_MULTI_COLUMN_LIST_ROW
			the_cursor: EV_DYNAMIC_LIST_CURSOR [EV_MULTI_COLUMN_LIST_ROW]
		do
			the_cursor := cursor
			from
				start
			until
				after
			loop
				the_row := item
				if the_row.is_selected then
					Result := Result + 1
				end
				forth
			end
			go_to (the_cursor)
		end

	refresh_visible_counts is
			-- Refresh "visible" and "hidden" counters
		do
			update_counter (visible_box, Visible_text, visible_count)
			update_counter (hidden_box, Hidden_text, hidden_count)
		end

	set_visible_counts is
			-- Set the "visible" and "hidden" counters
		do
			visible_count := count
			hidden_count := total_count - visible_count
		end

	update_counter (b: EV_BOX a_text: STRING n: INTEGER) is
			-- Update counter displayed in `b' with text `a_text'
			-- and value `n'
		require
			box_exists: b /= Void
			a_text_exists: a_text /= Void
			-- FIXME: add back in when Vision2 bug is fixed
			-- (select_actions and deselect_actions are not
			-- always called)
			-- value_nonnegative: n >= 0
		local
			label: EV_LABEL
		do
			create label.make_with_text (a_text + ":  " + n.out)
			b.wipe_out
			b.extend (label)
		end

	rebuild is
			-- Rebuild `Current' from rows in `all_rows'
			-- that are marked "visible" and from `visible_columns'
		local
			row: G
			k: INTEGER
			t: STRING
			col: MULTI_COLUMN_LIST_COLUMN [G]
			sort_sym: STRING
			old_visible_count, old_selected_count: INTEGER
		do
			save_selection
			old_visible_count := visible_count
			old_selected_count := selected_count
			wipe_out
			
			from
				visible_columns.start
				k := 1
			until
				visible_columns.after
			loop
				col := visible_columns.item
				t := col.title
				if col = last_sort_column then
					if sort_count \\ 2 = 1 then
						sort_sym := Sort_ascending
					else
						sort_sym := Sort_descending
					end
					if col.alignment.is_right_aligned then
						t := Sort_sym + Spacer + t
					else
						t := t + Spacer + Sort_sym
					end
				end
				set_column_title (t, k)
				k := k + 1
				visible_columns.forth
			end
			
			from
				k := visible_columns.count + 1
			until
				k > column_count
			loop
				set_column_title ("", k)
				set_column_width (0, k)
				k := k + 1
			end
			
			
			from
				all_rows.start
			until
				all_rows.after
			loop
				row := all_rows.item
				if row.is_visible then
					row.wipe_out
					from
						visible_columns.start
					until
						visible_columns.after
					loop
						row.extend (visible_columns.item.value (row))

						visible_columns.forth
					end
					extend (row)
				end
				all_rows.forth
			end
			
			resize_all_columns_to_content
			
			restore_selection
			set_visible_counts
			if selected_count /= old_selected_count then
				refresh_selected_count
			end
			if visible_count /= old_visible_count then
				refresh_visible_counts
			end
		end
			
	refresh_only_selected_rows is
			-- Refresh only rows that are currently selected
		do
			refresh (True)
		end
			
	refresh (selected_only: BOOLEAN) is
			-- Refresh all rows of `Current' or 
			-- only selected rows if `selected_only'
		local
			row: G
		do
			from
				start
			until
				after
			loop
				row := row_to_generic (item)
				if not selected_only or row.is_selected then
					refresh_row (row)
				end
				forth
			end
		end
			
	process_column_title_click (n: INTEGER) is
			-- Take actions needed when user clicks on column title
			-- for column `n'
		local
			comp: KL_COMPARATOR [G]
			col: like last_sort_column
		do
			if n <= visible_columns.count then
				col := visible_columns.i_th (n)
				comp := col.comparator
				if last_sort_column = col then
					sort_count := sort_count + 1
				else
					sort_count := 1
				end
				sort_all_rows (comp, sort_count \\ 2 = 0)
				last_sort_column := col
				last_sort_position := n
				rebuild
			end
		end

	sort_all_rows (a_comparator: KL_COMPARATOR [G] reverse: BOOLEAN) is
			-- Sort `all_rows' according to `a_comparator'
		local
			sorted_rows: ARRAY [G]
			sorter: DS_ARRAY_BUBBLE_SORTER [G]
			k: INTEGER
		do
			create sorted_rows.make (1, all_rows.count)
			from
				all_rows.start
				k := 1
			until
				all_rows.after
			loop
				sorted_rows.put (all_rows.item, k)
				all_rows.forth
				k := k + 1
			end
			create sorter.make (a_comparator)
			if reverse then
				sorter.reverse_sort (sorted_rows)
			else
				sorter.sort (sorted_rows)
			end
			create {ARRAYED_LIST [G]} all_rows.make (sorted_rows.count)
			from
				k := sorted_rows.lower
			until
				k > sorted_rows.upper
			loop
				all_rows.extend (sorted_rows.item (k))
				k := k + 1
			end
		end

	save_selection is
			-- Save `is_selected' status of each row
		local
			is_selected: BOOLEAN
			the_row: G
		do
			create selection_status.make (count)
			from
				start
			until
				after
			loop
				the_row := row_to_generic (item)
				is_selected := the_row.is_selected
				selection_status.put (is_selected, the_row.unique_identifier)
				forth
			end
		end

	restore_selection is
			-- Restore `is_selected' status of each row
		local
			is_selected: BOOLEAN
			the_row: G
			n: INTEGER
		do
			from
				start
			until
				after
			loop
				the_row := row_to_generic (item)
				is_selected := selection_status.item (the_row.unique_identifier)
				if is_selected then
					the_row.enable_select
					n := n + 1
				end
				forth
			end
			selected_count := n
		end

	last_sort_column: MULTI_COLUMN_LIST_COLUMN [G]
			-- Last column on which sort was done

	selection_status: HASH_TABLE [BOOLEAN, STRING]
			-- `is_selected' status of each entity

	selected_box: EV_HORIZONTAL_BOX
			-- Box to display number of selected items

	visible_box: EV_HORIZONTAL_BOX
			-- Box to display number of visible items

	hidden_box: EV_HORIZONTAL_BOX
			-- Box to display number of hidden items

	total_box: EV_HORIZONTAL_BOX
			-- Box to display total number of items

	selected_count: INTEGER
			-- Number of items currently selected

	visible_count: INTEGER
			-- Number of items currently visible

	hidden_count: INTEGER
			-- Number of items currently hidden

	total_count: INTEGER is
			-- Total number of items
		do
			Result := all_rows.count
		end
		
	selected_count_refresh_queued: BOOLEAN
			-- Is a call to `refresh_selected_count' currently
			-- queued for execution when things are idle?

	Sort_ascending: STRING is "+"

	Sort_descending: STRING is "-"

	Spacer: STRING is " "

	Selected_text: STRING is "Selected"
	Visible_text: STRING is "Visible"
	Hidden_text: STRING is "Hidden"
	Total_text: STRING is "Total"

end
