
indexing
	
	author: "David Hollenberg"
	date:	"December 6, 1995"
	description: "A set of Oracle named or numeric placeholders %
		%(bind variables)"

deferred class ORACLE_BINDSET

inherit
	ANY
	C_MEMORY_ALLOCATION
		export
			{NONE} all
		end
	ORACLE_VALIDITY
		export
			{ANY} is_valid_placeholder_value
			{NONE} all
		end
	ORACLE_EXCEPTIONS
		export
			{NONE} all
		end
	MEMORY
		export
			{NONE} all
		redefine
			dispose
		end

feature {NONE} -- Creation

	make (num: INTEGER the_cursor: ORACLE_CURSOR)
			-- Create a bindset capable of describing
			-- `num' placeholders and associated with cursor
			-- `the_cursor'
		require
			nonnegative_count: num >= 0
			cursor_exists: the_cursor /= Void
		do
			address := allocate_nongc_memory (num * placeholder_length)
			capacity := num
			cursor := the_cursor
			create_placeholder
			mark_not_in_use (address, num)
		ensure
			correct_capacity: capacity = num
			address_exists: address /= default_pointer
			placeholder_exists: placeholder /= Void
			correct_cursor: cursor = the_cursor
		end

feature {NONE}	-- Cleanup

	dispose
			-- Return memory allocated for bindset
		local
			null_ptr: POINTER
		do
			if address /= null_ptr then
				free_long_data_areas (address, placeholder_length, capacity)
				free_nongc_memory (address)
				address := null_ptr
			end
		end

feature -- Properties

	address: POINTER
			-- Address of area for placeholders

	count: INTEGER
			-- Number of placeholders in bindset

	capacity: INTEGER
			-- Maximum number of placeholders bindset is
			-- capable of holding

	reparse_required: BOOLEAN
			-- Does the SQL statement associated with this
			-- bindset need to be reparsed due to a
			-- rebind which changed external datatypes?

	cursor: ORACLE_CURSOR
			-- Cursor with which bindset is associated

	placeholder_name: STRING
			-- Display name of placeholder being bound
		do
			Result := placeholder.display_name
		end

feature -- Modification

	grow (num: INTEGER)
			-- Increase size of area for placeholders so
			-- that it can hold `num' placeholders.
			-- No effect if `num' is less than or equal
			-- `capacity'.
		require
			nonnegative_count: num >= 0
		local
			new_addr: POINTER
		do
			if num > capacity then
				new_addr := allocate_nongc_memory (num * placeholder_length)
				mark_not_in_use (new_addr, num)
				copy_nongc_memory(address, new_addr, capacity * placeholder_length)
				address := new_addr
				fix_placeholders
				capacity := num
			end
		ensure
			capacity_changed: num > old capacity implies num = capacity
			capacity_same: num <= old capacity implies capacity = old capacity
			address_changed: num > old capacity implies address /= old address
			address_same: num <= old capacity implies address = old address
		end

	bind_to_integer (value: INTEGER null: BOOLEAN)
			-- Bind `placeholder' to `value' if `null' is false
			-- otherwise, bind to null
		local
			ph: ORACLE_PLACEHOLDER
		do
			ph := placeholder
			if ph.in_use then
				if not ph.bound_to_integer then
					ph.set_needs_oci_binding (True)
				end
				if not (ph.bound_to_integer or ph.bound_to_real or ph.bound_to_double) then
					reparse_required := True
				end
			else
				ph.initialize
				count := count + 1
			end
			ph.bind_to_integer (value, null)
		end

	bind_to_real (value: REAL null: BOOLEAN)
			-- Bind `placeholder' to `value' if `null' is false
			-- otherwise, bind to null
		local
			ph: ORACLE_PLACEHOLDER
		do
			ph := placeholder
			if ph.in_use then
				if not ph.bound_to_real then
					ph.set_needs_oci_binding (True)
				end
				if not (ph.bound_to_integer or ph.bound_to_real or ph.bound_to_double) then
					reparse_required := True
				end
			else
				ph.initialize
				count := count + 1
			end
			ph.bind_to_real (value, null)
		end

	bind_to_double (value: DOUBLE null: BOOLEAN)
			-- Bind `placeholder' to `value' if `null' is false
			-- otherwise, bind to null
		local
			ph: ORACLE_PLACEHOLDER
		do
			ph := placeholder
			if ph.in_use then
				if not ph.bound_to_double then
					ph.set_needs_oci_binding (True)
				end
				if not (ph.bound_to_integer or ph.bound_to_real or ph.bound_to_double) then
					reparse_required := True
				end
			else
				ph.initialize
				count := count + 1
			end
			ph.bind_to_double (value, null)
		end

	bind_to_character (value: CHARACTER null: BOOLEAN)
			-- Bind `placeholder' to `value' if `null' is false;
			-- otherwise, bind to null
		local
			ph: ORACLE_PLACEHOLDER
		do
			ph := placeholder
			if ph.in_use then
				if not ph.bound_to_string then
					reparse_required := True
				end
			else
				ph.initialize
				count := count + 1
			end
			ph.bind_to_character (value, null)
		end

	bind_to_boolean (value: BOOLEAN null: BOOLEAN)
			-- Bind `placeholder' to `value' if `null' is false;
			-- otherwise, bind to null
		local
			c: CHARACTER
		do
			if value then
				c := 'Y'
			else
				c := 'N'
			end
			bind_to_character (c, null)
		end

	bind_to_string (value: STRING null: BOOLEAN)
			-- Bind `placeholder' to `value' if `null' is false;
			-- otherwise, bind to null.  If `value' is void,
			-- bind to null regardless of `null'
		local
			ph: ORACLE_PLACEHOLDER
		do
			ph := placeholder
			if ph.in_use then
				if not ph.bound_to_string then
					reparse_required := True
				end
			else
				ph.initialize
				count := count + 1
			end
			ph.bind_to_string (value, null)
		end

	bind_to_string_with_max_size (value: STRING max_out_len: INTEGER null: BOOLEAN)
			-- Bind `placeholder' to `value' if `null' is false;
			-- otherwise, bind to null.  If `value' is void,
			-- bind to null regardless of `null'.
			-- Use `max_out_len' as the maximum length of the
			-- string if it is greater than the actual length
		local
			ph: ORACLE_PLACEHOLDER
		do
			ph := placeholder
			if ph.in_use then
				if not ph.bound_to_string then
					reparse_required := True
				end
			else
				ph.initialize
				count := count + 1
			end
			ph.bind_to_string_with_max_size (value, max_out_len, null)
		end

	bind_to_raw (value: STRING null: BOOLEAN)
			-- Bind `placeholder' to `value' if `null' is false;
			-- otherwise, bind to null.  If `value' is void,
			-- bind to null regardless of `null'
		local
			ph: ORACLE_PLACEHOLDER
		do
			ph := placeholder
			if ph.in_use then
				if not ph.bound_to_raw then
					reparse_required := True
				end
			else
				ph.initialize
				count := count + 1
			end
			ph.bind_to_raw (value, null)
		end

	bind_to_raw_with_max_size (value: STRING max_out_len: INTEGER null: BOOLEAN)
			-- Bind `placeholder' to `value' if `null' is false;
			-- otherwise, bind to null.  If `value' is void,
			-- bind to null regardless of `null'.
			-- Use `max_out_len' as the maximum length of the
			-- string if it is greater than the actual length
		local
			ph: ORACLE_PLACEHOLDER
		do
			ph := placeholder
			if ph.in_use then
				if not ph.bound_to_raw then
					reparse_required := True
				end
			else
				ph.initialize
				count := count + 1
			end
			ph.bind_to_raw_with_max_size (value, max_out_len, null)
		end

	bind_to_date (value: ORACLE_DATE null: BOOLEAN)
			-- Bind `placeholder' to `value' if `null' is
			-- false; otherwise, bind to null.  If `value'
			-- is void or `value'.`is_null', bind to null
			-- regardless of `null'
		local
			ph: ORACLE_PLACEHOLDER
		do
			ph := placeholder
			if ph.in_use then
				if not ph.bound_to_date then
					reparse_required := True
				end
			else
				ph.initialize
				count := count + 1
			end
			ph.bind_to_date (value, null)
		end

	bind_to_rowid (value: ORACLE_ROWID null: BOOLEAN)
			-- Bind `placeholder' to `value' if `null' is false;
			-- otherwise, bind to null.  If `value' is void,
			-- bind to null regardless of `null'
		local
			ph: ORACLE_PLACEHOLDER
		do
			ph := placeholder
			if ph.in_use then
				if not ph.bound_to_rowid then
					reparse_required := True
				end
			else
				ph.initialize
				count := count + 1
			end
			ph.bind_to_rowid (value, null)
		end

	bind_to_any (value: ANY null: BOOLEAN)
			-- Bind `placeholder' to `value' if `null' is false;
			-- otherwise, bind to null
		require
			value_exists: value /= Void
			valid_value: is_valid_placeholder_value (value)
		local
			integer_ref: INTEGER_REF
			real_ref: REAL_REF
			double_ref: DOUBLE_REF
			boolean_ref: BOOLEAN_REF
			char_ref: CHARACTER_REF
			string: STRING
			date: ORACLE_DATE
			rowid: ORACLE_ROWID
			message: STRING
		do
			integer_ref ?= value
			real_ref ?= value
			double_ref ?= value
			boolean_ref ?= value
			char_ref ?= value
			string ?= value
			date ?= value
			rowid ?= value
			if integer_ref /= Void then
				bind_to_integer (integer_ref.item, null)
			elseif real_ref /= Void then
				bind_to_real (real_ref.item, null)
			elseif double_ref /= Void then
				bind_to_double (double_ref.item, null)
			elseif boolean_ref /= Void then
				bind_to_boolean (boolean_ref.item, null)
			elseif char_ref /= Void then
				bind_to_character (char_ref.item, null)
			elseif string /= Void then
				bind_to_string (string, null)
			elseif date /= Void then
				bind_to_date (date, null)
			elseif rowid /= Void then
				bind_to_rowid (rowid, null)
			else
				create message.make (0)
				message.append ("attempt to bind placeholder %"")
				message.append (placeholder.display_name)
				message.append ("%" to value whose type (")
				if value /= Void then
					message.append (value.generator)
				else
					message.append ("Void")
				end
				message.append (") is incompatible with database")
				cursor.handle_oracle_error (dbex.Db_invalid_placeholder_value,
						message)
			end
		end

	check_for_unbound_placeholders
			-- Check for "unbound placeholder" errors, which
			-- can occur when a cursor is reused
		local
			ph_addr: POINTER
			k, len, local_capacity: INTEGER
			item: ORACLE_PLACEHOLDER
			message: STRING
		do
			from
				ph_addr := address
				k := 1
				local_capacity := capacity
				len := placeholder_length
				item := placeholder
			until
				k > local_capacity
			loop
				item.set_address (ph_addr)
				if item.in_use and then not item.is_bound then
					create message.make (0)
					message.append ("placeholder %"")
					message.append (item.display_name)
					message.append ("%" not bound")
					cursor.handle_oracle_error (dbex.Db_unbound_placeholder,
						message)
				end
				ph_addr := next_placeholder (ph_addr, len)
				k := k + 1
			end
		end

	set_needs_oci_bind
			-- Mark all placeholders as needing OCI binding
		local
			k, len, local_capacity: INTEGER
			item: ORACLE_PLACEHOLDER
			ph_addr: POINTER
		do
			from
				k := 1
				ph_addr := address
				item := placeholder
				len := placeholder_length
				local_capacity := capacity
			until
				k > local_capacity
			loop
				item.set_address (ph_addr)
				if item.in_use then
					item.set_needs_oci_binding (True)
				end
				ph_addr := next_placeholder (ph_addr, len)
				k := k + 1
			end
		end

	set_reparse_required (newval: BOOLEAN)
			-- Set `reparse_required' to `newval'
		do
			reparse_required := newval
		ensure
			value_set: reparse_required = newval
		end

	oci_bind
			-- Bind all placeholders in `Current' which need
			-- binding via the Oracle Call Interface
		local
			ph_addr: POINTER
			k, len, local_capacity: INTEGER
			item: ORACLE_PLACEHOLDER
		do
			from
				ph_addr := address
				k := 1
				len := placeholder_length
				local_capacity := capacity
				item := placeholder
			until
				k > local_capacity
			loop
				item.set_address (ph_addr)
				if item.in_use and item.needs_oci_binding then
					item.oci_bind (cursor)
				end
				ph_addr := next_placeholder (ph_addr, len)
				k := k + 1
			end
		end

	unbind
			-- Unbind bindset so that it can be correctly reused
		local
			ph_addr: POINTER
			k, len, local_capacity: INTEGER
			item: ORACLE_PLACEHOLDER
		do
			from
				ph_addr := address
				k := 1
				len := placeholder_length
				local_capacity := capacity
				item := placeholder
			until
				k > local_capacity
			loop
				item.set_address (ph_addr)
				if item.in_use then
					item.unbind
				end
				ph_addr := next_placeholder (ph_addr, len)
				k := k + 1
			end
		end

feature -- Display

	append_description (dest: STRING)
			-- Append to `dest' a description of `Current'
		require
			destination_exists: dest /= Void
		local
			ph_addr: POINTER
			k, len, display_len, local_capacity: INTEGER
			item: ORACLE_PLACEHOLDER
		do
			from
				ph_addr := address
				k := 1
				len := placeholder_length
				display_len := max_display_name_length
				local_capacity := capacity
				item := placeholder
			until
				k > local_capacity
			loop
				item.set_address (ph_addr)
				if item.in_use then
					item.append_description (display_len, dest)
				end
				ph_addr := next_placeholder (ph_addr, len)
				k := k + 1
			end
		end

feature {NONE} -- Implementation

	placeholder: ORACLE_PLACEHOLDER
			-- Object used to access and modify a piece of C memory
			-- containing a placeholder as an Eiffel object.
			-- Having this as an attribute avoids repeated
			-- creation.

	found: BOOLEAN
			-- Was last placeholder for which `find' was
			-- called found?

	create_placeholder
			-- Create placeholder object and put in `placeholder'
		deferred
		ensure
			placeholder_exists: placeholder /= Void
		end

	mark_not_in_use (addr: POINTER num: INTEGER)
			-- Initialize and mark "not in use" `num'
			-- newly allocated placeholder slots starting
			-- at address `addr'
		local
			k, len: INTEGER
			item: ORACLE_PLACEHOLDER
			ph_addr: POINTER
		do
			from
				k := 1
				ph_addr := addr
				len := placeholder_length
				item := placeholder
			until
				k > num
			loop
				item.set_address (ph_addr)
				item.set_in_use (False)
				item.clear_long_data_address
				item.set_long_data_capacity (0)
				ph_addr := next_placeholder (ph_addr, len)
				k := k + 1
			end
		end

	fix_placeholders
			-- Correct each placeholder which is in use to account
			-- for the fact that it has been moved to a
			-- different memory location
		local
			ph_addr: POINTER
			k, len, old_capacity: INTEGER
			item: ORACLE_PLACEHOLDER
		do
			from
				ph_addr := address
				k := 1
				len := placeholder_length
				old_capacity := capacity
				item := placeholder
			until
				k > old_capacity
			loop
				item.set_address (ph_addr)
				if item.in_use then
					item.fix_after_move
				end
				ph_addr := next_placeholder (ph_addr, len)
				k := k + 1
			end
		end

	max_display_name_length: INTEGER
			-- Length of longest display name for a placeholder
			-- in `Current'
		local
			ph_addr: POINTER
			k, len, local_capacity: INTEGER
			item: ORACLE_PLACEHOLDER
		do
			from
				ph_addr := address
				k := 1
				len := placeholder_length
				local_capacity := capacity
				item := placeholder
			until
				k > local_capacity
			loop
				item.set_address (ph_addr)
				if item.in_use then
					Result := Result.max (item.display_name.count)
				end
				ph_addr := next_placeholder (ph_addr, len)
				k := k + 1
			end
		end

feature {NONE}	-- Externals

	placeholder_length: INTEGER
			-- Length of one placeholder in bytes
		external
			"C macro use %"oracle.h%""
		alias
			"ph_length"
		end

	next_placeholder (ph_addr: POINTER len: INTEGER): POINTER
			-- Address of placeholder following the one at
			-- `ph_addr', assuming placeholder length is `len'
		external
			"C macro use %"oracle.h%""
		alias
			"ph_next_address"
		end

	free_long_data_areas (ph_addr: POINTER ph_len, cap: INTEGER)
			-- Free long data areas in bindset whose
			-- first placeholder is at `ph_addr', with each
			-- placeholder having length `ph_len' and which
			-- holds `cap' placeholders
		external
			"C macro use %"oracle.h%""
		alias
			"ph_free_long_data_areas"
		end

invariant
	address_exists: address /= default_pointer
	placeholder_exists: placeholder /= Void
	valid_count: count >= 0
	valid_capacity: capacity >= 0
	count_small_enough: count <= capacity

end
