
indexing
	
	author: "David Hollenberg"
	date:	"November 22, 1995"
	description: "An Oracle Cursor Data Area (CDA)"

class ORACLE_CDA

inherit
	ANY
	ORACLE_CALL_INTERFACE
		export
			{NONE} all
		end
	C_MEMORY_ALLOCATION
		export
			{NONE} all
		end

feature {NONE} -- Creation

	make
			-- Allocate memory for Cursor Data Area and
			-- initialize it
		do
			address := allocate_nongc_memory (length)
			cda_set_self (address, address)
		end
			
feature -- Properties

	address: POINTER
			-- Address of start of Cursor Data Area

	sql_function_code: INTEGER
			-- Code indicating SQL function requested
		do
			Result := cda_sql_func_code (address)
		end
	
	rows_processed_count: INTEGER
			-- Number of rows processed by SQL operation
		do
			Result := cda_rows_proc_count (address)
		end
	
	parse_error_offset: INTEGER
			-- Starting position of parse error within SQL statement
		do
			Result := cda_parse_err_offset (address)
		end
	
	oci_function_code: INTEGER
			-- Code indicating which OCI function was requested
		do
			Result := cda_oci_func_code (address)
		end
	
	return_code: INTEGER
			-- Oracle error code for most recently
			-- executed statement which used this cursor
		do
			Result := cda_retcode (address)
		end
	
	is_warning: BOOLEAN
			-- Are there any warnings in the CDA's warning flags?
		do
			Result := cda_warnings (address)
		end
	
	data_truncated: BOOLEAN
			-- Was data item truncated on OFETCH?
		do
			Result := cda_data_trunc (address)
		end
	
	missing_where_clause: BOOLEAN
			-- Was UPDATE or DELETE done with no where clause?
		do
			Result := cda_no_where (address)
		end
	
	cursor_number: INTEGER
			-- Internal number used by Oracle to track cursor
		do
			Result := cda_cursor_num (address)
		end
	
	internal_rowid_address: POINTER
			-- Address of rowid in internal format
		do
			Result := cda_rowid_addr (address)
		end
	
	internal_rowid_length: INTEGER
			-- Length of rowid in internal format
		do
			Result := cda_rowid_len (address)
		end
	
	internal_rowid_first_half: DOUBLE
			-- First half of rowid in internal format
		do
			cda_get_rowid_first_half (address, $Result)
		end
	
	internal_rowid_second_half: DOUBLE
			-- Second half of rowid in internal format
		do
			cda_get_rowid_second_half (address, $Result)
		end
	
	os_error_code: INTEGER
			-- OS-dependent error code
		do
			Result := cda_osd_error_code (address)
		end
	

feature {NONE}	-- Externals

	length: INTEGER
			-- Length in bytes
		external
			"C macro use %"oracle.h%""
		alias
			"cda_length"
		end

	cda_sql_func_code (cda_address: POINTER): INTEGER
			-- Code indicating last SQL function requested
		external
			"C macro use %"oracle.h%""
		end
	
	cda_rows_proc_count (cda_address: POINTER): INTEGER
			-- Number of rows processed by SQL operation
		external
			"C macro use %"oracle.h%""
		end
	
	cda_parse_err_offset (cda_address: POINTER): INTEGER
			-- Starting position of parse error within SQL
			-- statement
		external
			"C macro use %"oracle.h%""
		end
	
	cda_oci_func_code (cda_address: POINTER): INTEGER
			-- Code indicating which OCI function was requested
		external
			"C macro use %"oracle.h%""
		end
	
	cda_retcode (cda_address: POINTER): INTEGER
			-- Current Oracle error code for most recently
			-- executed statement
		external
			"C macro use %"oracle.h%""
		end
	
	cda_warnings (cda_address: POINTER): BOOLEAN
			-- Are there any warnings in the CDA's warning flags?
		external
			"C macro use %"oracle.h%""
		end
	
	cda_data_trunc (cda_address: POINTER): BOOLEAN
			-- Was data item truncated on OFETCH?
		external
			"C macro use %"oracle.h%""
		end
	
	cda_aggr_func_null (cda_address: POINTER): BOOLEAN
			-- Was null encountered during evaluation
			-- of aggregate function?
		external
			"C macro use %"oracle.h%""
		end
	
	cda_no_where (cda_address: POINTER): BOOLEAN
			-- Was UPDATE or DELETE done with no where clause?
		external
			"C macro use %"oracle.h%""
		end
	
	cda_plsql_error (cda_address: POINTER): BOOLEAN
			-- Was there a PL/SQL compilation error?
		external
			"C macro use %"oracle.h%""
		end
	
	cda_cursor_num (cda_address: POINTER): INTEGER
			-- Internal number used by Oracle to track cursor
		external
			"C macro use %"oracle.h%""
		end
	
	cda_rowid_addr (cda_address: POINTER): POINTER
			-- Address of internal rowid in Cursor Data Area
		external
			"C macro use %"oracle.h%""
		end
	
	cda_rowid_len (cda_address: POINTER): INTEGER
			-- Length of internal rowid in Cursor Data Area
		external
			"C macro use %"oracle.h%""
		end
	
	cda_rowid_first_half (cda_address: POINTER): DOUBLE
			-- First half of internal rowid in Cursor Data Area
		external
			"C macro use %"oracle.h%""
		end
	
	cda_rowid_second_half (cda_address: POINTER): DOUBLE
			-- Second half of internal rowid in Cursor Data Area
		external
			"C macro use %"oracle.h%""
		end
	
	cda_get_rowid_first_half (cda_address: POINTER dest: POINTER)
			-- Get first half of internal rowid in Cursor Data Area
			-- and put it in `dest'
		external
			"C macro use %"oracle.h%""
		end
	
	cda_get_rowid_second_half (cda_address: POINTER dest: POINTER)
			-- Get second half of internal rowid in Cursor Data
			-- Area and put it in `dest'
		external
			"C macro use %"oracle.h%""
		end
	
	cda_osd_error_code (cda_address: POINTER): INTEGER
			-- OS-dependent error code
		external
			"C macro use %"oracle.h%""
		end
	
feature {NONE}	-- Externals for accessing and setting fields beyond end
		-- of Cursor Data Area proper

	cda_self_address (cda_address: POINTER): POINTER
			-- Address of self stored inside Cursor Data Area
		external
			"C macro use %"oracle.h%""
		end
	
	cda_set_self (cda_address, value: POINTER)
			-- Set address of self inside Cursor Data Area
			-- at address `cda_address' to `value'
		external
			"C macro use %"oracle.h%""
		end
	
invariant
	address_exists: address /= default_pointer
	correct_self_address: cda_self_address (address) = address

end
