
indexing
	
	author: "David Hollenberg"
	date:	"November 28, 1995"
	description: "An Oracle cursor associated with a parsed SQL statement"

class ORACLE_CURSOR

inherit
	SQL_STATEMENT
		rename
			text as sql_statement,
			select_count as select_item_capacity
		redefine
			is_equal, connection, query_select_items
		end
	SQL_CHANGE
		rename
			text as sql_statement,
			select_count as select_item_capacity
		redefine
			is_equal, connection, query_select_items
		end
	SQL_BLOCK
		rename
			text as sql_statement,
			select_count as select_item_capacity
		redefine
			is_equal, connection, query_select_items
		end
	SQL_QUERY
		rename
			text as sql_statement,
			select_count as select_item_capacity
		redefine
			is_equal, connection, query_select_items
		end
	ORACLE_EXTERNALS
		export
			{NONE} all
		redefine
			is_equal
		end
	ORACLE_CDA
		rename
			make as oracle_cda_make
		redefine
			is_equal
		end
	ORACLE_VALIDITY
		redefine
			is_equal
		end
	ORACLE_CURSOR_STATES
		export
			{ANY} is_valid_state
			{NONE} all
		redefine
			is_equal
		end
	ORACLE_EXCEPTIONS
		export
			{NONE} all
		redefine
			is_equal
		end
	ORACLE_STATEMENT_TYPES
		export
			{ANY} valid_statement_type
			{NONE} all
		redefine
			is_equal
		end
	OCI_CONSTANTS
		export
			{NONE} all
		redefine
			is_equal
		end
	COMPARABLE		redefine
			is_equal
		end
	OS_ACCESS
		redefine
			is_equal
		end
	UNIX_OS_ACCESS
		redefine
			is_equal
		end
	STRING_UTILITIES_ACCESS
		export
			{NONE} all
		redefine
			is_equal
		end
	MEMORY
		rename
			free as memory_free
		export
			{NONE} all
		redefine
			dispose, is_equal
		end

create {ORACLE_CURSOR_SERVER}
	make

feature {NONE} -- Creation

	make (conn: ORACLE_CONNECTION)
			-- Create a new closed cursor and associate it with
			-- connection `conn'
		do
			oracle_cda_make
			set_state (Cursor_closed)
			connection := conn
			lda_address := conn.lda_address
			end_of_data := True
		ensure
			cursor_closed: is_closed
		end
			
feature {ORACLE_CURSOR_SERVER} -- Reinitialization

	open
			-- Open cursor
		require	
			cursor_closed: is_closed
		local
			code: INTEGER
		do
			enter_oci
			code := ora_open_cursor (address, connection.lda_address)
			if code = Oci_success_code then
				set_state (Cursor_open)
			end
			leave_oci
			check_status (code)
		ensure
			cursor_open: is_open
		end
			
	try_to_open
			-- Try to open cursor.  Do not raise an exception
			-- if open fails.  Caller must check cursor
			-- state to determine whether open was successful
		require	
			cursor_closed: is_closed
		local
			code: INTEGER
		do
			enter_oci
			code := ora_open_cursor (address, connection.lda_address)
			if code = Oci_success_code then
				set_state (Cursor_open)
			end
			leave_oci
		end
			
	parse (sql_stmt: STRING)
			-- Associate `sql_stmt' with cursor
		require
			statement_exists: sql_stmt /= Void
			cursor_open: is_open
		local
			stmt_area: ANY
			code, count: INTEGER
		do
			sql_statement := sql_stmt.twin
			numeric_bindings := Void
			named_bindings := Void
			select_items := Void
			count := sql_stmt.count
			stmt_area := sql_stmt.area
			enter_oci
			code := ora_parse_sql_statement (address, $stmt_area, count, Nondeferred_mode)
			if code = Oci_success_code then
				set_state (Cursor_parsed)
			end
			leave_oci
			check_status (code)
		ensure
			cursor_parsed: is_parsed
		rescue
			sql_statement := Void
		end
			
	reparse
			-- Reparse SQL associated SQL statement.  This must
			-- be done when a placeholder is rebound to a
			-- different external datatype.
		require
			statement_exists: sql_statement /= Void
			cursor_reserved: is_reserved
		local
			stmt_area: ANY
			code: INTEGER
		do
			stmt_area := sql_statement.area
			enter_oci
			code := ora_parse_sql_statement (address, $stmt_area, sql_statement.count, Nondeferred_mode)
			leave_oci
			check_status (code)
		end
			
	reparse_new_statement (sql_stmt: STRING)
			-- Reparse open and parsed cursor, but with a
			-- new SQL statement.  Not currently used because
			-- if the parse fails, any parse locks held
			-- by the engine on tables in the SQL statement
			-- previously associated with this cursor are
			-- not released.
		require
			statement_exists: sql_stmt /= Void
			cursor_parsed: is_parsed
		do
			set_state (Cursor_open)
			sql_statement := Void
			parse (sql_stmt)
		end
			
	reserve
			-- Reserve cursor (mark it as "in use")
		require	
			cursor_parsed: is_parsed
		do
			set_state (Cursor_reserved)
		ensure
			cursor_reserved: is_reserved
		end
			
feature {ORACLE_CURSOR_SERVER} -- States

	state: INTEGER
			-- State of cursor
			
	set_state (s: INTEGER)
			-- Set state of `Current' to `s'
		require
			valid_state: is_valid_state (s)
		do
			state := s
		ensure
			state_set: state = s
		end
			
feature -- Numeric placeholder binding

	set_numeric_bind_capacity (num: INTEGER)
			-- Notify cursor to allocate space for
			-- at least `num' numeric placeholders
		do
			if numeric_bindings = Void then
				create numeric_bindings.make (num, Current)
			else
				numeric_bindings.grow (num)
			end
		end

	bind_num_to_integer (ph_number: INTEGER value: INTEGER null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null
		do
			make_numeric_bindings
			numeric_bindings.bind_to_integer (ph_number, value, null)
		end

	bind_num_to_real (ph_number: INTEGER value: REAL null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null
		do
			make_numeric_bindings
			numeric_bindings.bind_to_real (ph_number, value, null)
		end

	bind_num_to_double (ph_number: INTEGER value: DOUBLE null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null
		do
			make_numeric_bindings
			numeric_bindings.bind_to_double (ph_number, value, null)
		end

	bind_num_to_character (ph_number: INTEGER value: CHARACTER null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null
		do
			make_numeric_bindings
			numeric_bindings.bind_to_character (ph_number, value, null)
		end

	bind_num_to_boolean (ph_number: INTEGER value: BOOLEAN null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null
		do
			make_numeric_bindings
			numeric_bindings.bind_to_boolean (ph_number, value, null)
		end

	bind_num_to_string (ph_number: INTEGER value: STRING null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null
		do
			make_numeric_bindings
			numeric_bindings.bind_to_string (ph_number, value, null)
		end

	bind_num_to_string_with_max_size (ph_number: INTEGER value: STRING max_out_len: INTEGER null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null.
			-- Use `max_out_len' as the maximum length of the
			-- string if it is greater than the actual length
		do
			make_numeric_bindings
			numeric_bindings.bind_to_string_with_max_size (ph_number, value, max_out_len, null)
		end

	bind_num_to_raw (ph_number: INTEGER value: STRING null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null
		do
			make_numeric_bindings
			numeric_bindings.bind_to_raw (ph_number, value, null)
		end

	bind_num_to_raw_with_max_size (ph_number: INTEGER value: STRING max_out_len: INTEGER null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null.
			-- Use `max_out_len' as the maximum length of the
			-- string if it is greater than the actual length
		do
			make_numeric_bindings
			numeric_bindings.bind_to_raw_with_max_size (ph_number, value, max_out_len, null)
		end

	bind_num_to_date (ph_number: INTEGER value: like date_value null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null
		local
			date: ORACLE_DATE
		do
			date ?= value
			if date /= Void or value = Void or else value.is_null then
				make_numeric_bindings
				numeric_bindings.bind_to_date (ph_number, date, null)
			else
				raise (dbex.Db_conformance_error)
			end
		end

	bind_num_to_rowid (ph_number: INTEGER value: like rowid_value null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null
		local
			rowid: ORACLE_ROWID
		do
			rowid ?= value
			if rowid /= Void or value = Void then
				make_numeric_bindings
				numeric_bindings.bind_to_rowid (ph_number, rowid, null)
			else
				raise (dbex.Db_conformance_error)
			end
		end

	bind_num_to_any (ph_number: INTEGER value: ANY null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null
		do
			make_numeric_bindings
			numeric_bindings.bind_to_any (ph_number, value, null)
		end

feature -- Named placeholder binding

	set_named_bind_capacity (num: INTEGER)
			-- Notify cursor to allocate space for
			-- at least `num' named placeholders
		do
			if named_bindings = Void then
				create named_bindings.make (num, Current)
			else
				named_bindings.grow (num)
			end
		end

	bind_name_to_integer (ph_name: STRING value: INTEGER null: BOOLEAN)
			-- Bind placeholder named `ph_name' to `value' if
			-- `null' is false; otherwise, bind to null
		do
			make_named_bindings
			named_bindings.bind_to_integer (ph_name, value, null)
		end

	bind_name_to_real (ph_name: STRING value: REAL null: BOOLEAN)
			-- Bind placeholder named `ph_name' to `value' if
			-- `null' is false; otherwise, bind to null
		do
			make_named_bindings
			named_bindings.bind_to_real (ph_name, value, null)
		end

	bind_name_to_double (ph_name: STRING value: DOUBLE null: BOOLEAN)
			-- Bind placeholder named `ph_name' to `value' if
			-- `null' is false; otherwise, bind to null
		do
			make_named_bindings
			named_bindings.bind_to_double (ph_name, value, null)
		end

	bind_name_to_character (ph_name: STRING value: CHARACTER null: BOOLEAN)
			-- Bind placeholder named `ph_name' to `value' if
			-- `null' is false; otherwise, bind to null
		do
			make_named_bindings
			named_bindings.bind_to_character (ph_name, value, null)
		end

	bind_name_to_boolean (ph_name: STRING value: BOOLEAN null: BOOLEAN)
			-- Bind placeholder named `ph_name' to `value' if
			-- `null' is false; otherwise, bind to null
		do
			make_named_bindings
			named_bindings.bind_to_boolean (ph_name, value, null)
		end

	bind_name_to_string (ph_name: STRING value: STRING null: BOOLEAN)
			-- Bind placeholder named `ph_name' to `value' if
			-- `null' is false; otherwise, bind to null
		do
			make_named_bindings
			named_bindings.bind_to_string (ph_name, value, null)
		end

	bind_name_to_string_with_max_size (ph_name: STRING value: STRING max_out_len: INTEGER null: BOOLEAN)
			-- Bind placeholder named `ph_name' to `value' if
			-- `null' is false; otherwise, bind to null.
			-- Use `max_out_len' as the maximum length of the
			-- string if it is greater than the actual length
		do
			make_named_bindings
			named_bindings.bind_to_string_with_max_size (ph_name, value, max_out_len, null)
		end

	bind_name_to_raw (ph_name: STRING value: STRING null: BOOLEAN)
			-- Bind placeholder named `ph_name' to `value' if
			-- `null' is false; otherwise, bind to null
		do
			make_named_bindings
			named_bindings.bind_to_raw (ph_name, value, null)
		end

	bind_name_to_raw_with_max_size (ph_name: STRING value: STRING max_out_len: INTEGER null: BOOLEAN)
			-- Bind placeholder named `ph_name' to `value' if
			-- `null' is false; otherwise, bind to null.
			-- Use `max_out_len' as the maximum length of the
			-- string if it is greater than the actual length
		do
			make_named_bindings
			named_bindings.bind_to_raw_with_max_size (ph_name, value, max_out_len, null)
		end

	bind_name_to_date (ph_name: STRING value: like date_value null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null
		local
			date: ORACLE_DATE
		do
			date ?= value
			if date /= Void or value = Void or else value.is_null then
				make_named_bindings
				named_bindings.bind_to_date (ph_name, date, null)
			else
				raise (dbex.Db_conformance_error)
			end
		end

	bind_name_to_rowid (ph_name: STRING value: like rowid_value null: BOOLEAN)
			-- Bind placeholder named `ph_name' to `value' if
			-- `null' is false; otherwise, bind to null
		local
			rowid: ORACLE_ROWID
		do
			rowid ?= value
			if rowid /= Void or value = Void then
				make_named_bindings
				named_bindings.bind_to_rowid (ph_name, rowid, null)
			else
				raise (dbex.Db_conformance_error)
			end
		end

	bind_name_to_any (ph_name: STRING value: ANY null: BOOLEAN)
			-- Bind placeholder named `ph_name' to `value' if
			-- `null' is false; otherwise, bind to null
		do
			make_named_bindings
			named_bindings.bind_to_any (ph_name, value, null)
		end

feature -- Numeric placeholder properties

	is_bound_num (ph_number: INTEGER): BOOLEAN
			-- Is placeholder number `ph_number' bound to anything?
		do
			if numeric_bindings /= Void then
				Result := numeric_bindings.is_bound_any (ph_number)
			end
		end

	is_bound_null_num (ph_number: INTEGER): BOOLEAN
			-- Is placeholder number `ph_number' bound to null?
		do
			if numeric_bindings /= Void then
				Result := numeric_bindings.is_bound_null (ph_number)
			end
		end

	is_bound_integer_num (ph_number: INTEGER): BOOLEAN
			-- Is placeholder number `ph_number' bound to an
			-- integer?
		do
			if numeric_bindings /= Void then
				Result := numeric_bindings.is_bound_integer (ph_number)
			end
		end

	is_bound_real_num (ph_number: INTEGER): BOOLEAN
			-- Is placeholder number `ph_number' bound to a real?
		do
			if numeric_bindings /= Void then
				Result := numeric_bindings.is_bound_real (ph_number)
			end
		end

	is_bound_double_num (ph_number: INTEGER): BOOLEAN
			-- Is placeholder number `ph_number' bound to a
			-- double?
		do
			if numeric_bindings /= Void then
				Result := numeric_bindings.is_bound_double (ph_number)
			end
		end

	is_bound_character_num (ph_number: INTEGER): BOOLEAN
			-- Is placeholder number `ph_number' bound to a
			-- character?
		do
			if numeric_bindings /= Void then
				Result := numeric_bindings.is_bound_character (ph_number)
			end
		end

	is_bound_boolean_num (ph_number: INTEGER): BOOLEAN
			-- Is placeholder number `ph_number' bound to a
			-- boolean?
		do
			if numeric_bindings /= Void then
				Result := numeric_bindings.is_bound_boolean (ph_number)
			end
		end

	is_bound_string_num (ph_number: INTEGER): BOOLEAN
			-- Is placeholder number `ph_number' bound to a
			-- string?
		do
			if numeric_bindings /= Void then
				Result := numeric_bindings.is_bound_string (ph_number)
			end
		end

	is_bound_raw_num (ph_number: INTEGER): BOOLEAN
			-- Is placeholder number `ph_number' bound to a
			-- raw string?
		do
			if numeric_bindings /= Void then
				Result := numeric_bindings.is_bound_raw (ph_number)
			end
		end

	is_bound_date_num (ph_number: INTEGER): BOOLEAN
			-- Is placeholder number `ph_number' bound to a
			-- database date?
		do
			if numeric_bindings /= Void then
				Result := numeric_bindings.is_bound_date (ph_number)
			end
		end

	is_bound_rowid_num (ph_number: INTEGER): BOOLEAN
			-- Is placeholder number `ph_number' bound to a
			-- database rowid?
		do
			if numeric_bindings /= Void then
				Result := numeric_bindings.is_bound_rowid (ph_number)
			end
		end

	integer_bound_to_num (ph_number: INTEGER): INTEGER
			-- Value of integer bound to placeholder number
			-- `ph_number'
		do
			Result := numeric_bindings.integer_value (ph_number)
		end

	real_bound_to_num (ph_number: INTEGER): REAL
			-- Value of real bound to placeholder number
			-- `ph_number'
		do
			Result := numeric_bindings.real_value (ph_number)
		end

	double_bound_to_num (ph_number: INTEGER): DOUBLE
			-- Value of double bound to placeholder number
			-- `ph_number'
		do
			Result := numeric_bindings.double_value (ph_number)
		end

	character_bound_to_num (ph_number: INTEGER): CHARACTER
			-- Value of character bound to placeholder number
			-- `ph_number'
		do
			Result := numeric_bindings.character_value (ph_number)
		end

	boolean_bound_to_num (ph_number: INTEGER): BOOLEAN
			-- Value of boolean bound to placeholder number
			-- `ph_number'
		do
			Result := numeric_bindings.boolean_value (ph_number)
		end

	string_bound_to_num (ph_number: INTEGER): STRING
			-- Value of string bound to placeholder number
			-- `ph_number'
		do
			Result := numeric_bindings.string_value (ph_number)
		end

	raw_bound_to_num (ph_number: INTEGER): STRING
			-- Value of raw string bound to placeholder number
			-- `ph_number'
		do
			Result := numeric_bindings.raw_value (ph_number)
		end

	date_bound_to_num (ph_number: INTEGER): ORACLE_DATE
			-- Value of database date string bound to placeholder
			-- number `ph_number'
		do
			Result := numeric_bindings.date_value (ph_number)
		end

	rowid_bound_to_num (ph_number: INTEGER): ORACLE_ROWID
			-- Value of database rowid string bound to placeholder
			-- number `ph_number'
		do
			Result := numeric_bindings.rowid_value (ph_number)
		end

	value_bound_to_num (ph_number: INTEGER): ANY
			-- Value bound to placeholder number `ph_number'
		do
			Result := numeric_bindings.any_value (ph_number)
		end

feature -- Named placeholder properties
	
	is_bound_name (ph_name: STRING): BOOLEAN
			-- Is placeholder named `ph_name' bound to anything?
		do
			if named_bindings /= Void then
				Result := named_bindings.is_bound_any (ph_name)
			end
		end

	is_bound_null_name (ph_name: STRING): BOOLEAN
			-- Is placeholder named `ph_name' bound to null?
		do
			if named_bindings /= Void then
				Result := named_bindings.is_bound_null (ph_name)
			end
		end

	is_bound_integer_name (ph_name: STRING): BOOLEAN
			-- Is placeholder named `ph_name' bound to an
			-- integer?
		do
			if named_bindings /= Void then
				Result := named_bindings.is_bound_integer (ph_name)
			end
		end

	is_bound_real_name (ph_name: STRING): BOOLEAN
			-- Is placeholder named `ph_name' bound to a real?
		do
			if named_bindings /= Void then
				Result := named_bindings.is_bound_real (ph_name)
			end
		end

	is_bound_double_name (ph_name: STRING): BOOLEAN
			-- Is placeholder named `ph_name' bound to a double?
		do
			if named_bindings /= Void then
				Result := named_bindings.is_bound_double (ph_name)
			end
		end

	is_bound_character_name (ph_name: STRING): BOOLEAN
			-- Is placeholder named `ph_name' bound to a character?
		do
			if named_bindings /= Void then
				Result := named_bindings.is_bound_character (ph_name)
			end
		end

	is_bound_boolean_name (ph_name: STRING): BOOLEAN
			-- Is placeholder named `ph_name' bound to a boolean?
		do
			if named_bindings /= Void then
				Result := named_bindings.is_bound_boolean (ph_name)
			end
		end

	is_bound_string_name (ph_name: STRING): BOOLEAN
			-- Is placeholder named `ph_name' bound to a string?
		do
			if named_bindings /= Void then
				Result := named_bindings.is_bound_string (ph_name)
			end
		end

	is_bound_raw_name (ph_name: STRING): BOOLEAN
			-- Is placeholder named `ph_name' bound to a raw
			-- string?
		do
			if named_bindings /= Void then
				Result := named_bindings.is_bound_raw (ph_name)
			end
		end

	is_bound_date_name (ph_name: STRING): BOOLEAN
			-- Is placeholder named `ph_name' bound to a database
			-- date?
		do
			if named_bindings /= Void then
				Result := named_bindings.is_bound_date (ph_name)
			end
		end

	is_bound_rowid_name (ph_name: STRING): BOOLEAN
			-- Is placeholder named `ph_name' bound to a database
			-- rowid?
		do
			if named_bindings /= Void then
				Result := named_bindings.is_bound_rowid (ph_name)
			end
		end

	integer_bound_to_name (ph_name: STRING): INTEGER
			-- Value of integer bound to placeholder named
			-- `ph_name'
		do
			Result := named_bindings.integer_value (ph_name)
		end

	real_bound_to_name (ph_name: STRING): REAL
			-- Value of real bound to placeholder named
			-- `ph_name'
		do
			Result := named_bindings.real_value (ph_name)
		end

	double_bound_to_name (ph_name: STRING): DOUBLE
			-- Value of double bound to placeholder named
			-- `ph_name'
		do
			Result := named_bindings.double_value (ph_name)
		end

	character_bound_to_name (ph_name: STRING): CHARACTER
			-- Value of character bound to placeholder named
			-- `ph_name'
		do
			Result := named_bindings.character_value (ph_name)
		end

	boolean_bound_to_name (ph_name: STRING): BOOLEAN
			-- Value of boolean bound to placeholder named
			-- `ph_name'
		do
			Result := named_bindings.boolean_value (ph_name)
		end

	string_bound_to_name (ph_name: STRING): STRING
			-- Value of string bound to placeholder named
			-- `ph_name'
		do
			Result := named_bindings.string_value (ph_name)
		end

	raw_bound_to_name (ph_name: STRING): STRING
			-- Value of raw string bound to placeholder named
			-- `ph_name'
		do
			Result := named_bindings.raw_value (ph_name)
		end

	date_bound_to_name (ph_name: STRING): ORACLE_DATE
			-- Value of database date bound to placeholder named
			-- `ph_name'
		do
			Result := named_bindings.date_value (ph_name)
		end

	rowid_bound_to_name (ph_name: STRING): ORACLE_ROWID
			-- Value of database rowid bound to placeholder named
			-- `ph_name'
		do
			Result := named_bindings.rowid_value (ph_name)
		end

	value_bound_to_name (ph_name: STRING): ANY
			-- Value bound to placeholder named `ph_name'
		do
			Result := named_bindings.any_value (ph_name)
		end

feature -- Select item definition

	set_select_count (num: INTEGER)
			-- Set number of select items in query to `num'
		do
			if select_items = Void then
				create select_items.make (num, Current)
			else
				select_items.resize (num)
			end
		end

	select_integer (k: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- integers
		do
			make_select_items
			select_items.select_integer (k)
		end

	select_real (k: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- reals
		do
			make_select_items
			select_items.select_real (k)
		end

	select_double (k: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- doubles
		do
			make_select_items
			select_items.select_double (k)
		end

	select_character (k: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- characters
		do
			make_select_items
			select_items.select_character (k)
		end

	select_boolean (k: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- booleans
		do
			make_select_items
			select_items.select_boolean (k)
		end

	select_date (k: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- database dates
		do
			make_select_items
			select_items.select_date (k)
		end

	select_rowid (k: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- database rowids
		do
			make_select_items
			select_items.select_rowid (k)
		end

	select_string (k: INTEGER maxlen: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- strings of maximum length `maxlen'.  Zero means
			-- the implementation should consult the database
			-- engine to determine the maximum length
		do
			make_select_items
			select_items.select_string (k, maxlen)
		end

	select_raw (k: INTEGER maxlen: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- raw strings of maximum length `maxlen'.  Zero means
			-- the implementation should consult the database
			-- engine to determine the maximum length
		do
			make_select_items
			select_items.select_raw (k, maxlen)
		end

feature -- Select item properties

	select_item_capacity: INTEGER
			-- Number of select items for which `Current'
			-- has space
		do
			if select_items /= Void then
				Result := select_items.capacity
			end
		end

	end_of_data: BOOLEAN
			-- Has end of data been reached on a query?

	query_in_progress: BOOLEAN
			-- Does `Current' represent a query which is
			-- in progress (between `execute' and `terminate')?

	is_busy: BOOLEAN
			-- Was a required resource busy (i.e., was a
			-- row that needed to be locked already locked
			-- by another session) during last call to
			-- `execute_with_busy_status'?

	is_unique_key_error: BOOLEAN
			-- Did a unique constraint violation occur
			-- during last call to
			-- `execute_with_unique_status'?

	is_defined_select_position (k: INTEGER): BOOLEAN
			-- Is position `k' a valid select position?
		do
			Result := k >= 1 and k <= select_item_capacity
		end

	is_integer_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel integers?
		do
			Result := select_items.is_integer_column (k)
		end

	is_real_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel reals?
		do
			Result := select_items.is_real_column (k)
		end

	is_double_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel doubles?
		do
			Result := select_items.is_double_column (k)
		end

	is_character_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel characters?
		do
			Result := select_items.is_character_column (k)
		end

	is_boolean_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel characters?
		do
			Result := select_items.is_boolean_column (k)
		end

	is_date_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel database
			-- dates?
		do
			Result := select_items.is_date_column (k)
		end

	is_rowid_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel database
			-- rowids?
		do
			Result := select_items.is_rowid_column (k)
		end

	is_string_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel strings?
		do
			Result := select_items.is_string_column (k)
		end

	is_raw_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel raw
			-- strings?
		do
			Result := select_items.is_raw_column (k)
		end

feature -- Select item fetching

	is_null (pos: INTEGER): BOOLEAN
			-- Is the `pos'-th select item of the last fetched row
			-- null?
		do
			Result := select_items.is_null (pos)
		end
	
	integer_value (pos: INTEGER): INTEGER
			-- Integer value for `pos'-th select item of
			-- last fetched row
		do
			Result := select_items.integer_value (pos)
		end
	
	real_value (pos: INTEGER): REAL
			-- Real value for `pos'-th select item of
			-- last fetched row
		do
			Result := select_items.real_value (pos)
		end
	
	double_value (pos: INTEGER): DOUBLE
			-- Double value for `pos'-th select item of
			-- last fetched row
		do
			Result := select_items.double_value (pos)
		end
	
	character_value (pos: INTEGER): CHARACTER
			-- Character value for `pos'-th select item of
			-- last fetched row
		do
			Result := select_items.character_value (pos)
		end
	
	boolean_value (pos: INTEGER): BOOLEAN
			-- Boolean value for `pos'-th select item of
			-- last fetched row
		do
			Result := select_items.boolean_value (pos)
		end
	
	string_value (pos: INTEGER): STRING
			-- String value for `pos'-th select item of
			-- last fetched row
		do
			Result := select_items.string_value (pos)
		end
	
	raw_value (pos: INTEGER): STRING
			-- Raw string value for `pos'-th select item of
			-- last fetched row
		do
			Result := select_items.raw_value (pos)
		end
	
	date_value (pos: INTEGER): ORACLE_DATE
			-- Database date value for `pos'-th select item of
			-- last fetched row
		do
			Result := select_items.date_value (pos)
		end
	
	rowid_value (pos: INTEGER): ORACLE_ROWID
			-- Database rowid value for `pos'-th select item of
			-- last fetched row
		do
			Result := select_items.rowid_value (pos)
		end
	
	get_string_value (pos: INTEGER dest: STRING)
			-- Get string value for `pos'-th select item of
			-- last fetched row and store into `dest'
		do
			select_items.get_string_value (pos, dest)
		end
	
	get_raw_value (pos: INTEGER dest: STRING)
			-- Get raw string value for `pos'-th select item of
			-- last fetched row and store into `dest'
		do
			select_items.get_raw_value (pos, dest)
		end
	
	get_date_value (pos: INTEGER dest: like date_value)
			-- Get date value for `pos'-th select item of
			-- last fetched row and store into `dest'
		local
			date: ORACLE_DATE
		do
			date ?= dest
			if date /= Void then
				select_items.get_date_value (pos, date)
			else
				raise (dbex.Db_conformance_error)
			end
		end
	
	get_rowid_value (pos: INTEGER dest: like rowid_value)
			-- Get rowid value for `pos'-th select item of
			-- last fetched row and store into `dest'
		local
			rowid: ORACLE_ROWID
		do
			rowid ?= dest
			if rowid /= Void then
				select_items.get_rowid_value (pos, rowid)
			else
				raise (dbex.Db_conformance_error)
			end
		end
	
feature -- Modification

	set_statement_type (type: INTEGER)
			-- Set statement type to `type'
		require
			valid_statement_type: valid_statement_type (type)
		do
			statement_type := type
		ensure
			type_set: statement_type = type
		end
	
feature -- Operations

	set_query_buffer_size_limit (value: INTEGER)
			-- Set `query_buffer_size_limit' to `value'
		do
			make_select_items
			select_items.set_requested_buffer_limit (value)
		end
	
	set_query_rows_per_fetch (value: INTEGER)
			-- Set `query_rows_per_fetch' to `value'
		do
			make_select_items
			select_items.set_requested_row_capacity (value)
		end
	
	describe
			-- Construct list of select items for query
			-- and put in `query_select_items'.  Useful
			-- for determining number of select items in
			-- dynamically generated SQL, as well as the
			-- type, max length and max display length of each
			-- item
		local
			k, retcode: INTEGER
			done: BOOLEAN
			item: ORACLE_DESCRIBE_ITEM
		do
			create {ARRAYED_LIST [ORACLE_DESCRIBE_ITEM]} query_select_items.make (20)
			from
				k := 1
			until
				done
			loop
				create item.make
				item.describe_without_error (Current, k)
				retcode := item.return_code
				if retcode = Oci_success_code then
					query_select_items.extend (item)
				elseif return_code = Oci_invalid_select_item_code then
					-- Describe failed - done if cursor
					-- return code "no such select item"
					done := True
				else
					-- Otherwise signal error
					check_status (retcode)
				end
				k := k + 1
			end
		end

	execute
			-- Execute the SQL statement associated with
			-- `Current'.
		local
			code: INTEGER
		do
			execute_with_status_only
			code := status_code
			if code /= Oci_success_code and return_code = Oci_resource_busy_code then
				code := retry_execute
			end
			check_status (code)
			is_busy := False
			if select_items /= Void then
				select_items.execute
			end
			if is_query_statement then
				end_of_data := False
				query_in_progress := True
			end
		end

	execute_with_busy_status
			-- Execute the SQL statement associated with
			-- `Current'.  Set `is_busy' to indicate whether
			-- a required resource was busy, instead of raising
			-- an exception if a "busy" condition is detected
		local
			code: INTEGER
		do
			execute_with_status_only
			code := status_code
			is_busy := code /= Oci_success_code and return_code = Oci_resource_busy_code
			if is_busy then
				if is_query_statement then
					end_of_data := True
					query_in_progress := False
				end
			else
				check_status (code)
				if select_items /= Void then
					select_items.execute
				end
				if is_query_statement then
					end_of_data := False
					query_in_progress := True
				end
			end
		end

	execute_with_unique_status
			-- Execute the SQL statement associated with
			-- `Current'.  Set `is_unique_key_error' to indicate
			-- whether a unique constraint violation
			-- occurred, instead of raising an exception
			-- if a unique constraint violation is
			-- detected
		local
			code: INTEGER
		do
			execute_with_status_only
			code := status_code
			if code /= Oci_success_code and return_code = Oci_unique_constraint_code then
				is_unique_key_error := True
			else
				check_status (code)
				is_busy := False
				if select_items /= Void then
					select_items.execute
				end
				if is_query_statement then
					end_of_data := False
					query_in_progress := True
				end
			end
		end

	fetch
			-- Fetch next row for a query
		do
			select_items.fetch
			end_of_data := select_items.end_of_data
			if end_of_data then
				query_in_progress := False
			end
		end

	unbind
			-- Unbind all placeholders for this cursor,
			-- leaving it open and parsed.
		do
			if numeric_bindings /= Void then
				numeric_bindings.unbind
			end
			if named_bindings /= Void then
				named_bindings.unbind
			end
		end

	undefine_all
			-- Undefine all select items for this cursor,
			-- leaving it open and parsed.
		do
			if select_items /= Void then
				select_items.undefine_all
			end
		end

	terminate
			-- Terminate query if it is in progress.  No
			-- effect if query not in progress
		local
			code: INTEGER
		do
			if query_in_progress then
				enter_oci
				code := ora_cancel_query (address)
				leave_oci
				check_status (code)
			end
			end_of_data := True
			query_in_progress := False
			if select_items /= Void then
				select_items.terminate
			end
		end

	release
			-- Release this cursor for use by others,
			-- leaving it open and parsed.  Notify the cursor
			-- server of the state change.
		do
			if is_query_statement or is_dml_statement then
				if is_query_statement then
					terminate
					undefine_all
				end
				query_in_progress := False
				unbind
				set_state (cursor_parsed)
				connection.release_cursor (Current)
			else
				-- DDL statements must be closed when caller
				-- requests a release because it is the
				-- parse which actually executes the
				-- statement
				query_in_progress := False
				close
			end
		ensure then
			std_cursor_parsed: not old is_ddl_statement implies is_parsed or is_closed
				-- Cursor server may decide to close cursor
				-- when it is released
			ddl_cursor_closed: old is_ddl_statement implies is_closed
		end

	close
			-- Release this cursor for use by others and
			-- close it to free any associated resources allocated
			-- by Oracle.  Notify the cursor server of the
			-- state change.
		do
			raw_close
			connection.close_cursor (Current)
		ensure then
			cursor_closed: is_closed
		end

	free
			-- Release this cursor for use by others and
			-- close it to free any associated resources allocated
			-- by Oracle.  Notify the cursor server for this
			-- cursor's database connection that the cursor
			-- Eiffel object may be freed.
		do
			raw_close
			connection.free_cursor (Current)
		ensure then
			cursor_closed: is_closed
		end

	append_bindings_description (dest: STRING)
			-- Append to `dest' information about the named and
			-- numeric placeholders associated with `Current',
			-- if any
		do
			if numeric_bindings /= Void and then numeric_bindings.count > 0 then
				dest.extend ('%N')
				dest.append ("Numeric bindings")
				dest.extend ('%N')
				numeric_bindings.append_description (dest)
			end
			if named_bindings /= Void and then named_bindings.count > 0 then
				dest.extend ('%N')
				dest.append ("Named bindings")
				dest.extend ('%N')
				named_bindings.append_description (dest)
			end
		end

feature {ORACLE_CURSOR_SERVER} -- Implementation

	raw_close
			-- Release this cursor for use by others and
			-- close it to free any associated resources allocated
			-- by Oracle.  Does not notify cursor server of
			-- change in status.
		require
			cursor_open: not is_closed
		local
			code: INTEGER
		do
			set_state (Cursor_closed)
			sql_statement := Void
			numeric_bindings := Void
			named_bindings := Void
			select_items := Void
			enter_oci
			code := ora_close_cursor (address)
			leave_oci
			check_status (code)
		ensure
			cursor_closed: is_closed
		end

feature {NONE} -- Implementation

	statement_type: INTEGER
			-- Type of SQL statement associated with this cursor

	Nondeferred_mode: INTEGER = 0
			-- Constant to indicate to Oracle that parse should
			-- not be deferred (should be done now)

	status_code: INTEGER
			-- Return code from last call to
			-- `execute_with_status_only'

	make_numeric_bindings
			-- If `numeric_bindings' does not exist, create it
			-- with default number of placeholders
		do
			if numeric_bindings = Void then
				create numeric_bindings.make (Default_placeholder_count, Current)
			end
		ensure
			numeric_bindings_exist: numeric_bindings /= Void
		end

	make_named_bindings
			-- If `named_bindings' does not exist, create it
			-- with default number of placeholders
		do
			if named_bindings = Void then
				create named_bindings.make (Default_placeholder_count, Current)
			end
		ensure
			named_bindings_exist: named_bindings /= Void
		end

	make_select_items
			-- If `select_items' does not exist, create it
			-- with default number of select items
		do
			if select_items = Void then
				create select_items.make (Default_select_item_count, Current)
			end
		ensure
			select_items_exist: select_items /= Void
		end

	execute_with_status_only
			-- Execute the SQL statement associated with
			-- `Current' and put the return code from
			-- the execution in `status_code'
		do
			connection.check_access_restrictions (sql_statement,
				statement_type)
			if (numeric_bindings /= Void and then numeric_bindings.reparse_required) or
			   (named_bindings /= Void and then named_bindings.reparse_required) or
			   (select_items /= Void and then select_items.reparse_required) then
				reparse
				if numeric_bindings /= Void then
					numeric_bindings.set_reparse_required (False)
					numeric_bindings.set_needs_oci_bind
				end
				if named_bindings /= Void then
					named_bindings.set_reparse_required (False)
					named_bindings.set_needs_oci_bind
				end
				if select_items /= Void then
					select_items.set_reparse_required (False)
					select_items.set_needs_oci_define
				end
			end
			if numeric_bindings /= Void then
				numeric_bindings.check_for_unbound_placeholders
			end
			if named_bindings /= Void then
				named_bindings.check_for_unbound_placeholders
			end
			if select_items /= Void then
				select_items.check_for_undefined_select_items
			end
			if numeric_bindings /= Void then
				numeric_bindings.oci_bind
			end
			if named_bindings /= Void then
				named_bindings.oci_bind
			end
			if select_items /= Void then
				select_items.fill_in_zero_lengths
				select_items.oci_define
			end
			enter_oci
			status_code := ora_execute_sql_statement (address)
			leave_oci
		end

	retry_execute: INTEGER
			
			-- A call to `ora_execute_sql_statement' returned
			-- a non-success code and the return code in the
			-- Cursor Data Area was `Oci_resource_busy_code'.
			-- Retry the OCI execute with wait intervals according
			-- to the wait options specified for the connection.
			-- Keep retrying until the execute suceeds or the
			-- return code is not `Oci_resource_busy_code' or
			-- until either the maximum number of retries has
			-- been attempted or the maximum total wait time
			-- is reached.  Return the OCI return code from
			-- the last execute attempt
		local
			conn: ORACLE_CONNECTION
			retries: INTEGER
			wait, total_wait: DOUBLE
			max_retries: INTEGER
			max_wait, max_total_wait, notify_time: DOUBLE
			const_coeff, linear_coeff, exp_coeff, base: DOUBLE
			notified: BOOLEAN
		do
			Result := Oci_resource_busy_code
				-- Anything but `Oci_success_code'
			conn := connection
			max_retries := conn.lock_fail_max_retries
			max_total_wait := conn.lock_fail_max_total_wait
			max_wait := conn.lock_fail_max_wait
			notify_time := conn.lock_fail_notify_time
			const_coeff := conn.lock_fail_constant_coeff
			linear_coeff := conn.lock_fail_linear_coeff
			exp_coeff := conn.lock_fail_exponential_coeff
			base := conn.lock_fail_exponential_base
			from
			until retries >= max_retries or
			      total_wait >= max_total_wait or
			      Result = Oci_success_code or
			      (Result /= Oci_success_code and return_code /= Oci_resource_busy_code)
			loop
				retries := retries + 1
					-- Number of this retry
				wait := exp_coeff * (base ^ (retries - 1)) + linear_coeff * retries + const_coeff
				wait := wait.min (max_wait)
				wait := wait.min (max_total_wait - total_wait)
				wait := wait.max (1.0E-6)
					-- Minimum wait is 1 microsecond
				if notify_time >= total_wait and notify_time < total_wait + wait and not conn.status_display_suppressed then
					notify_user_about_wait (total_wait)
					notified := True
				end
				os.sleep_seconds (wait)
				enter_oci
				Result := ora_execute_sql_statement (address)
				leave_oci
				total_wait := total_wait + wait
			end
			if notified then
				notify_user_wait_ended (retries, max_retries, total_wait, max_total_wait, Result, return_code)
			end
		end

	notify_user_about_wait (total_wait: DOUBLE)
			-- Notify user that we've waited `total_wait' seconds
			-- for database resource which is still busy
		local
			s: STRING
		do
			create s.make (255)
			s.extend ('%N')
			s.append (os.current_date_string)
			s.extend ('%N')
			s.append ("SQL statement:%N")
			su.append_string_indented (Statement_prefix, sql_statement, 0, s)
			s.append ("Database resource still busy after ")
			s.append_double (total_wait)
			s.append (" seconds - waiting ...")
			io.put_string (s)
			io.standard_default.flush
		end

	Statement_prefix: STRING = "    "
			-- Prefix displayed before each line of a SQL
			-- statement which is displayed because of a
			-- "resource busy" condition
	
	notify_user_wait_ended (retries, max_retries: INTEGER total_wait, max_total_wait: DOUBLE func_code, retcode: INTEGER)
			-- Notify user that we're done waiting for a database
			-- resource which was busy, and why
		local
			s: STRING
		do
			create s.make (255)
			s.extend (' ')
			if func_code = Oci_success_code then
				s.append ("[acquired]")
			elseif retcode /= Oci_resource_busy_code then
				s.append ("[error]")
			else
				s.append ("[gave up]")
				s.extend ('%N')
				s.append ("	(still busy after ")
				s.append_integer (retries)
				s.append (" retries and ")
				s.append_real (total_wait)
				s.append (" seconds waiting)")
			end
			s.extend ('%N')
			io.put_string (s)
		end

feature {ORACLE_CURSOR_SERVER, ORACLE_CURSOR} -- Timestamp

	time_parse_requested: INTEGER
			-- Latest "time" when parse was requested for
			-- `sql_statement'.  Time is measured in seconds
			-- since the start of the epoch (00:00:00 GMT,
			-- Jan.  1,  1970)

	set_time_parse_requested (time: INTEGER)
			-- Set `time_parse_requested' to `time'
		require
			positive_time: time > 0
		do
			time_parse_requested := time
		end

feature -- Properties

	connection: ORACLE_CONNECTION
			-- Connection with which `Current' is associated

	sql_statement: STRING
			-- SQL statement associated with this cursor
	
	last_rowid: ORACLE_ROWID
			-- Rowid of last row inserted, updated, deleted
			-- or selected for update
		local
			first_half, second_half: DOUBLE
		do
			first_half := internal_rowid_first_half
			second_half := internal_rowid_second_half
			create Result.make_internal (first_half, second_half)
		end
	
	is_ddl_statement: BOOLEAN
			-- Is associated SQL statement a Data Definition
			-- Language (DDL) statement?
		do
			Result := statement_type = Ddl_statement
		end

	is_dml_statement: BOOLEAN
			-- Is associated SQL statement a Data Manipulation
			-- Language (DML) statement?  Queries are not
			-- considered DML statements.
		do
			Result := statement_type = Dml_statement
		end

	is_query_statement: BOOLEAN
			-- Is associated SQL statement a query?
		do
			Result := statement_type = Query_statement
		end

	is_block_statement: BOOLEAN
			-- Is associated SQL statement a PL/SQL block?
		do
			Result := statement_type = Block_statement
		end

	
	is_bindable_statement: BOOLEAN
			-- Is `Current' a statement which allows bind
			-- variables?
		do
			inspect statement_type
			when Query_statement, Dml_statement, Block_statement then
				Result := True
			else
				-- Not bindable
			end
		end

	consistent_with_ddl_statement: BOOLEAN
			-- Is SQL statement type indicated in
			-- Cursor Data Area consistent with a DDL statement?
		do
			Result := sql_function_code = Unknown_operation_code
				or (not consistent_with_dml_statement and not consistent_with_query and not consistent_with_block_statement)
		end

	consistent_with_dml_statement: BOOLEAN
			-- Is SQL statement type indicated in
			-- Cursor Data Area consistent with a DML statement?
		local
			code: INTEGER
		do
			code := sql_function_code
			inspect code
			when Insert_code, Update_code, Delete_code,
			     Explain_code, Set_transaction_code,
			     Commit_code, Rollback_code, Savepoint_code,
			     Lock_table_code, Set_role_code, Alter_user_code,
			     Alter_session_code, Alter_system_switch_log_code,
			     Unknown_operation_code then
				Result := True
			else
				Result := False
			end
		end

	consistent_with_block_statement: BOOLEAN
			-- Is SQL statement type indicated in
			-- Cursor Data Area consistent with a PL/SQL block?
		do
			Result :=  sql_function_code = Plsql_execute_code
		end

	consistent_with_query: BOOLEAN
			-- Is SQL statement type indicated in
			-- Cursor Data Area consistent with a query?
		do
			Result := sql_function_code = Select_code
		end

	is_reserved: BOOLEAN
			-- Is cursor reserved for use by `Current'?
		do
			Result := state = Cursor_reserved and
			   sql_statement /= Void
		end

	is_parsed: BOOLEAN
			-- Is cursor open and associated with a parsed
			-- SQL statement, but not reserved?
		do
			Result := state = Cursor_parsed and
			   sql_statement /= Void
		end

	is_open: BOOLEAN
			-- Is cursor open, but not associated with a
			-- parsed SQL statement?
		do
			Result := state = Cursor_open and sql_statement = Void
		end

	is_closed: BOOLEAN
			-- Is cursor closed?
		do
			Result := state = Cursor_closed and
			   sql_statement = Void and numeric_bindings = Void
			   and named_bindings = Void
		end

	numeric_bindings: ORACLE_NUMERIC_BINDSET
			-- Bindings of numeric placeholders like :3
	
	numeric_bind_capacity: INTEGER
			-- Number of numeric placeholders for which `Current'
			-- has space
		do
			if numeric_bindings /= Void then
				Result := numeric_bindings.capacity
			end
		end
	
	named_bindings: ORACLE_NAMED_BINDSET
			-- Bindings of named placeholders like :design_no
	
	named_bind_capacity: INTEGER
			-- Number of named placeholders for which `Current'
			-- has space
		do
			if named_bindings /= Void then
				Result := named_bindings.capacity
			end
		end

	query_buffer_size: INTEGER
			-- Size of buffer for query results
		do
			Result := select_items.output_buffer_capacity
		end

	query_buffer_size_limit: INTEGER
			-- Requested maximum size of buffer to fetch
			-- query results.  This is only a hint, so
			-- the buffer size may be larger.
			-- Zero means no limit.
		do
			if select_items /= Void then
				Result := select_items.output_buffer_limit
			else
				Result := connection.query_buffer_size_limit
			end
		end

	query_rows_per_fetch: INTEGER
			-- Requested number of rows to fetch from database
			-- engine each time a fetch is needed.  Zero means
			-- use `query_buffer_size_limit' to determine
			-- the number of rows
		do
			if select_items /= Void then
				Result := select_items.requested_row_capacity
			else
				Result := connection.query_rows_per_fetch
			end
		end

	select_items: ORACLE_SELECT_SET
			-- Definitions of select items for a query
	
	Default_placeholder_count: INTEGER = 10
			-- Default number of placeholders to allocate in
			-- a named or numeric bindset, if a bind is
			-- requested before placeholders have been allocated
	
	Default_select_item_count: INTEGER = 10
			-- Default number of select items to allocate in
			-- a select item set, if a select is
			-- requested before select items have been allocated
	
	query_select_items: LIST [ORACLE_DESCRIBE_ITEM]
			-- List of descriptions of select items for
			-- query computed by last call to `describe',
			-- in order items appear in query
			-- (useful for dynamic SQL)

feature -- Display

	display
			-- Display `Current'
		local
			current_time: INTEGER
		do
			current_time := unix_os.current_time_in_seconds
			io.put_string ("State: ") 
			io.put_integer (state)
			io.put_string ("  Age: ")
			io.put_integer (current_time - time_parse_requested) 
			io.new_line
			if sql_statement /= Void then
				io.put_string ("SQL: ")
				io.put_string (sql_statement)
				io.new_line
			end
		end

feature -- Comparison

	infix "<" (other: like Current): BOOLEAN
		do
			Result := time_parse_requested > other.time_parse_requested
		end

	is_equal (other: like Current): BOOLEAN
		do
			Result := time_parse_requested = other.time_parse_requested
		end

feature -- Status checking

	check_status (code: INTEGER)
			-- Check that return status `code' from OCI routine is
			-- OK and call error handler if not
		do
			if code /= Oci_success_code and then connection.error_checking_enabled then
				connection.set_last_error_code (return_code)
				connection.handle_oci_error (Current)
			end
		end

	check_code (code: INTEGER)
			-- Check that CDA return code or column return
			-- code `code' (which is NOT a value returned
			-- from an OCI routine call) is OK and call
			-- error handler if not
		do
			if code /= Oci_success_code then
				connection.set_last_error_code (code)
				connection.handle_oci_error_with_code (Current, code)
			end
		end

	handle_oracle_error (except, message: STRING)
			-- Handle an Oracle error which is detected by
			-- us rather than by the Oracle Call Interface,
		do
			connection.handle_oracle_user_error (Current, except, message)
		end

feature -- Interrupt handling
	
	enter_oci
			-- Take actions needed when about to enter
			-- an Oracle Call Interface routine
		do
			connection.enter_oci
		end

	leave_oci
			-- Take actions needed when we have just left
			-- an Oracle Call Interface routine
		do
			connection.leave_oci
		end

feature {NONE}	-- Cleanup

	lda_address: POINTER
			-- Address of LDA for connection with which
			-- cursor is associated

	dispose
			-- Close cursor if it is open and return memory
			-- allocated for CDA
		local
			null_ptr: POINTER
			code: INTEGER
		do
			if state /= Cursor_closed then
				if is_allocated_lda (lda_address) then
					block_interrupts
					code := ora_close_cursor (address)
						-- Do not raise an exception,
						-- which would mess up the
						-- garbage collector
					set_state (Cursor_closed)
					unblock_interrupts
				else
					set_state (Cursor_closed)
				end
			end
			if address /= null_ptr then
				free_nongc_memory (address)
				address := null_ptr
			end
		end

	block_interrupts
			-- Block delivery of interrupts
		external
			"C"
		end
	
	unblock_interrupts
			-- Unblock delivery of interrupts
		external
			"C"
		end
	
	Cursor_not_released_exception: POINTER
			-- Address of null terminated string which serves
			-- as tag for a "Cursor_not_released" exception
		external
			"C macro use %"oracle.h%""
		end
	
	Cursor_close_failed_exception: POINTER
			-- Address of null terminated string which serves
			-- as tag for a "Cursor_close_failed" exception
		external
			"C macro use %"oracle.h%""
		end
	
invariant
	valid_state: is_valid_state (state)

end
