
indexing
	
	author: "David Hollenberg"
	date:	"November 28, 1995"
	description: "An Oracle cursor server.  The server manages cursors, %
		%allowing them to be reused, and can allocate a parsed %
		%cursor for an SQL statement."

class ORACLE_CURSOR_SERVER

inherit
	UNIX_OS_ACCESS
	ORACLE_CURSOR_STATES
		export
			{NONE} all
		end
	MEMORY
		export
			{NONE} all
		end

create
	make

feature {NONE} -- Creation

	make (conn: ORACLE_CONNECTION)
		do
			connection := conn
			max_closed_cursors := Default_max_closed_cursors
			max_open_cursors := Default_max_open_cursors
			max_parsed_cursors := Default_max_parsed_cursors
			closed_age_limit := Default_closed_age_limit
			open_age_limit := Default_open_age_limit
			parsed_age_limit := Default_parsed_age_limit
			age_check_interval := Default_age_check_interval
			create closed_cursors.make
			create open_cursors.make
			create parsed_cursors.make
			create reserved_cursors.make (10)
		end
			
feature -- Properties

	connection: ORACLE_CONNECTION
			-- Connection for which server manages cursors

	error_on_connection: BOOLEAN
			-- Has an error occurred on this connection
			-- to which the server has not yet reacted?

	max_closed_cursors: INTEGER
			-- Maximum number of closed cursors to keep

	max_open_cursors: INTEGER
			-- Maximum number of open cursors to keep
	
	max_parsed_cursors: INTEGER
			-- Maximum number of open and parsed cursors to keep

	closed_age_limit: INTEGER
			-- Closed cursor with age in seconds at or above
			-- this limit will be freed next time server
			-- has control
	
	open_age_limit: INTEGER
			-- Open cursor with age in seconds at or above
			-- this limit will be closed next time server
			-- has control

	parsed_age_limit: INTEGER
			-- Open and parsed cursor with age in seconds at
			-- or above this limit will be closed next time server
			-- has control

	age_check_interval: INTEGER
			-- Minimum interval in seconds between checks for
			-- cursors which are at or above their age limit

	cached_open_cursor_count: INTEGER
			-- Number of cursors that are open and cached
			-- within cursor server for future use
		do
			Result := open_cursors.count
		end

	cached_parsed_cursor_count: INTEGER
			-- Number of cursors that are open and parsed
			-- and cached within cursor server for future use
		do
			Result := parsed_cursors.count
		end

	available_cursor_count: INTEGER
			-- Number of additional cursors that could be
			-- opened before running into database
			-- server's limit
		local
			list: ARRAYED_LIST [ORACLE_CURSOR]
			cursor: ORACLE_CURSOR
			failed: BOOLEAN
		do
			-- Keep opening cursors until we get a failure
			create list.make (50)
			from
			until
				failed
			loop
				create cursor.make (connection)
				cursor.try_to_open
				if cursor.is_open then
					list.extend (cursor)
					Result := Result + 1
				else
					failed := True
				end
			end

			-- Close all cursors we opened
			from
				list.start
			until
				list.after
			loop
				cursor := list.item
				cursor.raw_close
				list.forth
			end
			
		end

	reserved_cursor_count: INTEGER
			-- Number of cursors that have been reserved
			-- but not yet released, closed or freed
			-- (i.e., the number of cursors still in use).
			-- Cursors lost and closed due to garbage
			-- collection do not decrement this count

	keep_reserved_cursor_list: BOOLEAN
			-- Should reserved cursors be kept on list
			-- `reserved_cursors' until they are released,
			-- closed or freed?  If so, reserved cursors can
			-- never become garbage so they won't be closed
			-- if client loses track of them

	reserved_cursors_description: STRING
			-- Text of the SQL statements of all reserved
			-- cursors.  If `keep_reserved_cursor_list'
			-- is false, will only tell number of reserved
			-- statements
		local
			list: ARRAYED_LIST [ORACLE_CURSOR]
			k, len: INTEGER
			cursor: ORACLE_CURSOR
		do
			from
				list := reserved_cursors
				len := 0
				list.start
			until
				list.after
			loop
				cursor := list.item
				if cursor.sql_statement /= Void then
					len := len + cursor.sql_statement.count + 1
				else
					len := len + 10
				end
				list.forth
			end
			
			create Result.make (len)
			Result.append_integer (reserved_cursor_count)
			Result.append (" cursors in use%N")
			from
				list := reserved_cursors
				k := 1
				list.start
			until
				list.after
			loop
				cursor := list.item
				Result.extend ('(')
				Result.append_integer (k)
				Result.append (") Parse date: ")
				unix_os.append_date_string_from_integer (cursor.time_parse_requested, Result)
				Result.append ("   State: ")
				Result.append (text_for_state (cursor.state))
				Result.extend ('%N')
				if cursor.sql_statement /= Void then
					Result.append (cursor.sql_statement)
				else
					Result.append ("<Void>")
				end
				Result.extend ('%N')
				k := k + 1
				list.forth
			end
		ensure
			result_exists: Result /= Void
		end

	Default_max_closed_cursors: INTEGER = 100

	Default_max_open_cursors: INTEGER = 10

	Default_max_parsed_cursors: INTEGER = 200

	Default_closed_age_limit: INTEGER = 600

	Default_open_age_limit: INTEGER = 600

	Default_parsed_age_limit: INTEGER = 600

	Default_age_check_interval: INTEGER = 1

feature -- Modification

	set_max_closed_cursors (value: INTEGER)
			-- Set `max_closed_cursors' to `value'
		require
			non_negative_value: value >= 0
		do
			max_closed_cursors := value
		ensure
			value_set: max_closed_cursors = value
		end

	set_max_open_cursors (value: INTEGER)
			-- Set `max_open_cursors' to `value'
		require
			non_negative_value: value >= 0
		do
			max_open_cursors := value
		ensure
			value_set: max_open_cursors = value
		end

	set_max_parsed_cursors (value: INTEGER)
			-- Set `max_parsed_cursors' to `value'
		require
			non_negative_value: value >= 0
		do
			max_parsed_cursors := value
		ensure
			value_set: max_parsed_cursors = value
		end

	set_closed_age_limit (value: INTEGER)
			-- Set `closed_age_limit' to `value'
		require
			non_negative_value: value >= 0
		do
			closed_age_limit := value
		ensure
			value_set: closed_age_limit = value
		end

	set_open_age_limit (value: INTEGER)
			-- Set `open_age_limit' to `value'
		require
			non_negative_value: value >= 0
		do
			open_age_limit := value
		ensure
			value_set: open_age_limit = value
		end

	set_parsed_age_limit (value: INTEGER)
			-- Set `parsed_age_limit' to `value'
		require
			non_negative_value: value >= 0
		do
			parsed_age_limit := value
		ensure
			value_set: parsed_age_limit = value
		end

	set_age_check_interval (value: INTEGER)
			-- Set `age_check_interval' to `value'
		require
			non_negative_value: value >= 0
		do
			age_check_interval := value
		ensure
			value_set: age_check_interval = value
		end

	set_keep_reserved_cursor_list (b: BOOLEAN)
			-- Set `keep_reserved_cursor_list' to `b'
		do
			keep_reserved_cursor_list := b
			if not b then
				reserved_cursors.wipe_out
				reserved_cursor_count := 0
			end
		ensure
			value_set: keep_reserved_cursor_list = b
		end

feature -- Cursor operations

	parsed_cursor (sql_stmt: STRING): ORACLE_CURSOR
			-- An open and parsed Oracle cursor for SQL statement
			-- `sql_stmt'
		require
			statement_exists: sql_stmt /= Void
		local
			cursors: SORTED_TWO_WAY_LIST [ORACLE_CURSOR]
			save, tried, gc_done, open_in_progress: BOOLEAN
		do
			if not tried then
				process_connection_error
				from
					cursors := parsed_cursors
					cursors.start
				until
					cursors.exhausted or else equal (sql_stmt, cursors.item.sql_statement)
				loop
					cursors.forth
				end
				if not cursors.exhausted then -- Found matching stmt
					Result := cursors.item
					cursors.remove
				elseif not open_cursors.is_empty then
					open_cursors.start
					Result := open_cursors.item
					open_cursors.remove
					Result.parse (sql_stmt)
				else
					if closed_cursors.is_empty then
						create Result.make (connection)
					else
						closed_cursors.start
						Result := closed_cursors.item
						closed_cursors.remove
					end
					open_in_progress := True
					save := connection.error_display_suppressed
					connection.set_error_display_suppressed (True)
					Result.open
					connection.set_error_display_suppressed (save)
					open_in_progress := False
					Result.parse (sql_stmt)
				end
			elseif not parsed_cursors.is_empty then	
				-- Open failed but there are parsed cursors
				-- which are not reserved.  Use last one.
				parsed_cursors.finish
				Result := parsed_cursors.item
				parsed_cursors.remove
				Result.raw_close
				Result.open
				Result.parse (sql_stmt)
			else	
				-- Cannot find already parsed cursor for
				-- statement and there are no already open
				-- cursors and no parsed cursors for different
				-- statements and open failed for new cursor
				-- even after a GC
				connection.handle_oci_error (Result)
			end
			Result.reserve
			Result.set_time_parse_requested (unix_os.current_time_in_seconds)
			process_call_event
			add_reserved_cursor (Result)
		ensure
			result_exists: Result /= Void
			result_reserved: Result.is_reserved
			valid_cursor_lists: valid_cursor_lists
		rescue
			if Result /= Void then
				add_to_correct_list (Result)
			end
			if open_in_progress then
				open_in_progress := False
				connection.set_error_display_suppressed (save)
				if not gc_done then
					full_collect
					gc_done := True
				else
					tried := True
				end
				retry
			end
		end
	
	release_cursor (cursor: ORACLE_CURSOR)
			-- Notify server that `cursor' is no longer in
			-- use.  Cursor server may leave cursor open
			-- and parsed or close it or free it now or
			-- later, at its discretion.
		require
			cursor_exists: cursor /= Void
		local
			item: ORACLE_CURSOR
		do
			process_call_event
			remove_reserved_cursor (cursor)
			parsed_cursors.extend (cursor)
			if parsed_cursors.count > max_parsed_cursors then
				parsed_cursors.finish
				item := parsed_cursors.item
				parsed_cursors.remove
				item.raw_close
				closed_cursors.extend (item)
			end
		ensure
			valid_cursor_lists: valid_cursor_lists
		rescue
			if item /= Void then
				add_to_correct_list (item)
			end
		end
	
	close_cursor (cursor: ORACLE_CURSOR)
			-- Notify cursor server that cursor `cursor' has
			-- been closed.  Cursor server may keep cursor
			-- around and reopen it later or free it now or later,
			-- at its discretion.
		require
			cursor_exists: cursor /= Void
		do
			process_call_event
			remove_reserved_cursor (cursor)
			closed_cursors.extend (cursor)
			if closed_cursors.count > max_closed_cursors then
				closed_cursors.finish
				closed_cursors.remove
			end
		ensure
			valid_cursor_lists: valid_cursor_lists
		end
	
	free_cursor (cursor: ORACLE_CURSOR)
			-- Notify cursor server that cursor `cursor' has
			-- been closed and must be freed now.
		require
			cursor_exists: cursor /= Void
		do
			process_call_event
			remove_reserved_cursor (cursor)
		ensure
			valid_cursor_lists: valid_cursor_lists
		end
	
	flush_cursor_cache
			-- Flush all cursors from the cache served by
			-- this cursor server (reserved cursors are
			-- not in the cache).  This includes closing any
			-- open or parsed cursors and freeing all closed
			-- cursors.  Also, do a full garbage collection
			-- to ensure that any cursors which have become
			-- garbage are closed.
		do
			closed_cursors.wipe_out
			close_all_cursors (open_cursors)
			close_all_cursors (parsed_cursors)
			open_cursors.wipe_out
			parsed_cursors.wipe_out
			close_garbage_cursors
		end

feature -- Event processing
	
	process_call_event
			-- Do any processing which should be done on
			-- every call to the cursor server
		do
			process_old_cursors
			process_connection_error
		end

	process_elapsed_time_event
			-- Do any processing which is time-sensitive,
			-- such as closing or freeing cursors which have
			-- exceeded their age limit
		do
			process_call_event
		end

	process_logoff_event
			-- Do any processing which needs to be done
			-- when database connection served by this server
			-- is closed
		do
			flush_cursor_cache
		end

	process_error_event
			-- Do any processing which needs to be done
			-- when an error occurs on the database connection
			-- served by this server
		do
			error_on_connection := True
		end

feature {NONE} -- Implementation

	display_lists
			-- Display all cursor lists
		do
			display_list (closed_cursors, "Closed cursors")
			display_list (open_cursors, "Open cursors")
			display_list (parsed_cursors, "Parsed cursors")
		end

	display_list (list: LIST [ORACLE_CURSOR] name: STRING)
			-- Display cursor list `list'
		local
		do
			if not list.is_empty then
				io.put_string (name) io.put_string (": ")
				io.put_integer (list.count) io.new_line
				from
					list.start
				until
					list.exhausted
				loop
					list.item.display
					list.forth
				end
			end
		end

	add_to_correct_list (cursor: ORACLE_CURSOR)
			-- Add `cursor' to the correct list for its state,
			-- if it is not already there
		local
			list: LIST [ORACLE_CURSOR]
		do
			inspect cursor.state
			when Cursor_closed then
				list := closed_cursors
			when Cursor_open then
				list := open_cursors
			when Cursor_parsed then
				list := parsed_cursors
			when Cursor_reserved then
				list := Void
			end
			if list /= Void and not list.has (cursor) then
				list.extend (cursor)
			end
		end

	add_reserved_cursor (cursor: ORACLE_CURSOR)
			-- Increment count of reserved cursors.
			-- Add `cursor' to `reserved_cursors' if
			-- `keep_reserved_cursor_list' is true.
		do
			reserved_cursor_count := reserved_cursor_count + 1
			if keep_reserved_cursor_list then
				reserved_cursors.extend (cursor)
			end
		end

	remove_reserved_cursor (cursor: ORACLE_CURSOR)
			-- Decrement count of reserved cursors.
			-- Remove `cursor' from `reserved_cursors' if
			-- `keep_reserved_cursor_list' is true.
		do
			reserved_cursor_count := reserved_cursor_count - 1
			if keep_reserved_cursor_list then
				reserved_cursors.start
				reserved_cursors.search (cursor)
				if not reserved_cursors.exhausted then
					reserved_cursors.remove
				end
			end
		end

	process_old_cursors
			-- Check for cursors which are at or above the age
			-- limit for their category and process them
			-- accordingly.  Parsed or open cursors which
			-- are too old are closed.  Closed cursors which
			-- are too old are freed.
		local
			time: INTEGER
		do
			time := unix_os.current_time_in_seconds
			if (time - age_limit_check_time) >= age_check_interval then
				close_old_open_cursors (parsed_cursors,
					time - parsed_age_limit)
				close_old_open_cursors (open_cursors,
					time - open_age_limit)
				free_old_closed_cursors (time - closed_age_limit)
				age_limit_check_time := time
			end
		end

	close_old_open_cursors (cursors: SORTED_TWO_WAY_LIST [ORACLE_CURSOR]
		recycle_time: INTEGER)
			-- Close all cursors in list `cursors' whose
			-- `time_parse_requested' is less than or equal
			-- to `recycle_time'.  Add each closed cursor to
			-- `closed_cursors'.
		local
			cursor: ORACLE_CURSOR
		do
			from
				cursors.finish
			until
				cursors.is_empty or else cursors.item.time_parse_requested > recycle_time
			loop
				cursor := cursors.item
				cursor.raw_close
				cursors.remove
				closed_cursors.extend (cursor)
				cursors.back
			end
		end

	free_old_closed_cursors (recycle_time: INTEGER)
			-- Free all cursors in `closed_cursors' whose
			-- `time_parse_requested' is less than or equal
			-- to `recycle_time'.  Also if the number of closed
			-- cursors exceeds `max_closed_cursors' then free
			-- enough cursors to bring the number down to the
			-- limit.
		local
			cursors: SORTED_TWO_WAY_LIST [ORACLE_CURSOR]
		do
			from
				cursors := closed_cursors
				cursors.finish
			until
				cursors.is_empty or else (cursors.item.time_parse_requested > recycle_time
				and cursors.count <= max_closed_cursors)
			loop
				cursors.remove
				cursors.back
			end
		end

	close_all_cursors (cursors: SORTED_TWO_WAY_LIST [ORACLE_CURSOR])
			-- Close all cursors in list `cursors'
		local
			cursor: ORACLE_CURSOR
		do
			from
				cursors.start
			until
				cursors.exhausted
			loop
				cursor := cursors.item
				-- Note: None of the cursors in the list
				--	 should be in the "closed" state.
				--       However, there seem to be some
				--	 exceptional circumstances where
				--	 a cursor on the open list is
				--	 marked as closed in its state.
				--	 Rather than try to track this down
				--	 we'll just check first to avoid
				--	 precondition violations on
				--	`raw_close'.
				if not cursor.is_closed then
					cursor.raw_close
				end
				cursors.forth
			end
		end

	close_garbage_cursors
			-- Do a full garbage collection to cause any
			-- cursors which have become garbage to be closed
			-- if they are in state open, parsed or reserved
		do
			full_collect
		end

	process_connection_error
			-- Process error if one has occurred on this
			-- connection since server last one server processed
		do
			if error_on_connection then
				close_garbage_cursors
				error_on_connection := False
			end
		end

	valid_cursor_lists: BOOLEAN
			-- Are all cursor lists valid - that is, do they
			-- all contain only cursors of the proper type?
		do
			Result := valid_cursor_list (closed_cursors, Cursor_closed)
			   and valid_cursor_list (open_cursors, Cursor_open)
			   and valid_cursor_list (parsed_cursors, Cursor_parsed)
		end

	valid_cursor_list (list: LIST [ORACLE_CURSOR] state: INTEGER): BOOLEAN
			-- Are all cursors in `list' in state `state'?
		do
			from
				Result := True
				list.start
			until
				not Result or list.exhausted
			loop
				if list.item.state /= state then
					Result := False
				end
				list.forth
			end
		end

	closed_cursors: SORTED_TWO_WAY_LIST [ORACLE_CURSOR]
			-- Cursors which are closed, but have Context	
			-- Data Area allocated

	open_cursors: SORTED_TWO_WAY_LIST [ORACLE_CURSOR]
			-- Cursors which are open

	parsed_cursors: SORTED_TWO_WAY_LIST [ORACLE_CURSOR]
			-- Cursors which are open and have a parsed
			-- SQL statement associated with them

	reserved_cursors: ARRAYED_LIST [ORACLE_CURSOR]
			-- Cursors which are open and have a parsed
			-- SQL statement associated with them and
			-- are reserved for use by a particular client.
			-- Reserved cursors are only added to this list
			-- if `keep_reserved_cursor_list' is true

	age_limit_check_time: INTEGER
			-- Time that server last checked for cursors which
			-- were at or above the age limit for their
			-- category

invariant

	connection_exists: connection /= Void	
	closed_cursors_exists: closed_cursors /= Void	
	open_cursors_exists: open_cursors /= Void	
	parsed_cursors_exists: parsed_cursors /= Void	
	reserved_cursors_exists: reserved_cursors /= Void	
	valid_closed_cursor_max: max_closed_cursors >= 0
	valid_open_cursor_max: max_open_cursors >= 0
	valid_parsed_cursor_max: max_parsed_cursors >= 0
	valid_closed_age_limit: closed_age_limit >= 0
	valid_open_age_limit: open_age_limit >= 0
	valid_parsed_age_limit: parsed_age_limit >= 0
	valid_age_check_interval: age_check_interval >= 0
	
end
