
indexing
	
	author: "David Hollenberg"
	date:	"December 5, 1995"
	description: "An Oracle named or numeric placeholder (e.g., :1 or %
		%:design_no)"

deferred class ORACLE_PLACEHOLDER

inherit
	ANY
	C_MEMORY_ALLOCATION
		export
			{NONE} all
		end
	ORACLE_DATATYPE_CODES
		export
			{ANY} is_valid_external_datatype
			{NONE} all
		end
	ORACLE_VALIDITY
		export
			{NONE} all
		end
	ORACLE_CALL_INTERFACE
		export
			{NONE} all
		end
	STRING_UTILITIES_ACCESS
		export
			{NONE} all
		end
	STRING_FILLER_ACCESS
		export
			{NONE} all
		end
	ANY

feature -- Creation

	set_address (addr: POINTER)
			-- Set address of structure describing placeholder
			-- to `addr'
		require
			valid_address: addr /= default_pointer
		do
			address := addr
		ensure
			address_set: address = addr
		end

feature -- Properties

	address: POINTER
			-- Address of structure describing placeholder

	in_use: BOOLEAN
			-- Is placeholder slot in use (i.e., does it
			-- contain a placeholder)?
		do
			Result := ph_in_use (address)
		end

	display_name: STRING
			-- Name for display purposes
		deferred
		ensure
			result_exists: Result /= Void
		end

	max_display_name_length: INTEGER
			-- Maximum length of display name
		deferred
		ensure
			result_positive: Result > 0
		end

	data_length: INTEGER
			-- Length of data to which placeholder is bound
		do
			Result := ph_data_length (address)
		ensure
			nonnegative_result: Result >= 0
		end
	
	data_address: POINTER
			-- Address of data to which placeholder is bound
		do
			Result := ph_data_address (address)
		end
	
	long_data_address: POINTER
			-- Address of area for long data for placeholder
			-- (data that is not local to placeholder structure)
		do
			Result := ph_long_data_address (address)
		end
	
	long_data_capacity: INTEGER
			-- Capacity of area for long data for placeholder
		do
			Result := ph_long_data_capacity (address)
		ensure
			nonnegative_result: Result >= 0
		end
	
	max_output_length: INTEGER
			-- Maximum length of output for placeholder
			-- bound to string or raw.  Useful only for
			-- PL/SQL OUT or IN/OUT arguments
		do
			Result := ph_max_output_length (address)
		ensure
			nonnegative_result: Result >= 0
		end
	
	std_null_indicator_address: POINTER
			-- Standard null indicator address, which
			-- points to a location inside placeholder structure
		do
			Result := ph_std_null_ind_address (address)
		end
	
	external_datatype: INTEGER
			-- External datatype code as defined by the
			-- Oracle Call Interface
		do
			Result := ph_ext_datatype (address)
		ensure
			valid_datatype: is_valid_external_datatype (Result)
		end
	
	is_bound: BOOLEAN
			-- Has `Current' been bound by caller of placeholder
			-- binding interface?  A placeholder which has been
			-- bound from caller's point of view may or may not
			-- need binding from the Oracle Call Interface's
			-- point of view.
		do
			Result := ph_is_bound (address)
		end
	
	needs_oci_binding: BOOLEAN
			-- Does `Current' need binding or rebinding via
			-- an `obndrn' or `obndrv' routine call?
		do
			Result := ph_needs_oci_binding (address)
		end
	
	bound_to_null: BOOLEAN
			-- Is `Current' bound to null?
		do
			Result := ph_is_null (address)
		end
	
	bound_to_integer: BOOLEAN
			-- Is `Current' bound to an integer?
		do
			Result := external_datatype = integer_external_type
				and data_length = integer_external_length
		end
	
	bound_to_real: BOOLEAN
			-- Is `Current' bound to a real?
		do
			Result := external_datatype = float_external_type
				and data_length = real_external_length
		end
	
	bound_to_double: BOOLEAN
			-- Is `Current' bound to a double?
		do
			Result := external_datatype = float_external_type
				and data_length = double_external_length
		end
	
	bound_to_character: BOOLEAN
			-- Is `Current' bound to a character?
		do
			Result := external_datatype = varchar2_external_type
				and data_length = character_external_length
		end
	
	bound_to_boolean: BOOLEAN
			-- Is `Current' bound to a boolean?
		local
			dest: CHARACTER
		do
			if bound_to_character then
				dest := character_value
				Result := dest = 'Y' or dest = 'N'
			end
		end
	
	bound_to_string: BOOLEAN
			-- Is `Current' bound to a string?
		do
			Result := external_datatype = long_varchar_external_type
		end
	
	bound_to_raw: BOOLEAN
			-- Is `Current' bound to a raw?
		do
			Result := external_datatype = long_varraw_external_type
		end
	
	bound_to_date: BOOLEAN
			-- Is `Current' bound to a date?
		do
			Result := external_datatype = date_external_type
		end
	
	bound_to_rowid: BOOLEAN
			-- Is `Current' bound to a rowid?
		do
			Result := external_datatype = rowid_external_type
		end
	
	data_is_local: BOOLEAN
			-- Is data to which placeholder is bound local
			-- (i.e., inside the placeholder structure itself)?
		do
			Result := ph_data_is_local (address)
		end
	
	max_local_data_length: INTEGER
			-- Maximum length of data which can be bound to
			-- placeholder locally (i.e., inside the placeholder
			-- structure itself)
		do
			Result := ph_max_local_data_length (address)
		end
	
	integer_value: INTEGER
			-- Integer value to which `Current' is bound
		require
			bound_to_integer: bound_to_integer
		do
			Result := ph_integer_value (address)
		end
	
	real_value: REAL
			-- Real value to which `Current' is bound
		require
			bound_to_real: bound_to_real
		do
			Result := ph_real_value (address)
		end
	
	double_value: DOUBLE
			-- Double value to which `Current' is bound
		require
			bound_to_double: bound_to_double
		do
			Result := ph_double_value (address)
		end
	
	character_value: CHARACTER
			-- Character value to which `Current' is bound
		require
			bound_to_character: bound_to_character
		do
			ph_get_string_value (address, $Result)
		end
	
	boolean_value: BOOLEAN
			-- Boolean value to which `Current' is bound
		require
			bound_to_boolean: bound_to_boolean
		local
			c: CHARACTER
		do
			c := character_value
			Result := c = 'Y'
		end
	
	string_value: STRING
			-- String value to which `Current' is bound
		require
			bound_to_string: bound_to_string
		do
			Result := string_or_raw_value
		ensure
			result_exists: Result /= Void
		end
	
	raw_value: STRING
			-- Raw string value to which `Current' is bound
		require
			bound_to_raw: bound_to_raw
		do
			Result := string_or_raw_value
		ensure
			result_exists: Result /= Void
		end
	
	string_or_raw_value: STRING
			-- String or raw string value to which `Current'
			-- is bound
		require
			bound_to_string_or_raw: bound_to_string or bound_to_raw
		local
			len: INTEGER
		do
			len := signed_integer_value (data_address)
			create Result.make (len)
			strfill.fill_string (Result, offset (data_address, string_count_length), len)
		ensure
			result_exists: Result /= Void
		end
	
	date_value: ORACLE_DATE
			-- Date value to which `Current' is bound
		require
			bound_to_date: bound_to_date
		do
			create Result.make_internal (ph_double_value (address))
				-- Dates in Oracle internal format are stored
				-- as doubles in Eiffel
		ensure
			result_exists: Result /= Void
		end
	
	rowid_value: ORACLE_ROWID
			-- Rowid value to which `Current' is bound
		require
			bound_to_rowid: bound_to_rowid
		do
			create Result.make_internal (ph_rowid_first_value (address),
				ph_rowid_second_value (address))
		ensure
			result_exists: Result /= Void
		end
	
	any_value: ANY
			-- Value to which `Current' is bound
		require
			is_bound: is_bound
		do
			if bound_to_integer then
				Result := integer_value
			elseif bound_to_real then
				Result := real_value
			elseif bound_to_double then
				Result := double_value
			elseif bound_to_boolean then
				Result := boolean_value
			elseif bound_to_character then
				Result := character_value
			elseif bound_to_string then
				Result := string_value
			elseif bound_to_raw then
				Result := raw_value
			elseif bound_to_date then
				Result := date_value
			elseif bound_to_rowid then
				Result := rowid_value
			end
		ensure
			result_exists: Result /= Void
		end
	
	Long_data_multiple: INTEGER = 32
			-- Capacity of area for long data is allocated
			-- in multiples of this number

feature -- Modification

	initialize
			-- Initialize `Current'
		do
			set_in_use (True)
			set_is_bound (True)
			set_data_length (0)
			set_max_output_length (0)
			clear_data_address
			clear_external_datatype
			set_needs_oci_binding (True)
		ensure
			correct_data_length: data_length = 0
			correct_max_out_length: max_output_length = 0
			data_local: data_is_local
			placeholder_in_use: in_use
			needs_oci_binding: needs_oci_binding
		end
	
	fix_after_move
			-- Correct placeholder to account for the
			-- fact that it has been moved to a different
			-- memory location
		do
			if data_is_local then
				set_data_address (address)
			end
			set_needs_oci_binding (True)
		end

	oci_bind (c: ORACLE_CURSOR)
			-- Bind `Current' via Oracle Call Interface using
			-- `c' as the cursor
		require
			cursor_exists: c /= Void
			needs_binding: needs_oci_binding
		do
			c.check_status (do_oci_bind (c))
			set_needs_oci_binding (False)
		ensure
			oci_bound: not needs_oci_binding
		end
	
	do_oci_bind (c: ORACLE_CURSOR): INTEGER
			-- Bind `Current' via the appropriate Oracle Call
			-- Interface routine using  `c' as the cursor.
			-- Return the return code from this routine.
		require
			cursor_exists: c /= Void
			needs_binding: needs_oci_binding
		deferred
		end
	
	unbind
			-- Unbind placeholder so that it can be
			-- correctly reused by someone who reuses
			-- the associated open and parsed cursor
		do
			set_is_bound (False)
		ensure
			not_bound: not is_bound
		end

	bind_to_integer (value: INTEGER null: BOOLEAN)
			-- Bind `Current' to `value' if `null' is false;
			-- otherwise, bind to null
		do	
			bind_value (integer_external_type,
				integer_external_length, null)
			set_integer_value (value)
		ensure
			bound_to_integer: bound_to_integer
			null_bound_iff_null: null = bound_to_null
		end

	bind_to_real (value: REAL null: BOOLEAN)
			-- Bind `Current' to `value' if `null' is false;
			-- otherwise, bind to null
		do	
			bind_value (float_external_type,
				real_external_length, null)
			set_real_value (value)
		ensure
			bound_to_real: bound_to_real
			null_bound_iff_null: null = bound_to_null
		end

	bind_to_double (value: DOUBLE null: BOOLEAN)
			-- Bind `Current' to `value' if `null' is false;
			-- otherwise, bind to null
		do	
			bind_value (float_external_type,
				double_external_length, null)
			set_double_value (value)
		ensure
			bound_to_double: bound_to_double
			null_bound_iff_null: null = bound_to_null
		end

	bind_to_string (value: STRING null: BOOLEAN)
			-- Bind `Current' to `value' if `null' is false;
			-- otherwise, bind to null.  If `value' is void,
			-- bind to null regardless of `null'
		do	
			bind_to_string_or_raw (value, 0, null, long_varchar_external_type)
		ensure
			bound_to_string: bound_to_string
			null_bound_iff_null: (null or value = Void or else value.count = 0) = bound_to_null
			correct_nonvoid_max_length: value /= Void implies max_output_length = value.count
			correct_void_max_length: value = Void implies max_output_length = 0
		end

	bind_to_string_with_max_size (value: STRING max_out_len: INTEGER null: BOOLEAN)
			-- Bind `Current' to `value' if `null' is false;
			-- otherwise, bind to null.  If `value' is void,
			-- bind to null regardless of `null'.  Use
			-- `max_out_len' as the maximum length of the string
			-- if it is greater than the actual length
		do	
			bind_to_string_or_raw (value, max_out_len, null, long_varchar_external_type)
		ensure
			bound_to_string: bound_to_string
			null_bound_iff_null: (null or value = Void or else value.count = 0) = bound_to_null
			correct_nonvoid_max_length: value /= Void implies max_output_length = max_out_len.max (value.count)
			correct_void_max_length: value = Void implies max_output_length = max_out_len
		end

	bind_to_raw (value: STRING null: BOOLEAN)
			-- Bind `Current' to `value' if `null' is false;
			-- otherwise, bind to null.  If `value' is void,
			-- bind to null regardless of `null'
		do	
			bind_to_string_or_raw (value, 0, null, long_varraw_external_type)
		ensure
			bound_to_raw: bound_to_raw
			null_bound_iff_null: (null or value = Void or else value.count = 0) = bound_to_null
			correct_nonvoid_max_length: value /= Void implies max_output_length = value.count
			correct_void_max_length: value = Void implies max_output_length = 0
		end

	bind_to_raw_with_max_size (value: STRING max_out_len: INTEGER null: BOOLEAN)
			-- Bind `Current' to `value' if `null' is false;
			-- otherwise, bind to null.  If `value' is void,
			-- bind to null regardless of `null'.  Use
			-- `max_out_len' as the maximum length of the string
			-- if it is greater than the actual length
		do	
			bind_to_string_or_raw (value, max_out_len, null, long_varraw_external_type)
		ensure
			bound_to_raw: bound_to_raw
			null_bound_iff_null: (null or value = Void or else value.count = 0) = bound_to_null
			correct_nonvoid_max_length: value /= Void implies max_output_length = max_out_len.max (value.count)
			correct_void_max_length: value = Void implies max_output_length = max_out_len
		end

	bind_to_string_or_raw (value: STRING max_out_len: INTEGER null: BOOLEAN ext_type: INTEGER)
			-- Bind `Current' to `value' if `null' is false;
			-- otherwise, bind to null.  If `value' is void,
			-- bind to null regardless of `null'.  Use `ext_type'
			-- as the external datatype.  Use
			-- `max_out_len' as the maximum length of the string
			-- if it is greater than the actual length
		require
			nonnegative_max_length: max_out_len >= 0
			valid_datatype: is_valid_external_datatype (ext_type)
		local
			data_len, str_len, prev_len: INTEGER
			prev_addr: POINTER
		do	
			if value /= Void then
				str_len := value.count
			end
			data_len := str_len.max (max_out_len) + string_count_length
			-- Work around apparent Oracle bug when
			-- string length is zero (won't allow buffer
			-- length of 4 - have to use 5)

			if str_len = 0 then
				data_len := data_len + 1
			end
			
			prev_addr := data_address
			prev_len := data_length
			
			bind_value (ext_type, data_len, null or value = Void or str_len = 0)
			set_max_output_length (str_len.max (max_out_len))
			set_string_value (value, str_len, data_len)
			if prev_len /= data_length or prev_addr /= data_address then
				set_needs_oci_binding (True)
			end
		end

	bind_to_character (value: CHARACTER null: BOOLEAN)
			-- Bind `Current' to `value' if `null' is false;
			-- otherwise, bind to null
		local
			prev_addr: POINTER
			prev_len: INTEGER
		do	
			prev_addr := data_address
			prev_len := data_length
			bind_value (varchar2_external_type, 1, null)
			set_character_value (value)
			if prev_len /= data_length or prev_addr /= data_address then
				set_needs_oci_binding (True)
			end
		ensure
			bound_to_character: bound_to_character
			null_bound_iff_null: null = bound_to_null
		end

	bind_to_boolean (value: BOOLEAN null: BOOLEAN)
			-- Bind `Current' to `value' if `null' is false;
			-- otherwise, bind to null
		local
			c: CHARACTER
		do	
			if value then
				c := 'Y'
			else
				c := 'N'
			end
			bind_to_character (c, null)
		ensure
			bound_to_boolean: bound_to_boolean
			null_bound_iff_null: null = bound_to_null
		end

	bind_to_date (value: ORACLE_DATE null: BOOLEAN)
			-- Bind `Current' to `value' if `null' is
			-- false; otherwise, bind to null.  If `value'
			-- is void or `value'.`is_null', bind to null
			-- regardless of `null'
		do	
			bind_value (date_external_type, date_external_length,
				null or value = Void or else value.is_null)
			if value /= Void and then not value.is_null then
				set_date_value (value)
			end
		ensure
			bound_to_date: bound_to_date
			null_bound_iff_null: (null or value = Void or else value.is_null) = bound_to_null
		end

	bind_to_rowid (value: ORACLE_ROWID null: BOOLEAN)
			-- Bind `Current' to `value' if `null' is false;
			-- otherwise, bind to null.  If `value' is void,
			-- bind to null regardless of `null'
		do	
			bind_value (rowid_external_type, rowid_external_length,
				null or value = Void)
			if value /= Void then
				set_rowid_value (value)
			end
		ensure
			bound_to_rowid: bound_to_rowid
			null_bound_iff_null: (null or value = Void) = bound_to_null
		end

	bind_value (ext_type, len: INTEGER null: BOOLEAN)
			-- Bind `Current' to external datatype
			-- `ext_type' with data length `len'.  Set
			-- null indicator to `null'.  Set
			-- `data_address' to point to local data
			-- within placeholder (caller must override if
			-- data is not local).
		require
			valid_datatype: is_valid_external_datatype (ext_type)
			nonnegative_length: len >= 0
		do	
			set_external_datatype (ext_type)
			set_data_length (len)
			set_data_address (address)
			set_null_indicator (null)
			set_is_bound (True)
		end

	set_data_length (len: INTEGER)
			-- Set length of data to which placeholder is bound
			-- to `len'
		require
			nonnegative_length: len >= 0
		do
			ph_set_data_length (address, len)
		ensure
			correct_length: data_length = len
		end
	
	set_data_address (addr: POINTER)
			-- Set address of data to which placeholder is bound
			-- to `addr'
		require
			valid_address: addr /= default_pointer
		do
			ph_set_data_address (address, addr)
		ensure
			correct_data_address: data_address = addr
		end
	
	set_long_data_address (addr: POINTER)
			-- Set address of area for long data for placeholder
			-- to `addr'
		require
			valid_address: addr /= default_pointer
		do
			ph_set_long_data_address (address, addr)
		ensure
			correct_data_address: long_data_address = addr
		end
	
	clear_data_address
			-- Set address of data to which placeholder is bound
			-- to the null pointer
		local
			null_ptr: POINTER
		do
			ph_set_data_address (address, null_ptr)
		ensure
			correct_data_address: data_address = default_pointer
		end
	
	clear_long_data_address
			-- Set address of area for long strings to the
			-- null pointer
		local
			null_ptr: POINTER
		do
			ph_set_long_data_address (address, null_ptr)
		ensure
			correct_long_data_address: long_data_address = default_pointer
		end
	
	set_long_data_capacity (len: INTEGER)
			-- Set capacity of long data area for placeholder
			-- to `len'
		require
			nonnegative_length: len >= 0
		do
			ph_set_long_data_capacity (address, len)
		ensure
			correct_length: long_data_capacity = len
		end
	
	set_max_output_length (len: INTEGER)
			-- Set maximum length of output for string or raw
			-- placeholder to `len'
		require
			nonnegative_length: len >= 0
		do
			ph_set_max_output_length (address, len)
		ensure
			correct_length: max_output_length = len
		end
	
	set_external_datatype (code: INTEGER)
			-- Set external datatype to `code'
		require
			valid_datatype: is_valid_external_datatype (code)
		do
			ph_set_ext_datatype (address, code)
		ensure
			correct_ext_datatype: external_datatype = code
		end
	
	clear_external_datatype
			-- Set external datatype to 0
		do
			ph_set_ext_datatype (address, 0)
		ensure
			correct_ext_datatype: external_datatype = 0
		end
	
	set_is_bound (b: BOOLEAN)
			-- Set "bound by caller" indicator to `b'
		do
			ph_set_is_bound (address, b)
		ensure
			is_bound_set: is_bound = b
		end
	
	set_needs_oci_binding (b: BOOLEAN)
			-- Set "needs OCI binding" indicator to `b'
		do
			ph_set_needs_oci_binding (address, b)
		ensure
			needs_binding_set: needs_oci_binding = b
		end
	
	set_in_use (b: BOOLEAN)
			-- Set indication that placeholder is in use to `b`
		do
			ph_set_in_use (address, b)
		ensure
			is_in_use: in_use = b
		end
	
	set_null_indicator (b: BOOLEAN)
			-- Set null indicator to `b'.  True means placeholder
			-- is bound to null.
		do
			ph_set_null_indicator (address, b)
		ensure
			null_indicator_set: bound_to_null = b
		end
	
	set_integer_value (value: INTEGER)
			-- Set integer value to which placeholder is bound
			-- to `value'
		do
			ph_set_integer_value (address, value)
		ensure
			value_set: integer_value = value
		end
	
	set_real_value (value: REAL)
			-- Set real value to which placeholder is bound
			-- to `value'
		do
			ph_set_real_value (address, value)
		ensure
			value_set: real_value = value
		end
	
	set_double_value (value: DOUBLE)
			-- Set double value to which placeholder is bound
			-- to `value'
		do
			ph_set_double_value (address, value)
		ensure
			value_set: double_value = value
		end
	
	set_string_value (value: STRING str_len, data_len: INTEGER)
			-- Set string value to which placeholder is bound
			-- to `value', which has length `str_len' and
			-- whose maximum data length is `data_len'
		local
			area: ANY
			newcap: INTEGER
		do
			if data_len <= max_local_data_length then
				-- Nothing special to do
			elseif data_len <= long_data_capacity then
				set_data_address (long_data_address)
			else
				-- Long data area not big enough - reallocate
				if long_data_capacity > 0 then
					free_nongc_memory (long_data_address)
					set_long_data_capacity (0)
			 	end
				if data_len \\ Long_data_multiple = 0 then
					newcap := data_len
				else
					newcap := data_len + (Long_data_multiple - (data_len \\ Long_data_multiple))
				end
				set_long_data_address (allocate_nongc_memory (newcap))
				set_long_data_capacity (newcap)
				set_data_address (long_data_address)
			end
			ph_set_string_count (address, str_len)
			if value /= Void then
				area := value.area
				ph_set_string_value (address, $area, str_len)
			end
		ensure
			value_set_if_nonvoid: value /= Void implies value.is_equal (string_value)
			value_set_if_void: value = Void implies string_value.is_equal ("")
		end
	
	set_character_value (value: CHARACTER)
			-- Set character value to which placeholder is bound
			-- to `value'
		do
			ph_set_character_value (address, value)
		ensure
			value_set: character_value = value
		end
	
	set_date_value (value: ORACLE_DATE)
			-- Set date value to which placeholder is bound
			-- to `value'
		require
			value_exists: value /= Void
		do
			ph_set_double_value (address, value.contents)
				-- Dates in Oracle internal format
				-- are stored as doubles in Eiffel
		ensure
			value_set: equal (date_value, value)
		end
	
	set_rowid_value (value: ORACLE_ROWID)
			-- Set rowid value to which placeholder is bound
			-- to `value'
		require
			value_exists: value /= Void
		do
			ph_set_rowid_first_value (address, value.first_half)
			ph_set_rowid_second_value (address, value.second_half)
		ensure
			value_set: equal (rowid_value, value)
		end
	
	
feature -- Display

	append_description (name_len: INTEGER dest: STRING)
			-- Append to `dest' a description of `Current',
			-- padding the identifying  name or number to
			-- `name_len'
		local
			s: STRING
			name: STRING
			len, max_len: INTEGER
		do
			create s.make (100)
			s.append ("    ")
			name := display_name
			su.right_pad (name, name_len)
			s.append (name)
			s.append ("   ")
			name := name_for_external_code (external_datatype)
			name.append_character ('(')
			if bound_to_string or bound_to_raw then
				len := signed_integer_value (data_address)
				max_len := max_output_length
				if max_len /= len then
					name.append_integer (max_len)
					name.extend (',')
				end
				name.append_integer (len)
			else
				name.append_integer (data_length)
			end
			name.append_character (')')
			su.right_pad (name, 20)
			s.append (name)
			s.append ("   ")
			if not is_bound then
				s.append ("<unbound>")
			elseif bound_to_null then
				s.append ("<null>")
			elseif bound_to_integer then
				s.append_integer (integer_value)
			elseif bound_to_real then
				s.append_real (real_value)
			elseif bound_to_double then
				s.append_double (double_value)
			elseif bound_to_character then
				s.append_character ('%'')
				s.append_character (character_value)
				s.append_character ('%'')
			elseif bound_to_string then
				s.append_character ('%"')
				s.append (string_value)
				s.append_character ('%"')
			elseif bound_to_raw then
				s.append (su.string_to_hex (raw_value))
				s.append (" (raw)")
			elseif bound_to_date then
				s.append (date_value.out)
			elseif bound_to_rowid then
				s.append (rowid_value.out)
			else
				s.append ("<???>")
			end
			dest.append (s)
			dest.extend ('%N')
		end

feature {NONE}	-- Externals

	length: INTEGER
			-- Length in bytes
		external
			"C macro use %"oracle.h%""
		alias
			"ph_length"
		end

	ph_data_length (ph_address: POINTER): INTEGER
			-- Length of data to which placeholder is bound
		external
			"C macro use %"oracle.h%""
		end
	
	ph_data_address (ph_address: POINTER): POINTER
			-- Address of data to which placeholder is bound
		external
			"C macro use %"oracle.h%""
		end
	
	ph_long_data_address (ph_address: POINTER): POINTER
			-- Address of long data area for placeholder
		external
			"C macro use %"oracle.h%""
		end
	
	ph_long_data_capacity (ph_address: POINTER): INTEGER
			-- Capacity of long data area for placeholder
		external
			"C macro use %"oracle.h%""
		end
	
	ph_max_output_length (ph_address: POINTER): INTEGER
			-- Maximum length of output for placeholder
			-- bound to string or raw
		external
			"C macro use %"oracle.h%""
		end
	
	ph_std_null_ind_address (ph_address: POINTER): POINTER
			-- Standard address of null indicator, which
			-- is inside placeholder structure
		external
			"C macro use %"oracle.h%""
		end
	
	ph_ext_datatype (ph_address: POINTER): INTEGER
			-- External datatype
		external
			"C macro use %"oracle.h%""
		end
	
	ph_is_bound (ph_address: POINTER): BOOLEAN
			-- Has placeholder been bound by caller of
			-- placeholder binding interface?
		external
			"C macro use %"oracle.h%""
		end
	
	ph_needs_oci_binding (ph_address: POINTER): BOOLEAN
			-- Does placeholder need binding or rebinding
			-- via OCI call?
		external
			"C macro use %"oracle.h%""
		end
	
	ph_number (ph_address: POINTER): INTEGER
			-- Numeric placeholder number or length of
			-- named placeholder name
		external
			"C macro use %"oracle.h%""
		end
	
	ph_is_null (ph_address: POINTER): BOOLEAN
			-- Is placeholder bound to null?
		external
			"C macro use %"oracle.h%""
		end
	
	ph_in_use (ph_address: POINTER): BOOLEAN
			-- Is placeholder in use?
		external
			"C macro use %"oracle.h%""
		end
	
	ph_data_is_local (ph_address: POINTER): BOOLEAN
			-- Is data to which placeholder is bound local
			-- (i.e., inside the placeholder structure itself)?
		external
			"C macro use %"oracle.h%""
		end
	
	ph_max_local_data_length (ph_address: POINTER): INTEGER
			-- Maximum length of data which can be bound to
			-- placeholder locally (i.e., inside the placeholder
			-- structure itself)
		external
			"C macro use %"oracle.h%""
		end
	
	ph_integer_value (ph_address: POINTER): INTEGER
			-- Integer value to which placeholder is bound
		external
			"C macro use %"oracle.h%""
		end
	
	ph_real_value (ph_address: POINTER): REAL
			-- Real value to which placeholder is bound
		external
			"C macro use %"oracle.h%""
		end
	
	ph_double_value (ph_address: POINTER): DOUBLE
			-- Double value to which placeholder is bound
		external
			"C macro use %"oracle.h%""
		end
	
	ph_rowid_first_value (ph_address: POINTER): DOUBLE
			-- First half of rowid value to which placeholder
			-- is bound
		external
			"C macro use %"oracle.h%""
		end
	
	ph_rowid_second_value (ph_address: POINTER): DOUBLE
			-- Second half of rowid value to which placeholder
			-- is bound
		external
			"C macro use %"oracle.h%""
		end
	
	ph_rowid_row_number (ph_address: POINTER): INTEGER
			-- Row number portion of rowid
		external
			"C macro use %"oracle.h%""
		end
	
	ph_rowid_block_number (ph_address: POINTER): INTEGER
			-- Block number portion of rowid
		external
			"C macro use %"oracle.h%""
		end
	
	ph_rowid_file_number (ph_address: POINTER): INTEGER
			-- Database file number portion of rowid
		external
			"C macro use %"oracle.h%""
		end
	
	ph_get_string_value (ph_address: POINTER dest: POINTER)
			-- Copy string value to which placeholder is bound
			-- to `dest'
		external
			"C macro use %"oracle.h%""
		end
	
	ph_set_data_length (ph_address: POINTER len: INTEGER)
			-- Set length of data to which placeholder is bound
			-- to `len'
		external
			"C macro use %"oracle.h%""
		end
	
	ph_set_data_address (ph_address: POINTER data_addr: POINTER)
			-- Set address of data to which placeholder is bound
			-- to `data_addr'
		external
			"C macro use %"oracle.h%""
		end
	
	ph_set_long_data_address (ph_address: POINTER data_addr: POINTER)
			-- Set address of long data area `data_addr'
		external
			"C macro use %"oracle.h%""
		end
	
	ph_set_long_data_capacity (ph_address: POINTER len: INTEGER)
			-- Set capacity of long data area for placeholder
			-- to `len'
		external
			"C macro use %"oracle.h%""
		end
	
	ph_set_max_output_length (ph_address: POINTER len: INTEGER)
			-- Set maximum output length for placeholder bound
			-- to string or raw to `len'
		external
			"C macro use %"oracle.h%""
		end
	
	ph_set_ext_datatype (ph_address: POINTER code: INTEGER)
			-- Set external datatype code to `code'
		external
			"C macro use %"oracle.h%""
		end
	
	ph_set_is_bound (ph_address: POINTER newval: BOOLEAN)
			-- Set "is bound by caller" indicator to `newval'
		external
			"C macro use %"oracle.h%""
		end
	
	ph_set_needs_oci_binding (ph_address: POINTER newval: BOOLEAN)
			-- Set "needs binding" indicator to `newval'
		external
			"C macro use %"oracle.h%""
		end
	
	ph_set_in_use (ph_address: POINTER newval: BOOLEAN)
			-- Set "in use" indicator to `newval'
		external
			"C macro use %"oracle.h%""
		end
	
	ph_set_number (ph_address: POINTER value: INTEGER)
			-- Set placeholder number or placeholder name
			-- length to `value'
		external
			"C macro use %"oracle.h%""
		end

	ph_set_null_indicator (ph_address: POINTER newval: BOOLEAN)
			-- Set value of null indicator to `newval'
		external
			"C macro use %"oracle.h%""
		end
	
	ph_set_integer_value (ph_address: POINTER newval: INTEGER)
			-- Set integer value to which placeholder is bound
			-- to `newval'
		external
			"C macro use %"oracle.h%""
		end
	
	ph_set_real_value (ph_address: POINTER newval: REAL)
			-- Set real value to which placeholder is bound
			-- to `newval'
		external
			"C macro use %"oracle.h%""
		end
	
	ph_set_double_value (ph_address: POINTER newval: DOUBLE)
			-- Set double value to which placeholder is bound
			-- to `newval'
		external
			"C macro use %"oracle.h%""
		end
	
	ph_set_character_value (ph_address: POINTER newval: CHARACTER)
			-- Set character value to which placeholder is bound
			-- to `newval'
		external
			"C macro use %"oracle.h%""
		end
	
	ph_set_string_value (ph_address: POINTER str_addr: POINTER len: INTEGER)
			-- Set string value to which placeholder is bound
			-- to string at address `str_addr' which has
			-- length `len'
		external
			"C macro use %"oracle.h%""
		end
	
	ph_set_string_count (ph_address: POINTER len: INTEGER)
			-- Set count for string value to which placeholder
			-- is bound to `len'
		external
			"C macro use %"oracle.h%""
		end
	
	ph_set_rowid_first_value (ph_address: POINTER newval: DOUBLE)
			-- Set first half of rowid value to which
			-- placeholder is bound to `newval'
		external
			"C macro use %"oracle.h%""
		end
	
	ph_set_rowid_second_value (ph_address: POINTER newval: DOUBLE)
			-- Set second half of rowid value to which
			-- placeholder is bound to `newval'
		external
			"C macro use %"oracle.h%""
		end
	
feature {NONE} -- Obsolete
	
	ph_get_date_value (ph_address: POINTER dest: POINTER)
			-- Copy date value to which placeholder is bound
			-- to `dest'
		external
			"C macro use %"oracle.h%""
		end
	
	ph_set_date_value (ph_address: POINTER date_addr: POINTER)
			-- Set date value to which placeholder is bound
			-- to date at address `date_addr'
		external
			"C macro use %"oracle.h%""
		end
	
	ph_get_rowid_value (ph_address: POINTER dest: POINTER)
			-- Copy row id value to which placeholder is bound
			-- to `dest'
		external
			"C macro use %"oracle.h%""
		end
	
	ph_set_rowid_value (ph_address: POINTER rowid_addr: POINTER)
			-- Set rowid value to which placeholder is bound
			-- to rowid at address `rowid_addr'
		external
			"C macro use %"oracle.h%""
		end
	
invariant
	address_exists: address /= default_pointer

end
