
indexing
	
	author: "David Hollenberg"
	date:	"December 21, 1995"
	description: "An Oracle select item in a query"

class ORACLE_SELECT_ITEM

inherit
	ANY
	ORACLE_DATATYPE_CODES
		export
			{ANY} is_valid_external_datatype
			{NONE} all
		end
	EIFFEL_DATATYPE_CODES
		export
			{ANY} is_valid_eiffel_datatype
			{NONE} all
		end
	ORACLE_CALL_INTERFACE
		export
			{NONE} all
		end
	ORACLE_VALIDITY
		export
			{ANY} is_valid_select_position,
				is_valid_string_length
			{NONE} all
		end
	C_MEMORY_ALLOCATION
		export
			{NONE} all
		end
	STRING_UTILITIES_ACCESS
		export
			{NONE} all
		end
	STRING_FILLER_ACCESS
		export
			{NONE} all
		end

create
	set_address

feature -- Creation

	set_address (addr: POINTER)
			-- Set address of structure describing select item
			-- to `addr'
		require
			valid_address: addr /= default_pointer
		do
			address := addr
		ensure
			address_set: address = addr
		end

feature -- Properties

	address: POINTER
			-- Address of structure describing select item

	position: INTEGER
			-- Position of select item in query
		do
			Result := sel_position (address)
		ensure
			nonnegative_result: Result > 0
		end
	
	buffer_address: POINTER
			-- Address of buffer which holds one or more
			-- fetched values (one per row)
		do
			Result := sel_buffer_address (address)
		end
	
	value_length: INTEGER
			-- Length of each fetched value (amount of space
			-- allocated for each one)
		do
			Result := sel_value_length (address)
		ensure
			nonnegative_result: Result >= 0
		end
	
	max_value_length: INTEGER
			-- Specified maximum length of each fetched value
			-- for a string or raw string.  Zero indicates that
			-- the implementation should ask the database
			-- server for the maximum length
		do
			Result := sel_max_value_length (address)
		ensure
			nonnegative_result: Result >= 0
		end
	
	max_select_length: INTEGER = 65536
			-- Largest select item buffer length Oracle can
			-- handle.  Actually, it can probably handle larger
			-- select items in Oracle 7
	
	eiffel_datatype: INTEGER
			-- Code indicating the Eiffel datatype to which the
			-- fetched values will be converted
		do
			Result := sel_eiffel_datatype (address)
		ensure
			valid_datatype: is_valid_eiffel_datatype (Result)
		end
	
	external_datatype: INTEGER
			-- External datatype code as defined by the
			-- Oracle Call Interface
		do
			Result := sel_external_datatype (address)
		ensure
			valid_datatype: is_valid_external_datatype (Result)
		end
	
	is_defined: BOOLEAN
			-- Has select item been defined from user's viewpoint?
		do
			Result := sel_is_defined (address)
		end
	
	is_oci_defined: BOOLEAN
			-- Has select item been defined from viewpoint
			-- of Oracle Call Interface?
		do
			Result := sel_is_oci_defined (address)
		end
	
	row_count: INTEGER
			-- Number of rows fetched on each fetch

	integer_value (k: INTEGER): INTEGER
			-- Integer value for `Current' for `k'-th row
		require
			index_large_enough: k >= 1
			index_small_enough: k <= row_count
			column_is_integer: is_integer
		do
			Result := sel_integer_item (address, k - 1)
		end
	
	real_value (k: INTEGER): REAL
			-- Real value for `Current' for `k'-th row
		require
			index_large_enough: k >= 1
			index_small_enough: k <= row_count
			column_is_real: is_real
		do
			Result := sel_real_item (address, k - 1)
		end
	
	double_value (k: INTEGER): DOUBLE
			-- Double value for `Current' for `k'-th row
		require
			index_large_enough: k >= 1
			index_small_enough: k <= row_count
			column_is_double: is_double
		do
			Result := sel_double_item (address, k - 1)
		end
	
	character_value (k: INTEGER): CHARACTER
			-- Character value for `Current' for `k'-th row
		require
			index_large_enough: k >= 1
			index_small_enough: k <= row_count
			column_is_character: is_character
		do
			Result := sel_character_item (address, k - 1)
		end
	
	boolean_character_value (k: INTEGER): CHARACTER
			--Character value for boolean for `k'-th row
		require
			index_large_enough: k >= 1
			index_small_enough: k <= row_count
			column_is_boolean: is_boolean
		do
			Result := sel_character_item (address, k - 1)
		end
	
	string_value (k, len: INTEGER): STRING
			-- String value of length `len'
			-- for `k'-th row of `Current'
		require
			index_large_enough: k >= 1
			index_small_enough: k <= row_count
			column_is_string: is_string
		do
			Result := string_or_raw_value (k, len)
		ensure
			result_exists: Result /= Void
		end
	
	raw_value (k, len: INTEGER): STRING
			-- Raw string value of length `len'
			-- for `k'-th row of `Current'
		require
			index_large_enough: k >= 1
			index_small_enough: k <= row_count
			column_is_raw: is_raw
		do
			Result := string_or_raw_value (k, len)
		ensure
			result_exists: Result /= Void
		end
	
	date_value (k: INTEGER): ORACLE_DATE
			-- Date value for `Current' for `k'-th row
		require
			index_large_enough: k >= 1
			index_small_enough: k <= row_count
			column_is_date: is_date
		do
			create Result.make_internal (0.0)
			get_date_value (k, Result)
		ensure
			result_exists: Result /= Void
		end
	
	rowid_value (k: INTEGER): ORACLE_ROWID
			-- Rowid value for `Current' for `k'-th row
		require
			index_large_enough: k >= 1
			index_small_enough: k <= row_count
			column_is_rowid: is_rowid
		do
			create Result.set_null
			get_rowid_value (k, Result)
		ensure
			result_exists: Result /= Void
		end
	
feature -- Access to values
	
	get_string_value (k, len: INTEGER dest: STRING)
			-- Get string value of length `len'
			-- for `k'-th row of `Current' and store into `dest'
		require
			index_large_enough: k >= 1
			index_small_enough: k <= row_count
			column_is_string: is_string
			destination_exists: dest /= Void
		do
			if dest.capacity < len then
				dest.grow (len)
			end
			get_string_or_raw_value (k, dest, len)
		ensure
			destination_set: equal (dest, string_value (k, len))
		end
	
	get_raw_value (k, len: INTEGER dest: STRING)
			-- Get raw string value of length `len'
			-- for `k'-th row of `Current' and store into `dest'
		require
			index_large_enough: k >= 1
			index_small_enough: k <= row_count
			column_is_raw: is_raw
			destination_exists: dest /= Void
		do
			if dest.capacity < len then
				dest.grow (len)
			end
			get_string_or_raw_value (k, dest, len)
		ensure
			destination_set: equal (dest, raw_value (k, len))
		end
	
	get_date_value (k: INTEGER dest: ORACLE_DATE)
			-- Get date value for `Current' for `k'-th row
			-- and store it in `dest'
		require
			index_large_enough: k >= 1
			index_small_enough: k <= row_count
			column_is_date: is_date
			destination_exists: dest /= Void
		local
			addr: POINTER
			len: INTEGER
			d: DOUBLE
		do
			len := sel_value_length (address)
			addr := sel_date_item_address (address, k - 1)
			copy_nongc_memory (addr, $d, len)
			dest.make_internal (d)
		ensure
			destination_set: equal (dest, date_value (k))
		end
	
	get_rowid_value (k: INTEGER dest: ORACLE_ROWID)
			-- Rowid value for `Current' for `k'-th row
		require
			index_large_enough: k >= 1
			index_small_enough: k <= row_count
			column_is_rowid: is_rowid
			destination_exists: dest /= Void
		local
			first, second: DOUBLE
		do
			first := sel_rowid_item_first_half (address, k - 1)
			second := sel_rowid_item_second_half (address, k - 1)
			dest.make_internal (first, second)
		ensure
			destination_set: equal (dest, rowid_value (k))
		end
	
feature {NONE} -- Implementation

	filler: STRING_FILLER

	string_or_raw_value (k, len: INTEGER): STRING
			-- String or raw string value of length `len'
			-- for `k'-th row of `Current'
		require
			index_large_enough: k >= 1
			index_small_enough: k <= row_count
			column_is_string_or_raw: is_string or is_raw
		do
			create Result.make (len)
			get_string_or_raw_value (k, Result, len)
		end
	
	get_string_or_raw_value (k: INTEGER dest: STRING len: INTEGER)
			-- Get string or raw string value for `Current' for
			-- `k'-th row, which is of length `len' and store
			-- it in `dest'
		require
			index_large_enough: k >= 1
			index_small_enough: k <= row_count
			column_is_string_or_raw: is_string or is_raw
			destination_exists: dest /= Void
			destination_large_enough: dest.capacity >= len
		local
			addr: POINTER
			alloc_len: INTEGER
			fill: STRING_FILLER
		do
			alloc_len := sel_value_length (address)
			addr := sel_string_item_address (address, k - 1, alloc_len)
			fill := filler
			if fill = Void then
				fill := strfill
				filler := fill
			end
			fill.fill_string (dest, addr, len)
		end
	
feature -- Properties

	is_integer: BOOLEAN
			-- Are the values for `Current' Eiffel integers?
		do
			Result := sel_eiffel_datatype (address) = integer_eiffel_type
		end

	is_real: BOOLEAN
			-- Are the values for `Current' Eiffel reals?
		do
			Result := sel_eiffel_datatype (address) = real_eiffel_type
		end

	is_double: BOOLEAN
			-- Are the values for `Current' Eiffel doubles?
		do
			Result := sel_eiffel_datatype (address) = double_eiffel_type
		end

	is_character: BOOLEAN
			-- Are the values for `Current' Eiffel characters?
		do
			Result := sel_eiffel_datatype (address) = character_eiffel_type
		end

	is_boolean: BOOLEAN
			-- Are the values for `Current' Eiffel booleans?
		do
			Result := sel_eiffel_datatype (address) = boolean_eiffel_type
		end

	is_date: BOOLEAN
			-- Are the values for `Current' Eiffel database dates?
		do
			Result := sel_eiffel_datatype (address) = date_eiffel_type
		end

	is_rowid: BOOLEAN
			-- Are the values for `Current' Eiffel database rowids?
		do
			Result := sel_eiffel_datatype (address) = rowid_eiffel_type
		end

	is_string: BOOLEAN
			-- Are the values for `Current' Eiffel strings?
		do
			Result := sel_eiffel_datatype (address) = string_eiffel_type
		end

	is_raw: BOOLEAN
			-- Are the values for `Current' Eiffel strings which
			-- are "raw" database values allowing any bit pattern
			-- for each byte?
		do
			Result := sel_eiffel_datatype (address) = raw_eiffel_type
		end

feature -- Modification

	initialize (k: INTEGER)
			-- Initialize `Current' as select item `k'
		local
			null_ptr: POINTER
		do
			set_position (k)
			set_is_defined (False)
			set_is_oci_defined (False)
			set_buffer_address (null_ptr)
			set_value_length (0)
			set_max_value_length (0)
		ensure
			correct_position: position = k
			not_defined: not is_defined
		end
	
	set_row_count (num: INTEGER)
			-- Set number of values to be fetched for select item
			-- on each OCI fetch to `num'
		do
			row_count := num
		ensure
			row_count_set: row_count = num
		end

	oci_define (c: ORACLE_CURSOR var_addr, retlen_addr, col_retcode_addr, indicator_addr: POINTER)
			-- Define `Current' via Oracle Call Interface using
			-- `c' as the cursor
		require
			cursor_exists: c /= Void
		local
			code: INTEGER
		do
			set_buffer_address (var_addr)
			c.enter_oci
			code := ora_bind_select_item (c.address,
				position, var_addr, value_length,
				external_datatype, indicator_addr, retlen_addr,
				col_retcode_addr)
			c.leave_oci
			c.check_status (code)
			set_is_oci_defined (True)
		end
	
	select_integer
			-- Specify that values for `Current' are Eiffel
			-- integers
		do	
			set_eiffel_datatype (integer_eiffel_type)
			set_external_datatype (integer_external_type)
			set_value_length (integer_external_length)
			set_is_defined (True)
		end

	select_real
			-- Specify that values for `Current' are Eiffel reals
		do	
			set_eiffel_datatype (real_eiffel_type)
			set_external_datatype (float_external_type)
			set_value_length (real_external_length)
			set_is_defined (True)
		end

	select_double
			-- Specify that values for `Current' are Eiffel doubles
		do	
			set_eiffel_datatype (double_eiffel_type)
			set_external_datatype (float_external_type)
			set_value_length (double_external_length)
			set_is_defined (True)
		end

	select_character
			-- Specify that values for `Current' are Eiffel
			-- characters
		do	
			set_eiffel_datatype (character_eiffel_type)
			set_external_datatype (varchar2_external_type)
			set_value_length (character_external_length)
			set_is_defined (True)
		end

	select_boolean
			-- Specify that values for `Current' are Eiffel
			-- booleans
		do	
			set_eiffel_datatype (boolean_eiffel_type)
			set_external_datatype (varchar2_external_type)
			set_value_length (character_external_length)
			set_is_defined (True)
		end

	select_string (maxlen: INTEGER)
			-- Specify that values for `Current' are Eiffel
			-- strings of maximum length `maxlen'
		require
			valid_max_length: is_valid_string_length (maxlen)
		local
			old_maxlen, new_maxlen: INTEGER
			needs_value_len: BOOLEAN
		do	
			old_maxlen := max_value_length
			new_maxlen := maxlen.min (max_select_length)
			needs_value_len := not is_oci_defined or else not is_string or else old_maxlen /= new_maxlen
			set_eiffel_datatype (string_eiffel_type)
			set_external_datatype (varchar2_external_type)
			set_max_value_length (new_maxlen)
			if needs_value_len then
				set_value_length (new_maxlen)
			end
			set_is_defined (True)
		end

	select_raw (maxlen: INTEGER)
			-- Specify that values for `Current' are Eiffel
			-- raw strings of maximum length `maxlen'
		require
			valid_max_length: is_valid_string_length (maxlen)
		local
			old_maxlen, new_maxlen: INTEGER
			needs_value_len: BOOLEAN
		do	
			old_maxlen := max_value_length
			new_maxlen := maxlen.min (max_select_length)
			needs_value_len := not is_oci_defined or else not is_raw or else old_maxlen /= new_maxlen
			set_eiffel_datatype (raw_eiffel_type)
			set_external_datatype (raw_external_type)
			set_max_value_length (new_maxlen)
			if needs_value_len then
				set_value_length (new_maxlen)
			end
			set_is_defined (True)
		end

	select_date
			-- Specify that values for `Current' are Eiffel
			-- database dates
		do	
			set_eiffel_datatype (date_eiffel_type)
			set_external_datatype (date_external_type)
			set_value_length (date_external_length)
			set_is_defined (True)
		end

	select_rowid
			-- Specify that values for `Current' are Eiffel
			-- database rowids
		do	
			set_eiffel_datatype (rowid_eiffel_type)
			set_external_datatype (rowid_external_type)
			set_value_length (rowid_external_length)
			set_is_defined (True)
		end

	set_position (k: INTEGER)
			-- Set position of `Current' in select clause to `k'
		require
			valid_select_position: is_valid_select_position (k)
		do
			sel_set_position (address, k)
		ensure
			position_set: position = k
		end
	
	set_buffer_address (addr: POINTER)
			-- Set address of buffer for fetched values for
			-- `Current' to `addr'
		do
			sel_set_buffer_address (address, addr)
		ensure
			adress_set: buffer_address = addr
		end
	
	set_value_length (len: INTEGER)
			-- Set length of each fetched value (amount of space
			-- allocated for each one) to `len'
		require
			nonnegative_length: len >= 0
		do
			sel_set_value_length (address, len)
		ensure
			correct_length: value_length = len
		end
	
	set_max_value_length (len: INTEGER)
			-- Set requested maximum length of each fetched value
			-- to `len'
		require
			nonnegative_length: len >= 0
		do
			sel_set_max_value_length (address, len)
		ensure
			correct_length: max_value_length = len
		end
	
	set_external_datatype (code: INTEGER)
			-- Set external datatype to `code'
		require
			valid_datatype: is_valid_external_datatype (code)
		do
			sel_set_external_datatype (address, code)
		ensure
			correct_ext_datatype: external_datatype = code
		end
	
	set_eiffel_datatype (code: INTEGER)
			-- Set Eiffel datatype to `code'
		require
			valid_datatype: is_valid_eiffel_datatype (code)
		do
			sel_set_eiffel_datatype (address, code)
		ensure
			correct_eiffel_datatype: eiffel_datatype = code
		end
	
	set_is_defined (b: BOOLEAN)
			-- Set "is defined" indicator to `b'
		do
			sel_set_is_defined (address, b)
		ensure
			is_defined_set: is_defined = b
		end
	
	set_is_oci_defined (b: BOOLEAN)
			-- Set "is OCI defined" indicator to `b'
		do
			sel_set_is_oci_defined (address, b)
		ensure
			is_defined_set: is_defined = b
		end
	
feature -- Display

	display
			-- Display `Current'
		local
			s: STRING
			name: STRING
		do
			create s.make (100)
			s.append ("    ")
			create name.make (10)
			name.append_integer (position)
			su.right_pad (name, 3)
			s.append (name)
			s.append ("   ")
			if not is_defined then
				s.append ("<undefined>")
			else
				name := name_for_eiffel_type (eiffel_datatype)
				su.right_pad (name, 10)
				s.append (name)
				s.append ("   ")
				name := name_for_external_code (external_datatype)
				su.right_pad (name, 11)
				s.append (name)
				s.append ("   ")
				create name.make (10)
				name.append_integer (value_length)
				s.append (name)
			end
			io.put_string (s)
			io.new_line
		end

feature {NONE}	-- Externals

	length: INTEGER
			-- Length of representation in bytes
		external
			"C macro use %"oracle.h%""
		alias
			"sel_length"
		end

	sel_position (sel_address: POINTER): INTEGER
			-- Position of select item in query
		external
			"C macro use %"oracle.h%""
		end
	
	sel_buffer_address (sel_address: POINTER): POINTER
			-- Address of buffer which holds one or more
			-- fetched values (one per row)
		external
			"C macro use %"oracle.h%""
		end
	
	sel_value_length (sel_address: POINTER): INTEGER
			-- Length of each fetched value (amount of space
			-- allocated for each one)
		external
			"C macro use %"oracle.h%""
		end
	
	sel_max_value_length (sel_address: POINTER): INTEGER
			-- Specified maximum length of each fetched value
			-- for a string or raw string.  Zero indicates that
			-- the implementation should ask the database
			-- server for the maximum length
		external
			"C macro use %"oracle.h%""
		end
	
	sel_eiffel_datatype (sel_address: POINTER): INTEGER
			-- Code indicating the Eiffel datatype to which the
			-- fetched values will be converted
		external
			"C macro use %"oracle.h%""
		end
	
	sel_external_datatype (sel_address: POINTER): INTEGER
			-- External datatype code as defined by the
			-- Oracle Call Interface
		external
			"C macro use %"oracle.h%""
		end
	
	sel_is_defined (sel_address: POINTER): BOOLEAN
			-- Has select item been defined from user's viewpoint?
		external
			"C macro use %"oracle.h%""
		end
	
	sel_is_oci_defined (sel_address: POINTER): BOOLEAN
			-- Has select item been defined from viewpoint
			-- of Oracle Call Interface?
		external
			"C macro use %"oracle.h%""
		end
	
	sel_integer_item (sel_address: POINTER k: INTEGER): INTEGER
			-- Integer value for `Current' for `k'-th row
		external
			"C macro use %"oracle.h%""
		end
	
	sel_real_item (sel_address: POINTER k: INTEGER): REAL
			-- Real value for `Current' for `k'-th row
		external
			"C macro use %"oracle.h%""
		end
	
	sel_double_item (sel_address: POINTER k: INTEGER): DOUBLE
			-- Double value for `Current' for `k'-th row
		external
			"C macro use %"oracle.h%""
		end
	
	sel_character_item (sel_address: POINTER k: INTEGER): CHARACTER
			-- Character value for `Current' for `k'-th row
		external
			"C macro use %"oracle.h%""
		end
	
	sel_string_item_address (sel_address: POINTER k: INTEGER len: INTEGER): POINTER
			-- Address of string value for `Current' for `k'-th row
		external
			"C macro use %"oracle.h%""
		end
	
	sel_date_item_address (sel_address: POINTER k: INTEGER): POINTER
			-- Address of date value for `Current' for `k'-th row
		external
			"C macro use %"oracle.h%""
		end
	
	sel_rowid_item_first_half (sel_address: POINTER k: INTEGER): DOUBLE
			-- First half of rowid value for `Current' for `k'-th row
		external
			"C macro use %"oracle.h%""
		end
	
	sel_rowid_item_second_half (sel_address: POINTER k: INTEGER): DOUBLE
			-- Second half of rowid value for `Current' for `k'-th row
		external
			"C macro use %"oracle.h%""
		end
	
	sel_set_position (sel_address: POINTER k: INTEGER)
			-- Set position of `Current' in select clause to `k'
		external
			"C macro use %"oracle.h%""
		end
	
	sel_set_buffer_address (sel_address: POINTER addr: POINTER)
			-- Set address of buffer for fetched values for
			-- `Current' to `addr'
		external
			"C macro use %"oracle.h%""
		end
	
	sel_set_value_length (sel_address: POINTER len: INTEGER)
			-- Set length of each fetched value (amount of space
			-- allocated for each one) to `len'
		external
			"C macro use %"oracle.h%""
		end
	
	sel_set_max_value_length (sel_address: POINTER len: INTEGER)
			-- Set requested maximum length of each fetched value
			-- to `len'
		external
			"C macro use %"oracle.h%""
		end
	
	sel_set_external_datatype (sel_address: POINTER code: INTEGER)
			-- Set external datatype to `code'
		external
			"C macro use %"oracle.h%""
		end
	
	sel_set_eiffel_datatype (sel_address: POINTER code: INTEGER)
			-- Set eiffel datatype to `code'
		external
			"C macro use %"oracle.h%""
		end
	
	sel_set_is_defined (sel_address: POINTER newval: BOOLEAN)
			-- Set "is defined" indicator to `newval'
		external
			"C macro use %"oracle.h%""
		end
	
	sel_set_is_oci_defined (sel_address: POINTER newval: BOOLEAN)
			-- Set "is OCI defined" indicator to `newval'
		external
			"C macro use %"oracle.h%""
		end
	
invariant
	address_exists: address /= default_pointer

end
