
indexing
	
	author: "David Hollenberg"
	date:	"December 22, 1995"
	description: "A set of Oracle select items for a query"

class ORACLE_SELECT_SET

inherit
	ANY
	ORACLE_VALIDITY
		export
			{ANY} is_valid_select_position,
				is_valid_string_length
			{NONE} all
		end
	ORACLE_EXCEPTIONS
		export
			{NONE} all
		end
	C_MEMORY_ALLOCATION
		export
			{NONE} all
		end
	OCI_CONSTANTS
		export
			{NONE} all
		end
	ORACLE_CALL_INTERFACE
		export
			{NONE} all
		end
	MEMORY
		export
			{NONE} all
		redefine
			dispose
		end

create
	make

feature {NONE} -- Creation
	
	make (num: INTEGER the_cursor: ORACLE_CURSOR)
			-- Create a select set capable of describing
			-- `num' select items and associated with cursor
			-- `the_cursor'
		require
			nonnegative_count: num >= 0
			cursor_exists: the_cursor /= Void
		do
			cursor := the_cursor
			create select_item.set_address ($make)
				-- Use bogus address since no select items
				-- exist yet
			resize (num)
			raw_requested_row_capacity := -1
			requested_buffer_limit := -1
			set_row_capacity (0)
		ensure
			correct_capacity: capacity = num
			address_exists: address /= default_pointer
			select_item_exists: select_item /= Void
			correct_cursor: cursor = the_cursor
		end

feature {NONE}	-- Cleanup

	dispose
			-- Return memory allocated for select set
		local
			null_ptr: POINTER
		do
			if address /= null_ptr then
				free_nongc_memory (address)
				address := null_ptr
			end
			if output_buffer /= null_ptr then
				free_nongc_memory (output_buffer)
				output_buffer := null_ptr
			end
		end

feature -- Properties

	address: POINTER
			-- Address of area for select item descriptions

	output_buffer: POINTER
			-- Address of area where select item values,
			-- column returned lengths, column return codes
			-- and column indicator variables will be placed

	column_return_codes: POINTER
			-- Address of area where select item column return
			-- codes will be placed

	column_return_lengths: POINTER
			-- Address of area where select item returned column
			-- lengths will be placed

	column_indicator_vars: POINTER
			-- Address of area where select item indicator
			-- variables will be placed

	output_buffer_length: INTEGER
			-- Length of output buffer (space actually used)

	output_buffer_capacity: INTEGER
			-- Length of output buffer (space allocated)

	requested_buffer_limit: INTEGER
			-- Requested maximum length of output buffer.
			-- -1 means use default value for connection.
			-- If 0, then no limit is requested.

	output_buffer_limit: INTEGER
			-- Maximum length of output buffer.
			-- This is only a hint, so the actual maximum
			-- may be greater.
		do
			if requested_buffer_limit >= 0 then
				Result := requested_buffer_limit
			else
				Result := cursor.connection.query_buffer_size_limit
			end
		end

	capacity: INTEGER
			-- Maximum number of select items set is
			-- capable of holding

	raw_requested_row_capacity: INTEGER
			-- Raw requested number of rows in output buffer.
			-- -1 means use default value for connection.
			-- If 0, then calculate actual row capacity
			-- from `output_buffer_limit'.

	requested_row_capacity: INTEGER
			-- Calculated requested number of rows in output
			-- buffer, taking into account default for connection.
		do
			if raw_requested_row_capacity >= 0 then
				Result := raw_requested_row_capacity
			else
				Result := cursor.connection.query_rows_per_fetch
			end
		end

	reparse_required: BOOLEAN
			-- Does the SQL statement associated with this
			-- select set need to be reparsed due to a
			-- redefine which changed external datatypes or
			-- lengths?

	needs_oci_define: BOOLEAN
			-- Do the select items in this set need
			-- to be defined to the Oracle Call Interface?

	fetched_row_count: INTEGER
			-- Number of rows fetched on last OCI fetch

	row_index: INTEGER
			-- Position of next row in output buffer to be
			-- returned on `fetch' call

	end_of_data: BOOLEAN
			-- Has end of data been reached on a query?

	cursor: ORACLE_CURSOR
			-- Cursor with which `Current' is associated

	select_item_position: INTEGER
			-- Position of select item being defined
		do
			Result := select_item.position
		end

	is_integer_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel integers?
		require
			valid_index: k >= 1 and k <= capacity
		do
			search (k)
			Result := select_item.is_defined and then
				select_item.is_integer
		end

	is_real_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel reals?
		require
			valid_index: k >= 1 and k <= capacity
		do
			search (k)
			Result := select_item.is_defined and then
				select_item.is_real
		end

	is_double_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel doubles?
		require
			valid_index: k >= 1 and k <= capacity
		do
			search (k)
			Result := select_item.is_defined and then
				select_item.is_double
		end

	is_character_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel characters?
		require
			valid_index: k >= 1 and k <= capacity
		do
			search (k)
			Result := select_item.is_defined and then
				select_item.is_character
		end

	is_boolean_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel booleans?
		require
			valid_index: k >= 1 and k <= capacity
		do
			search (k)
			Result := select_item.is_defined and then
				select_item.is_boolean
		end

	is_date_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel database
			-- dates?
		require
			valid_index: k >= 1 and k <= capacity
		do
			search (k)
			Result := select_item.is_defined and then
				select_item.is_date
		end

	is_rowid_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel database
			-- rowids?
		require
			valid_index: k >= 1 and k <= capacity
		do
			search (k)
			Result := select_item.is_defined and then
				select_item.is_rowid
		end

	is_string_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel strings?
		require
			valid_index: k >= 1 and k <= capacity
		do
			search (k)
			Result := select_item.is_defined and then
				select_item.is_string
		end

	is_raw_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel raw
			-- strings?
		require
			valid_index: k >= 1 and k <= capacity
		do
			search (k)
			Result := select_item.is_defined and then
				select_item.is_raw
		end

	is_null (pos: INTEGER): BOOLEAN
			-- Is the `pos'-th select item in the
			-- `row_index'-th row null?
		require
			index_large_enough: row_index >= 1
			index_small_enough: row_index <= fetched_row_count
		do
			Result := unsigned_short_value (offset (column_return_codes, column_retcode_length * ((pos - 1) * row_capacity + (row_index - 1)))) = Oci_fetched_col_null_code
		end
	
	column_return_code (pos: INTEGER): INTEGER
			-- Column return code for the `pos'-th select item
			-- in the `row_index'-th row
		require
			index_large_enough: row_index >= 1
			index_small_enough: row_index <= fetched_row_count
		do
			Result := unsigned_short_value (offset (column_return_codes, column_retcode_length * ((pos - 1) * row_capacity + (row_index - 1))))
		end
	
	column_return_length (pos: INTEGER): INTEGER
			-- Returned column length for the `pos'-th select item
			-- in the `row_index'-th row
		require
			index_large_enough: row_index >= 1
			index_small_enough: row_index <= fetched_row_count
		do
			Result := unsigned_short_value (offset (column_return_lengths, column_length_length * ((pos - 1) * row_capacity + (row_index - 1))))
		end
	
	column_indicator_var (pos: INTEGER): INTEGER
			-- Value of indicator variable for the `pos'-th
			-- select item in the `row_index'-th row
		require
			index_large_enough: row_index >= 1
			index_small_enough: row_index <= fetched_row_count
		do
			Result := signed_short_value (offset (column_indicator_vars, column_indicator_length * ((pos - 1) * row_capacity + (row_index - 1))))
		end
	
	integer_value (pos: INTEGER): INTEGER
			-- Integer value for `pos'-th select item in
			-- `row_index'-th row
		require
			index_large_enough: row_index >= 1
			index_small_enough: row_index <= fetched_row_count
			column_is_integer: is_integer_column (pos)
		local
			sel: ORACLE_SELECT_ITEM
		do
			if is_null (pos) then
				Result := 0
			else
				sel := select_item
				sel.set_address (i_th_select_item (pos))
				Result := sel.integer_value (row_index)
			end
		end
	
	real_value (pos: INTEGER): REAL
			-- Real value for `pos'-th select item in
			-- `row_index'-th row
		require
			index_large_enough: row_index >= 1
			index_small_enough: row_index <= fetched_row_count
			column_is_real: is_real_column (pos)
		local
			sel: ORACLE_SELECT_ITEM
		do
			if is_null (pos) then
				Result := 0.0
			else
				sel := select_item
				sel.set_address (i_th_select_item (pos))
				Result := sel.real_value (row_index)
			end
		end
	
	double_value (pos: INTEGER): DOUBLE
			-- Double value for `pos'-th select item in
			-- `row_index'-th row
		require
			index_large_enough: row_index >= 1
			index_small_enough: row_index <= fetched_row_count
			column_is_double: is_double_column (pos)
		local
			sel: ORACLE_SELECT_ITEM
		do
			if is_null (pos) then
				Result := 0.0
			else
				sel := select_item
				sel.set_address (i_th_select_item (pos))
				Result := sel.double_value (row_index)
			end
		end
	
	character_value (pos: INTEGER): CHARACTER
			-- Character value for `pos'-th select item in
			-- `row_index'-th row
		require
			index_large_enough: row_index >= 1
			index_small_enough: row_index <= fetched_row_count
			column_is_character: is_character_column (pos)
		local
			sel: ORACLE_SELECT_ITEM
		do
			if is_null (pos) then
				Result := '%U'
			else
				sel := select_item
				sel.set_address (i_th_select_item (pos))
				Result := sel.character_value (row_index)
			end
		end
	
	boolean_value (pos: INTEGER): BOOLEAN
			-- Boolean value for `pos'-th select item in
			-- `row_index'-th row
		require
			index_large_enough: row_index >= 1
			index_small_enough: row_index <= fetched_row_count
			column_is_boolean: is_boolean_column (pos)
		local
			sel: ORACLE_SELECT_ITEM
			c: CHARACTER
			message: STRING
		do
			if is_null (pos) then
				Result := False
			else
				sel := select_item
				sel.set_address (i_th_select_item (pos))
				c := sel.boolean_character_value (row_index)
				if c = 'Y' then
					Result := True
				elseif c = 'N' then
					Result := False
				else
					create message.make (0)
					message.append ("invalid boolean for select item ")
					message.append_integer (pos)
					cursor.handle_oracle_error (dbex.Db_invalid_boolean,
						message)

				end
			end
		end
	
	string_value (pos: INTEGER): STRING
			-- String value for `pos'-th select item in
			-- `row_index'-th row
		require
			index_large_enough: row_index >= 1
			index_small_enough: row_index <= fetched_row_count
			column_is_string: is_string_column (pos)
		local
			sel: ORACLE_SELECT_ITEM
		do
			if is_null (pos) then
				Result := ""
			else
				sel := select_item
				sel.set_address (i_th_select_item (pos))
				Result := sel.string_value (row_index,
					column_return_length (pos))
			end
		end
	
	raw_value (pos: INTEGER): STRING
			-- Raw string value for `pos'-th select item in
			-- `row_index'-th row
		require
			index_large_enough: row_index >= 1
			index_small_enough: row_index <= fetched_row_count
			column_is_raw: is_raw_column (pos)
		local
			sel: ORACLE_SELECT_ITEM
		do
			if is_null (pos) then
				Result := ""
			else
				sel := select_item
				sel.set_address (i_th_select_item (pos))
				Result := sel.raw_value (row_index,
					column_return_length (pos))
			end
		end
	
	date_value (pos: INTEGER): ORACLE_DATE
			-- Database date value for `pos'-th select item in
			-- `row_index'-th row
		require
			index_large_enough: row_index >= 1
			index_small_enough: row_index <= fetched_row_count
			column_is_date: is_date_column (pos)
		local
			sel: ORACLE_SELECT_ITEM
		do
			if is_null (pos) then
				create Result.set_null
			else
				sel := select_item
				sel.set_address (i_th_select_item (pos))
				Result := sel.date_value (row_index)
				Result.set_nonnull_indicator
			end
		end
	
	rowid_value (pos: INTEGER): ORACLE_ROWID
			-- Database rowid value for `pos'-th select item in
			-- `row_index'-th row
		require
			index_large_enough: row_index >= 1
			index_small_enough: row_index <= fetched_row_count
			column_is_rowid: is_rowid_column (pos)
		local
			sel: ORACLE_SELECT_ITEM
		do
			if is_null (pos) then
				create Result.set_null
			else
				sel := select_item
				sel.set_address (i_th_select_item (pos))
				Result := sel.rowid_value (row_index)
			end
		end
	
feature -- Access to values

	get_string_value (pos: INTEGER dest: STRING)
			-- Get string value for `pos'-th select item in
			-- `row_index'-th row and store into `dest'
		require
			index_large_enough: row_index >= 1
			index_small_enough: row_index <= fetched_row_count
			column_is_string: is_string_column (pos)
			destination_exists: dest /= Void
		local
			sel: ORACLE_SELECT_ITEM
		do
			if is_null (pos) then
				dest.keep_head (0)
			else
				sel := select_item
				sel.set_address (i_th_select_item (pos))
				sel.get_string_value (row_index, column_return_length (pos), dest)
			end
		ensure
			destination_set: equal (dest, string_value (pos))
		end
	
	get_raw_value (pos: INTEGER dest: STRING)
			-- Get raw string value for `pos'-th select item in
			-- `row_index'-th row and store into `dest'
		require
			index_large_enough: row_index >= 1
			index_small_enough: row_index <= fetched_row_count
			column_is_raw: is_raw_column (pos)
			destination_exists: dest /= Void
		local
			sel: ORACLE_SELECT_ITEM
		do
			if is_null (pos) then
				dest.keep_head (0)
			else
				sel := select_item
				sel.set_address (i_th_select_item (pos))
				sel.get_raw_value (row_index, column_return_length (pos), dest)
			end
		ensure
			destination_set: equal (dest, raw_value (pos))
		end
	
	get_date_value (pos: INTEGER dest: ORACLE_DATE)
			-- Get date value for `pos'-th select item in
			-- `row_index'-th row and store into `dest'
		require
			index_large_enough: row_index >= 1
			index_small_enough: row_index <= fetched_row_count
			column_is_date: is_date_column (pos)
			destination_exists: dest /= Void
		local
			sel: ORACLE_SELECT_ITEM
		do
			if is_null (pos) then
				dest.set_null
			else
				sel := select_item
				sel.set_address (i_th_select_item (pos))
				sel.get_date_value (row_index, dest)
				dest.set_nonnull_indicator
			end
		ensure
			destination_set: equal (dest, date_value (pos))
		end
	
	get_rowid_value (pos: INTEGER dest: ORACLE_ROWID)
			-- Get rowid value for `pos'-th select item in
			-- `row_index'-th row and store into `dest'
		require
			index_large_enough: row_index >= 1
			index_small_enough: row_index <= fetched_row_count
			column_is_rowid: is_rowid_column (pos)
			destination_exists: dest /= Void
		local
			sel: ORACLE_SELECT_ITEM
		do
			if is_null (pos) then
				dest.set_null
			else
				sel := select_item
				sel.set_address (i_th_select_item (pos))
				sel.get_rowid_value (row_index, dest)
			end
		ensure
			destination_set: equal (dest, rowid_value (pos))
		end

feature -- Modification

	set_needs_oci_define
			-- Set "needs OCI define" indicator to True
		do
			needs_oci_define := True
		ensure
			indicator_set: needs_oci_define
		end

	set_reparse_required (newval: BOOLEAN)
			-- Set `reparse_required' to `newval'
		do
			reparse_required := newval
		ensure
			value_set: reparse_required = newval
		end

	set_requested_row_capacity (num: INTEGER)
			-- Set requested number of rows for which output buffer
			-- space should be allocated for each select item
			-- to `num'
		do
			raw_requested_row_capacity := num
		ensure
			row_capacity_set: requested_row_capacity = num
		end

	set_requested_buffer_limit (num: INTEGER)
			-- Set `requested_buffer_limit' to 'num'
		do
			requested_buffer_limit := num
		ensure
			value_set: requested_buffer_limit = num
		end

	resize (num: INTEGER)
			-- Change size of area for select items so
			-- that it can hold `num' items.
			-- No effect if `num' is equal to `capacity'.
		require
			nonnegative_count: num >= 0
		local
			new_addr: POINTER
			copy_len: INTEGER
		do
			if num /= capacity then
				new_addr := allocate_nongc_memory (num * select_item_length)
				initialize (new_addr, num)
				if address /= default_pointer then
					copy_len := capacity.min (num) * select_item_length
					copy_nongc_memory(address, new_addr, copy_len)
				end
				needs_oci_define := True
				if num < capacity then
					reparse_required := True
				end
				address := new_addr
				capacity := num
			end
		ensure
			correct_capacity: capacity = num
			address_changed: num /= old capacity implies address /= old address
			address_same: num = old capacity implies address = old address
		end

	search (sel_number: INTEGER)
			-- Find slot for select item number `sel_number'
			-- and put in `select_item'.  If slot is not
			-- already in use, set its number to `sel_number'.
		local
			sel: ORACLE_SELECT_ITEM
		do
			if sel_number > capacity then
				resize (sel_number)
			end
			sel := select_item
			sel.set_address (i_th_select_item (sel_number))
		end

	i_th_select_item (k: INTEGER): POINTER
			-- Address of slot for select item number `k'
		require
			positive_position: k > 0
		do
			Result := sel_i_th_address (address, k - 1)
		end

	select_integer (k: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- integers
		require
			valid_select_position: is_valid_select_position (k)
		local
			sel: ORACLE_SELECT_ITEM
		do
			search (k)
			sel := select_item
			if sel.is_oci_defined and then not sel.is_integer then
				reparse_required := True
			end
			sel.select_integer
		ensure
			is_integer_column: is_integer_column (k)
		end

	select_real (k: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- reals
		require
			valid_select_position: is_valid_select_position (k)
		local
			sel: ORACLE_SELECT_ITEM
		do
			search (k)
			sel := select_item
			if sel.is_oci_defined and then not sel.is_real then
				reparse_required := True
			end
			sel.select_real
		ensure
			is_real_column: is_real_column (k)
		end

	select_double (k: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- doubles
		require
			valid_select_position: is_valid_select_position (k)
		local
			sel: ORACLE_SELECT_ITEM
		do
			search (k)
			sel := select_item
			if sel.is_oci_defined and then not sel.is_double then
				reparse_required := True
			end
			sel.select_double
		ensure
			is_double_column: is_double_column (k)
		end

	select_character (k: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- character
		require
			valid_select_position: is_valid_select_position (k)
		local
			sel: ORACLE_SELECT_ITEM
		do
			search (k)
			sel := select_item
			if sel.is_oci_defined and then not sel.is_character then
				reparse_required := True
			end
			sel.select_character
		ensure
			is_character_column: is_character_column (k)
		end

	select_boolean (k: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- booleans
		require
			valid_select_position: is_valid_select_position (k)
		local
			sel: ORACLE_SELECT_ITEM
		do
			search (k)
			sel := select_item
			if sel.is_oci_defined and then not sel.is_boolean then
				reparse_required := True
			end
			sel.select_boolean
		ensure
			is_boolean_column: is_boolean_column (k)
		end

	select_date (k: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- database dates
		require
			valid_select_position: is_valid_select_position (k)
		local
			sel: ORACLE_SELECT_ITEM
		do
			search (k)
			sel := select_item
			if sel.is_oci_defined and then not sel.is_date then
				reparse_required := True
			end
			sel.select_date
		ensure
			is_date_column: is_date_column (k)
		end

	select_rowid (k: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- database rowids
		require
			valid_select_position: is_valid_select_position (k)
		local
			sel: ORACLE_SELECT_ITEM
		do
			search (k)
			sel := select_item
			if sel.is_oci_defined and then not sel.is_rowid then
				reparse_required := True
			end
			sel.select_rowid
		ensure
			is_rowid_column: is_rowid_column (k)
		end

	select_string (k: INTEGER maxlen: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- strings of maximum length `maxlen'
		require
			valid_select_position: is_valid_select_position (k)
			valid_max_length: is_valid_string_length (maxlen)
		local
			sel: ORACLE_SELECT_ITEM
		do
			search (k)
			sel := select_item
			if sel.is_oci_defined and then (not sel.is_string or
			   sel.max_value_length /= maxlen) then
				reparse_required := True
			end
			sel.select_string (maxlen)
		ensure
			is_string_column: is_string_column (k)
		end

	select_raw (k: INTEGER maxlen: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- raw strings of maximum length `maxlen'
		require
			valid_select_position: is_valid_select_position (k)
			valid_max_length: is_valid_string_length (maxlen)
		local
			sel: ORACLE_SELECT_ITEM
		do
			search (k)
			sel := select_item
			if sel.is_oci_defined and then (not sel.is_raw or
			   sel.max_value_length /= maxlen) then
				reparse_required := True
			end
			sel.select_raw (maxlen)
		ensure
			is_raw_column: is_raw_column (k)
		end

	check_for_undefined_select_items
			-- Check for "undefined select item" errors, which
			-- can occur when a cursor is reused
		local
			k, local_capacity: INTEGER
			item: ORACLE_SELECT_ITEM
			message: STRING
		do
			from
				k := 1
				item := select_item
				local_capacity := capacity
			until
				k > local_capacity
			loop
				item.set_address (i_th_select_item (k))
				if not item.is_defined then
					create message.make (0)
					message.append ("select item ")
					message.append_integer (item.position)
					message.append (" not defined")
					cursor.handle_oracle_error (dbex.Db_undefined_select_item,
						message)
				end
				
				k := k + 1
			end
		end
			
	fill_in_zero_lengths
			-- Fill in the length for all select items which
			-- have a zero length, by describing them
		local
			k, local_capacity, len: INTEGER
			item: ORACLE_SELECT_ITEM
			desc: ORACLE_DESCRIBE_ITEM
		do
			from
				k := 1
				item := select_item
				local_capacity := capacity
				desc := describe_item
			until
				k > local_capacity
			loop
				item.set_address (i_th_select_item (k))
				if item.value_length = 0 then
					if desc = Void then
						create describe_item.make
						desc := describe_item
					end
					desc.describe (cursor, item.position)
					len := desc.max_display_size
					if len = 0 then
						len := item.max_select_length
					end
					item.set_value_length (len)
				end
				
				k := k + 1
			end
		end
			

	oci_define
			-- Define all select item in `Current' which need
			-- defining via the Oracle Call Interface
		local
			k, local_capacity, space: INTEGER
			item: ORACLE_SELECT_ITEM
			buf_addr: POINTER
			col_retcodes, col_retlens, col_indicators: POINTER
		do
			-- Set actual row capacity, based on requested
			-- row capacity and output buffer size limit
			compute_row_capacity
			allocate_buffer_space (output_space_required)
			if needs_oci_define then
				from
					k := 1
					item := select_item
					local_capacity := capacity
					buf_addr := output_buffer
					col_retlens := column_return_lengths
					col_retcodes := column_return_codes
					col_indicators := column_indicator_vars
				until
					k > local_capacity
				loop
					item.set_address (i_th_select_item (k))
					item.oci_define (cursor, buf_addr, offset (col_retlens, column_length_length * row_capacity * (k - 1)),
						offset (col_retcodes, column_retcode_length * row_capacity * (k - 1)),
						offset (col_indicators, column_indicator_length * row_capacity * (k - 1)))
					space := aligned (item.value_length * row_capacity)
					buf_addr := offset (buf_addr, space)
					k := k + 1
				end
			
				needs_oci_define := False
			end
		end

	execute
			-- Do any processing needed when query is
			-- executed, before fetches are done
		do
			fetched_row_count := 0
			fetch_return_code := Oci_success_code
			row_index := 0
			end_of_data := False
		end

	terminate
			-- Do any processing needed when query is
			-- terminated
		do
			fetched_row_count := 0
			fetch_return_code := Oci_success_code
			row_index := 0
			end_of_data := True
			
			-- Revert to connection defaults
			
			raw_requested_row_capacity := -1
			requested_buffer_limit := -1
		end

	fetch
			-- Fetch next row for a query.  May not actually
			-- do an OCI fetch if another row is
			-- already available in output buffer
		local
			code, old_rows: INTEGER
		do
			if row_index >= fetched_row_count then
				inspect fetch_return_code
				when Oci_success_code then -- May be more rows
					old_rows := cursor.rows_processed_count
					cursor.enter_oci
					code := ora_fetch_multiple_rows (cursor.address, row_capacity)
					fetch_return_code := cursor.return_code
					fetched_row_count := cursor.rows_processed_count - old_rows
					cursor.leave_oci
					if fetched_row_count > 0 then
						row_index := 1
					elseif fetch_return_code = Oci_no_data_found_code then
						end_of_data := True
					else
						cursor.check_code (fetch_return_code)
					end
				when Oci_no_data_found_code then -- End of data
					end_of_data := True
				else	-- Error
					cursor.check_code (fetch_return_code)
				end
			else
				row_index := row_index + 1
			end
			if not end_of_data then
				check_column_return_codes
			end
		end

	undefine_all
			-- Undefine all items in select set so that it can
			-- be correctly reused
		local
			k, local_capacity: INTEGER
			item: ORACLE_SELECT_ITEM
		do
			from
				k := 1
				item := select_item
				local_capacity := capacity
			until
				k > local_capacity
			loop
				item.set_address (i_th_select_item (k))
				item.set_is_defined (False)
				k := k + 1
			end
		end

feature -- Display

	display
			-- Display `Current'
		local
			k, local_capacity: INTEGER
			item: ORACLE_SELECT_ITEM
		do
			from
				k := 1
				item := select_item
				local_capacity := capacity
			until
				k > local_capacity
			loop
				item.set_address (i_th_select_item (k))
				item.display
				k := k + 1
			end
		end

feature {NONE} -- Implementation

	row_capacity: INTEGER
			-- Number of rows output buffer can hold.

	set_row_capacity (num: INTEGER)
			-- Set number of rows for which output buffer
			-- space should be allocated for each select item
			-- to `num'
		do
			if num /= row_capacity then
				row_capacity := num
				needs_oci_define := True
				select_item.set_row_count (num)
			end
		ensure
			row_capacity_set: row_capacity = num
		end

	select_item: ORACLE_SELECT_ITEM
			-- Object used to access and modify a piece of C memory
			-- containing a select item as an Eiffel object.
			-- Having this as an attribute avoids repeated
			-- creation.

	describe_item: ORACLE_DESCRIBE_ITEM
			-- Object used to access and modify a piece of C memory
			-- containing a describe item for describing select
			-- items as an Eiffel object.
			-- Having this as an attribute avoids repeated
			-- creation.

	initialize (addr: POINTER num: INTEGER)
			-- Initialize `num' select item slots
			-- starting at address `addr'
		local
			k: INTEGER
			item: ORACLE_SELECT_ITEM
		do
			from
				k := 1
				item := select_item
			until
				k > num
			loop
				item.set_address (sel_i_th_address (addr, k - 1))
				item.initialize (k)
				k := k + 1
			end
		end

	fetch_return_code: INTEGER
			-- Return code from last OCI fetch

	check_column_return_codes
			-- Check column return codes for the `row_index'-th
			-- row and signal error if appropriate
		local
			pos, local_capacity, code: INTEGER
		do
			from
				pos := 1
				local_capacity := capacity
			until
				pos > local_capacity
			loop
				code := column_return_code (pos)
				if code /= Oci_success_code and code /= Oci_fetched_col_null_code then
					cursor.check_code (code)
				end
				pos := pos + 1
			end
			
		end

	allocate_buffer_space (len: INTEGER)
			-- Allocate output buffer space needed for this
			-- select set, resizing existing space if necessary.
		local
			size, decr1, decr2, decr3: INTEGER
			null_ptr: POINTER
		do
			if len > output_buffer_capacity then
				if output_buffer /= null_ptr then
					free_nongc_memory (output_buffer)
				end
				output_buffer := allocate_nongc_memory (len)
				output_buffer_capacity := len
			end
			if len /= output_buffer_length then
				needs_oci_define := True
				output_buffer_length := len
			end
			if needs_oci_define then
				size := capacity * row_capacity
				decr1 := column_length_length * size
				column_return_lengths := offset (output_buffer, output_buffer_length - decr1)
				decr2 := decr1 + column_retcode_length * size
				column_return_codes := offset (output_buffer, output_buffer_length - decr2)
				decr3 := decr2 + column_indicator_length * size
				column_indicator_vars := offset (output_buffer, output_buffer_length - decr3)
			end
		end

	Alignment: INTEGER = 8
			-- Most general alignment boundary

	aligned (num: INTEGER): INTEGER
			-- Value of `num' aligned to next higher alignment
			-- boundary if it is not already on one
		do
			if num \\ Alignment = 0 then
				Result := num
			else
				Result := (num // Alignment + 1) * Alignment
			end
		end

	compute_row_capacity
			-- Set `row_capacity' to the number of rows
			-- to be retrieved from database engine on each
			-- fetch.
		local
			rows, limit: INTEGER
		do
			rows := requested_row_capacity
			limit := output_buffer_limit
			check
				rows_or_limit_positive: rows > 0 or limit > 0
			end
			if rows > 0 then
				if limit > 0 then
					rows := rows.min (rows_dictated_by_limit (limit))
				end
			else	-- Limit must be positive
				
				rows := rows_dictated_by_limit (limit)
			end
			set_row_capacity (rows)
		ensure
			positive_row_capacity: row_capacity > 0
		end

	rows_dictated_by_limit (limit: INTEGER): INTEGER
			-- Largest number of rows which require output buffer
			-- space less than or equal to limit
		require
			positive_limit: limit > 0
		do
			Result := limit // approximate_space_per_row
			if Result = 0 then	
				Result := 1	-- At least one row
			end
		ensure
			nonnegative_result: Result > 0
		end

	approximate_space_per_row: INTEGER
			-- Approximate amount of output buffer space required
			-- for one row (does not include any padding needed
			-- for alignment)
		local
			k, local_capacity: INTEGER
			item: ORACLE_SELECT_ITEM
		do
			from
				k := 1
				item := select_item
				local_capacity := capacity
			until
				k > local_capacity
			loop
				item.set_address (i_th_select_item (k))
				Result := Result + item.value_length
				k := k + 1
			end
			Result := Result + (column_length_length + column_retcode_length + column_indicator_length) * local_capacity
			if Result = 0 then
				Result := 1
			end
		ensure
			positive_result: Result > 0
		end

	output_space_required: INTEGER
			-- Length of output buffer required to provide space
			-- for `row_capacity' rows for each select item
		local
			k, local_capacity, space: INTEGER
			item: ORACLE_SELECT_ITEM
		do
			from
				k := 1
				item := select_item
				local_capacity := capacity
			until
				k > local_capacity
			loop
				item.set_address (i_th_select_item (k))
				space := aligned (item.value_length * row_capacity)
				Result := Result + space
				k := k + 1
			end
			Result := Result + (column_length_length + column_retcode_length + column_indicator_length) * local_capacity * row_capacity
		ensure
			nonnegative_result: Result >= 0
		end

feature {NONE}	-- Externals

	sel_i_th_address (sel_addr: POINTER k: INTEGER): POINTER
			-- Address of select item `k' in the select
			-- set at address `sel_addr'
		external
			"C macro use %"oracle.h%""
		end

	select_item_length: INTEGER
			-- Length of one select item
		external
			"C macro use %"oracle.h%""
		alias
			"sel_length"
		end

	column_retcode_length: INTEGER
			-- Length of one column return code
		external
			"C macro use %"oracle.h%""
		end

	column_length_length: INTEGER
			-- Length of one returned column length
		external
			"C macro use %"oracle.h%""
		end

	column_indicator_length: INTEGER
			-- Length of one column indicator variable
		external
			"C macro use %"oracle.h%""
		end

invariant
	address_exists: address /= default_pointer
	select_item_exists: select_item /= Void
	valid_capacity: capacity >= 0
	valid_output_buffer: output_buffer_length <= output_buffer_capacity
	fetched_count_big_enough: fetched_row_count >= 0
	fetched_count_small_enough: fetched_row_count <= row_capacity
	row_index_big_enough: row_index >= 0
	row_index_small_enough: row_index <= row_capacity + 1

end
