
indexing
	
	author: "David Hollenberg"
	date:	"November 28, 1995"
	description: "The standard error handler for errors related to the %
		%Oracle database"

class ORACLE_STD_ERROR_HANDLER

inherit
	ORACLE_ERROR_HANDLER
	ORACLE_EXCEPTIONS
		export
			{NONE} all
		end
	OCI_CONSTANTS
		export
			{NONE} all
		end
	OS_ACCESS
		export
			{NONE} all
		end
	UNIX_OS_ACCESS
		export
			{NONE} all
		end
	STRING_UTILITIES_ACCESS
		export
			{NONE} all
		end

feature

	handle_oci_error (connection: ORACLE_CONNECTION cursor: ORACLE_CURSOR)
			-- Handle an Oracle Call Interface error which
			-- occurred on connection `connection' and which
			-- involves cursor `cursor' if non-void
		local
			desc: STRING
		do
			-- We are trying to track down a bug where the
			-- ftp_daemon gets stuck doing a rollback.
			-- We will set partial description of last error first,
			-- then set full description.  Partial description
			-- can be constructed without trying to talk to the
			-- database.  It will appear in error message that
			-- is generated if the ftp_daemon gets stuck trying
			-- to get the message corresponding to an error code
			-- and we have to kill it.
			
			desc := oci_partial_error_description (connection, cursor)
			connection.set_last_error_description (desc)
			desc := oci_error_description (connection, cursor)
			connection.set_last_error_description (desc)
			if not error_display_suppressed then
				display_error (desc)
			end
			raise (dbex.Db_call_interface_error)
		end

	handle_oci_error_with_code (connection: ORACLE_CONNECTION cursor: ORACLE_CURSOR code: INTEGER)
			-- Handle an Oracle Call Interface error which
			-- occurred on connection `connection' and which
			-- involves cursor `cursor' if non-void.  Use
			-- `code' as the error code instead of getting it
			-- from the return code field of the CDA (if
			-- `cursor' non-Void) or from the LDA
		local
			desc: STRING
		do
			desc := oci_partial_error_description (connection, cursor)
			connection.set_last_error_description (desc)
			desc := oci_error_description_for_code (connection, cursor, code)
			connection.set_last_error_description (desc)
			if not error_display_suppressed then
				display_error (desc)
			end
			raise (dbex.Db_call_interface_error)
		end

	handle_oracle_user_error (connection: ORACLE_CONNECTION cursor: ORACLE_CURSOR except, message: STRING)
			-- Handle an Oracle error which is detected by
			-- us rather than by the Oracle Call Interface.
			-- The error occurred on connection `connection' and
			-- involves cursor `cursor' if non-void.
			-- `except' is the exception tag to use if
			-- an exception is raised and `message' is the
			-- message to display.
		local
			desc: STRING
		do
			desc := oracle_user_error_description (connection, cursor, message)
			connection.set_last_error_description (desc)
			if not error_display_suppressed then
				display_error (desc)
			end
			raise (except)
		end

	display_error (s: STRING)
			-- Display Oracle database error whose description
			-- is `s'
		require
			description_exists: s /= Void
		do
			io.put_string (Separator)
			io.put_string (s)
		end

	oci_partial_error_description (connection: ORACLE_CONNECTION cursor: ORACLE_CURSOR): STRING
			-- Return a partial description of an Oracle
			-- Call Interface error which occurred on
			-- connection `connection' and which involves
			-- cursor `cursor' if non-void.  The description
			-- is partial in that the text corresponding
			-- to the error code is not included
		require
			connection_exists: connection /= Void
		do
			Result := oci_error_description_for_message (
				connection, cursor, Message_not_yet_available)
		ensure
			result_exists: Result /= Void
		end

	oci_error_description (connection: ORACLE_CONNECTION cursor: ORACLE_CURSOR): STRING
			-- Return a description of an Oracle Call Interface
			-- error which occurred on connection `connection'
			-- and which involves cursor `cursor' if non-void
		local
			code: INTEGER
		do
			if cursor /= Void then
				code := cursor.return_code
			else
				code := connection.return_code
			end
			Result := oci_error_description_for_message (
				connection, cursor,
				connection.error_message_for_code (code))
		end

	oci_error_description_for_code (connection: ORACLE_CONNECTION cursor: ORACLE_CURSOR code: INTEGER): STRING
			-- Return a description of an Oracle Call Interface
			-- error which occurred on connection `connection'
			-- and which involves cursor `cursor' if non-void,
			-- using `code' as the code of the error that
			-- occurred
		do
			Result := oci_error_description_for_message (
				connection, cursor,
				connection.error_message_for_code (code))
		end

	oci_error_description_for_message (connection: ORACLE_CONNECTION cursor: ORACLE_CURSOR message: STRING): STRING
			-- Return a description of an Oracle Call Interface
			-- error which  occurred on connection `connection'
			-- and which involves cursor `cursor' if non-void.
			-- Use `message' as the explanation for the error code
		require
			connection_exists: connection /= Void
			message_exists: message /= Void
		local
			os_code, oci_code, mark: INTEGER
		do
			create Result.make (256)
			if cursor /= Void then
				os_code := cursor.os_error_code
			else
				os_code := connection.os_error_code
			end
			append_date (Result)
			Result.append ("Oracle Call Interface error%N")
			Result.append (message)
			if message.item (message.count) /= '%N' then
				Result.append_character ('%N')
			end
			Result.append_character ('%N')
			append_os_error_code (os_code, Result)
			if cursor /= Void then
				if cursor.sql_statement /= Void then
					append_placeholder_name (cursor, Result)
					oci_code := cursor.oci_function_code
					inspect oci_code
					when Obndrn_code, Obndrv_code, Odescr_code, Odefin_code, Oexec_code then
						mark := cursor.parse_error_offset + 1
					else
						mark := 0
					end
					append_sql_statement (cursor, mark, Result)
					cursor.append_bindings_description (Result)
					Result.extend ('%N')
				end
				append_marked_position (cursor, Result)
				append_oci_function (cursor, Result)
				append_sql_function (cursor, Result)
				append_rowid (cursor, Result)
				append_rows_processed (cursor, Result)
			end
			append_connection (connection, Result)
		ensure
			result_exists: Result /= Void
		end

	oracle_user_error_description (connection: ORACLE_CONNECTION cursor: ORACLE_CURSOR message: STRING): STRING
			-- Return a description of an Oracle error which is
			-- detected by us rather than by the Oracle Call
			-- Interface.
			-- The error occurred on connection `connection' and
			-- involves cursor `cursor' if non-void.
			-- `message' is the message to display.
		do
			create Result.make (256)
			append_date (Result)
			Result.append ("Oracle error: ")
			Result.append (message)
			Result.extend ('%N')
			Result.extend ('%N')
			if cursor /= Void then
				append_sql_statement (cursor, 0, Result)
				cursor.append_bindings_description (Result)
				Result.extend ('%N')
			end
			append_connection (connection, Result)
		ensure
			result_exists: Result /= Void
		end

feature {NONE} -- Implementation

	Separator: STRING = "%N------------------------------------------------------------------------------%N"

	Message_not_yet_available: STRING = "(Message is not available yet)"

	append_date (dest: STRING)
			-- Append the current date to `dest'
		require
			destination_exists: dest /= Void
		do
			dest.append (os.current_date_string)
			dest.extend ('%N') dest.extend ('%N')
		end

	append_connection (connection: ORACLE_CONNECTION dest: STRING)
			-- Append to `dest' information about `connection'
		require
			connection_exists: connection /= Void
			destination_exists: dest /= Void
		do
			if connection.user_id /= Void then
				dest.append ("Oracle user id:    %"")
				dest.append (connection.user_id)
				dest.append ("%"%N")
			end
			dest.append ("Connect string:    ")
			dest.append (connection.printable_connect_string)
			dest.extend ('%N')
		end

	append_sql_statement (cursor: ORACLE_CURSOR mark: INTEGER dest: STRING)
			-- Append the SQL statement associated with `cursor',
			-- if any, to `dest'.  An error occurred at
			-- position `mark' within the statement (no error
			-- if `mark' is 0)
		require
			cursor_exists: cursor /= Void
			nonnegative_position: mark >= 0
			destination_exists: dest /= Void
		local
			stmt: STRING
		do
			stmt := cursor.sql_statement
			if stmt /= Void then
				dest.append ("SQL statement: %N")
				su.append_string_indented (Statement_prefix, stmt, mark, dest)
			end
		end
	
	Statement_prefix: STRING = "    "
			-- Prefix displayed before each line of a SQL
			-- statement

	append_marked_position (cursor: ORACLE_CURSOR dest: STRING)
			-- Append to `dest' the numerical position marked
			-- by the asterisk which is at the parse error
			-- offset position
		require
			cursor_exists: cursor /= Void
			destination_exists: dest /= Void
		do
			dest.append ("Marked position:   ")
			dest.append_integer (cursor.parse_error_offset + 1)
			dest.extend ('%N')
		end

	append_sql_function (cursor: ORACLE_CURSOR dest: STRING)
			-- Append to `dest' the SQL function indicated
			-- by `cursor'
		require
			cursor_exists: cursor /= Void
			destination_exists: dest /= Void
		local
			sql_function: INTEGER
		do
			sql_function := cursor.sql_function_code
			dest.append ("SQL function code: ")
			dest.append (name_for_sql_code (sql_function))
			dest.append (" (")
			dest.append_integer (sql_function)
			dest.append (")%N")
		end

	append_oci_function (cursor: ORACLE_CURSOR dest: STRING)
			-- Append to `dest' the OCI function indicated
			-- by `cursor'
		require
			cursor_exists: cursor /= Void
			destination_exists: dest /= Void
		local
			oci_function: INTEGER
		do		
			oci_function := cursor.oci_function_code
			dest.append ("OCI function code: ")
			dest.append (name_for_oci_code (oci_function))
			dest.append (" (")
			dest.append_integer (oci_function)
			dest.append (")%N")
		end
			
	append_placeholder_name (cursor: ORACLE_CURSOR dest: STRING)
			-- Append the placeholder on which bind error
			-- occurred, if any, to `dest'
		require
			cursor_exists: cursor /= Void
			destination_exists: dest /= Void
		local
			oci_function: INTEGER
			ph_name: STRING
		do		
			oci_function := cursor.oci_function_code
			if oci_function = Obndrn_code then
				ph_name := cursor.numeric_bindings.placeholder_name
			elseif oci_function = Obndrv_code then
				ph_name := cursor.named_bindings.placeholder_name
			end
			if ph_name /= Void then
				dest.append ("Placeholder name: ")
				dest.append (ph_name)
				dest.extend ('%N')
			end
		end
			
	append_rowid (cursor: ORACLE_CURSOR dest: STRING)
			-- Append to `dest' the rowid of last row
			-- processed for `cursor' (all zeroes if none)
		require
			cursor_exists: cursor /= Void
			destination_exists: dest /= Void
		local
			rowid: ORACLE_ROWID
		do
			rowid := cursor.last_rowid
			dest.append ("Last rowid:        ")
			dest.append (rowid.out)
			dest.extend ('%N')
		end

	append_rows_processed (cursor: ORACLE_CURSOR dest: STRING)
			-- Append to `dest' the count of rows processed
			-- for `cursor'
		require
			cursor_exists: cursor /= Void
			destination_exists: dest /= Void
		do
			dest.append ("Rows processed:    ")
			dest.append_integer (cursor.rows_processed_count)
			dest.extend ('%N')
		end

	append_os_error_code (code: INTEGER dest: STRING)
			-- Append description of OS error code `code' to `dest'
		require
			destination_exists: dest /= Void
		do
			if code /= 0 then
				dest.append ("OS error code:    ")
				dest.append_integer (code)
				dest.extend ('%N')
				dest.append ("OS error message: ")
				dest.append (unix_os.error_string (code))
				dest.extend ('%N')
			end
		end

end
