
indexing
	
	author: "David Hollenberg"
	date:	"February 21, 1996"
	description: "A database transaction"

class ORACLE_TRANSACTION

inherit
	ANY
	DB_EXCEPTION_ACCESS
		export
			{NONE} all
		end

create
	make

feature -- Creation

	make (conn: ORACLE_CONNECTION)
			-- Create transaction associated database
			-- connection `conn'
		do
			create item_list.make (Initial_item_count)
			create posted_list.make (Initial_item_count)
			create rowid_table.make (Initial_item_count)
			connection := conn
		ensure
			connection_set: connection = conn
		end

feature -- Properites

	connection: ORACLE_CONNECTION
			-- Connection to which this transaction applies

	number: INTEGER
			-- Number of this transaction

	date: ORACLE_DATE
			-- Date associated with this transaction

	is_changed: BOOLEAN
			-- Have any changes been made to the database
			-- during this transaction (posted or unposted)?

	is_registered (item: POSTABLE): BOOLEAN
			-- Is `item' registered as part of `Current'?
		local
			rowid: DB_ROWID
		do
			rowid := item.rowid
			Result := item_list.has (item)
			if rowid /= Void then
				Result := Result and rowid_table.has (rowid)
			end
		end

	has_registered_items: BOOLEAN
			-- Are there any registered items in this transaction
			-- (i.e.,  unposted items)?
		do
			Result := not item_list.is_empty
		end

feature -- Operations

	start
			-- Start a new transaction
		do
			number := number + 1
			date := Void
			is_changed := False
		ensure
			number_set: number = old number + 1
			no_transaction_date: date = Void
			not_changed: not is_changed
		end

	set_date (d: ORACLE_DATE)
		require
			date_exists: d /= Void
		do
			date := d
		ensure
			date_set: date = d
		end

	register (item: POSTABLE)
			-- Register `item' as part of this transaction
		require
			item_exists: item /= Void
		local
			rowid: DB_ROWID
		do
			rowid := item.rowid
			if is_registered (item) then
				handle_reregister_error (rowid)
			end
			if item.connection /= connection then
				handle_wrong_connection_error (rowid)
			end
			item_list.extend (item)
			is_changed := True
			if rowid /= Void then
				rowid_table.put (item, rowid)
			end
		ensure
			item_registered: is_registered (item)
		end

	reregister_first (item: POSTABLE)
			-- Re-register already registered `item' as first
			-- element in registration list
		require
			item_exists: item /= Void
			item_registered: item.is_registered
		local
			rowid: DB_ROWID
		do
			rowid := item.rowid
			if not is_registered (item) then
				handle_unregistered_error (rowid)
			end
			item_list.prune_all (item)
			item_list.put_front (item)
		ensure
			item_reregistered: item_list.first = item
		end

	reregister_last (item: POSTABLE)
			-- Re-register already registered `item' as last
			-- element in registration list
		require
			item_exists: item /= Void
			item_registered: item.is_registered
		local
			rowid: DB_ROWID
		do
			rowid := item.rowid
			if not is_registered (item) then
				handle_unregistered_error (rowid)
			end
			item_list.prune_all (item)
			item_list.extend (item)
		ensure
			item_reregistered: item_list.last = item
		end

	set_item_rowid (item: POSTABLE)
			-- Associate rowid with `item' after insert
		require
			item_not_void: item /= Void
		do
			if item.rowid /= Void then
				rowid_table.put (item, item.rowid)
			end
		end

	deregister (item: POSTABLE)
			-- Deregister `item'
		require
			is_registered: is_registered (item)
		local
			rowid: DB_ROWID
		do
			rowid := item.rowid
			if not is_registered (item) then
				handle_unregistered_error (rowid)
			end
			item_list.prune_all (item)
			if rowid /= Void then
				rowid_table.remove (rowid)
			end
		ensure
			not_register: not is_registered (item)
		end

	deregister_all
			-- Deregister all items which are part of this
			-- transaction, without posting them
		do
			item_list.wipe_out
			rowid_table.clear_all
		ensure
			no_registered_items: not has_registered_items
		end

	rollback
			-- Roll back current transaction
		local
			list: ARRAYED_LIST [POSTABLE]
		do
			-- Mark all registered and previously posted items
			-- as corrupt
			from
				create list.make (item_list.count + posted_list.count)
				list.append (item_list)
				list.append (posted_list)
				list.start
			until
				list.exhausted
			loop
				list.item.set_corrupted
				list.forth
			end
			deregister_all
			posted_list.wipe_out
		ensure
			no_registered_items: not has_registered_items
		end

	commit
			-- Commit current transaction
		do
			posted_list.wipe_out
		end

	post
			-- Post all items which are part of this
			-- transaction and then deregister them
		local
			list: ARRAYED_LIST [POSTABLE]
		do
			from
				list := item_list
				list.start
			until
				list.exhausted
			loop
				list.item.post
				list.forth
			end
			posted_list.append (list)
			deregister_all
		ensure
			no_registered_items: not has_registered_items
		end

feature {NONE} -- Implementation

	Initial_item_count: INTEGER = 10
			-- Initial number of items for which `item_list'
			-- has space

	item_list: ARRAYED_LIST [POSTABLE]
			-- List of items which are part of this transaction

	posted_list: ARRAYED_LIST [POSTABLE]
			-- List of items which have been posted during
			-- current transaction

	rowid_table: HASH_TABLE [POSTABLE, DB_ROWID]
			-- Postable items with a rowid which are part
			-- of this transaction, indexed by rowid

	
	handle_reregister_error (rowid: DB_ROWID)
			-- Handle attempt to register an item with a
			-- rowid which is already registered
		do		
			handle_error (dbex.Db_already_registered,
				"attempt to register item which is already registered", rowid)
		end
	
	handle_unregistered_error (rowid: DB_ROWID)
			-- Handle attempt to re-register an item which is
			-- not registered
		do		
			handle_error (dbex.Db_not_registered,
				"attempt to re-register item which is not registered", rowid)
		end
	
	handle_wrong_connection_error (rowid: DB_ROWID)
			-- Handle attempt to register an item on the wrong
			-- connection
		do		
			handle_error (dbex.Db_wrong_connection,
				"item registered with wrong database connection", rowid)
		end

	handle_error (except, msg: STRING rowid: DB_ROWID)
			-- Handle registration error with exception tag
			-- `except', message `msg' and rowid `rowid'
		local
			message: STRING
		do		
			create message.make (0)
			message.append (msg)
			if rowid /= Void then
				message.append (" - rowid is ")
				message.append (rowid.out)
			end
			connection.handle_oracle_user_error (Void,
				except, message)
		end

invariant
	valid_transaction_number: number >= 0
	item_list_exists: item_list /= Void
	rowid_table_exists: rowid_table /= Void
	connection_exists: connection /= Void

end
