
indexing
	
	description: "List of packer boxes as used in ICL"
	author: "David Hollenberg"
	date: "January 26, 2006"

class PACKER_BOX_LIST
inherit
	ARRAYED_LIST [PACKER_BOX]
	STRING_UTILITIES_ACCESS
		export
			{NONE} all
		undefine
			is_equal, copy
		end
	NUMBER_UTILITIES_ACCESS
		export
			{NONE} all
		undefine
			is_equal, copy
		end
	CONDITION_UTILITIES
		export
			{NONE} all
		undefine
			is_equal, copy
		end

create
	make, make_from_named_file, make_from_file, make_from_packing_version

create {PACKER_BOX_LIST}
	make_filled

feature {NONE} -- Creation

	make_from_named_file (fname: STRING)
			-- Create from file named `fname'
		local
			open_op: FILE_OPEN_READ_OP
			close_op: FILE_CLOSE_OP
			f: like file
		do
			create f.make (fname)
			create open_op
			open_op.open (f)
			make_from_file (f)
			create close_op
			close_op.close (f)
		end

	make_from_file (f: like file)
			-- Create from file `f'
		do
			make (0)

			file := f
			done := False
			in_boxes := False
			line_number := 0
		
			from
				read_line
			until
				file.end_of_file or done
			loop
				process_line
				read_line
			end
			
			if file.end_of_file then
				report_read_error ("unexpected premature end of file")
			end
		end

	make_from_packing_version (pv: PACKING_VERSIONS_ROW)
			-- Create from packing version `pv'
		require
			pv_not_void: pv /= Void
		local
			dies: LIST [PACKING_DIES_ROW]
		do
			dies := pv.dies
			packing := pv.packing
			family_processes_table := pv.family_processes_table
			wafer_techns_row := packing.wafer_techns_row
			runs_row := packing.run
			make (dies.count)
			from
				dies.start
			until
				dies.after
			loop
				append_packer_box (dies.item)
				dies.forth
			end
		end

feature -- Access

	cust_die (m_fab_no: INTEGER): PACKER_BOX
			-- Cust die packer box with `m_fab_no' or Void if none
		do
			Result := cust_die_table.item (m_fab_no)
		end

	test_vehicle (design_no: INTEGER): PACKER_BOX
			-- Test vehicle packer box with `design_no' or
			-- Void if none
		do
			Result := test_vehicle_table.item (design_no)
		end

	vendor_geometry (geom_id: STRING): PACKER_BOX
			-- Vendor geometry/ghost packer box with `geom_id' or
			-- Void if none
		do
			Result := vendor_geometry_table.item (geom_id)
		end

feature -- Operations

	write_to_named_file (fname: STRING)
			-- Write to file named `fname'
		local
			open_op: FILE_OPEN_WRITE_OP
			close_op: FILE_CLOSE_OP
			f: like file
		do
			create f.make (fname)
			create open_op
			open_op.open (f)
			write_to_file (f)
			create close_op
			close_op.close (f)
		end

	write_to_file (f: like file)
			-- Write to file `f'
		do
			file := f
			write_line (" ")
			write_line ("%"Written by Unix Delta%"")
			write_line (" ")
			write_line ("begin   var pbs = packer_boxes;")
			write_line ("do")
			write_line ("pbs := nil;")
			
			from
				start
			until
				after
			loop
				write_line ("pbs ::= $$ {")
				write_box_to_file (item)
				write_line (" };")
				forth
			end
			
			write_line ("pbs := refresh(pbs);")
			write_line ("give pbs")
			write_line ("end")
		end

	minus_external_projects_and_macrocells: like Current
			-- New copy of `Current' with external projects and
			-- macrocells removed
		local
			die_type: DIE_TYPE_VALUES
			pb: PACKER_BOX
		do
			Result := twin
			Result.wipe_out
			from
				start
			until
				after
			loop
				pb := item
				if not pb.project_type.is_equal (die_type.Macrocell) and not pb.is_external then
					Result.extend (pb)
				end
				forth
			end
		end

	setup_for_fast_access
			-- Create hash tables used to quickly find
			-- packer box of particular type and identification
		do
			create cust_die_table.make (count)
			create test_vehicle_table.make (count)
			create vendor_geometry_table.make (count)
			do_all (agent add_packer_box_to_table)
		end

feature {NONE} -- Implementation

	add_packer_box_to_table (pb: PACKER_BOX)
			-- Add `pb' to appropriate table, based on project
			-- type
		local
			die_type: DIE_TYPE_VALUES
			pb_type: STRING
		do
			pb_type := pb.project_type
			if pb_type.is_equal (die_type.cust_die) then
				cust_die_table.put (pb, pb.m_fab_no)
				if not cust_die_table.inserted then
					handle_hard_error ("duplicate m_fab_no among cust dies in packer box list")
				end
			elseif pb_type.is_equal (die_type.test_vehicle) then
				test_vehicle_table.put (pb, pb.name.to_integer)
				if not test_vehicle_table.inserted then
					handle_hard_error ("duplicate design_no among test vehicles in packer box list")
				end
			elseif pb_type.is_equal (die_type.vendor_geometry) or pb_type.is_equal (die_type.vendor_ghost) then
				vendor_geometry_table.put (pb, pb.name)
				if not vendor_geometry_table.inserted then
					handle_hard_error ("duplicate geomtry ID among vendor geometries/ghosts in packer box list")
				end
			else
				-- Ignore other project types
			end
		end

	append_packer_box (die: PACKING_DIES_ROW)
			-- Append to `Current' a packer box representation
			-- of `die'
		require
			die_not_void: die /= Void
		local
			pb: PACKER_BOX
			proj: PACKING_PROJECTS_ROW
			acct: CUST_ACCTS_ROW
			max_die: CONST_POINT
			rot_disallowed, move_disallowed: BOOLEAN
			vend_geom: GEOMETRIES_ROW
			vend_locs: LIST [GEOMETRYXFIXED_LOCATIONS_ROW]
			vend_quants: LIST [GEOMETRY_QUANTITIES_ROW]
			req_num, rot: INTEGER
			die_locs: LIST [PACKING_DIE_LOCATIONS_ROW]
			die_loc: PACKING_DIE_LOCATIONS_ROW
			packer_die_locs: ARRAYED_LIST [PACKER_DIE_LOCATION]
			pdl: PACKER_DIE_LOCATION
			etest_id, assigned_etest_id: STRING
		do
			create pb
			if die.has_project then
				proj := die.project
				acct := proj.cust_account
			end
			
			pb.set_project_type (die.die_type)
			pb.set_is_external (die.is_external_to_payload)
			if die.has_project then
				pb.set_design_name (proj.design_name)
			end
			if die.has_project then
				pb.set_name (die.design_no.out)
			elseif die.is_vendor_geometry_or_ghost then
				pb.set_name (die.geometry_id)
			elseif die.is_horizontal_scribe then
				pb.set_name (pb.horizontal_scribe_name + die.die_id)
			elseif die.is_vertical_scribe then
				pb.set_name (pb.vertical_scribe_name + die.die_id)
			elseif die.is_filler then
				pb.set_name (die.filler_name)
			end
			if die.has_project then
				pb.set_cust_account_display_name (acct.display_name)
			end
			if die.is_cust_die then
				pb.set_m_fab_no (proj.m_fab_no)
				pb.set_cust_account_no (acct.cust_acct_no)
				pb.set_cust_account_type (acct.acct_type)
				if not acct.is_null_acct_sponsor_id then
					pb.set_cust_account_sponsor (acct.acct_sponsor_id)
				end
			end
			
			if die.has_project then
				pb.set_min_die_size (icl_minimum_die_size (proj))
			else
				pb.set_min_die_size (die.upright_printed_size)
			end

			if die.has_project then
				max_die := proj.max_die_size
				if max_die /= Void then
					pb.set_max_die_size (max_die)
				end
			end

			if die.has_project then
				rot_disallowed := not proj.is_null_design_wafer_rotation_reqd
			elseif die.is_vendor_geometry_or_ghost then
				rot_disallowed := True
			else
				rot_disallowed := False
			end
			pb.set_not_rotatable (rot_disallowed)

			if die.is_vendor_geometry_or_ghost then
				vend_geom := die.vendor_geometry
				vend_locs := vend_geom.fixed_locations (packing.field_set_name)
				vend_quants := vend_geom.quantities (packing.field_set_name)
				move_disallowed := not vend_locs.is_empty
				if not vend_locs.is_empty and not vend_quants.is_empty then
					handle_hard_error ("vendor geometry " + vend_geom.geometry_id + " has both fixed locations and quantities")
					
				end
			else
				move_disallowed := False
			end
			pb.set_not_movable (move_disallowed)
			
			pb.set_is_tiny_chip (False)
			
			if die.is_cust_die then
				req_num := proj.fab_instance.work_order.ic_quantity_ordered
			elseif die.is_test_vehicle then
				req_num := -proj.test_vehicle.quantity_sites
			elseif die.is_vendor_geometry_or_ghost then
				if not vend_quants.is_empty then
					req_num := -vend_quants.first.quantity_sites
				else
					req_num := 0
				end
			else
				req_num := 0
			end
			pb.set_requested_number (req_num)

			-- Skip requested_display_name (not used by ICL code)
			-- Skip display_name (always set by packer)
			-- Skip wafer_good_sites (not used by ICL code)

			if die.is_cust_die then
				etest_id := family_processes_table.item (die.wafer_process).wafer_etest_id
				assigned_etest_id := die.assigned_wafer_etest_id
				if not etest_id.is_equal (assigned_etest_id) then
					handle_hard_error ("assigned wafer etest ID " + su.double_quoted (assigned_etest_id) + " different from etest ID " + su.double_quoted (etest_id) + " computed from process " + die.wafer_process)
				end
				pb.set_process_name (die.wafer_process)
				pb.set_wafer_etest_id (etest_id)
			end

			rot := die.die_wafer_rotation
			inspect rot
			when 0, 90, 180, 270 then
			else
				handle_hard_error ("die has rotation that is not 0, 90, 180 or 270")
			end
			pb.set_is_rotated (rot = 90 or rot = 270)

			if die.has_die_size then
				pb.set_die_size (die.printed_size_on_wafer)
			end
			
			if not die.is_null_die_id then
				pb.set_die_id (die.die_id)
			end
			
			die_locs := die.die_locations
			create packer_die_locs.make (die_locs.count)
			from
				die_locs.start
			until
				die_locs.after
			loop
				die_loc := die_locs.item
				create pdl
				pdl.set_field_name (die_loc.field_name)
				pdl.set_location (die_loc.location)
				-- Skip mask_site_no (always set by packer)
				pdl.set_mask_site_no (die_loc.mask_site_no)
				-- Skip section_number, is_new and chip_number
				-- (obsolete)
				packer_die_locs.extend (pdl)
				die_locs.forth
			end
			pb.set_positions (packer_die_locs)
			extend (pb)
		end

	write_box_to_file (pb: PACKER_BOX)
			-- Write `pb' to `file'
		local
			locs: LIST [PACKER_DIE_LOCATION]
			loc: PACKER_DIE_LOCATION
			is_first: BOOLEAN
		do
			write ("[")
			
			if pb.project_type /= Void then
				write ("project_type: ")
				write_line (pb.project_type)
			end
			
			if pb.name /= Void then
				write (" name: ")
				write_line (quoted (pb.name))
			end
			
			if pb.design_name /= Void then
				write (" design_name: ")
				write_line (quoted (pb.design_name))
			end
			
			if pb.cust_account_display_name /= Void then
				write (" cust_account_display_name: ")
				write_line (quoted (pb.cust_account_display_name))
			end
			
			if pb.cust_account_no /= 0 then
				write (" cust_account_no: ")
				write_line (pb.cust_account_no.out)
			end
			
			if pb.cust_account_type /= Void then
				write (" cust_account_type: ")
				write_line (quoted (pb.cust_account_type))
			end
			
			if pb.cust_account_sponsor /= Void then
				write (" cust_account_sponsor: ")
				write_line (quoted (pb.cust_account_sponsor))
			end
			
			if pb.is_external then
				write (" external: ")
				write_line (boolean_to_string (pb.is_external))
			end
			
			if pb.m_fab_no /= 0 then
				write (" m_fab_no: ")
				write_line (pb.m_fab_no.out)
			end
			
			if pb.priority /= 0 then
				write (" priority: ")
				write_line (pb.priority.out)
			end
			
			if pb.min_die_size /= Void then
				write (" min_die_size: ")
				write_line (point_to_string (pb.min_die_size))
			end
			
			if pb.max_die_size /= Void then
				write (" max_die_size: ")
				write_line (point_to_string (pb.max_die_size))
			end
			
			if pb.not_rotatable then
				write (" not_rotatable: ")
				write_line (boolean_to_string (pb.not_rotatable))
			end
			
			if pb.not_movable then
				write (" not_movable: ")
				write_line (boolean_to_string (pb.not_movable))
			end
			
			if pb.is_tiny_chip then
				write (" tiny_chip: ")
				write_line (boolean_to_string (pb.is_tiny_chip))
			end
			
			if pb.requested_number /= 0 then
				write (" requested_number: ")
				write_line (pb.requested_number.out)
			end
			
			if pb.requested_display_name /= Void then
				write (" requested_display_name: ")
				write_line (quoted (pb.requested_display_name))
			end
			
			if pb.process_name /= Void then
				write (" process_name: ")
				write_line (quoted (pb.process_name))
			end
			
			if pb.wafer_etest_id /= Void then
				write (" wafer_etest_id: ")
				write_line (quoted (pb.wafer_etest_id))
			end
			
			if pb.display_name /= Void then
				write (" display_name: ")
				write_line (quoted (pb.display_name))
			end
			
			if pb.wafer_good_sites /= 0 then
				write (" wafer_good_sites: ")
				write_line (pb.wafer_good_sites.out)
			end
			
			if pb.is_rotated then
				write (" rotated: ")
				write_line (boolean_to_string (pb.is_rotated))
			end
			
			if pb.die_size /= Void then
				write (" die_size: ")
				write_line (point_to_string (pb.die_size))
			end
			
			if pb.die_id /= Void then
				write (" die_id: ")
				write_line (quoted (pb.die_id))
			end
			
			if pb.positions /= Void and then not pb.positions.is_empty then
				write_line (" positions: {")
				from
					locs := pb.positions
					is_first := True
					locs.start
				until
					locs.after
				loop
					loc := locs.item
					if not is_first then
						write_line (" ;")
					end
					
					write ("    [")
					
					if loc.field_name /= Void then
						write ("field_name: ")
						write (quoted (loc.field_name))
					end
			
					if loc.location /= Void then
						write ("   location: ")
						write (point_to_string (loc.location))
					end
			
					if loc.mask_site_no /= 0 then
						write ("   mask_site_no: ")
						write (loc.mask_site_no.out)
					end
			
					if loc.section_number /= 0 then
						write (" section_number: ")
						write (loc.section_number.out)
					end
			
					if loc.is_new then
						write (" new: ")
						write (boolean_to_string (loc.is_new))
					end
			
					if loc.chip_number /= 0 then
						write (" chip_number: ")
						write (loc.chip_number.out)
					end
			
					write ("]")
					
					is_first := False
					locs.forth
				end
				write (" }")
			end
			
			write ("]")
		end

	write_line (s: STRING)
			-- Write `s' to `file', followed by newline
		do
			write (s)
			write ("%N")
		end

	write (s: STRING)
			-- Write `s' to `file'
		do
			file.put_string (s)
		end

	point_to_string (p: CONST_POINT): STRING
			-- `p' as a string in format ICL expects
		do
			Result := double_to_string (p.x) + " # " + double_to_string (p.y)
		end
	
	double_to_string (d: DOUBLE): STRING
			-- `d' as a string in format ICL expects
		do
			Result := nu.fixed_formatted_number (d, icl_precision (d))
			if Result.item (1) = '0' and not Result.is_equal (Zero_constant) then
				Result.remove_head (1)
			elseif Result.item (1) = '-' and then  Result.item (2) = '0' then
				Result.remove (2)
			end
		end
	
	Zero_constant: STRING = "0.0"

	icl_precision (d: DOUBLE): INTEGER
			-- Number of digits we expect ICL to put after
			-- decimal point when writing `d' as a string
		local
			k: INTEGER
			found: BOOLEAN
			val: DOUBLE
		do
			val := d.abs
			Result := 1
			from
				k := 0
			until
				k > 20 or found
			loop
				if val < 10.0 ^ k - 0.5 * 10.0 ^ (k - 7) then
					Result := 7 - k
					found := True
				end
				k := k + 1
			end
			if Result < 1 or d = 0.0 then
				Result := 1
			end
		end

	boolean_to_string (b: BOOLEAN): STRING
			-- `b' as a string in format ICL expects
		do
			Result := b.out.as_upper
		end

	string_to_integer (s: STRING): INTEGER
			-- `s' as an integer
		do
			if s.is_integer then
				Result := s.to_integer
			else
				report_read_error ("value " + su.double_quoted (s) + " is not an integer")
			end
		end

	string_to_boolean (s: STRING): BOOLEAN
			-- `s' as a boolean
		do
			if s.is_boolean then
				Result := s.to_boolean
			else
				report_read_error ("value " + su.double_quoted (s) + " is not a boolean")
			end
		end

	string_to_point (s: STRING): CONST_POINT
			-- `s' as a point
		local
			pos: INTEGER
			xstr, ystr: STRING
		do
			pos := s.index_of ('#', 1)
			if pos = 0 then
				report_read_error ("value " + su.double_quoted (s) + " is not a valid ICL point")
			end
			xstr := s.substring (1, pos - 1)
			ystr := s.substring (pos + 1, s.count)
			xstr.right_adjust
			ystr.left_adjust
			if not xstr.is_double then
				report_read_error ("value " + xstr + " is not a valid real number")
			end
			if not ystr.is_double then
				report_read_error ("value " + ystr + " is not a valid real number")
			end
			create Result.make_xy (xstr.to_double, ystr.to_double)
		end

	quoted (s: STRING): STRING
			-- `s' as a string in format ICL expects
		do
			Result := s.twin
			Result.replace_substring_all ("%'", "%'%'")
			Result := su.single_quoted (Result)
		end

	unquoted (s: STRING): STRING
			-- `s' as a string with ICL leading, trailing and
			-- internal doubled quotes removed
		do
			Result := s.twin
			if not Result.is_empty and then Result.item (1) = '%'' then
				Result.remove_head (1)
			end
			if not Result.is_empty and then Result.item (Result.count) = '%'' then
				Result.remove_tail (1)
			end
			Result.replace_substring_all ("%'%'", "%'")
		end

	read_line
			-- Read next line from `file' and put in `line'.
			-- Remove leading and trailing white space
		do
			file.read_line
			if not file.end_of_file then
				line_number := line_number + 1
				line := file.last_string
				line.left_adjust
				line.right_adjust
				if not line.is_empty and then line.item (1) = '[' then
					line.remove_head (1)
					line.left_adjust
				end
				
				from
				until
					line.is_empty or else Ignored_trailing_chars.index_of (line.item (line.count), 1) = 0
				loop
					line.remove_tail (1)
				end
			end
		end

	process_line
			-- Process `line', which was read from packer box
			-- file and is not a delimiter line
		do
			if line.is_empty then
				-- Skip it
			elseif is_pb_start_delimiter or is_file_end_delimiter then
				-- Add `box' if not Void
				if box /= Void then
					extend (box)
					box := Void
				end
				if is_pb_start_delimiter then
					-- Start of new packer box
					create box
					in_boxes := True
				end
				done := is_file_end_delimiter
				in_positions := False
			elseif in_boxes then
				-- Process line
				if is_positions_start_delimiter then
					in_positions := True
				else
					process_non_delimiter_line
				end
			end
		end

	process_non_delimiter_line
		local
			kw, val: STRING
			pos, end_pos: INTEGER
			loc: PACKER_DIE_LOCATION
		do
			if not in_positions then
				-- Reading non-positions packer box field
				pos := line.index_of (':', 1)
				if pos = 0 then
					report_read_error ("no %":%" found in line: " + line)
				end
				kw := line.substring (1, pos - 1)
				val := line.substring (pos + 1, line.count)
				val.left_adjust
				if kw.is_equal ("project_type") then
					val.to_upper
					box.set_project_type (val)
				elseif kw.is_equal ("name") then
					box.set_name (unquoted (val))
				elseif kw.is_equal ("design_name") then
					box.set_design_name (unquoted (val))
				elseif kw.is_equal ("cust_account_display_name") then
					box.set_cust_account_display_name (unquoted (val))
				elseif kw.is_equal ("cust_account_no") then
					box.set_cust_account_no (string_to_integer (val))
				elseif kw.is_equal ("cust_account_type") then
					box.set_cust_account_type (unquoted (val))
				elseif kw.is_equal ("cust_account_sponsor") then
					box.set_cust_account_sponsor (unquoted (val))
				elseif kw.is_equal ("external") then
					box.set_is_external (string_to_boolean (val))
				elseif kw.is_equal ("m_fab_no") then
					box.set_m_fab_no (string_to_integer (val))
				elseif kw.is_equal ("priority") then
					box.set_priority (string_to_integer (val))
				elseif kw.is_equal ("min_die_size") then
					box.set_min_die_size (string_to_point (val))
				elseif kw.is_equal ("max_die_size") then
					box.set_max_die_size (string_to_point (val))
				elseif kw.is_equal ("not_rotatable") then
					box.set_not_rotatable (string_to_boolean (val))
				elseif kw.is_equal ("not_movable") then
					box.set_not_movable (string_to_boolean (val))
				elseif kw.is_equal ("tiny_chip") then
					box.set_is_tiny_chip (string_to_boolean (val))
				elseif kw.is_equal ("requested_number") then
					box.set_requested_number (string_to_integer (val))
				elseif kw.is_equal ("requested_display_name") then
					box.set_requested_display_name (unquoted (val))
				elseif kw.is_equal ("process_name") then
					box.set_process_name (unquoted (val))
				elseif kw.is_equal ("wafer_etest_id") then
					box.set_wafer_etest_id (unquoted (val))
				elseif kw.is_equal ("display_name") then
					box.set_display_name (unquoted (val))
				elseif kw.is_equal ("wafer_good_sites") then
					box.set_wafer_good_sites (string_to_integer (val))
				elseif kw.is_equal ("rotated") then
					box.set_is_rotated (string_to_boolean (val))
				elseif kw.is_equal ("die_size") then
					box.set_die_size (string_to_point (val))
				elseif kw.is_equal ("die_id") then
					box.set_die_id (unquoted (val))
				else
					report_read_error ("unknown packer_box field name " + su.double_quoted (kw))
				end
			else
				-- Reading positions - each line has one location
					
				create loc
				
				from
				until
					line.is_empty
				loop
					pos := line.index_of (':', 1)
					if pos = 0 then
						report_read_error ("no %":%" found in line: " + line)
					end
					kw := line.substring (1, pos - 1)
					line.remove_head (pos)
					pos := line.index_of (':', 1)
					if pos /= 0 then
						end_pos := su.last_white_position (line, pos)
						val := line.substring (1, end_pos)
						line.remove_head (end_pos)
					else
						val := line.twin
						line.wipe_out
					end
					val.left_adjust
					val.right_adjust
					
					if kw.is_equal ("field_name") then
						loc.set_field_name (unquoted (val))
					elseif kw.is_equal ("location") then
						loc.set_location (string_to_point (val))
					elseif kw.is_equal ("mask_site_no") then
						loc.set_mask_site_no (string_to_integer (val))
					elseif kw.is_equal ("section_number") then
						loc.set_section_number (string_to_integer (val))
					elseif kw.is_equal ("new") then
						loc.set_is_new (string_to_boolean (val))
					elseif kw.is_equal ("chip_number") then
						loc.set_chip_number (string_to_integer (val))
					else
						report_read_error ("unknown die_location field name " + su.double_quoted (kw))
					end
				end
				
				if box.positions = Void then
					box.set_positions (create {ARRAYED_LIST [PACKER_DIE_LOCATION]}.make (5))
				end
				box.positions.extend (loc)
			end
		end

	report_read_error (s: STRING)
		do
			handle_hard_error (s + " at line " + line_number.out + " of packer box file " + file.name)
		end

	is_pb_start_delimiter: BOOLEAN
		do
			Result := line.is_equal (pb_start_delimiter)
		end

	is_file_end_delimiter: BOOLEAN
		do
			Result := line.is_equal (pb_file_end_delimiter)
		end

	is_positions_start_delimiter: BOOLEAN
		do
			Result := line.is_equal (positions_start_delimiter)
		end

	icl_minimum_die_size (proj: PACKING_PROJECTS_ROW): CONST_POINT
			-- Minimum die size for project, when using old
			-- ICL code on VMS to pack
		do
			Result := wafer_techns_row.die_size_with_overhead (packing.effective_overhead_style (proj.overhead_style), packing.other_border, proj.design_size)
		end
	
	pb_start_delimiter: STRING = "pbs ::= $$"
			-- Signals start of next packer box

	positions_start_delimiter: STRING = "positions:"
			-- Signals start of positions in packer box

	pb_file_end_delimiter: STRING = "pbs := refresh(pbs)"
			-- Signals end of packer box file

	Ignored_trailing_chars: STRING = " %T{]};"
			-- Characters to be removed from end of each line

	file: LARGE_PLAIN_TEXT_FILE
			-- Used for reading and writing packer box files

	box: PACKER_BOX
			-- Current packer box being filled in while
			-- reading packer box file
	
	line: STRING
			-- Current packer box file line being considered
	
	done: BOOLEAN
			-- Are we done reading packer box file?

	in_boxes: BOOLEAN
			-- Have we encountered the fist packer box yet?

	in_positions: BOOLEAN
			-- Are we in the "positions" field of packer box?

	line_number: INTEGER
			-- Number of current line

	packing: PACKINGS_ROW
			-- Packing for which `Current' is being generated

	wafer_techns_row: WAFER_TECHNS_ROW
			-- Wafer techn for which `Current' is being
			-- generated

	runs_row: RUNS_ROW
			-- Run for which `Current' is being generated

	cust_die_table: HASH_TABLE [PACKER_BOX, INTEGER]
			-- Packer boxes of type cust_die indexed by m_fab_no

	test_vehicle_table: HASH_TABLE [PACKER_BOX, INTEGER]
			-- Packer boxes of type test_vehicle indexed by
			-- design_no

	vendor_geometry_table: HASH_TABLE [PACKER_BOX, STRINg]
			-- Packer boxes of type vendor_geometry or
			-- vendor_ghost indexed by geomtry_id

	family_processes_table: HASH_TABLE [WAFER_PROCESSES_ROW, STRING]
			-- Processes for packing's wafer techn family,
			-- indexed by process name

end

