
indexing
	
	description: "Object that can create a detached PACKINGS row %
		%with all dependent rows filled in, based on data from %
		%official database tables"
	author: "David Hollenberg"
	date: "February 27, 2007"

class PACKING_CREATOR
inherit
	ANY
	CONDITION_UTILITIES
		export
			{NONE} all
		end
	ITEM_COUNT_RANGES_ACCESS
		export
			{NONE} all
		end
	DB_OPTIONS_ACCESS
		export
			{NONE} all
		end

create
	make

feature {NONE} -- Creation

	make (a_prod_id: like production_id a_packing_id: like packing_id a_runs_row: like run)
			-- Create packing for production `a_prod_id' 
			-- identified by `a_packing_id', using `a_runs_row'
		do
			production_id := a_prod_id
			packing_id := a_packing_id
			run := a_runs_row
		end

feature -- Properties

	run: RUNS_ROW
			-- Run for which packing is to be synthesized

	production_id: STRING
			-- Production for which packing is to be synthesized

	packing_id: STRING
			-- Identifying name for packing

	dies: LIST [DIES_ROW]
			-- List of dies

	die_locations: LIST [DIE_LOCATIONS_ROW]
			-- List of die locations

	cut_patterns: LIST [RUN_CUT_PATTERNS_ROW]
			-- List of cut patterns

	cut_locations: LIST [CUT_LOCATIONS_ROW]
			-- List of cut locations

	cut_die_sites: LIST [CUT_DIE_SITES_ROW]
			-- List of cut die sites

	packing: RC_PACKING
			-- Packing that looks like it was fetched from
			-- the PACKINGS table, with no current version

	packings_row: PACKINGS_MROW
			-- Detached row for packing synthesized from
			-- official database tables

	packing_version: PACKING_VERSIONS_MROW
			-- Detached row for packing version with all 
			-- dependent rows filled in from official 
			-- (non-packing) database tables

feature -- Modification

	set_run (a_run: like run)
		require
			a_run_not_void: a_run /= Void
		do
			run := a_run
		ensure
			run_set: run = a_run
		end

	set_dies (a_list: like dies)
		require
			a_list_not_void: a_list /= Void
		do
			dies := a_list
		ensure
			dies_set: dies = a_list
		end

	set_die_locations (a_list: like die_locations)
		require
			a_list_not_void: a_list /= Void
		do
			die_locations := a_list
		ensure
			die_locations_set: die_locations = a_list
		end

	set_cut_patterns (a_list: like cut_patterns)
		require
			a_list_not_void: a_list /= Void
		do
			cut_patterns := a_list
		ensure
			cut_patterns_set: cut_patterns = a_list
		end

	set_cut_locations (a_list: like cut_locations)
		require
			a_list_not_void: a_list /= Void
		do
			cut_locations := a_list
		ensure
			cut_locations_set: cut_locations = a_list
		end

	set_cut_die_sites (a_list: like cut_die_sites)
		require
			a_list_not_void: a_list /= Void
		do
			cut_die_sites := a_list
		ensure
			cut_die_sites_set: cut_die_sites = a_list
		end

feature -- Operations

	synthesize
			-- Synthesize detached packing versions row 
			-- with all dependent rows filled in and leave
			-- in `packing_version'.  Also set `packing'
			-- and `packings_row'
		require
			dies_not_void: dies /= Void
			die_locations_not_void: die_locations /= Void
			cut_patterns_not_void: cut_patterns /= Void
			cut_locations_not_void: cut_locations /= Void
			cut_die_sites_not_void: cut_die_sites /= Void
		do
			connection := run.connection
			create packing
			
			-- Create `packings_row' and put in packing

			create packings_row.make_detached (connection)
			packings_row.set_packing_id (packing_id)
			packings_row.initialize_from_run (run)
			
			packing.set_packings_row (packings_row)
			packing.set_production_id (production_id)

			fetch_data
			synthesize_packing_version

			packing.set_current_version (packing_version)
		end

feature {NONE} -- Implementation

	fetch_data
			-- Fetch required data
		local
			die: DIES_ROW
			cut_pat: RUN_CUT_PATTERNS_ROW
		do
			create die.make_template
			dies := die.dies_for_run (run.run_id, connection, expected_count.At_least_one)
			create cut_pat.make_template
			cut_patterns := cut_pat.rows_for_production (production_id, connection, expected_count.At_least_zero)
		end

	synthesize_packing_version
			-- Synthesize detached row `packing_version' and
			-- all its dependent rows from fetched data
		do
			synthesize_version_row
			synthesize_and_add_dies
			synthesize_and_add_cut_patterns
		end

	synthesize_version_row
			-- Synthesize detached row `packing_version'
		local
			pv: like packing_version
		do
			create pv.make_detached (connection)
			pv.set_packing_id (packing_id)
			pv.set_version_no (0)
			pv.set_date_created (connection.transaction_date)
			pv.set_creator (connection.user_name)
			pv.set_packing_name (packing_id)
			pv.set_version_id ("production " + production_id)
			pv.set_packing (packings_row)
			packing_version := pv
		end

	synthesize_and_add_dies
			-- Synthesize detached row for each die and
			-- append to list of dies for `packing_version'
		do
			packing_version.set_selection_date
			packing.setup_for_fast_project_access
			create {ARRAYED_LIST [PACKING_DIES_MROW]} packing_dies.make (dies.count)
			create wafer_processes.make (11)
			create packing_die_locations.make (200)
			from
				dies.start
			until
				dies.after
			loop
				add_packing_die (dies.item)
				dies.forth
			end
			packing_version.set_dies (packing_dies)
		end

	add_packing_die (a_die: DIES_ROW)
			-- Synthesize a packing die for `a_die' and append
			-- to `packing_dies'
		local
			dt, process_name: STRING
			type: DIE_TYPE_VALUES
			new_die: PACKING_DIES_MROW
			ifi: IC_FAB_INSTANCES_ROW
			wafer_process: WAFER_PROCESSES_ROW
		do
			dt := a_die.die_type
			if dt.is_equal (type.Cust_die) then
				ifi := a_die.ic_fab_instance
				new_die := packing_version.new_cust_die (ifi.m_fab_no)
				process_name := ifi.ordered_wafer_process
				wafer_processes.search (process_name)
				if wafer_processes.found then
					wafer_process := wafer_processes.found_item
				else
					create wafer_process.make_by_key (connection, option.Old_row, process_name)
					wafer_processes.put (wafer_process, process_name)
				end
				new_die.set_wafer_etest_name (wafer_process.etest_name.wafer_etest_name)
			elseif dt.is_equal (type.Test_vehicle) then
				new_die := packing_version.new_test_vehicle_die (a_die.test_vehicle_design_no)
			elseif dt.is_equal (type.Vendor_geometry) or dt.is_equal (type.Vendor_ghost) then
			elseif dt.is_equal (type.Scribe) then
				new_die := packing_version.new_scribe_with_orientation (a_die.scribe_orientation)
			elseif dt.is_equal (type.Filler) then
				new_die := packing_version.new_filler_with_name (a_die.filler_name)
			elseif dt.is_equal (type.Macrocell) then
				-- Ignore
			elseif dt.is_equal (type.Placard) or dt.is_equal (type.Start_die) then
				handle_hard_error ("die " + a_die.die_id + " has invalid die type " + dt + " for reticle packer")
			else
				handle_hard_error ("die " + a_die.die_id + " has unknown die type " + dt)
			end
			if new_die /= Void then
				new_die.set_die_id (a_die.die_id)
				new_die.set_die_size_printed_x (a_die.die_size_printed_x)
				new_die.set_die_size_printed_y (a_die.die_size_printed_y)
				new_die.set_die_wafer_rotation (a_die.die_wafer_rotation)
				synthesize_and_add_die_locations (a_die, new_die)
				packing_dies.extend (new_die)
			end
		end

	synthesize_and_add_die_locations (a_die: DIES_ROW new_die: PACKING_DIES_MROW)
			-- Synthesize detached row for each die
			-- location for `a_die' and add to locations for
			-- `new_die'
		local
			locations: LIST [DIE_LOCATIONS_ROW]
			dl : DIE_LOCATIONS_ROW
			pdl: PACKING_DIE_LOCATIONS_MROW
			msn: INTEGER
		do
			locations := a_die.locations
			from
				locations.start
			until
				locations.after
			loop
				dl := locations.item
				new_die.add_die_location (dl.field_name, dl.location_x, dl.location_y)
				pdl := new_die.last_die_location
				msn := dl.mask_site_no
				pdl.set_mask_site_no (msn)
				packing_die_locations.put (pdl, msn)
				locations.forth
			end
		end

	synthesize_and_add_cut_patterns
			-- Synthesize detached row for each cut pattern
			-- and append to list of cut patterns for `packing_version'
		do
			create {ARRAYED_LIST [PACKING_CUT_PATTERNS_MROW]} packing_cut_patterns.make (cut_patterns.count)
			create {ARRAYED_LIST [PACKING_WAFER_PROCESSES_MROW]} packing_wafer_processes.make (0)
			packing_version.set_cut_patterns (packing_cut_patterns)
			packing_version.set_wafer_processes (packing_wafer_processes)
			from
				cut_patterns.start
			until
				cut_patterns.after
			loop
				add_packing_cut_pattern (cut_patterns.item)
				cut_patterns.forth
			end
		end

	add_packing_cut_pattern (a_cut_pattern: RUN_CUT_PATTERNS_ROW)
			-- Synthesize a packing cut pattern for 
			-- `a_cut_pattern' and add to cut patterns for
			-- `packing_version
		local
			new_pattern: PACKING_CUT_PATTERNS_MROW
		do
			packing_version.add_cut_pattern (a_cut_pattern.cut_pattern_no, a_cut_pattern.field_name, a_cut_pattern.packaging_lot.process.etest_name.wafer_etest_name, a_cut_pattern.quantity_wafers_needed, True)
			new_pattern := packing_version.cut_patterns.i_th (packing_version.last_cut_pattern_position)
			new_pattern.set_packaging_lot_no (a_cut_pattern.packaging_lot_no)
			new_pattern.set_is_completed (True)
			add_packing_cut_locations (new_pattern, a_cut_pattern.cuts)
			add_packing_cut_die_sites (new_pattern, a_cut_pattern.cut_die_sites)
		end

	add_packing_cut_locations (a_cut_pattern: PACKING_CUT_PATTERNS_MROW cuts: LIST [CUT_LOCATIONS_ROW])
			-- Add to `a_cut_pattern' one cut for each element of
			-- `cuts'
		do
			a_cut_pattern.set_cut_locations (create {ARRAYED_LIST [PACKING_CUT_LOCATIONS_MROW]}.make (cuts.count))
			from
				cuts.start
			until
				cuts.after
			loop
				add_packing_cut_location (a_cut_pattern, cuts.item)
				cuts.forth
			end
		end

	add_packing_cut_location (a_cut_pattern: PACKING_CUT_PATTERNS_MROW cut: CUT_LOCATIONS_ROW)
			-- Add to `a_cut_pattern' a cut corresponding to `cut'
		local
			new_cut: PACKING_CUT_LOCATIONS_MROW
			orientation: ORIENTATION_VALUES
		do
			if cut.orientation.is_equal (orientation.Horizontal) then
				new_cut := a_cut_pattern.new_horizontal_cut (cut.location)
			else
				new_cut := a_cut_pattern.new_vertical_cut (cut.location)
			end
			a_cut_pattern.add_cut (new_cut)
		end

	add_packing_cut_die_sites (a_cut_pattern: PACKING_CUT_PATTERNS_MROW sites: LIST [CUT_DIE_SITES_ROW])
			-- Add to `a_cut_pattern' one cut die site for each 
			-- element of `sites'
		do
			a_cut_pattern.set_cut_die_sites (create {ARRAYED_LIST [PACKING_CUT_DIE_SITES_MROW]}.make (sites.count))
			from
				sites.start
			until
				sites.after
			loop
				add_packing_cut_die_site (a_cut_pattern, sites.item)
				sites.forth
			end
		end

	add_packing_cut_die_site (a_cut_pattern: PACKING_CUT_PATTERNS_MROW site: CUT_DIE_SITES_ROW)
			-- Add to `a_cut_pattern' a cut die site corresponding
			-- to `site'
		local
			pdl: PACKING_DIE_LOCATIONS_MROW
			msn: INTEGER
			mbb: CONST_RECTANGLE
		do
			msn := site.mask_site_no
			packing_die_locations.search (msn)
			if packing_die_locations.found then
				pdl := packing_die_locations.found_item
				create mbb.make_by_size_and_center (site.die_size_cut_x, site.die_size_cut_y, site.location_x, site.location_y)
				a_cut_pattern.isolate (pdl, mbb)
			else
				handle_hard_error ("die location with mask site number " + msn.out + " not found")
			end
		end

	connection: DB_CONNECTION
			-- Database connection

	packing_dies: LIST [PACKING_DIES_MROW]
			-- Dies to be added to packing version

	packing_cut_patterns: LIST [PACKING_CUT_PATTERNS_MROW]
			-- Cut patterns to be added to packing version

	packing_wafer_processes: LIST [PACKING_WAFER_PROCESSES_MROW]
			-- Wafer processes to be added to packing version

	wafer_processes: HASH_TABLE [WAFER_PROCESSES_ROW, STRING]
			-- Table of wafer processes ordered by a customer
			-- project so far, indexed by wafer process name

	packing_die_locations: HASH_TABLE [PACKING_DIE_LOCATIONS_MROW, INTEGER]
			-- Table of packing die locations, indexed by
			-- mask site number

end
