
indexing

	author: "David Hollenberg"

class PACKING_PROJECTS_ROW
inherit
	PACKING_PROJECTS_RAW
	PACKING_QUERIES
	DB_HASHABLE_ROW
		undefine
			initialize_new, initialize_old, fill_in
		end
	ROW_COUNT_CHECK_ACCESS
		export
			{NONE} all
		end
	DB_OPTIONS_ACCESS
		export
			{NONE} all
		end
	ITEM_COUNT_RANGES_ACCESS
		export
			{NONE} all
		end
	NULL_VALUE_CHECK_ACCESS
		export
			{NONE} all
		end
	WAFER_OPTION_UTILITIES_INTERFACE
		export
			{NONE} compute_processes_for_options
		end

create
	make_template, make_null, make_by_key, make_detached

feature -- Queries

	rows_for_project_version (a_packing_id: STRING a_project_no, a_proj_version: INTEGER conn: DB_CONNECTION range: ITEM_COUNT_RANGE): LIST [like Current] is
			-- Projects for `a_packing_id', `a_project_no' and
			-- `a_proj_version'
		require
			a_packing_id_not_void: a_packing_id /= Void
			connection_exists: conn /= Void
			range_exists: range /= Void
		local
			b: DB_BIND_SII
		do
			create b.make (a_packing_id, a_project_no, a_proj_version)
			Result := all_rows_filtered (conn, "packing_id = :1 and project_no = :2 and version_no = :3", b)
			check_row_count (Result, range)
		ensure
			result_exists: Result /= Void
		end

	highest_version_rows_for_packing (a_packing_id: STRING conn: DB_CONNECTION range: ITEM_COUNT_RANGE): LIST [like Current] is
			-- All projects for `a_packing_id' (highest numbered 
			-- version only).  Includes both currently selected
			-- projects and projects that have been deselected
		require
			a_packing_id_not_void: a_packing_id /= Void
			connection_exists: conn /= Void
			range_exists: range /= Void
		local
			b: DB_BIND_S
		do
			create b.make (a_packing_id)
			Result := all_rows_filtered (conn, "packing_id = :1 and version_no = (select max(version_no) from packing_projects where packing_id = pckp.packing_id and project_no = pckp.project_no)", b)
			check_row_count (Result, range)
		ensure
			result_exists: Result /= Void
		end

	highest_used_version: INTEGER is
			-- Highest version number previously used for
			-- `packing_id' and `project_no'
		require
			packing_id_not_null: not is_null_packing_id
			project_no_not_null: not is_null_project_no
		local
			b: DB_BIND_SI
		do
			create b.make (packing_id, project_no)
			Result := any_connection.selected_integer ("max (version_no)", table_name, "packing_id = :1 and project_no = :2", b)
		end

feature -- Properties

	hash_code: INTEGER is
		local
			s: STRING
		do
			create s.make (packing_id.count + 10)
			s.append (packing_id)
			s.extend ('%U')
			s.append_integer (project_no)
			Result := s.hash_code
		end
	
	packagings: LIST [like anchored_packaging] is
			-- Packagings for `Current'
		do
			if cached_packagings = Void then
				fetch_packagings
			end
			Result := cached_packagings
		end

	is_cached_packagings: BOOLEAN is
			-- Are packagings currently cached?
		do
			Result := cached_packagings /= Void
		end

	packaged_packages: LIST [PACKAGES_ROW] is
			-- Packages in which `Current' will be packaged
		do
			if cached_packaged_packages = Void then
				fetch_packaged_packages
			end
			Result := cached_packaged_packages
		end

	is_cached_packaged_packages: BOOLEAN is
			-- Are packages in which `Current' will be
			-- packaged currently cached?
		do
			Result := cached_packaged_packages /= Void
		end

	packaging_description: STRING is
			-- Description of packages in which `Current' will 
			-- be packaged
		local
			list: like packaged_packages
			pkg: PACKAGES_ROW
		do
			list := packaged_packages
			create Result.make (list.count * 40)
			from
				list.start
			until
				list.after
			loop
				pkg := list.item
				Result.append (pkg.package_id)
				if list.count > 1 then
					Result.append ("   ")
				else
					Result.extend ('%N')
				end
				if pkg.leadframe then
					Result.append ("PLASTIC")
				else
					Result.append ("CERAMIC")
				end
				list.forth
				if list.count > 1 and not list.after then
					Result.extend ('%N')
				end
			end
		end

	has_ceramic_packaging: BOOLEAN is
			-- Will project be packaged in at least one ceramic
			-- package?
		local
			list: like packaged_packages
			pkg: PACKAGES_ROW
		do
			list := packaged_packages
			from
				list.start
			until
				list.after
			loop
				pkg := list.item
				if not pkg.leadframe then
					Result := True
				end
				list.forth
			end
		end

	has_plastic_packaging: BOOLEAN is
			-- Will project be packaged in at least one plastic
			-- package?
		local
			list: like packaged_packages
			pkg: PACKAGES_ROW
		do
			list := packaged_packages
			from
				list.start
			until
				list.after
			loop
				pkg := list.item
				if pkg.leadframe then
					Result := True
				end
				list.forth
			end
		end

	options: LIST [like anchored_option] is
			-- Options for `Current'
		do
			if cached_options = Void then
				fetch_options
			end
			Result := cached_options
		end

	is_cached_options: BOOLEAN is
			-- Are options currently cached?
		do
			Result := cached_options /= Void
		end

	is_test_vehicle: BOOLEAN is
			-- Is `Current' a test vehicle?
		local
			type: IC_FAB_INSTANCE_TYPE_VALUES
		do
			Result := ic_fab_instance_type.is_equal (type.Test_vehicle)
		end

	is_customer_project: BOOLEAN is
			-- Is `Current' a customer project?
		local
			type: IC_FAB_INSTANCE_TYPE_VALUES
		do
			Result := ic_fab_instance_type.is_equal (type.Customer_fab)
		end

	is_customer_project_or_test_vehicle: BOOLEAN is
			-- Is `Current' a customer project or test vehicle?
		do
			Result := is_customer_project or is_test_vehicle
		end

	is_neither_customer_project_nor_test_vehicle: BOOLEAN is
			-- Is `Current' neither a customer project nor 
			-- a test vehicle?
		do
			Result := not is_customer_project and not is_test_vehicle
		end

	is_selected: BOOLEAN
			-- Is project currently selected?

	is_currently_locked: BOOLEAN is
			-- Is project currently locked against updates
			-- by customer?
		local
			f: like fab_instance
		do
			if not is_test_vehicle then
				f := fab_instance
				if f.is_old then
					-- Fab instance still there (could
					-- disappear for a refab that is
					-- cancelled)
					Result := f.is_frozen
				else
					Result := False
				end
			else
				Result := False
			end
		end

	is_unlocked_customer_project: BOOLEAN is
			-- Is `Current' a customer project that is not
			-- currently locked against updates?
		do
			Result := is_customer_project and then not is_currently_locked
		end

	is_locked_customer_project: BOOLEAN is
			-- Is `Current' a customer project that is
			-- currently locked against updates?
		do
			Result := is_customer_project and then is_currently_locked
		end

	is_locked_assigned_customer_project (a_run_id: STRING): BOOLEAN is
			-- Is `Current' a customer project that is
			-- locked and assigned to
			-- `a_run_id'?
		do
			Result := is_locked_customer_project and then is_assigned_to_run (a_run_id)
		end

	is_locked_unassigned_customer_project (a_run_id: STRING): BOOLEAN is
			-- Is `Current' a customer project that is
			-- locked and not currently assigned to
			-- `a_run_id' (is either not assigned to any
			-- run or is assigned to a run other than
			-- `a_run_id')?
		do
			Result := is_locked_customer_project and then not is_assigned_to_run (a_run_id)
		end

	is_unlocked_or_unassigned_customer_project (a_run_id: STRING): BOOLEAN is
			-- Is `Current' a customer project that is
			-- either unlocked or is locked but not
			-- currently assigned to `a_run_id' (i.e,, is
			-- either not assigned to any run or is
			-- assigned to a run other than `a_run_id')?
		do
			Result := is_customer_project and then (not is_currently_locked or not is_assigned_to_run (a_run_id))
		end

	is_assigned_to_run (a_run_id: STRING): BOOLEAN is
			-- Is `Current' assigned to run identified by
			-- `a_run_id'?
		require
			has_fab_no: not is_null_m_fab_no
		do
			Result := is_assigned and then not fab_instance.is_null_run_id and then fab_instance.run_id.is_equal (a_run_id)
		end

	assigned_run_id: STRING is
			-- Run ID of run to which `Current' is assigned
			-- (empty if none)
		require
			has_fab_no: not is_null_m_fab_no
		do
			Result := fab_instance.run_id
		end

	is_assigned: BOOLEAN is
			-- Is project currently assigned to a run?
		local
			f: like fab_instance
		do
			if not is_test_vehicle then
				f := fab_instance
				if f.is_old then
					-- Fab instance still there (could
					-- disappear for a refab that is
					-- cancelled)
					Result := f.is_on_run
				else
					Result := False
				end
			else
				Result := False
			end
		end

	locked_and_assigned_status: STRING is
			-- Text string describing "locked" and "assigned"
			-- status of `Current'
		require
			is_customer_project: is_customer_project
		local
			lck, assn: BOOLEAN
			run: STRING
		do
			lck := is_currently_locked
			assn := is_assigned
			run := assigned_run_id
			if lck and assn then
				Result := "locked and assigned to run " + run
			elseif lck and not assn then
				Result := "locked but not assigned to any run"
			elseif not lck and assn then
				Result := "not locked but assigned to run " + run
			else
				Result := "not locked and not assigned to any run"
			end
		end

	fab_instance: like anchored_fab_instance is
			-- Fab instance
		require
			has_fab_no: not is_null_m_fab_no
		do
			if not is_cached_fab_instance then
				fetch_fab_instance
			end
			Result := cached_fab_instance
		ensure
			result_not_void: Result /= Void
		end

	is_cached_fab_instance: BOOLEAN is
			-- Is `fab_instance' cached?
		do
			Result := cached_fab_instance /= Void
		end

	cust_ic_design: like anchored_cust_ic_design is
			-- Cust IC design
		require
			has_design_no: not is_null_design_no
		do
			if not is_cached_cust_ic_design then
				fetch_cust_ic_design
			end
			Result := cached_cust_ic_design
		ensure
			result_not_void: Result /= Void
		end

	is_cached_cust_ic_design: BOOLEAN is
			-- Is `cust_ic_design' cached?
		do
			Result := cached_cust_ic_design /= Void
		end

	cust_account: like anchored_cust_account is
			-- Customer account
		do
			if not is_cached_cust_account then
				fetch_cust_account
			end
			Result := cached_cust_account
		ensure
			result_not_void: Result /= Void
		end

	is_cached_cust_account: BOOLEAN is
			-- Is `cust_account' cached?
		do
			Result := cached_cust_account /= Void
		end

	test_vehicle: like anchored_test_vehicle is
			-- Test vehicle row
		require
			is_test_vehicle: is_test_vehicle
		do
			if not is_cached_test_vehicle then
				fetch_test_vehicle
			end
			Result := cached_test_vehicle
		ensure
			result_not_void: Result /= Void
		end

	is_cached_test_vehicle: BOOLEAN is
			-- Is `test_vehicle' cached?
		do
			Result := cached_test_vehicle /= Void
		end

	packing: like anchored_packing is
			-- Packing to which `Current' belongs
		do
			if cached_packing = Void then
				fetch_packing
			end
			Result := cached_packing
		end

	is_cached_packing: BOOLEAN is
			-- Is packing currently cached?
		do
			Result := cached_packing /= Void
		end

	non_null_design_size_x: like design_size_x is
			-- Value of `design_size_x', with error check 
			-- for nullness
		do
			check_not_null ("design_size_x", is_null_design_size_x, Current)
			Result := design_size_x
		end

	non_null_design_size_y: like design_size_y is
			-- Value of `design_size_y', with error check 
			-- for nullness
		do
			check_not_null ("design_size_y", is_null_design_size_y, Current)
			Result := design_size_y
		end

	design_size: CONST_POINT is
			-- Design size of `Current'
		do
			create Result.make_xy (non_null_design_size_x, non_null_design_size_y)
		end

	max_die_size: CONST_POINT is
			-- Maximum die size, as constrained by package
			-- cavity sizes and user-specified limits.
			-- Void indicates no maximum
		local
			max_size: CONST_POINT
			max_die_x, max_die_y: DOUBLE
			tv: like test_vehicle
			has_max: BOOLEAN
			list: LIST [like anchored_packaging]
		do
			if is_customer_project then
				
				-- Compute size limit based on packagings

				from
					max_die_x := 1000000
					max_die_y := 1000000
					list := packagings
					list.start
				until
					list.after
				loop
					max_size := list.item.max_die_size
					if max_size /= Void then
						has_max := True
						max_die_x := max_die_x.min (max_size.x)
						max_die_y := max_die_y.min (max_size.y)
					end
					list.forth
				end
				
				-- Factor in user-specified limits

				if not is_null_max_allowed_die_size_x then
					max_die_x := max_die_x.min (max_allowed_die_size_x)
					has_max := True
				end
				if not is_null_max_allowed_die_size_y then
					max_die_y := max_die_y.min (max_allowed_die_size_y)
					has_max := True
				end
			else -- Test vehicle
				tv := test_vehicle
				if not tv.is_null_max_die_size_x then
					max_die_x := tv.max_die_size_x
					has_max := True
				end
				if not tv.is_null_max_die_size_y then
					max_die_y := tv.max_die_size_y
					has_max := True
				end
			end
			if has_max then
				create Result.make_xy (max_die_x, max_die_y)
			end
		end

	is_oversize_die_always_allowed: BOOLEAN
			-- Is a die that is bigger than `max_die_size'
			-- allowed unconditionally?
		do
			Result := not is_null_allow_oversize_die_on_run and allow_oversize_die_on_run
		end

	is_oversize_die_possibly_allowed: BOOLEAN
			-- Is a die that is bigger than `max_die_size'
			-- allowed but only under specific conditions?
		do
			Result := is_null_allow_oversize_die_on_run
		end

	oversize_die_inclusion_conditions: STRING
			-- Description of conditions under which 
			-- die that is bigger than `max_die_size'
			-- allowed
		do
			result := oversize_die_incl_conditions
		end

	assigned_process: STRING
			-- Name of compatible process to which `Current'
			-- was assigned by last call to `assign_process'

	reason_unselectable: STRING
			-- Reason `Current' is not selectable

	compatible_wafer_etest_ids: LIST [STRING]
			-- Distinct wafer etest IDs for `compatible_processes'

	compatible_wafer_etest_names: LIST [WAFER_ETEST_NAMES_ROW]
			-- Distinct wafer etest names rows for 
			-- `compatible_processes'

	assigned_wafer_etest_name: WAFER_ETEST_NAMES_ROW
			-- Assigned wafer etest name row or Void if no
			-- wafer etest ID assigned yet

	assigned_wafer_etest_id: STRING is
			-- Wafer etest ID assigned to cust die or empty
			-- string if none
		require
			is_customer_project: is_customer_project
		local
			row: like assigned_wafer_etest_name
		do
			row := assigned_wafer_etest_name
			if row /= Void then
				Result := row.wafer_etest_id
			else
				Result := ""
			end
		end

	compatible_processes_with_assigned_etest_name: like compatible_processes
			-- Processes in `compatible_processes' that
			-- have the wafer etest name indicated by
			-- `assigned_wafer_etest_name' (no restriction if
			-- `assigned_wafer_etest_name' is Void)

	test_vehicle_header: STRING is
			-- Header lines for `Current' for test vehicle
			-- notification message
		do
			create Result.make (200)
			Result.append ("    Design  Design Name                     Rotation   Project Type%N")
			Result.append ("    ------  ------------------------------  ---------  ------------%N%N")
		end

	test_vehicle_description (rot: INTEGER): STRING is
			-- Description of `Current' for test vehicle
			-- notification message, where `rot' is the rotation
			-- in degrees clockwise
		do
			create Result.make (100)
			Result.append ("    ")
			Result.append (su.left_padded (design_no.out, 6))
			Result.append ("  ")
			Result.append (su.right_padded (design_name, 30))
			Result.append ("  ")
			Result.append (su.left_padded (rot.out, 9))
			Result.append ("  ")
			if is_test_vehicle then
				Result.append ("TEST VEHICLE")
			else
				Result.append ("CUST DIE")
			end
			Result.extend ('%N')
		end

	is_reworkable: BOOLEAN
			-- Is project currently reworkable?
		do
			Result := fab_instance.is_reworkable
		end

feature -- Modification

	uncache_packagings is
			-- Clear packaging cache
		do
			cached_packagings := Void
		ensure
			packagings_not_cached: not is_cached_packagings
		end

	uncache_packaged_packages is
			-- Clear cached list of packages in which
			-- `Current' will be packaged
		do
			cached_packaged_packages := Void
		ensure
			packages_not_cached: not is_cached_packaged_packages
		end

	set_packagings (list: like packagings) is
			-- Set `packagings' to `list'
		require
			list_not_void: list /= Void
		do
			cached_packagings := list
		ensure
			packagings_set: packagings = list
		end

	uncache_options is
			-- Clear option cache
		do
			cached_options := Void
		ensure
			options_not_cached: not is_cached_options
		end

	set_options (list: like options) is
			-- Set `options' to `list'
		require
			list_not_void: list /= Void
		do
			cached_options := list
		ensure
			options_set: options = list
		end

	set_selected (b: BOOLEAN) is
			-- Set `is_selected' to `b'
		do
			is_selected := b
		ensure
			value_set: is_selected = b
		end

	uncache_fab_instance is
			-- Clear `fab_instance' from cache
		do
			cached_fab_instance := Void
		ensure
			not_cached: not is_cached_fab_instance
		end

	uncache_cust_ic_design is
			-- Clear `cust_ic_design' from cache
		do
			cached_cust_ic_design := Void
		ensure
			not_cached: not is_cached_cust_ic_design
		end

	uncache_cust_account is
			-- Clear `cust_account' from cache
		do
			cached_cust_account := Void
		ensure
			not_cached: not is_cached_cust_account
		end

	uncache_test_vehicle is
			-- Clear `test_vehicle' from cache
		do
			cached_test_vehicle := Void
		ensure
			not_cached: not is_cached_test_vehicle
		end

	uncache_packing is
			-- Clear packing cache
		do
			cached_packing := Void
		ensure
			packing_not_cached: not is_cached_packing
		end

	uncache_dependents is
			-- Clear cached dependent rows
		do
			uncache_packagings
			uncache_packaged_packages
			uncache_options
		end

	uncache_foreign_keys is
			-- Clear cached rows that are foreign keys to
			-- other tables)
		do
			uncache_fab_instance
			uncache_cust_ic_design
			uncache_cust_account
		end

	compute_compatible_processes (table: HASH_TABLE [WAFER_PROCESSES_ROW, STRING]) is
			-- Compute processes on which `Current' can be
			-- fabricated and put in `compatible_processes'.
			-- Use `table' to find wafer etest ID of each
			-- process and leave list of distinct wafer etest
			-- IDs for compatible processes in `wafer_etest_ids'
			-- and list of distinct wafer etest names in
			-- `wafer_etest_names'
		do
			compute_processes_for_options (any_connection, packing.wafer_techn, layer_options, explicit_options)
			if compatible_processes.is_empty then
				handle_hard_error ("no compatible processes for design_no " + design_no.out)
			end
			create {ARRAYED_LIST [STRING]} compatible_wafer_etest_ids.make (1)
			create {ARRAYED_LIST [WAFER_ETEST_NAMES_ROW]} compatible_wafer_etest_names.make (1)
			compatible_wafer_etest_ids.compare_objects
			compatible_processes.do_all (agent add_etest_id (?, table))
		end

	assign_process (ord_list: LIST [STRING]) is
			-- Find the first process in `compatible_processes' 
			-- (i.e. the cheapest one) which is in
			-- `ord_list' (the current list of processes
			-- to be ordered) and put it in `assigned_process'
		require
			ord_list_not_void: ord_list /= Void
			list_object_comparison: ord_list.object_comparison
			is_customer_project: is_customer_project
		local
			procs: like compatible_processes
			proc, sep: STRING
		do
			sep := "%N   "
			assigned_process := Void
			from
				procs := compatible_processes_with_assigned_etest_name
				procs.start
			until
				procs.after or assigned_process /= Void
			loop
				proc := procs.item
				if ord_list.has (proc) then
					assigned_process := proc
				end
				procs.forth
			end
			if assigned_process = Void then
				handle_hard_error ("design_no " + design_no.out + " not compatible with any of the processes to be ordered.%NProcesses to order:" + sep + su.joined_by (ord_list, sep) + "%NCompatible processes:" + sep + su.joined_by (procs, sep))
			end
		end

	fetch_layer_options (wt: STRING) is
			-- Fetch options determined by layers from database,
			-- for wafer techn `wt'
		do
			compute_layer_options (any_connection, design_no, design_techn_id, wt)
		end

	set_reason_unselectable (s: STRING) is
			-- Set `reason_unselectable' to `s'
		do
			reason_unselectable := s
		ensure
			value_set: reason_unselectable = s
		end

feature {NONE} -- Implementation

	add_etest_id (s: STRING table: HASH_TABLE [WAFER_PROCESSES_ROW, STRING]) is
			-- Add wafer etest ID for process `s' to
			-- `compatible_wafer_etest_ids' if it is not already
			-- there, using `table' to map process name to
			-- wafer_processes row
		local
			etest_id: STRING
			wp: WAFER_PROCESSES_ROW
		do
			table.search (s)
			if table.found then
				wp := table.found_item
				etest_id := wp.wafer_etest_id
			else
				handle_hard_error ("wafer process " + s + " not found in lookup table - cannot get its wafer etest ID")
			end
			if not compatible_wafer_etest_ids.has (etest_id) then
				compatible_wafer_etest_ids.extend (etest_id)
				compatible_wafer_etest_names.extend (wp.etest_name)
			end
		end

	layer_options: like options is
			-- Subset of `options' that is determined by
			-- layers present in design
		do
			Result := options.twin
			Result.wipe_out
			options.do_if (agent {like anchored_option}.add_to_list (Result), agent {like anchored_option}.is_layer_option)
		end

	explicit_options: like options is
			-- Subset of `options' that is explicitly requested
			-- by customer (can't be determined from layers
			-- present)
		do
			Result := options.twin
			Result.wipe_out
			options.do_if (agent {like anchored_option}.add_to_list (Result), agent {like anchored_option}.is_explicit_option)
		end

	fetch_packagings is
			-- Fetch packagings for `Current'
		local
			row: like anchored_packaging
		do
			create row.make_template
			cached_packagings := row.rows_for_project_version (packing_id, project_no, version_no, any_connection, expected_count.At_least_zero)
		end
		
	fetch_packaged_packages is
			-- Fetch list of packages in which `Current' will
			-- be packaged
		local
			pkgs: ARRAYED_LIST [STRING]
			pkging: like anchored_packaging
			l_packagings: like packagings
		do
			l_packagings := packagings
			create {ARRAYED_LIST [PACKAGES_ROW]} cached_packaged_packages.make (l_packagings.count)
			create pkgs.make (l_packagings.count)
			pkgs.compare_objects
			from
				l_packagings.start
			until
				l_packagings.after
			loop
				pkging := l_packagings.item
				if not pkging.is_null_package_id and pkging.packaged and then not pkgs.has (pkging.package_id) then
					pkgs.extend (pkging.package_id)
					cached_packaged_packages.extend (pkging.package)
				end
				l_packagings.forth
			end
		end
		
	anchored_packaging: PACKING_PROJECT_PACKAGINGS_ROW is
			-- For anchoring only
		do
		ensure
			result_void: Result = Void
		end

	fetch_options is
			-- Fetch options for `Current'
		local
			row: like anchored_option
		do
			create row.make_template
			cached_options := row.rows_for_project_version (packing_id, project_no, version_no, any_connection, expected_count.At_least_zero)
		end
		
	anchored_option: PACKING_PROJECT_OPTIONS_ROW is
			-- For anchoring only
		do
		ensure
			result_void: Result = Void
		end

	fetch_fab_instance is
			-- Fetch fab instance
		do
			create cached_fab_instance.make_by_key (any_connection, option.Old_no_exception_row, m_fab_no)
		end
		
	fetch_cust_ic_design is
			-- Fetch cust IC design
		do
			create cached_cust_ic_design.make_by_key (any_connection, option.Old_row, design_no)
			cached_cust_ic_design.use_final_layout
		end
		
	fetch_cust_account is
			-- Fetch cust account
		do
			create cached_cust_account.make_by_key (any_connection, option.Old_row, cust_acct_no)
		end
		
	fetch_test_vehicle is
			-- Fetch test vehicle for `Current'
		local
			row: M_TEST_VEHICLES_ROW
		do
			create row.make_template
			cached_test_vehicle := row.test_vehicle_with_design_no (test_vehicle_design_no, packing.wafer_techn, any_connection, expected_count.Exactly_one).first
		end
		
	fetch_packing is
			-- Fetch packing of which `Current' is part
		do
			create cached_packing.make_by_key (any_connection, option.Old_row, packing_id)
		end
		
	anchored_fab_instance: IC_FAB_INSTANCES_ROW is
			-- For anchoring only
		do
		ensure
			result_void: Result = Void
		end

	anchored_cust_ic_design: CUST_IC_DESIGNS_ROW is
			-- For anchoring only
		do
		ensure
			result_void: Result = Void
		end

	anchored_cust_account: CUST_ACCTS_ROW is
			-- For anchoring only
		do
		ensure
			result_void: Result = Void
		end

	anchored_test_vehicle: M_TEST_VEHICLES_ROW is
		do
		ensure
			result_void: Result = Void
		end

	anchored_packing: PACKINGS_ROW is
			-- For anchoring only
		do
		ensure
			result_void: Result = Void
		end

feature {PACKING_PROJECTS_ROW} -- Implementation

	cached_packagings: like packagings
			-- Cached list of packagings
		
	cached_packaged_packages: like packaged_packages
			-- Cached list of packages in which `Current' will
			-- be packaged
		
	cached_options: like options
			-- Cached list of options
		
	cached_packing: like packing
			-- Cached packing
		
	cached_cust_account: like cust_account
			-- Cust account row
		
	cached_test_vehicle: like test_vehicle
			-- Test vehicle row, cached in memory
		
	cached_cust_ic_design: like cust_ic_design
			-- Cust IC designs row
		
	cached_fab_instance: like fab_instance
			-- Fab instance
		
end
