class PART_AND_PRICE
		-- M_PARTS row and its price
		-- deals with standard prices and quoted prices
		-- (see set_price et al. for locating a price for me)

inherit
	M_PARTS_ROW

	MOSIS_PRICE_CALCULATOR

	GENERIC_SYSTEM_SUPPORT			-- (punts in 2 places)
		export
			{NONE} all
		end

create

	make_by_key, make_template, make_null

feature -- attributes of the price

	have_a_price: BOOLEAN is
		-- do we have a price
		-- (either as a result of a successful set_price or
		--  because our part is priceless)
	do  Result := is_queried_price or is_priceless
	end;

	is_priceless: BOOLEAN is
		-- is this a part that does not need a price?
	do  Result := (is_package and not (packaged and cust_orderable))
			or
		      is_on_the_house
	end;

	is_on_the_house: BOOLEAN;
		-- is this part gratis (see set_on_the_house; default no)

	is_queried_price: BOOLEAN is
		-- is the price set from the database?
	do  Result := my_price /= Void
	end;

	is_standard_price: BOOLEAN is
		-- is this a standard price_schedule price?
	do  Result := is_queried_price and then my_price.is_standard_price
	end;

	is_quoted_price: BOOLEAN is
		-- is this price based on a custom quote?
	do  Result := is_queried_price and then my_price.is_quoted_price
	end;

	have_quoted_prices: BOOLEAN is
		-- did we find some quoted prices (need to pick one;
		--  note that we may not yet have_a_price)
	do  Result := my_prices /= Void and then (not my_prices.is_empty)
			and then my_prices.first.is_quoted_price
	end;

	quantities_quoted: LIST[INTEGER] is
	require	have_quoted_prices
	do  create {ARRAYED_LIST[INTEGER]} Result.make(my_prices.count);
	    from  my_prices.start
	    until my_prices.after
	    loop  Result.extend(my_prices.item.quantity_quoted)
		  my_prices.forth
	    end
	end;

	m_price_no: INTEGER is
	require  is_standard_price
	do  Result := my_price.price_identification_no
	end;

	pricing_schedule: STRING is
	require  is_standard_price
	do  Result := my_price.pricing_schedule
	end;

	date_effective: DB_DATE is
	require  is_queried_price
	do  Result := my_price.date_effective
	end;

	date_expires: DB_DATE is
	require  is_queried_price
	do  if not is_null_date_expires
	     then Result := my_price.date_expires
	    end
	end;

	is_null_date_expires: BOOLEAN is
	require  is_queried_price
	do  Result := my_price.is_null_date_expires
	end;

	flat_amount: DOUBLE is
	-- require  have_a_price
	do  if is_queried_price
	     then Result := my_price.flat_amount
	     else Result := 0
	    end
	end;

	area_amount: DOUBLE is
	-- require  have_a_price
	do  if is_queried_price
	     then Result := my_price.area_amount
	     else Result := 0
	    end
	end;

	edu_units: INTEGER is
	-- require  have_a_price
	do  if is_standard_price
	     then Result := my_price.edu_units
	     else Result := 0		-- quoted or priceless
	    end
	end;

	pricing_descr: STRING is
		-- readable descriptor of the price source
	require have_price: have_a_price or have_quoted_prices
	local b: DB_BIND_S; tgd: STRING;
	do  if have_quoted_prices
	     then  -- can describe even if haven't picked one out yet
		  Result := su.string_of_2(
				"MOSIS quote #", my_prices.first.descr)
	    elseif is_standard_price
	     then -- first the price family
		  if is_IC_or_prep
		   then create b.make(techn_group_family);
			tgd := connection.safe_selected_string(
				"descr",
				"techn_group_families",
				"techn_group_family = :1", b)
		   else tgd := package_group
		  end;
		  -- put it all together
		  Result := su.string_of_4(
				tgd, su.space, my_price.descr, " price list")
	     else Result := "no additional cost"
	    end;
	end;

feature -- getting at the price info

	set_on_the_house(gratis: BOOLEAN) is
		-- marks this part as being free regardless of prices
		-- (used for dedicated run IC parts)
	do  is_on_the_house := gratis;
	    if gratis then my_prices := Void end;
	ensure is_on_the_house = gratis
	end;
	    
	set_price(pe: PRICING_ENVIRONMENT) is
		-- (may fail; see have_a_price on return)
	do  my_prices := Void;
	    -- try for a quoted price first
	    if not (is_on_the_house or pe.is_null_quote_no)
	     then set_quoted_price(pe)
	    end;
	    -- failing that, try for a standard price
	    if not (is_on_the_house or my_prices /= Void)
	     then set_standard_price(pe)
	    end
	    -- (both might fail, leaving a Void)
	end;

	select_by_quantity_quoted_and_area(quant: INTEGER; area: DOUBLE) is
		-- pick my_price from the list based on matching quantity
		-- in case of multiple quant matches, use closest in area
	require have_quoted_prices
	local bst: INTEGER; pdif, bdif: DOUBLE; pr: MOSIS_PRICE
	do  from  my_prices.start
	    until my_prices.after
	    loop  pr := my_prices.item;
		  if pr.quantity_quoted = quant
		   then pdif := (pr.assumed_area - area).abs;
			if bst = 0 or else pdif < bdif
			 then -- best line item so far
			      bst := my_prices.index; bdif := pdif
			end
		  end
		  my_prices.forth
	    end;
	    if bst > 0
	     then -- bst is exact quant; closest in area
		  my_prices.go_i_th(bst)
	     else -- no quant match; leave at my.prices.after
	    end
	ensure have_a_price implies my_price.quantity_quoted = quant
	end;

feature {NONE} -- features for handling multiple prices

	my_prices: LIST[MOSIS_PRICE]

	set_standard_prices(pe: PRICING_ENVIRONMENT) is
		-- (may fail; see my_prices on return)
	    do
		my_prices := Void;
		set_standard_price(pe)
	    end

feature {NONE} --

	my_price: MOSIS_PRICE is
		-- uses my_prices.item, whichever element that may be
	    do
		if my_prices /= Void and then not my_prices.off then
		  Result := my_prices.item
		end
	    end

	set_quoted_price(pe: PRICING_ENVIRONMENT) is
		-- (often "fails", leaving Void my_price, but 
		--  there is still have_quoted_prices)
	local	rq: CUST_QUOTE_LINE_ITEMS_ROW; bq: DB_BIND_II;
	do  create rq.make_template;
	    create bq.make(m_part_no, pe.quote_no);
	    my_prices := rq.all_rows_filtered_ordered
				(connection, part_and_quote, bq,
				 "quantity_quoted");
	    -- fix up position
	    my_prices.start
	    if my_prices.is_empty
	     then my_prices := Void
	    elseif my_prices.count > 1
	     then my_prices.back	-- puts us off the list
	    end;
	end;

	set_standard_price(pe: PRICING_ENVIRONMENT) is
		-- (may fail; see have_a_price on return)
	local	r: M_PRICES_ROW;
		b: DB_BIND_SID; b1: DB_BIND_ID; psched: STRING
	do  psched := pe.pricing_schedule;
	    create r.make_template;
	    if su.is_empty(psched)
	     then create b1.make(m_part_no, pe.date_for_pricing)
		  my_prices := r.all_rows_filtered
				(connection, part_date, b1)
	     else create b.make(psched, m_part_no, pe.date_for_pricing);
		  my_prices := r.all_rows_filtered
				(connection, sched_part_date, b);
	    end
	    if my_prices.count = 1
	     then my_prices.start
	    elseif my_prices.count > 1 and su.not_empty(psched)
	     then system_punt(su.string_of_6
				("SET_PRICE: multiple price for part ",
				 m_part_no.out, " on schedule ",
				 psched, " for date ",
				 pe.date_for_pricing.out))
	    end;
	end;

	sched_part_date: STRING is
		-- filter for my price
	"pricing_schedule = :1 and m_part_no = :2 and %
	%trunc(date_effective) <= :3 and %
	%(date_expires is null or date_expires >= trunc(:3))"

	part_date: STRING is
		-- filter for all my prices
	"m_part_no = :1 and %
	%trunc(date_effective) <= :2 and %
	%(date_expires is null or date_expires >= trunc(:2))"

	part_and_quote: STRING is
		-- filter for my quote
	"m_part_no = :1 and m_cust_quote_no = :2"

feature -- price_breakdowns_rows

	the_breakdowns: LIST[GENERIC_PRICE_BREAKDOWN] is
	    require
		is_priced: have_a_price
	    do
		if my_breakdowns = Void then
		    get_my_breakdowns
		elseif is_standard_price and then
		      m_price_no /= my_breakdowns_id then
		    get_my_breakdowns
		end;
		Result := my_breakdowns
	    ensure
		Result /= Void
	    end;

feature {NONE} -- implementation for price_breakdowns_rows

	my_breakdowns: LIST[GENERIC_PRICE_BREAKDOWN]

	my_breakdowns_id: INTEGER

	get_my_breakdowns is
		-- like fetch_the_list, but with check on sums
	local	b: DB_BIND_I; r: PRICE_BREAKDOWNS_ROW;
		rg: GENERIC_PRICE_BREAKDOWN; flat, area: DOUBLE;
	do  if is_standard_price
	     then -- get PRICE_BREAKDOWNS
		  create b.make(m_price_no);
		  create r.make_template;
		  my_breakdowns := r.all_rows_filtered
					(connection, "m_price_no = :1", b);
		  my_breakdowns_id := m_price_no;
	     else create {LINKED_LIST[GENERIC_PRICE_BREAKDOWN]}
				my_breakdowns.make
		  my_breakdowns_id := 0;
	    end;
	    -- now check that they add up to the price we have
	    --  or that there are no breakdowns (allowed as of 13-Feb-2001)
	    from  my_breakdowns.start
	    until my_breakdowns.after
	    loop  rg := my_breakdowns.item;
		  flat := flat + rg.flat_amount
		  area := area + rg.area_amount
		  my_breakdowns.forth
	    end;
	    if not (my_breakdowns.is_empty or
		    (flat = flat_amount and area = area_amount))
	     then system_punt_int("Breakdowns do not add up for m_part_no",
				m_part_no)
	    end;
	end;

end -- class PART_AND_PRICE
