class PATH
	-- depicted as a center-line path (the trace), a width, and an end type

inherit
	TRACE
	rename	make as trace_make,
		make_copy as trace_make_copy,
		make_transformed as trace_make_transformed,
		bounding_box as trace_bounding_box,
		out as path_out
	end;

	GEOMETRY_SUPPORT
	export {NONE} all
	redefine out
	select	out
	end;

create
	make, make_frame, make_copy, make_transformed

feature	-- other attributes (beside the list_of_points)

	width: DOUBLE

	ends: PATH_OPTIONS

feature -- useful stuff

	bounding_box: CONST_RECTANGLE is
	local poly: POLYGON
	do  create poly.make(3*count);
	    drawn_out(poly);
	    Result := poly.bounding_box
	end;

	out: STRING is
	do  create Result.make(80);
	    Result.append("W: "); Result.append_double(width);
	    Result.append("; E: "); Result.append(ends.out);
	    Result.append("; P: "); Result.append(path_out);
	end;

feature	-- change the width

	set_width(w: DOUBLE) is
		-- reset the width
	require w >= 0
	do  width := w
	end;

feature	-- change the end type (may require adjusting first and last points)
	-- needed, at least, by the cif writer

	can_set_ends(e: PATH_OPTIONS): BOOLEAN is
		-- is this end change one we can handle?
	do  
	    Result := (ends.is_flush = e.is_flush)	-- no change
			or
		      (width = 0)			-- nothing to change
			or
		      (count > 1 and then		-- changing
			(e.is_flush			--  to flush is OK
			  or
			 -- from flush to extended; must retract ends
			(count = 2 and segment_length(1) > width)
			  or
			(count > 2 and
				segment_length(1) > width/2 and
				segment_length(count-1) > width/2)))
	end;

	set_ends(e: PATH_OPTIONS) is
		-- change the end type;
		-- if we go from flush to non-flush or vice versa,
		--  we must adjust the first and last points
	require can_set_ends(e)
	local amt: DOUBLE; ls: LINE_SEGMENT
	do  if e.is_flush and not ends.is_flush
	     then -- going from extended to flush; extend end points
		amt := width/2
	    elseif ends.is_flush and not e.is_flush
	     then -- going from flush to extended; retract end points
		amt := -width/2
	    end;
	    if amt /= 0		-- adjust each end by amt
	     then ls := scratch_line_segment;
		  -- first segment, but backwards so pt #1 is extended
		  ls.make_xy(i_th_x(2), i_th_y(2), i_th_x(1), i_th_y(1));
		  ls.extend(amt)
		  put_i_th_x(ls.x2, 1);
		  put_i_th_y(ls.y2, 1);
		  -- last segment
		  put_i_th_segment(count-1, ls);
		  ls.extend(amt);
		  put_i_th_x(ls.x2, count);
		  put_i_th_y(ls.y2, count);
	    end;
	    ends := e
	ensure ends = e
	end;
	
feature	-- represent as a polygon

	drawn_out(poly: SIMPLE_LIST_OF_POINTS) is
		-- draw self as a closed polygon into poly
		-- (which must be empty coming in)
	require have_poly: poly /= Void
		empty_poly: poly.is_empty
	local ndx: INTEGER; pseg, nseg: LINE_SEGMENT
	do  if is_empty
	     then -- return empty poly
	    elseif count = 1
	     then -- return "expanded" point
		  draw_a_path_point(poly)
	     else -- normal path 
		  nseg := i_th_segment(1);
		  draw_a_path_end(poly, nseg, ends.begin_extension);
		  from  ndx := 2
		  until not valid_segment(ndx)
		  loop  pseg := nseg;
			nseg := i_th_segment(ndx)
			draw_a_path_internal(poly, pseg, nseg)
			ndx := ndx + 1
		  end;
		  -- now go back around the other side of the wire
		  nseg.reverse		--
		  draw_a_path_end(poly, nseg, ends.end_extension);
		  from  ndx := count-2
		  until not valid_segment(ndx)
		  loop	pseg := nseg;
			nseg := i_th_segment(-ndx)
			draw_a_path_internal(poly, pseg, nseg)
			ndx := ndx - 1
		  end
	    end
	end;

feature {NONE} -- drawn_out implementation

	draw_a_path_point(poly: SIMPLE_LIST_OF_POINTS) is
	local x, y, hw: DOUBLE; m: MATRIX; p: POLYGON
	do  x := i_th_x(1); y := i_th_y(1);	-- our single point
	    hw := width / 2;
	    if width = 0 or ends.is_flush
	     then poly.extend_xy(x,y)
	    elseif ends.is_half_square or ends.is_custom
	     then poly.extend_xy(x-hw, y-hw);
		  poly.extend_xy(x-hw, y+hw);
		  poly.extend_xy(x+hw, y+hw);
		  poly.extend_xy(x+hw, y-hw)
	     else m := magnify_by(width).and_then(translate_by_xy(x, y));
		  create p.make_transformed(the_unit_octagon, m);
		  poly.append(p)
	    end
	end;

	draw_a_path_end(poly: SIMPLE_LIST_OF_POINTS; seg: LINE_SEGMENT;
			cust_ext: DOUBLE) is
		-- add points for the outline end at seg.pt1
	local hw, ext: DOUBLE; 
	do  hw := width/2; 
	    -- compute extension beyond pt1
	    if ends.is_flush
	     then ext := 0
	    elseif ends.is_half_square
	     then ext := hw
	    elseif ends.is_rounded
	     then ext := hw/2		-- half octagon approximation
	     else ext := cust_ext
	    end;
	    add_an_end_point(poly, seg, -hw, ext);
	    if ends.is_rounded
	     then -- other two points of the 1/2 octagon
		  add_an_end_point(poly, seg, -ext, hw);
		  add_an_end_point(poly, seg, ext, hw)
	    end;
	    add_an_end_point(poly, seg, hw, ext)
	end;

	add_an_end_point(poly: SIMPLE_LIST_OF_POINTS; seg: LINE_SEGMENT;
			 pdist, ext: DOUBLE) is
	local edge: LINE_SEGMENT;
	do  if pdist > 0	-- parallel at pdist
	     then edge := seg.right_parallel(pdist)
	     else edge := seg.left_parallel(pdist.abs)
	    end;
	    edge.reverse;		-- path end is now pt2
	    edge.extend(ext.max(-edge.length))	-- stretch pt2
	    poly.extend_xy(edge.x2, edge.y2)
	end;

	draw_a_path_internal(poly: SIMPLE_LIST_OF_POINTS; 
				prv, nxt: LINE_SEGMENT) is
		-- add point(s) for the outline turn at juncture of prv and nxt
		-- on the "right" side of the path
	-- require prv.pt2 = nxt.pt1
	local hw, bend: DOUBLE; pedge, nedge: LINE_SEGMENT; pt: POINT
	do  hw := width / 2;
	    bend := normalized_degrees(nxt.degrees - prv.degrees);
	    pedge := prv.right_parallel(hw);
	    nedge := nxt.right_parallel(hw);
	    if bend = 0
	     then -- nothing; skip this point
	    elseif bend <= 90
	     then -- inside bend or obtuse bend; use intersection
		  pt := pedge.intersects_at(nedge);
		  poly.extend_xy(pt.x, pt.y)
	     else -- outside acute bend; bevel
		  pedge.extend(hw)
		  nedge.reverse; nedge.extend(hw);
		  poly.extend_xy(pedge.x2, pedge.y2);
		  poly.extend_xy(nedge.x2, nedge.y2)
	    end;
	end;

feature {NONE} -- creation

	make(w: DOUBLE; e: PATH_OPTIONS) is
		-- path is built piecemeal via extend ops
	require w >= 0; e /= Void;
	do  trace_make(5);
	    width := w;
	    ends := e
	end;

	make_frame(box: CONST_RECTANGLE; w: DOUBLE) is
		-- path is a "frame" around box
	require box /= Void; w >= 0
	do  make(w, half_square_ends);
	    extend(box.ul);  extend(box.ur);
	    extend(box.lr);  extend(box.ll);
	    extend(box.ul)		-- close explicitly
	end;

	make_copy(src: PATH) is
		-- copy of another path
	require src /= Void
	do  trace_make_copy(src);
	    width := src.width
	    ends := src.ends
	end;

	make_transformed(src: PATH; at: MATRIX) is
		-- transformation of another path
	require src /= Void; at /= Void;
		src.width = 0 or at.is_gds_legal;
	do  trace_make_transformed(src, at);
	    if src.width > 0
	     then width := src.width * at.magnification
	     else width := 0
	    end;
	    ends := src.ends
	end;

end -- PATH
