class POLYGON
		-- list of points representing a closed polygon
		-- the last point folds back to the first point

inherit
	LIST_OF_POINTS

create
	make,
	make_copy,
	make_transformed

feature	-- status

	valid_segment(i: INTEGER): BOOLEAN is
		-- there are count vertices and count edges
	do  Result := 1 <= i and i <= count
	end;

	other_vertex(i: INTEGER): INTEGER is
		-- find the vertex index of the second vertex in segment i
	require valid_segment(i)
	do  if i < count
	     then Result := i + 1
	     else Result := 1
	    end
	end;

feature	-- area

	area: DOUBLE is
	local i1, i2: INTEGER
	do  from  i1 := 1; Result := 0
	    until i1 > count
	    loop  i2 := other_vertex(i1);
		  Result := Result + 
			    ((i_th_x(i2) - i_th_x(i1)) *
			     (i_th_y(i1) + i_th_y(i2)) / 2)
		  i1 := i1 + 1
	    end;
	    Result := Result.abs
	end;

feature	-- rectangularity

	is_rectangular: BOOLEAN is
		-- is this polygon a closet rectangle
	do  Result :=
		count = 4 and then (
		(gres.eq(i_th_x(1), i_th_x(2)) and
		 gres.eq(i_th_y(2), i_th_y(3)) and
		 gres.eq(i_th_x(3), i_th_x(4)) and
		 gres.eq(i_th_y(4), i_th_y(1)))
		 or
		(gres.eq(i_th_y(1), i_th_y(2)) and
		 gres.eq(i_th_x(2), i_th_x(3)) and
		 gres.eq(i_th_y(3), i_th_y(4)) and
		 gres.eq(i_th_x(4), i_th_x(1))) )
	end;

	to_rectangle: RECTANGLE is
	require is_rectangular
	do  create Result.make_by_corners(i_th_x(1), i_th_y(1),
				     i_th_x(3), i_th_y(3))
	end;

feature	-- trapezoidality

	is_horizontal_trapezoid: BOOLEAN is
		-- is this polygon a trapezoid with two horizontal edges?
	do  Result :=
		count = 4 and then
		((gres.eq(i_th_y(1), i_th_y(2)) and
		  gres.eq(i_th_y(3), i_th_y(4)))
		 or
		 (gres.eq(i_th_y(2), i_th_y(3)) and
		  gres.eq(i_th_y(4), i_th_y(1))))
	end;

	is_vertical_trapezoid: BOOLEAN is
		-- is this polygon a trapezoid with two vertical edges?
	do  Result :=
		count = 4 and then
		((gres.eq(i_th_x(1), i_th_x(2)) and
		  gres.eq(i_th_x(3), i_th_x(4)))
		 or
		 (gres.eq(i_th_x(2), i_th_x(3)) and
		  gres.eq(i_th_x(4), i_th_x(1))))
	end;

feature -- winding number for a point

	winding_number(p: VIRTUAL_POINT): INTEGER is
		-- winding number for p
	require have_point: p /= Void
	local i1,i2: INTEGER; px,py,x1,y1,x2,y2: DOUBLE
	do  from  i1 := 1; Result := 0;
		  px := p.x; py := p.y;
	    until i1 > count
	    loop  -- find other end of this segment
		  i2 := other_vertex(i1);
		  -- load up this segment's end points
		  x1 := i_th_x(i1); y1 := i_th_y(i1);
		  x2 := i_th_x(i2); y2 := i_th_y(i2);
		  -- we draw a ray from p to the right
		  if (gres.lt(y1, py) and gres.le(py, y2)) and then
			gres.lt(px, x2 + (x1 - x2)*(y2 - py)/(y2 - y1))
		   then Result := Result + 1
		  elseif (gres.lt(y2, py) and gres.le(py, y1)) and then
			gres.lt(px, x1 + (x2 - x1)*(y1 - py)/(y1 - y2))
		   then Result := Result - 1
		  end;
		  i1 := i1 + 1
	    end;
	end;

feature	-- contains point

	contains(p: VIRTUAL_POINT): BOOLEAN
		-- is p inside this polygon?
		-- aka, is winding_number(p) odd?
		-- Note: approx half the points that are ON one of my
		--  edges will be contained and half not contained!!
	require have_point: p /= Void
	do  Result := ((winding_number(p).abs) \\ 2) = 1
	end;

feature	-- encloses rectangle

	probably_encloses(r: CONST_RECTANGLE): BOOLEAN
		-- is r (probably) enclosed by me?
		-- for convex polygons, the answer is definitive, but
		-- for concave polygons, it is not
	require have_rect: r /= Void
	do  Result := contains(r.ul) and contains(r.ur) and
		      contains(r.ll) and contains(r.lr)
	end;

end -- POLYGON
