
indexing
	
	description: "Parser for PostScript file that separates it into %
		%pages, to allow creation of a file with a subset of the pages"
	author: "David Hollenberg"
	date: "April 7, 2005"

class POSTSCRIPT_PAGE_PARSER
inherit
	ANY
	CONDITION_UTILITIES
		export
			{NONE} all
		end
	STRING_UTILITIES_ACCESS
		export
			{NONE} all
		end

feature -- Operations

	parse_named_file (fname: STRING) is
			-- Parse PostScript file named `fname' and
			-- set `preamble', `pages' and `postamble'
		require
			fname_not_void: fname /= Void
			fname_not_empty: not fname.is_empty
		local
			open_op: FILE_OPEN_READ_OP
			close_op: FILE_CLOSE_OP
			f: LARGE_PLAIN_TEXT_FILE
		do
			create open_op
			create close_op
			create f.make (fname)
			open_op.open (f)
			parse_file (f)
			close_op.close (f)
		end

	parse_file (f: LARGE_PLAIN_TEXT_FILE) is
			-- Parse PostScript file `f' and
			-- set `preamble', `pages' and `postamble'
		require
			f_not_void: f /= Void
			f_open_read: f.is_open_read
		local
			s: DATA_SOURCE_FILE
		do
			create s.make_entire (f)
			parse_source (s)
		end

	parse_source (s: DATA_SOURCE) is
			-- Parse PostScript file in `s' and
			-- set `preamble', `pages' and `postamble'
		require
			s_not_void: s /= Void
		local
			line: STRING
			is_page_start, is_end, is_eof: BOOLEAN
			state: INTEGER
			offset: INTEGER_64
			page: POSTSCRIPT_PAGE
		do
			create preamble
			preamble.set_start_offset (0)
			preamble.set_end_offset (0)
			create {ARRAYED_LIST [POSTSCRIPT_PAGE]} pages.make (0)
			create postamble
			postamble.set_start_offset (0)
			postamble.set_end_offset (0)
			
			state := 0
			s.reset
			from
				offset := s.position
				page := preamble
				s.read_line
			until
				s.after
			loop
				line := s.last_string
				is_page_start := su.is_prefix (Page_comment, line)
				is_end := su.is_prefix (Trailer_comment, line)
				is_eof := su.is_prefix (Eof_comment, line)
				inspect state
				when 0 then
					if is_page_start then
						page.set_end_offset (offset)
						page := new_page (line, offset)
						pages.extend (page)
						state := 1
					end
				when 1 then
					if is_page_start or is_end or is_eof then
						page.set_end_offset (offset)
						if is_page_start then
							page := new_page (line, offset)
							pages.extend (page)
						elseif is_end then
							page := postamble
							page.set_start_offset (offset)
							state := 2
						elseif is_eof then
							page := postamble
							page.set_start_offset (offset)
							state := 3
						end
					end
				when 2 then
					if is_eof then
						state := 3
					end
				when 3 then
					page.set_end_offset (offset)
					state := 4
				when 4 then
					-- Skip
				end
				offset := s.position
				s.read_line
			end
			if state /= 4 then
				page.set_end_offset (s.count)
			end
		end

feature -- Properties

	preamble: POSTSCRIPT_PAGE
			-- Preamble (part before first page).  Only the
			-- offset and count for page are applicable
			-- since this isn't really a PostScript page

	postamble: POSTSCRIPT_PAGE
			-- Postamble (part after last page).  Only the
			-- offset and count for page are applicable
			-- since this isn't really a PostScript page

	pages: LIST [POSTSCRIPT_PAGE]

feature {NONE} -- Implementation

	Page_comment: STRING is "%%%%Page:"
			-- Comment that marks start of PostScript page

	Trailer_comment: STRING is "%%%%Trailer"
			-- Comment that marks start of PostScript trailer
			-- (after all pages)

	Eof_comment: STRING is "%%%%EOF"
			-- Comment that marks PostScript end of file

	last_number: INTEGER
			-- Last ordinal page number from `parse_page_comment'

	last_label: STRING
			-- Last label from `parse_page_comment'

	new_page (line: STRING offset: INTEGER_64): POSTSCRIPT_PAGE is
			-- New page initialized from `line' and `offset'
		do
			create Result
			Result.set_start_offset (offset)
			parse_page_comment (line)
			Result.set_number (last_number)
			Result.set_label (last_label)
		ensure
			result_not_void: Result /= Void
		end

	parse_page_comment (line: STRING) is
			-- Parse page comment `line' and set `last_number'
			-- and `last_label'
		local
			c: CHARACTER
			in_parens: BOOLEAN
			state, paren_count, k: INTEGER
			num, octal: STRING
		do
			create last_label.make (line.count)
			create num.make (line.count)
			create octal.make (3)
			from
				k := Page_comment.count + 1
				state := 0
				in_parens := False
				paren_count := 0
			until
				k > line.count
			loop
				c := line.item (k)
				inspect state
				when 0 then
					inspect c
					when ' ', '%T' then
						state := 0
					when '(' then
						in_parens := True
						paren_count := paren_count + 1
						state := 1
					else
						k := k - 1
							-- Reprocess curr char
						state := 1
					end
				when 1 then
					inspect c
					when '(' then
						paren_count := paren_count + 1
						last_label.extend (c)
					when ')' then
						paren_count := paren_count - 1
						if in_parens and paren_count = 0 then
							state := 5
						else
							last_label.extend (c)
						end
					when '\' then
						state := 2
					when ' ', '%T' then
						if in_parens then
							last_label.extend (c)
						else
							state := 5
						end
					else
						last_label.extend (c)
					end
				when 2 then
					state := 1
					inspect c
					when 'b' then
						last_label.extend ('%B')
					when 'f' then
						last_label.extend ('%F')
					when 'n' then
						last_label.extend ('%N')
					when 'r' then
						last_label.extend ('%R')
					when 't' then
						last_label.extend ('%T')
					when '0'..'7' then
						octal.keep_head (0)
						octal.extend (c)
						state := 3
					else
						last_label.extend (c)
					end
				when 3 then
					inspect c
					when '0'..'7' then
						octal.extend (c)
						state := 4
					else
						append_octal_char (octal)
						state := 1
						k := k - 1
							-- Reprocess curr char
					end
				when 4 then
					inspect c
					when '0'..'7' then
						octal.extend (c)
						append_octal_char (octal)
					else
						append_octal_char (octal)
						k := k - 1
							-- Reprocess curr char
					end
					state := 1
				when 5 then
					num.extend (c)
				end
				k := k + 1
			end
			num.left_adjust
			num.right_adjust
			if num.is_integer then
				last_number := num.to_integer
			else
				last_number := 0
			end
		end

	append_octal_char (s: STRING) is
			-- Extend `last_label' with octal character
			-- represented by `s'
		require
			s_not_void: s /= Void
			s_not_empty: not s.is_empty
			octal_digits_only: -- `s' has only octal digits
		local
			n: INTEGER
		do
			n := octal_to_decimal (s)
			if n.is_valid_character_8_code then
				last_label.extend (n.to_character_8)
			end
		end

	octal_to_decimal (s: STRING): INTEGER is
			-- `s' as a decimal integer
		require
			s_not_void: s /= Void
			s_not_empty: not s.is_empty
			octal_digits_only: -- `s' has only octal digits
		local
			k, count: INTEGER
		do
			count := s.count
			from
				k := 1
			until
				k > count
			loop
				Result := Result * 8 + (s.item (k) |-| '0')
				k := k + 1
			end
		end

end
