
indexing
	
	author: "David Hollenberg"
	date: "May 3, 2002"
	description: "A file descriptor for a spawned process"

deferred class PROCESS_FD

inherit
	ANY
	UNIX_OS_ACCESS
		export
			{NONE} all
		end

feature -- Properties

	number: INTEGER
			-- Number of file descriptor

	is_pipe: BOOLEAN
			-- Is file descriptor associated with a pipe connected
			-- to parent process?
		do
		end

	is_input_pipe: BOOLEAN
			-- Is file descriptor associated with a pipe connected
			-- to parent process, from which child can read input?
		do
			if is_pipe then
				Result := is_input
			end
		end

	is_output_pipe: BOOLEAN
			-- Is file descriptor associated with a pipe connected
			-- to parent process, to which child can write output?
		do
			if is_pipe then
				Result := not is_input
			end
		end

	is_file: BOOLEAN
			-- Is file descriptor associated with a file?
		do
		end

	is_descriptor: BOOLEAN
			-- Is file descriptor setup from another file
			-- descriptor?
		do
		end

	is_duplicate: BOOLEAN
			-- Is file descriptor duplicated from another file
			-- descriptor?
		do
		end

	is_unchanged: BOOLEAN
			-- Is file descriptor left the same in child as
			-- it was in parent?
		do
		end

	is_close: BOOLEAN
			-- Should file descriptor be closed in spawned process?
		do
		end

	is_input: BOOLEAN
			-- Is pipe an input pipe (input to child)?
		require
			is_pipe: is_pipe
		do
		end

	compatible_descriptor: INTEGER
			-- Descriptor for `Current' which is compatible
			-- with old UNIX_PROCESS interface
		do
			if is_descriptor then
				Result := descriptor
			else
				Result := unix_os.Invalid_file_descriptor
			end
		end

	compatible_file_name: STRING
			-- File name for `Current' which is compatible
			-- with old UNIX_PROCESS interface
		do
			if is_file then
				Result := file_name
			elseif is_close then
				Result := ""
			else
				Result := Void
			end
		end

	descriptor: INTEGER
			-- Descriptor with which `number' is to be associated
		require
			is_descriptor: is_descriptor
		do
			Result := unix_os.Invalid_file_descriptor
		end

	dup_descriptor: INTEGER
			-- Descriptor from which `number' is to be duplicated
		require
			is_duplicate: is_duplicate
		do
			Result := unix_os.Invalid_file_descriptor
		end

	file_name: STRING
			-- File name with which `number' is to be associated
		require
			is_file: is_file
		do
			Result := Void
		end

	pipe_file: LARGE_RAW_FILE
			-- File with which `number' is associated in parent,
			-- for a pipe
		require
			is_pipe: is_pipe
		do
			Result := Void
		end

feature -- Validity

	valid_file_descriptor (fd: INTEGER): BOOLEAN
			-- Is `fd' a valid value for a file descriptor?
		do
			Result := unix_os.valid_file_descriptor (fd)
		end

feature {PROCESS_FD_SET} -- Operations

	prepare_for_fork
			-- Do any preparation needed prior to a fork
			-- to setup this file descriptor in spawned
			-- process
		deferred
		end

	setup_for_parent
			-- Take any actions needed to setup this file
			-- descriptor in parent process after a fork
		deferred
		end

	setup_for_child_early
			-- Take any early actions needed to setup this file
			-- descriptor in child process after a fork
		do
		end

	setup_for_child (fd_set: PROCESS_FD_SET)
			-- Take any actions needed to setup this file
			-- descriptor in child process after a fork,
			-- where `fd_set' is the entire set of file
			-- descriptors to be setup
		deferred
		end

	setup_for_child_late
			-- Take any late actions needed to setup this file
			-- descriptor in child process after a fork
		do
		end

feature {PROCESS_FD_SET} -- Properties

	collides_with (dest_fd: INTEGER): BOOLEAN
			-- Does source descriptor reference `dest_fd'
			-- in a manner which is a "collision",
			-- requiring that source descriptor be moved
			-- to another descriptor?
		do
			Result := number > dest_fd and source_descriptor = dest_fd and not is_duplicate
		end

	source_descriptor: INTEGER
			-- Descriptor in child that is source from which
			-- `number' is to be set

feature {PROCESS_FD_SET} -- Modification

	move_source_descriptor (new_fd: INTEGER)
			-- Move source descriptor to `new_fd', since
			-- it collides with existing descriptor
		do
			source_descriptor := new_fd
			moved := True
		end

feature {NONE} -- Implementation

	moved: BOOLEAN
			-- Has `source_descriptor' been moved to a clone
			-- of the original descriptor it referenced, due
			-- to a collision (or was descriptor not
			-- originally open)?

	close_source_descriptor_if_ok (fd_set: PROCESS_FD_SET)
			-- Close `source_descriptor' if `fd_set' indicates
			-- that it is OK to do so
		do
			if moved and not fd_set.references_descriptor (source_descriptor, number) then
				unix_os.close_file_descriptor (source_descriptor)
			end
		end

	move_desc (source, dest: INTEGER)
			-- If descriptor `source' is different than
			-- descriptor `dest', duplicate `source' onto `dest'
			-- and close `source'
		do
			if source /= dest then
				unix_os.duplicate_file_descriptor (source, dest)
				unix_os.close_file_descriptor (source)
			end
			
		end
	
invariant
	valid_source_desc: unix_os.valid_file_descriptor (source_descriptor) or source_descriptor = -1

end
