
indexing
	
	author: "David Hollenberg"
	date: "May 3, 2002"
	description: "A set of file descriptors and how they are to be %
		%setup for a spawned process"

class PROCESS_FD_SET

inherit
	ARRAY [PROCESS_FD]
	UNIX_OS_ACCESS
		export
			{NONE} all
		undefine
			copy, is_equal
		end

create
	make

feature {UNIX_PROCESS} -- Operations

	mark_descriptors_close_on_exec
			-- Mark all file descriptors whose disposition
			-- in child is not set so that they are "close on exec"
		local
			k, max_desc: INTEGER
		do
			from
				max_desc := unix_os.max_file_descriptors
				k := 0
			until
				k >= max_desc
			loop
				if k < lower or k > upper or else item (k) = Void then
					unix_os.mark_file_descriptor_close_on_exec (k)
				end
				k := k + 1
			end
		end

	prepare_for_fork
			-- Do any preparation needed prior to a fork
			-- to setup this file descriptor in spawned
			-- process
		local
			k: INTEGER
			elem: PROCESS_FD
		do
			from
				k := lower
			until
				k > upper
			loop
				elem := item (k)
				if elem /= Void then
					elem.prepare_for_fork
				end
				k := k + 1
			end
		end

	setup_for_parent
			-- Take any actions needed to setup this file
			-- descriptor in parent process after a fork
		local
			k: INTEGER
			elem: PROCESS_FD
		do
			from
				k := lower
			until
				k > upper
			loop
				elem := item (k)
				if elem /= Void then
					elem.setup_for_parent
				end
				k := k + 1
			end
		end

	setup_for_child
			-- Take any actions needed to setup this file
			-- descriptor in child process after a fork
		local
			k: INTEGER
			elem: PROCESS_FD
		do
			-- Do pass for "early" setup

			from
				k := lower
			until
				k > upper
			loop
				elem := item (k)
				if elem /= Void then
					elem.setup_for_child_early
				end
				k := k + 1
			end
			
			from
				k := lower
			until
				k > upper
			loop
				elem := item (k)
				if elem /= Void then
					move_desc_if_collision (elem.number)
					elem.setup_for_child (Current)
				end
				k := k + 1
			end
			
			-- Now do second pass for "late" setup

			from
				k := lower
			until
				k > upper
			loop
				elem := item (k)
				if elem /= Void then
					elem.setup_for_child_late
				end
				k := k + 1
			end
		end

feature {PROCESS_FD} -- Operations

	move_desc_if_collision (dest_fd: INTEGER)
			-- Move source descriptors in elements of set
			-- that reference `dest_fd' to another
			-- descriptor which is cloned from `dest_fd'
		local
			k, new_fd: INTEGER
			elem: PROCESS_FD
		do
			new_fd := unix_os.Invalid_file_descriptor
			from
				k := lower
			until
				k > upper
			loop
				elem := item (k)
				if elem /= Void and then elem.collides_with (dest_fd) then
					if new_fd = unix_os.Invalid_file_descriptor then
						new_fd := unix_os.clone_file_descriptor (elem.source_descriptor)
						
					end
					elem.move_source_descriptor (new_fd)
				end
				k := k + 1
			end
		end

feature {PROCESS_FD} -- Proerties

	references_descriptor (fd, pos: INTEGER): BOOLEAN
			-- Do any elements of set at positions greater than
			-- `pos' reference descriptor `fd' in a way
			-- that prevents it from being closed?
		local
			k: INTEGER
			elem: PROCESS_FD
		do
			from
				k := pos + 1
			until
				k > upper or Result
			loop
				elem := item (k)
				if elem /= Void and then (elem.source_descriptor = fd and not elem.is_unchanged) then
					Result := True
				end
				k := k + 1
			end
		end

end
