
indexing
	
	author: "David Hollenberg"
	date:	"December 1, 1998"
	description: "An encoder that encodes a stream of data in %
		%quoted-printable format"

class QUOTED_PRINTABLE_ENCODER

inherit
	DATA_TRANSFORM_SYNCHRONOUS
		rename
			consume as encode,
			consume_eof as finish,
			consume_internal as encode_internal,
			consume_eof_internal as finish_internal
		redefine
			reset
		end
	C_MEMORY_ALLOCATION
		export
			{NONE} all
		end
	MEMORY
		export
			{NONE} all
		redefine
			dispose
		end
	STRING_FILLER_ACCESS
		export
			{NONE} all
		end

create
	
	make

feature {NONE} -- Creation

	make
		do
			encoder := allocate_nongc_memory (qp_encoder_length)
			strfiller := strfill
			reset
		ensure
			text_mode: not binary_mode
			not_strong_encoding: not strong_encoding
		end

feature -- Properties

	in_initial_state: BOOLEAN
			-- Is `Current' in its initial state?
	
	binary_mode: BOOLEAN
			-- Will `Current' be used to encode files
			-- that are binary rather than text files?
	
	strong_encoding: BOOLEAN
			-- Will encoding be done in a "strong" manner,
			-- making it less immune to corruption by
			-- mail gateways but also less readable by people?
			-- Has no effect if not using binary mode
	
	Default_name: STRING = "Quoted printable decode"
			-- Default name of `Current'

feature -- Modification

	set_binary_mode (b: BOOLEAN)
			-- Set `binary_mode' to `b'
		do
			binary_mode := b
		ensure
			mode_set: binary_mode = b
		end

	set_strong_encoding (b: BOOLEAN)
			-- Set `strong_encoding' to `b'
		do
			strong_encoding := b
		ensure
			strong_encoding_set: strong_encoding = b
		end

feature -- Operations

	reset
			-- Reset `Current' to its initial state
		do
			precursor
			initialize_qp_encoder (encoder, binary_mode, strong_encoding)
			in_initial_state := True
		end
	
feature {NONE} -- Implementation
	
	encode_internal (the_data: STRING)
			-- Encode some more data in quoted printable
			-- format in `the_data' and put results in
			-- `output_buffer'
		local
			in_area, out_area: ANY
			outbuf: POINTER
			in_count, out_count, needed, new_cap: INTEGER
			added: INTEGER
		do
			in_initial_state := False
			in_count := the_data.count
			out_count := output_buffer.count
			needed := bytes_needed (in_count)
			new_cap := out_count + needed
			if new_cap > output_buffer.capacity then
				output_buffer.resize (new_cap)
			end
			in_area := the_data.area
			out_area := output_buffer.area
			outbuf := offset ($out_area, out_count)
			qp_encode (encoder, $in_area, in_count, outbuf)
			added := qp_encode_bytes_added (encoder)
			strfiller.set_count (output_buffer, added + out_count)
		end

	finish_internal
			-- Finish quoted printable encode operation
		local
			out_area: ANY
			outbuf: POINTER
			out_count, needed, new_cap: INTEGER
			added: INTEGER
		do
			in_initial_state := False
			out_count := output_buffer.count
			needed := bytes_needed (0)
			new_cap := out_count + needed
			if new_cap > output_buffer.capacity then
				output_buffer.resize (new_cap)
			end
			out_area := output_buffer.area
			outbuf := offset ($out_area, out_count)
			qp_encode_final (encoder, outbuf)
			added := qp_encode_bytes_added (encoder)
			strfiller.set_count (output_buffer, added + out_count)
		end

	encoder: POINTER
			-- Address of C memory with encoder state

	strfiller: STRING_FILLER
			-- Local copy of string filler object

	bytes_needed (count: INTEGER): INTEGER
			-- Number of bytes needed to store output
			-- generated from encoding `count' more bytes of input
		do
			Result := (count + 4) * 3 + ((count + 4) * 2.0 / 25.0).ceiling
				-- May be up to 4 bytes pending in encoder.
				-- Each byte expands to 3 bytes in worst
				-- case (e.g., =2E) and each 25 bytes of
				-- input can produce 75 bytes of encoded
				-- output which may need to be followed by
				-- 2 chars (= and LF)
		end

	dispose
			-- Deallocate resources when object is
			-- garbage collected
		local
			null_ptr: POINTER
		do
			if encoder /= null_ptr then
				free_nongc_memory (encoder)
				encoder := null_ptr
			end
		end

feature {NONE} -- Externals

	qp_encoder_length: INTEGER
			-- Length of C memory area required to store
			-- encoder's state
		external
			"C"
		end
	
	initialize_qp_encoder (e: POINTER binary, strong: BOOLEAN)
			-- Initialize encoder C memory area at address `e'.
			-- Prepare to encode binary data if `binary' is
			-- true, text data otherwise.  Use a stronger
			-- encoding if `strong' is true
		external
			"C"
		end
	
	qp_encode (e: POINTER inp: POINTER count: INTEGER outp: POINTER)
			-- Encode in quoted printable some more input
			-- at address `inp' and of length `count', using
			-- encoder whose C memory is at address `e'.
			-- Put the resulting output at `outp'
		external
			"C"
		end
	
	qp_encode_final (e: POINTER outp: POINTER)
			-- Get final quoted printable output and put in
			-- `outp', using encoder whose C memory is at
			-- address `e'
		external
			"C"
		end
	
	qp_encode_bytes_added (e: POINTER): INTEGER
			-- Number of bytes added to output buffer by
			-- last call to `qp_encode' or `qp_encode_final',
			-- for encoder described by C memory at address `e'
		external
			"C"
		end
	
end
