
indexing
	
	description: "A composite run closing step"
	author: "David Hollenberg"
	date: "April 16, 2003"

class RC_COMPOSITE_STEP
inherit
	RC_STEP
		redefine
			initialize
		end
	RUN_CLOSING_STEPS_ROW

create
	make_template

feature -- Initialization

	initialize (sess: RC_RUN_CLOSING_SESSION)
			-- Initialize step that will be executed in context
			-- of `sess'
		do
			precursor (sess)
			create {ARRAYED_LIST [RC_STEP]} component_steps.make (5)
		end


feature -- Properties

	is_primitive: BOOLEAN = False
			-- Is this a primitive step?
	
	type_name: STRING = "composite"
			-- Name for step type (primitive or composite)
	
	component_steps: LIST [RC_STEP]
			-- Steps that comprise this step

	primitive_steps: LIST [RC_PRIMITIVE_STEP]
			-- Primitive steps included in this step,
			-- either directly or indirectly, in order
		do
			create {ARRAYED_LIST [RC_PRIMITIVE_STEP]} Result.make (component_steps.count)
			from
				component_steps.start
			until
				component_steps.after
			loop
				Result.append (component_steps.item.primitive_steps)
				component_steps.forth
			end
		end

	was_applicable_to_run: BOOLEAN
			-- Was completed step applicable to run?
		do
			Result := False
			from
				component_steps.start
			until
				component_steps.after or Result
			loop
				if component_steps.item.was_applicable_to_run then
					Result := True
				end
				component_steps.forth
			end
		end

	is_done: BOOLEAN
			-- Has step been fully completed?
		do
			Result := True
			from
				component_steps.start
			until
				component_steps.after or not Result
			loop
				if not component_steps.item.is_done then
					Result := False
				end
				component_steps.forth
			end
		end

	is_undone: BOOLEAN
			-- Is step fully uncompleted (not even started)?
		do
			Result := True
			from
				component_steps.start
			until
				component_steps.after or not Result
			loop
				if not component_steps.item.is_undone then
					Result := False
				end
				component_steps.forth
			end
		end

	is_soft_undone: BOOLEAN
			-- Is step uncompleted, but was previously
			-- done and then undone with a "soft undo",
			-- possibly saving some state?
		do
			Result := False
			from
				component_steps.start
			until
				component_steps.after or Result
			loop
				if component_steps.item.is_soft_undone then
					Result := True
				end
				component_steps.forth
			end
		end

	is_partially_done: BOOLEAN
			-- Is step partially completed?
		local
			found_done, found_undone: BOOLEAN
			item: RC_STEP
		do
			Result := False
			from
				component_steps.start
			until
				component_steps.after or Result
			loop
				item := component_steps.item
				if item.is_partially_done then
					Result := True
				elseif item.is_done then
					found_done := True
				elseif item.is_undone then
					found_undone := True
				end
				component_steps.forth
			end
			Result := Result or (found_done and found_undone)
		end

	is_partially_undone: BOOLEAN
			-- Has step been partially undone?
		local
			item: RC_STEP
		do
			Result := False
			from
				component_steps.start
			until
				component_steps.after or Result
			loop
				item := component_steps.item
				if item.is_partially_undone then
					Result := True
				end
				component_steps.forth
			end
		end

	has_saved_state: BOOLEAN
			-- Does step have saved state from a previous
			-- execution?
		do
			Result := False
			from
				component_steps.start
			until
				component_steps.after or Result
			loop
				if component_steps.item.has_saved_state then
					Result := True
				end
				component_steps.forth
			end
		end

	repeated_step_text: STRING
			-- Text describing all primitive steps that occur
			-- more than once in `primitive_steps' or empty
			-- string if no repeated primitive steps
		local
			list: LIST [RC_PRIMITIVE_STEP]
			table: HASH_TABLE [INTEGER, STRING]
			name: STRING
			count: INTEGER
		do
			create Result.make (0)
			list := primitive_steps
			create table.make (list.count)
			from
				list.start
			until
				list.after
			loop
				name := list.item.step_name
				if table.has (name) then
					table.replace (table.item (name) + 1, name)
				else
					table.put (1, name)
				end
				list.forth
			end
			
			from
				list.start
			until
				list.after
			loop
				name := list.item.step_name
				count := table.item (name)
				if count > 1 then
					Result.append ("    Composite step " + step_name + " includes " + count.out + " occurrences of primitive step " + name + "%N")
					table.replace (1, name)
				end
				list.forth
			end
		end

	invalid_step_order_text: STRING
			-- Text describing all steps in
			-- `primitive_steps' that occur in an order
			-- that violates dependency constraints, or
			-- empty if no such steps
		local
			item1, item2: RC_PRIMITIVE_STEP
			list1, list2: LIST [RC_PRIMITIVE_STEP]
		do
			create Result.make (0)
			list1 := primitive_steps
			from
				list1.start
			until
				list1.after
			loop
				item1 := list1.item
				list2 := list1.twin
				from
					list2.forth
				until
					list2.after
				loop
					item2 := list2.item
					if item1.required_steps.has (item2) then
						Result.append ("    In composite step " + step_name + " step " + item1.step_name + " is executed before step " + item2.step_name + " but it depends on it%N")
					end
					list2.forth
				end
				list1.forth
			end
		end

end
