
indexing
	
	description: "A generator for objects that produce run closing maps"
	author: "David Hollenberg"
	date: "January 21, 2005"

class RC_MAP_GENERATOR

create
	make

feature {NONE} -- Creation

	make (sess: RC_RUN_CLOSING_SESSION)
			-- Create generator with `sess' as session context
		require
			sess_exists: sess /= Void
		do
			session := sess
			cache := sess.cache
			file_generator := sess.file_generator
			connection := session.connection
		end

feature -- Properties

	site_numbers_map: SITE_NUMBERS_MAP
			-- Map that shows mask site numbers and die ids,
			-- with one page for each field of view
		do
			create Result.make (session.run_id, connection.system_date)
			set_detailed_payload_map_data (Result)
		ensure
			result_exists: Result /= Void
		end

	organizations_map: ORGANIZATIONS_MAP
			-- Map that shows sponsors and organizations,
			-- with one page for each field of view
		do
			create Result.make (session.run_id, connection.system_date)
			set_detailed_payload_map_data (Result)
			Result.set_wafer_techn_organization (cache.wt_organizations_row)
		ensure
			result_exists: Result /= Void
		end

	design_names_map: DESIGN_NAMES_MAP
			-- Map that shows design names and die ids,
			-- with one page for each field of view
		do
			create Result.make (session.run_id, connection.system_date)
			set_detailed_payload_map_data (Result)
		ensure
			result_exists: Result /= Void
		end

	design_numbers_map: DESIGN_NUMBERS_MAP
			-- Map that shows design numbers and die ids,
			-- with one page for each field of view
		do
			create Result.make (session.run_id, connection.system_date)
			set_detailed_payload_map_data (Result)
		ensure
			result_exists: Result /= Void
		end

	step_pattern_map: STEP_PATTERN_MAP
			-- Map that shows fields of view stepped across wafer
		do
			create Result.make (session.run_id, connection.system_date)
			set_wafer_map_data (Result)
			Result.set_die_locations (cache.extended_die_locations_table)
		ensure
			result_exists: Result /= Void
		end

	detailed_step_pattern_map: DETAILED_STEP_PATTERN_MAP
			-- Map that shows fields of view stepped across wafer,
			-- with dies shown inside each field of view
		do
			create Result.make (session.run_id, connection.system_date)
			set_wafer_map_data (Result)
			Result.set_die_locations (cache.extended_die_locations_table)
		ensure
			result_exists: Result /= Void
		end

	cut_patterns_map (lot: RUN_PACKAGING_LOTS_ROW): CUT_PATTERNS_MAP
			-- Map that shows cut patterns for packaging lot
			-- `lot' (or cut patterns not assigned to a packaging
			-- lot if `lot' is Void)
		local
			cut_patterns: LIST [RUN_CUT_PATTERNS_ROW]
			table: HASH_TABLE [RUN_CUT_PATTERNS_ROW, INTEGER]
			process: WAFER_PROCESSES_ROW
		do
			create Result.make (session.run_id, connection.system_date)
			set_detailed_payload_map_data (Result)
			cut_patterns := cache.cut_pattern_rows
			table := cut_patterns_table (cut_patterns)
			if lot /= Void then
				process := wafer_processes_table (cache.wafer_process_rows).item (lot.non_null_wafer_process)
				lot.set_process (process)
				lot.set_organization (cache.organizations_row_for_org_no (lot.packaging_org_no))
				lot.set_wafer_lots (wafer_lots_for_process (cache.wafer_lot_rows, lot.non_null_wafer_process))
			end
			cut_patterns.do_all (agent {RUN_CUT_PATTERNS_ROW}.clear_cuts_and_sites)
			cut_patterns.do_all (agent {RUN_CUT_PATTERNS_ROW}.set_packaging_lot (lot))
			cache.cut_location_rows.do_all (agent assign_cut (?, table))
			cache.cut_die_site_rows.do_all (agent assign_isolated_site (?, table))
			cut_patterns := cut_patterns_for_lot (lot, cut_patterns)
			Result.set_cut_patterns (cut_patterns)
			Result.set_packaging_lot (lot)
		ensure
			result_exists: Result /= Void
		end

feature {NONE} -- Implementation

	session: RC_RUN_CLOSING_SESSION
			-- Session context

	cache: RUN_CACHE
			-- Run-specific cache

	connection: DB_CONNECTION
			-- Database connection to use

	environment: RC_ENVIRONMENT
			-- Environment for run closing
		do
			Result := session.environment
		end

	file_generator: RC_FILE_GENERATOR
			-- Generator that produces file category and
			-- file objects

	set_wafer_map_data (map: WAFER_MAP)
			-- Fill in data required for `map' from cache
		require
			map_exists: map /= Void
		local
			first_lot: RUN_WAFER_LOTS_ROW
		do
			map.set_runs_row (cache.runs_row)
			map.set_litho_sizes_row (cache.litho_sizes_row)
			map.set_wafer_types_row (cache.wafer_types_row)
			map.set_field_names (cache.field_names)
			map.set_field_step_patterns_row (cache.field_step_patterns_row)
			
			-- FIXME: just use step pattern and offset for
			-- wafer lot 1 for now
			
			first_lot := cache.first_wafer_lot_row
			map.set_step_pattern_offset (first_lot.field_step_pattern_offset)
			map.set_step_locations (cache.field_step_location_rows (first_lot.field_step_pattern))
			map.set_wafer_radius (cache.wafer_radius)
			map.set_tweezer_radius (cache.tweezer_radius)
			map.set_wafer_primary_flat_box (cache.wafer_primary_flat_box)
			map.set_tweezer_primary_flat_box (cache.tweezer_primary_flat_box)
		end

	set_detailed_payload_map_data (map: DETAILED_PAYLOAD_MAP)
			-- Fill in data required for `map' from cache
		require
			map_exists: map /= Void
		do
			map.set_runs_row (cache.runs_row)
			map.set_payload_box (cache.payload_box)
			map.set_field_names (cache.field_names)
			map.set_die_locations (cache.extended_die_locations_table)
		end

	cut_patterns_for_lot (lot: RUN_PACKAGING_LOTS_ROW pats: LIST [RUN_CUT_PATTERNS_ROW]): LIST [RUN_CUT_PATTERNS_ROW]
			-- List of cut patterns in `pats' that are assigned
			-- to `lot' (or that are unassigned if `lot' is Void)
		require
			pats_not_void: pats /= Void
		local
			pat: RUN_CUT_PATTERNS_ROW
		do
			create {ARRAYED_LIST [RUN_CUT_PATTERNS_ROW]} Result.make (pats.count)
			from
				pats.start
			until
				pats.after
			loop
				pat := pats.item
				if pat.assigned_to_lot (lot) then
					Result.extend (pat)
				end
				pats.forth
			end
		ensure
			result_exists: Result /= Void
		end

	wafer_lots_for_process (lots: LIST [RUN_WAFER_LOTS_ROW] proc: STRING): LIST [RUN_WAFER_LOTS_ROW]
			-- Wafer lots patterns in list for process `proc'
		require
			lots_not_void: lots /= Void
			proc_not_void: proc /= Void
		local
			lot: RUN_WAFER_LOTS_ROW
		do
			create {ARRAYED_LIST [RUN_WAFER_LOTS_ROW]} Result.make (lots.count)
			from
				lots.start
			until
				lots.after
			loop
				lot := lots.item
				if lot.wafer_process.is_equal (proc) then
					Result.extend (lot)
				end
				lots.forth
			end
		ensure
			result_exists: Result /= Void
		end

	cut_patterns_table (pats: LIST [RUN_CUT_PATTERNS_ROW]): HASH_TABLE [RUN_CUT_PATTERNS_ROW, INTEGER]
			-- Cut patterns `pats', indexed by cut pattern number
		require
			pats_not_void: pats /= Void
		local
			pat: RUN_CUT_PATTERNS_ROW
		do
			create Result.make (pats.count)
			from
				pats.start
			until
				pats.after
			loop
				pat := pats.item
				Result.put (pat, pat.cut_pattern_no)
				pats.forth
			end
		ensure
			result_exists: Result /= Void
		end

	wafer_processes_table (procs: LIST [WAFER_PROCESSES_ROW]): HASH_TABLE [WAFER_PROCESSES_ROW, STRING]
			-- Wafer processes `procs', indexed by wafer process
			-- name
		require
			procs_not_void: procs /= Void
		local
			proc: WAFER_PROCESSES_ROW
		do
			create Result.make (procs.count)
			from
				procs.start
			until
				procs.after
			loop
				proc := procs.item
				Result.put (proc, proc.wafer_process)
				procs.forth
			end
		ensure
			result_exists: Result /= Void
		end

	assign_cut (cut: CUT_LOCATIONS_ROW table: HASH_TABLE [RUN_CUT_PATTERNS_ROW, INTEGER])
			-- Assign cut to its cut pattern, using `table' to
			-- find the right one
		require
			cut_not_void: cut /= Void
			table_not_void: table /= Void
		do
			table.item (cut.cut_pattern_no).cuts.extend (cut)
		end

	assign_isolated_site (site: CUT_DIE_SITES_ROW table: HASH_TABLE [RUN_CUT_PATTERNS_ROW, INTEGER])
			-- Assign isolated site to its cut pattern,
			-- using `table' to find the right one
		require
			site_not_void: site /= Void
			table_not_void: table /= Void
		do
			table.item (site.cut_pattern_no).isolated_sites.put (site, site.mask_site_no)
		end

invariant
	session_exists: session /= Void
	cache_exists: cache /= Void
	generator_exists: file_generator /= Void
end
