
indexing
	
	description: "A primitive run closing step"
	author: "David Hollenberg"
	date: "April 16, 2003"

class RC_PRIMITIVE_STEP
inherit
	RC_STEP
		undefine
			is_equal
		redefine
			initialize
		end
	RUN_CLOSING_PRIMITIVE_STEPS_ROW
		rename
			primitive_step_name as step_name,
			is_null_primitive_step_name as is_null_step_name
		undefine
			is_equal
		end
	DB_QUERY_OPTIONS
		undefine
			is_equal
		end
	COMPARABLE

create
	make_template

feature -- Initialization

	initialize (sess: RC_RUN_CLOSING_SESSION)
			-- Initialize step that will be executed in context
			-- of `sess'
		do
			precursor (sess)
			create {ARRAYED_LIST [RUN_CLOSING_DATA_ITEMS_ROW]} input_data_items.make (5)
			create {ARRAYED_LIST [RUN_CLOSING_DATA_ITEMS_ROW]} applicability_data_items.make (5)
			create {ARRAYED_LIST [RUN_CLOSING_DATA_ITEMS_ROW]} output_data_items.make (5)
			create {ARRAYED_LIST [like Current]} required_steps.make (5)
			create {ARRAYED_LIST [like Current]} required_applicability_steps.make (5)
			create {ARRAYED_LIST [like Current]} dependent_steps.make (5)
		end

feature -- Properties

	is_primitive: BOOLEAN = True
			-- Is this a primitive step?
	
	type_name: STRING = "primitive"
			-- Name for step type (primitive or composite)
	
	input_data_items: LIST [RUN_CLOSING_DATA_ITEMS_ROW]
			-- Data items step requires before it can
			-- be executed

	applicability_data_items: LIST [RUN_CLOSING_DATA_ITEMS_ROW]
			-- Data items step requires before we can
			-- determine whether step applies to run

	output_data_items: LIST [RUN_CLOSING_DATA_ITEMS_ROW]
			-- Data items step produces

	required_steps: LIST [like Current]
			-- Steps that must be executed before this step
			-- (direct dependencies only)

	dependent_steps: LIST [like Current]
			-- Steps that depend on this step

	required_applicability_steps: LIST [like Current]
			-- Steps that must be executed before we can
			-- determine whether this step is applicable to
			-- this run (direct dependencies only)

	all_required_steps_done: BOOLEAN
			-- Are all steps that must be executed before this
			-- step done?
		do
			Result := True
			from
				required_steps.start
			until
				required_steps.after or not Result
			loop
				if not required_steps.item.is_done then
					Result := False
				end
				required_steps.forth
			end
		end

	all_applicability_steps_done: BOOLEAN
			-- Are all steps that must be executed before we
			-- can determine whether step is applicable done?
		do
			Result := True
			from
				required_applicability_steps.start
			until
				required_applicability_steps.after or not Result
			loop
				if not required_applicability_steps.item.is_done then
					Result := False
				end
				required_applicability_steps.forth
			end
		end

	max_required_step: like Current
			-- Step with maximum execution order of any steps that
			-- must be executed before this step, either
			-- directly or indirectly

	max_required_step_or_self: like Current
			-- Step with maximum execution order of any steps that
			-- must be executed before this step, either
			-- directly or indirectly, or `Current' if Void
		do
			if max_required_step /= Void then
				Result := max_required_step
			else
				Result := Current
			end
		end

	has_dependents: BOOLEAN
			-- Does this step have any steps that depend on it?
		do
			Result := not dependent_steps.is_empty
		end

	has_step_name (s: STRING): BOOLEAN
			-- Does this step have name `s'?
		require
			name_not_void: s /= Void
		do
			Result := step_name.is_equal (s)
		end

	component_steps: LIST [RC_STEP]
			-- Sub-steps of this step, if any
		do
			create {ARRAYED_LIST [RC_STEP]} Result.make (0)
		end

	primitive_steps: LIST [RC_PRIMITIVE_STEP]
			-- Primitive steps included in this step,
			-- either directly or indirectly, in order
		do
			create {ARRAYED_LIST [RC_PRIMITIVE_STEP]} Result.make (1)
			Result.extend (Current)
		end

	run_step: RUN_PRIMITIVE_STEPS_MROW
			-- Database status record for this step for
			-- run being closed.  Void if step has never
			-- been done or if it has been undone with
			-- no state saved
	
	modifiable_run_step (run_id: STRING): RUN_PRIMITIVE_STEPS_MROW
			-- Database status record for this step for `run_id'
			-- which is safe to modify
		do
			create Result.make_by_key (session.connection, Old_no_exception_row, run_id, step_name)
			if Result.is_empty then
				create Result.make_by_key (session.connection, New_row, run_id, step_name)
			end
		ensure
			result_exists: Result /= Void
		end
	
	was_applicable_to_run: BOOLEAN
			-- Was step applicable to this run?
		do
			Result := run_step.is_applicable
		end

	is_done: BOOLEAN
			-- Has step been fully completed?
		do
			Result := run_step /= Void and then (not run_step.is_null_date_completed and run_step.is_null_date_undo_started)
		end
	
	is_undone: BOOLEAN
			-- Is step fully uncompleted (not even started)?
		do
			Result := run_step = Void or else not run_step.is_null_date_undo_completed
		end

	is_soft_undone: BOOLEAN
			-- Is step fully uncompleted, but was previously
			-- done and then undone with a "soft undo",
			-- possibly saving some state?
		do
			Result := is_undone and run_step /= Void
		end

	is_partially_done: BOOLEAN
			-- Is step partially completed?
		do
			Result := run_step /= Void and then (run_step.is_null_date_completed and run_step.is_null_date_undo_started)
		end

	is_partially_undone: BOOLEAN
			-- Has step been partially undone?
		do
			Result := run_step /= Void and then (not run_step.is_null_date_undo_started and run_step.is_null_date_undo_completed)
		end

	has_saved_state: BOOLEAN
			-- Does step have saved state from a previous
			-- execution?
		do
			Result := run_step /= Void and then not run_step.is_null_saved_state_filename
		end

	infix "<" (other: like Current): BOOLEAN
		do
			Result := execution_order < other.execution_order
		end

feature -- Modification

	set_max_required_step (s: RC_PRIMITIVE_STEP)
			-- Set `max_required_step' to `s'
		do
			max_required_step := s
		ensure
			value_set: max_required_step = s
		end
			
	set_run_step (s: RUN_PRIMITIVE_STEPS_MROW)
			-- Set `run_step' to `s'
		do
			run_step := s
		ensure
			step_set: run_step = s
		end

end
