
indexing
	
	description: "A run closing session for closing existing run"
	author: "David Hollenberg"
	date: "September 25, 2002"

deferred class RC_RUN_CLOSING_SESSION
inherit
	RC_SESSION
		rename
			make as rc_session_make
		export
			{NONE} rc_session_make
		end
	RC_ABSTRACT_PACKING_SESSION
		rename
			make as rc_session_make
		export
			{NONE} rc_session_make
		end
	DB_QUERY_OPTIONS
		export
			{NONE} all
		end
	RC_UTILITIES_ACCESS
		export
			{NONE} all
		end
	FILE_UTILITIES_ACCESS
		export
			{NONE} all
		end
	ITEM_COUNT_RANGES_ACCESS
		export
			{NONE} all
		end

	WAFER_TECHN_FAMILY_ACCESS

feature {NONE} -- Creation

	make (conn: DB_CONNECTION r: STRING a_log_file: LARGE_PLAIN_TEXT_FILE)
		require
			connection_exists: conn /= Void
			connection_open: conn.is_open
			run_id_exists: r /= Void
			a_log_file_exists: a_log_file /= Void
		do
			connection := conn
			set_run_id (r)
			log_file := a_log_file
			create commit_actions
			create cache.make (run_id, connection, file_generator)
		end

feature -- Creation

	make_run_session (sess: RC_SESSION a_run_id: STRING)
			-- Create writable session from `sess'
		require
			session_exists: sess /= Void
			a_run_id_exists: a_run_id /= Void
		do
			connection := sess.connection
			factory := sess.factory
			application := sess.application
			log_file := sess.log_file
			is_read_only := False
			create commit_actions
			set_run_id (a_run_id)
			create cache.make (run_id, connection, file_generator)
		end

feature -- Properties

	type: STRING
			-- Type of session
		deferred
		end

	id_name: STRING = "Run ID"
			-- Name of session-identifying value
	
	id_value: STRING
			-- Value that identifies session
		do
			Result := run_id
		end

	has_steps: BOOLEAN = True
			-- Is session activity broken into separate steps?
			-- (Answer: yes)

	run_id: STRING
			-- Run id of run we're working on, in upper case

	lower_run_id: STRING
			-- Run id of run we're working on, in lower case

	is_trial_run: BOOLEAN
			-- Is this a session for closing a trial run?
		deferred
		end

	is_standard_run_id (s: STRING): BOOLEAN
			-- Is `s' a standard run id for this session?
		require
			string_exists: s /= Void
		deferred
		end

	run_reality_type: STRING
			-- Text describing whether run is real
		deferred
		ensure
			result_exists: Result /= Void
		end

	run_directory: STRING
			-- Name of run directory
		deferred
		end

	probing_run_directory: STRING
			-- Name of run directory for probing purposes
		deferred
		end

	title_with_run_id (s: STRING): STRING
			-- Title `s' with run id information added to it
			-- in standard way
		require
			title_exists: s /= Void
		do
			Result := s + " for " + run_id
		end

	log_file_name: STRING
			-- Name of log file
		do
			Result := environment.log_name_for_run (run_id)
		end
	
	error_file_name: STRING
			-- Name of file for error messages
		do
			Result := environment.error_name_for_run (run_id)
		end
	
	shared_lock_name: STRING
			-- Name of shared lock file
		do
			Result := environment.shared_lock_name
		end

	exclusive_lock_name: STRING
			-- Name of exclusive lock file
		do
			Result := environment.lock_name_for_run (run_id)
		end

	is_read_only: BOOLEAN
			-- Is session read-only (no modification to
			-- run allowed)?

	file_generator: RC_FILE_GENERATOR
			-- Object for generating file and file category
			-- objects for run

	cache: RUN_CACHE
			-- Run-specific cache

	cache_auto_flush_time: INTEGER
			-- Flush run cache automatically if there has been
			-- no step activity (do, undo, redo) before this
			-- time in seconds since beginning of epoch,
			-- but only if value is greater than 0 and auto
			-- flushing is enabled

	primitive_steps: LIST [RC_PRIMITIVE_STEP]
			-- Primitive run closing steps

	composite_steps: LIST [RC_COMPOSITE_STEP]
			-- Composite run closing steps

	forward_steps (step: RC_STEP): LIST [RC_PRIMITIVE_STEP]
			-- List of primitive steps to be done, in order,
			-- to do `step'.  Each step is marked as
			-- being implicitly or explicitly selected
		require
			step_exists: step /= Void
		local
			explicit_steps: LIST [RC_PRIMITIVE_STEP]
			next_step: RC_PRIMITIVE_STEP
		do
			create {ARRAYED_LIST [RC_PRIMITIVE_STEP]} Result.make (primitive_steps.count)
			unselect (primitive_steps)
			explicit_steps := done_steps_removed (step.primitive_steps)
			select_explicit (explicit_steps)
			from
				explicit_steps.start
			until
				explicit_steps.after
			loop
				next_step := explicit_steps.item
				create needed_steps.make
				compute_required_steps (next_step)
				Result.append (needed_steps)
				Result.extend (next_step)
				explicit_steps.forth
			end
		ensure
			result_exists: Result /= Void
		end

	backward_steps (step: RC_STEP save_state: BOOLEAN): LIST [RC_PRIMITIVE_STEP]
			-- List of primitive steps to be undone, in
			-- order, to undo `step'.  `save_state'
			-- indicates whether state should be saved
			-- before undoing step.  Each step is marked
			-- as being implicitly or explicitly selected
		require
			step_exists: step /= Void
		local
			explicit_steps: LIST [RC_PRIMITIVE_STEP]
			next_step: RC_PRIMITIVE_STEP
		do
			create {ARRAYED_LIST [RC_PRIMITIVE_STEP]} Result.make (primitive_steps.count)
			unselect (primitive_steps)
			explicit_steps := undone_steps_removed (step.primitive_steps, save_state)
			select_explicit (explicit_steps)
			from
				explicit_steps.finish
			until
				explicit_steps.before
			loop
				next_step := explicit_steps.item
				create needed_steps.make
				compute_dependent_steps (next_step)
				from
					needed_steps.finish
				until
					needed_steps.before
				loop
					Result.extend (needed_steps.item)
					needed_steps.back
				end
				Result.extend (next_step)
				explicit_steps.back
			end
		ensure
			result_exists: Result /= Void
		end

	step_list_implementation (list: LIST [RC_PRIMITIVE_STEP]): RC_PRIMITIVE_STEPS_OP
			-- Implementation for `list'
		require
			list_exists: list /= Void
		do
			create {RC_PRIMITIVE_STEPS_OP} Result.make (list, Current)
		ensure
			result_exists: Result /= Void
		end

	primitive_step_implementation (step: RC_PRIMITIVE_STEP): RC_PRIMITIVE_STEP_OP
			-- New object which is implementation for
			-- primitive step `step' or Void if implementation
			-- is not the same for both real and trial runs
		require
			step_exists: step /= Void
		local
			name: STRING
			v: PRIMITIVE_STEP_NAME_VALUES
		do
			name := step.step_name
			if equal (name, v.Set_closer) then
				create {SET_CLOSER_OP} Result.make (step, Current)
			elseif equal (name, v.Set_clerk) then
				create {SET_CLERK_OP} Result.make (step, Current)
			elseif equal (name, v.Set_design_engineer) then
				create {SET_DESIGN_ENGINEER_OP} Result.make (step, Current)
			elseif equal (name, v.Set_test_engineer) then
				create {SET_TEST_ENGINEER_OP} Result.make (step, Current)
			elseif equal (name, v.Set_run_type) then
				create {SET_RUN_TYPE_OP} Result.make (step, Current)
			elseif equal (name, v.Set_overhead_run_id) then
				create {SET_OVERHEAD_RUN_ID_OP} Result.make (step, Current)
			elseif equal (name, v.Set_taxi_status) then
				create {SET_TAXI_STATUS_OP} Result.make (step, Current)
			elseif equal (name, v.Set_educational_status) then
				create {SET_EDUCATIONAL_STATUS_OP} Result.make (step, Current)
			elseif equal (name, v.Set_wafer_techn) then
				create {SET_WAFER_TECHN_OP} Result.make (step, Current)
			elseif equal (name, v.Set_mask_techn) then
				create {SET_MASK_TECHN_OP} Result.make (step, Current)
			elseif equal (name, v.Set_field_set) then
				create {SET_FIELD_SET_OP} Result.make (step, Current)
			elseif equal (name, v.Set_close_date) then
				create {SET_RUN_CLOSE_DATE_OP} Result.make (step, Current)
			elseif equal (name, v.Set_sched_run_no) then
				create {SET_SCHED_RUN_NO_OP} Result.make (step, Current)
			elseif equal (name, v.Set_die_overhead_parameters) then
				create {SET_DIE_OVERHEAD_PARAMETERS_OP} Result.make (step, Current)
			elseif equal (name, v.Set_vendor_part_sequence) then
				create {SET_VENDOR_PART_SEQUENCE_OP} Result.make (step, Current)
			elseif equal (name, v.Set_wafer_part_id_source) then
				create {SET_WAFER_PART_ID_SOURCE_OP} Result.make (step, Current)
			elseif equal (name, v.Set_wafer_part_id) then
				create {SET_WAFER_PART_ID_OP} Result.make (step, Current)
			elseif equal (name, v.Set_mask_part_id) then
				create {SET_MASK_PART_ID_OP} Result.make (step, Current)
			elseif equal (name, v.Assign_packing) then
				create {ASSIGN_PACKING_OP} Result.make (step, Current)
			elseif equal (name, v.Select_and_pack_projects) then
				create {SELECT_AND_PACK_PROJECTS_OP} Result.make (step, Current)
			elseif equal (name, v.Set_wafer_processes) then
				create {SET_WAFER_PROCESSES_OP} Result.make (step, Current)
			elseif equal (name, v.Assign_die_ids) then
				create {ASSIGN_DIE_IDS_OP} Result.make (step, Current)
			elseif equal (name, v.Create_field_step_pattern) then
				create {CREATE_FIELD_STEP_PATTERN_OP} Result.make (step, Current)
			elseif equal (name, v.Create_initial_wafer_lots) then
				create {CREATE_INITIAL_WAFER_LOTS_OP} Result.make (step, Current)
			elseif equal (name, v.Create_initial_packaging_lots) then
				create {CREATE_INITIAL_PACKAGING_LOTS_OP} Result.make (step, Current)
			elseif equal (name, v.Generate_probing_files) then
				create {GENERATE_PROBING_FILES_OP} Result.make (step, Current)
			elseif equal (name, v.Gen_cut_patterns_map) then
				create {GEN_CUT_PATTERNS_MAP_OP} Result.make (step, Current)
			elseif equal (name, v.Gen_design_names_map) then
				create {GEN_DESIGN_NAMES_MAP_OP} Result.make (step, Current)
			elseif equal (name, v.Gen_design_numbers_map) then
				create {GEN_DESIGN_NUMBERS_MAP_OP} Result.make (step, Current)
			elseif equal (name, v.Gen_detailed_step_pattern_map) then
				create {GEN_DETAILED_STEP_PATTERN_MAP_OP} Result.make (step, Current)
			elseif equal (name, v.Gen_organizations_map) then
				create {GEN_ORGANIZATIONS_MAP_OP} Result.make (step, Current)
			elseif equal (name, v.Gen_site_numbers_map) then
				create {GEN_SITE_NUMBERS_MAP_OP} Result.make (step, Current)
			elseif equal (name, v.Gen_step_pattern_map) then
				create {GEN_STEP_PATTERN_MAP_OP} Result.make (step, Current)
			elseif equal (name, v.Gen_project_summary) then
				create {GEN_PROJECT_SUMMARY_OP} Result.make (step, Current)
			elseif equal (name, v.Create_merge_job_file) then
				create {CREATE_MERGE_JOB_FILE_OP} Result.make (step, Current)
			elseif equal (name, v.Create_final_job_file) then
				create {CREATE_FINAL_JOB_FILE_OP} Result.make (step, Current)
			elseif equal (name, v.Create_test_vehicle_fabs) then
				create {CREATE_TEST_VEHICLE_FABS_OP} Result.make (step, Current)
			elseif equal (name, v.Create_test_vehicle_orders) then
				create {CREATE_TEST_VEHICLE_ORDERS_OP} Result.make (step, Current)
			elseif equal (name, v.Setup_packaging) then
				create {SETUP_PACKAGING_OP} Result.make (step, Current)
			elseif equal (name, v.Check_unresolved_issues) then
				create {CHECK_UNRESOLVED_ISSUES_OP} Result.make (step, Current)
			elseif equal (name, v.Create_production_job) then
				create {CREATE_PRODUCTION_JOB_OP} Result.make (step, Current)
			elseif equal (name, v.Adjust_charges) then
				create {ADJUST_CHARGES_OP} Result.make (step, Current)
			elseif equal (name, v.Set_designated_design_techn) then
				create {SET_DESIGNATED_DESIGN_TECHN_OP} Result.make (step, Current)
			elseif equal (name, v.Gen_wafer_techn_survey) then
				create {GEN_WAFER_TECHN_SURVEY_OP} Result.make (step, Current)
			elseif equal (name, v.Gen_die_fracture_scripts) then
				create {GEN_DIE_FRACTURE_SCRIPTS_OP} Result.make (step, Current)
			elseif equal (name, v.Gen_dynamic_vend_geom_files) then
				create {GEN_DYNAMIC_VEND_GEOM_FILES_OP} Result.make (step, Current)
			elseif equal (name, v.Gen_scribe_files) then
				create {GEN_SCRIBE_FILES_OP} Result.make (step, Current)
			elseif equal (name, v.Gen_filler_files) then
				create {GEN_FILLER_FILES_OP} Result.make (step, Current)
			elseif equal (name, v.Gen_die_overhead_files) then
				create {GEN_DIE_OVERHEAD_FILES_OP} Result.make (step, Current)
			elseif equal (name, v.Convert_scmos_layout) then
				create {CONVERT_SCMOS_LAYOUT_OP} Result.make (step, Current)
			elseif equal (name, v.Gen_merged_gds_files) then
				create {GEN_MERGED_GDS_FILES_OP} Result.make (step, Current)
			elseif equal (name, v.Generate_mebes) then
				create {GENERATE_MEBES_OP} Result.make (step, Current)
			elseif equal (name, v.Merge_mebes) then
				create {MERGE_MEBES_OP} Result.make (step, Current)
			elseif equal (name, v.Gen_fill_exclusion) then
				create {GEN_FILL_EXCLUSION_OP} Result.make (step, Current)
			elseif equal (name, v.Gen_fill_area_selector) then
				create {GEN_FILL_AREA_SELECTOR_OP} Result.make (step, Current)
			elseif equal (name, v.Fill_mebes) then
				create {FILL_MEBES_OP} Result.make (step, Current)
			elseif equal (name, v.Send_density_report) then
				create {SEND_DENSITY_REPORT_OP} Result.make (step, Current)
			elseif equal (name, v.Rename_mebes) then
				create {RENAME_MEBES_OP} Result.make (step, Current)
			elseif equal (name, v.Complete_run_assembly) then
				create {COMPLETE_RUN_ASSEMBLY_OP} Result.make (step, Current)
			elseif equal (name, v.Create_mask_shop_files) then
				create {CREATE_MASK_SHOP_FILES_OP} Result.make (step, Current)
			elseif equal (name, v.Verify_sales_orders) then
				create {VERIFY_SALES_ORDERS_OP} Result.make (step, Current)
			elseif equal (name, v.Fill_vendor_forms) then
				create {FILL_VENDOR_FORMS_OP} Result.make (step, Current)
			elseif equal (name, v.Gen_foundry_eccn_report) then
				create {GEN_FOUNDRY_ECCN_REPORT_OP} Result.make (step, Current)
			elseif equal (name, v.Send_foundry_eccn_report) then
				create {SEND_FOUNDRY_ECCN_REPORT_OP} Result.make (step, Current)
			elseif equal (name, v.Transfer_mask_shop_files) then
				create {TRANSFER_MASK_SHOP_FILES_OP} Result.make (step, Current)
			elseif equal (name, v.Notify_mask_shop) then
				create {NOTIFY_MASK_SHOP_OP} Result.make (step, Current)
			elseif equal (name, v.Build_fab_run_order) then
				create {BUILD_FAB_RUN_ORDER_OP} Result.make (step, Current)
			elseif equal (name, v.Set_turnaround_dates) then
				create {SET_TURNAROUND_DATES_OP} Result.make (step, Current)
			elseif equal (name, v.Create_fab_run_order) then
				create {CREATE_FAB_RUN_ORDER_OP} Result.make (step, Current)
			elseif equal (name, v.Assign_cut_patterns) then
				create {ASSIGN_CUT_PATTERNS_OP} Result.make (step, Current)
			elseif equal (name, v.Assign_packaging_line_items) then
				create {ASSIGN_PACKAGING_LINE_ITEMS_OP} Result.make (step, Current)
			elseif equal (name, v.Post_run) then
				create {POST_RUN_OP} Result.make (step, Current)
			elseif equal (name, v.Gen_public_status) then
				create {GEN_PUBLIC_STATUS_OP} Result.make (step, Current)
			elseif equal (name, v.Send_submission_messages) then
				create {SEND_SUBMISSION_MESSAGES_OP} Result.make (step, Current)
			elseif equal (name, v.Lock_production_charges) then
				create {LOCK_PRODUCTION_CHARGES_OP} Result.make (step, Current)
			elseif equal (name, v.Run_accounting_reports) then
				create {RUN_ACCOUNTING_REPORTS_OP} Result.make (step, Current)
			elseif equal (name, v.Release_to_manufacturing) then
				create {RELEASE_TO_MANUFACTURING_OP} Result.make (step, Current)
			elseif equal (name, v.Send_rtm_messages) then
				create {SEND_RTM_MESSAGES_OP} Result.make (step, Current)
			elseif equal (name, v.Prepare_final_bond_diagrams) then
				create {PREPARE_FINAL_BOND_DIAGRAMS_OP} Result.make (step, Current)
			end
		end

	next_state_file (step: RC_PRIMITIVE_STEP): RC_FILE
			-- Next state file for `step'
		require
			step_exists: step /= Void
		do
			Result := file_generator.state_file (step.step_name, step.run_step.saved_undo_count + 1)
		ensure
			result_exists: Result /= Void
		end

feature -- Modification

	set_run_id (s: STRING)
			-- Set `run_id' to `s'
		do
			run_id := s.as_upper
			lower_run_id := s.as_lower
			create file_generator.make (run_directory, probing_run_directory, lower_run_id)
		ensure then
			lower_run_id_set: equal (lower_run_id, s.as_lower)
		end

	set_read_only
			-- Set `is_read_only' to True
		do
			is_read_only := True
		ensure
			read_only: is_read_only
		end

	set_primitive_steps (list: LIST [RC_PRIMITIVE_STEP])
			-- Set `primitive_steps' to `list'
		require
			list_exists: list /= Void
		do
			primitive_steps := list
		ensure
			value_set: primitive_steps = list
		end

	set_composite_steps (list: LIST [RC_COMPOSITE_STEP])
			-- Set `composite_steps' to `list'
		require
			list_exists: list /= Void
		do
			composite_steps := list
		ensure
			value_set: composite_steps = list
		end

	set_primitive_step_status
			-- Set the status of each primitive step by
			-- setting it's `run_step'
		local
			list: LIST [RUN_PRIMITIVE_STEPS_MROW]
			run_steps: HASH_TABLE [RUN_PRIMITIVE_STEPS_MROW, STRING]
			row: RUN_PRIMITIVE_STEPS_MROW
			step: RC_PRIMITIVE_STEP
		do
			create row.make_template
			list := row.steps_for_run (run_id, connection, expected_count.At_least_zero)
			create run_steps.make (list.count)
			from
				list.start
			until
				list.after
			loop
				row := list.item
				run_steps.put (row, row.primitive_step_name)
				list.forth
			end

			from
				primitive_steps.start
			until
				primitive_steps.after
			loop
				step := primitive_steps.item
				step.set_run_step (run_steps.item (step.step_name))
				primitive_steps.forth
			end
		end

	start_applicable_step (op: RC_PRIMITIVE_STEP_OP)
			-- Take actions needed when starting applicable
			-- step operation `op'
		require
			op_exists: op /= Void
		local
			row: RUN_PRIMITIVE_STEPS_MROW
			step: RC_PRIMITIVE_STEP
		do
			step := op.step
			start_step (step, True)
			if op.creates_external_state then
				row := step.modifiable_run_step (run_id)
				if row.is_old then
					lock_row (row)
				else
					row.set_saved_undo_count (highest_used_undo_version (step))
				end
				row.set_is_applicable (True)
				row.set_date_started (date_started)
				row.set_null_date_completed
				row.set_null_date_undo_started
				row.set_null_date_undo_completed
				row.set_username (connection.user_name.as_upper)
				commit_and_update_step (step, row)
				publish_step_changed_state (step)
			end
		end

	start_inapplicable_step (op: RC_PRIMITIVE_STEP_OP)
			-- Take actions needed when starting inapplicable
			-- step operation `op'
		require
			op_exists: op /= Void
		do
			start_step (op.step, False)
		end

	start_soft_undo_step (op: RC_PRIMITIVE_STEP_OP applicable: BOOLEAN)
			-- Take actions needed when starting soft undo
			-- of step operation `op', where `applicable'
			-- indicates whether step applied to run
		require
			op_exists: op /= Void
		local
			row: RUN_PRIMITIVE_STEPS_MROW
			step: RC_PRIMITIVE_STEP
		do
			step := op.step
			if op.is_currently_undoable then
				start_undo_step (step, True, applicable)
				row := step.modifiable_run_step (run_id)
				lock_row (row)
				row.set_date_undo_started (date_started)
				row.set_null_date_undo_completed
				row.set_null_saved_state_filename
				row.set_username (connection.user_name.as_upper)
				commit_and_update_step (step, row)
				publish_step_changed_state (step)
			else
				report_undo_not_allowed (op)
			end
		end

	report_undo_not_allowed (op: RC_PRIMITIVE_STEP_OP)
			-- Report to operator that undo is not allowed,
			-- including reason if available
		require
			op_exists: op /= Void
		local
			step: RC_PRIMITIVE_STEP
			reason: STRING
		do
			step := op.step
			reason := op.not_undoable_reason
			if not reason.is_empty then
				reason := "%Nbecause " + reason
			end
			op.inform ("Step %"" + step.presentation_text + "%" cannot be undone" + reason)
			op.abort
		end

	start_hard_undo_step (op: RC_PRIMITIVE_STEP_OP applicable: BOOLEAN)
			-- Take actions needed when starting hard undo
			-- of step operation `op', where `applicable'
			-- indicates whether step applied to run
		require
			op_exists: op /= Void
		local
			step: RC_PRIMITIVE_STEP
			row: RUN_PRIMITIVE_STEPS_MROW
			fname: STRING
			file: RC_FILE
		do
			step := op.step
			if op.is_currently_undoable then
				start_undo_step (step, False, applicable)
				row := step.run_step
				if not row.is_null_date_undo_started then
					file := next_state_file (step)
					fname := file.file_pathname
					if path_exists (fname) then
						delete_file (fname)
					end
				end
				if op.creates_external_state then
					row := step.modifiable_run_step (run_id)
					lock_row (row)
					row.set_date_undo_started (date_started)
					row.set_null_date_undo_completed
					row.set_null_saved_state_filename
					row.set_username (connection.user_name.as_upper)
					commit_and_update_step (step, row)
					publish_step_changed_state (step)
				end
			else
				report_undo_not_allowed (op)
			end
		end

	finish_applicable_step (op: RC_PRIMITIVE_STEP_OP)
			-- Take actions needed when finishing applicable
			-- step operation `op'
		require
			op_exists: op /= Void
		do
			finish_step (op, True)
		end

	finish_inapplicable_step (op: RC_PRIMITIVE_STEP_OP)
			-- Take actions needed when finishing inapplicable
			-- step operation `op'
		require
			op_exists: op /= Void
		do
			finish_step (op, False)
		end

	finish_soft_undo_step (op: RC_PRIMITIVE_STEP_OP applicable: BOOLEAN)
			-- Take actions needed when finishing soft undo
			-- of step operation `op', where `applicable'
			-- indicates whether step applied to run
		require
			op_exists: op /= Void
		local
			row: RUN_PRIMITIVE_STEPS_MROW
			fname: STRING
			file: RC_FILE
			step: RC_PRIMITIVE_STEP
			saved_state: BOOLEAN
		do
			step := op.step
			saved_state := op.state /= Void
			if saved_state then
				file := next_state_file (step)
				fname := file.file_pathname
			end
			row := step.modifiable_run_step (run_id)
			lock_row (row)
			row.set_date_undo_completed (connection.system_date)
			if saved_state then
				row.set_saved_state_filename (fname)
				row.set_saved_undo_count (row.saved_undo_count + 1)
			elseif row.is_null_saved_undo_count then
				row.set_saved_undo_count (0)
			end
			row.set_username (connection.user_name.as_upper)
			commit_and_update_step (step, row)
			publish_step_changed_state (step)
			finish_undo_step (step, True, applicable)
		end

	finish_hard_undo_step (op: RC_PRIMITIVE_STEP_OP applicable: BOOLEAN)
			-- Take actions needed when finishing hard undo
			-- of step operation `op', where `applicable'
			-- indicates whether step applied to run
		require
			op_exists: op /= Void
		local
			row: RUN_PRIMITIVE_STEPS_MROW
			step: RC_PRIMITIVE_STEP
		do
			step := op.step
			row := step.modifiable_run_step (run_id)
			lock_row (row)
			row.delete
			commit_and_update_step (step, Void)
			publish_step_changed_state (step)
			finish_undo_step (step, False, applicable)
		end

	reenter_completed_step (step: RC_PRIMITIVE_STEP)
			-- Take actions needed to re-enter modifiable
			-- step `step' that is already completed
		require
			step_exists: step /= Void
		local
			row: RUN_PRIMITIVE_STEPS_MROW
		do
			set_role (step.database_role)
			row := step.modifiable_run_step (run_id)
			lock_row (row)
			row.set_null_date_completed
			commit_and_update_step (step, row)
			publish_step_changed_state (step)
			publish_step_reentered (step)
			set_default_role
		end

	abort_step (step: RC_PRIMITIVE_STEP)
			-- Take actions needed when `step' is aborted
		require
			step_exists: step /= Void
		do
			cleanup_for_abort
			publish_step_aborted (step)
		end

	abort_soft_undo_step (step: RC_PRIMITIVE_STEP)
			-- Take actions needed when undo of `step' is aborted
		require
			step_exists: step /= Void
		do
			cleanup_for_abort
			publish_step_undo_aborted (step, True)
		end

	abort_hard_undo_step (step: RC_PRIMITIVE_STEP)
			-- Take actions needed when hard of `step' is aborted
		require
			step_exists: step /= Void
		do
			cleanup_for_abort
			publish_step_undo_aborted (step, False)
		end

feature -- Operations

	check_valid_run_id (s: STRING)
			-- Check that `s' is a possibly valid run id
		require
			string_exists: s /= Void
		local
			run_checker: RUN_ID_CHECK_OP
		do
			create run_checker
			run_checker.check_run_id (run_id, is_trial_run)
		end
			
	ensure_directory_exists (cat: RC_FILE_CATEGORY)
			-- Ensure directory for `cat' exists.  Create
			-- it if not there
		require
			category_exists: cat /= Void
		local
			dir_name: STRING
			old_mask: INTEGER
			mask_changed: BOOLEAN
		do
			dir_name := cat.directory_pathname
			if not path_exists (dir_name) then
				old_mask := unix_os.set_file_creation_mask (fu.permission_string_to_mask (cat.directory_permission_string).bit_not)
				mask_changed := True
				create_directory (dir_name)
				change_group (dir_name, cat.group_name)
				old_mask := unix_os.set_file_creation_mask (old_mask)
			end
		rescue
			if mask_changed then
				old_mask := unix_os.set_file_creation_mask (old_mask)
			end
		end
			
	open_file_for_write (f: LARGE_FILE file: RC_FILE)
			-- Open `f' for write.  Create directory
			-- containing it if not there
		require
			file_exists: f /= Void
			rc_file_exists: file /= Void
		local
			old_mask: INTEGER
			mask_changed: BOOLEAN
			file_opener: FILE_OPEN_WRITE_OP
		do
			ensure_directory_exists (file)
			old_mask := unix_os.set_file_creation_mask (fu.permission_string_to_mask (file.file_permission_string).bit_not)
			mask_changed := True
			create file_opener
			file_opener.open (f)
			change_group (file.file_pathname, file.group_name)
			old_mask := unix_os.set_file_creation_mask (old_mask)
		rescue
			if mask_changed then
				old_mask := unix_os.set_file_creation_mask (old_mask)
			end
		end
			
	set_cache_auto_flush_time
			-- Set time for next auto-flush of cache, based
			-- on current time
		do
			cache_auto_flush_time := unix_os.current_time_in_seconds + environment.cache_auto_flush_interval
		end
			
	clear_cache_auto_flush_time
			-- Clear time for next auto-flush of cache
		do
			cache_auto_flush_time := 0
		end
			
	auto_flush_cache_if_needed
			-- Flush run closing cache if auto-flush requested,
			-- there is non-zero flush time and the current
			-- time is after the flush time
		do
			if preferences.is_cache_auto_flush_enabled and then cache_auto_flush_time /= 0 and cache_auto_flush_time <= unix_os.current_time_in_seconds then
				cache.uncache_all
			end
		end
			
feature {NONE} -- Implementation

	date_started: DB_DATE
			-- Date last primitive step do or undo started

	needed_steps: SORTED_TWO_WAY_LIST [RC_PRIMITIVE_STEP]
			-- Steps that need to be done or undone to
			-- do or undo a (composite or primitive) step,
			-- in order

	done_steps_removed (list: LIST [RC_PRIMITIVE_STEP]): like list
			-- `list' with all steps that are done removed
		local
			step: RC_PRIMITIVE_STEP
		do
			Result := list.twin
			Result.wipe_out
			from
				list.start
			until
				list.after
			loop
				step := list.item
				if not step.is_done then
					Result.extend (step)
				end
				list.forth
			end
		end

	undone_steps_removed (list: LIST [RC_PRIMITIVE_STEP] save_state: BOOLEAN): like list
			-- `list' with all steps removed that are undone.
			-- If `save_state' is False, keep steps that are
			-- undone, but have saved state
		local
			step: RC_PRIMITIVE_STEP
		do
			Result := list.twin
			Result.wipe_out
			from
				list.start
			until
				list.after
			loop
				step := list.item
				if not step.is_undone or (not save_state and step.is_soft_undone) then
					Result.extend (step)
				end
				list.forth
			end
		end

	unselect (list: LIST [RC_STEP])
			-- Unselect all steps in `list'
		local
			step: RC_STEP
		do
			from
				list.start
			until
				list.after
			loop
				step := list.item
				step.unselect
				list.forth
			end
		end

	select_explicit (list: LIST [RC_STEP])
			-- Set all steps in `list' to be explicitly selected
		local
			step: RC_STEP
		do
			from
				list.start
			until
				list.after
			loop
				step := list.item
				step.select_explicit
				list.forth
			end
		end

	compute_required_steps (step: RC_PRIMITIVE_STEP)
			-- Add to `needed_steps' all primitive steps
			-- required by `step' that are not done,
			-- in order they should be executed
		require
			step_exists: step /= Void
		local
			list: LIST [RC_PRIMITIVE_STEP]
			s: RC_PRIMITIVE_STEP
		do
			list := step.required_steps
			from
				list.start
			until
				list.after
			loop
				s := list.item
				if not s.is_selected and not s.is_done then
					compute_required_steps (s)
					add_needed_step (s)
				end
				list.forth
			end
		end

	compute_dependent_steps (step: RC_PRIMITIVE_STEP)
			-- Add to `needed_steps' all primitive steps
			-- that depend on `step' that are not undone,
			-- in order they should be undone
		require
			step_exists: step /= Void
		local
			list: LIST [RC_PRIMITIVE_STEP]
			s: RC_PRIMITIVE_STEP
		do
			list := step.dependent_steps
			from
				list.start
			until
				list.after
			loop
				s := list.item
				if not s.is_selected and not s.is_undone then
					compute_dependent_steps (s)
					add_needed_step (s)
				end
				list.forth
			end
		end

	add_needed_step (step: RC_PRIMITIVE_STEP)
			-- Add `step' to `needed_steps' and mark
			-- it as implicitly selected
		require
			step_exists: step /= Void
		do
			step.select_implicit
			needed_steps.extend (step)
		end

	start_step (step: RC_PRIMITIVE_STEP applicable: BOOLEAN)
			-- Take actions needed when starting `step'
		require
			step_exists: step /= Void
		do
			publish_step_started (step, applicable)
			set_date_started
			set_role (step.database_role)
		end

	finish_step (op: RC_PRIMITIVE_STEP_OP applicable: BOOLEAN)
			-- Take actions needed when finishing step
			-- operation `op'
		require
			op_exists: op /= Void
		local
			row: RUN_PRIMITIVE_STEPS_MROW
			step: RC_PRIMITIVE_STEP
		do
			step := op.step
			row := step.modifiable_run_step (run_id)
			if row.is_old then
				lock_row (row)
			else
				row.set_saved_undo_count (highest_used_undo_version (step))
			end
			row.set_is_applicable (applicable)
			if not op.creates_external_state or not applicable then
				row.set_date_started (date_started)
			end
			row.set_date_completed (connection.system_date)
			row.set_null_date_undo_started
			row.set_null_date_undo_completed
			row.set_null_saved_state_filename
			row.set_username (connection.user_name.as_upper)
			commit_and_update_step (step, row)
			publish_step_changed_state (step)
			publish_step_ended (step, applicable)
			set_default_role
			date_started := Void
		end

	start_undo_step (step: RC_PRIMITIVE_STEP save_state, applicable: BOOLEAN)
			-- Take actions needed when starting to undo `step'
		require
			step_exists: step /= Void
		do
			publish_step_undo_started (step, save_state, applicable)
			set_date_started
			set_role (step.database_role)
		end

	finish_undo_step (step: RC_PRIMITIVE_STEP save_state, applicable: BOOLEAN)
			-- Take actions needed when finishing undo of `step'
		require
			step_exists: step /= Void
		do
			publish_step_undo_ended (step, save_state, applicable)
			set_default_role
			date_started := Void
		end

	commit_and_update_step (step: RC_PRIMITIVE_STEP row: RUN_PRIMITIVE_STEPS_MROW)
			-- Prepend to `commit_actions' an action to update
			-- the `run_step' of `step' to `row'.  This
			-- needs to be done before any other actions,
			-- so that they see the new value of `row'
		require
			step_exists: step /= Void
		do
			commit_actions.put_front (agent step.set_run_step (row))
			commit
		end

	cleanup_for_abort
			-- Take actions needed on abort of step
		do
			rollback
			set_default_role
		end

	set_date_started
			-- Set `date_started' to current date
			-- according to database
		do
			date_started := connection.system_date
		ensure
			date_started_set: date_started /= Void
		end

	highest_used_undo_version (step: RC_PRIMITIVE_STEP): INTEGER
			-- Highest undo version number used for `step'
			-- or 0 if none
		require
			step_exists: step /= Void
		local
			cat: RC_FILE_CATEGORY
			dir_name, lower_step_name, entry, version: STRING
			dir: LARGE_DIRECTORY
		do
			cat := file_generator.state_directory
			dir_name := cat.directory_pathname
			lower_step_name := step.step_name.as_lower
			lower_step_name.extend ('.')
			create dir.make (dir_name)
			if path_exists (dir_name) then
				open_directory_for_read (dir)
				from
					dir.start
					dir.readentry
				until
					dir.lastentry = Void
				loop
					entry := dir.lastentry
					if su.is_prefix (lower_step_name, entry) then
						version := entry.twin
						version.keep_tail (version.count - lower_step_name.count)
						if version.is_integer then
							Result := Result.max (version.to_integer)
						end
					end
					dir.readentry
				end
				close_directory (dir)
			end
		ensure
			result_nonnegative: Result >= 0
		end

	publish_step_started (step: RC_PRIMITIVE_STEP applicable: BOOLEAN)
		local
			do_event: RC_STEP_DO_STARTED_EVENT
		do
			create do_event
			do_event.set_step (step)
			do_event.set_is_applicable (applicable)
			do_event.publish
		end

	publish_step_undo_started (step: RC_PRIMITIVE_STEP save, applicable: BOOLEAN)
		local
			undo_event: RC_STEP_UNDO_STARTED_EVENT
		do
			create undo_event
			undo_event.set_step (step)
			undo_event.set_save_state (save)
			undo_event.set_is_applicable (applicable)
			undo_event.publish
		end

	publish_step_ended (step: RC_PRIMITIVE_STEP applicable: BOOLEAN)
			-- Publish step done event
		require
			step_exists: step /= Void
		local
			done_event: RC_STEP_DO_ENDED_EVENT
		do
			create done_event
			done_event.set_step (step)
			done_event.set_is_applicable (applicable)
			done_event.publish
		end

	publish_step_reentered (step: RC_PRIMITIVE_STEP)
			-- Publish step re-entered event
		require
			step_exists: step /= Void
		local
			reentered_event: RC_STEP_REENTERED_EVENT
		do
			create reentered_event
			reentered_event.set_step (step)
			reentered_event.publish
		end

	publish_step_undo_ended (step: RC_PRIMITIVE_STEP save_state, applicable: BOOLEAN)
			-- Publish step undone event
		require
			step_exists: step /= Void
		local
			undone_event: RC_STEP_UNDO_ENDED_EVENT
		do
			create undone_event
			undone_event.set_step (step)
			undone_event.set_save_state (save_state)
			undone_event.set_is_applicable (applicable)
			undone_event.publish
		end

	publish_step_changed_state (step: RC_PRIMITIVE_STEP)
			-- Publish step changed state event
		require
			step_exists: step /= Void
		local
			changed_event: RC_STEP_STATE_CHANGE_EVENT
		do
			create changed_event
			changed_event.set_step (step)
			changed_event.publish
		end

	publish_step_aborted (step: RC_PRIMITIVE_STEP)
			-- Publish step aborted event
		require
			step_exists: step /= Void
		local
			aborted_event: RC_STEP_DO_ABORTED_EVENT
		do
			create aborted_event
			aborted_event.set_step (step)
			aborted_event.publish
		end

	publish_step_undo_aborted (step: RC_PRIMITIVE_STEP save_state: BOOLEAN)
			-- Publish step undo aborted event
		require
			step_exists: step /= Void
		local
			aborted_event: RC_STEP_UNDO_ABORTED_EVENT
		do
			create aborted_event
			aborted_event.set_step (step)
			aborted_event.set_save_state (save_state)
			aborted_event.publish
		end

end
