
indexing
	
	description: "A run closing session, either for a specific run %
		%or for a non-run specific activity."
	author: "David Hollenberg"
	date: "September 17, 2002"

deferred class RC_SESSION

inherit
	STRING_UTILITIES_ACCESS
		export
			{NONE} all
		end
	OS_ACCESS
		export
			{NONE} all
		end
	UNIX_OS_ACCESS
		export
			{NONE} all
		end
	DB_ACCESS
		export
			{NONE} all
		end
	RC_ENVIRONMENT_ACCESS
		rename
			rc_env as environment
		end
	RC_SESSION_MANAGER_ACCESS
		export
			{NONE} all
		end
	RC_FILE_CATEGORY_NAMES
		export
			{NONE} all
		end
	GUI_UTILITIES_ACCESS
		export
			{NONE} all
		end
	CONTROL_FLOW_EXCEPTIONS_ACCESS
		export
			{NONE} all
		end

feature {NONE} -- Creation

	make
			-- Create new run closing session
		do
			create {ORACLE_CONNECTION} connection.make
			set_db (connection)
			create commit_actions
		end

	make_from_session (sess: RC_SESSION)
			-- Create new session from `sess'
		require
			session_exists: sess /= Void
		do
			connection := sess.connection
			factory := sess.factory
			application := sess.application
			log_file := sess.log_file
			create commit_actions
		end

feature -- Properties

	connection: DB_CONNECTION
			-- Connection to use for database access

	commit_actions: CLOSED_ACTION_SEQUENCE
			-- Actions to be executed once on commit

	type: STRING
			-- Type of session
		deferred
		end

	id_name: STRING
			-- Name of session-identifying value
		deferred
		end
	
	id_value: STRING
			-- Value that identifies session
		deferred
		end
	
	has_steps: BOOLEAN
			-- Is session activity broken into separate steps?
		deferred
		end

	current_step_name: STRING
			-- Name of step being executed, if not Void

	current_activity: STRING
			-- Description of current activity, if not Void
		
	current_state: STRING
			-- Description of current state, including what
			-- is being done, which user is doing it, date, etc
		local
			id, db_error: STRING
		do
			create Result.make (256)
			Result.append ("Date: ")
			unix_os.append_current_date_string (Result)
			Result.extend ('%N')
			id := id_name
			if su.not_empty (id) then
				Result.append (id)
				Result.append (": ")
				id := id_value
				if su.not_empty (id_value) then
					Result.append (id)
				end
				Result.extend ('%N')
			end
			if has_steps then
				Result.append ("Step: ")
				if su.not_empty (current_step_name) then
					Result.append (current_step_name)
				end
				Result.extend ('%N')
			end
			Result.append ("User: ")
			Result.append (environment.user_name)
			Result.extend ('%N')
			Result.append ("Host: ")
			Result.append (environment.host_name)
			Result.extend ('%N')
			Result.append ("Process id: ")
			Result.append_integer (environment.process_id)
			Result.extend ('%N')
			if su.not_empty (current_activity) then
				Result.append ("Activity: ")
				Result.append (current_activity)
			end
			Result.extend ('%N')
			if connection /= Void then
				db_error := connection.last_error_description
				if not db_error.is_empty then
					Result.append ("Database error:%N")
					Result.append (db_error)
				end
			end
		end

	path_exists (path: STRING): BOOLEAN
			-- Does `path' exist?
		require
			path_exists: path /= Void
		local
			checker: FILE_EXISTENCE_CHECK_OP
		do
			create checker
			checker.check_existence (path)
			Result := checker.path_exists
		end

	log_file: LARGE_PLAIN_TEXT_FILE
			-- Session log file or Void if not opened yet

	log_file_name: STRING
			-- Name of log file
		deferred
		end
	
	error_file_name: STRING
			-- Name of file for error messages
		deferred
		end

	shared_lock_name: STRING
			-- Name of shared lock file, if not Void
		deferred
		end

	exclusive_lock_name: STRING
			-- Name of exclusive lock file, if not Void
		deferred
		end

	exclusive_lock: LOCK
			-- Session exclusive lock, if not Void
		
	shared_lock: LOCK
			-- Session shared lock, if not Void

	factory: RC_INTERFACE_FACTORY
			-- Factory from which to get interface elements

	application: RC_APPLICATION
			-- Application control

	preferences: RC_PREFERENCE_SET
			-- User preferences for session
		do
			if session_preferences = Void then
				create session_preferences
			end
			Result := session_preferences
		end

	is_abort_requested: BOOLEAN
			-- Has user requested abort?

	is_known_user_email_address: BOOLEAN
			-- Is e-mail address of user running program known?
		local
			addr: STRING
		do
			addr := environment.user_email_address
			Result := addr /= Void and then not addr.is_empty
		end

feature -- Modification

	set_current_step_name (s: STRING)
			-- Set `current_step_name' to `s'
		require
			session_has_steps: has_steps
		do
			current_step_name := s
		ensure
			step_name_set: equal (current_step_name, s)
		end

	set_current_activity (s: STRING)
			-- Set `current_activity' to `s'
		do
			current_activity := s
		ensure
			activity_set: equal (current_activity, s)
		end

	set_exclusive_lock (l: LOCK)
			-- Set `exclusive_lock' to `l'
		do
			exclusive_lock := l
		end

	set_shared_lock (l: LOCK)
			-- Set `shared_lock' to `l'
		do
			shared_lock := l
		end

	set_factory (f: RC_INTERFACE_FACTORY)
			-- Set `factory' to `f'
		require
			a_factory_exists: f /= Void
		do
			factory := f
		ensure
			factory_set: factory = f
		end

	set_application (a: RC_APPLICATION)
			-- Set `application' to `a'
		require
			an_application_exists: a /= Void
		do
			application := a
		ensure
			application_set: application = a
		end

	set_abort_requested
			-- Set `is_abort_requested' to True
		do
			is_abort_requested := True
		ensure
			abort_requested: is_abort_requested
		end

	clear_abort_requested
			-- Set `is_abort_requested' to False
		do
			is_abort_requested := False
		ensure
			no_abort_requested: not is_abort_requested
		end

feature -- Operations

	open_log
			-- Open log file, if there is a session-specific file
		local
			name: STRING
			file: LARGE_PLAIN_TEXT_FILE
			file_opener: FILE_OPEN_APPEND_OP
		do
			name := log_file_name
			if name /= Void then
				create file.make (name)
				create file_opener
				file_opener.open (file)
				log_file := file
			end
		end

	close_log
			-- Close log file, if there is a session-specific file
		do
			if log_file /= Void then
				if not log_file.is_closed then
					log_file.close
				end
				log_file := Void
			end
		end

	lock
			-- Lock shared and exclusive lock files, if needed
		do
			lock_shared
			lock_exclusive
		end

	lock_shared
			-- Lock shared lock file, if needed
		local
			locker: LOCK_OP
			the_lock: LOCK
			name: STRING
		do
			create locker
			name := shared_lock_name
			if name /= Void then
				the_lock := os.new_lock (name)
				the_lock.set_shared
				locker.lock_nowait (the_lock)
				shared_lock := the_lock
			end
		end

	lock_exclusive
			-- Lock exclusive lock file, if needed
		local
			locker: LOCK_OP
			the_lock: LOCK
			name: STRING
		do
			create locker
			name := exclusive_lock_name
			if name /= Void then
				the_lock := os.new_lock (name)
				the_lock.set_exclusive
				locker.lock_nowait (the_lock)
				exclusive_lock := the_lock
				the_lock.set_auxiliary_data ("Last exclusive lock by user " + environment.user_name + " process id " + environment.process_id.out + " on host " + environment.host_name)
			end
		end

	lock_shared_exclusively
			-- Lock shared lock file in exclusive mode
		local
			locker: LOCK_OP
			the_lock: LOCK
			name: STRING
		do
			create locker
			name := shared_lock_name
			if name /= Void then
				the_lock := os.new_lock (name)
				the_lock.set_exclusive
				locker.lock_nowait (the_lock)
				shared_lock := the_lock
				the_lock.set_auxiliary_data ("Last exclusive lock by user " + environment.user_name + " process id " + environment.process_id.out + " on host " + environment.host_name)
			end
		end

	unlock
			-- Lock shared and exclusive lock files, if needed
		local
			unlocker: UNLOCK_OP
		do
			create unlocker
			if exclusive_lock /= Void then
				unlocker.unlock (exclusive_lock)
				exclusive_lock := Void
			end
			if shared_lock /= Void then
				unlocker.unlock (shared_lock)
				shared_lock := Void
			end
		end

	set_role (role: STRING)
			-- Set database role `role'
		local
			role_setter: ROLE_SET_OP
		do
			create role_setter
			role_setter.set_role (connection, role, environment.db_access_control)
		end
			
	set_default_role
			-- Set default database role
		local
			role_setter: ROLE_SET_OP
		do
			create role_setter
			role_setter.set_role (connection, environment.db_default_role, Void)
		end
			
	set_run_closer_role
			-- Set run closer database role
		local
			role_setter: ROLE_SET_OP
		do
			create role_setter
			role_setter.set_role (connection, environment.db_run_closer_role, environment.db_access_control)
		end
			
	lock_row (row: DB_LOCKABLE_ROW)
			-- Lock database row `row'
		require
			row_exists: row /= Void
		local
			row_locker: ROW_LOCK_OP
		do
			create row_locker
			row_locker.lock_nowait (row)
		end
			
	create_directory (name: STRING)
			-- Create directory `name'
		require
			name_exists: name /= Void
			name_not_empty: not name.is_empty
		local
			dir_creator: DIRECTORY_CREATE_OP
		do
			create dir_creator
			dir_creator.create_dir (name)
		end
			
	remove_directory (name: STRING)
			-- Remove directory `name'
		require
			name_exists: name /= Void
			name_not_empty: not name.is_empty
		local
			dir_remover: DIRECTORY_DELETE_OP
		do
			create dir_remover
			dir_remover.remove_dir (name)
		end
			
	open_directory_for_read (d: LARGE_DIRECTORY)
			-- Open `d' for read
		require
			directory_exists: d /= Void
			directory_close: d.is_closed
		local
			dir_opener: DIR_OPEN_READ_OP
		do
			create dir_opener
			dir_opener.open (d)
		end
			
	close_directory (d: LARGE_DIRECTORY)
			-- Close `d'
		require
			directory_exists: d /= Void
			directory_open: not d.is_closed
		local
			dir_closer: DIR_CLOSE_OP
		do
			create dir_closer
			dir_closer.close (d)
		end
			
	open_file_for_read (f: LARGE_FILE)
			-- Open `f' for read
		require
			file_exists: f /= Void
		local
			file_opener: FILE_OPEN_READ_OP
		do
			create file_opener
			file_opener.open (f)
		end
			
	close_file (f: LARGE_FILE)
			-- Close `f'
		require
			file_exists: f /= Void
			file_open: not f.is_closed
		local
			file_closer: FILE_CLOSE_OP
		do
			create file_closer
			file_closer.close (f)
		end
			
	delete_file_if_exists (fname: STRING)
			-- Delete file named `fname' if it exists
		require
			name_exists: fname /= Void
			name_not_empty: not fname.is_empty
		do
			if path_exists (fname) then
				delete_file (fname)
			end
		end
			
	delete_file (fname: STRING)
			-- Delete file named `fname'
		require
			name_exists: fname /= Void
			name_not_empty: not fname.is_empty
		local
			file_remover: FILE_DELETE_OP
		do
			create file_remover
			file_remover.delete (fname)
		end
			
	change_group (fname, grp_name: STRING)
			-- Change group name of `fname' to `grp_name'
		require
			file_name_exists: fname /= Void
			file_name_nonempty: not fname.is_empty
			group_name_nonempty: not grp_name.is_empty
		local
			group_changer: FILE_GROUP_CHANGE_OP
		do
			create group_changer
			group_changer.change_group (fname, grp_name)
		end
			
	add_commit_action (p: PROCEDURE [ANY, TUPLE])
			-- Append `p' to `commit_actions'
		do
			commit_actions.extend (p)
		end

	commit
			-- Commit transaction for `connection'
		require
			connection_exists: connection /= Void
			connection_open: connection.is_open
		local
			commit_event: COMMIT_EVENT
			actions: CLOSED_ACTION_SEQUENCE
		do
			connection.commit
			create commit_event
			commit_event.publish
			actions := commit_actions.twin
			commit_actions.wipe_out
			actions.call ([])
		end

	post
			-- Post transaction for `connection'
		require
			connection_exists: connection /= Void
			connection_open: connection.is_open
		local
			post_event: POST_EVENT
		do
			connection.post
			create post_event
			post_event.publish
		end

	rollback
			-- Rollback transaction for `connection'
		require
			connection_exists: connection /= Void
			connection_open: connection.is_open
		local
			rollback_event: ROLLBACK_EVENT
		do
			connection.rollback
			commit_actions.wipe_out
			create rollback_event
			rollback_event.publish
		end

	log_operation (s: STRING)
			-- Log operation described by `s'
		require
			text_exists: s /= Void
		do
			if log_file /= Void and then log_file.is_open_write then
				environment.log_operation_to_file (log_file, s)
			else
				environment.log_operation (s)
			end
		end

	log_error (s: STRING)
			-- Log error described by `s'
		require
			text_exists: s /= Void
		do
			log_operation (s)
		end

	send_error_notification (s: STRING)
			-- Send error notification with error message `s'
		require
			text_exists: s /= Void
		local
			message: OUTGOING_MESSAGE
			subject: STRING
		do
			subject := "Delta error"
			if su.not_empty (id_name) then
				subject := subject + " on " + id_name + " " + id_value
			end
			create message.make_alt_mailer (environment.mailer_command)
			message.set_header_field (message.msg_from, environment.from_line)
			message.set_header_field (message.msg_to, environment.crash_msg_to_list)
			message.set_header_field (message.msg_reply_to, environment.crash_msg_to_list)
			message.set_header_field (message.msg_subject, subject)
			message.set_message_body (s + "%N%N" + current_state)
			message.send
		end

	send_email_to_user (subj, body: STRING)
			-- Send e-mail message to user running program
		require
			email_address_known: is_known_user_email_address
			subject_exists: subj /= Void
			body_exists: body /= Void
		do
			send_email_to_list (environment.user_email_address, subj, body, Void)
		end

	send_email_to_list (addr, subj, body, repl_to: STRING)
			-- Send e-mail message with subject `subj' and `body'
			-- to `addr'.  Use `repl_to' if not
			-- as reply-to address if not Void
		require
			email_address_exists: addr /= Void
			email_address_nonempty: not addr.is_empty
			subject_exists: subj /= Void
			body_exists: body /= Void
		local
			message: OUTGOING_MESSAGE
			subject: STRING
		do
			if su.not_empty (id_name) then
				subject := id_value + " " + subj
			else
				subject := subj
			end
			create message.make_alt_mailer (environment.mailer_command)
			message.set_header_field (message.msg_from, environment.from_line)
			message.set_header_field (message.msg_to, addr)
			message.set_header_field (message.msg_subject, subject)
			if repl_to /= Void then
				message.set_header_field (message.msg_reply_to, repl_to)
			end
			message.set_message_body (body)
			message.send
		end

	process_pending_abort
			-- Process events to determine whether there
			-- are any abort requests.  If so, raise
			-- an Abort developer exception
		do
			gui.process_events
			if is_abort_requested then
				cfe.raise_abort_exception
			end
		end

	terminate
			-- Terminate `Current'
		do
			if connection.is_open then
				rollback
			end
			close_log
			unlock
			set_current_activity ("Terminating session")
			session_manager.terminate_session (Current)
		end

feature {NONE} -- Implementation

	session_preferences: RC_PREFERENCE_SET
			-- User preferences for this session

invariant
	connection_exists: connection /= Void
	commit_actions_exists: connection /= Void

end
