
indexing
	
	description: "An operation that reads run closing step information from the database, validates it and stores it in session"
	author: "David Hollenberg"
	date: "April 12, 2003"

class RC_STEP_READ_OP
inherit
	OPERATION
		rename
			try_operation as try_to_read_steps
		redefine
			default_create
		end
	DB_OPTIONS_ACCESS
		export
			{NONE} all
		redefine
			default_create
		end

create
	default_create

feature {NONE} -- Creation

	default_create
		do
			precursor {OPERATION}
		end

feature -- Operations

	read_steps (sess: RC_RUN_CLOSING_SESSION)
			-- Read and validate run closing steps using `sess'
		require
			session_exists: sess /= Void
			session_has_connection: sess.connection /= Void
			session_connection_open: sess.connection.is_open
		do
			session := sess
			connection := sess.connection
			execute
		end

feature {NONE} -- Implementation

	session: RC_RUN_CLOSING_SESSION
			-- Session context for step

	connection: DB_CONNECTION
			-- Database connection to use

	primitive_steps: LIST [RC_PRIMITIVE_STEP]

	composite_steps: LIST [RC_COMPOSITE_STEP]

	composite_step_components: LIST [RUN_CLOSING_STEP_COMPONENTS_ROW]

	primitive_step_data_items: LIST [RUN_CLOSING_STEP_DATA_ITEMS_ROW]

	data_items: LIST [RUN_CLOSING_DATA_ITEMS_ROW]

	primitive_step_table: HASH_TABLE [RC_PRIMITIVE_STEP, STRING]

	composite_step_table: HASH_TABLE [RC_COMPOSITE_STEP, STRING]

	data_item_table: HASH_TABLE [RUN_CLOSING_DATA_ITEMS_ROW, STRING]

	step_list: LIST [RC_STEP]
			-- List of steps in current traversal path

	message: STRING
			-- For building composite step error message

	try_to_read_steps
			-- Try to read and validate run closing steps
		do
			fetch_step_data
			populate_step_tables
			link_steps_to_data_items
			validate_step_data_items
			link_primitive_steps
			check_primitive_steps_for_cycles
			report_primitive_step_cycles
			set_max_required_steps
			report_primitive_execution_order_errors
			link_composite_steps
			report_empty_composite_steps
			check_composite_steps_for_cycles
			report_composite_step_cycles
			report_composite_primitive_errors
			report_composite_presentation_order_errors
			report_primitive_input_item_mismatches
			session.set_primitive_steps (primitive_steps)
			session.set_composite_steps (composite_steps)
			session.set_primitive_step_status
			-- FIXME: check that for every primitive step
			-- that is done all of its required steps are also
			-- done.  Error if not.
		end

	fetch_step_data
			-- Fetch run closing step data
		local
			prim_step: RC_PRIMITIVE_STEP
			comp_step: RC_COMPOSITE_STEP
			comp_step_part: RUN_CLOSING_STEP_COMPONENTS_ROW
			step_data_item: RUN_CLOSING_STEP_DATA_ITEMS_ROW
			data_item: RUN_CLOSING_DATA_ITEMS_ROW
		do
			create prim_step.make_template
			primitive_steps := prim_step.all_primitive_steps_ordered (connection, expected_count.at_least_zero)
			create comp_step.make_template
			composite_steps := comp_step.all_composite_steps_ordered (connection, expected_count.at_least_zero)
			create comp_step_part.make_template
			composite_step_components := comp_step_part.all_components_ordered (connection, expected_count.at_least_zero)
			create step_data_item.make_template
			primitive_step_data_items := step_data_item.all_rows (connection)
			create data_item.make_template
			data_items := data_item.all_rows (connection)
		end

	populate_step_tables
			-- Populate hash tables for faster lookup
			-- (so we don't have to search through a list).
			-- Also, initialize primitive steps and composite steps
		local
			prim_step: RC_PRIMITIVE_STEP
			comp_step: RC_COMPOSITE_STEP
			data_item: RUN_CLOSING_DATA_ITEMS_ROW
		do
			create primitive_step_table.make (primitive_steps.count)
			from
				primitive_steps.start
			until
				primitive_steps.after
			loop
				prim_step := primitive_steps.item
				prim_step.initialize (session)
				primitive_step_table.put (prim_step, prim_step.step_name)
				primitive_steps.forth
			end
			
			create composite_step_table.make (composite_steps.count)
			from
				composite_steps.start
			until
				composite_steps.after
			loop
				comp_step := composite_steps.item
				comp_step.initialize (session)
				composite_step_table.put (comp_step, comp_step.step_name)
				composite_steps.forth
			end
			
			create data_item_table.make (data_items.count)
			from
				data_items.start
			until
				data_items.after
			loop
				data_item := data_items.item
				data_item_table.put (data_item, data_item.data_item)
				data_items.forth
			end
		end

	link_steps_to_data_items
			-- Link primitive steps to their input and output
			-- data items
		local
			step_data_item: RUN_CLOSING_STEP_DATA_ITEMS_ROW
			step: RC_PRIMITIVE_STEP
			data_item: RUN_CLOSING_DATA_ITEMS_ROW
		do
			from
				primitive_step_data_items.start
			until
				primitive_step_data_items.after
			loop
				step_data_item := primitive_step_data_items.item
				step := primitive_step_table.item (step_data_item.primitive_step_name)
				data_item := data_item_table.item (step_data_item.data_item)
				if step_data_item.is_input then
					step.input_data_items.extend (data_item)
					if step_data_item.determines_applicability then
						step.applicability_data_items.extend (data_item)
					end
				else
					step.output_data_items.extend (data_item)
				end
				primitive_step_data_items.forth
			end
		end

	validate_step_data_items
			-- Make sure that the output data items of each
			-- primitive step have that step's name as
			-- their producing step name.  Otherwise,
			-- their is a database inconsistency
		local
			prim_step: RC_PRIMITIVE_STEP
			output_items: LIST [RUN_CLOSING_DATA_ITEMS_ROW]
			output_item: RUN_CLOSING_DATA_ITEMS_ROW
			data_item: RUN_CLOSING_DATA_ITEMS_ROW
			msg, prod_step_name, prim_step_name: STRING
		do
			create msg.make (0)
			
			from
				primitive_steps.start
			until
				primitive_steps.after
			loop
				prim_step := primitive_steps.item
				output_items := prim_step.output_data_items
				from
					output_items.start
				until
					output_items.after
				loop
					output_item := output_items.item
					prod_step_name := output_item.producing_primitive_step_name
					prim_step_name := prim_step.step_name
					if not equal (prod_step_name, prim_step_name) then
						msg.append ("Data item " + output_item.data_item + " produced by step " + prod_step_name + " but is output item for step " + prim_step_name + "%N")
					end
					output_items.forth
				end
				primitive_steps.forth
			end
			
			from
				data_items.start
			until
				data_items.after
			loop
				data_item := data_items.item
				prod_step_name := data_item.producing_primitive_step_name
				prim_step := primitive_step_table.item (prod_step_name)
				if not prim_step.output_data_items.has (data_item) then
					msg.append ("Data item " + data_item.data_item + " produced by step " + prod_step_name + " but is not output item for that step%N")
				end
				data_items.forth
			end
			
			if not msg.is_empty then
				handle_hard_error (msg)
			end
		end

	link_primitive_steps
			-- Link primitive steps to their required,
			-- dependent and applicability primitive steps
		local
			prim_step, prod_step: RC_PRIMITIVE_STEP
			input_items: LIST [RUN_CLOSING_DATA_ITEMS_ROW]
			input_item: RUN_CLOSING_DATA_ITEMS_ROW
			applicability_items: LIST [RUN_CLOSING_DATA_ITEMS_ROW]
			app_item: RUN_CLOSING_DATA_ITEMS_ROW
			prod_step_name: STRING
		do
			from
				primitive_steps.start
			until
				primitive_steps.after
			loop
				prim_step := primitive_steps.item
				input_items := prim_step.input_data_items
				
				from
					input_items.start
				until
					input_items.after
				loop
					input_item := input_items.item
					prod_step_name := input_item.producing_primitive_step_name
					prod_step := primitive_step_table.item (prod_step_name)
					prim_step.required_steps.extend (prod_step)
					prod_step.dependent_steps.extend (prim_step)
					input_items.forth
				end
				
				applicability_items := prim_step.applicability_data_items
				from
					applicability_items.start
				until
					applicability_items.after
				loop
					app_item := applicability_items.item
					prod_step_name := app_item.producing_primitive_step_name
					prod_step := primitive_step_table.item (prod_step_name)
					prim_step.required_applicability_steps.extend (prod_step)
					applicability_items.forth
				end
				primitive_steps.forth
			end
		end

	check_primitive_steps_for_cycles
			-- Check whether there are any cycles among
			-- primitive steps (i.e., whether there are any
			-- steps A and B such that A depends on B and
			-- B depends on A, directly or indirectly)
		local
			step: RC_PRIMITIVE_STEP
		do
			unmark_and_unvisit (primitive_steps)
			create {ARRAYED_LIST [RC_STEP]} step_list.make (primitive_steps.count)
			from
				primitive_steps.start
			until
				primitive_steps.after
			loop
				step := primitive_steps.item
				if not step.is_visited then
					check_primitive_step_for_cycles (step)
				end
				primitive_steps.forth
			end
		end

	
	check_primitive_step_for_cycles (step: RC_PRIMITIVE_STEP)
			-- Check whether there are any cycles from
			-- `step'
		local
			dependent: RC_PRIMITIVE_STEP
			dependents: LIST [RC_PRIMITIVE_STEP]
		do
			step.mark
			step_list.extend (step)
			dependents := step.dependent_steps
			from
				dependents.start
			until
				dependents.after
			loop
				dependent := dependents.item
				if dependent.is_marked then
					set_in_cycle (dependent)
				elseif not dependent.is_visited then
					check_primitive_step_for_cycles (dependent)
				end
				dependents.forth
			end
			step_list.finish
			step_list.remove
			step.unmark
			step.visit
		end

	set_max_required_steps
			-- Set `max_required_step' for all primitive
			-- steps
		local
			step: RC_PRIMITIVE_STEP
		do
			unmark_and_unvisit (primitive_steps)
			from
				primitive_steps.start
			until
				primitive_steps.after
			loop
				step := primitive_steps.item
				if not step.has_dependents then
					set_max_required_step (step)
				end
				primitive_steps.forth
			end
		end

	
	set_max_required_step (step: RC_PRIMITIVE_STEP)
			-- Step `max_required_step' for `step' and
			-- all unvisited steps it requires
		local
			req_step, max_step, child_max_step: RC_PRIMITIVE_STEP
			req_steps: LIST [RC_PRIMITIVE_STEP]
			max_order: DOUBLE
		do
			max_order := step.execution_order
			max_step := Void
			req_steps := step.required_steps
			from
				req_steps.start
			until
				req_steps.after
			loop
				req_step := req_steps.item
				if not req_step.is_visited then
					set_max_required_step (req_step)
				end
				child_max_step := req_step.max_required_step_or_self
				if child_max_step.execution_order > max_order then
					max_order := child_max_step.execution_order
					max_step := req_step
				end
				req_steps.forth
			end
			step.set_max_required_step (max_step)
			step.visit
		end

	report_primitive_execution_order_errors
			-- Report errors where the maximum execution
			-- order of any of the primitive steps
			-- required by a primitive step (directly or
			-- indirectly) is greater than the execution
			-- order of that step.  This indicates an
			-- execution order that violates ordering
			-- constraints imposed by data items
		local
			step, child_max_step: RC_PRIMITIVE_STEP
			msg: STRING
		do
			create msg.make (0)
			from
				primitive_steps.start
			until
				primitive_steps.after
			loop
				step := primitive_steps.item
				child_max_step := step.max_required_step_or_self
				if child_max_step.execution_order > step.execution_order then
					msg.append (Indent + "Step " + step.step_name + " with order " + step.execution_order.out + " depends on step " + child_max_step.step_name + " with order " + child_max_step.execution_order.out + "%N")
				end
				primitive_steps.forth
			end
			if not msg.is_empty then	
				handle_hard_error ("The following primitive steps depend on steps whose execution order%Nis greater than theirs:%N%N" + msg)
			end
		end
	
	set_in_cycle (step: RC_STEP)
			-- Mark all steps in `step_list' from
			-- end back to `step' as being in a cycle
		do
			from
				step_list.finish
			until
				step_list.item = step
			loop
				step_list.item.set_in_cycle
				step_list.back
			end
			step_list.item.set_in_cycle
		end

	report_primitive_step_cycles
			-- Report cycles among primitive steps, if any
		local
			msg: STRING
		do
			msg := steps_in_cycle (primitive_steps)
			if not msg.is_empty then
				handle_hard_error ("The following primitive steps are involved in a cycle:%N%N" + msg)
			end
		end

	link_composite_steps
			-- Link composite steps to their components
		local
			step: RC_STEP
			composite_step: RC_COMPOSITE_STEP
			component_step: RUN_CLOSING_STEP_COMPONENTS_ROW
			composite_step_name: STRING
		do
			from
				composite_step_components.start
			until
				composite_step_components.after
			loop
				component_step := composite_step_components.item
				composite_step_name := component_step.step_name
				composite_step := composite_step_table.item (composite_step_name)
				if not component_step.is_null_primitive_step_name then
					step := primitive_step_table.item (component_step.primitive_step_name)
				else
					step := composite_step_table.item (component_step.nonprimitive_step_name)
				end
				composite_step.component_steps.extend (step)
					-- Can just add to end because steps
					-- were ordered when fetched
				step.parents.extend (composite_step)
				composite_step_components.forth
			end
		end

	report_empty_composite_steps
			-- Report composite steps that do not have any
			-- components
		local
			step: RC_COMPOSITE_STEP
			msg: STRING
		do
			create msg.make (0)
			from
				composite_steps.start
			until
				composite_steps.after
			loop
				step := composite_steps.item
				if step.component_steps.is_empty then
					msg.append (Indent + step.step_name + "%N")
				end
				composite_steps.forth
			end
			if not msg.is_empty then
				handle_hard_error ("The following composite steps do not have any component steps:%N%N" + msg)
			end
		end

	check_composite_steps_for_cycles
			-- Check whether there are any cycles among
			-- composite steps (i.e., whether there are any
			-- steps A and B such that A includes B as a
			-- component and B includes A as a component,
			-- directly or indirectly)
		local
			step: RC_COMPOSITE_STEP
		do
			unmark_and_unvisit (composite_steps)
			create {ARRAYED_LIST [RC_STEP]} step_list.make (composite_steps.count)
			from
				composite_steps.start
			until
				composite_steps.after
			loop
				step := composite_steps.item
				if not step.is_visited then
					check_composite_step_for_cycles (step)
				end
				composite_steps.forth
			end
		end

	
	check_composite_step_for_cycles (step: RC_COMPOSITE_STEP)
			-- Check whether there are any cycles from
			-- `step'
		local
			component: RC_STEP
			components: LIST [RC_STEP]
			composite_step: RC_COMPOSITE_STEP
		do
			step.mark
			step_list.extend (step)
			components := step.component_steps
			from
				components.start
			until
				components.after
			loop
				component := components.item
				if component.is_composite then
					composite_step ?= component
					check
						composite_step_exists: composite_step /= Void
					end
					if composite_step.is_marked then
						set_in_cycle (composite_step)
					elseif not composite_step.is_visited then
						check_composite_step_for_cycles (composite_step)
					end
				end
				components.forth
			end
			step_list.finish
			step_list.remove
			step.unmark
			step.visit
		end

	report_composite_step_cycles
			-- Report cycles among composite steps, if any
		local
			msg: STRING
		do
			msg := steps_in_cycle (composite_steps)
			if not msg.is_empty then
				handle_hard_error ("The following composite steps are involved in a cycle:%N%N" + msg)
			end
		end

	steps_in_cycle (list: LIST [RC_STEP]): STRING
			-- Steps in `list' involved in a cycle or empty
			-- string if none
		local
			step: RC_STEP
		do
			create Result.make (0)
			from
				list.start
			until
				list.after
			loop
				step := list.item
				if step.is_in_cycle then
					Result.append (Indent + step.step_name + "%N")
				end
				list.forth
			end
		ensure
			result_exists: Result /= Void
		end

	report_composite_primitive_errors
			-- Report "top level" composite steps (ones with
			-- no parents) that include more than one
			-- occurrence of a primitive step (directly or
			-- indirectly) or that have primitive steps in
			-- an order that violates dependency constraints
		local
			step: RC_COMPOSITE_STEP
			repeated_msg, bad_order_msg: STRING
		do
			create repeated_msg.make (0)
			create bad_order_msg.make (0)
			from
				composite_steps.start
			until
				composite_steps.after
			loop
				step := composite_steps.item
				if step.is_top_level then
					repeated_msg.append (step.repeated_step_text)
					bad_order_msg.append (step.invalid_step_order_text)
				end
				composite_steps.forth
			end
			
			if not repeated_msg.is_empty then
				handle_hard_error ("The following composite steps contain more than one occurrence%Nof a primitive step (directly or indirectly):%N%N" + repeated_msg)
			end
			if not bad_order_msg.is_empty then
				handle_hard_error ("The following composite steps include primitive steps to be executed before%Nother primitive steps that they depend on:%N%N" + bad_order_msg)
			end
		end

	report_composite_presentation_order_errors
			-- Report cases where the presentation order
			-- of a top-level composite step is null
		local
			step: RC_COMPOSITE_STEP
			msg: STRING
		do
			create msg.make (0)
			from
				composite_steps.start
			until
				composite_steps.after
			loop
				step := composite_steps.item
				if step.is_top_level and step.is_null_presentation_order then
					msg.append (Indent + step.step_name + "%N")
				end
				composite_steps.forth
			end
			if not msg.is_empty then	
				handle_hard_error ("The following top-level composite steps have a null presentation order:%N%N" + msg)
			end
		end
	
	report_primitive_input_item_mismatches
			-- Report cases where the list of input data items
			-- for a step specified in the database is
			-- different than the list specified by the
			-- step's implementation
		local
			step: RC_PRIMITIVE_STEP
			input_items: LIST [RUN_CLOSING_DATA_ITEMS_ROW]
			op: RC_PRIMITIVE_STEP_OP
			msg, s: STRING
			set1: TWO_WAY_SORTED_SET [STRING]
			set2: TWO_WAY_SORTED_SET [STRING]
			items1, items2: STRING
		do
			create msg.make (0)
			from
				primitive_steps.start
			until
				primitive_steps.after
			loop
				step := primitive_steps.item
				op := session.primitive_step_implementation (step)
				create set1.make
				set1.compare_objects
				set1.fill (op.input_data_items)
				create set2.make
				set2.compare_objects
				input_items := step.input_data_items
				input_items.do_all (agent add_to_set (?, set2))
				
				if not (set1.is_subset (set2) and set2.is_subset (set1)) then
					create items1.make (0)
					create items2.make (0)
					set1.do_all (agent append_string (?, items1))
					set2.do_all (agent append_string (?, items2))
					msg.append (Indent + step.step_name + " in database: " + items2 + "%N")
					msg.append (Indent + step.step_name + " implementation: " + items1 + "%N")
				end
				primitive_steps.forth
			end
			if not msg.is_empty then	
				s := "The following primitive steps have a mismatch between the input%Ndata items listed in the database and the input data items returned by%Nthe step's implementation:%N%N" + msg
				if session.environment.is_production then
					handle_hard_error (s)
				else
					warn_with_notify (s)
				end
			end
		end
	
	unmark_and_unvisit (list: LIST [RC_STEP])
			-- Unmark and unvisit all steps in `list'
		local
			step: RC_STEP
		do
			from
				list.start
			until
				list.after
			loop
				step := list.item
				step.unmark
				step.unvisit
				list.forth
			end
		end

	publish_start_close_event
		local
			close_event: RUN_START_CLOSE_EVENT
		do
			create close_event
			close_event.publish
		end

	append_string (s: STRING dest: STRING)
		require
			string_exists: s /= Void
			dest_exists: dest /= Void
		do
			if not dest.is_empty then
				dest.append (", ")
			end
			dest.append (s)
		end

	add_to_set (r: RUN_CLOSING_DATA_ITEMS_ROW dest: SET [STRING])
		require
			row_exists: r /= Void
			dest_exists: dest /= Void
		do
			dest.extend (r.data_item)
		end

	Indent: STRING = "    "

end
