class RESPONSE_COLLECTOR
		-- RESPONSE_GENERATOR plus
		--  a linked list to collect the generated responses

inherit
	ARRAYED_LIST[RESPONSE_ITEM]	-- the list of responses
	rename	make as make_list,
		extend as extend_list,
		append as append_list
	end;

	RESPONSE_GENERATOR
	rename	make as make_generator
	undefine copy, is_equal
	end;

create
	make

create {RESPONSE_COLLECTOR}
	make_filled

feature -- status

	severity: INTEGER is
		-- highest severity level in our list
	local ndx: INTEGER;	-- (don't use index)
	do  from  ndx := 1
	    until ndx > count
	    loop  Result := Result.max(i_th(ndx).severity)
		  ndx := ndx + 1
	    end
	end;

	is_in_error: BOOLEAN is
		-- true if we have any errors (including fatals)
	do  Result := severity >= severity_is_error
	end;

	is_in_urgent: BOOLEAN is
		-- true if we have any urgent warnings OR errors
	do Result := severity >= severity_is_urgent
	end;

	is_in_warning: BOOLEAN is
		-- true if we have any warnings OR errors
	do  Result := severity >= severity_is_warn
        end;

	suppressed_count: INTEGER
		-- number of responses suppressed (due to repetition)

	summary_total_key: STRING is "TOTAL"
	summary_addl_key:  STRING is "ADDL"	-- rvalue keys for summaries

	summary_id: STRING is "SUMMARY"		
				-- id for all summary RESPONSE_ITEMs

	has_id(id: STRING): BOOLEAN is
		-- do we have a response with id = id in our list?
	local ndx: INTEGER	-- (dont use index)
	do  from  ndx := 1
	    until (ndx > count) or Result
	    loop  Result := equal(id, i_th(ndx).id)
		  ndx := ndx + 1
	    end
	end;

feature -- access

	i_th_at_severity(i: INTEGER; sev: INTEGER): RESPONSE_ITEM is
		-- return the i_th item that is at severity sev
		-- (may be Void)
	require i > 0; is_severity_level(sev);
	local ndx, j: INTEGER; it: RESPONSE_ITEM;
	do  from  ndx := 1
	    until (ndx > count) or Result /= Void
	    loop  it := i_th(ndx);
		  if it.severity = sev
		   then j := j + 1;
			if j = i then Result := it end;
		   end;
		  ndx := ndx + 1
	    end;
	ensure Result /= Void implies Result.severity = sev
	end;

feature -- modification

	extend(ri: RESPONSE_ITEM) is
		-- (not used to tack on the summary items)
	do  remove_summaries;
	    extend_list(ri)	-- add to list
	end;

	append(ris: LIST[RESPONSE_ITEM]) is
	do  remove_summaries;
	    append_list(ris)
	end;

	add_response_item(id: STRING; retr: BOOLEAN) is
		-- build an item and append to end of our list
	require id /= Void
	local rc: RESPONSE_COUNTER;  omit: BOOLEAN;
	do  build_response_item(id);
	    last_response_item.set_is_retractable(retr);
	    if last_max_occurrences > 0
	     then	-- this is a counted response
		  rc := the_counter;
		  rc.bump;
		  omit := rc.maxxed_out;
	    end;
	    if omit
	     then suppressed_count := suppressed_count + 1
	     else extend(last_response_item);	-- add to end of list
	    end;
	ensure (count+suppressed_count) = old (count+suppressed_count)+1
	end;

	head(n: INTEGER) is
		-- keep the first n (or fewer) items in the list
	require valid_arg: n >= 0;
		nothing_suppressed: n = 0 or suppressed_count = 0
	do  if n = 0
	     then wipe_out;
		  suppressed_count := 0
	    elseif n < count
	     then from  index := n+1
		  until count = n
		  loop  remove
		  end;
	     else --leave it alone
	    end;
	end;

	add_summaries is
		-- add summaries of suppressed lines to our list
		-- (called implicitly by the display routines)
	local rc: RESPONSE_COUNTER;  ri: RESPONSE_ITEM;
	do  if suppressed_count > 0 and first_summary = 0
	     then from  cntrs.start
		  until cntrs.after
		  loop  rc := cntrs.item_for_iteration;
			if rc.maxxed_out
			 then update_rvalue_int(summary_total_key,
					rc.actual_occurrences);
			      update_rvalue_int(summary_addl_key,
					rc.additional_occurences);
			      create ri.make(rc.severity, summary_id,
					substituted_text(rc.summary_template));
			      extend_list(ri);
			      if first_summary = 0
			       then first_summary := count;
			      end;
			end;
			cntrs.forth
		  end;
		  check first_summary /= 0 end
	    end;
	end;

	remove_summaries is
		-- remove the summaries from our list
		-- (called implicitly by add_response_item)
	do  if first_summary /= 0
	     then count := first_summary - 1;
		  first_summary := 0
	    end
	end

	retract is
		-- remove the retractable items
	do  remove_summaries;
	    from  start
	    until after
	    loop  if item.is_retractable
		   then remove
		   else forth
		  end
	    end;
	end;

feature -- Display

	all_plain_text(ehdr, uhdr, whdr, ihdr: STRING): STRING is
		-- all RESPONSE_ITEMS as plain text, grouped by errors,
		--  urgent warnings, warnings, and infos
		--  (in that order; nulls suppressed).
		--  Each section is omitted if there are no items.
		--  Items in each section
		--  are in list order and are separated by blank lines
	require ehdr /= Void and whdr /= Void and ihdr /= Void
	do  Result := plain_text_spaced(ehdr, severity_is_error,
						severity_is_fatal);
	    Result.append(
		      plain_text_spaced(uhdr, severity_is_urgent,
						severity_is_urgent));
	    Result.append(
		      plain_text_spaced(whdr, severity_is_warn,
						severity_is_warn));
	    Result.append(
		      plain_text_spaced(ihdr, severity_is_info,
						severity_is_info));
	end;

	all_plain_text_alt(ehdr, uhdr, whdr, ihdr: STRING): STRING is
		-- all RESPONSE_ITEMS as plain text, grouped by errors,
		--  urgent warnings, warnings, and infos
		--  (in that order).  Each section is
		--  omitted if there are no items.  Items in each section
		--  are in list order and are separated by blank lines,
		--  except for the info section which is not
	require ehdr /= Void and whdr /= Void and ihdr /= Void
	do  Result := plain_text_spaced(ehdr, severity_is_error,
						severity_is_fatal);
	    Result.append(
		      plain_text_spaced(uhdr, severity_is_urgent,
						severity_is_urgent));
	    Result.append(
		      plain_text_spaced(whdr, severity_is_warn,
						severity_is_warn));
	    Result.append(
		      plain_text(ihdr, severity_is_info,
						severity_is_info));
	end;

	all_html_text(ehdr, uhdr, whdr, ihdr: STRING): STRING is
		--all RESPONSE_ITEMS as html text
	require ehdr /= Void and whdr /= Void and ihdr /= Void
	do  Result := html_text(ehdr, severity_is_error, severity_is_fatal);
	    Result.append(
		      html_text(uhdr, severity_is_urgent, severity_is_urgent));
	    Result.append(
		      html_text(whdr, severity_is_warn, severity_is_warn));
	    Result.append(
		      html_text(ihdr, severity_is_info,	severity_is_info));
	end;

	plain_text(hdr: STRING;  lvl1, lvl2: INTEGER): STRING is
		-- concatenate hdr and all the reply items with levels
		-- between lvl1 and lvl2, inclusive, with %N separation
		-- but return empty if nothing
	require hdr /= Void and lvl1 <= lvl2
	do  if plain_fixer = Void then set_plain_fixer(null_fixer) end;
	    Result := arbitrary_text(hdr, "", "%N", "%N",
					lvl1, lvl2, plain_fixer,
					text_id_exclusion_list);
	end;

	plain_text_spaced(hdr: STRING;  lvl1, lvl2: INTEGER): STRING is
		-- concatenate hdr and all the reply items with levels
		-- between lvl1 and lvl2, inclusive, with %N%N separation
		-- but return empty if nothing
	require hdr /= Void and lvl1 <= lvl2
	do  if plain_fixer = Void then set_plain_fixer(null_fixer) end;
	    Result := arbitrary_text(hdr, "", "%N%N", "%N",
					lvl1, lvl2, plain_fixer,
					text_id_exclusion_list);
	end;

	html_text(hdr: STRING;  lvl1, lvl2: INTEGER): STRING is
		-- return string is an html unordered list
		-- with hdr in front (probably of form "<hi>text</hi>")
		-- empty list returns an empty string
	require hdr /= Void and lvl1 <= lvl2
	do  if html_fixer = Void then set_html_fixer(std_html_fixer) end;
	    Result := arbitrary_text(hdr, "%N<ul>", "%N<li>", "%N</ul>%N",
					lvl1, lvl2, html_fixer,
					html_id_exclusion_list);
	end;

	arbitrary_text(hdr1, hdr2, isep, trl: STRING;
			lvl1, lvl2: INTEGER;
			fx: OBJECT_MODIFIER[STRING];
			xid: LIST[STRING]): STRING is
		-- build a single message containing all the RESPONSE_ITEMs
		--  that are between lvl1 and lvl2 inclusive
		-- begins with hdr1 and hdr2
		--  isep before each successive item
		--   and followed by trl
		-- but, Result is empty if no RESPONSE_ITEMs match
		-- each RESPONSE_ITEM text is fixed (but not the headers)
	require hdr1/=Void and hdr2/=Void and isep/=Void and trl/=Void;
		lvl1 <= lvl2
	local ri: RESPONSE_ITEM; byts, ohd: INTEGER; foo: STRING;
	do add_summaries;
	   from  start;  ohd := isep.count;
	   until after		-- count bytes in result
	   loop  ri := item;
		 if lvl1 <= ri.severity and ri.severity <= lvl2
		    and not xid.has(ri.id)
		  then	byts := byts + ri.text.count + ohd;
			if with_ids then byts := byts + ri.id.count + 2 end;
		 end;
		 forth;
	   end;
	   if byts=0		-- (there may not be any!!)
	    then create Result.make(0);		-- empty string
	    else create Result.make(hdr1.count+hdr2.count+trl.count+byts);
		 Result.append(hdr1)
		 Result.append(hdr2);
		 from  start;
		 until after
		 loop ri := item;
		      if lvl1 <= ri.severity and ri.severity <= lvl2
			 and not xid.has(ri.id)
		       then Result.append(isep);
			    if with_ids
			     then foo := su.string_of_3(ri.id, su.colon_space,
							ri.text)
			     else foo := ri.text
			    end;
			    Result.append(fx.modify(foo));
		      end;
		      forth;
		 end;
		 Result.append(trl);
	   end;
	end;

feature -- string modifiers

	plain_fixer: OBJECT_MODIFIER[STRING]
		-- the fixer of choice for plain text

	set_plain_fixer(f: OBJECT_MODIFIER[STRING]) is
	do  plain_fixer := f
	end;

	null_fixer: NULL_OBJECT_MODIFIER[STRING] is
	once  create Result
	end;

	html_fixer: OBJECT_MODIFIER[STRING]
		-- the fixer of choice for html text

	set_html_fixer(f: OBJECT_MODIFIER[STRING]) is
        do  html_fixer := f
	end;

	std_html_fixer: HTML_STRING_MODIFIER is
	once  create Result
	end;

feature -- Display Control

	with_ids: BOOLEAN
		-- are we including the id in our display strings?

	set_with_ids(wid: BOOLEAN) is
	do  with_ids := wid
	ensure with_ids = wid
	end;

	add_text_excluded_id(id: STRING) is
	require not id.is_empty
	do  if not text_id_exclusion_list.has(id)
	     then text_id_exclusion_list.extend(id)
	    end
	end;

	add_html_excluded_id(id: STRING) is
	require not id.is_empty
	do  if not html_id_exclusion_list.has(id)
	     then html_id_exclusion_list.extend(id)
	    end
	end;

feature {NONE} -- creation

	make(app: STRING) is
	do  make_list(0);
	    make_generator(app);
	    with_ids := true
	    create text_id_exclusion_list.make(1);
	    text_id_exclusion_list.compare_objects
	    create html_id_exclusion_list.make(1);
	    html_id_exclusion_list.compare_objects
	end;

feature {NONE} -- implementation details

	first_summary: INTEGER
		-- index of first of the summaries (after add_summaries)
		-- 0 if none in list
		-- summaries are always consecutive and at the end

	cntrs: HASH_TABLE[RESPONSE_COUNTER, STRING]
		-- response counters for the repeatable items

	the_counter: RESPONSE_COUNTER is
		-- find or create RESPONSE_COUNTER for last_response_item
	require last_max_occurrences > 0
	local id: STRING;
	do  if cntrs = Void then create cntrs.make(10)  end;
	    id := last_response_item.id;
	    Result := cntrs.item(id);
	    if Result = Void
	     then create Result.make(id, last_response_item.severity,
				last_max_occurrences,
				last_summary_template);
		  cntrs.put(Result, id);
	    end
	end;

	text_id_exclusion_list: ARRAYED_LIST[STRING];
	html_id_exclusion_list: ARRAYED_LIST[STRING];

invariant
  --	count > 0 implies is_severity_level(severity)

end  -- RESPONSE_COLLECTOR
