class RESPONSE_GENERATOR

	-- storage for the rvalues (response values) that will be used to
	--  fill out MESSAGE_TEMPLATE <key> parts
	-- NOTE: key names are case-insensitive in both calls and templates
	-- this class also builds a new RESPONSE_ITEM, with actual rvalue
	--  substitution, at each call to build_response_item,
	--  but does not attempt to do anything to store, or otherwise
	--  organize, those RESPONSE_ITEMs.

inherit
	SEVERITY_LEVELS

	DB_ACCESS			-- db.xxx
	export  {NONE} all	end;

	STRING_UTILITIES_ACCESS
		export
			{NONE} all
		end
	
	EXCEPTIONS_ACCESS
		export
			{NONE} all
		end

	FILE_UTILITIES_ACCESS
		export
			{NONE} all
		end

create
	make

feature -- application name

	application: STRING;
		-- APPLICATION name (for database queries)
		-- set at creation time

feature -- rvalue management

	valid_key(key: STRING): BOOLEAN is
		-- is key a valid key value?
	do  Result := su.not_empty(key)
	end;

	check_rvalue(key: STRING): STRING is
		-- return Void if no associated value, or the
		-- associated value (which may be empty) if there is one
	require valid_key: valid_key(key)
	do  Result := response_rvalues.item(key.as_upper)
	end;

	update_rvalue(key, value: STRING) is
		-- set value as the string now associated with key
		-- (there may or may not have been a prior value
		--  associated with key; if there was, it is lost.)
		-- NOTE: the value is NOT cloned.  This class will not
		--  change the contents of the string; neither should
		--  anyone else
	require valid_key: valid_key(key);
		value_exists: value /= Void
	do  response_rvalues.force(value, key.as_upper);
	ensure	check_rvalue(key) = value
	end;

	update_rvalue_int(key: STRING; value: INTEGER) is
		-- shortcut for an integer value; uses value.out
	require valid_key: valid_key(key)
	do  update_rvalue(key, value.out)
	end;

	remove_rvalue(key: STRING) is
		-- remove key and its last associated value (if any)
	require valid_key: valid_key(key)
	do  response_rvalues.remove(key.as_upper);
	ensure  check_rvalue(key) = Void
	end;

	left_key_delimiter,
	right_key_delimiter: CHARACTER
		-- left and right key delimiters in message_templates
		--  and summary_templates.  Default is '<' and '>'

	set_key_delimiters(lft, rgt: CHARACTER) is
		-- change the delimiter characters for my templates
	do  left_key_delimiter := lft;
	    right_key_delimiter := rgt
	end;

feature -- data from last build_response_item call

	last_message_id: STRING

	last_response_item: RESPONSE_ITEM

	last_max_occurrences: INTEGER

	last_summary_template: STRING

feature -- implementation

	have_template_for(id: STRING): BOOLEAN is
		-- do we have a template for message_id id?
	require id /= Void
	do  Result := try_to_get_template_row(id) /= Void
	end;

	build_response_item(id: STRING) is
		-- fetch a RESPONSE_TEXT_TEMPLATES row for this application
		--  and id; perform rvalue and literal substitution on
		--  its MESSAGE_TEMPLATE, and build a new RESPONSE_ITEM in
		--  last_response_item
		-- an exception occurs if either id is not found in the
		-- database, or a MESSAGE_TEMPLATE rvalue is not found
		-- in response_rvalues
	require id /= Void
	local  resp: RESPONSE_TEXT_TEMPLATES_ROW;
	do  last_message_id := id;
	    -- fetch row from our cache
	    resp := get_template_row(id);
	    -- build response_item to return
	    create last_response_item.make(
			severity_level(resp.severity),
			id,
			substituted_text(resp.message_template) );
	    last_max_occurrences := resp.max_occurrences;
	    last_summary_template := resp.summary_template;
	ensure  last_response_item /= Void and then
		id = last_response_item.id
	end;

	make_response_item(id, temp: STRING) is
		-- similar to build_response_item, except that if we are
		-- not already connected to a database, we use temp as
		-- the message_template and error as the severity
	require id /= Void and temp /= Void
	do  if db.is_open
	     then build_response_item(id)	-- temp is ignored here
	     else create last_response_item.make(
			severity_is_error,
			id,
			substituted_text(temp) );
		  last_max_occurrences := 0;
		  last_summary_template := ""
	    end;
	ensure  last_response_item /= Void and then
		id = last_response_item.id
	end;

	response_text (id: STRING): STRING is
			-- Text of response for application
			-- `application' and message id `id', with
			-- values substituted for substitution
			-- parameters.  Will punt if `id' cannot be
			-- found or if a substitution parameter in
			-- template is not defined
		local
			s: STRING
		do
			build_response_item (id)
			s := last_response_item.text
			if s.count > 0 then
				s.put (s.item (1).upper, 1)
			end
			Result := s
		end

	substituted_text(txt: STRING): STRING is
		-- rvalue substitution directly on txt
		-- (raises an exception if any rvalues are missing)
	require txt /= Void
	local i, state: INTEGER; c: CHARACTER; t, key: STRING;
	do  t := maybe_use_file_contents(txt);
	    create Result.make(2 * t.count);
	    create key.make(16);
	    from  i := 1;
	    until i > t.count
	    loop  c := t @ i;		-- fetch
		  i := i + 1;		-- and advance
		  if state = 0		-- collecting Result
		   then if c = left_key_delimiter
			 then key.wipe_out;
			      state := 1;
			elseif c = right_key_delimiter
			 then state := 3;
			 else Result.extend(c)	-- normal Result collector
			end;
		  elseif state = 1	-- after left_key_delimiter
		   then if c = left_key_delimiter  -- "<<"
			 then Result.extend(c);	   --  treat as literal '<'
			      state := 0
			elseif	c = right_key_delimiter
			 then state := 0	-- ignore "<>"
			 else key.extend(c.as_upper);
			      state := 2
			end;
		  elseif state = 2	-- collecting key name
		   then if c = right_key_delimiter
			 then Result.append(subst(key));
			      state := 0;
			 else key.extend(c.as_upper);	-- normal key name coll
			end;
		   else			-- after right_key_delimiter
			if c = right_key_delimiter
			 then Result.extend(c);
			      state := 0;
			 else -- oops
			end;
		  end;
	    end;
	    if state /= 0 then -- oops
	    end;
	    Result := maybe_use_file_contents(Result);
	end;

feature {NONE} -- creation

	make(app: STRING) is
	require app /= Void;
	do  application := app
	    set_key_delimiters ('<', '>')
	    -- create rvalue hash table
	    create response_rvalues.make(17);
	end;

feature {NONE} -- implementation

	response_rvalues: HASH_TABLE[STRING, STRING]
		-- all the latest known rvalue values

	subst(p: STRING): STRING is
	do  Result := check_rvalue(p);
	    if Result = Void
	     then exc.raise(su.string_of_4("RESPONSE_GENERATOR: no value for ",
				p, " in ", last_message_id))
	    end;
	end;

	the_templates: LIST[RESPONSE_TEXT_TEMPLATES_ROW] is
	local r:RESPONSE_TEXT_TEMPLATES_ROW; b: DB_BIND_S;
	do  if cached_templates = Void
	     then  -- fetch all templates for this application
		   create r.make_template;
		   create b.make(application);
		   cached_templates := r.all_rows_filtered
				(db, "application = :1", b)
	    end;
	    Result := cached_templates
	end;

	cached_templates: like the_templates

	try_to_get_template_row(id: STRING): RESPONSE_TEXT_TEMPLATES_ROW is
		-- fetch from the cached list (Void if not there)
	local foo: RESPONSE_TEXT_TEMPLATES_ROW
	do  from  the_templates.start
	    until the_templates.after or Result /= Void
	    loop  foo := the_templates.item;
		  if equal(foo.message_id, id) then Result := foo end;
		  the_templates.forth
	    end;
	end;

	get_template_row(id: STRING): RESPONSE_TEXT_TEMPLATES_ROW is
		-- fetch from the cached list (exception if not there)
	do  Result := try_to_get_template_row(id)
	    if Result = Void
	     then exc.raise(su.string_of_4(
				"RESPONSE_GENERATOR: no template for ",
				application, su.comma_space, id))
	    end
	ensure Result /= Void
	end;

	maybe_use_file_contents(t: STRING): STRING is
		-- kludge until we have long CHARs in the database
	require t /= Void
	local fnam: STRING
	do  if t.count > 1 and then (t @ 1) = '@'
	     then -- we have "@<file-name>"
		  fnam := t.substring(2,t.count);
		  Result := fu.file_contents(fnam)
		  if Result = Void 
		   then exc.raise(su.string_of_2(
			"RESPONSE_GENERATOR: could not read ", fnam))
		  end;
	     else -- normal case; Result is just the argument
		  Result := t
	    end;
	end;

end  -- RESPONSE_GENERATOR
