
indexing
	
	description: "One copy of a die element (of any type) of reticle model"
	author: "David Hollenberg"
	date: "July 28, 2006"

deferred class RETICLE_DIE
inherit
	RETICLE_MODEL_ELEMENT
		redefine
			is_die
		end

feature {NONE} -- Creation

	make (a_die: like die a_die_location: like die_location a_model: like anchored_model)
		require
			a_die_not_void: a_die /= Void
			a_die_location_not_void: a_die_location /= Void
			a_model_not_void: a_model /= Void
			is_valid_for_creation: is_valid_for_creation (a_die)
		do
			die := a_die
			die_location := a_die_location
			model := a_model
			die.detach_actions.extend (agent update_die)
			die_location.detach_actions.extend (agent update_die_location)
			create temp_point.make_xy (0, 0)
			
			-- Make sure preferences for movability, etc
			-- are set appropriately if object is
			-- stationary, etc.  Not strictly needed but
			-- might avoid confusion for people looking at
			-- contents of packing database tables
			
			if is_stationary_in_x then
				set_is_fixed_in_x_internal (True)
				set_is_movable_in_x_internal (False)
			end
			if is_stationary_in_y then
				set_is_fixed_in_y_internal (True)
				set_is_movable_in_y_internal (False)
			end
			if is_stationary_orientation then
				set_is_rotatable_internal (False)
			end
			if is_never_duplicatable then
				set_is_duplicatable_internal (False)
			end
			if is_never_deletable then
				set_is_deletable_internal (False)
			end
		end

feature -- Properties

	die: PACKING_DIES_MROW
			-- Die of which this is one location
	
	die_location: PACKING_DIE_LOCATIONS_MROW
			-- Location of this copy of die
	
	max_die_size: CONST_POINT
			-- Maximum die size, as constrained by package
			-- cavity sizes and user-specified limits.
			-- Void indicates no maximum
		do
			Result := Void
		end

	is_oversize_die_always_allowed: BOOLEAN
			-- Is a die that is bigger than `max_die_size'
			-- allowed unconditionally?
		do
			Result := False
		end

	is_oversize_die_possibly_allowed: BOOLEAN
			-- Is a die that is bigger than `max_die_size'
			-- allowed but only under specific conditions?
		do
			Result := False
		end

	oversize_die_inclusion_conditions: STRING
			-- Description of conditions under which 
			-- die that is bigger than `max_die_size'
			-- allowed
		do
			Result := Void
		end

	is_die: BOOLEAN = True
			-- Is element a die?

	is_selectable: BOOLEAN = True
			-- May object be selected?

	rotation: INTEGER
			-- Rotation of object in degrees clockwise
		do
			Result := die.die_wafer_rotation
		end

	is_rotated_sideways: BOOLEAN
			-- Is object rotated by 90 degrees clockwise or
			-- 90 degrees counterclockwise?
		do
			Result := die.is_rotated_sideways
		end

	is_fixed_in_x_internal: BOOLEAN
			-- Is object temporarily fixed in the X dimension
			-- so that it cannot move in that dimension?
		do
			Result := die_location.is_fixed_in_x
		end

	is_fixed_in_y_internal: BOOLEAN
			-- Is object temporarily fixed in the Y dimension
			-- so that it cannot move in that dimension?
		do
			Result := die_location.is_fixed_in_y
		end

	is_movable_in_x_internal: BOOLEAN
			-- May object currently be moved in the X dimension?
		do
			Result := die_location.is_movable_in_x
		end

	is_movable_in_y_internal: BOOLEAN
			-- May object currently be moved in the X dimension?
		do
			Result := die_location.is_movable_in_y
		end

	is_rotatable_internal: BOOLEAN
			-- May object currently be moved in the X dimension?
		do
			Result := die_location.is_rotatable
		end

	is_duplicatable_internal: BOOLEAN
			-- May object currently be duplicated?
		do
			Result := die_location.is_duplicatable
		end

	is_deletable_internal: BOOLEAN
			-- May object currently be deleted?
		do
			Result := die_location.is_deletable
		end

	can_be_fixed_in_x: BOOLEAN = True
			-- Can object be temporarily fixed in the X dimension
			-- so that it cannot move in that dimension?

	can_be_fixed_in_y: BOOLEAN = True
			-- Can object be temporarily fixed in the Y dimension
			-- so that it cannot move in that dimension?

	is_stationary_in_x: BOOLEAN
			-- Can object never be moved in the X dimension?
		do
			Result := False
		end

	is_stationary_in_y: BOOLEAN
			-- Can object never be moved in the Y dimension?
		do
			Result := False
		end

	is_stationary_orientation: BOOLEAN
			-- Can orientation of object never be changed?
		do
			Result := False
		end

	is_never_duplicatable: BOOLEAN
			-- Can object never be duplicated?
		do
			Result := False
		end

	is_never_deletable: BOOLEAN
			-- Can object never be deleted?
		do
			Result := False
		end

	size_x: DOUBLE
			-- Size of minimum bounding box in X dimension
			-- of reticle field of view
		do
			Result := die.non_null_die_size_printed_x
		end

	size_y: DOUBLE
			-- Size of minimum bounding box in Y dimension
			-- of reticle field of view
		do
			Result := die.non_null_die_size_printed_y
		end

	center_x: DOUBLE
			-- Center of minimum bounding box in X dimension
		do
			Result := field.center_x + field_relative_center_x
		end

	center_y: DOUBLE
			-- Center of minimum bounding box in Y dimension
		do
			Result := field.center_y + field_relative_center_y
		end

	field_relative_center_x: DOUBLE
			-- Center of minimum bounding box in X dimension,
			-- relative to center of field with which element
			-- is associated
		do
			Result := die_location.location_x
		end

	field_relative_center_y: DOUBLE
			-- Center of minimum bounding box in Y dimension,
			-- relative to center of field with which element
			-- is associated
		do
			Result := die_location.location_y
		end

	db_maximum_move_up: DOUBLE
			-- Maximum distance element can move up and still
			-- fit in database
		local
			list: like all_fields
			coord: DOUBLE
		do
			Result := Infinity
			list := all_fields
			from
				list.start
			until
				list.after
			loop
				coord := die_location.location_y + field.center_y - list.item.center_y
				Result := Result.min (die_location.max_db_value - coord)
				list.forth
			end
			Result := Result.max (0.0)
		end

	db_maximum_move_down: DOUBLE
			-- Maximum distance element can move down and still
			-- fit in database
		local
			list: like all_fields
			coord: DOUBLE
		do
			Result := Infinity
			list := all_fields
			from
				list.start
			until
				list.after
			loop
				coord := die_location.location_y + field.center_y - list.item.center_y
				Result := Result.min (coord - -die_location.max_db_value)
				list.forth
			end
			Result := Result.max (0.0)
		end

	db_maximum_move_left: DOUBLE
			-- Maximum distance element can move left and still
			-- fit in database
		local
			list: like all_fields
			coord: DOUBLE
		do
			Result := Infinity
			list := all_fields
			from
				list.start
			until
				list.after
			loop
				coord := die_location.location_x + field.center_x - list.item.center_x
				Result := Result.min (coord - -die_location.max_db_value)
				list.forth
			end
			Result := Result.max (0.0)
		end

	db_maximum_move_right: DOUBLE
			-- Maximum distance element can move right and still
			-- fit in database
		local
			list: like all_fields
			coord: DOUBLE
		do
			Result := Infinity
			list := all_fields
			from
				list.start
			until
				list.after
			loop
				coord := die_location.location_x + field.center_x - list.item.center_x
				Result := Result.min (die_location.max_db_value - coord)
				list.forth
			end
			Result := Result.max (0.0)
		end

feature -- Modification

	set_is_fixed_in_x (b: BOOLEAN)
			-- Set `is_fixed_in_x' to `b'
		do
			set_is_fixed_in_x_internal (b)
		end

	set_is_fixed_in_y (b: BOOLEAN)
			-- Set `is_fixed_in_y' to `b'
		do
			set_is_fixed_in_y_internal (b)
		end

	set_is_movable_in_x (b: BOOLEAN)
			-- Set `is_movable_in_x' to `b'
		do
			set_is_movable_in_x_internal (b)
		end

	set_is_movable_in_y (b: BOOLEAN)
			-- Set `is_movable_in_y' to `b'
		do
			set_is_movable_in_y_internal (b)
		end

	set_is_rotatable (b: BOOLEAN)
			-- Set `is_rotatable' to `b'
		do
			set_is_rotatable_internal (b)
		end

	set_is_duplicatable (b: BOOLEAN)
			-- Set `is_duplicatable' to `b'
		do
			set_is_duplicatable_internal (b)
		end

	set_is_deletable (b: BOOLEAN)
			-- Set `is_deletable' to `b'
		do
			set_is_deletable_internal (b)
		end

	set_rotation (n: INTEGER)
			-- Set rotation of `Current' to `n' degrees clockwise
		do
			die.set_rotation (n)
		end

	toggle_rotation
			-- Toggle `is_rotated'
		do
			die.toggle_rotation
		end

	rotate_90_clockwise
			-- Rotate `Current' 90 degrees clockwise
		do
			die.rotate_90_clockwise
		end

	rotate_90_counterclockwise
			-- Rotate `Current' 90 degrees counterclockwise
		do
			die.rotate_90_counterclockwise
		end

	move_by (a_distance: CONST_POINT)
			-- Move `Current' by `a_distance'
		do
			die_location.move_by (a_distance)
			round_edge_to_grid
			set_closest_field
		end

	round_edge_to_grid
			-- Move lower left corner of `Current' to nearest
			-- grid point.  Used to prevent accumulation of
			-- errors from multiple moves
		local
			diff_x, diff_y: DOUBLE
		do
			diff_x := grid.rounded_edge (left_x) - left_x
			diff_y := grid.rounded_edge (bottom_y) - bottom_y
			die_location.set_location_x (grid.rounded_center (die_location.location_x + diff_x))
			die_location.set_location_y (grid.rounded_center (die_location.location_y + diff_y))
		end

	duplicate
			-- Duplicate element and add it to model
		local
			dl: like die_location
		do
			die.add_die_location (die_location.field_name, die_location.location_x, die_location.location_y)
			dl := die.last_die_location
			model.add_die_location (die, dl)
		end

	move_left_edge (a_distance: DOUBLE)
			-- Move left edge of `Current' by `a_distance'
		do
			die.set_die_size_printed_x (grid.rounded_size (die.die_size_printed_x - a_distance))
			die_location.set_location_x (die_location.location_x + a_distance / 2.0)
			round_edge_to_grid
			set_closest_field
		end

	move_right_edge (a_distance: DOUBLE)
			-- Move right edge of `Current' by `a_distance'
		do
			die.set_die_size_printed_x (grid.rounded_size (die.die_size_printed_x + a_distance))
			die_location.set_location_x (die_location.location_x + a_distance / 2.0)
			round_edge_to_grid
			set_closest_field
		end

	move_top_edge (a_distance: DOUBLE)
			-- Move top edge of `Current' by `a_distance'
		do
			die.set_die_size_printed_y (grid.rounded_size (die.die_size_printed_y + a_distance))
			die_location.set_location_y (die_location.location_y + a_distance / 2.0)
			round_edge_to_grid
			set_closest_field
		end

	move_bottom_edge (a_distance: DOUBLE)
			-- Move bottom edge of `Current' by `a_distance'
		do
			die.set_die_size_printed_y (grid.rounded_size (die.die_size_printed_y - a_distance))
			die_location.set_location_y (die_location.location_y + a_distance / 2.0)
			round_edge_to_grid
			set_closest_field
		end

feature -- Contract support

	is_valid_for_creation (a_die: like die): BOOLEAN
			-- Does `a_die' have a die type that is valid
			-- for creation of `Current'?
		require
			a_die_not_void: a_die /= Void
		deferred
		end

feature {NONE} -- Implementation

	set_is_fixed_in_x_internal (b: BOOLEAN)
			-- Set `is_fixed_in_x_internal' to `b'
		do
			die_location.set_is_fixed_in_x (b)
		end

	set_is_fixed_in_y_internal (b: BOOLEAN)
			-- Set `is_fixed_in_y_internal' to `b'
		do
			die_location.set_is_fixed_in_y (b)
		end

	set_is_movable_in_x_internal (b: BOOLEAN)
			-- Set `is_movable_in_x_internal' to `b'
		do
			die_location.set_is_movable_in_x (b)
		end

	set_is_movable_in_y_internal (b: BOOLEAN)
			-- Set `is_movable_in_y_internal' to `b'
		do
			die_location.set_is_movable_in_y (b)
		end

	set_is_rotatable_internal (b: BOOLEAN)
			-- Set `is_rotatable_internal' to `b'
		do
			die_location.set_is_rotatable (b)
		end

	set_is_duplicatable_internal (b: BOOLEAN)
			-- Set `is_duplicatable_internal' to `b'
		do
			die_location.set_is_duplicatable (b)
		end

	set_is_deletable_internal (b: BOOLEAN)
			-- Set `is_deletable_internal' to `b'
		do
			die_location.set_is_deletable (b)
		end

	y_coordinate_translated_to_field (a_field: RETICLE_FIELD): DOUBLE
			-- Y coordinate of field-relative center
			-- translated from its value for current associated
			-- field to the value it would have if associated
			-- field was changed to `a_field'
		require
			a_field_not_void: a_field /= Void
		local
			old_field: like field
		do
			old_field := field
			Result := die_location.location_y + field.center_y - a_field.center_y
		end

	set_closest_field
			-- Set `field' to closest field
		local
			fld_name: STRING
			old_field, new_field: like field
			field_changed: BOOLEAN
		do
			temp_point.make_xy (center_x, center_y)
			fld_name := model.closest_field (temp_point).name
			field_changed := not die_location.field_name.is_equal (fld_name)
			if field_changed then
				-- Set new field name and adjust
				-- center so that it is relative to
				-- new closest field
				old_field := field
				die_location.set_field_name (fld_name)
				uncache_field
				new_field := field
				temp_point.make_xy (old_field.center_x, old_field.center_y)
				temp_point.move_by (-new_field.center_x, -new_field.center_y)
				die_location.move_by (temp_point)
			end
		end

	update_die (an_original, a_replacement: like die)
			-- If `die' equals `an_original', replace it
			-- with `a_replacement'
		do
			if die = an_original then
				die := a_replacement
			end
		end

	update_die_location (an_original, a_replacement: like die_location)
			-- If `die_location' equals `an_original', replace it
			-- with `a_replacement'
		do
			if die_location = an_original then
				die_location := a_replacement
			end
		end

	associated_field_name: STRING
		do
			Result := die_location.field_name
		end
	
	temp_point: POINT
			-- Temporary to avoid repeated object creation
	
invariant
	die_not_void: die /= Void
	die_location_not_void: die_location /= Void

end
