
indexing
	
	description: "Project element of reticle model (cust die or test vehicle)"
	author: "David Hollenberg"
	date: "July 28, 2006"

deferred class RETICLE_PROJECT
inherit
	RETICLE_DIE
		redefine
			design_name, die_id, account_number, account_type,
			organization_short_id, effective_overhead_style,
			effective_overhead_style_description, is_commercial_project,
			is_mep_project, is_other_project, is_project,
			max_die_size, is_oversize_die_always_allowed,
			is_oversize_die_possibly_allowed, 
			oversize_die_inclusion_conditions, info_name, 
			packaging_description, has_ceramic_packaging, 
			has_plastic_packaging, design_number, 
			uniform_white_space_border,
			needs_min_sep_white_space, can_be_isolated
		end

feature -- Properties

	identification: STRING
			-- Identifying name of element
		do
			Result := project.design_no.out
		end

	design_number: INTEGER
			-- Design number of element
		do
			Result := project.design_no
		end

	design_name: STRING
			-- Design name of element or identifying name if
			-- no associated design name
		do
			Result := project.design_name
		end

	die_id: STRING
			-- Die ID of element or empty string if
			-- no associated die ID
		do
			Result := die.die_id
		end

	info_name: STRING
			-- Name of element for packer information window
		do
			Result := project.design_name
		end

	account_number: STRING
			-- Account number of element or identifying name if
			-- no associated account number
		do
			Result := project.cust_acct_no.out
		end

	account_type: STRING
			-- Account type of element or identifying name if
			-- no associated account type
		local
			acct: CUST_ACCTS_ROW
			acct_type, acct_sub_type, acct_sponsor: STRING
			type: CUST_ACCT_TYPE_VALUES
			subtype: ACCT_SUBTYPE_VALUES
			sponsor: ACCT_SPONSOR_VALUES
		do
			acct := project.cust_account
			acct_type := acct.acct_type
			acct_sub_type := acct.acct_sub_type
			acct_sponsor := acct.acct_sponsor_id
			if acct_sponsor.is_equal (sponsor.Mep) then
				Result := acct_sponsor
			elseif acct_type.is_equal (type.Com) then
				if acct_sub_type.is_equal (subtype.Univ) or acct_sub_type.is_equal (subtype.Expuniv) then
					Result := acct_type + Account_field_separator + acct_sub_type
				else
					Result := acct_type
				end
			else
				Result := acct_sponsor.twin
				if not Result.is_empty then
					Result.append (Account_field_separator)
				end
				Result.append (acct_type)
				if not acct_sub_type.is_empty then
					Result.append (Account_field_separator)
					Result.append (acct_sub_type)
				end
			end
		end

	organization_short_id: STRING
			-- Short name of organization type for element or
			-- identifying name if no associated organization
		local
			acct: CUST_ACCTS_ROW
		do
			acct := project.cust_account
			Result := acct.organization.short_id
		end

	is_commercial_project: BOOLEAN
			-- Is element a commercial project?
		local
			acct: CUST_ACCTS_ROW
			acct_type, acct_sponsor: STRING
			type: CUST_ACCT_TYPE_VALUES
			sponsor: ACCT_SPONSOR_VALUES
		do
			acct := project.cust_account
			acct_type := acct.acct_type
			acct_sponsor := acct.acct_sponsor_id
			Result := not acct_sponsor.is_equal (sponsor.Mep) and
				acct_type.is_equal (type.Com)
		end

	is_mep_project: BOOLEAN
			-- Is element an MEP project?
		local
			acct: CUST_ACCTS_ROW
			acct_sponsor: STRING
			sponsor: ACCT_SPONSOR_VALUES
		do
			acct := project.cust_account
			acct_sponsor := acct.acct_sponsor_id
			Result := acct_sponsor.is_equal (sponsor.Mep)
		end

	is_other_project: BOOLEAN
			-- Is element a project whose account type is not
			-- commercial or MEP?
		do
			Result := not is_commercial_project and not is_mep_project
		end

	is_project: BOOLEAN = True
			-- Is element a project?

	can_be_isolated: BOOLEAN = True
			-- Is `Current' allowed to be isolated, based on its
			-- type?

	max_die_size: CONST_POINT
			-- Maximum die size, as constrained by package
			-- cavity sizes and user-specified limits.
			-- Void indicates no maximum
		do
			Result := project.max_die_size
		end

	is_oversize_die_always_allowed: BOOLEAN
			-- Is a die that is bigger than `max_die_size'
			-- allowed unconditionally?
		do
			Result := project.is_oversize_die_always_allowed
		end

	is_oversize_die_possibly_allowed: BOOLEAN
			-- Is a die that is bigger than `max_die_size'
			-- allowed but only under specific conditions?
		do
			Result := project.is_oversize_die_possibly_allowed
		end

	oversize_die_inclusion_conditions: STRING
			-- Description of conditions under which 
			-- die that is bigger than `max_die_size'
			-- allowed
		do
			Result := project.oversize_die_inclusion_conditions
		end

	effective_overhead_style: STRING
			-- Effective overhead style for element
		do
			if not model.global_overhead_style.is_empty then
				Result := model.global_overhead_style
			else
				Result := project.overhead_style
			end
		end

	effective_overhead_style_description: STRING
			-- Description of effective overhead style for
			-- element
		do
			Result := effective_overhead_style
			if not Result.is_equal (project.overhead_style) then
				Result := Result + " (" + project.overhead_style + ")"
			end
		end

	packaging_description: STRING
			-- Description of packaging for element
		do
			Result := project.packaging_description
		end

	has_ceramic_packaging: BOOLEAN
			-- Will element be packaged in at least one ceramic
			-- package?
		do
			Result := project.has_ceramic_packaging
		end

	has_plastic_packaging: BOOLEAN
			-- Will element be packaged in at least one plastic
			-- package?
		do
			Result := project.has_plastic_packaging
		end

	uniform_white_space_border: DOUBLE
			-- Uniform amount of white space around element
		do
			Result := die.uniform_white_space_border
		end

	needs_min_sep_white_space: BOOLEAN = True
			-- Does element need "minimum separation" amount of
			-- white space around it?

feature {NONE} -- Implementation

	Account_field_separator: STRING = "-"
			-- Separator between parts of account identifying name

end
