class RFC822_PARSERS

inherit
	MESSAGE_SUPPORT

	-- This class is just a set of routines to enable the parsing of
	-- RFC822 style message header fields (or similar strings).
	-- There are a few classes of routines, each with a similar interface.

	-- is_xxxx(ch: CHARACTER): BOOLEAN
	--  determines whether ch is a member of the character class xxxx

	-- parse_xxxx(str: STRING  pos: INTEGER): INTEGER
	--  determines whether or not there is an instance of the construct
	--  xxxx in str beginning at pos.
	-- if there is, it returns the string position just beyond the end
	--  of the xxxx it found (may be off the string if xxxx used it up)
	-- if there is not, it returns 0
	-- these routines all require that (pos > 0) (but it may be after
	--  the string)
	-- these routines all ensure that (Result = 0 or Result > pos)

	-- parse0_xxxx(str: STRING  pos: INTEGER): INTEGER
	--  like the above, except that upon failure, it returns pos
	-- these routines all ensure that (Result >= pos)

feature  -- address collection and group options

	-- the parsers at the level of addr-spec and above (route-addr, 
	--  mailbox[es], group, address[es]) take two additional arguments
	--  ...; addrs: LIST[STRING]; addr_mode: INTEGER)
	-- if addrs is not Void, then each mailbox is appended to addrs as
	--  it is parsed (as a new string); the parsing is not affected.
	-- (if addrs is Void, the routines only parse)

	-- the addr_mode variables are defined in MESSAGE_SUPPORT

feature  -- character classifiers

	is_lwsp(ch: CHARACTER): BOOLEAN is
		-- is ch an RFC822 linear white space character?
	do  Result := (ch = ' ' or ch = '%T');
	end;

  --	is_alpha(ch: CHARACTER): BOOLEAN is
  --	do  Result := ch.is_alpha
  --	end;

  --	is_digit(ch: CHARACTER): BOOLEAN is
  --	do  Result := ch.is_digit
  --	end;

	is_special(ch: CHARACTER): BOOLEAN is
		-- is ch an RFC822 special character?
	do  inspect ch
	    when '(', ')', '<', '>', '@',
		 ',', ';', ':', '\', '%"',
		 '.', '[', ']'
		then Result := true
	    else
	    end;
	end;

	is_control(ch: CHARACTER): BOOLEAN is
		-- is ch an RFC822 control character?
	do  inspect ch
	    when '%/0/'..'%/31/','%/127/'
		then Result := true
	    else
	    end;
	end;

feature  -- parsers

	parse_char(str: STRING; pos: INTEGER; ch: CHARACTER): INTEGER is
	require str /= Void and pos > 0
	do  if pos <= str.count and then (str @ pos) = ch
	     then Result := pos + 1;
	     else Result := 0;
	    end;
	ensure
	    Result = 0 or Result > pos
	end;

	parse_alphas(str: STRING; pos: INTEGER): INTEGER is
        require str /= Void and pos > 0
	local len: INTEGER;
	do  len := str.count;
	    from  Result := pos
	    until Result > len or else not (str @ Result).is_alpha
	    loop  Result := Result + 1;
	    end;
	    if Result = pos then Result := 0; end;
        ensure
            Result = 0 or Result > pos
        end;

	parse_digits(str: STRING; pos: INTEGER): INTEGER is
        require str /= Void and pos > 0
	local len: INTEGER;
	do  len := str.count;
	    from  Result := pos
	    until Result > len or else not (str @ Result).is_digit
	    loop  Result := Result + 1;
	    end;
	    if Result = pos then Result := 0; end;
        ensure
            Result = 0 or Result > pos
        end;

	parse_n_digits(str: STRING; pos: INTEGER; n: INTEGER): INTEGER is
        require str /= Void and pos > 0
	do  Result := parse_digits(str, pos);
	    if (Result - pos) /= n then Result := 0; end;
        ensure
            Result = 0 or Result > pos
        end;

	parse_atom(str: STRING; pos: INTEGER): INTEGER is
		-- 1*<any chars except special or control or space>
	require str /= Void and pos > 0
	local ch: CHARACTER; len: INTEGER; not_atomic: BOOLEAN
	do  from  Result := pos;
		  len := str.count;
	    until Result > len or not_atomic
	    loop  ch := str @ Result;
		  if is_special(ch) or is_control(ch) or ch = ' '
		   then not_atomic := true;
		   else Result := Result + 1;
		  end;
	    end;
	    if Result = pos then Result := 0; end	-- (empty atom)
	ensure
	    Result = 0 or Result > pos
	end;

	parse_quoted_string(str: STRING; pos: INTEGER): INTEGER is
		-- <"> *(qtext | quoted-pair) <">
	require str /= Void and pos > 0
	local ch: CHARACTER; len: INTEGER;
	do  len := str.count;
	    if pos < len and then (str @ pos) = '%"' then
		from  Result := pos+1;
		until Result > len or ch = '%"'
		loop  ch := str @ Result;
		      if ch = '\'
			then Result := Result + 2  -- quoted pair
			else Result := Result + 1
		      end;
		end;
		if ch /= '%"' then Result := 0; end;  -- oops (no close quote)
	    end;
	ensure
	    Result = 0 or Result > pos
	end;

	parse_word(str: STRING; pos: INTEGER): INTEGER is
		-- atom | quoted string
	require str /= Void and pos > 0
	do  Result := parse_atom(str, pos);
	    if Result = 0 then Result := parse_quoted_string(str, pos); end
	ensure
	    Result = 0 or Result > pos
	end;

	parse_domain_literal(str: STRING; pos: INTEGER): INTEGER is
		-- "[" *( dtext | quoted-pair ) "]"
	require str /= Void and pos > 0
	local ch: CHARACTER; len: INTEGER;
	do  len := str.count;
	    if pos < len and then (str @ pos) = '[' then
		from  Result := pos+1;
		until Result > len or ch = '[' or ch = ']'
		loop  ch := str @ Result;
		      if ch = '\' then Result := Result + 2  -- quoted pair
				  else Result := Result + 1
		      end;
		end;
		if ch /= ']' then Result := 0; end;	-- oops
	    end;
	ensure
	    Result = 0 or Result > pos
	end;

	parse_subdomain(str: STRING; pos: INTEGER): INTEGER is
		-- domain-ref | domain-literal
		-- domain-ref :: atom
	require str /= Void and pos > 0
	do  Result := parse_atom(str, pos);
	    if Result = 0 then Result := parse_domain_literal(str, pos); end;
	ensure
	    Result = 0 or Result > pos
	end;

	parse_dns_subdomain(str: STRING; pos: INTEGER): INTEGER is
		-- leters, digits, and hyphen (but no hyphen at ends)
		-- (this is more restrictive than RFC822)
	require str /= Void and pos > 0
	local len: INTEGER; ch, lch: CHARACTER; done: BOOLEAN;
	do  from  Result := pos;
		  len := str.count;
	    until Result > len or done
	    loop  ch := str @ Result;
		  if ch.is_alpha or ch.is_digit or
		     (ch = '-' and Result > pos)	-- not first char
		   then Result := Result + 1;
			lch := ch			-- last char accepted
		   else done := true
		  end;
	    end;
	    if Result = pos			-- no subdomain
		or Result > (pos + 63)		-- too long for subdomain
		or lch = '-'			-- terminal '-'
	     then Result := 0
	    end;
	ensure
	    Result = 0 or Result > pos
	end;

	parse_comment(str: STRING; pos: INTEGER): INTEGER is
		-- "(" *( ctext | quoted-pair | comment ) ")"
	require str /= Void and pos > 0
	local ch: CHARACTER; len, lcnt: INTEGER;
	do  len := str.count;
	    if pos < len and then (str @ pos) = '(' then
		from  Result := pos+1;
		      lcnt := 1;
		until Result > len or lcnt = 0
		loop  ch := str @ Result;
		      if ch = '(' then lcnt := lcnt + 1; end;
		      if ch = ')' then lcnt := lcnt - 1; end;
		      if ch = '\' then Result := Result + 2  -- quoted pair
				  else Result := Result + 1
		      end;
		end;
		if lcnt /= 0 then Result := 0; end;	-- oops
	    end;
	ensure
	    Result = 0 or Result > pos
	end;

	parse0_lwsp(str: STRING; pos: INTEGER): INTEGER is
		-- 0*( lwsp-char )
		-- (this one may parse nothing, returning pos)
	require str /= Void and pos > 0
	local ch: CHARACTER;  len: INTEGER;  stop: BOOLEAN;
	do  from  Result := pos;
		  len := str.count;
	    until Result > len or stop
	    loop  ch := str @ Result;
		  if is_lwsp(ch)
		   then Result := Result + 1;
		   else stop := true
		  end
	    end
	ensure
	    Result >= pos
	end;

	parse0_lwsp_or_comment(str: STRING; pos: INTEGER): INTEGER is
		-- 0*( lwsp-char | comment )
		-- (this one may parse nothing, returning pos)
	require str /= Void and pos > 0
	local ch: CHARACTER;  len, cend: INTEGER;  stop: BOOLEAN;
	do  from  Result := pos;
		  len := str.count;
	    until Result > len or stop
	    loop  ch := str @ Result;
		  if is_lwsp(ch)
		   then Result := Result + 1;
		   else cend := parse_comment(str, Result);
			if cend > 0
			 then Result := cend;
			 else stop := true
			end
		  end
	    end
	ensure
	    Result >= pos
	end;

	parse_lwsp_and_char(str: STRING; pos: INTEGER; ch: CHARACTER):
				INTEGER is
		-- (common combination)
        require str /= Void and pos > 0
	do  Result := parse0_lwsp_or_comment(str, pos);
	    Result := parse_char(str, Result, ch);
        ensure
            Result = 0 or Result > pos
        end;

	parse_lwsp_or_comment(str: STRING; pos: INTEGER): INTEGER is
		-- 1*( lwsp-char | comment )
	require str /= Void and pos > 0
	do  Result := parse0_lwsp_or_comment(str, pos);
	    if Result = pos then Result := 0; end;
	ensure
	    Result = 0 or Result > pos
	end;

	parse_phrase(str: STRING; pos: INTEGER): INTEGER is
		-- 1* word
	require str /= Void and pos > 0
	local wend: INTEGER;
	do  from   wend := parse_word(str, pos);
	    until  wend = 0
	    loop   Result := wend;	-- without trailing delimiter
		   wend := parse0_lwsp_or_comment(str, wend);
		   wend := parse_word(str, wend);
	    end
	ensure
	    Result = 0 or Result > pos
	end;

	parse_local_part(str: STRING; pos: INTEGER): INTEGER is
		-- word *("." word)
	require str /= Void and pos > 0
	local wend: INTEGER;
	do  from  wend := parse_word(str, pos);
	    until wend = 0
	    loop  Result := wend;	-- without trailing delimiter
		  wend := parse_lwsp_and_char(str, wend, '.');
		  if wend > 0
		   then wend := parse0_lwsp_or_comment(str, wend);
			wend := parse_word(str, wend);
		  end
	    end;
	ensure
	    Result = 0 or Result > pos
	end;

	parse_domain(str: STRING; pos: INTEGER): INTEGER is
		-- sub-domain *("." sub-domain)
		-- changed (7/29/99) to
		--   dns-subdomain *("." dns-subdomain)
	require str /= Void and pos > 0
	local wend: INTEGER;
	do  from  wend := parse_dns_subdomain(str, pos);
	    until wend = 0
	    loop  Result := wend;	-- without trailing delimiter
		  wend := parse_lwsp_and_char(str, wend, '.');
		  if wend > 0
		   then wend := parse0_lwsp_or_comment(str, wend);
			wend := parse_dns_subdomain(str, wend);
		  end
	    end;
	ensure
	    Result = 0 or Result > pos
	end;

	parse_addr_spec(str: STRING; pos: INTEGER;
		addrs: LIST[STRING]): INTEGER is
		-- local-part "@" domain
	require str /= Void and pos > 0
	do  Result := parse_local_part(str, pos);
	    if Result > 0
	     then Result := parse_lwsp_and_char(str, Result, '@');
	    end;
	    if Result > 0
	     then Result := parse0_lwsp_or_comment(str, Result);
		  Result := parse_domain(str, Result);
	    end;
	    if Result > 0
	      then extend_addrs(addrs, str, pos, Result-1);
	    end;
	ensure
	    Result = 0 or Result > pos
	end;

	parse_route(str: STRING; pos: INTEGER): INTEGER is
		-- 1#("@" domain) ":"
	require str /= Void and pos > 0
	local
	do
	ensure
	    Result = 0 or Result > pos
	end;	

	parse_route_addr(str: STRING; pos: INTEGER;
		addrs: LIST[STRING]): INTEGER is
		-- "<" [route] addr_spec ">"
	require str /= Void and pos > 0
	local rend: INTEGER;
	do  Result := parse_char(str, pos, '<');
	    if Result > 0
	     then Result := parse0_lwsp_or_comment(str, Result);
		  rend := parse_route(str, Result);
		  if rend > 0		-- route is optional
		   then Result := parse0_lwsp_or_comment(str, rend); end;
		  Result := parse_addr_spec(str, Result, addrs);
	    end;
	    if Result > 0
	     then Result := parse_lwsp_and_char(str, Result, '>');
	    end;
	ensure
	    Result = 0 or Result > pos
	end;

	parse_mailbox(str: STRING; pos: INTEGER;
		addrs: LIST[STRING]; addr_mode: INTEGER): INTEGER is
		-- addr_spec | phrase  route_addr
		-- this one picks up leading and trailing lwsp_or_comment
	require str /= Void and pos > 0
	local pos2: INTEGER;  addrs2: LIST[STRING];
	do  pos2 := parse0_lwsp_or_comment(str, pos);
	    if addr_mode = save_addr_specs
	     then addrs2 := addrs; end		-- else leave Void
	    Result := parse_addr_spec(str, pos2, addrs2);
	    if Result = 0 then
		Result := parse_phrase(str, pos2);
		if Result > 0 then
		   Result := parse0_lwsp_or_comment(str, Result);
		   Result := parse_route_addr(str, Result, addrs2);
		end;
	    end;
	    if Result > 0
	     then Result := parse0_lwsp_or_comment(str, Result);
		  if addr_mode /= save_addr_specs
		   then extend_addrs(addrs, str, pos, Result-1);
		  end;
	    end;
        ensure
            Result = 0 or Result > pos
        end;

	parse0_mailboxes(str: STRING; pos: INTEGER;
		addrs: LIST[STRING]; addr_mode: INTEGER): INTEGER is
		-- #(mailbox)
	require str /= Void and pos > 0
	local npos: INTEGER;
	do  from  npos := pos;
	    until npos = 0	-- meaning no comma found last try
	    loop  Result := npos;
		  npos := parse_mailbox(str, Result, addrs, addr_mode);
		  if npos > 0		-- a mailbox
		   then Result := npos;
		  end;
		  npos := parse_lwsp_and_char(str, Result, ',');
	    end;
	ensure
	    Result >= pos
	end;	

	parse_group(str: STRING; pos: INTEGER;
		addrs: LIST[STRING]; addr_mode: INTEGER): INTEGER is
		-- phrase ":" [#mailbox] ";"
		-- this one picks up leading and trailing lwsp_or_comment
	require str /= Void and pos > 0
	local addrs_for_boxes: LIST[STRING];
	do  Result := parse0_lwsp_or_comment(str, pos);
	    Result := parse_phrase(str, Result);
	    if Result > 0
	     then Result := parse_lwsp_and_char(str, Result, ':');
	    end;
	    if Result > 0
	     then if addr_mode = expand_groups or
		     addr_mode = save_addr_specs
		   then addrs_for_boxes := addrs; end;  -- else leave Void
		  Result := parse0_mailboxes(str, Result, 
					addrs_for_boxes, addr_mode);
		  Result := parse_char(str, Result, ';');
	    end;
	    if Result > 0 and addr_mode = retain_groups
	     then extend_addrs(addrs, str, pos, Result-1);
	    end;
	ensure
	    Result = 0 or Result > pos
	end;	

	parse_address(str: STRING; pos: INTEGER;
		addrs: LIST[STRING]; addr_mode: INTEGER): INTEGER is
		-- mailbox | group
	require str /= Void and pos > 0
	do  Result := parse_mailbox(str, pos, addrs, addr_mode);
	    if Result = 0
	     then Result := parse_group(str, pos, addrs, addr_mode);
	    end;
	ensure
	    Result = 0 or Result > pos
	end;	

	parse0_addresses(str: STRING; pos: INTEGER;
                addrs: LIST[STRING]; addr_mode: INTEGER): INTEGER is
		-- # address
        require str /= Void and pos > 0
	local npos: INTEGER;
	do  from  npos := pos;
	    until npos = 0	-- meaning no comma found last try
	    loop  Result := npos;
		  npos := parse_address(str, Result, addrs, addr_mode);
		  if npos > 0		-- a mailbox
		   then Result := npos;
		  end;
		  npos := parse_lwsp_and_char(str, Result, ',');
	    end;
        ensure
            Result >= pos
        end;

	parse_msg_id(str: STRING; pos: INTEGER): INTEGER is
		-- "<" addr_spec ">"
	require str /= Void and pos > 0
	do  Result := parse_char(str, pos, '<');
	    if Result > 0
	     then Result := parse0_lwsp_or_comment(str, Result);
		  Result := parse_addr_spec(str, Result, Void);
	    end;
	    if Result > 0
	     then Result := parse_lwsp_and_char(str, Result, '>');
	    end;
	ensure
	    Result = 0 or Result > pos
	end;

	parse_day(str: STRING; pos: INTEGER): INTEGER is
		-- Mon | Tue | ... | Sun
        require str /= Void and pos > 0
	local d: INTEGER;
        do  Result := parse_alphas(str, pos);
	    if Result > 0
	     then d := matches_one(str, pos, Result-1, days_of_the_week);
		  if d = 0 then Result := 0; end;
	     end;
        ensure
            Result = 0 or Result > pos
        end;

	parse_date(str: STRING; pos: INTEGER): INTEGER is
		-- 1*2DIGIT month (2DIGIT | 4DIGIT)
        require str /= Void and pos > 0
	local m: INTEGER;
        do  Result := parse_digits(str, pos);
	    if Result > 0 and Result <= pos+2	-- 1*2DIGIT
	     then m := parse0_lwsp(str, Result);
		  Result := parse_alphas(str, m);
	     else Result := 0;
	    end
            if Result > 0
             then m := matches_one(str, m, Result-1, months_of_the_year);
                  if m = 0 then Result := 0; end;
             end;
	    if Result > 0
	     then m := parse0_lwsp(str, Result);	-- start of year
		  Result := parse_digits(str, m);
		  if Result = m+2 or Result = m+4
		   then 
		   else Result := 0;
		  end
	    end;
        ensure
            Result = 0 or Result > pos
        end;

	time_zones: ARRAY[STRING] is
	once  Result := 
		<<"UT",  "GMT", "EST", "EDT",
		  "CST", "CDT", "MST", "MDT", "PST", "PDT">>
	end;

	parse_zone(str: STRING; pos: INTEGER): INTEGER is
		-- "UT" | ... | ("+"|"-") 4DIGIT
	require str /= Void and pos > 0
	do  Result := parse_alphas(str, pos);	-- named time zone
	    if Result > 0
	     then if Result > pos+1 and then	-- any one alpha is OK
		     matches_one(str, pos, Result-1, time_zones) = 0
		   then Result := 0
		  end
	     else				-- numeric differential
		  Result := parse_char(str, pos, '+') +
			    parse_char(str, pos, '-');
		  if Result > 0
		   then Result := parse0_lwsp(str, Result);
			Result := parse_n_digits(str, Result, 4);
		  end;
	    end;
        ensure
            Result = 0 or Result > pos
        end;

	parse_time(str: STRING; pos: INTEGER): INTEGER is
		-- 2DIGIT ":" 2DIGIT [ ":" 2DIGIT ]
        require str /= Void and pos > 0
	local opt: INTEGER;
        do  Result := parse_n_digits(str, pos, 2);	-- hh
	    if Result > 0
	     then Result := parse_lwsp_and_char(str, Result, ':');
	    end;
	    if Result > 0
	     then Result := parse0_lwsp(str, Result);
		  Result := parse_n_digits(str, Result, 2);  -- mm
	    end;
	    if Result > 0
	     then opt := parse_lwsp_and_char(str, Result, ':');
	    end;
	    if opt > 0
	     then opt := parse0_lwsp(str, opt);
		  opt := parse_n_digits(str, opt, 2);	-- ss
	    end;
	    if opt > 0 then Result := opt; end;
	    if Result > 0 
	     then Result := parse0_lwsp(str, Result);
		  Result := parse_zone(str, Result);
	     end;
        ensure
            Result = 0 or Result > pos
        end;

	parse_date_time(str: STRING; pos: INTEGER): INTEGER is
		-- [ day "," ] date time
        require str /= Void and pos > 0
	local m: INTEGER;
	do  -- optional day part
	    Result := parse_day(str, pos);
	    if Result > 0
             then -- make comma optional for old systems [non-822]
		  m := parse_lwsp_and_char(str, Result, ',');
		  if m > 0 then Result := m; end;
	     end;
	    if Result > 0 then Result := parse0_lwsp(str, Result); end;
	    if Result = 0 then Result := pos; end;	-- no day part; reset
	    -- date and time part
	    Result := parse_date(str, Result);
	    if Result > 0
             then Result := parse0_lwsp_or_comment(str, Result);
		  Result := parse_time(str, Result);
	     end;
        ensure
            Result = 0 or Result > pos
	end;

feature {NONE}

	extend_addrs(a: LIST[STRING]; src: STRING; n1,n2: INTEGER) is
	local s: STRING;
	do  if a /= Void
	     then s := src.substring(n1,n2);
		  s.left_adjust;		-- clear lwsp
		  s.right_adjust;
		  a.extend(s);			-- add to end
	    end;
	end;

	matches_one(str: STRING; n1, n2: INTEGER; master: ARRAY[STRING]):
			INTEGER is
	local  it: STRING;  i: INTEGER;
	do  it := str.substring(n1, n2);
	    it.to_upper;
	    from  i := master.lower;
	    until i > master.upper or Result > 0
	    loop  if it.is_equal(master @ i)
		   then Result := i;
		   else i := i + 1
		  end;
	    end;
	end;

	parse_xxxx(str: STRING; pos: INTEGER): INTEGER is
		-- this is just a template for the generic parser routine
        require str /= Void and pos > 0
        do
        ensure
            Result = 0 or Result > pos
        end;

feature	-- a couple more (not really RFC822) parsers

	parse0_noise(str: STRING; pos: INTEGER; xtra: STRING): INTEGER is
		-- noise is all sequential characters that are either
		-- lwsp or special (a la RFC822); however,
		-- characters in the xtra string have their noisiness
		--  reversed from the default
	require str /= Void and pos > 0 and xtra /= Void
	local ch: CHARACTER;  len: INTEGER;  stop: BOOLEAN;
	do  from  Result := pos;
		  len := str.count;
	    until Result > len or stop
	    loop  ch := str @ Result;
		  if (is_lwsp(ch) or is_special(ch)) /= xtra.has(ch)
		   then Result := Result + 1;
		   else stop := true
		  end
	    end
	ensure
	    Result >= pos
	end;

	parse_integer(str: STRING; pos: INTEGER): INTEGER is
		-- [-]<digits>
	require str /= Void and pos > 0
	do  Result := parse_char(str, pos, '-');
	    if Result = 0 then Result := pos end;	-- optional -
	    Result := parse_digits(str, Result)
	ensure
	    Result = 0 or Result > pos
	end;

	parse_real(str: STRING; pos: INTEGER): INTEGER is
		-- [-][<digits>][.[<digits>]], but one digit required
	require str /= Void and pos > 0
	local ch: CHARACTER;  len: INTEGER;  stop, dot, dig: BOOLEAN;
	do  from  Result := pos;
                  len := str.count;
            until Result > len or stop
            loop  ch := str @ Result;
                  if ch.is_digit
		   then Result := Result + 1;  dig := true;
		  elseif ch = '.' and not dot
		   then Result := Result + 1;  dot := true;
		  elseif ch = '-' and Result = pos	-- at head only
		   then Result := Result + 1;
                   else stop := true
                  end
	    end;
	    if not dig then Result := 0; end	-- require one digit
	ensure
	    Result = 0 or Result > pos
	end;

end -- RFC822_PARSERS
