
indexing
	
	description: "An operation that creates a run"
	author: "David Hollenberg"
	date: "March 12, 2003"

class RUN_CREATION_OP
inherit
	OPERATION
		rename
			try_operation as try_to_create_run
		end
	RC_UTILITIES_ACCESS
		export
			{NONE} all
		end
	EXCEPTIONS_ACCESS
		export
			{NONE} all
		undefine
			default_create
		end
	DB_OPTIONS_ACCESS
		export
			{NONE} all
		undefine
			default_create
		end
	RC_SESSIONABLE
		export
			{NONE} set_session
		end
	RUN_QUERIES
		export
			{NONE} all
		end

create
	default_create

feature -- Operations

	create_run (s: STRING sess: RC_RUN_CLOSING_SESSION a_role: STRING)
			-- Create run with run id `s', using `sess' and
			-- with `a_role' as the only database role
		require
			name_exists: s /= Void
			session_exists: sess /= Void
			session_has_connection: sess.connection /= Void
			session_connection_open: sess.connection.is_open
			a_role_exists: a_role /= Void
		do
			session := sess
			run_id := s
			role := a_role
			execute
		end

feature {NONE} -- Implementation

	run_id: STRING
			-- Run id for which to try to create run

	role: STRING
			-- Role to enable while creating run

	is_run_dir_created: BOOLEAN
			-- Has run directory been created yet?

	try_to_create_run
			-- Try to create run
		local
			b: DB_BIND_S
			run_exists: BOOLEAN
		do
			is_run_dir_created := False
			session.set_role (role)
			session.check_valid_run_id (run_id)

			session.lock

			-- Check whether run already exists

			create b.make (run_id)
			run_exists := is_existing_run (run_id, connection)
			if run_exists then
				handle_user_error ("Run already exists: " + run_id)
			end
			if not session.environment.is_production then
				possibly_delete_run_dir
			end
			if not session.is_standard_run_id (run_id) then
				warn (session.run_reality_type + " run id does not conform to MOSIS naming conventions: " + run_id)
			end
	
			-- Create run directory and initial database rows
			-- for run
	
			session.open_log
			create_run_dir
			insert_initial_rows
			session.commit
			publish_created_event
			session.terminate
			inform (session.run_reality_type + " run " + run_id + " created")
		rescue
			if not exc.assertion_violation and is_run_dir_created then
				delete_run_dir
			end
		end

	possibly_delete_run_dir
			-- Delete the run directory tree, if it 
			-- already exists and user wants to delete it
		local
			remover: DIR_TREE_REMOVER
			lister: DIR_TREE_LISTER
			dir_name: STRING
			n: INTEGER
		do
			dir_name := session.run_directory
			if session.path_exists (dir_name) then
				create lister.make (dir_name)
				lister.set_files_excluded (False)
				lister.set_directories_excluded (False)
				lister.set_directory_symlinks_followed (False)
				lister.set_root_excluded (True)
				lister.make_pathname_list
				n := lister.path_names.count
				
				if question ("Run directory " + dir_name + " already exists and contains " + n.out + " files/directories.%NDelete it before proceeding?") then
					create remover.make (dir_name)
					remover.set_directory_symlinks_followed (False)
					remover.remove
					if session.path_exists (dir_name) then
						warn ("cannot remove entire directory tree " + dir_name + "%N" + remover.error_count.out + " errors encountered%N%NProceed anyway?")
					end
				end
			end
		end

	create_run_dir
			-- Create the run directory
		local
			cat: RC_FILE_CATEGORY
		do
			cat := session.file_generator.run_directory
			is_run_dir_created := True
			session.ensure_directory_exists (cat)
		end

	delete_run_dir
			-- Try to delete the run directory.  Ignore any
			-- errors
		local
			dir: LARGE_DIRECTORY
			tried: BOOLEAN
		do
			if not tried then
				create dir.make (session.run_directory)
				dir.delete
			end
		rescue
			tried := True
			retry
		end

	insert_initial_rows
			-- Insert row in RUNS table, PRODUCTIONS table
			-- and MASK_SETS table
		local
			r: RUNS_MROW
			p: PRODUCTIONS_MROW
			m: MASK_SETS_MROW
			prod_type: PRODUCTION_TYPE_VALUES
		do
			create r.make_by_key (connection, option.New_row, run_id)
			r.set_educational (False)
			r.set_wafer_scale_integration (False)
			r.set_is_trial_run (session.is_trial_run)
			
			create p.make_by_key (connection, option.New_row, run_id)
			p.set_run_id (run_id)
			p.set_production_type (prod_type.Initial_run)
			p.set_is_completed (False)
			p.set_production_charges_locked (False)
			
			create m.make_by_key (connection, option.New_row, run_id, rc.Initial_mask_revision)
			m.set_production_id (run_id)
		end

	publish_created_event
		local
			created_event: RUN_CREATED_EVENT
		do
			create created_event
			created_event.set_run_id (run_id)
			created_event.publish
		end

end
