
indexing
	
	description: "A map for a specific run"
	author: "David Hollenberg"
	date: "January 21, 2005"

deferred class RUN_MAP
inherit
	MOSIS_MAP
		redefine
			set_fonts
		end

feature {NONE} -- Creation

	make (run: STRING a_date: DB_DATE)
			-- Create with run id `run' and map date `a_date'
		require
			run_exists: run /= Void
			a_date_exists: a_date /= Void
		do
			make_with_date (a_date)
			run_id := run
		end

feature -- Properties

	run_id: STRING
			-- Run id of run map is for

	runs_row: RUNS_ROW
			-- Information about run

	field_names: LIST [STRING]
			-- List of field names for run, ordered by field name

	has_single_field: BOOLEAN
			-- Is there only a single field of view?
		do
			Result := field_names.count = 1
		end

	has_multiple_fields: BOOLEAN
			-- Are there multiple fields of view?
		do
			Result := not has_single_field
		end

feature -- Modification

	set_runs_row (r: like runs_row)
			-- Set `runs_row' to `r'
		require
			row_not_void: r /= Void
		do
			runs_row := r
		ensure
			value_set: runs_row = r
		end

	set_field_names (list: like field_names)
			-- Set `field_names' to `list'
		require
			list_not_void: list /= Void
		do
			field_names := list
		ensure
			field_names_set: field_names = list
		end

feature {NONE} -- Implementation

	set_fonts
			-- Set font used for text size calculations
			-- (not the same font as used in PostScript but
			-- should be a good guess)
		do
			create font
			font.set_height (20)
			precursor
		end

	optional_text_pair_in_box (s1, s2: STRING a_rect: DRW_BASIC_RECTANGLE rot: INTEGER): DRW_GROUP
			-- Strings `s1' and `s2' displayed nicely in `a_rect',
			-- where `rot' is the clockwise rotation of text
			-- in degrees.  If one of the strings is empty,
			-- display the other one nicely in `a_rect'
		require
			s1_exists: s1 /= Void
			s2_exists: s2 /= Void
			a_rect_exists: a_rect /= Void
			rot_valid: is_valid_rotation (rot)
		local
			rect: like a_rect
		do
			rect := a_rect.twin
			shrink_box (rect)
			if s1.is_empty and s2.is_empty then
				create Result.make (0)
			elseif s1.is_empty then
				Result := rotated_text_in_box (s2, rect, rot)
			elseif s2.is_empty then
				Result := rotated_text_in_box (s1, rect, rot)
			else
				Result := text_pair_in_box (s1, s2, a_rect, rot)
			end
		end

	text_pair_in_box (s1, s2: STRING a_rect: DRW_BASIC_RECTANGLE rot: INTEGER): DRW_GROUP
			-- Strings `s1' and `s2' displayed nicely in `a_rect',
			-- where `rot' is the clockwise rotation of text
			-- in degrees
		require
			s1_exists: s1 /= Void
			s2_exists: s2 /= Void
			a_rect_exists: a_rect /= Void
			rot_valid: is_valid_rotation (rot)
		local
			left_box, right_box: DRW_BASIC_RECTANGLE
			top_box, bottom_box: DRW_BASIC_RECTANGLE
			size: TUPLE [INTEGER, INTEGER]
			w1, w2, h1, h2: INTEGER
			sf1, sf2, horiz_sf, vert_sf: DOUBLE
			old_width, old_height: DOUBLE
			m: DRW_MATRIX
			rect: like a_rect
		do
			create Result.make (2)
			rect := a_rect.twin
			if rot = 90 or rot = 270 then
				old_width := rect.width
				old_height := rect.height
				rect.set_width (old_height)
				rect.set_height (old_width)
			end
			
			-- Compute sizes of strings using `font'
			
			size := font.string_size (s1)
			w1 := size.integer_item (1)
			h1 := size.integer_item (2)
			size := font.string_size (s2)
			w2 := size.integer_item (1)
			h2 := size.integer_item (2)
				
			-- Compute horizontal scale factor (`s1' in left half
			-- of `rect', `s2' in right half of `rect')
			
			left_box := rect.twin
			left_box.set_x (rect.x - rect.width / 4.0)
			left_box.set_width (rect.width / 2.0)
			shrink_box (left_box)
			sf1 := scale_factor_to_fit (w1, h1, left_box.width, left_box.height)
			right_box := rect.twin
			right_box.set_x (rect.x + rect.width / 4.0)
			right_box.set_width (rect.width / 2.0)
			shrink_box (right_box)
			sf2 := scale_factor_to_fit (w2, h2, right_box.width, right_box.height)
			horiz_sf := sf1.min (sf2)
			
			-- Compute vertical scale factor (`s1' in top half
			-- of `rect', `s2' in bottom half of `rect')
			
			bottom_box := rect.twin
			bottom_box.set_y (rect.y - rect.height / 4.0)
			bottom_box.set_height (rect.height / 2.0)
			shrink_box (bottom_box)
			sf1 := scale_factor_to_fit (w1, h1, bottom_box.width, bottom_box.height)
			top_box := rect.twin
			top_box.set_y (rect.y + rect.height / 4.0)
			top_box.set_height (rect.height / 2.0)
			shrink_box (top_box)
			sf2 := scale_factor_to_fit (w2, h2, top_box.width, top_box.height)
			vert_sf := sf1.min (sf2)
			
			if horiz_sf >= vert_sf then
				Result.extend (text_in_box (s1, left_box))
				Result.extend (text_in_box (s2, right_box))
			else
				Result.extend (text_in_box (s1, top_box))
				Result.extend (text_in_box (s2, bottom_box))
			end
			if rot /= 0 then
				create m.make_rotation_degrees_xy (- rot, rect.x, rect.y)
				Result.set_transform (m)
			end
		end

	append_std_title_info
			-- Generate standard title information and append
			-- to `last_title'
		local
			b, whole_box: DRW_BASIC_RECTANGLE
			box: DRW_RECTANGLE
			w: DOUBLE
		do
			-- Add title aux info box

			w := available_paper_width - title_aux_width
			
			create box.make (title_aux_width, title_height, w + title_aux_width / 2.0, 0)
			box.set_stroke_width (thick_width)
			last_title.extend (box)
			whole_box := box.bounding_box.twin
			
			-- Box with run type and wafer techn

			b := nth_vertical_section (whole_box, 3, 3)
			create box.make (b.width, b.height, b.x, b.y)
			box.set_stroke_width (thick_width)
			last_title.extend (box)
			b := box.bounding_box.twin
			shrink_box (b)
			last_title.extend (text_in_box (runs_row.run_type + "  " + runs_row.wafer_techn, b))
			
			-- Box with run closer

			b := nth_vertical_section (whole_box, 2, 3)
			create box.make (b.width, b.height, b.x, b.y)
			box.set_stroke_width (thick_width)
			last_title.extend (box)
			b := box.bounding_box.twin
			shrink_box (b)
			last_title.extend (text_in_box ("Closer: " + runs_row.closer, b))
			
			-- Box with map generation date

			b := nth_vertical_section (whole_box, 1, 3)
			create box.make (b.width, b.height, b.x, b.y)
			box.set_stroke_width (thick_width)
			last_title.extend (box)
			b := box.bounding_box.twin
			shrink_box (b)
			last_title.extend (text_in_box (date.to_date_and_time_string, b))
		end

	font: EV_FONT
			-- Font used for text size calculations (only to
			-- guess whether text pairs should be displayed
			-- side by side or one on top of the other)

invariant
	run_id_exists: run_id /= Void
end
