
indexing
	
	description: "Object that can access information about scribe edges"
	author: "David Hollenberg"
	date: "January 17, 2007"

class SCRIBE_EDGE_ACCESSOR
inherit
	GRIDABLE

feature -- Properties

	field_list: LIST [FIELD_ELEMENT_EDGES]
			-- List of fields of view and their horizontal and
			-- vertical edges

feature -- Modification

	set_field_list (a_list: like field_list)
			-- Set `field_list' to `a_list'
		require
			a_list_not_void: a_list /= Void
		do
			field_list := a_list
		ensure
			field_list_set: field_list = a_list
		end

feature {NONE} -- Implementation

feature {NONE} -- Implementation

	is_floating_scribe (an_element: RETICLE_VIEW_ELEMENT): BOOLEAN
			-- Is `an_element' a floating scribe?
		do
			if an_element.is_scribe and an_element.intersects_any_field then
				Result := is_floating (an_element)
			end
		end

	is_floating (an_element: RETICLE_VIEW_ELEMENT): BOOLEAN
			-- Is either end of `an_element' floating (not
			-- on a field of view edge of an edge of scribe
			-- of opposite orientation)?
		local
			low, high: DOUBLE
		do
			if an_element.is_horizontal_scribe then
				low := grid.rounded_edge (an_element.bottom_y)
				high := grid.rounded_edge (an_element.top_y)
				Result := is_horizontal_floating_end (an_element.field_containing_left_edge, grid.rounded_edge (an_element.left_x), low, high, True) or else is_horizontal_floating_end (an_element.field_containing_right_edge, grid.rounded_edge (an_element.right_x), low, high, False)
			else
				low := grid.rounded_edge (an_element.left_x)
				high := grid.rounded_edge (an_element.right_x)
				Result := is_vertical_floating_end (an_element.field_containing_bottom_edge, grid.rounded_edge (an_element.bottom_y), low, high, True) or else is_vertical_floating_end (an_element.field_containing_top_edge, grid.rounded_edge (an_element.top_y), low, high, False)
			end
		end

	is_horizontal_floating_end (field: RETICLE_FIELD pos, low, high: DOUBLE is_low_end: BOOLEAN): BOOLEAN
			-- Is end of horizontal scribe which is in
			-- `field' and at `pos' floating, where end
			-- extends from `low' to `high' and is the
			-- left end if `is_low_end'?
		local
			edges: LIST [MODEL_ELEMENT_EDGE]
			next_edge: MODEL_ELEMENT_EDGE
		do
			if field /= Void then
				edges := edges_for_field (field).vertical_edges
				next_edge := closest_meetable_edge (pos, low, high, edges, is_low_end)
				Result := next_edge = Void or else next_edge.position /= pos or else not next_edge.contains (low, high)
			else
				Result := not meets_field_edge (True, pos, low, high, is_low_end)
			end
		end

	is_vertical_floating_end (field: RETICLE_FIELD pos, low, high: DOUBLE is_low_end: BOOLEAN): BOOLEAN
			-- Is end of vertical scribe which is in
			-- `field' and at `pos' floating, where end
			-- extends from `low' to `high' and is the
			-- bottom end if `is_low_end'?
		local
			edges: LIST [MODEL_ELEMENT_EDGE]
			next_edge: MODEL_ELEMENT_EDGE
		do
			if field /= Void then
				edges := edges_for_field (field).horizontal_edges
				next_edge := closest_meetable_edge (pos, low, high, edges, is_low_end)
				Result := next_edge = Void or else next_edge.position /= pos or else not next_edge.contains (low, high)
			else
				Result := not meets_field_edge (False, pos, low, high, is_low_end)
			end
		end

	meets_field_edge (is_horiz: BOOLEAN pos, low, high: DOUBLE is_low_end: BOOLEAN): BOOLEAN
			-- Does scribe end at `pos' which extends from
			-- `low' to `high' and is horizontal if `is_horiz' and
			-- is left or bottom end if `is_low_end' meet the
			-- edge of any field of view along entire length
			-- of end?
		local
			field: RETICLE_FIELD
			edge: DOUBLE
			lower_end, upper_end: DOUBLE
		do
			from
				field_list.start
			until
				field_list.after or Result
			loop
				field := field_list.item.field
				if is_horiz then
					if is_low_end then
						edge := grid.rounded_edge (field.left_x)
					else
						edge := grid.rounded_edge (field.right_x)
					end
					lower_end := grid.rounded_edge (field.bottom_y)
					upper_end := grid.rounded_edge (field.top_y)
				else
					if is_low_end then
						edge := grid.rounded_edge (field.bottom_y)
					else
						edge := grid.rounded_edge (field.top_y)
					end
					lower_end := grid.rounded_edge (field.left_x)
					upper_end := grid.rounded_edge (field.right_x)
				end
				Result := edge = pos and low >= lower_end and high <= upper_end
				field_list.forth
			end
		end

	edges_for_field (a_field: RETICLE_FIELD): FIELD_ELEMENT_EDGES
			-- Object with edges for `a_field'
		local
			f: like edges_for_field
		do
			from
				field_list.start
			until
				field_list.after or Result /= Void
			loop
				f := field_list.item
				if f.field = a_field then
					Result := f
				end
				field_list.forth
			end
		end

feature {NONE} -- Implementation

	closest_meetable_edge (pos, low, high: DOUBLE edges: LIST [MODEL_ELEMENT_EDGE] is_low_end: BOOLEAN): MODEL_ELEMENT_EDGE
			-- Closest edge to end of `an_element' at
			-- `pos' (which direction to search is
			-- indicated by `is_low_end' - bottom or left
			-- end if true) that truly intersects segment
			-- (`low', `high') or Void if closest edge is
			-- not meetable
		local
			e, closest_edge: MODEL_ELEMENT_EDGE
			done: BOOLEAN
			next_position, last_position: DOUBLE
		do
			if is_low_end then
				from
					done := False
					edges.start
					last_position := edges.item.position - 1.0E6
				until
					edges.after or done
				loop
					e := edges.item
					next_position := e.position
					done := next_position > pos
					if not done and e.truly_intersects (low, high) then
						if next_position > last_position or (next_position = last_position and not closest_edge.is_meetable) then
							closest_edge := e
							last_position := next_position
						end
					end
					edges.forth
				end
				
				if closest_edge.is_meetable then
					Result := closest_edge
				end
			else
				from
					done := False
					edges.finish
					last_position := edges.item.position + 1.0E6
				until
					edges.before or done
				loop
					e := edges.item
					next_position := e.position
					done := next_position < pos
					if not done and e.truly_intersects (low, high) then
						if next_position < last_position or (next_position = last_position and not closest_edge.is_meetable) then
							closest_edge := e
							last_position := next_position
						end
					end
					edges.back
				end
				
				if closest_edge.is_meetable then
					-- Can extend to closest edge and
					-- end does not already meet it
					Result := closest_edge
				end
			end
		end

end
