
indexing
	
	description: "A run closing step operation that sends an %
		%e-mail report on density of mask levels for run that %
		%have a minimum density requirement"
	author: "David Hollenberg"
	date: "March 14, 2004"

class SEND_DENSITY_REPORT_OP
inherit
	RC_PRIMITIVE_STEP_OP
	NO_USER_INTERFACE
	NO_EXTERNAL_STATE
	NUMBER_UTILITIES_ACCESS
		export
			{NONE} all
		undefine
			default_create
		end

create
	make

feature -- Properties

	input_data_items: LIST [STRING]
			-- Data items required before step can be executed
		do
			Result := array_as_list (<<
				data_item.Filled_mebes_files
				>>)
		end

feature {NONE} -- Implementation

	is_applicable: BOOLEAN
			-- Is step applicable to run for session?
		do
			Result := cache.run_needs_mebes_files and cache.run_level_rows.there_exists (agent {RUN_LEVELS_ROW}.needs_density_report)
		end

	do_step
			-- Execute step
		do
			send_density_report
			call_completion_actions
		end

	undo_step
			-- Undo execution of step
		do
			-- Nothing to undo
			call_completion_actions
		end

	set_state
			-- Set `state'
		do
			state := Void
		end

	send_density_report
			-- Send a report on all levels with a density
			-- requirement, which includes the minimum
			-- density, maximum density and the actual
			-- density of both merged and filled pattern
			-- files, if available
		do
			compute_body
			compute_subject
			publish_density_event
		end

	compute_body
			-- Compute body of report and put in `body'
		local
			levels: LIST [RUN_LEVELS_ROW]
			level: RUN_LEVELS_ROW
			fields: LIST [FIELD_NAMES_ROW]
			field_name: STRING
			row: PATTERN_FILES_ROW
			pattern_type: PATTERN_TYPE_VALUES
			merged_pattern_files: LIST [PATTERN_FILES_ROW]
			filled_pattern_files: LIST [PATTERN_FILES_ROW]
			merged_pf, filled_pf: PATTERN_FILES_ROW
		do
			levels := cache.run_level_rows
			fields := cache.field_name_rows
			create row.make_template
			merged_pattern_files := row.rows_for_run_and_type (session.run_id, pattern_type.Merge, connection, expected_count.At_least_zero)
			filled_pattern_files := row.rows_for_run_and_type (session.run_id, pattern_type.Fill, connection, expected_count.At_least_zero)
			out_of_spec := False
			create body.make (200)
	    		body.append ("Level               Field   Minimum   Maximum    Merged    Filled%N")
	    		body.append ("-----               -----   -------   -------    ------    ------%N")
			from
				levels.start
			until
				levels.after
			loop
				level := levels.item
				if level.needs_density_report then
					from
						fields.start
					until
						fields.after
					loop
						field_name := fields.item.field_name
						merged_pf := pattern_file (level.level_no, field_name, merged_pattern_files)
						filled_pf := pattern_file (level.level_no, field_name, filled_pattern_files)
						append_density_line (level, field_name, merged_pf, filled_pf)
						fields.forth
					end
					
				end
				levels.forth
			end
		end

	pattern_file (level_no: INTEGER field_name: STRING list: LIST [PATTERN_FILES_ROW]): PATTERN_FILES_ROW
			-- Row from `list' that for `level' and `field_name'
			-- or Void if none
		require
			field_name_exists: field_name /= Void
			list_exists: list /= Void
		local
			row: PATTERN_FILES_ROW
		do
			from
				list.start
			until
				list.after or Result /= Void
			loop
				row := list.item
				if row.level_no = level_no and row.field_name.is_equal (field_name) then
					Result := row
				end
				list.forth
			end
		end

	append_density_line (level: RUN_LEVELS_ROW field_name: STRING merged_pf, filled_pf: PATTERN_FILES_ROW)
			-- Append line describing minimum density and
			-- densities for `merged_pf' and `filled_pf'
			-- (if available) for `level' and `field_name'
			-- to `body'
		require
			level_exists: level /= Void
			field_name_exists: field_name /= Void
			body_exists: body /= Void
		local
			line, temp, spacer: STRING
			actual_density, min_density, max_density: DOUBLE
		do
			min_density := level.non_null_minimum_density
			max_density := level.non_null_maximum_density
			if filled_pf /= Void then
				actual_density := filled_pf.non_null_density
			elseif merged_pf /= Void then
				actual_density := merged_pf.non_null_density
			else
				actual_density := -1
					-- Indicate density not known
			end
			spacer := "   "
			create line.make (80)
			line.append (level.level_name)
			su.right_pad (line, 17)
			line.append (spacer)
			temp := field_name.twin
			su.right_pad (temp, 5)
			line.append (temp)
			line.append (spacer)
			temp := as_percent (min_density)
			su.left_pad (temp, 7)
			line.append (temp)
			line.append (spacer)
			temp := as_percent (max_density)
			su.left_pad (temp, 7)
			line.append (temp)
			line.append (spacer)
			temp := pattern_file_density (merged_pf)
			su.left_pad (temp, 7)
			line.append (temp)
			line.append (spacer)
			temp := pattern_file_density (filled_pf)
			su.left_pad (temp, 7)
			line.append (temp)
			line.append (spacer)
			if actual_density < 0 then
				out_of_spec := True
				line.append ("<<< MISSING")
			elseif actual_density < min_density then
				out_of_spec := True
				line.append ("<<< LOW")
			elseif actual_density > max_density then
				out_of_spec := True
				line.append ("<<< HIGH")
			end
			line.extend ('%N')
			body.append (line)
		end

	pattern_file_density (row: PATTERN_FILES_ROW): STRING
			-- Actual density of pattern file in `row' or
			-- N/A if density is null (or "----" if row is Void)
		do
			if row = Void then
				Result := " ---- "
	    		elseif row.is_null_density then
				Result := "  N/A "
	     		else
				Result := as_percent (row.density)
	    		end
		end

	as_percent (d: DOUBLE): STRING
			-- `d' as a nicely formatted percentage
		do
			Result := nu.money_formatted_number (d * 100.0)
			Result.extend ('%%')
		end

	compute_subject
			-- Compute subject of report and put in `subject'
		do
			if out_of_spec then
				subject := "OUT OF SPEC "
			else
				subject := ""
			end
			subject.append ("level density information")
		end

	publish_density_event
			-- Publish a level density event
		local
			event: LEVEL_DENSITY_EVENT
		do
			create event
			event.set_subject (subject)
			event.set_message (body)
			event.publish
		end

	out_of_spec: BOOLEAN
			-- Are any of the densities for level out of spec?

	subject: STRING
			-- Subject for level density message

	body: STRING
			-- Body of level density message

end
