
indexing
	
	description: "A run closing step operation that sets wafer %
		%part id source"
	author: "David Hollenberg"
	date: "June 3, 2003"

class SET_WAFER_PART_ID_SOURCE_OP
inherit
	SET_PART_ID_OP
		redefine
			state, set_state
		end

create
	make

feature -- Properties

	input_data_items: LIST [STRING]
			-- Data items required before step can be executed
		do
			Result := array_as_list (<<
				data_item.Vendor_part_sequence
				 >>)
		end

	state: STRING
			-- State saved by soft undo

feature {NONE} -- Implementation

	is_applicable: BOOLEAN
			-- Is step applicable to run for session?
		do
			if session.is_trial_run then
				Result := False
			else
				fetch_data
				Result := run_sequence /= Void or die_sequence /= Void or site_sequence /= Void
			end
		end

	do_step
		local
			acq_method: STRING
			method: PART_ACQUISITION_METHOD_VALUES
		do
			fetch_data
			check_legal_sequences
			select_sequence
			fetch_part_sequence
			warn_about_other_runs
			check_part_sequences
			fetch_next_part_id
			acq_method := sequence_row.part_acquisition_method
			if equal (acq_method, method.Preloaded) then
				if part_id_row /= Void then
					set_part_id
				else
					report_out_of_parts
				end
			elseif equal (acq_method, method.Manual) then
				ask_for_manual_part_id
			else
				handle_hard_error ("Unknown part id acquisition method " + acq_method)
				
			end
		end

	ask_for_manual_part_id
			-- Ask user for part id for a manual part sequence.
			-- Allow user to enter new part id or pick
			-- existing unused one for run's sequence
		local
			unassigned_list: DYNAMIC_LIST [STRING]
			template_row: VENDOR_PART_IDS_ROW
		do
			create template_row.make_template
			unassigned_list := template_row.unassigned_part_ids (sequence_row.vendor_part_sequence_id, connection, Expected_count.at_least_zero)
			unassigned_list.compare_objects
			if unassigned_list.is_empty then
				ask_for_new_part_id
			else
				ask_for_old_part_id (unassigned_list)
			end
		end

	ask_for_old_part_id (list: DYNAMIC_LIST [STRING])
			-- Ask user for existing part id from `list'
			-- (with option to create new part id)
		local
			ifc: SINGLE_CONSTRAINED_VALUE_IFC
			curr_part_id, part_id, default_choice: STRING
		do
			ifc := factory.single_constrained_value_interface (session.title_with_run_id ("Select Wafer Part Id"))
			ifc.set_label ("Wafer Part ID")
			ifc.enable_uppercase_only
			curr_part_id := cache.db_productions_row.wafer_part_id
			if not curr_part_id.is_empty and then not list.has (curr_part_id) then
				list.put_front (curr_part_id)
			end
			from
				list.start
			until
				list.after
			loop
				part_id := list.item
				ifc.add_choice_and_text (part_id, part_id)
				list.forth
			end
			
			if not curr_part_id.is_empty then
				default_choice := curr_part_id
			elseif state /= Void and then list.there_exists (agent {STRING}.is_equal (state)) then
				default_choice := state.out
			else
				default_choice := list.first
			end
			ifc.set_default_choice (default_choice)
			ifc.enable_free_form_entry
			ifc.build
			-- FIXME: Need to check whether there are unassigned
			-- parts and offer to delete them.
			ifc.accept_actions.extend (agent ifc.execute_action (agent call_safely_with_string (?, agent try_to_set_part_id, agent reset (ifc))))
			ifc.decline_actions.append (abort_actions)
		end

	ask_for_new_part_id
			-- Ask user for new part id and put it in
			-- database as "unassigned"
		local
			ifc: SINGLE_FREE_VALUE_IFC
			curr_part_id: STRING
		do
			ifc := factory.single_free_value_interface (session.title_with_run_id ("Enter New Wafer Part Id"))
			ifc.set_label ("Wafer Part ID (" + sequence_row.descr + ")")
			ifc.enable_uppercase_only
			ifc.set_minimum_width_in_chars (10)
			curr_part_id := cache.db_productions_row.wafer_part_id
			if not curr_part_id.is_empty then
				ifc.set_default_value (curr_part_id)
			end
			ifc.build
			ifc.accept_actions.extend (agent ifc.execute_action (agent call_safely_with_string (?, agent try_to_set_part_id, agent reset (ifc))))
			ifc.decline_actions.append (abort_actions)
		end

	matches_saved_value (s: STRING): BOOLEAN
			-- Is `s' the same part id as one saved in
			-- `state'?
		require
			value_exists: s /= Void
		do
			Result := equal (s, state)
		end

	try_to_set_part_id (s: STRING)
			-- Try to set part id source to `s'.
			-- Add it to list of parts for sequence if not present.
			-- Otherwise, select it if it is not assigned
		require
			part_id_exists: s /= Void
		local
			row_no: INTEGER
			row: VENDOR_PART_IDS_ROW
			b: DB_BIND_S
			part_id: STRING
		do
			part_id := s
			part_id.left_adjust
			part_id.right_adjust
			check_part_id (part_id)
			if not part_exists then
				fetch_part_sequence
				session.lock_row (sequence_row)
				create row.make_template
				create b.make (sequence_name)
				row_no := connection.safe_selected_integer ("nvl(max(row_no), 0)", row.table_name, "vendor_part_sequence_id = :1", b)
				row_no := row_no + 1
				create part_id_row.make_by_key (connection, option.New_row, sequence_name, row_no)
				part_id_row.set_vendor_part_id_source (part_id)
				part_id_row.set_assigned (False)
			end
			set_part_id
		end

	check_part_id (s: STRING)
			-- Check that part id `s' entered by user
			-- is valid for `sequence_name'.  Set `part_exists'
			-- to indicate whether part already exists in database
		require
			part_id_exists: s /= Void
		local
			len, max_len: INTEGER
			list: LIST [VENDOR_PART_IDS_MROW]
			template_row: VENDOR_PART_IDS_MROW
		do
			len := s.count
			max_len := {VENDOR_PART_IDS_ROW}.Maxlen_vendor_part_id_source
			if len = 0 then
				handle_user_error ("Part id is empty")
			elseif len > max_len then
				handle_hard_error ("Part id length (" + len.out + ") exceeds maximum length (" + max_len.out + ")")
			elseif s.has (' ') or s.has ('%T') then
				handle_user_error ("Part id cannot include blanks or tabs")
			else
				create template_row.make_template
				list := template_row.sequence_part_id_rows (sequence_name, s, connection, Expected_count.at_most_one)
				part_exists := not list.is_empty
				if part_exists then
					part_id_row := list.first
					session.lock_row (part_id_row)
					if part_id_row.assigned then
						handle_user_error ("Part id " + s + " for sequence " + sequence_name + " is already assigned")
					end
				end
			end
		end

	check_new_part_id (s: STRING)
			-- Check that new part id `s' entered by user
			-- is valid for `sequence_name'
		require
			part_id_exists: s /= Void
		local
			len, max_len: INTEGER
			list: LIST [VENDOR_PART_IDS_ROW]
			template_row: VENDOR_PART_IDS_ROW
		do
			len := s.count
			max_len := {VENDOR_PART_IDS_ROW}.Maxlen_vendor_part_id_source
			if len = 0 then
				handle_user_error ("Part id is empty")
			elseif len > max_len then
				handle_hard_error ("Part id length (" + len.out + ") exceeds maximum length (" + max_len.out + ")")
			elseif s.has (' ') or s.has ('%T') then
				handle_user_error ("Part id cannot include blanks or tabs")
			else
				create template_row.make_template
				list := template_row.sequence_part_id_rows (sequence_name, s, connection, Expected_count.at_most_one)
				if not list.is_empty then
					handle_user_error ("Part id " + s + " already exists for sequence " + sequence_name)
				end
			end
		end

	set_part_id
			-- Mark part id assigned and update wafer
			-- part id source in database
		do
			if not part_id_row.is_private then
				session.lock_row (part_id_row)
			end
			part_id_row.set_assigned (True)
			set_wafer_part_id_source (part_id_row.vendor_part_id_source)
			call_completion_actions
		end

	fetch_data
			-- Fetch needed data
		do
			field_sets_row := cache.field_sets_row
			fetch_part_sequences (cache.wafer_techn, field_sets_row.litho_size_id)
		end

	fetch_part_sequences (wt, ls: STRING)
			-- Fetch sequence names and put in `run_sequence',
			-- `die_sequence' and `site_sequence' (may be Void)
		require
			wafer_techn_exists: wt /= Void
			litho_size_id_exists: ls /= Void
		local
			labeled_entity: LABELED_ENTITY_VALUES
		do
			run_sequence := part_sequence_row (wt, ls, labeled_entity.Run)
			die_sequence := part_sequence_row (wt, ls, labeled_entity.Die)
			site_sequence := part_sequence_row (wt, ls, labeled_entity.Site)
		end

	part_sequence_row (wt, ls, ent: STRING): WAFER_LITHO_SIZE_PART_SEQ_IDS_ROW
			-- Row for wafer techn `wt', litho size id `ls'
			-- and labeled entity `ent' in wafer litho size
			-- part seq ids table, or Void if none
		require
			wafer_techn_exists: wt /= Void
			litho_size_id_exists: ls /= Void
			entity_exists: ent /= Void
		do
			create Result.make_by_key (connection, option.Old_no_exception_row, wt, ls, ent)
			if Result.is_empty then
				Result := Void
			end
		end

	select_sequence
			-- Select part id sequence to be used and
			-- put in `sequence'
		require
			have_sequence: run_sequence /= Void or die_sequence /= Void or site_sequence /= Void
		do
			if run_sequence /= Void then
				sequence := run_sequence
			elseif die_sequence /= Void then
				sequence := die_sequence
			else
				sequence := site_sequence
			end
			sequence_name := sequence.vendor_part_sequence_id
		ensure
			sequence_exists: sequence /= Void
		end

	fetch_assigned_part_id
			-- Fetch part id assigned to this run
			-- and put in `part_id_row'
		require
			sequence_exists: sequence /= Void
		local
			template_row: VENDOR_PART_IDS_MROW
		do
			create template_row.make_template
			part_id_row := template_row.sequence_part_id_rows (sequence_name, cache.wafer_part_id_source, connection, expected_count.exactly_one).first
		end

	warn_about_other_runs
			-- Warn about other runs that use the same
			-- vendor part sequence as this run and that
			-- are not fully closed
		do
			warn_about_sequence (sequence_name)
		end

	warn_about_sequence (seq: STRING)
			-- Warn about other runs that use `seq' as
			-- their vendor part sequence and that
			-- are not fully closed
		require
			sequence_exists: seq /= Void
		local
			template_row: RUN_PART_SEQUENCES_EROW
			list: LIST [RUN_PART_SEQUENCES_EROW]
			msg: STRING
			prod_row: PRODUCTIONS_ROW
			prod_rows: LIST [PRODUCTIONS_ROW]
			production_type: PRODUCTION_TYPE_VALUES
			part_id: STRING
		do
			create template_row.make_template
			create prod_row.make_template
			list := template_row.unclosed_runs_using_sequence (seq, session.run_id, connection)
			if not list.is_empty then
				msg := "The following runs use the same part sequence as " + session.run_id + " and are not fully closed:%N%N"
				from
					list.start
				until
					list.after
				loop
					prod_rows := prod_row.productions_for_run_and_type (list.item.run_id, production_type.Initial_run, connection, expected_count.At_most_one)
					msg.append ("    " + list.item.run_id)
					if not prod_rows.is_empty then
						part_id := prod_rows.first.wafer_part_id
						if part_id.is_empty then
							part_id := prod_rows.first.wafer_part_id_source
						end
						if not part_id.is_empty then
							msg.append ("    " + part_id)
						end
					end
					msg.extend ('%N')
					list.forth
				end
				warn_with_notify (msg)
			end
		end

	check_legal_sequences
			-- Check that combination of part sequences is legal
		local
			labeled_entity: LABELED_ENTITY_VALUES
			row: WAFER_LITHO_SIZE_PART_SEQ_IDS_ROW
		do
			if die_sequence /= Void and site_sequence /= Void then
				create row.make_template
				handle_hard_error ("Run has both a " + labeled_entity.Die + " part and a " + labeled_entity.Site + " part in table " + row.table_name)
			end
		end

	check_part_sequences
			-- Check whether part sequence(s) for run
			-- have any lost part ids (assigned but not
			-- referenced by any run) or any references
			-- to unassigned parts
		local
			set: TWO_WAY_SORTED_SET [STRING]
		do
			create set.make
			set.compare_objects
			if run_sequence /= Void then
				set.extend (run_sequence.vendor_part_sequence_id)
			end
			if die_sequence /= Void then
				set.extend (die_sequence.vendor_part_sequence_id)
			end
			if site_sequence /= Void then
				set.extend (site_sequence.vendor_part_sequence_id)
			end
			
			from
				set.start
			until
				set.after
			loop
				check_part_sequence (set.item)
				set.forth
			end
		end

	check_part_sequence (seq: STRING)
			-- Check whether sequence `seq' has any lost
			-- part ids (assigned but not referenced by
			-- any run) or whether there are any references
			-- to any of its unassigned parts
		require
			sequence_exists: seq /= Void
		local
			assigned_parts: LIST [STRING]
			references: LIST [PART_ID_USAGE]
			lost_parts, unassigned_references: STRING
			template_row: VENDOR_PART_IDS_ROW
		do
			create template_row.make_template
			assigned_parts := template_row.assigned_part_ids (seq, connection, expected_count.at_least_zero)
			references := all_part_id_usages (seq, connection)
			
			create lost_parts.make (0)
			from
				assigned_parts.start
			until
				assigned_parts.after
			loop
				if not references.there_exists (agent {PART_ID_USAGE}.matches_part (assigned_parts.item)) then
					lost_parts.append ("Part id " + assigned_parts.item + " is marked assigned but is not used by any run%N")
				end
				assigned_parts.forth
			end
			
			create unassigned_references.make (0)
			from
				references.start
			until
				references.after
			loop
				if not assigned_parts.there_exists (agent {STRING}.is_equal (references.item.part_id)) then
					unassigned_references.append ("Unassigned part id " + references.item.part_id + " is used by " + references.item.usage + "%N")
				end
				references.forth
			end
			
			if not lost_parts.is_empty or not unassigned_references.is_empty then
				-- FIXME: change back to handle_hard_error
				-- when data is fixed
				handle_soft_error ("Error on vendor part sequence " + seq + "%N%N" + lost_parts + unassigned_references)
				-- handle_hard_error ("Error on vendor part sequence " + seq + "%N%N" + lost_parts + unassigned_references)
			end
		end

	set_wafer_part_id_source (s: STRING)
			-- Set wafer part id source to `s'
		require
			string_exists: s /= Void
			string_not_empty: not s.is_empty
		do
			fetch_and_modify_run (s)
			add_commit_action (agent publish_new_value (s))
			add_commit_action (agent warn_if_low_on_parts (sequence_name))
		end

	undo_step
			-- Undo execution of step
		do
			fetch_data
			check_legal_sequences
			select_sequence
			fetch_assigned_part_id
			session.lock_row (part_id_row)
			part_id_row.set_assigned (False)
			fetch_and_modify_run (Void)
			add_commit_action (agent publish_new_value (""))
			call_completion_actions
		end

	set_state
			-- Set `state'
		do
			state := cache.wafer_part_id_source
		end

	fetch_and_modify_run (s: STRING)
			-- Fetch productions row, lock it and set
			-- wafer part id source to `s'
		local
			production: PRODUCTIONS_MROW
		do
			production := cache.productions_mrow
			session.lock_row (production)
			production.set_wafer_part_id_source (s)
		end


	publish_new_value (s: STRING)
			-- Publish new value of wafer part id source
		local
			e: WAFER_PART_ID_SRC_CHANGE_EVENT
		do
			create e
			e.set_wafer_part_id_source (s)
			e.publish
		end

	run_sequence, die_sequence, site_sequence: WAFER_LITHO_SIZE_PART_SEQ_IDS_ROW
	
	sequence: WAFER_LITHO_SIZE_PART_SEQ_IDS_ROW
			-- Sequence to be used to select part id
	
	field_sets_row: FIELD_SETS_ROW
			-- Row from FIELD_SETS table
	
	part_exists: BOOLEAN
			-- Does part selected by user for sequence with
			-- manual acquisition exist in database?
	
end
