
indexing
	
	description: "Factory to generate a combo box with a set of choices and descriptions"
	author: "David Hollenberg"
	date: "March 5, 2007"

class SINGLE_CONSTRAINED_VALUE_FACTORY

inherit
	CASE_SETTABLE
		redefine
			default_create
		end

feature {NONE} -- Creation

	default_create is
			-- Create factory
		do
			create {ARRAYED_LIST [STRING]} choices.make (10)
			create {ARRAYED_LIST [STRING]} descriptions.make (10)
			create table.make (10)
			choices.compare_objects
		end

feature -- Modification

	set_combo_box (a_combo_box: EV_COMBO_BOX)
			-- Set `combo_box' to `a_combo_box'
		do
			combo_box := a_combo_box
		end

	add_choice_and_text (s, descr: STRING) is
			-- Add `s' to end of list of choices, which is
			-- displayed to the user as `descr'.
		require
			choice_not_void: s /= Void
			new_choice: not has_choice (s)
			description_not_void: descr /= Void
		do
			choices.extend (s)
			descriptions.extend (descr)
			table.put (choices.count, s)
		end

	prepend_choice_and_text (s, descr: STRING) is
			-- Add `s' to beginning of list of choices,
			-- which is displayed to the user as `descr'.
		require
			choice_not_void: s /= Void
			new_choice: not has_choice (s)
			description_not_void: descr /= Void
		do
			choices.put_front (s)
			descriptions.put_front (descr)
			-- Increase index of every existing choice by 1
			from
				table.start
			until
				table.off
			loop
				table.replace (table.item_for_iteration + 1, table.key_for_iteration)
				table.forth
			end
			table.put (1, s)
		end

	set_default_choice (s: STRING) is
			-- Make `s' the default choice
		require
			choice_not_void: s /= Void
			has_choice: has_choice (s)
		do
			default_index := table.item (s)
		end

	enable_choices_prepended is
			-- Set `choices_prepended' to True
		do
			choices_prepended := True
		ensure
			choices_prepended: choices_prepended
		end

	disable_choices_prepended is
			-- Set `choices_prepended' to False
		do
			choices_prepended := False
		ensure
			choices_not_prepended: not choices_prepended
		end

	enable_free_form_entry is
			-- Set `free_form_entry_allowed' to True
		do
			free_form_entry_allowed := True
		ensure
			free_form_entry_enabled: free_form_entry_allowed
		end

	disable_free_form_entry is
			-- Set `free_form_entry_allowed' to False
		do
			free_form_entry_allowed := False
		ensure
			free_form_entry_disabled: not free_form_entry_allowed
		end

feature -- Properties

	has_choice (s: STRING): BOOLEAN is
			-- Is `s' an existing choice?
		require
			string_not_void: s /= Void
		do
			Result := table.has (s)
		end

	choices: DYNAMIC_LIST [STRING]
			-- Choices to offer user, in order

	descriptions: DYNAMIC_LIST [STRING]
			-- Descriptions to display to user for choices, 
			-- in order

	choices_prepended: BOOLEAN
			-- Should the text for each choice be prepended
			-- to the description for that choice, with
			-- enough spaces so that things line up nicely?

	free_form_entry_allowed: BOOLEAN
			-- Should free form entry be allowed in addition
			-- to selection from list of choices?

	default_choice: STRING is
			-- Default choice.  If none set, first choice is
			-- the default
		require
			has_choices: not choices.is_empty
		do
			if default_index > 0 then
				Result := choices.i_th (default_index)
			else
				Result := choices.i_th (1)
			end
		ensure
			result_not_void: Result /= Void
		end

	combo_box: EV_COMBO_BOX
			-- Interface element to use to allow user to
			-- select a choice

feature -- Generation

	generate_combo_box is
			-- Generate `combo_box'
		local
			cbox: COMBO_BOX
		do
			if is_uppercase_only then
				create cbox
				cbox.enable_uppercase_only
				combo_box := cbox
			elseif is_lowercase_only then
				create cbox
				cbox.enable_lowercase_only
				combo_box := cbox
			else
				create combo_box
			end
			
			fill_in_combo_box
		end

	fill_in_combo_box
			-- Fill in `combo_box'
		local
			list_item, default_item: EV_LIST_ITEM
			choice, descr, full_descr, the_default: STRING
			max_choice_width, space_width, min_width: INTEGER
			choice_width: INTEGER
                        font: EV_FONT
		do
			font := combo_box.font
			max_choice_width := 0
			from
				choices.start
			until
				choices.after 
			loop
				choice := choices.item
				max_choice_width := max_choice_width.max (font.string_width (choice))
				choices.forth
			end
			space_width := font.string_width (" ")
			
			the_default := default_choice
			from
				choices.start
				descriptions.start
			until
				choices.after 
			loop
				choice := choices.item
				descr := descriptions.item
				if choices_prepended then
					full_descr := choice.twin
					from
						choice_width := font.string_width (choice)
					until
						choice_width >= max_choice_width
					loop
						full_descr.extend (' ')
						choice_width := choice_width + space_width
					end
					full_descr.append (Spacer)
					full_descr.append (Left_delimiter)
					full_descr.append (descr)
					full_descr.append (Right_delimiter)
					descr := full_descr
				end
				min_width := min_width.max (font.string_width (descr))
				create list_item.make_with_text (descr)
				list_item.set_data (choice)
				combo_box.extend (list_item)
				if choice.is_equal (the_default) then
					default_item := list_item
				end
				choices.forth
				descriptions.forth
			end
			default_item.enable_select
			combo_box.set_minimum_width (min_width + 40)
				-- FIXME: 40 is a hack
			if free_form_entry_allowed then
				combo_box.enable_edit
			else
				combo_box.disable_edit
			end
			
		end

feature {NONE} -- Implementation

	table: HASH_TABLE [INTEGER, STRING]
			-- Table used to determine position of choice
			-- in list, without linear scan
	
	default_index: INTEGER
			-- Index of default choice in list.  Zero means
			-- use first choice in list
	
	Spacer: STRING is "      "
			
	Left_delimiter: STRING is ""
			
	Right_delimiter: STRING is ""
			
invariant
	choice_list_exists: choices /= Void
	description_list_exists: descriptions /= Void
	table_exists: table /= Void
end
