deferred class TEXT_AS_LAYOUT
	-- turning text into simple layout (synthesis)
	-- basic routine is
	--	text_as_layout(txt: STRING): LAYOUT

	-- units are arbitrary
	-- our baseline is at y = 0, and the left hand side of every
	-- character is at x = 0.
	-- standard height (for CAPS and digits) is 32; width is variable.
	-- descenders and shorter characters are allowed for other charss.

inherit
	STRING_UTILITIES_ACCESS

	LAYOUT_SUPPORT

feature	-- font attributes

	first_graphic: INTEGER is 33;	-- '!'.code
	last_graphic: INTEGER is 126	-- '~'.code

	is_graphic_code(cod: INTEGER): BOOLEAN is
	do  Result := first_graphic <= cod and cod <= last_graphic
	end;

	min_width: DOUBLE is
		-- width of the narrowest character
	local gr: INTEGER
	do  Result := blank_width;
	    from  gr := first_graphic
	    until gr > last_graphic
	    loop  Result := Result.min(char_width(gr))
		  gr := gr + 1
	    end
	end;

	max_width: DOUBLE is
		-- width of the widest character
	local gr: INTEGER
	do  Result := blank_width;
	    from  gr := first_graphic
	    until gr > last_graphic
	    loop  Result := Result.max(char_width(gr))
		  gr := gr + 1
	    end
	end;

	max_descender: DOUBLE is
		-- greatest descent below baseline
	local gr: INTEGER
	do  Result := 0
	    from  gr := first_graphic
	    until gr > last_graphic
	    loop  Result := Result.min(depict(gr).bounding_box.low_y)
		  gr := gr + 1
	    end
	ensure Result <= 0
	end;

feature	-- assembly attributes

	blank_width: DOUBLE
		-- width of a blank (which has no body)

	char_spacing: DOUBLE
		-- horizontal spacing between characters

	line_spacing: DOUBLE
		-- vertical spacing between lines
		--  (baseline-to-baseline distance is 32 + line_spacing)

	line_height: DOUBLE is
	do  Result := 32 + line_spacing
	end;

	set_blank_width(bwd: DOUBLE) is
	do  blank_width := bwd
	end;

	set_char_spacing(csp: DOUBLE) is
	do  char_spacing := csp
	end;

	set_line_spacing(lsp: DOUBLE) is
	do  line_spacing := lsp
	end;

	make is
		-- setup
	do  create the_characters.make(first_graphic, last_graphic);
	    set_blank_width(20);
	    set_char_spacing(5);
	    set_line_spacing(8)
	end;

feature	-- a character representation (in our font)

	depict(cod: INTEGER): LAYOUT is
	require is_graphic: is_graphic_code(cod)
	do  if the_characters @ cod = Void
	     then the_characters.enter(char_layout(cod.to_character_8), cod)
	    end;
	    Result := the_characters @ cod
	ensure Result /= Void
	end;

	char_width(cod: INTEGER): DOUBLE is
		-- since left side is 0, width is just right side
	do  Result := depict(cod).bounding_box.high_x
	end;

feature {NONE} -- character access

	the_characters: ARRAY[LAYOUT]
		-- indexed by code

	char_layout(c: CHARACTER): LAYOUT is
		-- the creator
	deferred
	end;

feature	-- text assembly

	text_as_layout(txt: STRING): LAYOUT is
		-- '%N' is new-line;
		-- all other characters must be graphics (or blank)
	require txt /= Void
	local lines: LIST[STRING]; y: DOUBLE; foo: LAYOUT_COLLECTION
	do  if not txt.has('%N')
	     then Result := line_as_layout(txt)
	     else lines := su.broken_at(txt, '%N');
		  create foo.make_unlayered(lines.count)
		  from  lines.start; y := (lines.count-1) * line_height
		  until lines.after
		  loop  foo.extend(slide(line_as_layout(lines.item), 0, y));
			y := y - line_height
			lines.forth
		  end;
		  Result := foo
	    end
	end;

feature {NONE} -- text assembly

	line_as_layout(txt: STRING): LAYOUT is
	require txt /= Void
	local foo: LAYOUT_COLLECTION; i: INTEGER; c: CHARACTER;
		l: LAYOUT; x: DOUBLE
	do  create foo.make_unlayered(txt.count);
	    from  i := 1; x := 0
	    until i > txt.count
	    loop  c := txt @ i;
		  if c = ' '
		   then x := x + blank_width + char_spacing
		   else l := depict(c.code);
			foo.extend(slide(l, x, 0));
			x := x + l.bounding_box.high_x + char_spacing
		  end;
		  i := i + 1
	    end;
	    Result := foo
	end

end -- TEXT_AS_LAYOUT
