
indexing
	
	description: "Proxy for a titled window.  This allows selection %
		%of which titled window is used at run-time"
	author: "David Hollenberg";
	date: "November 2, 2003"

class TITLED_WINDOW_PROXY
inherit
	ANY
		redefine
			default_create
		end

create
	make, make_with_title

feature {NONE} -- Initialization

	initialize
			-- Mark `Current' as initialized.
			-- This must be called during the creation procedure
			-- to satisfy the `is_initialized' invariant.
			-- Descendants may redefine initialize to perform
			-- additional setup tasks.
		do
		end

	make (a_window: EV_TITLED_WINDOW)
			-- Create proxy that delegates to `a_window'
		require
			a_window_exists: a_window /= Void
			a_window_not_destroyed: not a_window.is_destroyed
		do
			window := a_window
			default_create
		end

	make_with_title (a_title: STRING_GENERAL a_window: EV_TITLED_WINDOW)
			-- Initialize with `a_title' and `a_window'
		require
			a_title_not_void: a_title /= Void
			a_window_not_void: a_window /= Void
			a_window_not_destroyed: not a_window.is_destroyed
		do
			make (a_window)
			set_title (a_title)
		end
	
feature -- Initialization

	frozen default_create
			-- Standard creation procedure.
		do
			initialize
			is_initialized := True
		ensure then
			is_initialized: is_initialized
		end
	
feature -- Access

	accelerators: EV_ACCELERATOR_LIST
			-- Key combination shortcuts associated with 
		require
			not_destroyed: not is_destroyed
		do
			Result := window.accelerators
		ensure
			not_void: Result /= Void
		end

	accept_cursor: EV_POINTER_STYLE
			-- `Result' is cursor displayed when the screen pointer is over a
			-- target that accepts `pebble' during pick and drop.
		do
			Result := window.accept_cursor
		ensure
			result_not_void: Result /= Void
		end

	actual_drop_target_agent: FUNCTION [ANY, TUPLE [INTEGER_32, INTEGER_32], EV_ABSTRACT_PICK_AND_DROPABLE]
			-- Overrides default drop target on a certain position.
			-- If Void, `Current' will use the default drop target.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.actual_drop_target_agent
		end

	background_color: EV_COLOR
			-- Color displayed behind foreground features.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.background_color
		ensure
			result_exists: Result /= Void
		end

	background_pixmap: EV_PIXMAP
			-- Result is pixmap displayed on background of `Current'.
			-- It is tessellated and fills whole of `Current'.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.background_pixmap
		end

	data: ANY
			-- Arbitrary user data may be stored here.
		do
			Result := window.data
		end

	default_key_processing_handler: PREDICATE [ANY, TUPLE [EV_KEY]] assign set_default_key_processing_handler
			-- Agent used to determine whether the default key processing should occur for Current.
			-- If agent returns True then default key processing continues as normal, False prevents
			-- default key processing from occurring.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.default_key_processing_handler
		end

	deny_cursor: EV_POINTER_STYLE
			-- Result is cursor displayed when the screen pointer is over a
			-- target that does not accept `pebble' during pick and drop.
		do
			Result := window.deny_cursor
		ensure
			result_not_void: Result /= Void
		end

	foreground_color: EV_COLOR
			-- Color of foreground features like text.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.foreground_color
		ensure
			result_not_void: Result /= Void
		end

	has_recursive (an_item: like item): BOOLEAN
			-- Does structure include `an_item' or
			-- does any structure recursively included by structure,
			-- include `an_item'.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.has_recursive (an_item)
		end

	help_context: FUNCTION [ANY, TUPLE, EV_HELP_CONTEXT]
			-- Agent that evaluates to help context sent to help engine when help is requested
		require
			not_destroyed: not is_destroyed
		do
			Result := window.help_context
		end

	icon_name: STRING_32
			-- Name displayed when `Current is minimized.
			-- If `is_empty' then `title' is displayed.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.icon_name
		ensure
			result_not_void: Result /= Void
		end

	icon_pixmap: EV_PIXMAP
			-- Window icon.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.icon_pixmap
		ensure
			result_exists: Result /= Void
		end

	frozen id_object (an_id: INTEGER_32): IDENTIFIED
			-- Object associated with `an_id' (void if no such object)
		do
			Result := window.id_object (an_id)
		ensure
			consistent: Result = Void or else Result.object_id = an_id
		end

	is_docking_enabled: BOOLEAN
			-- May `Current' be docked to?
			-- If True, `Current' will accept docking
			-- from a compatible EV_DOCKABLE_SOURCE.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.is_docking_enabled
		end

	item: EV_WIDGET
			-- Current item.
		require
			not_destroyed: not is_destroyed
			readable: readable
		do
			Result := window.item
		end

	lower_bar: EV_VERTICAL_BOX
			-- Room at bottom of window. (Example use: statusbar.)
		require
			not_destroyed: not is_destroyed
		do
			Result := window.lower_bar
		end

	Maximum_dimension: INTEGER_32
			-- Maximum width/height that a window can be set to.
		do
			Result := window.Maximum_dimension
		end

	maximum_height: INTEGER_32
			-- Upper bound on `height' in pixels.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.maximum_height
		end

	maximum_width: INTEGER_32
			-- Upper bound on `width' in pixels.
			-- (from EV_WINDOW)
		require -- from EV_WINDOW
			not_destroyed: not is_destroyed
		do
			Result := window.maximum_width
		end

	menu_bar: EV_MENU_BAR
			-- Horizontal bar at top of client area that contains menu's.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.menu_bar
		end

	frozen object_id: INTEGER_32
			-- Unique for current object in any given session
		do
			Result := window.object_id
		ensure
			valid_id: id_object (Result) = window
		end

	parent: EV_CONTAINER
			-- Contains `Current'.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.parent
		end

	pebble: ANY
			-- Data to be transported by pick and drop mechanism.
		do
			Result := window.pebble
		end

	pebble_function: FUNCTION [ANY, TUPLE, ANY]
			-- Returns data to be transported by pick and drop mechanism.
		do
			Result := window.pebble_function
		end

	pebble_positioning_enabled: BOOLEAN
			-- If True then pick and drop start coordinates are
			-- `pebble_x_position', `pebble_y_position'.
			-- If False then pick and drop start coordinates are
			-- the pointer coordinates.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.pebble_positioning_enabled
		end

	pebble_x_position: INTEGER_32
			-- Initial x position for pick and drop relative to `Current'.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.pebble_x_position
		end

	pebble_y_position: INTEGER_32
			-- Initial y position for pick and drop relative to `Current'.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.pebble_y_position
		end

	pointer_position: EV_COORDINATE
			-- Position of the screen pointer relative to `Current'.
		require
			not_destroyed: not is_destroyed
			is_show_requested: is_show_requested
		do
			Result := window.pointer_position
		end

	pointer_style: EV_POINTER_STYLE
			-- Cursor displayed when pointer is over this widget.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.pointer_style
		end

	real_target: EV_DOCKABLE_TARGET
			-- Result is target used during a dockable transport if
			-- mouse pointer is above `Current'.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.real_target
		end

	target_name: STRING_GENERAL
			-- Optional textual name describing `Current' pick and drop hole.
		do
			Result := window.target_name
		end

	title: STRING_32
			-- A textual name used by the window manager.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.title
		ensure
			not_void: Result /= void
		end

	upper_bar: EV_VERTICAL_BOX
			-- Room at top of window. (Example use: toolbars.)
			-- Positioned below menu bar.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.upper_bar
		end

	veto_dock_function: FUNCTION [ANY, TUPLE [EV_DOCKABLE_SOURCE], BOOLEAN]
			-- Function to determine whether current dock is allowed.
			-- If Result is True, dock will be disallowed.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.veto_dock_function
		end
	
feature -- Measurement

	client_height: INTEGER_32
			-- Height of the area available to children in pixels.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.client_height
		end

	client_width: INTEGER_32
			-- Width of the area available to children in pixels.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.client_width
		end

	screen_x: INTEGER_32
			-- Horizontal offset relative to left of screen in pixels.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.screen_x
		end

	screen_y: INTEGER_32
			-- Vertical offset relative to top of screen in pixels.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.screen_y
		end

feature -- Status report

	count: INTEGER_32
			-- Number of items in `Current'
		require
			not_destroyed: not is_destroyed
		do
			Result := window.count
		ensure
			valid_result: Result = 0 or Result = 1
		end

	extendible: BOOLEAN
			-- Is there no element?
		do
			Result := window.extendible
		end

	full: BOOLEAN
			-- Is structure filled to capacity?
		do
			Result := window.full
		end

	has (v: EV_WIDGET): BOOLEAN
			-- Does structure include `v'?
		do
			Result := window.has (v)
		ensure
			not_found_in_empty: Result implies not is_empty
		end

	has_capture: BOOLEAN
			-- Does widget have capture?
		require
			not_destroyed: not is_destroyed
		do
			Result := window.has_capture
		end

	has_focus: BOOLEAN
			-- Does widget have the keyboard focus?
		require
			not_destroyed: not is_destroyed
		do
			Result := window.has_focus
		end

	frozen id_freed: BOOLEAN
			-- Has `Current' been removed from the table?
		do
			Result := window.id_freed
		end

	is_border_enabled: BOOLEAN
			-- Is a border displayed around `Current'?
		require
			not_destroyed: not is_destroyed
		do
			Result := window.is_border_enabled
		end

	is_displayed: BOOLEAN
			-- Is `Current' visible on the screen?
			-- True when show requested and parent displayed.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.is_displayed
		end

	is_dockable: BOOLEAN
			-- Is `Current' dockable?
			-- If True, then `Current' may be dragged
			-- from its current parent.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.is_dockable
		end

	is_empty: BOOLEAN
			-- Is there no element?
		do
			Result := window.is_empty
		end

	is_external_docking_enabled: BOOLEAN
			-- Is `Current' able to be docked into an EV_DOCKABLE_DIALOG
			-- When there is no valid EV_DRAGABLE_TARGET upon completion
			-- of transport?
		require
			not_destroyed: not is_destroyed
		do
			Result := window.is_external_docking_enabled
		end

	is_external_docking_relative: BOOLEAN
			-- Will dockable dialog displayed when `Current' is docked externally
			-- be displayed relative to parent window of `Current'?
			-- Otherwise displayed as a standard window.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.is_external_docking_relative
		end

	is_inserted (v: EV_WIDGET): BOOLEAN
			-- Has `v' been inserted by the most recent insertion?
			-- (By default, the value returned is equivalent to calling
			-- `has (v)'. However, descendants might be able to provide more
			-- efficient implementations.)
		do
			Result := window.is_inserted (v)
		end

	is_maximized: BOOLEAN
			-- Is displayed at maximum size?
		require
			not_destroyed: not is_destroyed
		do
			Result := window.is_maximized
		end

	is_minimized: BOOLEAN
			-- Is displayed iconified/minimised?
		require
			not_destroyed: not is_destroyed
		do
			Result := window.is_minimized
		end

	is_sensitive: BOOLEAN
			-- Is object sensitive to user input.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.is_sensitive
		end

	is_show_requested: BOOLEAN
			-- Will `Current' be displayed when its parent is?
			-- See also `is_displayed'.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.is_show_requested
		end

	merged_radio_button_groups: ARRAYED_LIST [EV_CONTAINER]
			-- Result is all other radio button groups
			-- merged with `Current'. Void if no other containers
			-- are merged.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.merged_radio_button_groups
		ensure
			not_is_empty: Result /= Void implies not Result.is_empty
		end

	mode_is_drag_and_drop: BOOLEAN
			-- Is the user interface mode drag and drop?
		require
			not_destroyed: not is_destroyed
		do
			Result := window.mode_is_drag_and_drop
		end

	mode_is_pick_and_drop: BOOLEAN
			-- Is the user interface mode pick and drop?
		require
			not_destroyed: not is_destroyed
		do
			Result := window.mode_is_pick_and_drop
		end

	mode_is_target_menu: BOOLEAN
			-- Is the user interface mode a pop-up menu of targets?
		require
			not_destroyed: not is_destroyed
		do
			Result := window.mode_is_target_menu
		end

	prunable: BOOLEAN
			-- May items be removed?
		do
			Result := window.prunable
		end

	readable: BOOLEAN
			-- Is there a current item that may be accessed?
		require
			not_destroyed: not is_destroyed
		do
			Result := window.readable
		end

	real_source: EV_DOCKABLE_SOURCE
			-- Result is source to be dragged
			--  when a docking drag occurs on `Current'.
			-- If Void, `Current' is dragged.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.real_source
		end

	user_can_resize: BOOLEAN
			-- Can the user resize?
		require
			not_destroyed: not is_destroyed
		do
			Result := window.user_can_resize
		end

	writable: BOOLEAN
			-- Is there a current item that may be modified?
		require
			not_destroyed: not is_destroyed
		do
			Result := window.writable
		end
	
feature -- Status setting

	center_pointer
			-- Position screen pointer over center of `Current'.
		require
			not_destroyed: not is_destroyed
		do
			window.center_pointer
		end

	disable_border
			-- Ensure no border is displayed around `Current'.
			-- Has no immediate effect if user_can_resize, although
			-- is reflected at next call to disable_user_resize.
		require
			not_destroyed: not is_destroyed
		do
			window.disable_border
		ensure
			border_disabled: not user_can_resize implies not is_border_enabled
		end

	disable_capture
			-- Disable grab of all user input events.
		require
			not_destroyed: not is_destroyed
		do
			window.disable_capture
		ensure
			not_has_capture: not has_capture
		end

	disable_dockable
			-- Ensure `Current' is not dockable.
		require
			not_destroyed: not is_destroyed
		do
			window.disable_dockable
		ensure
			not_is_dockable: not is_dockable
		end

	disable_docking
			-- Ensure `is_docking_enabled' is False.
			-- `Current' will not accept docking.
		require
			not_destroyed: not is_destroyed
		do
			window.disable_docking
		ensure
			not_dockable: not is_docking_enabled
		end

	disable_external_docking
			-- Assign False to `is_external_docking_enabled'.
			-- Forbid `Current' to be docked into an EV_DOCKABLE_DIALOG
			-- When there is no valid EV_DRAGABLE_TARGET upon completion
			-- of transport.
		require
			not_destroyed: not is_destroyed
			is_dockable: is_dockable
		do
			window.disable_external_docking
		ensure
			not_externally_dockable: not is_external_docking_enabled
		end

	disable_external_docking_relative
			-- Assign False to `is_external_docking_relative', ensuring that
			-- a dockable dialog displayed when `Current' is docked externally
			-- is displayed as a standard window.
		require
			not_destroyed: not is_destroyed
			external_docking_enabled: is_external_docking_enabled
		do
			window.disable_external_docking_relative
		ensure
			external_docking_not_relative: not is_external_docking_relative
		end

	disable_pebble_positioning
			-- Assign False to `pebble_positioning_enabled'.
			-- The pick and drop will start at the pointer position.
		require
			not_destroyed: not is_destroyed
		do
			window.disable_pebble_positioning
		ensure
			pebble_positioning_updated: not pebble_positioning_enabled
		end

	disable_sensitive
			-- Make object non-sensitive to user input.
		require
			not_destroyed: not is_destroyed
		do
			window.disable_sensitive
		ensure
			is_unsensitive: not is_sensitive
		end

	disable_user_resize
			-- Prevent user from resizing.
		require
			not_destroyed: not is_destroyed
		do
			window.disable_user_resize
		ensure
			not_user_can_resize: not user_can_resize
		end

	enable_border
			-- Ensure a border is displayed around `Current'.
		require
			not_destroyed: not is_destroyed
		do
			window.enable_border
		ensure
			is_border_enabled: is_border_enabled
		end

	enable_capture
			-- Grab all user input events (mouse and keyboard).
			-- `disable_capture' must be called to resume normal input handling.
		require
			not_destroyed: not is_destroyed
			is_displayed: is_displayed
		do
			window.enable_capture
		ensure
			has_capture: has_capture
		end

	enable_dockable
			-- Ensure `Current' is dockable.
		require
			not_destroyed: not is_destroyed
		do
			window.enable_dockable
		ensure
			is_dockable: is_dockable
		end

	enable_docking
			-- Ensure `is_docking_enabled' is True.
		require
			not_destroyed: not is_destroyed
		do
			window.enable_docking
		ensure
			is_dockable: is_docking_enabled
		end

	enable_external_docking
			-- Assign True to `is_external_docking_enabled'.
			-- Allows `Current' to be docked into an EV_DOCKABLE_DIALOG
			-- When there is no valid EV_DRAGABLE_TARGET upon completion
			-- of transport.
		require
			not_destroyed: not is_destroyed
			is_dockable: is_dockable
		do
			window.enable_external_docking
		ensure
			is_externally_dockable: is_external_docking_enabled
		end

	enable_external_docking_relative
			-- Assign True to `is_external_docking_relative', ensuring that
			-- a dockable dialog displayed when `Current' is docked externally
			-- is displayed relative to the top level window.
		require
			not_destroyed: not is_destroyed
			external_docking_enabled: is_external_docking_enabled
		do
			window.enable_external_docking_relative
		ensure
			external_docking_not_relative: is_external_docking_relative
		end

	enable_pebble_positioning
			-- Assign True to `pebble_positioning_enabled'.
			-- Use `pebble_x_position' and `pebble_y_position' as the initial coordinates
			-- for the pick and drop in pixels relative to `Current'.
		require
			not_destroyed: not is_destroyed
		do
			window.enable_pebble_positioning
		ensure
			pebble_positioning_updated: pebble_positioning_enabled
		end

	enable_sensitive
			-- Make object sensitive to user input.
		require
			not_destroyed: not is_destroyed
		do
			window.enable_sensitive
		ensure
			is_sensitive: (parent = Void or parent_is_sensitive) implies is_sensitive
		end

	enable_user_resize
			-- Allow user to resize.
		require
			not_destroyed: not is_destroyed
		do
			window.enable_user_resize
		ensure
			user_can_resize: user_can_resize
		end

	hide
			-- Request that `Current' not be displayed even when its parent is.
			-- If successful, make `is_show_requested' False.
		require
			not_destroyed: not is_destroyed
		do
			window.hide
		end

	lock_update
			-- Lock drawing updates for this window on certain platforms until
			-- `unlock_update' is called.
			--
			-- Note: - Only one window can be locked at a time.
			--       - The window cannot be moved while update is locked.
		require
			not_destroyed: not is_destroyed
			no_locked_window: ((create {EV_ENVIRONMENT}).application.locked_window = Void)
		do
			window.lock_update
		ensure
			locked_window_is_current: (create {EV_ENVIRONMENT}).application.locked_window = window
		end

	lower
			-- Request that window be displayed below all other windows.
		require
			not_destroyed: not is_destroyed
		do
			window.lower
		end

	maximize
			-- Display at maximum size.
		require
			not_destroyed: not is_destroyed
			shown: is_show_requested
		do
			window.maximize
		ensure
			is_maximized: is_maximized
		end

	merge_radio_button_groups (other: EV_CONTAINER)
			-- Merge `Current' radio button group with that of `other'.
		require
			not_destroyed: not is_destroyed
			other_not_void: other /= Void
		do
			window.merge_radio_button_groups (other)
		end

	minimize
			-- Display iconified/minimised.
			-- It is not possible to guarantee this on some
			-- platform configurations.
		require
			not_destroyed: not is_destroyed
			shown: is_show_requested
		do
			window.minimize
		ensure
			not_displayed: not is_displayed
		end

	raise
			-- Request that window be displayed above all other windows.
		require
			not_destroyed: not is_destroyed
		do
			window.raise
		end

	remove_default_key_processing_handler
			-- Ensure default_key_processing_handler is Void.
		require
			not_destroyed: not is_destroyed
		do
			window.remove_default_key_processing_handler
		ensure
			default_key_processing_handler_removed: default_key_processing_handler = Void
		end

	remove_menu_bar
			-- Make `menu_bar' Void.
		require
			not_destroyed: not is_destroyed
		do
			window.remove_menu_bar
		ensure
			menu_bar_void: menu_bar = Void
		end

	remove_pebble
			-- Make `pebble' and `pebble_function' Void,
			-- removing transport.
		do
			window.remove_pebble
		ensure
			pebble_removed: pebble = Void and pebble_function = Void
		end

	remove_real_source
			-- Ensure `real_source' is Void.
		require
			not_destroyed: not is_destroyed
		do
			window.remove_real_source
		ensure
			real_source_void: real_source = Void
		end

	remove_real_target
			-- Ensure `real_target' is Void
		require
			not_destroyed: not is_destroyed
		do
			window.remove_real_target
		ensure
			real_target_void: real_target = Void
		end

	remove_title
			-- Make `title' empty.
		require
			not_destroyed: not is_destroyed
		do
			window.remove_title
		ensure
			title_empty: title.is_empty
		end

	restore
			-- Restore to original position when minimized or maximized.
		require
			not_destroyed: not is_destroyed
		do
			window.restore
		ensure
			minimize_restored: old is_minimized implies not is_minimized
			maximize_restored: old is_maximized implies not is_maximized
		end

	set_accept_cursor (a_cursor: like accept_cursor)
			-- Set `a_cursor' to be displayed when the screen pointer is over a
			-- target that accepts `pebble' during pick and drop.
		require
			a_cursor_not_void: a_cursor /= Void
		do
			window.set_accept_cursor (a_cursor)
		ensure
			accept_cursor_assigned: accept_cursor.is_equal (a_cursor)
		end

	set_actual_drop_target_agent (an_agent: like actual_drop_target_agent)
			-- Assign `an_agent' to `actual_drop_target_agent'.
		require
			not_destroyed: not is_destroyed
			an_agent_not_void: an_agent /= Void
		do
			window.set_actual_drop_target_agent (an_agent)
		ensure
			assigned: actual_drop_target_agent = an_agent
		end

	set_default_colors
			-- Set foreground and background color to their default values.
		require
			not_destroyed: not is_destroyed
		do
			window.set_default_colors
		end

	set_default_key_processing_handler (a_handler: like default_key_processing_handler)
			-- Assign default_key_processing_handler to `a_handler'.
		require
			not_destroyed: not is_destroyed
		do
			window.set_default_key_processing_handler (a_handler)
		end

	set_deny_cursor (a_cursor: like deny_cursor)
			-- Set `a_cursor' to be displayed when the screen pointer is not
			-- over a valid target.
		require
			a_cursor_not_void: a_cursor /= Void
		do
			window.set_deny_cursor (a_cursor)
		ensure
			deny_cursor_assigned: deny_cursor.is_equal (a_cursor)
		end

	set_drag_and_drop_mode
			-- Set user interface mode to drag and drop.
		require
			not_destroyed: not is_destroyed
		do
			window.set_drag_and_drop_mode
		ensure
			drag_and_drop_set: mode_is_drag_and_drop
		end

	set_focus
			-- Grab keyboard focus.
		require
			not_destroyed: not is_destroyed
			is_displayed: is_displayed
			is_sensitive: is_sensitive
		do
			window.set_focus
		end

	set_height (a_height: INTEGER_32)
			-- Assign `a_height' to `height' in pixels.
		require
			not_destroyed: not is_destroyed
			a_height_positive_or_zero: a_height >= 0
		do
			window.set_height (a_height)
		ensure
			height_assigned: height = minimum_height or else height = a_height
		end

	set_maximum_height (a_maximum_height: INTEGER_32)
			-- Assign `a_maximum_height' to `maximum_height' in pixels.
		require
			not_destroyed: not is_destroyed
			a_maximum_height_non_negative: a_maximum_height >= 0
			a_maximum_height_not_less_than_minimum_height: a_maximum_height >= minimum_height
			a_maximum_height_not_greater_than_maximum_dimension: a_maximum_height <= maximum_dimension
		do
			window.set_maximum_height (a_maximum_height)
		ensure
			maximum_height_assigned: maximum_height = a_maximum_height
		end

	set_maximum_size (a_maximum_width, a_maximum_height: INTEGER_32)
			-- Assign `a_maximum_width' to `maximum_width'
			-- and `a_maximum_height' to `maximum_height' in pixels.
		require
			not_destroyed: not is_destroyed
			a_maximum_width_not_less_than_minimum_width: a_maximum_width >= minimum_width
			a_maximum_height_not_less_than_minimum_height: a_maximum_height >= minimum_height
			a_maximum_width_not_greater_than_maximum_dimension: a_maximum_width <= maximum_dimension
			a_maximum_height_not_greater_than_maximum_dimension: a_maximum_height <= maximum_dimension
		do
			window.set_maximum_size (a_maximum_width, a_maximum_height)
		ensure
			maximum_width_assigned: maximum_width = a_maximum_width
			maximum_height_assigned: maximum_height = a_maximum_height
		end

	set_maximum_width (a_maximum_width: INTEGER_32)
			-- Assign `a_maximum_width' to `maximum_width' in pixels.
		require
			not_destroyed: not is_destroyed
			a_maximum_width_non_negative: a_maximum_width >= 0
			a_maximum_width_not_less_than_minimum_width: a_maximum_width >= minimum_width
			a_maximum_width_not_greater_than_maximum_dimension: a_maximum_width <= maximum_dimension
		do
			window.set_maximum_width (a_maximum_width)
		ensure
			maximum_width_assigned: maximum_width = a_maximum_width
		end

	set_menu_bar (a_menu_bar: EV_MENU_BAR)
			-- Assign `a_menu_bar' to `menu_bar'.
		require
			not_destroyed: not is_destroyed
			no_menu_bar_assigned: menu_bar = Void
			a_menu_bar_not_void: a_menu_bar /= Void
			a_menu_bar_not_parented: a_menu_bar.parent = Void
		do
			window.set_menu_bar (a_menu_bar)
		ensure
			assigned: menu_bar = a_menu_bar
		end

	set_pebble (a_pebble: like pebble)
			-- Assign `a_pebble' to `pebble'.
			-- Overrides `set_pebble_function'.
		require
			a_pebble_not_void: a_pebble /= Void
		do
			window.set_pebble (a_pebble)
		ensure
			pebble_assigned: pebble = a_pebble
		end

	set_pebble_function (a_function: FUNCTION [ANY, TUPLE, ANY])
			-- Set `a_function' to compute `pebble'.
			-- It will be called once each time a pick occurs, the result
			-- will be assigned to `pebble' for the duration of transport.
			-- When a pick occurs, the pick position in widget coordinates,
			-- <<x, y>> in pixels, is passed.
			-- To handle this data use `a_function' of type
			-- FUNCTION [ANY, TUPLE [INTEGER, INTEGER], ANY] and return the
			-- pebble as a function of x and y.
			-- Overrides `set_pebble'.
		require
			a_function_not_void: a_function /= Void
			a_function_takes_two_integer_open_operands: a_function.valid_operands ([1, 1])
		do
			window.set_pebble_function (a_function)
		ensure
			pebble_function_assigned: pebble_function = a_function
		end

	set_pebble_position (a_x, a_y: INTEGER_32)
			-- Set the initial position for pick and drop
			-- Coordinates are in pixels and are relative to position of `Current'.
			-- Pebble_positioning_enabled must be `True' for the position to be used,
			-- use enable_pebble_positioning.
		require
			not_destroyed: not is_destroyed
		do
			window.set_pebble_position (a_x, a_y)
		ensure
			pebble_position_assigned: pebble_x_position = a_x and pebble_y_position = a_y
		end

	set_pick_and_drop_mode
			-- Set user interface mode to pick and drop.
		require
			not_destroyed: not is_destroyed
		do
			window.set_pick_and_drop_mode
		ensure
			pick_and_drop_set: mode_is_pick_and_drop
		end

	set_position (a_x, a_y: INTEGER_32)
			-- Assign `a_x' to `x_position' and `a_y' to `y_position' in pixels.
		require
			not_destroyed: not is_destroyed
		do
			window.set_position (a_x, a_y)
		ensure
			x_position_assigned: x_position = a_x
			y_position_assigned: y_position = a_y
		end

	set_real_source (dockable_source: EV_DOCKABLE_SOURCE)
			-- Assign `dockable_source' to `real_source'.
		require
			not_destroyed: not is_destroyed
			is_dockable: is_dockable
			dockable_source_not_void: dockable_source /= Void
			dockable_source_is_parent_recursive: source_has_current_recursive (dockable_source)
		do
			window.set_real_source (dockable_source)
		ensure
			real_source_assigned: real_source = dockable_source
		end

	set_real_target (a_target: EV_DOCKABLE_TARGET)
			-- Assign `a_target' to `real_target'.
		require
			not_destroyed: not is_destroyed
			target_not_void: a_target /= Void
		do
			window.set_real_target (a_target)
		ensure
			assigned: real_target = a_target
		end

	set_size (a_width, a_height: INTEGER_32)
			-- Assign `a_width' to `width' and `a_height' to `height' in pixels.
		require
			not_destroyed: not is_destroyed
			a_width_positive_or_zero: a_width >= 0
			a_height_positive_or_zero: a_height >= 0
		do
			window.set_size (a_width, a_height)
		ensure
			width_assigned: width = minimum_width or else width = a_width
			height_assigned: height = minimum_height or else height = a_height
		end

	set_target_menu_mode
			-- Set user interface mode to pop-up menu of targets.
		require
			not_destroyed: not is_destroyed
		do
			window.set_target_menu_mode
		ensure
			target_menu_mode_set: mode_is_target_menu
		end

	set_target_name (a_name: STRING_GENERAL)
			-- Assign `a_name' to `target_name'.
		require
			a_name_not_void: a_name /= Void
		do
			window.set_target_name (a_name)
		ensure
			target_name_assigned: a_name /= target_name and a_name.is_equal (target_name)
		end

	set_title (a_title: STRING_GENERAL)
			-- Assign `a_title' to `title'.
		require
			not_destroyed: not is_destroyed
			a_title_not_void: a_title /= Void
		do
			window.set_title (a_title)
		ensure
			a_title_assigned: title.is_equal (a_title)
			cloned: title /= a_title
		end

	set_veto_dock_function (a_function: FUNCTION [ANY, TUPLE [EV_DOCKABLE_SOURCE], BOOLEAN])
			-- Assign `a_function' to `veto_dock_function'.
		require
			not_destroyed: not is_destroyed
			a_function_not_void: a_function /= Void
		do
			window.set_veto_dock_function (a_function)
		ensure
			veto_function_set: veto_dock_function = a_function
		end

	set_width (a_width: INTEGER_32)
			-- Assign `a_width' to `width' in pixels.
		require
			not_destroyed: not is_destroyed
			a_width_positive_or_zero: a_width >= 0
		do
			window.set_width (a_width)
		ensure
			width_assigned: width = minimum_width or else width = a_width
		end

	set_x_position (a_x: INTEGER_32)
			-- Assign `a_x' to `x_position' in pixels.
		require
			not_destroyed: not is_destroyed
		do
			window.set_x_position (a_x)
		ensure
			x_position_assigned: x_position = a_x
		end

	set_y_position (a_y: INTEGER_32)
			-- Assign `a_y' to `y_position' in pixels.
		require
			not_destroyed: not is_destroyed
		do
			window.set_y_position (a_y)
		ensure
			y_position_assigned: y_position = a_y
		end

	show
			-- Request that `Current' be displayed when its parent is.
			-- `True' by default.
			-- If successful, make is_show_requested `True'.
		require
			not_destroyed: not is_destroyed
		do
			window.show
		end

	show_relative_to_window (a_parent: EV_WINDOW)
			-- Show `Current' with respect to `a_parent'.
		require
			not_void: a_parent /= Void
			not_destroyed: not is_destroyed
			parent_not_destroyed: not a_parent.is_destroyed
			a_window_not_current: a_parent /= window
		do
			window.show_relative_to_window (a_parent)
		end

	unlock_update
			-- Unlock updates for this widget on certain platforms.
		require
			not_destroyed: not is_destroyed
			locked_window_is_current: (create {EV_ENVIRONMENT}).application.locked_window = window
		do
			window.unlock_update
		ensure
			no_locked_window: ((create {EV_ENVIRONMENT}).application.locked_window = Void)
		end

	unmerge_radio_button_groups (other: EV_CONTAINER)
			-- Remove `other' from radio button group of `Current'.
			-- If no radio button of `other' was checked before removal
			-- then first radio button contained will be checked.
		require
			not_destroyed: not is_destroyed
			other_is_merged: merged_radio_button_groups.has (other)
		do
			window.unmerge_radio_button_groups (other)
		ensure
			other_not_merged: other.merged_radio_button_groups = Void
			not_contained_in_this_group: merged_radio_button_groups /= Void implies not merged_radio_button_groups.has (other)
		end
	
feature -- Element change

	extend (v: like item)
			-- Ensure that structure includes `v'.
			-- Do not move cursor.
		require
			not_destroyed: not is_destroyed
			extendible: extendible
			v_not_void: v /= Void
			v_parent_void: v.parent = Void
			v_not_current: v /= window
			v_not_parent_of_current: not is_parent_recursive (v)
			v_containable: may_contain (v)
		do
			window.extend (v)
		ensure
			has_v: has (v)
		end

	fill (other: CONTAINER [EV_WIDGET])
			-- Fill with as many items of `other' as possible.
			-- The representations of `other' and current structure
			-- need not be the same.
		require
			other_not_void: other /= Void
			extendible: extendible
		do
			window.fill (other)
		end

	put (v: like item)
			-- Replace `item' with `v'.
		require
			not_destroyed: not is_destroyed
			writable: writable
			v_not_void: v /= Void
			v_parent_void: v.parent = Void
			v_not_current: v /= window
			v_not_parent_of_current: not is_parent_recursive (v)
			v_containable: may_contain (v)
		do
			window.put (v)
		ensure
			has_v: has (v)
			not_has_old_item: not has (old item)
		end

	remove_help_context
			-- Remove key press action associated with `EV_APPLICATION.help_key'.
		require
			not_destroyed: not is_destroyed
			help_context_not_void: help_context /= Void
		do
			window.remove_help_context
		ensure
			no_help_context: help_context = Void
		end

	remove_icon_name
			-- Make `icon_name' empty.
		require
			not_destroyed: not is_destroyed
		do
			window.remove_icon_name
		ensure
			icon_name_removed: icon_name.is_empty
		end

	remove_background_pixmap
			-- Remove image displayed on `Current'.
		require
			not_destroyed: not is_destroyed
		do
			window.remove_background_pixmap
		ensure
			pixmap_removed: background_pixmap = Void
		end

	replace (v: like item)
			-- Replace `item' with `v'.
		require
			not_destroyed: not is_destroyed
			writable: writable
			v_not_void: v /= Void
			v_parent_void: v.parent = Void
			v_not_current: v /= window
			v_not_parent_of_current: not is_parent_recursive (v)
			v_containable: may_contain (v)
		do
			window.replace (v)
		ensure
			has_v: has (v)
			not_has_old_item: not has (old item)
		end

	set_background_color (a_color: like background_color)
			-- Assign `a_color' to `background_color'.
		require
			not_destroyed: not is_destroyed
			a_color_not_void: a_color /= Void
		do
			window.set_background_color (a_color)
		ensure
			background_color_assigned: background_color.is_equal (a_color)
		end

	set_data (some_data: like data)
			-- Assign `some_data' to `data'.
		require
			not_destroyed: not is_destroyed
		do
			window.set_data (some_data)
		ensure
			data_assigned: data = some_data
		end

	set_foreground_color (a_color: like foreground_color)
			-- Assign `a_color' to `foreground_color'.
		require
			not_destroyed: not is_destroyed
			a_color_not_void: a_color /= Void
		do
			window.set_foreground_color (a_color)
		ensure
			foreground_color_assigned: foreground_color.is_equal (a_color)
		end

	set_help_context (an_help_context: like help_context)
			-- Assign `an_help_context' to `help_context'.
		require
			not_destroyed: not is_destroyed
			an_help_context_not_void: an_help_context /= Void
		do
			window.set_help_context (an_help_context)
		ensure
			help_context_assigned: help_context.is_equal (an_help_context)
		end

	set_icon_name (an_icon_name: STRING_GENERAL)
			-- Assign `an_icon_name' to `icon_name'.
		require
			not_destroyed: not is_destroyed
			an_icon_name_not_void: an_icon_name /= Void
		do
			window.set_icon_name (an_icon_name)
		ensure
			icon_name_assigned: icon_name.is_equal (an_icon_name)
			cloned: icon_name /= an_icon_name
		end

	set_icon_pixmap (an_icon: EV_PIXMAP)
			-- Assign `an_icon' to `icon'.
		require
			not_destroyed: not is_destroyed
			pixmap_not_void: an_icon /= Void
		do
			window.set_icon_pixmap (an_icon)
		ensure
			icon_pixmap_assigned: icon_pixmap.is_equal (an_icon)
		end

	set_minimum_height (a_minimum_height: INTEGER_32)
			-- Set `a_minimum_height' in pixels to `minimum_height'.
			-- If `height' is less than `a_minimum_height', resize.
			-- From now, `minimum_height' is fixed and will not be changed
			-- dynamically by the application anymore.
		require
			not_destroyed: not is_destroyed
			a_minimum_height_positive: a_minimum_height >= 0
		do
			window.set_minimum_height (a_minimum_height)
		ensure
			minimum_height_assigned: minimum_height = a_minimum_height
		end

	set_minimum_size (a_minimum_width, a_minimum_height: INTEGER_32)
			-- Assign `a_minimum_height' to `minimum_height'
			-- and `a_minimum_width' to `minimum_width' in pixels.
			-- If `width' or `height' is less than minimum size, resize.
			-- From now, minimum size is fixed and will not be changed
			-- dynamically by the application anymore.
		require
			not_destroyed: not is_destroyed
			a_minimum_width_positive: a_minimum_width >= 0
			a_minimum_height_positive: a_minimum_height >= 0
		do
			window.set_minimum_size (a_minimum_width, a_minimum_height)
		ensure
			minimum_width_assigned: minimum_width = a_minimum_width
			minimum_height_assigned: minimum_height = a_minimum_height
		end

	set_minimum_width (a_minimum_width: INTEGER_32)
			-- Assign `a_minimum_width' in pixels to `minimum_width'.
			-- If `width' is less than `a_minimum_width', resize.
			-- From now, `minimum_width' is fixed and will not be changed
			-- dynamically by the application anymore.
		require
			not_destroyed: not is_destroyed
			a_minimum_width_positive: a_minimum_width >= 0
		do
			window.set_minimum_width (a_minimum_width)
		ensure
			minimum_width_assigned: minimum_width = a_minimum_width
		end

	set_background_pixmap (a_pixmap: EV_PIXMAP)
			-- Display image of `a_pixmap' on `Current'.
			-- Image of `pixmap' will be a copy of `a_pixmap'.
			-- Image may be scaled in some descendents, i.e EV_TREE_ITEM
			-- See EV_TREE.set_pixmaps_size.
		require
			not_destroyed: not is_destroyed
			pixmap_not_void: a_pixmap /= Void
		do
			window.set_background_pixmap (a_pixmap)
		ensure
			pixmap_assigned: background_pixmap /= a_pixmap
		end

	set_pointer_style (a_cursor: like pointer_style)
			-- Assign `a_cursor' to `pointer_style'.
		require
			not_destroyed: not is_destroyed
			a_cursor_not_void: a_cursor /= Void
		do
			window.set_pointer_style (a_cursor)
		ensure
			pointer_style_assigned: pointer_style.is_equal (a_cursor)
		end
	
feature -- Removal

	frozen free_id
			-- Free the entry associated with object_id if any
		do
			window.free_id
		ensure
			object_freed: id_freed
		end

	prune (v: EV_WIDGET)
			-- Remove `v' if contained.
		require
			prunable: prunable
		do
			window.prune (v)
		ensure
			not_has_v: not has (v)
		end

	prune_all (v: EV_WIDGET)
			-- Remove all occurrences of `v'.
			-- (Reference or object equality,
			-- based on `object_comparison'.)
		require
			prunable: prunable
		do
			window.prune_all (v)
		ensure
			no_more_occurrences: not has (v)
		end

	wipe_out
			-- Remove `item'.
		require
			prunable: prunable
		do
			window.wipe_out
		ensure
			wiped_out: is_empty
		end
	
feature -- Conversion

	linear_representation: LINEAR [like item]
			-- Representation as a linear structure
		do
			Result := window.linear_representation
		end
	
feature -- Basic operations

	propagate_background_color
			-- Propagate `background_color' recursively, to all children.
		require
			not_destroyed: not is_destroyed
		do
			window.propagate_background_color
		end

	propagate_foreground_color
			-- Propagate `foreground_color' recursively, to all children.
		require
			not_destroyed: not is_destroyed
		do
			window.propagate_foreground_color
		end

	refresh_now
			-- Force an immediate redraw of `Current'.
		require
			not_destroyed: not is_destroyed
		do
			window.refresh_now
		end
	
feature -- Command

	destroy
			-- Destroy underlying native toolkit object.
			-- Render `Current' unusable.
		do
			window.destroy
		ensure
			is_destroyed: is_destroyed
		end
	
feature {NONE} -- Contract support

	parent_is_sensitive: BOOLEAN
		local
			sensitive_parent: EV_SENSITIVE
		do
			sensitive_parent ?= parent
			if sensitive_parent /= Void then
				Result := sensitive_parent.is_sensitive
			end
		end

feature -- Contract support

	is_parent_recursive (a_widget: EV_WIDGET): BOOLEAN
			-- Is `a_widget' `parent', or recursivly, `parent' of `parent'.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.is_parent_recursive (a_widget)
		end

	may_contain (v: EV_WIDGET): BOOLEAN
			-- May `v' be inserted in `Current'.
			-- Instances of EV_WINDOW may not be inserted
			-- in a container even though they are widgets.
		do
			Result := window.may_contain (v)
		end

	parent_of_source_allows_docking: BOOLEAN
			-- Does parent of source to be transported
			-- allow current to be dragged out? (See `real_source')
			-- Not all Vision2 containers are currently supported by this
			-- mechanism, only descendents of EV_DOCKABLE_TARGET
			-- are supported.
		do
			Result := window.parent_of_source_allows_docking
		end

	source_has_current_recursive (source: EV_DOCKABLE_SOURCE): BOOLEAN
			-- Does `source' recursively contain `Current'?
			-- As seen by parenting structure.
		require
			not_destroyed: not is_destroyed
			source_not_void: source /= Void
		do
			Result := window.source_has_current_recursive (source)
		end
	
feature -- Event handling

	close_request_actions: EV_NOTIFY_ACTION_SEQUENCE
			-- Actions to be performed whan a request to close window
			-- has been received.
		do
			Result := window.close_request_actions
		ensure
			not_void: Result /= Void
		end

	conforming_pick_actions: EV_NOTIFY_ACTION_SEQUENCE
			-- Actions to be performed when a pebble that fits here is picked.
		do
			Result := window.conforming_pick_actions
		ensure
			not_void: Result /= Void
		end

	dock_ended_actions: EV_NOTIFY_ACTION_SEQUENCE
			-- Actions to be performed after a dock completes from `Current'.
			-- Either to a dockable target or a dockable dialog.
		do
			Result := window.dock_ended_actions
		ensure
			not_void: Result /= Void
		end

	dock_started_actions: EV_NOTIFY_ACTION_SEQUENCE
			-- Actions to be performed when a dockable source is dragged.
		do
			Result := window.dock_started_actions
		ensure
			not_void: Result /= Void
		end

	docked_actions: EV_DOCKABLE_SOURCE_ACTION_SEQUENCE
			-- Actions to be performed when a dockable source completes a transport
			-- Fired only if source has been moved, after parenting.
		do
			Result := window.docked_actions
		ensure
			not_void: Result /= Void
		end

	drop_actions: EV_PND_ACTION_SEQUENCE
			-- Actions to be performed when a pebble is dropped here.
		do
			Result := window.drop_actions
		ensure
			not_void: Result /= Void
		end

	focus_in_actions: EV_NOTIFY_ACTION_SEQUENCE
			-- Actions to be performed when keyboard focus is gained.
		do
			Result := window.focus_in_actions
		ensure
			not_void: Result /= Void
		end

	focus_out_actions: EV_NOTIFY_ACTION_SEQUENCE
			-- Actions to be performed when keyboard focus is lost.
		do
			Result := window.focus_out_actions
		ensure
			not_void: Result /= Void
		end

	key_press_actions: EV_KEY_ACTION_SEQUENCE
			-- Actions to be performed when a keyboard key is pressed.
		do
			Result := window.key_press_actions
		ensure
			not_void: Result /= Void
		end

	key_press_string_actions: EV_KEY_STRING_ACTION_SEQUENCE
			-- Actions to be performed when a keyboard press generates a displayable character.
		do
			Result := window.key_press_string_actions
		ensure
			not_void: Result /= Void
		end

	key_release_actions: EV_KEY_ACTION_SEQUENCE
			-- Actions to be performed when a keyboard key is released.
		do
			Result := window.key_release_actions
		ensure
			not_void: Result /= Void
		end

	maximize_actions: EV_NOTIFY_ACTION_SEQUENCE
			-- Actions to be performed when window is maximized.
		do
			Result := window.maximize_actions
		ensure
			not_void: Result /= Void
		end

	minimize_actions: EV_NOTIFY_ACTION_SEQUENCE
			-- Actions to be performed when window is minimized.
		do
			Result := window.minimize_actions
		ensure
			not_void: Result /= Void
		end

	mouse_wheel_actions: EV_INTEGER_ACTION_SEQUENCE
			-- Actions to be performed when mouse wheel is rotated.
		do
			Result := window.mouse_wheel_actions
		ensure
			not_void: Result /= Void
		end

	move_actions: EV_GEOMETRY_ACTION_SEQUENCE
			-- Actions to be performed when window moves.
		do
			Result := window.move_actions
		ensure
			not_void: Result /= Void
		end

	pick_actions: EV_PND_START_ACTION_SEQUENCE
			-- Actions to be performed when `pebble' is picked up.
		do
			Result := window.pick_actions
		ensure
			not_void: Result /= Void
		end

	pick_ended_actions: EV_PND_FINISHED_ACTION_SEQUENCE
			-- Actions to be performed when a transport from `Current' ends.
			-- If transport completed successfully, then event data
			-- is target. If cancelled, then event data is Void.
		do
			Result := window.pick_ended_actions
		ensure
			not_void: Result /= Void
		end

	pointer_button_press_actions: EV_POINTER_BUTTON_ACTION_SEQUENCE
			-- Actions to be performed when screen pointer button is pressed.
		do
			Result := window.pointer_button_press_actions
		ensure
			not_void: Result /= Void
		end

	pointer_button_release_actions: EV_POINTER_BUTTON_ACTION_SEQUENCE
			-- Actions to be performed when screen pointer button is released.
		do
			Result := window.pointer_button_release_actions
		ensure
			not_void: Result /= Void
		end

	pointer_double_press_actions: EV_POINTER_BUTTON_ACTION_SEQUENCE
			-- Actions to be performed when screen pointer is double clicked.
		do
			Result := window.pointer_double_press_actions
		ensure
			not_void: Result /= Void
		end

	pointer_enter_actions: EV_NOTIFY_ACTION_SEQUENCE
			-- Actions to be performed when screen pointer enters widget.
		do
			Result := window.pointer_enter_actions
		ensure
			not_void: Result /= Void
		end

	pointer_leave_actions: EV_NOTIFY_ACTION_SEQUENCE
			-- Actions to be performed when screen pointer leaves widget.
		do
			Result := window.pointer_leave_actions
		ensure
			not_void: Result /= Void
		end

	pointer_motion_actions: EV_POINTER_MOTION_ACTION_SEQUENCE
			-- Actions to be performed when screen pointer moves.
		do
			Result := window.pointer_motion_actions
		ensure
			not_void: Result /= Void
		end

	resize_actions: EV_GEOMETRY_ACTION_SEQUENCE
			-- Actions to be performed when size changes.
		do
			Result := window.resize_actions
		ensure
			not_void: Result /= Void
		end

	restore_actions: EV_NOTIFY_ACTION_SEQUENCE
			-- Actions to be performed when window leaves `maximized' or
			-- `minimized' state.
		do
			Result := window.restore_actions
		ensure
			not_void: Result /= Void
		end

	show_actions: EV_NOTIFY_ACTION_SEQUENCE
			-- Actions to be performed when window is shown.
		do
			Result := window.show_actions
		ensure
			not_void: Result /= Void
		end
	
feature -- Measurement 

	height: INTEGER_32
			-- Vertical size in pixels.
			-- Same as `minimum_height' when not displayed.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.height
		end

	minimum_height: INTEGER_32
			-- Lower bound on `height' in pixels.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.minimum_height
		ensure
			positive_or_zero: Result >= 0
		end

	minimum_width: INTEGER_32
			-- Lower bound on `width' in pixels.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.minimum_width
		ensure
			positive_or_zero: Result >= 0
		end

	width: INTEGER_32
			-- Horizontal size in pixels.
			-- Same as `minimum_width' when not displayed.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.width
		end

	x_position: INTEGER_32
			-- Horizontal offset relative to parent `x_position' in pixels.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.x_position
		end

	y_position: INTEGER_32
			-- Vertical offset relative to parent `y_position' in pixels.
		require
			not_destroyed: not is_destroyed
		do
			Result := window.y_position
		end
	
feature -- Status Report

	is_destroyed: BOOLEAN
			-- Is `Current' no longer usable?
		do
			Result := window.is_destroyed
		end

feature {NONE} -- Contract support
	
	is_usable: BOOLEAN is
			-- Is Current usable?
		do
			Result := is_initialized and not is_destroyed
		end
	
	is_initialized: BOOLEAN
			-- Has Current been initialized properly?
	
feature -- Properties

	window: EV_TITLED_WINDOW
			-- Window to which this object delegates
		
feature -- Initialization

	reset_to_default is
			-- Reset window back to state right after default
			-- creation
		do
			conforming_pick_actions.wipe_out
			drop_actions.wipe_out
			pick_actions.wipe_out
			pick_ended_actions.wipe_out
			focus_in_actions.wipe_out
			focus_out_actions.wipe_out
			key_press_actions.wipe_out
			key_press_string_actions.wipe_out
			key_release_actions.wipe_out
			pointer_button_press_actions.wipe_out
			pointer_button_release_actions.wipe_out
			pointer_double_press_actions.wipe_out
			pointer_enter_actions.wipe_out
			pointer_leave_actions.wipe_out
			pointer_motion_actions.wipe_out
			resize_actions.wipe_out
			-- Leave `close_request_actions' alone
			-- This is really a property of the window itself
			-- close_request_actions.wipe_out
			move_actions.wipe_out
			show_actions.wipe_out
			
			if help_context /= Void then
				remove_help_context
			end
			
			wipe_out
			set_default_colors
			-- set_size (0, 0)
			enable_sensitive
			remove_background_pixmap
			set_data (Void)
			disable_pebble_positioning
			remove_pebble
			set_pick_and_drop_mode
			if has_capture then
				disable_capture
			end
			set_pointer_style ((create {EV_STOCK_PIXMAPS}).standard_cursor)
			enable_user_resize
			remove_menu_bar
			remove_title
			lower_bar.wipe_out
			upper_bar.wipe_out
			remove_icon_name
			accelerators.wipe_out
			
			-- FIXME: next two lines cause postcondition violation
			-- set_accept_cursor (Void)
			-- set_deny_cursor (Void)
			-- FIXME: next line should be set_target_name (Void)
			-- but it violates precondition
			set_target_name ("")
			-- FIXME: violates precondition
			-- set_actual_drop_target_agent (Void)
			-- FIXME: should be (0, 0) but violates precondition
			-- set_minimum_size (1, 1)

		end
	
	set_size_to_minimum is
			-- Set window to minimum size
		do
			if not is_destroyed then
				set_size (minimum_width, minimum_height)
			end
		end
	
invariant
	window_not_void: window /= Void
	accelerators_not_void: is_usable implies accelerators /= Void
	consistent_horizontal_bounds: is_usable implies maximum_width >= minimum_width
	consistent_vertical_bounds: is_usable implies maximum_height >= minimum_height
	upper_bar_not_void: is_usable implies upper_bar /= Void
	lower_bar_not_void: is_usable implies lower_bar /= Void
	client_width_within_bounds: is_usable implies client_width >= 0 and client_width <= width
	client_height_within_bounds: is_usable implies client_height >= 0 and client_height <= height
	pointer_position_not_void: is_usable and is_show_requested implies pointer_position /= Void
	is_displayed_implies_show_requested: is_usable and then is_displayed implies is_show_requested
	parent_contains_current: is_usable and then parent /= Void implies parent.has (window)
	user_interface_modes_mutually_exclusive: mode_is_pick_and_drop.to_integer + mode_is_drag_and_drop.to_integer + mode_is_target_menu.to_integer = 1
	is_initialized: is_initialized
	accept_cursor_not_void: accept_cursor /= Void
	deny_cursor_not_void: deny_cursor /= Void
	parent_permits_docking: is_dockable implies parent_of_source_allows_docking
	background_color_not_void: is_usable implies background_color /= Void
	foreground_color_not_void: is_usable implies foreground_color /= Void
	width_not_negative: is_usable implies width >= 0
	height_not_negative: is_usable implies height >= 0
	minimum_width_not_negative: is_usable implies minimum_width >= 0
	minimum_height_not_negative: is_usable implies minimum_height >= 0
end

