
indexing
	
	author: "David Hollenberg"
	date: "October 7, 1997"
	description: "A network-wide shared or exclusive advisory lock on a file"

class UNIX_FILE_LOCK

inherit
	LOCK
	UNIX_OS_ACCESS
		export
			{NONE} all
		end

create
	make

feature -- Creation

	make (lname: STRING)
			-- Create a lock object for file `lname',
			-- which must be the name of an existing
			-- file which is readable and writable by the
			-- caller.  Does not do any locking.
		require
			lock_name_exists: lname /= Void
		do
			create internal_file.make (lname)
			lock_name := lname
			is_exclusive := True
			minimize_retained_resources := True
		ensure
			lock_name_set: equal (lock_name, lname)
			exclusive_mode_lock: is_exclusive
			lock_not_locked: not is_locked_by_me
			resources_not_kept: minimize_retained_resources
		end

feature -- Status

	auxiliary_data: STRING
			-- Auxiliary data associated with lock
		do
			open_internal_file
			Result := lock_file_contents
			if not is_locked_by_me and minimize_retained_resources then
				internal_file.close
			end
		end
	
feature -- Modification

	lock_wait
			
			-- Obtain network-wide advisory lock on
			-- file `lock_name', with locking mode indicated by
			-- `is_exclusive'.  If file is currently locked
			-- by another process in a manner which prevents
			-- obtaining the lock, call blocks until lock
			-- is released.
		do
			open_internal_file
			unix_os.lock_file_wait (internal_file.descriptor, is_exclusive)
			is_locked_by_me := True
			conflicting_lock_exists := False
			lock_holder_process_id := 0
			lock_holder_system_id := 0
		end
	
	lock_nowait
			-- Attempt to obtain network-wide advisory
			-- lock on file `lock_name', with locking mode
			-- indicated by `is_exclusive'.  Set
			-- `is_locked_by_me' to indicate whether the
			-- lock was gotten.  Also sets
			-- `conflicting_lock_exists',
			-- `conflicting_lock_process_id' and
			-- `conflicting_lock_system_id' appropriately.
			-- Lock won't be gotten if it is currently
			-- locked by another process in a manner which
			-- prevents obtaining the lock.  Does not
			-- block if lock is not available
		do
			open_internal_file
			is_locked_by_me := unix_os.lock_file_nowait (internal_file.descriptor, is_exclusive)
			conflicting_lock_exists := not is_locked_by_me
			if conflicting_lock_exists then
				set_lock_holder_info
				conflicting_lock_exists := True
					-- Make sure this is true for
					-- consistency even though process
					-- holding conflicting lock may have
					-- released it
				
			else
				lock_holder_process_id := 0
				lock_holder_system_id := 0
			end
			if not is_locked_by_me and minimize_retained_resources then
				internal_file.close
			end
		end
	
	test_lock
			-- Set `conflicting_lock_exists' to indicate whether
			-- the network-wide lock on file `lock_name'
			-- is locked in a manner which would prevent
			-- obtaining the lock with mode specified by
			-- `is_exclusive'.  Also sets
			-- `conflicting_lock_process_id' and
			-- `conflicting_lock_system_id'.  Will not block
		do
			open_internal_file
			set_lock_holder_info
			if minimize_retained_resources then
				internal_file.close
			end
		end
	
	unlock
			-- Release lock on lock `lock_name'
		do
			unix_os.unlock_file (internal_file.descriptor)
			is_locked_by_me := False
			if minimize_retained_resources then
				internal_file.close
			end
		ensure then
			internal_file_closed: minimize_retained_resources implies internal_file.is_closed
		end

	set_auxiliary_data (data: STRING)
			-- Set auxiliary data associated with `Current'
			-- to `data'
		do
			internal_file.start
			unix_os.truncate_file (internal_file)
			internal_file.put_string (data)
			unix_os.flush_file_to_disk (internal_file)
		end

	append_auxiliary_data (data: STRING)
			-- Append `data' to auxiliary data associated with
			-- `Current'
		do
			internal_file.put_string (data)
			unix_os.flush_file_to_disk (internal_file)
		end

	clear_auxiliary_data
			-- Clear auxiliary data associated with `Current'
			-- (set it to the empty string)
		do
			set_auxiliary_data ("")
		end

feature {NONE} -- Implementation

	internal_file: LARGE_RAW_FILE
			-- For opening named file, since the operating
			-- system lock interface requires a file
			-- descriptor rather than a file name

	open_internal_file
			-- Open `internal_file' for read/write access.
			-- File will not be created if it does not exist
			-- nor will it be truncated if it does exist
		do
			if not internal_file.is_open_write then
				internal_file.open_read_append
			end
		ensure
			internal_file_readable: internal_file.is_open_read
			internal_file_writable: internal_file.is_open_append
		end

	set_lock_holder_info
			-- Set `lock_holder_process_id' and
			-- `lock_holder_system_id' to the process id and
			-- system id of the process holding a conflicting
			-- lock or 0 if no such process.  Also, set
			-- `conflicting_lock_exists' to indicate whether
			-- there is such a process
		do
			unix_os.get_conflicting_lock_info (internal_file.descriptor, is_exclusive, $lock_holder_process_id, $lock_holder_system_id)
			if lock_holder_process_id > 0 then
				conflicting_lock_exists := True
			else
				conflicting_lock_exists := False
			end
		end

	lock_file_contents: STRING
			-- Contents of `internal_file'
		require
			internal_file_readable: internal_file.is_open_read
		do
			create Result.make (256)
			internal_file.start
			from
				internal_file.read_stream (8192)
				Result.append (internal_file.last_string)
			until
				internal_file.end_of_file
			loop
				internal_file.read_stream (8192)
				Result.append (internal_file.last_string)
			end
		end

invariant
	internal_file_exists: internal_file /= Void

end
