
indexing
	
	description: "Unix-specific operating system services"
	author: "David Hollenberg"
	date: "October 7, 1997"

class UNIX_OS

inherit
	OPERATING_SYSTEM
		export
			{NONE} Directory_separator,
			Current_directory_name_representation,
			home_directory_supported, root_directory_supported,
			case_sensitive_path_names
		end
	UNIX_SIGNALS
		rename
			meaning as signal_meaning
		end
	UNIX_ERROR_CONSTANTS
	EXECUTION_ENVIRONMENT
		export
			{ANY} return_code 
			{NONE} all
		end
	STRING_FILLER_ACCESS
		export
			{NONE} all
		end
	INTERVAL_TIMER_SERVER_ACCESS
		export
			{NONE} all
		end
	C_MEMORY_ALLOCATION
		export
			{NONE} all
		end

feature -- Path names

	null_file_name: STRING = "/dev/null"
			-- File name which represents null input or output
	
	full_file_name (dir_name, f_name: STRING): STRING
			-- Full name of file in directory `dir_name'
			-- with name `f_name'.
		do
			create Result.make (dir_name.count + f_name.count + 1)
			if not dir_name.is_empty then
				Result.append (dir_name)
				if dir_name.item (dir_name.count) /= Directory_separator then
					Result.extend (Directory_separator)
				end
			end
			Result.append (f_name)
		end

	full_directory_name (dir_name, subdir: STRING): STRING
			-- Full name of subdirectory `subdir' of directory
			-- `dir_name'
		do
			Result := full_file_name (dir_name, subdir)
		end


feature -- User ids
	
	real_user_id: INTEGER
			-- Real user id of process, which indicates
			-- the user that is really running the process
		external
			"C use <unistd.h>"
		alias
			"getuid"
		end
	
	effective_user_id: INTEGER
			-- Effective user id of process, which controls
			-- what effective privileges process has
		external
			"C use <unistd.h>"
		alias
			"geteuid"
		end
	
	real_group_id: INTEGER
			-- Group user id of process, which indicates
			-- the group of the user that is really running
			-- the process
		external
			"C use <unistd.h>"
		alias
			"getgid"
		end
	
	effective_group_id: INTEGER
			-- Effective group id of process, which controls
			-- what effective privileges process has
		external
			"C use <unistd.h>"
		alias
			"getegid"
		end
	
	user_name_for_id (id: INTEGER): STRING
			-- User name corresponding to user id `id'
		local
			capacity, incr, count: INTEGER
			temp: STRING
			done: BOOLEAN
			area, temp_area: ANY
		do
			from
				incr := 256
				capacity := incr
				create temp.make (capacity)
				create Result.make (capacity)
				done := False
			until
				done
			loop
				temp.grow (capacity)
				Result.grow (capacity)
				Result.fill_character ('%U')
				area := Result.area
				temp_area := temp.area
				count := get_user_name_for_id (id, $area, capacity, $temp_area, capacity)
				done := count > 0
				capacity := capacity + incr
			end
			Result.keep_head (count)
		ensure
			result_exists: Result /= Void
		end
	
	group_name_for_id (id: INTEGER): STRING
			-- Group name corresponding to user id `id'
		local
			capacity, incr, count: INTEGER
			temp: STRING
			done: BOOLEAN
			area, temp_area: ANY
		do
			from
				incr := 256
				capacity := incr
				create temp.make (capacity)
				create Result.make (capacity)
				done := False
			until
				done
			loop
				temp.grow (capacity)
				Result.grow (capacity)
				Result.fill_character ('%U')
				area := Result.area
				temp_area := temp.area
				count := get_group_name_for_id (id, $area, capacity, $temp_area, capacity)
				done := count > 0
				capacity := capacity + incr
			end
			Result.keep_head (count)
		ensure
			result_exists: Result /= Void
		end
	
	user_id_for_name (name: STRING): INTEGER
			-- User id corresponding to user name `name' or -1
			-- if no such user
		local
			area: ANY
		do
			area := name.to_c
			Result := get_user_id_for_name ($area)
		end
	
	group_id_for_user_name (name: STRING): INTEGER
			-- Group id for user name `name' or -1
			-- if no such user
		local
			area: ANY
		do
			area := name.to_c
			Result := get_group_id_for_user_name ($area)
		end
	
	group_id_for_group_name (name: STRING): INTEGER
			-- Group id corresponding to group name `name' or -1
			-- if no such group
		local
			area: ANY
		do
			area := name.to_c
			Result := get_group_id_for_group_name ($area)
		end
	
	Root_user_name: STRING = "root"
			-- User name of super-user

	Root_user_id: INTEGER = 0
			-- User id for super-user (root)
	
	set_real_user_id (n: INTEGER)
			-- Set `real_user_id' to `n'.  May fail and
			-- raise exception even if `n' is a valid user id,
			-- due to permission problems
		require
			nonnegative_user_id: n >= 0
		external
			"C"
		ensure
			user_id_set: real_user_id = n
		end

	set_effective_user_id (n: INTEGER)
			-- Set `effective_user_id' to `n'.  May fail and
			-- raise exception even if `n' is a valid user id,
			-- due to permission problems
		require
			nonnegative_user_id: n >= 0
		external
			"C"
		ensure
			user_id_set: effective_user_id = n
		end

	set_real_group_id (n: INTEGER)
			-- Set `real_group_id' to `n'.  May fail and
			-- raise exception even if `n' is a valid group id,
			-- due to permission problems
		require
			nonnegative_group_id: n >= 0
		external
			"C"
		ensure
			group_id_set: real_group_id = n
		end

	set_effective_group_id (n: INTEGER)
			-- Set `effective_group_id' to `n'.  May fail and
			-- raise exception even if `n' is a valid group id,
			-- due to permission problems
		require
			nonnegative_group_id: n >= 0
		external
			"C"
		ensure
			group_id_set: effective_group_id = n
		end

	try_to_become_root
			-- Try to set `effective_user_id' to `Root_user_id'.
			-- Will not raise an exception if it fails.
			-- Caller must check `effective_user_id' to
			-- see if we succeeded
		local
			tried: BOOLEAN
		do
			if not tried then
				set_effective_user_id (Root_user_id)
			end
		rescue
			tried := True
			retry
		end

feature -- Resource limits

	soft_resource_limit (code: INTEGER): INTEGER
			-- Soft resource limit for resource with code
			-- `code'
		external
			"C"
		end

	hard_resource_limit (code: INTEGER): INTEGER
			-- Hard resource limit for resource with code
			-- `code'
		external
			"C"
		end

	set_soft_resource_limit (code, val: INTEGER)
			-- Set soft resource limit for resource with code
			-- `code' to `val'
		do
			set_resource_limits (code, val, hard_resource_limit (code))
		end

	set_soft_resource_limit_to_max (code: INTEGER)
			-- Set soft resource limit for resource with code
			-- `code' to `val'
		local
			lim: INTEGER
		do
			lim := hard_resource_limit (code)
			set_resource_limits (code, lim, lim)
		end

	set_hard_resource_limit (code, val: INTEGER)
			-- Set hard resource limit for resource with code
			-- `code' to `val'
		do
			set_resource_limits (code, soft_resource_limit (code), val)
		end

feature -- File creation mask

	set_file_creation_mask (mask: INTEGER): INTEGER
			-- Set file creation mask to `mask' and return
			-- previous value of mask
		external
			"C signature (mode_t): mode_t use <sys/stat.h>"
		alias
			"umask"
		end

feature -- Directory operations
	
	delete_directory_tree (dir_name: STRING)
			-- Try to delete the directory tree rooted at
			-- `dir_name'.  Ignore any errors.  Leave status
			-- code in `return_code'
		local
			cmd: STRING
		do
			create cmd.make (dir_name.count + 12)
			cmd.append ("/bin/rm -rf ")
			cmd.append (dir_name)
			system (cmd)
		end

	change_directory (path: STRING)
			-- Change working directory to `path'
		local
			area: ANY
		do
			area := path.to_c
			unix_change_directory ($area)
		end

feature -- File operations
	
	is_symbolic_link (path: STRING): BOOLEAN
			-- Is `path' a symbolic link?
		require
			path_not_void: path /= Void
		local
			area: ANY
		do
			area := path.to_c
			Result := unix_is_symbolic_link ($area)
		end

	make_symbolic_link (old_path, new_path: STRING)
			-- Create a symbolic link `new_path' to the
			-- file or directory `old_path'
		require
			old_path_not_void: old_path /= Void
			new_path_not_void: new_path /= Void
		local
			area1, area2: ANY
		do
			area1 := old_path.to_c
			area2 := new_path.to_c
			unix_symbolic_link ($area1, $area2)
		end

	make_hard_link (old_path, new_path: STRING)
			-- Create a hard link `new_path' to the existing
			-- file `old_path'
		require
			old_path_not_void: old_path /= Void
			new_path_not_void: new_path /= Void
		local
			area1, area2: ANY
		do
			area1 := old_path.to_c
			area2 := new_path.to_c
			unix_hard_link ($area1, $area2)
		end

	unlink (path: STRING)
			-- Unlink `path'
		require
			path_not_void: path /= Void
		local
			area: ANY
		do
			area := path.to_c
			unix_unlink ($area)
		end

	flush_file_to_disk (f: LARGE_FILE)
			-- Flush all changes made to `f' to the operating
			-- system and then to disk.  Does not return until
			-- all changes are on disk
		do
			f.flush
			flush_descriptor_to_disk (f.descriptor)
		end

	flush_descriptor_to_disk (d: INTEGER)
			-- Flush all writes to open file descriptor `d'
			-- to disk.  Does not return until
			-- all changes are on disk
		external
			"C"
		alias
			"unix_flush_file_to_disk"
		end

	truncate_file (f: LARGE_FILE)
			-- Truncate open file `f' to zero length
		do
			truncate_descriptor (f.descriptor)
		end

	truncate_descriptor (d: INTEGER)
			-- Truncate to zero length the file referenced
			-- by open file descriptor `d'
		external
			"C"
		alias
			"unix_truncate_descriptor"
		end

feature -- File locking

	new_lock (fname: STRING): LOCK
			-- New lock object for file named `fname'
		do
			create {UNIX_FILE_LOCK} Result.make (fname)
		end

feature -- File descriptor operations

	valid_file_descriptor (fd: INTEGER): BOOLEAN
			-- Is `fd' in the range of valid file descriptors?
		do
			Result := fd >= 0
		end

	max_file_descriptors: INTEGER
			-- Maximum number of file descriptors that may
			-- be open (current soft limit)
		external
			"C"
		end

	Stdin_descriptor: INTEGER = 0
			-- File descriptor for standard input

	Stdout_descriptor: INTEGER = 1
			-- File descriptor for standard output
	
	Stderr_descriptor: INTEGER = 2
			-- File descriptor for standard error

	Invalid_file_descriptor: INTEGER = -1
			-- File descriptor which is not in valid range

	duplicate_file_descriptor (old_fd, new_fd: INTEGER)
			-- Duplicate existing file descriptor `old_fd' and
			-- give the new file descriptor the value `new_fd'.
			-- If `new_fd' is in use, it is first deallocated
			-- as if it were closed
		require
			valid_old_descriptor: valid_file_descriptor (old_fd)
			valid_new_descriptor: valid_file_descriptor (new_fd)
		external
			"C"
		alias
			"unix_dup2"
		end
	
	clone_file_descriptor (old_fd: INTEGER): INTEGER
			-- Clone existing file descriptor `old_fd' and
			-- return the new file descriptor returned by
			-- the operating system
		require
			valid_old_descriptor: valid_file_descriptor (old_fd)
		external
			"C"
		alias
			"unix_dup"
		ensure
			valid_new_descriptor: valid_file_descriptor (Result)
		end
	
	read_file_descriptor (fd: INTEGER dest: STRING)
			-- Read at most `dest.capacity' - `dest.count' bytes
			-- from file descriptor `fd' and append to `dest'.  I
			-- Set `last_read_count' to number of bytes read.
			-- If zero, then end of file has been reached.
			-- If -1 then no input was available from non-blocking
			-- file descriptor
		require
			valid_read_descriptor: valid_file_descriptor (fd)
			destination_exists: dest /= Void
			destination_has_room: dest.capacity > dest.count
		local
			c_string: ANY
			capacity, count, rd_count: INTEGER
			p: POINTER
		do
			capacity := dest.capacity
			count := dest.count
			c_string := dest.area
			p := offset($c_string, count)
			unix_read (fd, p, capacity - count)
			rd_count := last_read_count
			if rd_count > 0 then
				strfill.set_count (dest, count + rd_count)
			end
		end
	
	read_socket_file_descriptor (fd: INTEGER dest: STRING always_exception: BOOLEAN)
			-- Read at most `dest.capacity' bytes from file
			-- descriptor `fd' which represents a socket
			-- and put in `dest'.
			-- Set `last_read_count' to number of bytes read.
			-- If zero, then end of file has been reached.
			-- If -1 then no input was available from non-blocking
			-- file descriptor.
			-- Raise an exception for all failures (even those
			-- caused by remote host or network) if
			-- `always_exception' is true
		require
			valid_read_descriptor: valid_file_descriptor (fd)
			destination_exists: dest /= Void
			destination_has_room: dest.capacity > 0
		local
			c_string: ANY
			len, rd_count: INTEGER
		do
			len := dest.capacity
			c_string := dest.area
			unix_socket_read (fd, $c_string, dest.capacity,
				always_exception)
			rd_count := last_read_count
			if rd_count < 0 then	
				-- No data for non-blocking descriptor
				rd_count := 0
			end
			strfill.set_count (dest, rd_count)
		end
	
	last_read_count: INTEGER
			-- Number of bytes read by last call to
			-- `unix_socket_read' or `unix_read'
			-- or -1 if non-blocking descriptor had no data ready
		external
			"C"
		end
	
	write_file_descriptor (fd: INTEGER src: STRING pos: INTEGER)
			-- Write as many bytes as possible from `src'
			-- starting at position `pos' to file
			-- descriptor `fd'.  Set `last_written_count'
			-- to number of bytes written.  Do not block
			-- if `fd' is non-blocking.  Raises a "Sigpipe
			-- signal" exception if end of file occurs
		require
			valid_write_descriptor: valid_file_descriptor (fd)
			source_exists: src /= Void
			position_large_enough: pos >= 1
			position_small_enough: pos <= src.count
		local
			c_string: ANY
			count: INTEGER
			p: POINTER
		do
			count := src.count
			c_string := src.area
			p := offset($c_string, pos - 1)
			unix_write (fd, p, count - pos + 1)
		end
	
	write_socket_file_descriptor (fd: INTEGER src: STRING always_exception: BOOLEAN)
			-- Write as many bytes as possible from `src'
			-- to file descriptor `fd', which represents a
			-- socket and which must be open for write.
			-- Set `last_written_count' to number of bytes
			-- written.  Raise an exception for all
			-- failures (even those caused by remote host
			-- or network) if `always_exception' is true
		require
			valid_write_descriptor: valid_file_descriptor (fd)
			source_exists: src /= Void
		local
			c_string: ANY
			len: INTEGER
		do
			len := src.count
			c_string := src.area
			unix_socket_write (fd, $c_string, len, always_exception)
		end
	
	last_written_count: INTEGER
			-- Number of bytes written by last call to
			-- `unix_socket_write' or `unix_write'
		external
			"C"
		end
	
	receive_file_descriptor (fd: INTEGER)
			-- Receive a file descriptor sent by another
			-- process on descriptor `fd', which must be a
			-- a stream pipe
		require
			valid_descriptor: valid_file_descriptor (fd)
		external
			"C"
		alias
			"unix_receive_descriptor"
		end
	
	last_received_descriptor: INTEGER
			-- Last descriptor received by a call to
			-- `receive_file_descriptor'
		external
			"C"
		end
	
	close_file_descriptor (fd: INTEGER)
			-- Close existing open file descriptor `fd'
		require
			valid_descriptor: valid_file_descriptor (fd)
		external
			"C"
		alias
			"unix_close"
		end

	close_file_descriptor_without_exception (fd: INTEGER)
			-- Close existing open file descriptor `fd'
			-- Do not raise an exception if there is
			-- an error.  Raising an exception during
			-- a call to `dispose' screws up the garbage
			-- collector
		require
			valid_descriptor: valid_file_descriptor (fd)
		local
			rc: INTEGER
		do
			rc := unix_close_without_exception (fd)
		end

	mark_file_descriptor_close_on_exec (fd: INTEGER)
			-- Mark file descriptor `fd' as "close on exec"
			-- if it is open.  Do nothing if it is not open
		require
			valid_descriptor: valid_file_descriptor (fd)
		external
			"C"
		alias
			"mark_descriptor_close_on_exec"
		end

feature -- Terminal operations

	is_terminal (fd: INTEGER): BOOLEAN
			-- Is file descriptor `fd' associated with a
			-- terminal device?
		require
			valid_descriptor: valid_file_descriptor (fd)
		do
			Result := is_a_tty (fd) = 1
		end

	is_terminal_echo_on (fd: INTEGER): BOOLEAN
			-- Is echoing of input characters turned on for
			-- file descriptor `fd', which must be associated
			-- with a terminal
		require
			valid_descriptor: valid_file_descriptor (fd)
		external
			"C"
		end

	turn_terminal_echo_off (fd: INTEGER)
			-- Turn off echoing of input characters for
			-- file descriptor `fd', which must be associated
			-- with a terminal
		require
			valid_descriptor: valid_file_descriptor (fd)
		external
			"C"
		ensure
			echo_off: not is_terminal_echo_on (fd)
		end

	turn_terminal_echo_on (fd: INTEGER)
			-- Turn on echoing of input characters for
			-- file descriptor `fd', which must be associated
			-- with a terminal
		require
			valid_descriptor: valid_file_descriptor (fd)
		external
			"C"
		ensure
			echo_off: is_terminal_echo_on (fd)
		end

	prompt_noecho (the_prompt, the_response: STRING)
			-- Display prompt and then read response while
			-- terminal echoing is off.  Restore terminal
			-- echoing to its initial state
		require
			prompt_exists: the_prompt /= Void
			response_exists: the_response /= Void
		local
			echo_on: BOOLEAN
			fd: INTEGER
		do
			fd := io.input.descriptor
			echo_on := is_terminal_echo_on (fd)
			io.put_string (the_prompt)
			if echo_on then
				turn_terminal_echo_off (fd)
				io.read_line
				the_response.copy (io.last_string)
				turn_terminal_echo_on (fd)
				io.new_line
			end
		ensure
			same_terminal_state: is_terminal_echo_on (io.input.descriptor) = old is_terminal_echo_on (io.input.descriptor)
		rescue
			if echo_on then
				turn_terminal_echo_on (fd)
			end
		end


feature -- Interval timer

	system_interval_timer: UNIX_SYSTEM_INTERVAL_TIMER
			-- Operating system's interval timer
		once
			create Result
		end

feature -- Pipes

	new_pipe: UNIX_PIPE
			-- New pipe object for interprocess communication
		local
			read_fd, write_fd: INTEGER
		do
			unix_pipe ($read_fd, $write_fd)
			create Result.make (read_fd, write_fd)
		ensure
			result_exists: Result /= Void
		end

feature -- Environment operations

	current_environment: HASH_TABLE [STRING, STRING]
			-- Table of values in current environment,
			-- indexed by environment variable name
		do
			Result := environment_from_c (unix_env_pointer)
		end

	environment_from_c (p: POINTER): HASH_TABLE [STRING, STRING]
			-- Table of values in environment indicated
			-- by memory pointed to by `p' (array of char *),
			-- indexed by environment variable name
		require
			environment_exists: p /= default_pointer
		local
			var, null_ptr: POINTER
			k, pos: INTEGER
			s, name, val: STRING
		do
			create Result.make (100)
			from
				k := 0
				var := nth_pointer (p, k)
			until
				var = null_ptr
			loop
				create s.make_from_c (var)
				pos := s.index_of ('=', 1)
				if pos > 0 then
					name := s.substring (1, pos - 1)
					val := s.substring (pos + 1, s.count)
					Result.force (val, name)
				end
				k := k + 1
				var := nth_pointer (p, k)
			end
		ensure
			result_exists: Result /= Void
		end

	environment_to_c (table: HASH_TABLE [STRING, STRING]): UNIX_ENVIRONMENT
			-- Environment corresponding to the one
			-- specified by `table', in a form suitable
			-- for passing to Unix `execve' routine
		require
			table_exists: table /= Void
		local
			keys: ARRAY [STRING]
			key, val, pair: STRING
			src: ANY
			env: UNIX_ENVIRONMENT
			dest: C_MEMORY_CHUNK
			k, len: INTEGER
			null_ptr, addr: POINTER
		do
			keys := table.current_keys
			create env.make (keys.count)
			env.clear
			from
				k := keys.lower
			until
				k > keys.upper
			loop
				key := keys.item (k)
				val := table.item (key)
				create pair.make (key.count + val.count + 1)
				pair.append (key)
				pair.extend ('=')
				pair.append (val)
				len := pair.count + 1
				create dest.make (len)
				src := pair.to_c
				dest.copy_to_address (src, len)
				addr := dest.address
				dest.clear_address
				env.store_pointer (addr, k - keys.lower)
				k := k + 1
			end
			env.store_pointer (null_ptr, k - keys.lower)
			Result := env
		ensure
			result_exists: Result /= Void
		end

feature -- Process operations

	my_process_id: INTEGER
			-- Process id of currently executing process
		do
			Result := unix_get_process_id
		end

	parent_process_id: INTEGER
			-- Process id of parent of currently executing process
		external
			"C"
		end

	process_group_id: INTEGER
			-- Process group id of currently executing process
		external
			"C"
		end

	local_host_name: STRING
			-- New string with name of host on which
			-- current process is executing
		local
			capacity, incr: INTEGER
			done: BOOLEAN
			area: ANY
		do
			from
				incr := 32
				capacity := incr
				create Result.make (capacity)
				done := False
			until
				done
			loop
				Result.grow (capacity)
				Result.fill_character ('%U')
				area := Result.area
				unix_get_host_name ($area, Result.capacity)
				done := Result.item (Result.count) = '%U'
				capacity := capacity + incr
			end
			Result.keep_head (Result.index_of ('%U', 1) - 1)
		end

	local_host_short_name: STRING
			-- New string with short name of host on which
			-- current process is executing
		local
			pos: INTEGER
		do
			Result := local_host_name
			pos := Result.index_of ('.', 1)
			if pos > 0 then
				Result.keep_head (pos - 1)
			end
		end

	is_sunos: BOOLEAN
			-- Is process executing on a host running SunOS?
		do
			Result := unix_is_sunos
		end

	is_solaris: BOOLEAN
			-- Is process executing on a host running Solaris?
		do
			Result := unix_is_solaris
		end

	new_process (fname: STRING): UNIX_PROCESS
			-- New process object for file named `fname'.
			-- Make routine calls `initialize' on the new object
		do
			create {UNIX_PROCESS} Result.make (fname)
		ensure
			result_exists: Result /= Void
		end

	fork_process: INTEGER
			-- Fork a new process.  Return process id of new
			-- process to the parent process and 0 to the child
		do
			Result := unix_fork_process
		ensure
			process_id_nonnegative: Result >= 0
		end

	exec_process (prog_file: STRING args: ARRAY [STRING]
			env_table: HASH_TABLE [STRING, STRING])
			-- Overlay the process with a new process,
			-- which will execute program `prog_file' with
			-- arguments `args'.  If `env_table' is not
			-- void, it specifies the environment variable
			-- names and values for the new process;
			-- otherwise, the new process has the same
			-- environment as the current one.  This
			-- routine never returns to the caller
			-- normally, although it may raise an
			-- exception
		require
			program_name_exists: prog_file /= Void
			arguments_exist: args /= Void
		local
			k, count, lower: INTEGER
			pname, area: ANY
			arguments, arg_copy, null_ptr, env_ptr: POINTER
			env: UNIX_ENVIRONMENT
		do
			count := args.count
			lower := args.lower
			pname := prog_file.to_c
			arguments := unix_allocate_arg_memory (count + 1)
			from
				k := 1
			until
				k > count
			loop
				area := args.item (lower + k - 1).to_c
				arg_copy := str_dup ($area)
				unix_set_arg_value (arguments, k - 1, arg_copy)
				k := k + 1
			end
			unix_set_arg_value (arguments, count, null_ptr)
			if env_table /= Void then
				env := environment_to_c (env_table)
				env_ptr := env.address
			else
				env_ptr := default_pointer
			end
			unix_exec_process ($pname, arguments, env_ptr)
		end

	wait_for_process_block (pid: INTEGER): INTEGER
			-- Wait for any process specified by process id
			-- `pid' to return status.  Block until there
			-- a process returns status.  Return the exit
			-- or termination status of the process that
			-- finished.  It is an error to wait if there
			-- are no processes to wait for.  If `pid' > 0,
			-- wait for just the process with that process id.
			-- If `pid' = -1, wait for any child process.
			-- For effect of other values of `pid', see waitpid(2).
		local
			dummy: BOOLEAN
		do
			Result := unix_waitpid (pid, True, $dummy)
		end
	
	wait_for_process_noblock (pid: INTEGER status_avail_addr: POINTER): INTEGER
			-- Wait for any process specified by process
			-- id `pid' to return status.  Do not block if
			-- no process has finished (the return value
			-- is not meaningful in this case).  Set
			-- boolean whose address is in
			-- `status_avail_addr' to indicate whether
			-- status was available for any process.
			-- Return the exit or termination status of
			-- the process that finished.
		do
			Result := unix_waitpid (pid, False, status_avail_addr)
		end
	
	send_signal (sig, pid: INTEGER)
			-- Send signal `sig' to the process(es)
			-- identified by `pid'.  If signal is 0, error
			-- checking is done but no signal is actually
			-- sent.  If `pid' > 0, send signal to the
			-- process with that id.  If `pid' = 0, send
			-- signal to all processes with the same
			-- process group id as current process (except
			-- special processes).  For `pid' < 0 and is
			-- not -1, send signal to all processes whose
			-- process group id is absolute value of `pid'
			-- to which this process has permission to
			-- send a signal.  For `pid' = -1, see man
			-- page for kill(2).
			-- "No such process or process group" errors
			-- are silently ignored and do not raise an exception.
		require
			valid_signal: is_defined(sig)
		do
			unix_kill (pid, sig)
		end
	
	send_signal_to_process_group (sig, pgid: INTEGER)
			-- Send signal `sig' to the processes whose
			-- process group id is `pgid'.  "No such
			-- process or process group" errors are
			-- silently ignored and do not raise an
			-- exception.
		require
			valid_signal: is_defined(sig)
		do
			send_signal (sig, -pgid)
		end
	
	show_signal_action (sig: INTEGER)
			-- Show action for signal `sig'
		require
			valid_signal: is_defined(sig)
		external
			"C"
		end
	
	terminate_soft (pid: INTEGER)
			-- Send a software termination signal (SIGTERM)
			-- to the process(es) identified by `pid'
		do
			send_signal (Sigterm, pid)
		end
	
	terminate_hard (pid: INTEGER)
			-- Send a kill signal (SIGKILL) to the process(es)
			-- identified by `pid'
		do
			send_signal (Sigkill, pid)
		end
	
	die_without_handlers (code: INTEGER)
			-- Terminate execution with exit status `code',
			-- without calling any termination handlers
		do
			unix_exit_without_handlers (code)
		end

	become_process_group_leader
			-- Make this process a process group leader.
		do
			set_process_group_id (0, 0)
		ensure
			is_group_leader: my_process_id = process_group_id
		end

	set_process_group_id (pid, pgid: INTEGER)
			-- Set process group id of process whose process
			-- id is `pid' to `pgid'.  If `pid' is 0, use
			-- process id of calling process.  If `pgid' is 0,
			-- the process specified by `pid' becomes a
			-- process group leader
		external
			"C"
		end
	
	become_daemon
			-- Become a daemon process.  Process will not
			-- have a controlling terminal and will not
			-- be able to automatically acquire one.
			-- Current directory is changed to root directory.
			-- Closes all descriptors and then opens standard
			-- input, output and error on /dev/null
		local
			pid: INTEGER
		do
			pid := unix_fork_process
			if pid /= 0 then	-- Parent
				die_without_handlers (0)
			end
			set_session_id
			ignore (Sighup)
			pid := unix_fork_process
			if pid /= 0 then	-- Parent
				die_without_handlers (0)
			end
			change_working_directory ("/")
			close_all_descriptors
			open_std_descriptors_null
		end

	set_session_id
			-- Set the process group ID and session ID of
     			-- the calling process to the process ID of
			-- the calling  process and release the process's
			-- controlling terminal
		external
			"C"
		alias
			"unix_set_session_id"
		end

	close_all_descriptors
			-- Close all possible descriptors for process
		external
			"C"
		end
	
	open_std_descriptors_null
			-- Open descriptors for standard input, output
			-- and error with /dev/null as filename
		external
			"C"
		end
	
feature -- Process status queries

	is_exit_status (status: INTEGER): BOOLEAN
			-- Does returned status `status' indicate that
			-- process finished due to a call to exit?
		external
			"C"
		alias
			"status_is_exit"
		end
	
	is_signal_status (status: INTEGER): BOOLEAN
			-- Does returned status `status' indicate that
			-- process was terminated by a signal?
		external
			"C"
		alias
			"status_is_signal"
		end
	
	is_stopped_status (status: INTEGER): BOOLEAN
			-- Does returned status `status' indicate that
			-- process was stopped by a signal?
		external
			"C"
		alias
			"status_is_stop"
		end
	
	exit_code (status: INTEGER): INTEGER
			-- Exit code indicated by `status'
		require
			process_called_exit: is_exit_status (status)
		external
			"C"
		alias
			"unix_exit_code"
		end
	
	signal_number (status: INTEGER): INTEGER
			-- Signal code indicated by `status'
		require
			process_got_signal: is_signal_status (status) or is_stopped_status (status)
		external
			"C"
		alias
			"unix_signal_number"
		end
	
	status_description (status: INTEGER): STRING
			-- Description of status `status' suitable for
			-- display
		do
			if is_exit_status (status) then
				Result := "exited with code "
				Result.append_integer (exit_code (status))
			elseif is_signal_status (status) then
				Result := "terminated by signal "
				Result.append_integer (signal_number (status))
			elseif is_stopped_status (status) then
				Result := "stopped by signal "
				Result.append_integer (signal_number (status))
			else
				Result := "unknown process status"
			end
		ensure
			result_exists: Result /= Void
		end
	
feature -- Exception queries

	is_broken_pipe: BOOLEAN
			-- Is last exception originally due to a broken
			-- pipe?
		local
			exc: EXCEPTIONS
			code: INTEGER
			tag: STRING
		do
			create exc
			code := exc.original_exception
			tag := exc.original_tag_name
			Result := ((code = exc.Io_exception or
				code = exc.Operating_system_exception) and
			  	equal (tag, error_string (Epipe))) or
			  (code = exc.Signal_exception and signal = Sigpipe)
		end

	is_hangup: BOOLEAN
			-- Is last exception originally due to a terminal
			-- hangup?
		local
			exc: EXCEPTIONS
		do
			create exc
			Result := exc.original_exception = exc.Signal_exception and signal = Sighup
		end

	is_interrupt: BOOLEAN
			-- Is last exception originally due to an interrupt?
		local
			exc: EXCEPTIONS
		do
			create exc
			Result := exc.original_exception = exc.Signal_exception and signal = Sigint
		end

	is_alarm: BOOLEAN
			-- Is last exception originally due to an alarm?
		local
			exc: EXCEPTIONS
		do
			create exc
			Result := exc.original_exception = exc.Signal_exception and signal = Sigalrm
		end

	is_software_termination: BOOLEAN
			-- Is last exception originally due to a software
			-- termination signal?
		local
			exc: EXCEPTIONS
		do
			create exc
			Result := exc.original_exception = exc.Signal_exception and signal = Sigterm
		end

feature -- Date and time
	
	current_date_string: STRING
			-- String representing current date and time
			-- in format produced by `asctime'
		do
			create Result.make (Date_string_length + 2)
				-- Allow room for trailing LF and NUL
			append_current_date_string (Result)
		end

	append_current_date_string (dest: STRING)
			-- Append to `dest' a string representing
			-- current date and time in format produced
			-- by `asctime'
		do
			append_date_string_from_integer (current_time_in_seconds, dest)
		end

	current_time_in_seconds: INTEGER
			-- Current time in seconds since the start of
			-- the epoch (00:00:00 GMT,  Jan.  1,  1970)
		external
			"C"
		end

	current_time_in_fine_seconds: DOUBLE
			-- Current time in seconds since the start of
			-- the epoch (00:00:00 GMT,  Jan.  1,  1970), with
			-- a fine resolution
		external
			"C"
		end

	integer_to_date_string (the_date: INTEGER): STRING
			-- Date represented by `the_date' in format
			-- produced by `asctime'.  `the_date' is a
			-- date represented as number of seconds since
			-- the start of the epoch ((00:00:00 GMT, Jan. 1, 1970)
		do
			create Result.make (Date_string_length + 2)
				-- Allow room for trailing LF and NUL
			append_date_string_from_integer (the_date, Result)
		ensure
			result_exists: Result /= Void
		end

	append_date_string_from_integer (the_date: INTEGER dest: STRING)
			-- Append date represented by `the_date' to `dest'
			-- in format produced by `asctime'.  `the_date' is a
			-- date represented as number of seconds since
			-- the start of the epoch ((00:00:00 GMT, Jan. 1, 1970)
		require
			destination_exists: dest /= Void
		do
			last_date_string.from_c (unix_date_to_string (the_date))
			last_date_string.keep_head (Date_string_length)
				-- Get rid of trailing newline and null
			dest.append (last_date_string)
		end

feature -- Sleeping
	
	sleep_microseconds (n: DOUBLE)
			-- Suspend execution for `n' microseconds.
			-- Actual time could be longer or shorter
			-- since routine is awakened by any signal
			-- (not just the SIGALRM signal)
		do
			sleep_seconds (n / 1.0E6)
		end

	sleep_seconds (n: DOUBLE)
			-- Suspend execution for `n' seconds.
			-- Actual time could be longer or shorter
			-- since routine is awakened by any signal
			-- (not just the SIGALRM signal)
		local
			timer: INTERVAL_TIMER
			tried: BOOLEAN
		do
			if not tried then
				create timer.set_seconds (n)
				unix_pause
			
				-- Shouldn't ever get here since signal will
				-- raise an exception and take us to
				-- rescue clause
				timer.clear
			end
		rescue
			if timer /= Void and then not timer.is_expired then
				timer.clear
			end
			if timer /= Void and then timer.is_expiration_exception then
				tried := True
				retry
			end
		end

feature -- Error handling
	
	current_error_number: INTEGER
			-- Current value of errno global variable
		external
			"C"
		alias
			"unix_error_number"
		end

	last_error_number: INTEGER
			-- Last saved value of errno global variable
			-- (saved when system call had error)
		external
			"C"
		alias
			"unix_last_error_number"
		end

	clear_last_error_number
			-- Set saved value of errno global variable to 0
		external
			"C"
		alias
			"unix_clear_last_error_number"
		end

	error_string (code: INTEGER): STRING
			-- Textual description of meaning of error code `code'
		do
			create Result.make (30)
			Result.from_c (unix_error_string (code))
		ensure
			result_exists: Result /= Void
		end

feature {NONE} -- Implementation

	Date_string_length: INTEGER = 24
			-- Length of string produced by `asctime', not
			-- including trailing newline and null character

	Last_date_string: STRING
		once
			create Result.make (Date_string_length + 2)
		end

	Infinity: INTEGER = 2147483647
			-- Largest positive integer (assumes 32 bit
			-- integers 2's complement representation)

feature {UNIX_FILE_LOCK} -- Externals

	lock_file_wait (fd: INTEGER exclusive: BOOLEAN)
			-- Obtain network-wide advisory lock
			-- on open file descriptor `fd', which must be
			-- open with read/write access to the file to be
			-- locked.  Lock mode is indicated by `exclusive'.
			-- If file is currently locked by
			-- another process, call blocks until lock is
			-- released.
		require
			valid_descriptor: valid_file_descriptor (fd)
		external
			"C"
		end
	
	lock_file_nowait (fd: INTEGER exclusive: BOOLEAN): BOOLEAN
			-- Attempt to obtain network-wide
			-- advisory lock on open file descriptor `fd',
			-- which must be open with read/write access to the
			-- file to be locked.  Lock mode is indicated by
			-- `exclusive'.  Do not block if another
			-- process has the lock.  The returned value
			-- indicates whether the lock was obtained
		require
			valid_descriptor: valid_file_descriptor (fd)
		external
			"C"
		end
	
	get_conflicting_lock_info (fd: INTEGER exclusive: BOOLEAN process_id_addr, system_id_addr: POINTER)
			-- Get process id and system id of process holding
			-- lock that would conflict with a shared or
			-- exclusive mode lock on file indicated by
			-- open file descriptor `fd'.  Put them in attributes
			-- at addresses `process_id_addr' and
			-- `system_id_addr' or set these attributes to
			-- 0 if no such process.  Does not block
		require
			valid_descriptor: valid_file_descriptor (fd)
		external
			"C"
		end
	
	unlock_file (fd: INTEGER)
			-- Unlock the file indicated by open file descriptor
			-- `fd'
		require
			valid_descriptor: valid_file_descriptor (fd)
		external
			"C"
		end

feature {NONE} -- Externals

	str_dup (area: POINTER): POINTER
			-- Return new copy of C string indicated by `area'
		external
			"C"
		end

	unix_error_string (code: INTEGER): POINTER
		external
			"C"
		end

	unix_is_symbolic_link (path: POINTER): BOOLEAN
		external
			"C"
		end

	unix_symbolic_link (old_path, new_path: POINTER)
		external
			"C"
		end

	unix_change_directory (path: POINTER)
		external
			"C"
		end

	unix_hard_link (old_path, new_path: POINTER)
		external
			"C"
		end

	unix_unlink (path: POINTER)
		external
			"C"
		end

	unix_close_without_exception (fd: INTEGER): INTEGER
			-- Close existing open file descriptor `fd' and
			-- return the return code from the OS.
			-- Do not raise an exception if there is
			-- an error
		require
			valid_descriptor: valid_file_descriptor (fd)
		external
			"C signature (int): int use <unistd.h>"
		alias
			"close"
		end

	unix_pipe (read_fd, write_fd: POINTER)
			-- Create a new pipe and put the read file descriptor
			-- in `read_fd' and the write file descriptor in
			-- `write_fd'
		external
			"C"
		end

	unix_kill (pid, sig: INTEGER)
			-- Send signal `sig' to process(es) identified by `pid'
		external
			"C"
		end

	unix_waitpid (pid: INTEGER block: BOOLEAN status_avail_addr: POINTER): INTEGER
			-- Wait for process specified by `pid'.  Block if
			-- no process has status available if `block' is
			-- true.  Set boolean at `status_avail_addr' to
			-- indicate whether status was available
		external
			"C"
		end

	unix_fork_process: INTEGER
			-- Create a new process.  Return the process id
			-- to the parent and 0 to the child
		external
			"C"
		end

	unix_exec_process (pname, args, env: POINTER)
			-- Call execv or execve to overlay current process with
			-- new one.  Does not return (raises exception
			-- if error doing the exec)
		external
			"C"
		end

	unix_get_process_id: INTEGER
			-- Process id of currently executing process
		external
			"C"
		end
	
	unix_get_host_name (buffer: POINTER len: INTEGER)
			-- Put name of current host in `buffer' which
			-- can hold `len' bytes.  Name is null-terminated
			-- if buffer has enough room
		external
			"C"
		end
	
	get_user_name_for_id (id: INTEGER buffer: POINTER len: INTEGER temp_buffer: POINTER temp_len: INTEGER): INTEGER
			-- Put name of user id `id' in `buffer' which
			-- can hold `len' bytes.  Use `temp_buffer'
			-- which can hold `temp_len' bytes for
			-- temporary storage.  Return length of user
			-- name or 0 if buffer wasn't big enough
		external
			"C"
		end
	
	get_group_name_for_id (id: INTEGER buffer: POINTER len: INTEGER temp_buffer: POINTER temp_len: INTEGER): INTEGER
			-- Put name of group id `id' in `buffer' which
			-- can hold `len' bytes.  Use `temp_buffer'
			-- which can hold `temp_len' bytes for
			-- temporary storage.  Return length of group
			-- name or 0 if buffer wasn't big enough
		external
			"C"
		end
	
	get_user_id_for_name (user_name: POINTER): INTEGER
			-- User id for user name `user_name' or -1 if
			-- no such user
		external
			"C"
		end
	
	get_group_id_for_user_name (user_name: POINTER): INTEGER
			-- Group id for user name `name' or -1
			-- if no such user
		external
			"C"
		end
	
	get_group_id_for_group_name (grp_name: POINTER): INTEGER
			-- Group id for group name `grp_name' or -1 if
			-- no such group
		external
			"C"
		end
	
	unix_is_sunos: BOOLEAN
			-- Is current host running SunOS operating system?
		external
			"C"
		end
	
	unix_is_solaris: BOOLEAN
			-- Is current host running Solaris operating system?
		external
			"C"
		end
	
	unix_exit_without_handlers (code: INTEGER)
		external
			"C signature (int) use <unistd.h>"
		alias
			"_exit"
		end

	unix_allocate_arg_memory (count: INTEGER): POINTER
			-- Return pointer to newly allocated memory
			-- for `count' arguments which is not subject
			-- to garbage collection.  `count' must
			-- include both argument 0 and the trailing
			-- null pointer that terminates the argument list
		external
			"C"
		end

	unix_set_arg_value (arg_array: POINTER pos: INTEGER arg: POINTER)
			-- Set the element of `arg_array' at position `pos'
			-- (relative to 0) to `arg'
		external
			"C"
		end

	unix_env_pointer: POINTER
			-- Pointer to current environment (null-terminated
			-- array of char *)
		external
			"C"
		end

	unix_date_to_string (the_date: INTEGER): POINTER
		external
			"C"
		end

	unix_pause
			-- Pause until signal is received
		external
			"C"
		end

	unix_read (fd: INTEGER dest: POINTER count: INTEGER)
			-- Read at most `count' bytes from file
			-- descriptor `fd' and put in `dest'
		external
			"C"
		end
	
	unix_socket_read (fd: INTEGER dest: POINTER count: INTEGER always_exception: BOOLEAN)
			-- Read at most `count' bytes from socket file
			-- descriptor `fd' and put in `dest'.  Raise an
			-- exception for all failures (even network failures)
			-- if `always_exception' is true
		external
			"C"
		end
	
	unix_write (fd: INTEGER src: POINTER count: INTEGER)
			-- Write at most `count' bytes from `src' to
			-- file descriptor `fd'.  Set
			-- `last_written_count' to number of bytes
			-- written.
		external
			"C"
		end
	
	unix_socket_write (fd: INTEGER src: POINTER count: INTEGER always_exception: BOOLEAN)
			-- Write up to `count' bytes from `src' to socket file
			-- descriptor `fd', which must be open for write.
			-- Set `last_written_count' to number of bytes
			-- written.
			-- Raise an exception for all failures (even
			-- network failures) if `always_exception' is true
		external
			"C"
		end
	
	is_a_tty (fd: INTEGER): INTEGER
			-- 1 if file descriptor `fd' is associated with a
			-- terminal device, 0 otherwise
		external
			"C signature (int): int use  <stdlib.h>"
		alias
			"isatty"
		end
	
	set_resource_limits (code: INTEGER soft_lim, hard_lim: INTEGER)
			-- Set resource limits for resource with code `code'
		external
			"C"
		end
	
end
