
indexing
	
	author: "David Hollenberg"
	date: "October 9, 1997"
	description: "A Unix process"

class UNIX_PROCESS

inherit
	UNIX_SIGNALS
		rename
			meaning as signal_meaning
		export
			{ANY} all
			{NONE} is_defined, is_ignored, signal, catch, ignore,
				reset_all_default, reset_default, is_caught,
				c_signal_map, c_signal_name
		end
	UNIX_OS
		rename
			send_signal as unix_send_signal,
			terminate_soft as unix_terminate_soft,
			terminate_hard as unix_terminate_hard
		export
			{ANY} is_exit_status, is_signal_status,
				is_stopped_status, exit_code, signal_number,
				status_description, valid_file_descriptor,
				Stdin_descriptor, Stdout_descriptor,
				Stderr_descriptor
			{NONE} all
		end
	OPERATING_ENVIRONMENT
		export
			{NONE} all
		end
create
	make

feature {UNIX_OS} -- Creation

	make (fname: STRING)
			-- Create a process object which represents an
			-- independent process that can execute the
			-- program residing in file `fname'
		require
			file_name_exists: fname /= Void
			file_name_not_empty: not fname.is_empty
		do
			initialize (fname)
		ensure
			file_name_set: program_file_name.is_equal (fname)
		end

feature -- Initialization

	initialize (fname: STRING)
			-- Initialize `Current' with program residing
			-- in file `fname'
		require
			file_name_exists: fname /= Void
			file_name_not_empty: not fname.is_empty
			process_not_executing: not is_executing
		local
			fd: PROCESS_FD
		do
			program_file_name := fname
			is_executing := False
			process_id := 0
			in_child := False
			status := 0
			
			create fd_set.make (Stdin_descriptor, Stderr_descriptor)
			create {FD_UNCHANGED} fd.make (Stdin_descriptor)
			fd_set.put (fd, Stdin_descriptor)
			create {FD_UNCHANGED} fd.make (Stdout_descriptor)
			fd_set.put (fd, Stdout_descriptor)
			create {FD_UNCHANGED} fd.make (Stderr_descriptor)
			fd_set.put (fd, Stderr_descriptor)

			arguments := Void
			process_name := Void
		ensure
			file_name_set: program_file_name.is_equal (fname)
			leave_input_unchanged: input_file_name = Void
			leave_output_unchanged: output_file_name = Void
			leave_error_unchanged: error_file_name = Void
			input_not_piped: not input_piped
			output_not_piped: not output_piped
			error_not_piped: not error_piped
			input_not_descriptor: not valid_file_descriptor (input_descriptor)
			output_not_descriptor: not valid_file_descriptor (output_descriptor)
			error_not_descriptor: not valid_file_descriptor (error_descriptor)
			no_input_pipe: input_to_child = Void
			no_output_pipe: output_from_child = Void
			no_error_pipe: error_from_child = Void
			no_arguments: arguments = Void
			no_process_name: process_name = Void
			process_not_executing: not is_executing
			no_process_id: process_id = 0
			not_in_child: not in_child
		end

feature -- Properties

	program_file_name: STRING
			-- Name of file containing program which will be
			-- executed when process is spawned

	process_name: STRING
			-- Name to be assigned to process.  Passed as
			-- argument 0 when process is spawned.  If Void,
			-- then last component of `program_file_name'
			-- is passed

	arguments: ARRAY [STRING]
			-- Arguments to passed to process when it is spawned,
			-- not including argument 0 (which is conventionally
			-- the name of the program).  If Void or if count
			-- is zero, then no arguments are passed

	environment: HASH_TABLE [STRING, STRING]
			-- Table of environment values for process, indexed
			-- by environment variable name.  If Void, process
			-- gets same environment as parent

	close_nonstandard_files: BOOLEAN
			-- Should nonstandard file descriptors (files whose
			-- disposition in child process is not specified)
			-- be closed in the spawned process?

	process_group_leader: BOOLEAN
			-- Should spawned process become a process group
			-- leader?

	input_descriptor: INTEGER
			-- Descriptor to be used as standard input in
			-- spawned process, if value is valid descriptor
		do
			Result := fd_set.item (Stdin_descriptor).compatible_descriptor
		end

	output_descriptor: INTEGER
			-- Descriptor to be used as standard output in
			-- spawned process, if value is valid descriptor
		do
			Result := fd_set.item (Stdout_descriptor).compatible_descriptor
		end

	error_descriptor: INTEGER
			-- Descriptor to be used as standard error in
			-- spawned process, if value is valid descriptor
		do
			Result := fd_set.item (Stderr_descriptor).compatible_descriptor
		end

	input_file_name: STRING
			-- Name of file to be used as standard input in
			-- spawned process if `input_descriptor' is not a
			-- valid descriptor and `input_piped' is false.
			-- A Void value leaves standard input same as
			-- parent's and an empty string closes standard input
		do
			Result := fd_set.item (Stdin_descriptor).compatible_file_name
		end

	output_file_name: STRING
			-- Name of file to be used as standard output in
			-- spawned process if `output_descriptor' is not a
			-- valid descriptor and `output_piped' is false.
			-- A Void value leaves standard output same as
			-- parent's and an empty string closes standard output
		do
			Result := fd_set.item (Stdout_descriptor).compatible_file_name
		end

	error_file_name: STRING
			-- Name of file to be used as standard error in
			-- spawned process if `error_descriptor' is not a
			-- valid descriptor and `error_piped' is false.
			-- A Void value leaves standard error same as
			-- parent's and an empty string closes standard error
		do
			Result := fd_set.item (Stderr_descriptor).compatible_file_name
		end

	input_piped: BOOLEAN
			-- Should standard input for spawned process
			-- come from a pipe connected to parent,
			-- instead of from a file descriptor or named file?
		do
			Result := fd_set.item (Stdin_descriptor).is_pipe
		end

	output_piped: BOOLEAN
			-- Should standard output for spawned process go to a
			-- pipe connected to parent, instead of to a
			-- file descriptor or named file?
		do
			Result := fd_set.item (Stdout_descriptor).is_pipe
		end

	error_piped: BOOLEAN
			-- Should standard error for spawned process go to a
			-- pipe connected to parent, instead of to a
			-- file descriptor or named file?
		do
			Result := fd_set.item (Stderr_descriptor).is_pipe
		end

	error_same_as_output: BOOLEAN
			-- Should standard error for spawned process be
			-- the same as standard output (i.e., identical
			-- file descriptor)?
		local
			fd: PROCESS_FD
		do
			fd := fd_set.item (Stderr_descriptor)
			Result := fd.is_duplicate and fd.dup_descriptor = Stdout_descriptor
		end

	process_descriptor (n: INTEGER): PROCESS_FD
			-- Process file descriptor `n' or Void if no
			-- disposition specified
		require
			valid_descriptor: valid_file_descriptor (n)
		do
			if n >= fd_set.lower and n <= fd_set.upper then
				Result := fd_set.item (n)
			end
		end

feature -- Execution properties

	is_executing: BOOLEAN
			-- Is process represented by `Current' currently
			-- executing?

	process_id: INTEGER
			-- Process id of last child process spawned or
			-- 0 if no processes have been spawned.

	in_child: BOOLEAN
			-- Are we currently executing in the spawned child
			-- process?  Intended only for use after
			-- a spawn call which raises an exception

	status: INTEGER
			-- Status from last child process that reported
			-- status (0 if none)

	is_success_exit: BOOLEAN
			-- Did last child process exit with a success status?
		do
			Result := is_exit_status (status) and then exit_code (status) = 0
		end

	output_from_child: LARGE_RAW_FILE
			-- File from which parent can read output coming
			-- from spawned process, if `output_piped' is true.
			-- If `error_same_as_output' is true, this output
			-- includes the child's standard error as well as
			-- standard output
		require
			child_output_piped: output_piped
		do
			Result := fd_set.item (Stdout_descriptor).pipe_file
		ensure
			result_exists: Result /= Void
		end

	input_to_child: LARGE_RAW_FILE
			-- File to which parent can write input going to
			-- spawned process, if `input_piped' is true
		require
			child_input_piped: input_piped
		do
			Result := fd_set.item (Stdin_descriptor).pipe_file
		ensure
			result_exists: Result /= Void
		end

	error_from_child: LARGE_RAW_FILE
			-- File from which parent can read error input coming
			-- from spawned process, if `error_piped' is true
		require
			child_error_piped: error_piped
		do
			Result := fd_set.item (Stderr_descriptor).pipe_file
		ensure
			result_exists: Result /= Void
		end

feature -- Modification

	set_program_file_name (fname: STRING)
			-- Set `program_file_name' to `fname'
		require
			file_name_exists: fname /= Void
			file_name_not_empty: not fname.is_empty
			process_not_executing: not is_executing
		do
			program_file_name := fname
		ensure
			file_name_set: program_file_name.is_equal (fname)
		end

	set_process_name (pname: STRING)
			-- Set `process_name' to `pname'
		require
			process_not_executing: not is_executing
		do
			process_name := pname
		ensure
			process_name_set: equal(process_name, pname)
		end

	set_arguments (args: ARRAY [STRING])
			-- Set `arguments' to `args'
		require
			process_not_executing: not is_executing
		do
			arguments := args
		ensure
			arguments_set: arguments = args
		end

	set_environment (env: HASH_TABLE [STRING, STRING])
			-- Set `environment' to `env'
		require
			process_not_executing: not is_executing
		do
			environment := env
		ensure
			environment_set: environment = env
		end

	set_close_nonstandard_files (b: BOOLEAN)
			-- Set `close_nonstandard_files' to `b'
		require
			process_not_executing: not is_executing
		do
			close_nonstandard_files := b
		ensure
			value_set: close_nonstandard_files = b
		end

	set_process_group_leader (b: BOOLEAN)
			-- Set `process_group_leader' to `b'
		require
			process_not_executing: not is_executing
		do
			process_group_leader := b
		ensure
			value_set: process_group_leader = b
		end

	set_input_descriptor (fd: INTEGER)
			-- Set `input_descriptor' to `fd', which must
			-- be readable
		require
			valid_descriptor: valid_file_descriptor (fd)
			process_not_executing: not is_executing
		local
			proc_fd: PROCESS_FD
		do
			create {FD_DESCRIPTOR} proc_fd.make (Stdin_descriptor, fd)
			fd_set.put (proc_fd, Stdin_descriptor)
		ensure
			descriptor_set: input_descriptor = fd
			input_not_file: input_file_name = Void
			input_not_piped: not input_piped
		end

	set_output_descriptor (fd: INTEGER)
			-- Set `output_descriptor' to `fd', which must
			-- be writable
		require
			valid_descriptor: valid_file_descriptor (fd)
			process_not_executing: not is_executing
		local
			proc_fd: PROCESS_FD
		do
			create {FD_DESCRIPTOR} proc_fd.make (Stdout_descriptor, fd)
			fd_set.put (proc_fd, Stdout_descriptor)
		ensure
			descriptor_set: output_descriptor = fd
			output_not_file: output_file_name = Void
			output_not_piped: not output_piped
		end

	set_error_descriptor (fd: INTEGER)
			-- Set `error_descriptor' to `fd', which must
			-- be writable
		require
			valid_descriptor: valid_file_descriptor (fd)
			process_not_executing: not is_executing
		local
			proc_fd: PROCESS_FD
		do
			create {FD_DESCRIPTOR} proc_fd.make (Stderr_descriptor, fd)
			fd_set.put (proc_fd, Stderr_descriptor)
		ensure
			descriptor_set: error_descriptor = fd
			error_not_file: error_file_name = Void
			error_not_piped: not error_piped
			error_not_output: not error_same_as_output
		end

	set_input_file_name (fname: STRING)
			-- Set `input_file_name' to `fname', which must
			-- be the name of an existing file readable by
			-- the parent process
		require
			process_not_executing: not is_executing
		local
			proc_fd: PROCESS_FD
		do
			-- FIXME: should this be binary or text
			-- Perhaps this routine should be marked obsolete
			if fname = Void then
				
				create {FD_UNCHANGED} proc_fd.make (Stdin_descriptor)
				fd_set.put (proc_fd, Stdin_descriptor)
			elseif fname.is_empty then
				create {FD_CLOSE} proc_fd.make (Stdin_descriptor)
				fd_set.put (proc_fd, Stdin_descriptor)
			else
				create {FD_FILE} proc_fd.make_read_binary (Stdin_descriptor, fname)
				fd_set.put (proc_fd, Stdin_descriptor)
			end
		ensure
			file_name_set: equal (input_file_name, fname)
			input_not_descriptor: not valid_file_descriptor (input_descriptor)
			input_not_piped: not input_piped
		end

	set_output_file_name (fname: STRING)
			-- Set `output_file_name' to `fname', which must
			-- be the name of a file writable by the parent
			-- process.  File is created if it does not exist
			-- and truncated if it does
		require
			process_not_executing: not is_executing
		local
			proc_fd: PROCESS_FD
		do
			-- FIXME: should this be binary or text
			-- Perhaps this routine should be marked obsolete
			if fname = Void then
				
				create {FD_UNCHANGED} proc_fd.make (Stdout_descriptor)
				fd_set.put (proc_fd, Stdout_descriptor)
			elseif fname.is_empty then
				create {FD_CLOSE} proc_fd.make (Stdout_descriptor)
				fd_set.put (proc_fd, Stdout_descriptor)
			else
				create {FD_FILE} proc_fd.make_write_binary (Stdout_descriptor, fname)
				fd_set.put (proc_fd, Stdout_descriptor)
			end
		ensure
			file_name_set: equal (output_file_name, fname)
			output_not_descriptor: not valid_file_descriptor (output_descriptor)
			output_not_piped: not output_piped
		end

	set_error_file_name (fname: STRING)
			-- Set `error_file_name' to `fname', which must
			-- be the name of a file writable by the parent
			-- process.  File is created if it does not exist
			-- and truncated if it does
		require
			process_not_executing: not is_executing
		local
			proc_fd: PROCESS_FD
		do
			-- FIXME: should this be binary or text
			-- Perhaps this routine should be marked obsolete
			if fname = Void then
				
				create {FD_UNCHANGED} proc_fd.make (Stderr_descriptor)
				fd_set.put (proc_fd, Stderr_descriptor)
			elseif fname.is_empty then
				create {FD_CLOSE} proc_fd.make (Stderr_descriptor)
				fd_set.put (proc_fd, Stderr_descriptor)
			else
				create {FD_FILE} proc_fd.make_write_binary (Stderr_descriptor, fname)
				fd_set.put (proc_fd, Stderr_descriptor)
			end
		ensure
			file_name_set: equal (error_file_name, fname)
			error_not_descriptor: not valid_file_descriptor (error_descriptor)
			error_not_piped: not error_piped
			error_not_output: not error_same_as_output
		end

	set_input_piped
			-- Set `input_piped' to true
		require
			process_not_executing: not is_executing
		local
			proc_fd: PROCESS_FD
		do
			create {FD_PIPE} proc_fd.make_input (Stdin_descriptor)
			fd_set.put (proc_fd, Stdin_descriptor)
		ensure
			piped_input_set: input_piped
			input_not_file: input_file_name = Void
			input_not_descriptor: not valid_file_descriptor (input_descriptor)
		end

	set_output_piped
			-- Set `output_piped' to true
		require
			process_not_executing: not is_executing
		local
			proc_fd: PROCESS_FD
		do
			create {FD_PIPE} proc_fd.make_output (Stdout_descriptor)
			fd_set.put (proc_fd, Stdout_descriptor)
		ensure
			piped_output_set: output_piped
			output_not_file: output_file_name = Void
			output_not_descriptor: not valid_file_descriptor (output_descriptor)
		end

	set_error_piped
			-- Set `error_piped' to true
		require
			process_not_executing: not is_executing
		local
			proc_fd: PROCESS_FD
		do
			create {FD_PIPE} proc_fd.make_output (Stderr_descriptor)
			fd_set.put (proc_fd, Stderr_descriptor)
		ensure
			piped_error_set: error_piped
			error_not_file: error_file_name = Void
			error_not_descriptor: not valid_file_descriptor (error_descriptor)
			error_not_output: not error_same_as_output
		end

	set_error_same_as_output
			-- Set `error_same_as_output' to true
		require
			process_not_executing: not is_executing
		local
			proc_fd: PROCESS_FD
		do
			create {FD_DUPLICATE} proc_fd.make (Stderr_descriptor, Stdout_descriptor)
			fd_set.put (proc_fd, Stderr_descriptor)
		ensure
			error_is_output: error_same_as_output
			error_not_piped: not error_piped
			error_not_file: error_file_name = Void
			error_not_descriptor: not valid_file_descriptor (error_descriptor)
		end

	set_descriptor (fd: INTEGER proc_fd: PROCESS_FD)
			-- Set disposition of file descriptor `fd' in
			-- spawned process to disposition indicated
			-- by `proc_fd'
		require
			process_not_executing: not is_executing
			valid_descriptor: valid_file_descriptor (fd)
			nonstandard_descriptor: fd > Stderr_descriptor
			process_desc_exists: proc_fd /= Void
			matching_descriptor: fd = proc_fd.number
		do
			fd_set.force (proc_fd, fd)
		ensure
			descriptor_set: process_descriptor (fd) = proc_fd
		end

feature -- Execution

	spawn_wait
			-- Spawn process and wait for it to report status
			-- before returning.  This routine may
			-- raise an exception in the parent and/or child
			-- process.  Feature `in_child' distinguishes whether
			-- we're executing in parent or child process
		require
			process_not_executing: not is_executing
			input_not_piped: not input_piped
			output_not_piped: not output_piped
			error_not_piped: not error_piped
		do
			spawn_nowait
			get_status_block
		ensure
			process_id_set: process_id /= 0
			process_not_executing: not is_executing
		end

	spawn_nowait
			-- Spawn process and do not wait for it to report
			-- status before returning.  This routine may
			-- raise an exception in the parent and/or child
			-- process.  Feature `in_child' distinguishes whether
			-- the child process got the exception
		require
			process_not_executing: not is_executing
		local
			gc: MEMORY
		do
			create gc
			build_argument_list
			open_files_and_pipes
			process_id := fork_process
			if process_id = 0 then	-- Child
				in_child := True
				gc.collection_off
				setup_child_process_files
				if process_group_leader then
					become_process_group_leader
				end
				if close_nonstandard_files then
					fd_set.mark_descriptors_close_on_exec
				end
				exec_process (program_file_name, arguments_for_exec, environment)
				-- Never returns.  Either exec works
				-- or an exception is raised
			else			-- Parent
				setup_parent_process_files
				arguments_for_exec := Void
			end
			is_executing := True
		ensure
			process_executing: is_executing
			process_id_set: process_id /= 0
			input_pipe_set: input_piped implies input_to_child /= Void
			output_pipe_set: output_piped implies output_from_child /= Void
			error_pipe_set: error_piped implies error_from_child /= Void
		end

	get_status_block
			-- Wait for executing child process to report status
			-- and put it in `status'.  If child has not
			-- reported status yet, block until it does
		require
			process_executing: is_executing
		do
			status := wait_for_process_block (process_id)
			is_executing := False
		ensure
			process_not_executing: not is_executing
		end

	get_status_noblock
			-- Wait for executing child process to report status
			-- and put it in `status'.  If child has not
			-- reported status yet, do not block.
			-- Caller must check `is_executing' to determine
			-- whether child reported status
		require
			process_executing: is_executing
		local
			done: BOOLEAN
			stat: INTEGER
		do
			stat := wait_for_process_noblock (process_id, $done)
			if done then
				status := stat
				is_executing := False
			end
		ensure
			execution_status_set:
				-- `is_executing' set to indicate whether
				-- child finished and reported status
		end

	send_signal (sig: INTEGER)
			-- Send signal number `sig' to child process
		require
			process_executing: is_executing
		do
			unix_send_signal (sig, process_id)
		end
	
	terminate_soft
			-- Send a software termination signal (SIGTERM)
			-- to child process
		require
			process_executing: is_executing
		do
			unix_terminate_soft (process_id)
		end
	
	terminate_hard
			-- Send a kill signal (SIGKILL) to child process
		require
			process_executing: is_executing
		do
			unix_terminate_hard (process_id)
		end

feature {NONE} -- Implementation

	build_argument_list
			-- Build argument list for `exec_process' call
			-- and put it in `arguments_for_exec'.
			-- Make `process_name' argument 0 and append
			-- `arguments' as the rest of the arguments
		local
			k, count, lower, pos: INTEGER
			pname: STRING
			a: ARRAY [STRING]
		do
			if arguments /= Void then
				count := arguments.count + 1
				lower := arguments.lower
			else
				count := 1
				lower := 1	-- Not applicable
			end
			create a.make (1, count)
			if process_name /= Void then
				pname := process_name
			else
				pname := program_file_name.mirrored
				pos := pname.index_of (Directory_separator, 1)
				if pos /= 0 then
					pname := pname.substring (1, pos - 1)
				end
				pname.mirror
			end
			a.put (pname, 1)
			from
				k := 2
			until
				k > count
			loop
				a.put (arguments.item (lower + k - 2), k)
				k := k + 1
			end
			arguments_for_exec := a
		end
	
	open_files_and_pipes
			-- Open any files that the child process will use
			-- for standard input, output or error.  Create
			-- any pipes that will be needed
		do
			fd_set.prepare_for_fork
		end
	
	setup_parent_process_files
			-- Setup files for parent after doing fork
		do
			fd_set.setup_for_parent
		end
	
	setup_child_process_files
			-- Setup standard input, output and error in
			-- child process after fork and before calling
			-- `exec_process'
		require
			in_child_process: in_child
		do
			fd_set.setup_for_child
		end
	
	fd_set: PROCESS_FD_SET
			-- File descriptors whose disposition in child
			-- process has been specified (only non-Void
			-- entries are meaningful)
	
	arguments_for_exec: ARRAY [STRING]
			-- Arguments to be passed to `exec_process'
	
invariant

	input_piped_no_desc: input_piped implies
		not valid_file_descriptor (input_descriptor)
	output_piped_no_desc: output_piped implies
		not valid_file_descriptor (output_descriptor)
	error_piped_no_desc: error_piped implies
		not valid_file_descriptor (error_descriptor)
	
	input_piped_no_file: input_piped implies input_file_name = Void
	output_piped_no_file: output_piped implies output_file_name = Void
	error_piped_no_file: error_piped implies error_file_name = Void
	
	input_named_no_desc: input_file_name /= Void implies
		not valid_file_descriptor (input_descriptor)
	output_named_no_desc: output_file_name /= Void implies
		not valid_file_descriptor (output_descriptor)
	error_named_no_desc: error_file_name /= Void implies
		not valid_file_descriptor (error_descriptor)
	
	input_named_no_pipe: input_file_name /= Void implies
		not input_piped
	output_named_no_pipe: output_file_name /= Void implies
		not output_piped
	error_named_no_pipe: error_file_name /= Void implies
		not error_piped
	
	input_desc_no_file: valid_file_descriptor (input_descriptor) implies
		input_file_name = Void
	output_desc_no_file: valid_file_descriptor (output_descriptor) implies
		output_file_name = Void
	error_desc_no_file: valid_file_descriptor (error_descriptor) implies
		error_file_name = Void
	
	input_desc_no_pipe: valid_file_descriptor (input_descriptor) implies
		not input_piped
	output_desc_no_pipe: valid_file_descriptor (output_descriptor) implies
		not output_piped
	error_desc_no_pipe: valid_file_descriptor (error_descriptor) implies
		not error_piped
	
	invalid_descriptor_invalid: not valid_file_descriptor (Invalid_file_descriptor)
	valid_stdin_descriptor: valid_file_descriptor (Stdin_descriptor)
	valid_stdout_descriptor: valid_file_descriptor (Stdout_descriptor)
	valid_stderr_descriptor: valid_file_descriptor (Stderr_descriptor)

	valid_desc_set_lower: fd_set.lower = 0

end
