
indexing
	
	description: "Object that is a graphical element of a view"
	author: "David Hollenberg"
	date: "July 25, 2006"

deferred class VIEW_GRAPHICAL_ELEMENT

feature -- Properties

	view: MODEL_VIEW
			-- View of which this is one element

	x: INTEGER
			-- X coordinate of upper left corner of bounding box.
			-- May not be the "true" value if very
			-- large coordinates are involved, but will work
			-- correctly for drawing the element

	y: INTEGER
			-- Y coordinate of upper left corner of bounding box.
			-- May not be the "true" value if very
			-- large coordinates are involved, but will work
			-- correctly for drawing the element

	height: INTEGER
			-- Size of minimum bounding box in Y dimension.
			-- May not be the "true" value if very
			-- large coordinates are involved, but will work
			-- correctly for drawing the element

	width: INTEGER
			-- Size of minimum bounding box in X dimension.
			-- May not be the "true" value if very
			-- large coordinates are involved, but will work
			-- correctly for drawing the element

	bounding_box: EV_RECTANGLE
			-- View bounding box for element, based on current
			-- scale and displayed area.
			-- May not be the "true" value if very
			-- large coordinates are involved, but will work
			-- correctly for drawing the element

	drawer: VIEW_ELEMENT_DRAWER
			-- Object that can draw a view element onto
			-- a drawable

	line_width: INTEGER
			-- Width of line to draw for border, in pixels
		deferred
		end

	label: STRING
			-- Current label to display
		deferred
		ensure
			result_not_void: Result /= Void
		end

	label_rotation: INTEGER
			-- Rotation of label in degrees clockwise
		deferred
		end

	is_border_drawn: BOOLEAN
			-- Should border be drawn for `Current'?
		do
			Result := True
		end

	border_color: EV_COLOR
			-- Color to use to display border
		deferred
		ensure
			result_not_void: Result /= Void
		end

	label_color: EV_COLOR
			-- Color to use to display label
		deferred
		ensure
			result_not_void: Result /= Void
		end

	is_interior_filled: BOOLEAN
			-- Should interior of element be filled when displayed?
		deferred
		end

	interior_color: EV_COLOR
			-- Color to use to fill interior, if filled or opaque
		deferred
		ensure
			result_not_void: Result /= Void
		end

	is_opaque: BOOLEAN
			-- Should element be displayed as opaque, so that
			-- it completely obscures the part of any element
			-- that is below it in stacking order?
		deferred
		end

	is_transparent: BOOLEAN
			-- Should element be displayed as transparent, so that
			-- we attempt to draw parts of elements that fall
			-- below it in stacking order so that they are visible?
		deferred
		end

	is_visible: BOOLEAN
			-- Is any part of element currently visible?
		local
			drw_w, drw_h: INTEGER
		do
			drw_w := view.width
			drw_h := view.height
			Result := x < drw_w and x + width > 0 and y < drw_h and y + height > 0
		end

	contains_point (a_coordinate: EV_COORDINATE): BOOLEAN
			-- Does element contain `a_coordinate'?
		require
			a_coordinate_not_void: a_coordinate /= Void
		do
			Result := bounding_box.has (a_coordinate)
		end

	intersects_area (a_rectangle: EV_RECTANGLE): BOOLEAN
			-- Does element intersect `a_rectangle'?
		require
			a_rectangle_not_void: a_rectangle /= Void
		do
			Result := bounding_box.intersects (a_rectangle)
		end

	enclosed_by_area (a_rectangle: EV_RECTANGLE): BOOLEAN
			-- Is element enclosed by `a_rectangle'?
		require
			a_rectangle_not_void: a_rectangle /= Void
		do
			Result := rectangle_contained_in (bounding_box, a_rectangle)
		end

feature -- Operations

	set_view_coordinates
			-- Set `width', `height', `x', `y' based on
			-- both size and location of model element
			-- on `view'
		deferred
		end

	unload_view_coordinates
			-- Set `width', `height', `x', `y' from `bounding_box'
		do
			x := bounding_box.x
			y := bounding_box.y
			width := bounding_box.width
			height := bounding_box.height
		end

	draw_without_fill (a_drawable: EV_DRAWABLE a_palette: VIEW_COLORS)
			-- Draw `Current' onto `a_drawable' using
			-- `a_palette', not including any needed interior fill
		require
			a_drawable_not_void: a_drawable /= Void
			a_palette_not_void: a_palette /= Void
			drawer_not_void: drawer /= Void
		do
			drawer.draw_without_fill (Current, a_drawable, a_palette)
		end

	draw_fill_only (a_drawable: EV_DRAWABLE a_palette: VIEW_COLORS)
			-- Draw only any needed interior fill for
			-- `Current' onto `a_drawable' using
			-- `a_palette'
		require
			a_drawable_not_void: a_drawable /= Void
			a_palette_not_void: a_palette /= Void
			drawer_not_void: drawer /= Void
		do
			drawer.draw_fill_only (Current, a_drawable, a_palette)
		end

	set_clip_area
			-- Set clip area for view to bounding box of
			-- element, if no clip region is already
			-- defined for view
		do
			view.set_clip_area (bounding_box)
		end

	remove_clip_area
			-- Remove clip area for view, if no clip
			-- region is already defined for view
		do
			view.remove_clip_area
		end

feature {NONE} -- Implementation

	rectangle_contained_in (r1, r2: EV_RECTANGLE): BOOLEAN
			-- Is `r1' contained in `r2' (i.e., is intersection
			-- of `r1' and `r2' equal to `r1')?
		require
			rectangle_1_not_void: r1 /= Void
			rectangle_2_not_void: r2 /= Void
		do
			Result := r2.has_x_y (r1.left, r1.top) and then
				r2.has_x_y (r1.left, r1.bottom) and then
				r2.has_x_y (r1.right, r1.top) and then
				r2.has_x_y (r1.right, r1.bottom)
		end

invariant
	drawer_not_void: drawer /= Void
	view_not_void: view /= Void
	bounding_box_not_void: bounding_box /= Void

end

