deferred class VIRTUAL_POINT
	-- a single point (units indeterminate) in the plane

inherit
	ANY
		redefine
			out
		end

	MORE_DOUBLE_MATH
	export {NONE} all
	redefine out
	end;

	GEOMETRY_RESOLUTION_ACCESS
	redefine out
	end;

	NUMBER_UTILITIES_ACCESS
	export {NONE} all
	redefine out
	end;

feature -- the point in Cartesian coordinates

	x: DOUBLE is
	deferred
	end;

	y: DOUBLE is
	deferred
	end;

feature -- the point in polar coordinates

	theta_radians: DOUBLE is
		-- angle between -pi and pi
	do  Result := angle_in_radians(x, y)
	end;

	theta_degrees: DOUBLE is
		-- angle between -180 and 180 degrees
		-- make multiples of 45 come out exact
	do  Result := angle_in_degrees(x, y)
	end;

	rho, abs: DOUBLE is
		-- radial distance to point
	do  Result := radius(x, y)
	ensure Result >= 0
	end;

feature -- vector arithmetic properties

	length_squared: DOUBLE is
		do
			Result := x*x + y*y
		end

	dot (other: VIRTUAL_POINT): DOUBLE is
			-- The dot product
		require
			other_not_void: other /= Void
		do
			Result := x*other.x + y*other.y
		end

feature -- point arithmetic

	infix "+"(other: VIRTUAL_POINT): POINT is
	do  create Result.make_xy(x + other.x, y + other.y)
	end;

	infix "-"(other: VIRTUAL_POINT): POINT is
	require other /= Void;
	do  create Result.make_xy(x - other.x, y - other.y)
	end;

	prefix "-": POINT is
	do  create Result.make_xy(-x, -y)
	end;

	infix "*"(mult: DOUBLE): POINT is
	do  create Result.make_xy(x*mult, y*mult)
	end;

	infix "/"(div: DOUBLE): POINT is
	require div /= 0
	do  create Result.make_xy(x/div, y/div)
	end;

	infix "|-|"(other: VIRTUAL_POINT): DOUBLE is
		-- distance from me to other
	require other /= Void;
	do  Result := radius(x - other.x, y - other.y)
	ensure Result >= 0
	end;

	max(other: VIRTUAL_POINT): POINT is
		-- point with larger x and larger y
		-- NOTE: Result need not be = one of the arguments!!
	require other /= Void;
	do  create Result.make_xy(x.max(other.x), y.max(other.y))
	end;

	min(other: VIRTUAL_POINT): POINT is
		-- point with larger x and larger y
		-- NOTE: Result need not be = one of the arguments!!
	require other /= Void;
	do  create Result.make_xy(x.min(other.x), y.min(other.y))
	end;

feature	-- point comparators: various three_way comparators

	ordered_by_angle(other: like Current): INTEGER is
	require other /= Void;
	do  Result := theta_radians.three_way_comparison
				(other.theta_radians)
	end;

	ordered_by_x(other: like Current): INTEGER is
	require other /= Void;
	do  Result := gres.sign(x - other.x)
	end;

	ordered_by_y(other: like Current): INTEGER is
	require other /= Void;
	do  Result := gres.sign(y - other.y)
	end;

feature	-- point transformation

	at(m: MATRIX): POINT is
		-- matrix really knows how to do it; we don't
	require m /= Void
	do  Result := m.transform_point(Current)
	end;

	snapped_to_grid(grid: DOUBLE): POINT is
		-- point nearest me whose x and y coordinates are both
		-- on the given grid (i.e., integral multiples of grid)
	require valid_grid: grid > 0
	do  create Result.make_xy(x, y);
	    Result.snap_to_grid(grid)
	end;

feature -- comparison

	is_same_point(other: VIRTUAL_POINT): BOOLEAN is
		-- is other the same point (w/in resolution) as me?
	do  if other /= Void then
		Result := gres.eq(x, other.x) and gres.eq(y, other.y)
	    end
	end;

feature -- display

	out: STRING is
	do  Result := out_with_sep(xy_separator)
	end;

	formatted(fmt: FORMAT_DOUBLE): STRING is
		-- like .out, but use the formatter
	require have_formatter: fmt /= Void
	do  Result := formatted_with_sep(fmt, xy_separator)
	end;

	out_with_sep(sep: STRING): STRING is
		-- like .out, but with a separator of your choice
	require have_separator: sep /= Void
	do  Result := string_of_3(x.out, sep, y.out)
	end;

	formatted_with_sep(fmt: FORMAT_DOUBLE; sep: STRING): STRING is
		-- yes, you can have it your way!
	require have_formatter: fmt /= Void; 
		have_separator: sep /= Void
	do  Result :=
		string_of_3(fmt.formatted(x), sep, fmt.formatted(y))
	end;

feature {NONE} -- implementation for formatting

	xy_separator: STRING is ","

	string_of_3(s1, s2, s3: STRING): STRING is
	do  create Result.make(s1.count + s2.count + s3.count);
	    Result.append(s1);
	    Result.append(s2);
	    Result.append(s3)
	end;

end -- VIRTUAL_POINT
