
indexing
	
	description: "Wafer file that describes wafer sites to be probed and %
		%inked (used by probing software)"
	author: "David Hollenberg"
	date: "August 30, 2004"

class WAFER_FILE
inherit
	ANY
	NUMBER_UTILITIES_ACCESS
		export
			{NONE} all
		end

feature -- Properties

	prober_id: STRING
			-- Id of prober for which file is intended

	is_main: BOOLEAN
			-- Is this the "main" wafer file for `prober_id'?
			-- If there are test vehicles with two different
			-- rotations, an auxiliary file will also be generated

	run_id: STRING
			-- Run id of run file describes

	run_type: STRING
			-- Type of run (scheduled, development, etc)

	rotation: INTEGER
			-- Rotation of test vehicles in wafer file, in
			-- degrees clockwise

	is_rotated: BOOLEAN
			-- Are test vehicles in wafer file rotated?
		do
			Result := rotation = 90 or rotation = 270
		end

	run_techn: STRING
			-- Run technology of run

	feature_size: DOUBLE
			-- Minimum feature size for dominant test vehicle

	process_vdd: DOUBLE
			-- Vdd for process

	wafer_vendor: STRING
			-- Name of wafer vendor

	mask_vendor: STRING
			-- Name of mask vendor

	wafer_techn: STRING
			-- Name of wafer technology

	mask_techn: STRING
			-- Name of mask technology

	mask_type: STRING
			-- Type of masks

	mask_size: CONST_POINT
			-- Size of masks in microns

	field_step_size: CONST_POINT
			-- Size of field step in microns when wafer is
			-- oriented correctly for probing (orientation
			-- depends on `rotation')

	wafer_diameter: DOUBLE
			-- Diameter of wafer in microns

	reference_pad_location: CONST_POINT
			-- Location of lower rightmost probe pad of
			-- reference site relative to center of field

	reference_site: PROBE_SITE
			-- Reference site

	probe_sites: LIST [PROBE_SITE]
			-- List of sites to be probed

	near_edge_sites: LIST [PROBE_SITE]
			-- List of sites that need to be inked because
			-- they are too near the edge of the wafer

	distinct_object_sites: LIST [PROBE_SITE]
			-- Subset of `probe_sites' with all distinct
			-- test object names appearing once
			-- (only object name and type are relevant)
		local
			table: HASH_TABLE [PROBE_SITE, STRING]
			site: PROBE_SITE
			sorter: LIST_BUBBLE_SORTER [PROBE_SITE]
			comparator: PROBE_SITE_COMPARATOR
		do
			create table.make (probe_sites.count)
			from
				probe_sites.start
			until
				probe_sites.after
			loop
				site := probe_sites.item
				table.force (site, site.object_name)
				probe_sites.forth
			end
			create sorter
			create comparator
			Result := sorter.sorted (table.linear_representation, comparator)
		end

feature -- Modification

	set_prober_id (s: STRING)
			-- Set `prober_id' to `s'
		require
			value_exists: s /= Void
		do
			prober_id := s
		ensure
			value_set: prober_id = s
		end

	set_main (b: BOOLEAN)
			-- Set `is_main' to `b'
		do
			is_main := b
		ensure
			value_set: is_main = b
		end

	set_run_id (s: STRING)
			-- Set `run_id' to `s'
		require
			value_exists: s /= Void
		do
			run_id := s
		ensure
			value_set: run_id = s
		end

	set_run_type (s: STRING)
			-- Set `run_type' to `s'
		require
			value_exists: s /= Void
		do
			run_type := s
		ensure
			value_set: run_type = s
		end

	set_rotation (n: INTEGER)
			-- Set `rotation' to `b'
		do
			rotation := n
		ensure
			value_set: rotation = n
		end

	set_run_techn (s: STRING)
			-- Set `run_techn' to `s'
		require
			value_exists: s /= Void
		do
			run_techn := s
		ensure
			value_set: run_techn = s
		end

	set_feature_size (d: DOUBLE)
			-- Set `feature_size' to `d'
		require
			value_nonnegative: d >= 0.0
		do
			feature_size := d
		ensure
			value_set: feature_size = d
		end

	set_process_vdd (d: DOUBLE)
			-- Set `process_vdd' to `d'
		require
			value_positive: d > 0.0
		do
			process_vdd := d
		ensure
			value_set: process_vdd = d
		end

	set_wafer_vendor (s: STRING)
			-- Set `wafer_vendor' to `s'
		require
			value_exists: s /= Void
		do
			wafer_vendor := s
		ensure
			value_set: wafer_vendor = s
		end

	set_mask_vendor (s: STRING)
			-- Set `mask_vendor' to `s'
		require
			value_exists: s /= Void
		do
			mask_vendor := s
		ensure
			value_set: mask_vendor = s
		end

	set_wafer_techn (s: STRING)
			-- Set `wafer_techn' to `s'
		require
			value_exists: s /= Void
		do
			wafer_techn := s
		ensure
			value_set: wafer_techn = s
		end

	set_mask_techn (s: STRING)
			-- Set `mask_techn' to `s'
		require
			value_exists: s /= Void
		do
			mask_techn := s
		ensure
			value_set: mask_techn = s
		end

	set_mask_type (s: STRING)
			-- Set `mask_type' to `s'
		require
			value_exists: s /= Void
		do
			mask_type := s
		ensure
			value_set: mask_type = s
		end

	set_mask_size (p: CONST_POINT)
			-- Set `mask_size' to `p'
		require
			value_exists: p /= Void
		do
			mask_size := p
		ensure
			value_set: mask_size = p
		end

	set_field_step_size (p: CONST_POINT)
			-- Set `field_step_size' to `p'
		require
			value_exists: p /= Void
		do
			field_step_size := p
		ensure
			value_set: field_step_size = p
		end

	set_wafer_diameter (d: DOUBLE)
			-- Set `wafer_diameter' to `d'
		require
			value_positive: d > 0.0
		do
			wafer_diameter := d
		ensure
			value_set: wafer_diameter = d
		end

	set_reference_pad_location (p: CONST_POINT)
			-- Set `reference_pad_location' to `p'
		require
			value_exists: p /= Void
		do
			reference_pad_location := p
		ensure
			value_set: reference_pad_location = p
		end

	set_reference_site (p: PROBE_SITE)
			-- Set `reference_site' to `p'
		require
			value_exists: p /= Void
		do
			reference_site := p
		ensure
			value_set: reference_site = p
		end

	set_probe_sites (list: LIST [PROBE_SITE])
			-- Set `probe_sites' to `list'
		require
			list_exists: list /= Void
		do
			probe_sites := list
		ensure
			value_set: probe_sites = list
		end

	set_near_edge_sites (list: LIST [PROBE_SITE])
			-- Set `near_edge_sites' to `list'
		require
			list_exists: list /= Void
		do
			near_edge_sites := list
		ensure
			value_set: near_edge_sites = list
		end

feature -- Operations

	write (f: LARGE_PLAIN_TEXT_FILE)
			-- Write `Current' to `f'
		require
			file_not_void: f /= Void
			file_open_write: f.is_open_write
		do
			file := f
			write_header
			write_body
		end

feature {NONE} -- Implementation

	file: LARGE_PLAIN_TEXT_FILE
			-- File to which wafer file should be written

	write_header
			-- Write wafer file header
		local
			size_x, size_y: DOUBLE
			virtual_mask_size: CONST_POINT
			feat_size: STRING
		do
			-- Compute and use `virtual_mask_size' for
			-- compatibility with probing software
			if mask_size.x < wafer_diameter then
				size_x := wafer_diameter
			else
				size_x := mask_size.x
			end
			if mask_size.y < wafer_diameter then
				size_y := wafer_diameter
			else
				size_y := mask_size.x
			end
			create virtual_mask_size.make_xy (size_x, size_y)
			
			file.put_string (run_id)
			file.new_line
			write_header_field ("Nickname", "No_nickname")
			write_header_field ("Run_type", run_type)
			write_header_field ("Service_name", "MOSIS")
				-- Service name required for compatibility 
				-- with probing software
			write_header_field ("Rotated", is_rotated.out.as_upper)
			write_header_field ("Technology", run_techn)
			feat_size := nu.std_formatted_number (feature_size)
			if feature_size = 0.0 then
				feat_size := feat_size + " (not calculated)"
			end
			write_header_field ("Featuresize", feat_size)
			write_header_field ("Process_Vdd", nu.std_formatted_number (process_vdd))
			write_header_field ("Fab", wafer_vendor)
			write_header_field ("Masks", mask_vendor)
			write_header_field ("Wafer_techn", wafer_techn)
			write_header_field ("Mask_techn", mask_techn)
			write_header_field ("MaskType", mask_type)
			write_header_field ("MaskSize", formatted_rounded_point (virtual_mask_size))
			write_header_field ("Field_step", formatted_point (field_step_size))
			write_header_field ("WaferDiameter", wafer_diameter.rounded.out)
			write_header_field ("Reference", formatted_rounded_point (reference_site.location))
			write_header_field ("Reference_offset", formatted_rounded_point (reference_pad_location))
			write_header_field ("Ref_Site", reference_site.description)
			file.put_character (' ')
			file.new_line
		end

	write_header_field (fname, val: STRING)
			-- Write field named `fname' with value `val' to
			-- wafer file header
		do
			file.put_string (fname)
			file.put_string (":  ")
			file.put_string (val)
			file.new_line
		end

	write_body
			-- Write body of wafer file
		do
			if probe_sites /= Void then
				probe_sites.do_all (agent write_body_line)
			end
			if near_edge_sites /= Void then
				near_edge_sites.do_all (agent write_body_line)
			end
		end

	write_body_line (p: PROBE_SITE)
			-- Write one line of body of wafer file for `p'
		do
			file.put_string (p.out)
			file.new_line
		end

	formatted_point (p: CONST_POINT): STRING
			-- String describing `p' suitable for inclusion
			-- in wafer file header
		do
			Result := nu.std_formatted_number (p.x) + Spacer + nu.std_formatted_number (p.y)
		ensure
			result_exists: Result /= Void
		end

	formatted_rounded_point (p: CONST_POINT): STRING
			-- String describing `p' rounded to integer X
			-- and Y which is suitable for inclusion in
			-- wafer file header
		do
			Result := p.x.rounded.out + Spacer + p.y.rounded.out
		ensure
			result_exists: Result /= Void
		end

	Spacer: STRING = "%T"

end
