
indexing
	
	description: "A map for a specific run which shows the whole wafer"
	author: "David Hollenberg"
	date: "January 21, 2005"

deferred class WAFER_MAP
inherit
	RUN_MAP
		redefine
			set_sizes,
			set_stroke_widths
		end
	DIE_LOCATIONS_MAP
		undefine
			set_sizes,
			set_stroke_widths
		end
	NUMBER_UTILITIES_ACCESS
		export
			{NONE} all
		end
	CONDITION_UTILITIES
		export
			{NONE} all
		end
	DOUBLE_MATH
		export
			{NONE} all
		end

feature -- Properties

	litho_sizes_row: LITHO_SIZES_ROW
			-- Row with lithography_sizes

	wafer_types_row: WAFER_TYPES_ROW
			-- Row with wafer type information

	field_step_patterns_row: FIELD_STEP_PATTERNs_ROW
			-- Row for field step pattern

	step_locations: LIST [FIELD_STEP_PATTERNXFIELD_LOCS_ROW]
			-- Locations where pattern is stepped on wafer,
			-- before any offset

	step_pattern_offset: CONST_POINT
			-- Amount by which step pattern should be "moved"
			-- for this lot

	wafer_radius: DOUBLE
			-- Wafer radius

	tweezer_radius: DOUBLE
			-- Wafer radius minus tweezer margin

	wafer_primary_flat_box: CONST_RECTANGLE
			-- Rectangle enclosing wafer with flat location
			-- taken into account

	tweezer_primary_flat_box: CONST_RECTANGLE
			-- Rectangle enclosing wafer with flat location
			-- taken into account, with tweezer margin subtracted

	is_preliminary: BOOLEAN
			-- Is map preliminary (not suitable for distribution)?
		do
			Result := field_step_patterns_row.is_preliminary
		end

feature -- Modification

	set_litho_sizes_row (r: like litho_sizes_row)
			-- Set `litho_sizes_row' to `r'
		require
			row_not_void: r /= Void
		do
			litho_sizes_row := r
		ensure
			row_set: litho_sizes_row = r
		end

	set_wafer_types_row (r: like wafer_types_row)
			-- Set `wafer_types_row' to `r'
		require
			row_not_void: r /= Void
		do
			wafer_types_row := r
		ensure
			row_set: wafer_types_row = r
		end

	set_field_step_patterns_row (r: like field_step_patterns_row)
			-- Set `field_step_patterns_row' to `r'
		require
			row_not_void: r /= Void
		do
			field_step_patterns_row := r
		ensure
			row_set: field_step_patterns_row = r
		end

	set_step_locations (list: like step_locations)
			-- Set `step_locations' to `list'
		require
			list_not_void: list /= Void
		do
			step_locations := list
		ensure
			locations_set: step_locations = list
		end

	set_step_pattern_offset (p: like step_pattern_offset)
			-- Set `step_pattern_offset' to `p'
		require
			offset_not_void: p /= Void
		do
			step_pattern_offset := p
		ensure
			offset_set: step_pattern_offset = p
		end

	set_wafer_radius (r: like wafer_radius)
			-- Set `wafer_radius' to `r'
		require
			radius_positive: r > 0.0
		do
			wafer_radius := r
		ensure
			radius_set: wafer_radius = r
		end

	set_tweezer_radius (r: like tweezer_radius)
			-- Set `tweezer_radius' to `r'
		require
			radius_positive: r > 0.0
		do
			tweezer_radius := r
		ensure
			radius_set: tweezer_radius = r
		end

	set_wafer_primary_flat_box (b: like wafer_primary_flat_box)
			-- Set `wafer_primary_flat_box' to `b'
		require
			box_not_void: b /= Void
		do
			wafer_primary_flat_box := b
		ensure
			flat_box_set: wafer_primary_flat_box = b
		end

	set_tweezer_primary_flat_box (b: like tweezer_primary_flat_box)
			-- Set `tweezer_primary_flat_box' to `b'
		require
			box_not_void: b /= Void
		do
			tweezer_primary_flat_box := b
		ensure
			flat_box_set: tweezer_primary_flat_box = b
		end

feature -- Operations

	generate_whole_document
			-- Generate entire document representing map
			-- and put in `last_document'
		do
			generate_page
			last_document.add_page (last_page)
		end

feature {NONE} -- Implementation

	generate_drawing_common_part
			-- Generate part of drawing that is common
			-- to a group of maps
		do
			add_invisible_bounding_box
			add_stepped_fields
			add_bad_die_shading
			add_wafer_outline
		end

	add_invisible_bounding_box
			-- Add invisible box to `last_drawing' to
			-- make bounding box come out right - might not work
			-- right for arcs
		local
			box: DRW_RECTANGLE
			diam: DOUBLE
		do
			diam := 2.0 * wafer_radius
			create box.make (diam, diam, 0, 0)
			box.disable_stroke
			last_drawing.extend (box)
		end

	add_stepped_fields
			-- Add stepped field locations to `last_drawing'
		local
			step_loc: FIELD_STEP_PATTERNXFIELD_LOCS_ROW
			fsize, psize: CONST_POINT
		do
			fsize := litho_sizes_row.field_size
			psize := litho_sizes_row.payload_size
			from
				step_locations.start
			until
				step_locations.after
			loop
				step_loc := step_locations.item
				add_stepped_field (fsize, psize, step_loc)
				step_locations.forth
			end
		end

	add_stepped_field (fsize, psize: CONST_POINT step_loc: FIELD_STEP_PATTERNXFIELD_LOCS_ROW)
			-- Add stepped field with field size `fsize',
			-- payload size `psize' at `step_loc' to
			-- `last_drawing'
		deferred
		end

	add_bad_die_shading
			-- Add shading for bad dies to `last_drawing'
		local
			step_loc: FIELD_STEP_PATTERNXFIELD_LOCS_ROW
			step_center: CONST_POINT
			fsize: CONST_POINT
		do
			create classifier
			classifier.set_wafer_radius (wafer_radius)
			classifier.set_tweezer_radius (tweezer_radius)
			classifier.set_wafer_primary_flat_box (wafer_primary_flat_box)
			classifier.set_tweezer_primary_flat_box (tweezer_primary_flat_box)
			fsize := litho_sizes_row.field_size
			from
				step_locations.start
			until
				step_locations.after
			loop
				step_loc := step_locations.item
				step_center := step_loc.center + step_pattern_offset
				if not classifier.good_die (fsize, step_center) then
					-- Field step not completely within tweezer margin.
					-- Check individual sites and shade ones that
					-- are bad
					add_bad_die_shading_for_field_step (step_loc.field_name, step_center, fsize)
				end
				step_locations.forth
			end
		end

	add_bad_die_shading_for_field_step (fname: STRING step_center, fsize: CONST_POINT)
			-- Add shading for bad dies to `last_drawing'
		local
			locs: LIST [DIE_LOCATIONS_EROW]
			loc: DIE_LOCATIONS_EROW
			size, die_center: CONST_POINT
			rects: LIST [CONST_RECTANGLE]
		do
			locs := die_locations.item (fname)
			if locs /= Void and then not locs.is_empty then
				create {ARRAYED_LIST [CONST_RECTANGLE]} rects.make (locs.count)
				from
					locs.start
				until
					locs.after
				loop
					loc := locs.item
					if loc.is_project then
						size := loc.printed_size_on_wafer
						die_center := step_center + create {CONST_POINT}.make_xy (loc.location_x, loc.location_y)
						if loc.included_in_payload_maps then
							rects.extend (create {CONST_RECTANGLE}.make_size_and_center (size, die_center))
							if classifier.bad_die (size, die_center) then
								add_bad_die_with_size_and_center (size, die_center)
							end
						end
					end
					locs.forth
				end
			end
		end

	add_bad_die_with_size_and_center (a_size, a_center: CONST_POINT)
		local
			box: DRW_RECTANGLE
		do
			create box.make (a_size.x, a_size.y, a_center.x, a_center.y)
			box.disable_stroke
			box.set_fill (partially_processed_or_edge_fill)
			last_drawing.extend (box)
		end

	add_wafer_outline
			-- Add outline of wafer edge and tweezer margin
			-- to `last_drawing'
		local
			ws: STRING
			shape: WAFER_SHAPE_VALUES
			flat_dist, half_flat_len: DOUBLE
			flat_indent: DOUBLE
			circle: DRW_CIRCLE
		do
			ws := wafer_types_row.wafer_shape
			flat_indent := wafer_types_row.non_null_max_primary_flat_indent
			flat_dist := wafer_radius - flat_indent
			half_flat_len := wafer_types_row.non_null_max_primary_flat_length / 2.0
			if ws.is_equal (shape.Flatted_circle) then
				
				add_flatted_wafer_outline (wafer_radius, flat_dist, half_flat_len, thick_width)
				flat_dist := tweezer_radius - flat_indent
				half_flat_len := sqrt (tweezer_radius * tweezer_radius - flat_dist * flat_dist)
				add_flatted_wafer_outline (tweezer_radius, flat_dist, half_flat_len, thin_width)
			elseif ws.is_equal (shape.Notched_circle) then
				add_notched_wafer_outline (wafer_radius, flat_dist, half_flat_len)
				create circle.make (tweezer_radius, 0.0, 0.0)
				circle.set_stroke_width (thin_width)
				last_drawing.extend (circle)
				add_notch_labels
			else
				handle_hard_error ("Unsupported wafer shape " + ws)
			end
		end

	add_flatted_wafer_outline (r, flat_dist, half_flat_len, stroke_w: DOUBLE)
			-- Add outline of flatted wafer with radius
			-- `r' to drawing.  `flat_dist' is distance of
			-- flat from center of wafer.  `stroke_w' is
			-- width of stroke to use
		require
			radius_positive: r > 0.0
			flat_dist_positive: flat_dist > 0.0
			half_flat_len_positive: half_flat_len > 0.0
			stroke_w_nonnegative: stroke_w >= 0.0
		local
			line: DRW_DRAWTO
			ls: like litho_sizes_row
			hfl, fd: DOUBLE
		do
			ls := litho_sizes_row
			hfl := half_flat_len
			fd := flat_dist
			create path.make (10)
			path.set_stroke_width (stroke_w)
			create line.make_moveto_abs (hfl, -fd)
			path.extend (line)
			add_arc (r, fd, -hfl)
			add_arc_or_line (ls.has_right_flat, r, fd, hfl)
			add_arc (r, hfl, fd)
			add_arc_or_line (ls.has_top_flat, r, -hfl, fd)
			add_arc (r, -fd, hfl)
			add_arc_or_line (ls.has_left_flat, r, -fd, -hfl)
			add_arc (r, -hfl, -fd)
			add_arc_or_line (ls.has_bottom_flat, r, hfl, -fd)
			last_drawing.extend (path)
		end

	add_notched_wafer_outline (r, flat_dist, half_flat_len: DOUBLE)
			-- Add outline of notched wafer with radius
			-- `r' to drawing.  `flat_dist' is distance of
			-- imaginary flat connecting start and end of notch
			-- from center of wafer
		require
			radius_positive: r > 0.0
			flat_dist_positive: flat_dist > 0.0
			half_flat_len_positive: half_flat_len > 0.0
		local
			line: DRW_DRAWTO
			ls: like litho_sizes_row
			hfl, fd, nd: DOUBLE
		do
			ls := litho_sizes_row
			hfl := half_flat_len
			fd := flat_dist
			nd := flat_dist - half_flat_len
				-- Distance of innermost notch point from
				-- wafer center
			create path.make (10)
			path.set_stroke_width (thick_width)
			create line.make_moveto_abs (hfl, -fd)
			path.extend (line)
			add_arc (r, fd, -hfl)
			add_arc_or_notch (ls.has_right_flat, r, nd, 0, fd, hfl)
			add_arc (r, hfl, fd)
			add_arc_or_notch (ls.has_top_flat, r, 0, nd, -hfl, fd)
			add_arc (r, -fd, hfl)
			add_arc_or_notch (ls.has_left_flat, r, -nd, 0, -fd, -hfl)
			add_arc (r, -hfl, -fd)
			add_arc_or_notch (ls.has_bottom_flat, r, 0, -nd, hfl, -fd)
			last_drawing.extend (path)
		end

	add_arc (r, x, y: DOUBLE)
			-- Add arc of radius `r' and absolute end point
			-- (`x', `y') to `path'
		local
			arc: DRW_ARC
		do
			create arc.make_abs (r, r, 0.0, False, True, x, y)
			path.extend (arc)
		end

	add_line (x, y: DOUBLE)
			-- Add line to absolute end point (`x', `y')
			-- to `path'
		local
			line: DRW_DRAWTO
		do
			create line.make_lineto_abs (x, y)
			path.extend (line)
		end

	add_notch (cx, cy, x, y: DOUBLE)
			-- Add curve to absolute end point (`x', `y')
			-- using control point (`cx', `cy')
		local
			curve: DRW_CURVE2
		do
			create curve.make_abs (cx, cy, x, y)
			path.extend (curve)
		end

	add_arc_or_line (is_flat: BOOLEAN r, x, y: DOUBLE)
			-- If `is_flat' then, add line to absolute end point
			-- (`x', `y') to `path'.  Otherwise, add arc of
			-- radius `r' and absolute end point
			-- (`x', `y') to `path'
		do
			if is_flat then
				add_line (x, y)
			else
				add_arc (r, x, y)
			end
		end

	add_arc_or_notch (is_notch: BOOLEAN r, cx, cy, x, y: DOUBLE)
			-- If `is_notch' then, add curve representing
			-- notch to absolute end point (`x', `y') to
			-- `path', using control point (`cx', `cy').
			-- Otherwise, add arc of radius `r'
			-- and absolute end point (`x', `y') to `path'.
			-- `hfl' is half flat length (half the distance
			-- of line segment from start of notch to end of notch)
		do
			if is_notch then
				add_notch (cx, cy, x, y)
			else
				add_arc (r, x, y)
			end
		end

	add_notch_labels
			-- Add label to drawing indicating position of each
			-- notch (should be only one)
		local
			ls: like litho_sizes_row
			b: DRW_BASIC_RECTANGLE
			max_hw, h, w, text_h: DOUBLE
		do
			ls := litho_sizes_row
			b := last_drawing.bounding_box
			w := b.width
			h := b.height
			max_hw := w.max (h)
			text_h := max_hw * label_fraction
			if ls.has_top_flat then
				add_notch_label (w, text_h, 0.0, b.upper_left.y + text_h, 0.0)
			end
			if ls.has_bottom_flat then
				add_notch_label (w, text_h, 0.0, b.lower_left.y - text_h, 0.0)
			end
			if ls.has_left_flat then
				add_notch_label (h, text_h, b.lower_left.x - text_h, 0.0, 270.0)
			end
			if ls.has_right_flat then
				add_notch_label (h, text_h, b.lower_right.x + text_h, 0.0, 90.0)
			end
		end

	add_notch_label (size_x, size_y, center_x, center_y, rot: DOUBLE)
			-- Add notch label to drawing in box which is
			-- `size_x' by `size_y' at (`center_x', `center_y'),
			-- rotated counterclockwise `rot' degrees
		local
			box: DRW_BASIC_RECTANGLE
			text: DRW_TEXT
			group: DRW_GROUP
			figure: DRW_FIGURE
			m: DRW_MATRIX
		do
			create box.make (size_x, size_y, center_x, center_y)
			text := text_in_box (Notch_label, box)
			if rot = 0.0 then
				figure := text
			else
				create group.make (1)
				group.extend (text)
				create m.make_rotation_degrees_xy (rot, box.x, box.y)
				group.set_transform (m)
				figure := group
			end
			last_drawing.extend (figure)
		end

	generate_whole_title
			-- Generate entire title and put in `last_title'
		local
			b, upper_box, left_box, right_box: DRW_BASIC_RECTANGLE
			box: DRW_RECTANGLE
			w, h, aux_h, center_x: DOUBLE
			text: DRW_TEXT
			t: STRING
		do
			-- Add invisible box that covers area for
			-- either PRELIMINARY label or (if not
			-- preliminary) explanation of Partial
			
			title_fraction := original_title_fraction
			w := available_paper_width
			h := title_height * (Title_scale_factor - 1.0)
			create box.make (w, h, w / 2.0, title_height / 2.0 + h / 2.0)
			box.disable_stroke
			last_title.extend (box)
			upper_box := box.bounding_box.twin
			
			if is_preliminary then
				upper_box.resize (upper_box.width, upper_box.height * 0.9)
				text := text_in_box (Preliminary_label, upper_box)
				text.set_stroke (preliminary_stroke)
				last_title.extend (text)
			else
				-- Final version - add explanation of partially
				-- processed fields and edge dies

				shrink_box (upper_box)
				left_box := horizontal_section (upper_box, 0.0, 0.1)
				right_box := horizontal_section (upper_box, 0.13, 1.0)
				right_box.resize (right_box.width, right_box.height * 0.9)
				create box.make_from_rectangle (left_box)
				box.set_stroke_width (field_stroke_width)
				box.set_fill (partially_processed_or_edge_fill)
				last_title.extend (box)
				text := text_in_box (Partial_or_edge_text, right_box)
				text.set_stroke (partial_stroke)
				last_title.extend (text)
			end
			
			-- Add title

			w := available_paper_width - (title_aux_width + title_step_info_width + title_logo_width)
			create box.make (w, title_height, w / 2.0 + title_logo_width, 0)
			box.set_stroke_width (thick_width)
			last_title.extend (box)
			
			b := box.bounding_box
			shrink_box (b)
			t := run_id + " " + title
			last_title.extend (text_in_box (t, b))

			-- Add box with wafer diameter

			center_x := w + title_logo_width + title_step_info_width / 2.0
			aux_h := title_height / 3.0
			create box.make (title_step_info_width, aux_h, center_x, aux_h)
			box.set_stroke_width (thick_width)
			last_title.extend (box)
			b := box.bounding_box.twin
			shrink_box (b)
			last_title.extend (text_in_box ("D = " + nu.std_formatted_number (2.0 * wafer_radius) + Microns_label, b))
			
			-- Add boxes with X and Y field step distances

			create box.make (title_step_info_width, aux_h, center_x, 0)
			box.set_stroke_width (thick_width)
			last_title.extend (box)
			b := box.bounding_box.twin
			shrink_box (b)
			last_title.extend (text_in_box ("X = " + nu.std_formatted_number (litho_sizes_row.field_size.x) + Microns_label, b))
			
			create box.make (title_step_info_width, aux_h, center_x, - aux_h)
			box.set_stroke_width (thick_width)
			last_title.extend (box)
			b := box.bounding_box.twin
			shrink_box (b)
			last_title.extend (text_in_box ("Y = " + nu.std_formatted_number (litho_sizes_row.field_size.y) + Microns_label, b))
			
			-- Add standard title information

			append_mosis_logo
			append_std_title_info
			set_new_title_fraction
		end

	current_page_label: STRING
			-- Descriptive label for current page of map
			-- (alternative to page number) or Void if none
		do
		end

	set_sizes
			-- Set various sizes
		do
			precursor
			label_fraction := 0.015
			title_step_info_width := 1.25
			original_title_fraction := title_fraction
			set_new_title_fraction
		end

	set_new_title_fraction
			-- Set new value of `title_fraction' to be used
		do
			title_fraction := title_fraction * Title_scale_factor
		end

	set_stroke_widths
			-- Set stroke values to use for various line
			-- thicknesses (relative numbers only - widths
			-- will be scaled later for each group)
		do
			precursor
			field_stroke_width := thick_width
		end

	set_stroke_and_fill_colors
			-- Set colors used for stroke and fill
		do
			create preliminary_stroke.make_with_rgb (1.0, 0.0, 0.0)
			create partial_stroke.make_with_rgb (1.0, 0.0, 0.0)
			create partially_processed_or_edge_fill.make_with_rgb (0.5, 1.0, 1.0)
		end

	approximate_drawing_width: DOUBLE
			-- Approximate width of drawing, in unspecified units.
			-- Used only to determine initial stroke width
			-- for various line thicknesses, so that stroke
			-- width will be a small fraction of drawing width
		do
			Result := 2.0 * wafer_radius
		end

	title_step_info_width: DOUBLE
			-- Width of box with information about field step
			-- size, in inches

	original_title_fraction: DOUBLE
			-- Value of title fraction in parent

	preliminary_stroke: DRW_COLOR
			-- Color for text labeling drawing as preliminary

	partial_stroke: DRW_COLOR
			-- Color for text explaining partially processed fields

	field_step_stroke: DRW_COLOR
			-- Color for outline of field steps

	partially_processed_or_edge_fill: DRW_COLOR
			-- Fill for interior of fields of view that are
			-- partially processed or that are edge dies
			-- (not completely inside tweezer margin)

	field_stroke_width: DOUBLE
			-- Stroke width to use for field-of-view bounding box

	path: DRW_PATH
			-- Path being created

	Notch_label: STRING = "Notch"
			-- Label indicating position of notch

	Preliminary_label: STRING = "PRELIMINARY - DO NOT DISTRIBUTE"
			-- Text identifying map as preliminary

	Partial_or_edge_text: STRING = "Edge dies --- do not use"
			-- Text explaining shading of edge die or partially 
			-- processed field

	Title_scale_factor: DOUBLE = 1.5
			-- Factor by which to scale title height

	classifier: WAFER_SITE_CLASSIFIER
			-- Object used to classify wafer sites as good, bad,
			-- marginal, etc

end
