
indexing
	
	description: "An object that can answer queries about wafer sites %
		%and points on the wafer"
	author: "David Hollenberg"
	date: "August 25, 2008"

class WAFER_SITE_CLASSIFIER

feature -- Queries

	reachable_die (size, center: CONST_POINT): BOOLEAN
			-- Is entire rectangle described by `size' and `center'
			-- reachable by prober (i..e, is it contained in
			-- `chuck_travel_box'?
		require
			probing_specific_parameters_set: probing_specific_parameters_set
		do
			Result := chuck_travel_box.encloses (create {CONST_RECTANGLE}.make_size_and_center (size, center))
		end

	safe_die (size, center: CONST_POINT): BOOLEAN
			-- Is entire rectangle described by `size' and
			-- `center' safe to move under probe pins
			-- without edge of wafer hitting an obstacle?
		require
			probing_specific_parameters_set: probing_specific_parameters_set
		local
			rect: CONST_RECTANGLE
			dist: DOUBLE
		do
			dist := obstacle_distance - wafer_radius
			create rect.make_size_and_center (size, center)
			Result := rect.ll.abs < dist and rect.lr.abs < dist and
rect.ul.abs < dist and rect.ur.abs < dist
		end

	marginal_die (size, center: CONST_POINT): BOOLEAN
			-- Is die described by `size' and `center' marginal
			-- (i.e., is it completely on the wafer but not
			-- entirely within the tweezer margin)?
		require
			basic_parameters_set: basic_parameters_set
		local
			rect: CONST_RECTANGLE
		do
			create rect.make_size_and_center (size, center)
			Result := on_wafer (rect.ll) and on_wafer (rect.lr) and
			   on_wafer (rect.ul) and on_wafer (rect.ur) and not
			  (in_tweezer (rect.ll) and in_tweezer (rect.lr) and
			   in_tweezer (rect.ul) and in_tweezer (rect.ur))
		end

	good_die (size, center: CONST_POINT): BOOLEAN
			-- Is die described by `size' and `center'
			-- good (i.e., is it completely within the
			-- tweezer margin)?
		require
			basic_parameters_set: basic_parameters_set
		local
			rect: CONST_RECTANGLE
		do
			create rect.make_size_and_center (size, center)
			Result := in_tweezer (rect.ll) and in_tweezer (rect.lr) and in_tweezer (rect.ul) and in_tweezer (rect.ur)
		end

	good_reference_die (size, center: CONST_POINT): BOOLEAN
			-- Is die described by `size' and `center'
			-- good for use as a reference die (i.e., is it 
			-- completely within the reference tweezer margin)?
		require
			basic_parameters_set: basic_parameters_set
		local
			rect: CONST_RECTANGLE
		do
			create rect.make_size_and_center (size, center)
			Result := in_reference_tweezer (rect.ll) and in_reference_tweezer (rect.lr) and in_reference_tweezer (rect.ul) and in_reference_tweezer (rect.ur)
		end

	bad_die (size, center: CONST_POINT): BOOLEAN
			-- Is die described by `size' and `center'
			-- bad (i.e., is it not completely within the
			-- tweezer margin)?
		require
			basic_parameters_set: basic_parameters_set
		do
			Result := not good_die (size, center)
		end

	on_wafer (p: CONST_POINT): BOOLEAN
			-- Is `p' completely on the wafer, taking flats
			-- into account?
		require
			basic_parameters_set: basic_parameters_set
		do
			Result := p.abs <= wafer_radius and wafer_primary_flat_box.contains (p)
		end

	in_tweezer (p: CONST_POINT): BOOLEAN
			-- Is `p' completely inside the tweezer margin,
			-- taking flats into account?
		require
			basic_parameters_set: basic_parameters_set
		do
			Result := p.abs <= tweezer_radius and tweezer_primary_flat_box.contains (p)
		end

	in_reference_tweezer (p: CONST_POINT): BOOLEAN
			-- Is `p' completely inside the reference
			-- tweezer margin, taking flats into account?
		require
			basic_parameters_set: basic_parameters_set
		do
			Result := p.abs <= reference_tweezer_radius and reference_tweezer_primary_flat_box.contains (p)
		end

	wafer_radius: DOUBLE
			-- Radius of wafer

	tweezer_radius: DOUBLE
			-- Radius of wafer minus tweezer margin

	reference_tweezer_radius: DOUBLE
			-- Radius of wafer minus tweezer margin and
			-- an extra margin for reference site

	wafer_primary_flat_box: CONST_RECTANGLE
			-- Rectangle enclosing wafer with flat location(s)
			-- taken into account

	tweezer_primary_flat_box: CONST_RECTANGLE
			-- Rectangle enclosing wafer with flat location(s)
			-- taken into account, with tweezer margin subtracted

	reference_tweezer_primary_flat_box: CONST_RECTANGLE
			-- Rectangle enclosing wafer with flat location(s)
			-- taken into account, with tweezer margin and
			-- an extra margin for reference site

	chuck_travel_box: CONST_RECTANGLE
			-- Box describing where prober chuck can be
			-- safely moved for current prober

	obstacle_distance: DOUBLE
			-- Distance in microns from center of probe pins
			-- to nearest obstacle wafer might hit in any
			-- direction, for current prober

feature -- Contract support

	basic_parameters_set: BOOLEAN
			-- Are basic parameters set?
		do
			Result := wafer_radius > 0.0 and tweezer_radius > 0.0 and wafer_primary_flat_box /= Void and tweezer_primary_flat_box /= Void
		end

	probing_specific_parameters_set: BOOLEAN
			-- Are parameters specific to probing set?
		do
			Result := reference_tweezer_radius > 0.0 and reference_tweezer_primary_flat_box /= Void and chuck_travel_box /= Void and obstacle_distance > 0
		end

	all_parameters_set: BOOLEAN
			-- Are all parameters set?
		do
			Result := basic_parameters_set and probing_specific_parameters_set
		end

feature -- Modification

	set_wafer_radius (r: like wafer_radius)
			-- Set `wafer_radius' to `r'
		do
			wafer_radius := r
		end

	set_tweezer_radius (r: like tweezer_radius)
			-- Set `tweezer_radius' to `r'
		do
			tweezer_radius := r
		end

	set_reference_tweezer_radius (r: like reference_tweezer_radius)
			-- Set `reference_tweezer_radius' to `r'
		do
			reference_tweezer_radius := r
		end

	set_wafer_primary_flat_box (b: like wafer_primary_flat_box )
			-- Set `wafer_primary_flat_box' to `b'
		do
			wafer_primary_flat_box := b
		end

	set_tweezer_primary_flat_box (b: like tweezer_primary_flat_box )
			-- Set `tweezer_primary_flat_box' to `b'
		do
			tweezer_primary_flat_box := b
		end

	set_reference_tweezer_primary_flat_box (b: like reference_tweezer_primary_flat_box )
			-- Set `reference_tweezer_primary_flat_box' to `b'
		do
			reference_tweezer_primary_flat_box := b
		end

	set_chuck_travel_box (b: like chuck_travel_box)
			-- Set `chuck_travel_box' to `b'
		do
			chuck_travel_box := b
		end


	set_obstacle_distance (d: like obstacle_distance)
			-- Set `obstacle_distance' to `d'
		do
			obstacle_distance := d
		end

end
