
indexing
	
	description: "Set of columns, where each column is composed of widgets"
	author: "David Hollenberg";
	date: "November 5, 2003"

class WIDGET_COLUMN_SET
inherit
	ANY
	WIDGET_ALIGNMENT
		export
			{NONE} all
		end

create
	make

feature {NONE} -- Creation

	make (n: INTEGER) is
		require
			number_nonnegative: n >= 0
		local
			k: INTEGER
		do
			count := n
			create {ARRAYED_LIST [WIDGET_COLUMN]} columns.make (n)
			from
				k := 1
			until
				k > n
			loop
				columns.extend (create {WIDGET_COLUMN}.make (Initial_column_length))
				k := k + 1
			end
		ensure
			count_set: count = n
		end

feature -- Properties

	count: INTEGER
			-- Number of columns in set (fixed at creation time)

	is_valid_index (n: INTEGER): BOOLEAN is
			-- Is `n' a valid index into column set?
		do
			Result := n >= 1 and n <= count
		end

	row_padding: INTEGER
			-- Padding between widgets in each row

	column_padding: INTEGER
			-- Padding between widgets in each column

	row_border_width: INTEGER
			-- Width of border around each row

	column_border_width: INTEGER
			-- Width of border around each column

	border_width: INTEGER
			-- Width of border around entire set of columns

	box: EV_HORIZONTAL_BOX is
			-- Box with widgets laid out as requested
		local
			vbox: EV_VERTICAL_BOX
			item: WIDGET_COLUMN
			w: EV_WIDGET
		do
			create Result
			Result.set_padding (row_padding)
			Result.set_border_width (border_width)
			from
				columns.start
			until
				columns.after
			loop
				item := columns.item
				vbox := item.box
				Result.extend (vbox)
				if not item.is_expanded then
					Result.disable_item_expand (vbox)
				end
				columns.forth
			end
			w := aligned_widget (Result)
			Result ?= w
			check
				result_not_void: Result /= Void
			end
		ensure
			result_exists: Result /= Void
		end

feature -- Modification

	append_row (list: ARRAY [EV_WIDGET]) is
			-- Append each widget in `list' to the end of 
			-- the column corresponding to its position
		require
			list_exists: list /= Void
			correct_count: list.count = count
		local
			k, n: INTEGER
		do
			from
				k := list.lower
			until
				k > list.upper
			loop
				n := k - (list.lower - 1)
				columns.i_th (n).extend (list.item (k))
				k := k + 1
			end
		end

	set_padding (n: INTEGER) is
			-- Set `row_padding' and `column_padding' to `n'
		require
			value_nonnegative: n >= 0
		do
			row_padding := n
			set_columns_padding (n)
		ensure
			row_padding_set: row_padding = n
			column_padding_set: column_padding = n
		end

	set_row_padding (n: INTEGER) is
			-- Set `row_padding' to `n'
		require
			value_nonnegative: n >= 0
		do
			row_padding := n
		ensure
			row_padding_set: row_padding = n
		end

	set_columns_padding (n: INTEGER) is
			-- Set `padding' for each column to `n'
		require
			value_nonnegative: n >= 0
		local
			k: INTEGER
		do
			column_padding := n
			from
				k := 1
			until
				k > count
			loop
				columns.i_th (k).set_padding (n)
				k := k + 1
			end
		end

	set_column_padding (n: INTEGER) is
			-- Set `column_padding' to `n'
		require
			value_nonnegative: n >= 0
		do
			set_columns_padding (n)
		ensure
			column_padding_set: column_padding = n
		end

	set_row_border_width (n: INTEGER) is
			-- Set `row_border_width' to `n'
		require
			value_nonnegative: n >= 0
		do
			row_border_width := n
		ensure
			row_border_width_set: row_border_width = n
		end

	set_columns_border_width (n: INTEGER) is
			-- Set `border_width' for each column to `n'
		require
			value_nonnegative: n >= 0
		local
			k: INTEGER
		do
			column_border_width := n
			from
				k := 1
			until
				k > count
			loop
				columns.i_th (k).set_border_width (n)
				k := k + 1
			end
		end

	set_column_border_width (n: INTEGER) is
			-- Set `column_border_width' to `n'
		require
			value_nonnegative: n >= 0
		do
			set_columns_border_width (n)
		ensure
			column_border_width_set: column_border_width = n
		end

	set_border_width (n: INTEGER) is
			-- Set `border_width' to `n'
		require
			value_nonnegative: n >= 0
		do
			border_width := n
		ensure
			border_width_set: border_width = n
		end

	align_column_none (n: INTEGER) is
			-- Disable alignment of widgets in column `n'
		require
			value_in_range: is_valid_index (n)
		do
			columns.i_th (n).align_none
		end

	align_column_left (n: INTEGER) is
			-- Left-align widgets in column `n'
		require
			value_in_range: is_valid_index (n)
		do
			columns.i_th (n).align_left
		end

	align_column_center (n: INTEGER) is
			-- Center-align widgets in column `n'
		require
			value_in_range: is_valid_index (n)
		do
			columns.i_th (n).align_center
		end

	align_column_right (n: INTEGER) is
			-- Right-align widgets in column `n'
		require
			value_in_range: is_valid_index (n)
		do
			columns.i_th (n).align_right
		end

	align_columns_none is
			-- Disable alignment for all columns in set
		local
			k: INTEGER
		do
			from
				k := 1
			until
				k > count
			loop
				align_column_none (k)
				k := k + 1
			end
		end

	align_columns_left is
			-- Left-align all columns in set
		local
			k: INTEGER
		do
			from
				k := 1
			until
				k > count
			loop
				align_column_left (k)
				k := k + 1
			end
		end

	align_columns_center is
			-- Center-align all columns in set
		local
			k: INTEGER
		do
			from
				k := 1
			until
				k > count
			loop
				align_column_center (k)
				k := k + 1
			end
		end

	align_columns_right is
			-- Right-align all columns in set
		local
			k: INTEGER
		do
			from
				k := 1
			until
				k > count
			loop
				align_column_right (k)
				k := k + 1
			end
		end

	align_none is
			-- Disable alignment of column set
		do
			alignment := No_alignment
		end

	align_left is
			-- Left-align column set
		do
			alignment := Ev_text_alignment_left
		end

	align_center is
			-- Center-align column set
		do
			alignment := Ev_text_alignment_center
		end

	align_right is
			-- Right-align column set
		do
			alignment := Ev_text_alignment_right
		end

	enable_expand is
			-- Make all items in all columns occupy available
			-- spare space
		local
			k: INTEGER
		do
			from
				k := 1
			until
				k > count
			loop
				enable_column_expand (k)
				k := k + 1
			end
		end

	disable_expand is
			-- Make all items in all columns not occupy available
			-- spare space
		local
			k: INTEGER
		do
			from
				k := 1
			until
				k > count
			loop
				disable_column_expand (k)
				k := k + 1
			end
		end

	enable_column_expand (n: INTEGER) is
			-- Make items in column `n' occupy available
			-- spare space
		require
			value_in_range: is_valid_index (n)
		do
			columns.i_th (n).enable_expand
			columns.i_th (n).align_none
		end

	disable_column_expand (n: INTEGER) is
			-- Make items in column `n' not occupy available
			-- spare space
		require
			value_in_range: is_valid_index (n)
		do
			columns.i_th (n).disable_expand
		end

feature {NONE} -- Implementation

	columns: LIST [WIDGET_COLUMN]
			-- Columns in set, in left-to-right order

	Initial_column_length: INTEGER is 5
			-- Initial number of widgets each column can hold

end
