
indexing
	
	description: "Portion of graphical interface with payload size, external %
		%kerf size and field size fields, in X and Y"
	author: "David Hollenberg"
	date: "September 30, 2008"

deferred class PAYLOAD_KERF_FIELD_INTERFACE

feature -- Properties

	payload_size_x_field: EV_TEXT_FIELD
			-- Field for setting/displaying payload size in X

	payload_size_y_field: EV_TEXT_FIELD
			-- Field for setting/displaying payload size in Y

	external_kerf_size_x_field: EV_TEXT_FIELD
			-- Field for setting/displaying external kerf
			-- size in X

	external_kerf_size_y_field: EV_TEXT_FIELD
			-- Field for setting/displaying external kerf
			-- size in Y

	field_size_x_field: EV_TEXT_FIELD
			-- Field for setting/displaying field size in X

	field_size_y_field: EV_TEXT_FIELD
			-- Field for setting/displaying field size in Y

	is_input_valid: BOOLEAN
			-- Is user input valid for `Current'?
		do
			Result := validation_errors = Void or else validation_errors.is_empty
		end

	validation_errors: STRING
			-- Errors detected while validating user input

feature {NONE} -- Implementation

	disable_field_editing
			-- Disable editing of payload, kerf and
			-- field size fields
		do
			payload_size_x_field.disable_edit
			payload_size_y_field.disable_edit
			external_kerf_size_x_field.disable_edit
			external_kerf_size_y_field.disable_edit
			field_size_x_field.disable_edit
			field_size_y_field.disable_edit
		end

	add_change_actions
			-- Add change actions to compute field size
			-- when payload or external kerf size changes
		local
			change_x_agent, change_y_agent: PROCEDURE [like Current, TUPLE]
		do
			change_x_agent := agent process_size_change (payload_size_x_field, external_kerf_size_x_field, field_size_x_field)
			change_y_agent := agent process_size_change (payload_size_y_field, external_kerf_size_y_field, field_size_y_field)
			payload_size_x_field.change_actions.extend (change_x_agent)
			external_kerf_size_x_field.change_actions.extend (change_x_agent)
			payload_size_y_field.change_actions.extend (change_y_agent)
			external_kerf_size_y_field.change_actions.extend (change_y_agent)
		end

	process_size_change (payload_fld, kerf_fld, field_fld: EV_TEXT_FIELD)
			-- Process change of `payload_fld' or `kerf_fld'
			-- and update `field_fld' appropriately
		local
			pay_val, kerf_val: STRING
			val: DOUBLE
		do
			pay_val := payload_fld.text
			kerf_val := kerf_fld.text
			if pay_val.is_double and kerf_val.is_double then
				val := pay_val.to_double + kerf_val.to_double
				set_text (field_fld, formatted_number (val))
			else
				set_text (field_fld, "")
			end
		end

	formatted_number (d: DOUBLE): STRING
			-- `d' as a nicely formatted string
		do
			Result := grid.formatted_number (d)
		end

	set_text (a_field: EV_TEXT_FIELD a_text: STRING)
		local
			l_editable: BOOLEAN
		do
			l_editable := a_field.is_editable
			if not l_editable then
				a_field.enable_edit
			end
			a_field.set_text (a_text)
			if not l_editable then
				a_field.disable_edit
			end
		end

	clear_validation_errors
			-- Make `validation_errors' empty
		do
			create validation_errors.make (0)
		end

	validate_payload_and_kerf_fields
			-- Validate payload size and external kerf size
			-- field.  Add any validation
			-- errors to `validation_errors'
		do
			validate_numeric_field (payload_size_x_field, initial_text (Payload_size_x_text))
			validate_numeric_field (payload_size_y_field, initial_text (Payload_size_y_text))
			validate_numeric_field (external_kerf_size_x_field, initial_text (External_kerf_size_x_text))
			validate_numeric_field (external_kerf_size_y_field, initial_text (External_kerf_size_y_text))
			if not has_validation_errors then
				validate_positive_field (payload_size_x_field, initial_text (Payload_size_x_text))
				validate_positive_field (payload_size_y_field, initial_text (Payload_size_y_text))
			end
		end

	validate_numeric_field (fld: EV_TEXT_FIELD fld_name: STRING)
			-- Verify that text in `fld' is a valid number
			-- and add validation error if not
		local
			str_val: STRING
			val, min_val, max_val: DOUBLE
		do
			str_val := fld.text
			if not str_val.is_double then
				add_validation_error (fld_name, "value is not a real number")
			else
				val := str_val.to_double
				min_val := 0.0
				max_val := 10 ^ ({WAFER_TECHNS_ROW}.Precision_top_border - grid.scale) - grid.size_resolution
				if val < min_val then
					add_validation_error (fld_name, "value is less than " + grid.formatted_number (min_val))
				elseif val > max_val then
					add_validation_error (fld_name, "value is greater than " + grid.formatted_number (max_val))
				elseif not grid.is_size_on_grid (val) then
					add_validation_error (fld_name, "value is not a multiple of grid resolution " + grid.size_resolution_text)
				end
			end
		end
	
	validate_positive_field (fld: EV_TEXT_FIELD fld_name: STRING)
			-- Verify that text in `fld' is a positive
			-- number and add validation error if not
		local
			val: DOUBLE
		do
			val := fld.text.to_double
			if val <= 0.0 then
				add_validation_error (fld_name, "value is not positive")
			end
		end
	
	add_validation_error (fld_name, msg: STRING)
		do
			validation_errors.append (fld_name + " " + msg + "%N")
		end

	has_validation_errors: BOOLEAN
			-- Were there any errors from last call to
			-- `validate_input'?
		do
			Result := not validation_errors.is_empty
		end

	initial_text (a_text: STRING): STRING
			-- `a_text' possibly prefixed with "Initial"
		deferred
		end

	Payload_size_x_text: STRING = "Payload Size X"

	Payload_size_y_text: STRING = "Payload Size Y"

	External_kerf_size_x_text: STRING = "External Kerf Size X"

	External_kerf_size_y_text: STRING = "External Kerf Size Y"

	Field_size_x_text: STRING = "Field Size X"

	Field_size_y_text: STRING = "Field Size Y"

	grid: GRID 
			-- Grid used for rounding and converting
			-- numbers
		deferred
		end

end
